Package: lares
Type: Package
Title: Lean Analytics and Robust Exploration Sidekick
Version: 5.4.0
Authors@R: c(
    person("Bernardo", "Lares", , "laresbernardo@gmail.com", c("aut", "cre")))
Maintainer: Bernardo Lares <laresbernardo@gmail.com>
Description: Auxiliary package for better/faster analytics, visualization, data mining, and machine
    learning tasks. With a wide variety of family functions, like Machine Learning, Data Wrangling,
    Marketing Mix Modeling (Robyn), Exploratory, API, and Scrapper, it helps the analyst or 
    data scientist to get quick and robust results, without the need of repetitive coding or 
    advanced R programming skills.
Depends: R (>= 3.5.0)
Imports: dplyr (>= 1.0.0), ggplot2 (>= 3.4.0), httr (>= 1.4.3),
        jsonlite (>= 1.8.0), lubridate (>= 1.7.9), openxlsx (>=
        3.5.10), patchwork (>= 1.3.0), pROC (>= 1.18.5), rlang (>=
        0.6.3), rpart (>= 4.1.0), rpart.plot (>= 3.1.0), rvest (>=
        1.0.0), stringr (>= 1.4.2), tidyr (>= 1.0.0), yaml (>= 2.2.4)
Suggests: beepr (>= 0.1.5), DALEX (>= 1.0.0), fansi (>= 1.0.0),
        googleAuthR (>= 1.0.0), googlesheets4 (>= 1.0.0), h2o (>=
        3.24.0), knitr (>= 1.31.0), MASS, quantmod (>= 0.4.20),
        rmarkdown (>= 2.7)
VignetteBuilder: knitr
URL: https://github.com/laresbernardo/lares,
        https://laresbernardo.github.io/lares/
BugReports: https://github.com/laresbernardo/lares/issues
RoxygenNote: 7.3.3
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-23 06:54:57 UTC; bernardo
Author: Bernardo Lares [aut, cre]
Repository: CRAN
Date/Publication: 2026-04-23 07:40:02 UTC
Built: R 4.5.2; ; 2026-04-24 13:30:03 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,corr_var)
S3method(plot,h2o_automl)
S3method(plot,h2o_shap)
S3method(plot,robyn_modelselector)
S3method(plot,stocks_hist)
S3method(plot,tree_var)
S3method(plot,x2y)
S3method(plot,x2y_preds)
S3method(print,h2o_automl)
S3method(print,maze_solve)
S3method(print,robyn_crossmmm)
S3method(print,wordle_check)
S3method(print,wordle_simulation)
export("%>%")
export(ROC)
export(are_binary)
export(are_constant)
export(are_id)
export(autoline)
export(balance_data)
export(bind_files)
export(bring_api)
export(cache_clear)
export(cache_exists)
export(cache_pipe)
export(cache_read)
export(cache_write)
export(cal_split)
export(categ_reducer)
export(check_attr)
export(check_opts)
export(chr2date)
export(chr2logical)
export(chr2num)
export(ci_lower)
export(ci_upper)
export(ci_var)
export(cleanNames)
export(cleanText)
export(clusterKmeans)
export(clusterOptimalK)
export(clusterVisualK)
export(conf_mat)
export(corr)
export(corr_cross)
export(corr_var)
export(cran_logs)
export(crosstab)
export(daily_portfolio)
export(daily_stocks)
export(dalex_local)
export(dalex_residuals)
export(dalex_variable)
export(date_cuts)
export(date_feats)
export(db_download)
export(db_upload)
export(df_str)
export(dist2d)
export(distr)
export(dont_sleep)
export(dont_sleep_time)
export(encrypt_file)
export(errors)
export(etf_sector)
export(export_plot)
export(export_results)
export(fb_accounts)
export(fb_ads)
export(fb_creatives)
export(fb_insights)
export(fb_process)
export(fb_report_check)
export(fb_rf)
export(fb_token)
export(file_name)
export(file_type)
export(filesGD)
export(files_functions)
export(font_exists)
export(forecast_arima)
export(formatColoured)
export(formatHTML)
export(formatNum)
export(freqs)
export(freqs_df)
export(freqs_list)
export(freqs_plot)
export(gain_lift)
export(gemini_ask)
export(gemini_image)
export(get_credentials)
export(get_creds)
export(get_currency)
export(get_tweets)
export(gg_colour_customs)
export(gg_fill_customs)
export(gg_text_customs)
export(gg_vals)
export(glued)
export(gpt_ask)
export(gpt_classify)
export(gpt_convert)
export(gpt_extract)
export(gpt_format)
export(gpt_history)
export(gpt_prompter)
export(gpt_table)
export(gpt_tag)
export(gpt_translate)
export(grepl_letters)
export(grepm)
export(gtrends_related)
export(gtrends_time)
export(h2o_automl)
export(h2o_explainer)
export(h2o_predict_API)
export(h2o_predict_MOJO)
export(h2o_predict_binary)
export(h2o_predict_model)
export(h2o_results)
export(h2o_selectmodel)
export(h2o_shap)
export(haveInternet)
export(hex_to_raw)
export(holidays)
export(image_metadata)
export(importxlsx)
export(impute)
export(install_recommended)
export(ip_data)
export(is_even)
export(is_ip)
export(is_odd)
export(is_url)
export(iter_seeds)
export(json2vector)
export(lares_logo)
export(lares_pal)
export(lasso_vars)
export(left)
export(list_cats)
export(listfiles)
export(loglossBinary)
export(mae)
export(mail_send)
export(mape)
export(markdown2df)
export(maze_gridsearch)
export(maze_solve)
export(missingness)
export(model_metrics)
export(model_preprocess)
export(move_files)
export(mp3_get)
export(mp3_trim)
export(mp3_update_tags)
export(mplot_conf)
export(mplot_cuts)
export(mplot_cuts_error)
export(mplot_density)
export(mplot_full)
export(mplot_gain)
export(mplot_importance)
export(mplot_lineal)
export(mplot_metrics)
export(mplot_response)
export(mplot_roc)
export(mplot_splits)
export(mplot_topcats)
export(mse)
export(msplit)
export(myip)
export(ngrams)
export(noPlot)
export(normalize)
export(num_abbr)
export(numericalonly)
export(ohe_commas)
export(ohse)
export(outlier_tukey)
export(outlier_turkey)
export(outlier_zscore)
export(outlier_zscore_plot)
export(plot_cats)
export(plot_chord)
export(plot_df)
export(plot_nums)
export(plot_palette)
export(plot_survey)
export(plot_timeline)
export(prophesize)
export(quants)
export(queryDB)
export(queryGA)
export(quiet)
export(raw_to_hex)
export(read.file)
export(readGS)
export(read_encrypted)
export(reduce_pca)
export(reduce_tsne)
export(remove_stopwords)
export(removenacols)
export(removenarows)
export(replaceall)
export(replacefactor)
export(right)
export(rmse)
export(robyn_hypsbuilder)
export(robyn_marginal)
export(robyn_modelselector)
export(robyn_performance)
export(robyn_xchannels)
export(robyn_xmodels)
export(rsq)
export(rsqa)
export(rtistry_sphere)
export(scale_x_abbr)
export(scale_x_comma)
export(scale_x_dollar)
export(scale_x_formatNum)
export(scale_x_percent)
export(scale_y_abbr)
export(scale_y_comma)
export(scale_y_dollar)
export(scale_y_formatNum)
export(scale_y_percent)
export(scrabble_dictionary)
export(scrabble_points)
export(scrabble_score)
export(scrabble_words)
export(sentimentBreakdown)
export(seq_surnames)
export(shap_var)
export(slackSend)
export(splot_change)
export(splot_divs)
export(splot_etf)
export(splot_growth)
export(splot_roi)
export(splot_summary)
export(splot_types)
export(spread_list)
export(statusbar)
export(stocks_file)
export(stocks_hist)
export(stocks_obj)
export(stocks_quote)
export(stocks_report)
export(sudoku_solver)
export(target_set)
export(textCloud)
export(textFeats)
export(textTokenizer)
export(theme_lares)
export(tic)
export(toc)
export(toon_reduction)
export(topics_rake)
export(tree_var)
export(try_require)
export(updateLares)
export(v2t)
export(vector2text)
export(warnifnot)
export(weighted_value)
export(what_size)
export(winsorize)
export(wordle_check)
export(wordle_dictionary)
export(wordle_simulation)
export(writeGS)
export(write_encrypted)
export(x2y)
export(x2y_metric)
export(x2y_preds)
export(year_month)
export(year_quarter)
export(year_week)
export(zerovar)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,any_of)
importFrom(dplyr,arrange)
importFrom(dplyr,as_tibble)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_match)
importFrom(dplyr,case_when)
importFrom(dplyr,contains)
importFrom(dplyr,count)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,distinct_at)
importFrom(dplyr,ensym)
importFrom(dplyr,everything)
importFrom(dplyr,filter)
importFrom(dplyr,first)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,lag)
importFrom(dplyr,last)
importFrom(dplyr,left_join)
importFrom(dplyr,matches)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_all)
importFrom(dplyr,mutate_at)
importFrom(dplyr,mutate_if)
importFrom(dplyr,n)
importFrom(dplyr,ntile)
importFrom(dplyr,one_of)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,rename_at)
importFrom(dplyr,right_join)
importFrom(dplyr,row_number)
importFrom(dplyr,rowwise)
importFrom(dplyr,sample_n)
importFrom(dplyr,select)
importFrom(dplyr,select_if)
importFrom(dplyr,slice)
importFrom(dplyr,slice_head)
importFrom(dplyr,starts_with)
importFrom(dplyr,summarise)
importFrom(dplyr,summarise_all)
importFrom(dplyr,summarise_if)
importFrom(dplyr,summarize)
importFrom(dplyr,summarize_all)
importFrom(dplyr,sym)
importFrom(dplyr,tally)
importFrom(dplyr,tibble)
importFrom(dplyr,transmute_if)
importFrom(dplyr,ungroup)
importFrom(dplyr,vars)
importFrom(ggplot2,aes)
importFrom(ggplot2,after_stat)
importFrom(ggplot2,annotate)
importFrom(ggplot2,coord_equal)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_line)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,enquo)
importFrom(ggplot2,enquos)
importFrom(ggplot2,expand_limits)
importFrom(ggplot2,facet_grid)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_area)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_boxplot)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_density)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_jitter)
importFrom(ggplot2,geom_label)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_path)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_polygon)
importFrom(ggplot2,geom_raster)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_smooth)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_tile)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggsave)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,guide_colorbar)
importFrom(ggplot2,guide_colourbar)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,last_plot)
importFrom(ggplot2,margin)
importFrom(ggplot2,position_dodge)
importFrom(ggplot2,position_fill)
importFrom(ggplot2,position_jitter)
importFrom(ggplot2,position_stack)
importFrom(ggplot2,quo_name)
importFrom(ggplot2,quos)
importFrom(ggplot2,rel)
importFrom(ggplot2,scale_color_brewer)
importFrom(ggplot2,scale_color_identity)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_colour_brewer)
importFrom(ggplot2,scale_colour_continuous)
importFrom(ggplot2,scale_colour_gradient)
importFrom(ggplot2,scale_colour_identity)
importFrom(ggplot2,scale_colour_manual)
importFrom(ggplot2,scale_fill_brewer)
importFrom(ggplot2,scale_fill_continuous)
importFrom(ggplot2,scale_fill_gradient)
importFrom(ggplot2,scale_fill_grey)
importFrom(ggplot2,scale_fill_identity)
importFrom(ggplot2,scale_fill_manual)
importFrom(ggplot2,scale_shape_identity)
importFrom(ggplot2,scale_size)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_x_date)
importFrom(ggplot2,scale_x_discrete)
importFrom(ggplot2,scale_x_reverse)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,sec_axis)
importFrom(ggplot2,stat_density_2d)
importFrom(ggplot2,stat_ecdf)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,unit)
importFrom(ggplot2,update_geom_defaults)
importFrom(ggplot2,xlab)
importFrom(ggplot2,xlim)
importFrom(ggplot2,ylab)
importFrom(ggplot2,ylim)
importFrom(grDevices,dev.off)
importFrom(grDevices,dev.size)
importFrom(grDevices,graphics.off)
importFrom(grDevices,png)
importFrom(graphics,box)
importFrom(graphics,grid)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,mtext)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,rasterImage)
importFrom(graphics,rect)
importFrom(httr,GET)
importFrom(httr,POST)
importFrom(httr,add_headers)
importFrom(httr,authenticate)
importFrom(httr,config)
importFrom(httr,content)
importFrom(httr,http_error)
importFrom(httr,oauth1.0_token)
importFrom(httr,oauth_app)
importFrom(httr,oauth_endpoint)
importFrom(httr,set_config)
importFrom(httr,stop_for_status)
importFrom(httr,upload_file)
importFrom(jsonlite,flatten)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(lubridate,"%m+%")
importFrom(lubridate,"%m-%")
importFrom(lubridate,ceiling_date)
importFrom(lubridate,date)
importFrom(lubridate,day)
importFrom(lubridate,days)
importFrom(lubridate,dmy)
importFrom(lubridate,dmy_hms)
importFrom(lubridate,floor_date)
importFrom(lubridate,hour)
importFrom(lubridate,minute)
importFrom(lubridate,month)
importFrom(lubridate,second)
importFrom(lubridate,wday)
importFrom(lubridate,week)
importFrom(lubridate,weeks)
importFrom(lubridate,year)
importFrom(lubridate,years)
importFrom(lubridate,ymd)
importFrom(lubridate,ymd_hms)
importFrom(openxlsx,addWorksheet)
importFrom(openxlsx,copyWorkbook)
importFrom(openxlsx,getSheetNames)
importFrom(openxlsx,loadWorkbook)
importFrom(openxlsx,read.xlsx)
importFrom(openxlsx,removeWorksheet)
importFrom(openxlsx,renameWorksheet)
importFrom(openxlsx,saveWorkbook)
importFrom(openxlsx,sheets)
importFrom(openxlsx,write.xlsx)
importFrom(pROC,ci)
importFrom(pROC,roc)
importFrom(patchwork,guide_area)
importFrom(patchwork,plot_annotation)
importFrom(patchwork,plot_layout)
importFrom(patchwork,wrap_plots)
importFrom(rlang,.data)
importFrom(rlang,as_label)
importFrom(rpart,rpart)
importFrom(rpart,rpart.control)
importFrom(rpart.plot,rpart.plot)
importFrom(rpart.plot,rpart.rules)
importFrom(rvest,html_attr)
importFrom(rvest,html_attrs)
importFrom(rvest,html_node)
importFrom(rvest,html_nodes)
importFrom(rvest,html_table)
importFrom(rvest,html_text)
importFrom(stats,as.formula)
importFrom(stats,complete.cases)
importFrom(stats,cor)
importFrom(stats,cor.test)
importFrom(stats,dist)
importFrom(stats,end)
importFrom(stats,kmeans)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,na.omit)
importFrom(stats,p.adjust)
importFrom(stats,prcomp)
importFrom(stats,predict)
importFrom(stats,pt)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,reorder)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,start)
importFrom(stats,var)
importFrom(stats,xtabs)
importFrom(stringr,fixed)
importFrom(stringr,str_count)
importFrom(stringr,str_length)
importFrom(stringr,str_pad)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_split)
importFrom(stringr,str_to_title)
importFrom(stringr,word)
importFrom(tidyr,gather)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,pivot_wider)
importFrom(tidyr,spread)
importFrom(tidyr,unite)
importFrom(utils,URLencode)
importFrom(utils,browseURL)
importFrom(utils,capture.output)
importFrom(utils,combn)
importFrom(utils,data)
importFrom(utils,download.file)
importFrom(utils,find)
importFrom(utils,flush.console)
importFrom(utils,getParseData)
importFrom(utils,globalVariables)
importFrom(utils,head)
importFrom(utils,install.packages)
importFrom(utils,installed.packages)
importFrom(utils,modifyList)
importFrom(utils,object.size)
importFrom(utils,packageDescription)
importFrom(utils,packageVersion)
importFrom(utils,read.csv)
importFrom(utils,read.delim)
importFrom(utils,read.table)
importFrom(utils,remove.packages)
importFrom(utils,tail)
importFrom(utils,type.convert)
importFrom(utils,write.csv)
importFrom(utils,write.table)
importFrom(yaml,read_yaml)

# lares <a href='https://laresbernardo.github.io/lares/index.html' target="_blank"><img src='man/figures/logo.png' align="right" height="139" /></a>

### Lean Analytics and Robust Exploration Sidekick
[![R-CMD-check](https://github.com/laresbernardo/lares/workflows/R-CMD-check/badge.svg)](https://github.com/laresbernardo/lares/actions?query=workflow%3AR-CMD-check) [![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version-ago/lares)](https://cran.r-project.org/package=lares) [![](https://cranlogs.r-pkg.org/badges/lares)](https://cran.r-project.org/package=lares) [![docs](https://github.com/laresbernardo/lares/workflows/documentation/badge.svg)](https://laresbernardo.github.io/lares/reference/index.html) [![CodeFactor](https://www.codefactor.io/repository/github/laresbernardo/lares/badge)](https://www.codefactor.io/repository/github/laresbernardo/lares/overview/main)

----

**lares** is an R package designed to automate, improve, and accelerate everyday **analytics and machine learning** tasks. It offers a wide variety of functions grouped in families for:

- **Machine Learning**: Streamlined model training and evaluation, including friendly AutoML pipelines.
- **Data Cleaning & Processing**: Functions to quickly prepare your data for modeling or analyzes.
- **Exploratory Data Analysis (EDA)**: Instantly visualize and summarize your data.
- **Reporting**: Easily generate comprehensive reports to share insights for MMM and ML models.
- **Visualization**: Out-of-the-box plotting for classification and regression models, timelines, and more.
- **API Integrations & Scrapers**: Simplify data collection from various sources.
- **Time Series & Portfolio Analysis**: Specialized utilities for financial and temporal data.
- **Credentials & Secrets Management**: Securely handle sensitive information in your analytics pipelines.
- **NLP & Text Analytics**: Tools to analyze and process text data.

> **Tip:** See all available functions and documentation [here](https://laresbernardo.github.io/lares/reference/index.html) or type `?lares::` in RStudio to explore interactively.

----

## Installation

```r
# CRAN VERSION
install.packages("lares")

# DEV VERSION (latest updates)
# If you don't have remotes installed yet, run: install.packages('remotes')
remotes::install_github("laresbernardo/lares")
# For a full installation with recommended dependencies:
remotes::install_github("laresbernardo/lares", dependencies = TRUE)
```

**Windows users:** You may need to install [RTools](https://cran.r-project.org/bin/windows/Rtools/) to build the dev version.

----

## Read about `lares` in action!

- **AutoML Quickstart:** [Introduction to AutoML using `lares`](https://laresbernardo.github.io/lares/articles/h2o_automl.html)
- **Model Results Visualization:** [Classification](https://www.r-bloggers.com/2018/07/machine-learning-results-in-r-one-plot-to-rule-them-all/) | [Regression](https://www.r-bloggers.com/2018/07/machine-learning-results-in-r-one-plot-to-rule-them-all-part-2-regression-models/)
- **Marketing Mix Model Selection:** [Select the right MMM candidate](https://medium.com/@laresbernardo/select-the-right-mmm-candidate-based-on-your-specific-criteria-and-business-knowledge-1f583c3cb97a)
- **Cross-Correlations:** [Find Insights with Ranked Cross-Correlations](https://www.r-bloggers.com/2019/08/find-insights-with-ranked-cross-correlations/)
- **Secure Credentials:** [Manage Secrets in R](https://www.r-bloggers.com/2020/08/how-to-manage-credentials-and-secrets-safely-in-r/)
- **Portfolio Analysis:** [Performance & Reporting](https://www.r-bloggers.com/2018/09/visualize-your-portfolios-performance-and-generate-a-nice-report-with-r/)
- **Fun games**: [Wordle](https://www.r-bloggers.com/2022/02/how-i-selected-my-starting-word-for-wordle-using-simulations-and-r/), [Scrabble](https://laresbernardo.github.io/lares/reference/scrabble.html), [Sudoku](https://laresbernardo.github.io/lares/reference/sudoku_solver.html), [Mazes](https://laresbernardo.github.io/lares/reference/maze_solve.html), etc.
- **More Examples:** [Read other posts](https://www.r-bloggers.com/author/bernardo-lares/)

----

## Popular Functions

- `h2o_automl()`, `mplot_full()` – Automated machine learning pipeline with optimal model selection and visualizations.
- `freqs()`, `distr()`, `corr_var()`, `corr_cross()` – Instantly summarize, visualize, and uncover relationships in your data.
- `ohse()` – Efficient and smart one-hot encoding for categorical variables.
- `cache_*` – Speed up workflows by caching expensive computations.
- `robyn_*` – Additional functions to support [Robyn](https://github.com/facebookexperimental/Robyn) inputs and outputs.
- `fb_*` – Interact with Meta's Marketing API
- `gpt_*` – Structured prompts builder and interact with OpenAI's API
- `read_encrypted()`, `write_encrypted()` – Interact with encrypted files to keep secrets safe
- ...and [many more!](https://laresbernardo.github.io/lares/reference/index.html)

![AutoML Map (lares)](man/figures/automl_map.png?raw=true)

----

## Getting Started & Help

- Browse all functions in the [online reference](https://laresbernardo.github.io/lares/reference/index.html).
- Use `?lares::function_name` in RStudio for detailed help on any function.
- Found a bug or have a feature request? [Open an issue](https://github.com/laresbernardo/lares/issues).
- For questions or suggestions, reach out to [laresbernardo](mailto:laresbernardo@gmail.com).

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  warning = FALSE,
  message = FALSE,
  eval = TRUE, # Show actual results like old vignette
  cache = FALSE, # Disabled for CRAN to avoid massive tarball sizes
  cache.lazy = FALSE
)

## ----eval=TRUE----------------------------------------------------------------
library(lares)
library(dplyr)

## -----------------------------------------------------------------------------
# Install h2o (run once)
# install.packages("h2o")
library(h2o)

# Initialize h2o quietly for vignette
Sys.unsetenv("http_proxy")
Sys.unsetenv("https_proxy")
h2o.init(nthreads = -1, max_mem_size = "2G", ip = "127.0.0.1")
h2o.no_progress() # Disable progress bars

## -----------------------------------------------------------------------------
data(dft)

# Train an AutoML model
# Binary classification
model <- h2o_automl(
  df = dft,
  y = "Survived",
  target = "TRUE",
  ignore = c("Ticket", "Cabin", "PassengerId"),
  max_models = 10,
  max_time = 120,
  impute = FALSE
)

# View results
print(model)

## -----------------------------------------------------------------------------
names(model)

## -----------------------------------------------------------------------------
# All metrics
model$metrics

# Specific metrics
model$metrics$AUC
model$metrics$Accuracy
model$metrics$Logloss

## -----------------------------------------------------------------------------
# Confusion matrix plot
mplot_conf(
  tag = model$scores_test$tag,
  score = model$scores_test$score,
  subtitle = sprintf("AUC: %.3f", model$metrics$metrics$AUC)
)

## -----------------------------------------------------------------------------
# ROC curve
mplot_roc(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)

## -----------------------------------------------------------------------------
# Gain and Lift charts for binary classification
mplot_gain(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)

## -----------------------------------------------------------------------------
# Variable importance dataframe
head(model$importance, 15)

# Plot top 15 important variables
top15 <- head(model$importance, 15)
mplot_importance(
  var = top15$variable,
  imp = top15$importance
)

## ----eval=FALSE---------------------------------------------------------------
# # Calculate SHAP values (computationally expensive)
# shap <- h2o_shap(model)
# 
# # Plot SHAP summary
# plot(shap)

## -----------------------------------------------------------------------------
model <- h2o_automl(
  df = dft,
  y = "Survived",
  # Ignore specific columns
  ignore = c("Ticket", "Cabin", "PassengerId"),
  # Use only specific algorithms (exclude_algos also available)
  include_algos = c("GBM", "DRF"), # Gradient Boosting & Random Forest
  # Data split
  split = 0.7,
  # Handle imbalanced data
  balance = TRUE,
  # Remove outliers (Z-score > 3)
  no_outliers = TRUE,
  # Impute missing values (requires mice package if TRUE)
  impute = FALSE,
  # Keep only unique training rows
  unique_train = TRUE,
  # Reproducible results
  seed = 123
)

## -----------------------------------------------------------------------------
model_multiclass <- h2o_automl(
  df = dft,
  y = "Pclass",
  ignore = c("Cabin", "PassengerId"),
  max_models = 10,
  max_time = 60
)

# Multi-class metrics
model_multiclass$metrics

# Confusion matrix for multi-class
mplot_conf(
  tag = model_multiclass$scores_test$tag,
  score = model_multiclass$scores_test$score
)

## -----------------------------------------------------------------------------
model_regression <- h2o_automl(
  df = dft,
  y = "Fare",
  ignore = c("Cabin", "PassengerId"),
  max_models = 10,
  exclude_algos = NULL
)

# Regression metrics
model_regression$metrics

## -----------------------------------------------------------------------------
# Create splits
splits <- msplit(dft, size = 0.8, seed = 123)
splits$train$split <- "train"
splits$test$split <- "test"

# Combine
df_split <- rbind(splits$train, splits$test)

# Train using split column
model <- h2o_automl(
  df = df_split,
  y = "Survived",
  train_test = "split",
  max_models = 5
)

## -----------------------------------------------------------------------------
# New data (same structure as training)
new_data <- dft[1:10, ]

# Predict
predictions <- h2o_predict_model(new_data, model$model)
head(predictions)

## -----------------------------------------------------------------------------
# Get probabilities
predictions <- h2o_predict_model(new_data, model$model)
head(predictions)

## -----------------------------------------------------------------------------
# Complete model evaluation plots
mplot_full(
  tag = model$scores_test$tag,
  score = model$scores_test$score,
  subtitle = model$model@algorithm
)

## -----------------------------------------------------------------------------
# Model performance over trees
mplot_metrics(model)

## -----------------------------------------------------------------------------
# Save model and plots
export_results(model, subdir = "models", thresh = 0.5)

## ----eval=FALSE---------------------------------------------------------------
# # Load model
# loaded_model <- readRDS("models/Titanic_Model/Titanic_Model.rds")
# 
# # Make predictions with MOJO (production-ready)
# predictions <- h2o_predict_MOJO(
#   model_path = "models/Titanic_Model",
#   df = dft[1:10, ]
# )

## -----------------------------------------------------------------------------
# Quick prototype
model <- h2o_automl(dft, "Survived", max_models = 3, max_time = 30)

## -----------------------------------------------------------------------------
# Refine based on results
model <- h2o_automl(
  dft, "Survived",
  max_models = 20,
  no_outliers = TRUE,
  balance = TRUE,
  ignore = c("PassengerId", "Name", "Ticket", "Cabin"),
  model_name = "Titanic_Model"
)

## -----------------------------------------------------------------------------
# Check multiple metrics
model$metrics

# Visual inspection
mplot_full(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)

# Variable importance
mplot_importance(
  var = model$importance$variable,
  imp = model$importance$importance
)

## -----------------------------------------------------------------------------
# Density plot
mplot_density(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)

## -----------------------------------------------------------------------------
# Save everything
export_results(model, subdir = "my_project", thresh = 0.5)

## ----eval=FALSE---------------------------------------------------------------
# # Manually initialize h2o with more memory
# h2o::h2o.init(max_mem_size = "8G", nthreads = -1)

## -----------------------------------------------------------------------------
# Remove all models
h2o::h2o.removeAll()

# Shutdown h2o
h2o::h2o.shutdown(prompt = FALSE)

## -----------------------------------------------------------------------------
# Open h2o's web interface
# Navigate to: http://localhost:54321/flow/index.html

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  warning = FALSE,
  message = FALSE,
  eval = FALSE # API examples require API keys
)

## ----eval=TRUE----------------------------------------------------------------
library(lares)

## -----------------------------------------------------------------------------
# # lares will prompt you to set the directory
# creds <- get_credentials("openai")
# 
# # Or specify manually
# creds <- get_credentials("openai", dir = "~/my_credentials")

## -----------------------------------------------------------------------------
# # Simple question
# response <- gpt_ask("Explain linear regression in 2 sentences")

## -----------------------------------------------------------------------------
# # Create a structured prompt
# prompt <- gpt_prompter(
#   instruction = "Classify these items",
#   input = c("Dog", "Rose", "Car", "Oak"),
#   context = c("Animal", "Plant", "Object"),
#   output = "table",
#   cols = c("Item", "Category")
# )
# 
# # Send to ChatGPT
# result <- gpt_ask(prompt$prompt)

## -----------------------------------------------------------------------------
# # Classify text into categories
# items <- c("Python tutorial", "R package", "Java course", "Statistics book")
# categories <- c("Programming", "Statistics", "Other")
# 
# result <- gpt_classify(items, categories)
# print(result)

## -----------------------------------------------------------------------------
# # Extract specific information
# texts <- c(
#   "My email is john@example.com",
#   "Call me at +1-555-0123",
#   "Visit us at 123 Main St, NYC"
# )
# 
# extractions <- c("email", "phone number", "city")
# 
# result <- gpt_extract(texts, extractions)
# print(result)

## -----------------------------------------------------------------------------
# # Translate text
# text <- rep("Hello, how are you?", 3)
# languages <- c("Spanish", "French", "German")
# 
# result <- gpt_translate(text, languages)
# print(result)

## -----------------------------------------------------------------------------
# # Standardize date formats
# dates <- c("March 15, 2024", "15/03/2024", "2024-03-15")
# format_spec <- "ISO 8601 (YYYY-MM-DD)"
# 
# result <- gpt_format(dates, format = format_spec)
# print(result)

## -----------------------------------------------------------------------------
# # Tag items with multiple labels
# items <- c("Machine learning tutorial", "Data visualization guide")
# tags <- c("AI", "Statistics", "Programming", "Visualization")
# 
# result <- gpt_tag(items, tags)
# print(result)

## -----------------------------------------------------------------------------
# # See all previous prompts and responses from this session
# history <- gpt_history()
# head(history)

## -----------------------------------------------------------------------------
# # Customize model and parameters
# response <- gpt_ask(
#   "Write a haiku about data science",
#   model = "gpt-4",
#   temperature = 0.9, # More creative (0-2)
#   num_retries = 3
# )

## -----------------------------------------------------------------------------
# # Text generation
# response <- gemini_ask(
#   "Explain the Central Limit Theorem",
#   api_key = get_credentials("gemini")$api_key
# )

## -----------------------------------------------------------------------------
# # Analyze an image
# response <- gemini_image(
#   prompt = "Describe this plot in detail",
#   image_path = "path/to/plot.png",
#   api_key = get_credentials("gemini")$api_key
# )

## ----eval=FALSE---------------------------------------------------------------
# # Fetch Apple stock data
# aapl <- stocks_hist(
#   symbols = "AAPL",
#   from = Sys.Date() - 90, # Last 90 days
#   to = Sys.Date()
# )
# 
# head(aapl, 3)

## ----eval=FALSE---------------------------------------------------------------
# # Compare multiple stocks
# tech_stocks <- stocks_hist(
#   symbols = c("AAPL", "GOOGL", "MSFT"),
#   from = "2024-01-01"
# )
# 
# head(tech_stocks, 3)

## ----eval=FALSE---------------------------------------------------------------
# # Real-time quotes
# quote <- stocks_quote("AAPL")
# print(quote)

## -----------------------------------------------------------------------------
# # Track a portfolio
# portfolio <- daily_portfolio(
#   symbols = c("AAPL", "GOOGL", "TSLA"),
#   shares = c(10, 5, 8),
#   from = "2024-01-01"
# )
# 
# # Visualize performance
# splot_summary(portfolio)

## -----------------------------------------------------------------------------
# # Growth over time
# splot_growth(tech_stocks, symbols = c("AAPL", "GOOGL"))
# 
# # Returns on investment
# splot_roi(tech_stocks, investment = 10000)
# 
# # Sector performance
# splot_types(tech_stocks)

## -----------------------------------------------------------------------------
# # Read a Google Sheet
# data <- readGS(
#   title = "My Spreadsheet",
#   sheet = "Sheet1",
#   creds = get_credentials("google")
# )

## -----------------------------------------------------------------------------
# # Write data to Google Sheets
# writeGS(
#   data = mtcars,
#   title = "My Data",
#   sheet = "Cars",
#   creds = get_credentials("google")
# )

## -----------------------------------------------------------------------------
# # Query a database
# query <- "SELECT * FROM users WHERE active = TRUE LIMIT 10"
# result <- queryDB(
#   query = query,
#   from = "my_database_creds"
# )

## -----------------------------------------------------------------------------
# # Get trend data
# trends <- gtrends_time(
#   keyword = "machine learning",
#   from = "2023-01-01",
#   to = "2024-01-01"
# )
# 
# head(trends)

## -----------------------------------------------------------------------------
# # Find related search terms
# related <- gtrends_related(
#   keyword = "data science",
#   categories = c("top", "rising")
# )

## ----eval=FALSE---------------------------------------------------------------
# # ❌ DON'T hardcode
# # api_key <- "sk-1234567890"
# 
# # ✅ DO use credential management
# api_key <- get_credentials("openai")$secret_key

## -----------------------------------------------------------------------------
# # Wrap in try-catch
# result <- tryCatch(
#   {
#     gpt_ask("Your question here")
#   },
#   error = function(e) {
#     message("API error: ", e$message)
#     NULL
#   }
# )

## -----------------------------------------------------------------------------
# # Add delays for bulk operations
# items <- c("item1", "item2", "item3")
# 
# results <- lapply(items, function(item) {
#   result <- gpt_ask(item)
#   Sys.sleep(1) # 1 second delay
#   result
# })

## ----eval=FALSE---------------------------------------------------------------
# # Cache expensive API calls
# result <- cache_pipe(
#   {
#     # This only runs once, then cached and loaded on subsequent calls
#     gpt_ask("Expensive query here")
#   },
#   base = "my_gpt_query"
# )

## -----------------------------------------------------------------------------
# # Use cheaper models for simple tasks
# simple_task <- gpt_ask(
#   "Summarize: The meeting is at 3pm",
#   model = "gpt-3.5-turbo" # Cheaper than gpt-4
# )
# 
# # Use expensive models only when needed
# complex_task <- gpt_ask(
#   "Analyze this complex dataset: ...",
#   model = "gpt-4"
# )

## ----eval=TRUE----------------------------------------------------------------
# In .Renviron file:
# LARES_GPT_MODEL=gpt-4
# LARES_GPT_URL=https://api.openai.com/v1/chat/completions
# LARES_GEMINI_API=https://generativelanguage.googleapis.com/v1beta/models/

# Check current settings
Sys.getenv(c("LARES_GPT_MODEL", "LARES_GEMINI_API"))

## -----------------------------------------------------------------------------
# # Verify credentials are loaded
# creds <- get_credentials("openai")
# if (is.null(creds$secret_key)) {
#   stop("OpenAI API key not found!")
# }

## ----eval=TRUE----------------------------------------------------------------
# Check internet connection
if (haveInternet()) {
  message("Connected to internet")
} else {
  stop("No internet connection")
}

## -----------------------------------------------------------------------------
# # 1. Load data
# data(dft)
# 
# # 2. Get AI summary of data structure
# prompt <- sprintf(
#   "Summarize this dataset structure: %d rows, columns: %s",
#   nrow(dft),
#   paste(colnames(dft), collapse = ", ")
# )
# summary <- gpt_ask(prompt)
# 
# # 3. Get stock data
# stocks <- stocks_hist("AAPL", from = Sys.Date() - 30)
# 
# # 4. Cache the analysis
# analysis <- cache_pipe(
#   {
#     gpt_ask(sprintf(
#       "Analyze this stock: Recent high: $%.2f, Low: $%.2f",
#       max(stocks$High, na.rm = TRUE),
#       min(stocks$Low, na.rm = TRUE)
#     ))
#   },
#   base = "aapl_analysis"
# )
# 
# print(analysis)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE,
    cache = TRUE
)
library(lares)
library(dplyr)
library(ggplot2)

## ----eval=TRUE----------------------------------------------------------------
# Define your input variables
paid_media_spends <- c("tv_spend", "digital_spend")
organic_vars <- c("newsletter")

# Generate hyperparameters
hyperparameters <- robyn_hypsbuilder(
    channels = c(paid_media_spends, organic_vars),
    media_type = "default",
    adstock = "weibull_pdf", # or "weibull_cdf", "weibull_pdf"
    date_type = "weekly",
    lagged = FALSE
)

# The output is a list ready to be passed to robyn_inputs()
print(hyperparameters)

## ----eval=FALSE---------------------------------------------------------------
# # Assuming you have a Robyn OutputCollect object
# # OutputCollect <- robyn_run(...)
# 
# # Select best models
# best_models <- robyn_modelselector(
#     InputCollect = InputCollect,
#     OutputCollect = OutputCollect,
#     metrics = c(
#         "rsq_train", "performance", "potential_improvement", "non_zeroes",
#         "incluster_models", "cluster_sd", "certainty", "baseline_dist"
#     ),
#     wt = c(2, 0.1, 0, 1, 0.1, 0, 1.5, 0)
# )
# 
# # View the results
# print(best_models)
# 
# # Plot the clusters
# plot(best_models)

## ----eval=FALSE---------------------------------------------------------------
# # Get performance for a specific model and date range
# performance <- robyn_performance(
#     InputCollect = InputCollect,
#     OutputCollect = OutputCollect,
#     select_model = "1_123_4",
#     # Calculate for specific date range
#     start_date = NULL, end_date = NULL,
#     non_promo = TRUE, # Add non-promo decomposition as well
#     marginals = TRUE, # Add mROAS or mCPA for spend levels
#     carryovers = FALSE # Add mean percentage of carryover
# )
# 
# # View performance table
# print(performance)
# 
# # Plot performance
# plot(performance)

## ----eval=FALSE---------------------------------------------------------------
# # Calculate marginal response
# marginal <- robyn_marginal(
#     InputCollect = InputCollect,
#     OutputCollect = OutputCollect,
#     select_model = "1_123_4",
#     metric_name = "tv_spend",
#     metric_value = 100000
# )

## ----eval=FALSE---------------------------------------------------------------
# # Load multiple models
# model_paths <- c("model_us.json", "model_uk.json", "model_de.json")
# models <- lapply(model_paths, Robyn::robyn_recreate)
# names(models) <- c("US", "UK", "DE")
# 
# # Allocate budget across models
# cross_allocation <- robyn_xmodels(
#     models = models,
#     total_budget = 5000000,
#     scenario = "max_response"
# )
# 
# # View results
# print(cross_allocation)

####################################################################
#' Model Metrics and Performance
#'
#' This function lets the user get a confusion matrix and accuracy, and
#' for for binary classification models: AUC, Precision, Sensitivity, and
#' Specificity, given the expected (tags) values and predicted values (scores).
#'
#' @family Machine Learning
#' @family Model metrics
#' @family Calculus
#' @inheritParams h2o_automl
#' @param tag Vector. Real known label
#' @param score Vector. Predicted value or model's result
#' @param multis Data.frame. Containing columns with each category score
#' (only used when more than 2 categories coexist)
#' @param abc Boolean. Arrange columns and rows alphabetically
#' when categorical values?
#' @param auto_n Add \code{n_} before digits when it's categorical and
#' not numerical, even though seems numerical?
#' @param thresh_cm Numeric. Value to splits the results for the
#' confusion matrix. Range of values: (0-1)
#' @param type Character. One of: "train", "test".
#' @param model_name Character. Model's name for reference.
#' @param subtitle Character. Subtitle for plots
#' @return List. Multiple performance metrics that vary depending on
#' the type of model (classification or regression). If \code{plot=TRUE},
#' multiple plots are also returned.
#' @examples
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr, head)
#'
#' # Metrics for Binomial Model
#' met1 <- model_metrics(dfr$class2$tag, dfr$class2$scores,
#'   model_name = "Titanic Survived Model",
#'   plots = FALSE
#' )
#' print(met1)
#'
#' # Metrics for Multi-Categorical Model
#' met2 <- model_metrics(dfr$class3$tag, dfr$class3$score,
#'   multis = subset(dfr$class3, select = -c(tag, score)),
#'   model_name = "Titanic Class Model",
#'   plots = FALSE
#' )
#' print(met2)
#'
#' # Metrics for Regression Model
#' met3 <- model_metrics(dfr$regr$tag, dfr$regr$score,
#'   model_name = "Titanic Fare Model",
#'   plots = FALSE
#' )
#' print(met3)
#' @export
model_metrics <- function(tag, score, multis = NA,
                          abc = TRUE,
                          thresh = 10,
                          auto_n = TRUE,
                          thresh_cm = 0.5,
                          target = "auto",
                          type = "test",
                          model_name = NA,
                          plots = TRUE,
                          quiet = FALSE,
                          subtitle = NA) {
  if (length(tag) != length(score)) {
    stop(sprintf("tag (%s) and score (%s) have different lengths", length(tag), length(score)))
  }

  metrics <- list()
  cats <- sort(unique(as.character(tag)))
  model_type <- ifelse(length(cats) <= thresh, "Classification", "Regression")

  # When seems numeric but is categorical
  if (model_type == "Classification" && sum(grepl("^[0-9]", cats)) > 0 && auto_n) {
    tag <- as.factor(as.character(ifelse(
      grepl("^[0-9]", tag), paste0("n_", tag), as.character(tag)
    )))
  }
  # When is regression should always be numerical
  if (model_type == "Regression") {
    tag <- as.numeric(tag)
  }

  if (model_type == "Classification") {
    dic <- c(
      "AUC: Area Under the Curve",
      "ACC: Accuracy",
      "PRC: Precision = Positive Predictive Value",
      "TPR: Sensitivity = Recall = Hit rate = True Positive Rate",
      "TNR: Specificity = Selectivity = True Negative Rate",
      "Logloss (Error): Logarithmic loss [Neutral classification: 0.69315]",
      "Gain: When best n deciles selected, what % of the real target observations are picked?",
      "Lift: When best n deciles selected, how much better than random is?"
    )
    metrics[["dictionary"]] <- dic

    tag <- as.character(tag)

    if (is.numeric(score)) {
      new <- data.frame(score = score) %>%
        mutate(new = ifelse(.data$score >= thresh_cm, cats[1], cats[2])) %>%
        .$new
    } else {
      new <- score
    }

    conf <- .square_table(tag, new)
    total <- sum(conf)
    trues <- sum(diag(conf))
    falses <- total - trues

    # For Binaries
    if (length(cats) == 2) {
      metrics[["confusion_matrix"]] <- conf
      tag <- factor(tag)
      score <- as.numeric(score)
      ROC <- pROC::roc(tag, score, ci = TRUE, quiet = TRUE)
      nums <- data.frame(
        AUC = ROC$auc,
        ACC = trues / total,
        PRC = conf[2, 2] / (conf[2, 2] + conf[1, 2]),
        TPR = conf[2, 2] / (conf[2, 2] + conf[2, 1]),
        TNR = conf[1, 1] / (conf[1, 1] + conf[1, 2])
      )
      metrics[["gain_lift"]] <- gain_lift(tag, score, target = target, quiet = quiet)
      metrics[["metrics"]] <- signif(nums, 5)
    } else {
      # For Multi-Categories
      tags <- sort(unique(tag))
      if (is.na(multis)[1]) {
        stop("You have to input a data.frame with each tag's probability into the multis parameter.")
      }
      if (!all(as.character(tags) %in% colnames(multis))) {
        stop(paste0(
          "Your multis data.frame colums should be ", vector2text(tags),
          " (not ", vector2text(colnames(multis)), ")"
        ))
      }
      if (!all(colnames(multis) %in% as.character(tags))) {
        multis <- multis[, colnames(multis) %in% as.character(tags)]
      }

      df <- data.frame(tag, score)
      metrics[["confusion_matrix"]] <- conf_mat(tag, score)
      AUCs <- t(ROC(tag, score, multis)$ci)[, 2]
      m <- data.frame(
        AUC = mean(AUCs[seq_along(cats)], na.rm = TRUE),
        ACC = trues / total
      )
      metrics[["metrics"]] <- signif(m, 5)
      nums <- NULL

      for (i in seq_along(cats)) {
        tagi <- ifelse(tag == cats[i], 1, 0)
        predi <- as.numeric(ifelse(score == cats[i], 1, 0))
        conf_mati <- .square_table(tagi, predi)
        if (nrow(data.frame(conf_mati)) == 4) {
          total <- sum(conf_mati)
          trues <- sum(diag(conf_mati))
          falses <- total - trues
          numsi <- data.frame(
            tag = cats[i],
            ACC = trues / total,
            PRC = conf_mati[2, 2] / (conf_mati[2, 2] + conf_mati[1, 2]),
            TPR = conf_mati[2, 2] / (conf_mati[2, 2] + conf_mati[2, 1]),
            TNR = conf_mati[1, 1] / (conf_mati[1, 1] + conf_mati[1, 2])
          )
          nums <- rbind(nums, numsi)
        }
      }
      nums$AUC <- AUCs[seq_len(nrow(nums))]
      nums <- left_join(freqs(select(df, .data$tag), .data$tag), nums, "tag") %>%
        select(.data$tag, .data$n, .data$p, .data$AUC, everything(), -.data$pcum)
      metrics[["metrics_tags"]] <- mutate_if(nums, is.numeric, list(~ signif(., 5)))
    }

    if (plots && type == "test") {
      plots <- list()
      # CUMULATIVE GAINS PLOT
      plots[["gains"]] <- mplot_gain(
        tag, score, multis,
        target = "auto", splits = 10, quiet = TRUE
      )
      # CUMULATIVE RESPONSE PLOT
      plots[["response"]] <- mplot_response(
        tag, score, multis,
        target = "auto", splits = 10, highlight = "auto", quiet = TRUE
      )
      # CONFUSION MATRIX PLOT
      plots[["conf_matrix"]] <- mplot_conf(
        tag, score, thresh_cm,
        abc = abc, subtitle = subtitle, model_name = model_name
      )
      # ROC CURVES PLOT
      plots[["ROC"]] <- invisible(mplot_roc(
        tag, score, multis,
        subtitle = subtitle, model_name = model_name
      ))
      # Bring them all!
      metrics[["plots"]] <- plots
    }
  }

  if (model_type == "Regression") {
    dic <- c(
      "RMSE: Root Mean Squared Error",
      "MAE: Mean Average Error",
      "MAPE: Mean Absolute Percentage Error",
      "MSE: Mean Squared Error",
      "RSQ: R Squared",
      "RSQA: Adjusted R Squared"
    )
    metrics[["dictionary"]] <- dic

    metrics[["metrics"]] <- errors(tag, score)
  }

  if (type == "train") {
    metrics$dictionary <- NULL
  }

  metrics
}


####################################################################
#' Confussion Matrix
#'
#' This function calculates a Confussion Matrix using crosstab for
#' 2 or more categories. You can either set the score and threshold
#' or the labels you wish to cross with.
#'
#' You may use \code{mplot_conf()} or set \code{plot=TRUE}.
#'
#' @family Machine Learning
#' @family Model metrics
#' @inheritParams model_metrics
#' @param sense Character. Inequation sense for threshold: <, <=, >=, >
#' @param diagonal Boolean. \code{FALSE} to convert diagonal numbers to
#' zeroes. Ideal to detect must confusing categories.
#' @param plot Boolean. Plot result? Uses \code{mplot_conf()}
#' @return data.frame. Result of counting \code{tag} and \code{score}'s tag
#' given a \code{thresh}old, similar to \code{base::table()}.
#' @examples
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr[c(1, 2)], head)
#'
#' # Results for Binomial Model
#' conf_mat(dfr$class2$tag, dfr$class2$scores)
#' conf_mat(dfr$class2$tag, dfr$class2$scores, thresh = 0.3)
#' conf_mat(dfr$class2$tag, dfr$class2$scores, sense = "<=")
#'
#' # Results for Multi-Categorical Model
#' conf_mat(dfr$class3$tag, dfr$class3$score)
#' @export
conf_mat <- function(tag, score, thresh = 0.5,
                     sense = ">=",
                     diagonal = TRUE,
                     plot = FALSE) {
  df <- data.frame(tag, score)
  if (!diagonal) df <- filter(df, .data$tag != .data$score)
  if (plot) {
    mplot_conf(df$tag, df$score, thresh = thresh)
  } else {
    # About tags
    labels <- df %>%
      group_by(.data$tag, .drop = FALSE) %>%
      tally(wt = NULL) %>%
      arrange(desc(.data$n)) %>%
      .$tag
    df <- df %>% mutate(tag = factor(.data$tag, levels = unique(.data$tag)))

    # About scores
    if (is.numeric(df$score) && length(unique(tag)) == 2) {
      check_opts(sense, c("<", "<=", ">=", ">"))
      s <- do.call(sense, list(df$score, thresh))
      df <- mutate(df, pred = ifelse(s, as.character(labels[1]), as.character(labels[2])))
    } else {
      df <- mutate(df, pred = .data$score)
    }

    # Confusion Matrix
    ret <- df %>%
      rename("Real" = .data$tag, "Pred" = .data$pred) %>%
      crosstab(.data$Real, .data$Pred, total = FALSE)
    myRows <- as.character(unlist(ret[, 1]))
    myCols <- colnames(ret[, -1])
    ret <- ret[, c("Real x Pred", myCols[order(match(myCols, myRows))])]
    ret[is.na(ret)] <- 0
    as_tibble(ret)
  }
}


####################################################################
#' Cumulative Gain, Lift and Response
#'
#' This function calculates cumulative gain, lift, and response
#' values for a predictive score of a specific target. You can use the
#' \code{mplot_gain()} function to create a plot.
#'
#' @family Machine Learning
#' @family Model metrics
#' @inheritParams model_metrics
#' @param target Value. Which is your target positive value? If
#' set to 'auto', the target with largest mean(score) will be
#' selected. Change the value to overwrite. Only used when binary
#' categorical model.
#' @param splits Integer. Number of percentiles to split the data
#' @param plot Boolean. Plot results? Uses \code{mplot_gain()}
#' @return data.frame when \code{plot=FALSE} or plot when \code{plot=TRUE}.
#' @examples
#' data(dfr) # Results for AutoML Predictions
#' head(dfr$class2)
#'
#' # Results for Binomial Model
#' gain_lift(dfr$class2$tag, dfr$class2$scores, target = "FALSE")
#' gain_lift(dfr$class2$tag, dfr$class2$scores, target = "TRUE", splits = 5)
#' @export
gain_lift <- function(tag, score, target = "auto", splits = 10,
                      plot = FALSE, quiet = FALSE) {
  if (splits <= 1) stop("You must set more than 1 split")
  aux <- target_set(tag, score, target, quiet)
  df <- aux$df
  which <- aux$which

  sc <- df %>%
    mutate(tag = ifelse(.data$tag == which, TRUE, FALSE)) %>%
    arrange(desc(.data$score)) %>%
    mutate(percentile = .bincode(
      .data$score, quantile(.data$score,
        probs = seq(0, 1, length = splits + 1),
        include.lowest = TRUE
      ),
      right = TRUE, include.lowest = TRUE
    )) %>%
    mutate(percentile = rev(factor(.data$percentile, 1:splits)))

  wizard <- sc %>%
    filter(.data$tag == TRUE) %>%
    mutate(percentile = sc$percentile[seq_along(sc$percentile[sc$tag == TRUE])]) %>%
    group_by(.data$percentile) %>%
    tally() %>%
    ungroup() %>%
    mutate(p = 100 * .data$n / sum(.data$n), pcum = cumsum(.data$p)) %>%
    select(.data$percentile, .data$pcum) %>%
    rename(optimal = .data$pcum)

  gains <- sc %>%
    group_by(.data$percentile) %>%
    summarise(total = n(), target = sum(.data$tag), score = 100 * min(.data$score)) %>%
    left_join(wizard, "percentile") %>%
    replace(is.na(.), 100) %>%
    ungroup() %>%
    mutate(
      gain = 100 * cumsum(.data$target) / sum(.data$target),
      random = 100 * cumsum(.data$total) / sum(.data$total),
      lift = 100 * (.data$gain / .data$random - 1),
      response = 100 * .data$target / sum(.data$target),
      value = which
    ) %>%
    select(
      .data$percentile, .data$value, .data$random, .data$target, .data$total,
      .data$gain, .data$optimal, .data$lift, .data$response, .data$score
    )

  if (plot) {
    plots <- list()
    plots[["gain"]] <- mplot_gain(
      tag, score,
      target = which, splits = splits
    )
    plots[["response"]] <- mplot_response(
      tag, score,
      target = which, splits = splits
    )
    plots
  } else {
    gains
  }
}


####################################################################
#' AUC and ROC Curves Data
#'
#' This function calculates ROC Curves and AUC values with 95\% confidence
#' range. It also works for multi-categorical models.
#'
#' @section Plot Results:
#' To plot results, use the \code{mplot_roc()} function.
#'
#' @family Machine Learning
#' @family Model metrics
#' @inheritParams model_metrics
#' @return List with ROC's results, area under the curve (AUC) and their CI.
#' @examples
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr[c(1, 2)], head)
#'
#' # ROC Data for Binomial Model
#' roc1 <- ROC(dfr$class2$tag, dfr$class2$scores)
#' lapply(roc1, head)
#'
#' # ROC Data for Multi-Categorical Model
#' roc2 <- ROC(dfr$class3$tag, dfr$class3$score,
#'   multis = subset(dfr$class3, select = -c(tag, score))
#' )
#' lapply(roc2, head)
#' @export
ROC <- function(tag, score, multis = NA) {
  if (length(tag) != length(score)) {
    message("The tag and score vectors should be the same length.")
    stop(message(paste("Currently, tag has", length(tag), "rows and score has", length(score))))
  }

  if (!is.numeric(score) && length(multis) == 1) {
    score <- as.numeric(score)
    warning("You should use the 'multis' parameter to add each category's score")
  }

  if (length(unique(tag)) <= 1) {
    tag[1] <- "dummy_label"
    warning("Only 1 unique label detected. Adding single noice observation.")
  }

  if (length(multis) == 1) {
    tag <- factor(tag)
    score <- as.numeric(score)
    roc <- pROC::roc(tag, score, ci = TRUE, quiet = TRUE)
    coords <- data.frame(
      fpr = rev(roc$specificities),
      tpr = rev(roc$sensitivities)
    ) %>%
      mutate(label = "2cats")
    ci <- data.frame(roc$ci, row.names = c("min", "AUC", "max"))
  } else {
    df <- bind_cols(tag = tag, score = score, multis)
    cols <- colnames(df)
    coords <- NULL
    rocs <- list()
    for (i in 1:(length(cols) - 2)) {
      which <- colnames(df)[2 + i]
      res <- unlist(df[, c(which)])
      label <- factor(unlist(ifelse(df[, 1] == which, which, "other_label")))
      if (length(unique(label)) <= 1) {
        label[1] <- "dummy_label"
        warning("Only 1 unique label detected for ", which, ". Adding single noice observation.")
      }
      roci <- pROC::roc(label, res, ci = TRUE, quiet = TRUE)
      rocs[[paste(cols[i + 2])]] <- roci
      iter <- data.frame(
        fpr = rev(roci$specificities),
        tpr = rev(roci$sensitivities),
        label = paste(round(100 * roci$auc, 2), which, sep = "% | ")
      )
      coords <- rbind(coords, iter)
    }
    ci <- data.frame(lapply(rocs, "[[", "ci")) %>% mutate(mean = rowMeans(.))
    row.names(ci) <- c("min", "AUC", "max")
  }
  ret <- list(ci = ci, roc = coords)
  if (!is.na(multis)[1]) {
    ret[["rocs"]] <- rocs
  }
  ret
}



####################################################################
#' Calculate Continuous Values Errors
#'
#' This function lets the user calculate all errors and R squared
#' simultaneously.
#'
#' @family Model metrics
#' @inheritParams model_metrics
#' @return data.frame or numeric values results for multiple error
#' metrics on continuous numerical vectors inputs.
#' @examples
#' data(dfr) # Results for AutoML Predictions
#' head(dfr$regr)
#' df <- errors(dfr$regr$tag, dfr$regr$score)
#' head(df)
#' @export
errors <- function(tag, score) {
  data.frame(
    rmse = rmse(tag, score),
    mae = mae(tag, score),
    mape = mape(tag, score),
    mse = mse(tag, score),
    rsq = rsq(tag, score),
    rsqa = rsqa(tag, score)
  )
}

####################################################################
#' Root Mean Squared Error (RMSE)
#'
#' This function lets the user calculate Root Mean Squared Error
#'
#' @export
#' @rdname errors
rmse <- function(tag, score) {
  error <- tag - score
  sqrt(mean(error^2))
}


####################################################################
#' Mean Absolute Error (MAE)
#'
#' This function lets the user calculate Mean Absolute Error
#'
#' @export
#' @rdname errors
mae <- function(tag, score) {
  error <- tag - score
  mean(abs(error))
}


####################################################################
#' Mean Squared Error (MSE)
#'
#' This function lets the user calculate Mean Squared Error
#'
#' @export
#' @rdname errors
mse <- function(tag, score) {
  error <- tag - score
  mean(error^2)
}


####################################################################
#' Mean Absolute Percentage Error (MAPE)
#'
#' This function lets the user calculate Mean Squared Error
#'
#' @export
#' @rdname errors
mape <- function(tag, score) {
  error <- (tag - score) / tag
  error <- error[!is.infinite(error)]
  tag <- tag[tag != 0]
  mean(abs(error / tag))
}


####################################################################
#' R Squared
#'
#' This function lets the user calculate R Squared
#'
#' @export
#' @rdname errors
rsq <- function(tag, score) {
  fit <- lm(score ~ tag)
  signif(summary(fit)$r.squared, 4)
}

####################################################################
#' Adjusted R Squared
#'
#' This function lets the user calculate Adjusted R Squared
#'
#' @export
#' @rdname errors
rsqa <- function(tag, score) {
  fit <- lm(score ~ tag)
  signif(summary(fit)$adj.r.squared, 4)
}


####################################################################
#' Logarithmic Loss Function for Binary Models
#'
#' This function calculates log loss/cross-entropy loss for binary
#' models. NOTE: when result is 0.69315, the classification is neutral;
#' it assigns equal probability to both classes.
#'
#' @family Model metrics
#' @inheritParams model_metrics
#' @param eps Numeric. Epsilon value
#' @export
loglossBinary <- function(tag, score, eps = 0.001) {
  if (length(unique(tag)) != 2) {
    stop("Your 'tag' vector is not binary!")
  }

  if (length(tag) != length(score)) {
    message("The tag and score vectors should be the same length.")
    stop(message(paste(
      "Currently, tag has", length(tag),
      "rows and score has", length(score)
    )))
  }

  if (!is.numeric(tag)) tag <- as.integer(tag) - 1

  score <- pmax(pmin(score, 1 - eps), eps)
  output <- -mean(tag * log(score) + (1 - tag) * log(1 - score))
  output
}

.square_table <- function(x, y) {
  x <- factor(x)
  y <- factor(y)
  commonLevels <- sort(unique(c(levels(x), levels(y))))
  Real <- factor(x, levels = commonLevels)
  Pred <- factor(y, levels = commonLevels)
  table(Real, Pred)
}

####################################################################
#' Wordle Game Validation
#'
#' Given and input and a word, validate each letter based on Wordle's
#' rules: correct letter in correct placement (green), correct letter in wrong
#' placement (yellow), letter is not present (red).
#'
#' @family Games
#' @param input Character. Word to validate (5-letters)
#' @param word Character. Word actually answer (5-letters).
#' @param dictionary Character vector. List of valid words. If set to NULL
#' then will use modified \code{scrabble_dictionary()} to fetch 5 letter words.
#' Use \code{lang_dic} param to set language.
#' @param lang_dic Character. Any of: "en", "es". Only used when \code{dictionary}
#' parameter is NULL. Requires internet connection the first time. Uses cache.
#' @param method Integer. 1 for \code{scrabble_dictionary()}, 3 for scrapping
#' the words taken straight from the game's source code.
#' @param print Boolean. Print validation results?
#' @return Invisible vector with results by letter.
#' @examples
#' word <- "ABBEY"
#' # Or pick a random one:
#' # word <- sample(wordle_dictionary("en"), 1)
#' wordle_check("OPENS", word)
#' wordle_check("BABES", word)
#' wordle_check("KEBAB", word, print = FALSE)
#' wordle_check("ABYSS", word)
#' wordle_check("ABBEY", word)
#' # Feel free to use scrabble_words() for hints
#' @export
#' @rdname wordle
wordle_check <- function(input, word, dictionary = NULL, lang_dic = "en", method = 3, print = TRUE) {
  if (!haveInternet()) {
    message("No internet connetion...")
    invisible(NULL)
  } else {
    wordle_valid(input, dictionary, lang_dic, method)
    out <- pos_check(input, word, len = 5, print = print)
    invisible(out)
  }
}

pos_check <- function(input, solution, len = 5, print = TRUE) {
  tiles <- toupper(unlist(strsplit(solution, "")))
  in_tiles <- toupper(unlist(strsplit(input, "")))
  init <- rep("_", len)
  init[which(!in_tiles %in% tiles)] <- "RED"
  init[which(in_tiles %in% tiles)] <- "YELLOW"
  init[which(in_tiles == tiles)] <- "GREEN"
  out <- init
  names(out) <- in_tiles
  class(out) <- c("wordle_check", class(out))
  print(out, print = print)
  out
}

#' @rdname wordle
#' @param x Object to print
#' @export
print.wordle_check <- function(x, print = TRUE, ...) {
  texts <- NULL
  for (i in seq_along(x)) {
    texts <- c(texts, formatColoured(names(x)[i], x[i], bold = TRUE, cat = FALSE))
  }
  txt <- paste(texts, collapse = " ")
  if (print) cat(txt) else txt
}

wordle_valid <- function(input, dictionary, lang_dic = "en", method = 3) {
  if (is.null(dictionary)) {
    dictionary <- wordle_dictionary(lang_dic, method, quiet = TRUE)
  }
  if (!toupper(input) %in% toupper(dictionary)) {
    stop(input, " is not a valid input word")
  }
}

#' @inheritParams scrabble_dictionary
#' @export
#' @rdname wordle
wordle_dictionary <- function(lang_dic = "en", method = 3, quiet = TRUE) {
  if (is.null(lang_dic)) {
    NULL
  } else {
    check_opts(lang_dic, c("en", "es"))
    cache_name <- c(lang_dic, method)
    if (cache_exists(cache_name)) {
      words <- cache_read(cache_name, quiet = quiet)
      if (!quiet) {
        message(sprintf(
          ">>> Loaded %s '%s' words",
          formatNum(length(words), 0), lang_dic
        ))
      }
      words
    } else {
      if (method == 1) {
        words <- scrabble_dictionary(lang_dic, quiet)[[1]]
      }
      if (method == 3) {
        url <- "https://raw.githubusercontent.com/tabatkins/wordle-list/main/words"
        words <- readLines(url, warn = FALSE)
      }
      out <- toupper(words[nchar(words) == 5])
      cache_write(out, cache_name, quiet = quiet)
      if (!quiet) message(sprintf(">>> Saved (%s words) into cache", formatNum(length(out), 0)))
      out
    }
  }
}

#' @inheritParams cache_write
#' @inheritParams wordle_check
#' @param seed Numeric. For reproducibility. Accepts more than one: will
#' run as many seeds there are.
#' @export
#' @examples
#' \donttest{
#' x <- wordle_simulation(input = "SAINT", word = "ABBEY", seed = 1:3)
#' print(x)
#' }
#' @rdname wordle
wordle_simulation <- function(input, word, seed = NULL, quiet = FALSE, ...) {
  if (!haveInternet()) {
    message("No internet connetion...")
    invisible(NULL)
  } else {
    tic("wordle_simulation")
    output <- NULL
    if (is.null(seed)) seed <- sample(1:100, 1)
    # First iteration with picked word (no random stuff happens here)
    iter_first <- wordle_opts(input, word, quiet = quiet, ...)
    for (s in seed) {
      # Print first word for nicer and more informative output
      if (s != seed[1]) {
        wordle_check(input, word, ...)
        if (!isFALSE(list(...)[["print"]])) cat("\n")
      }
      set.seed(s) # s = seed[1]
      seed_loop <- NULL
      i <- 1
      used_words <- input
      iter <- iter_first
      # Second iteration onwards
      while (length(iter) > 1) {
        random_word <- sample(iter, 1)
        used_words <- c(used_words, random_word)
        iter <- wordle_opts(random_word, word, iter, quiet = quiet, ...)
        # If random word picked is the word
        if (random_word == word) break
        # If last word remaining is the one
        if (length(iter) == 1 && all(iter == word)) {
          iter <- wordle_opts(word, word, iter, quiet = quiet, ...)
        }
        i <- i + 1
      }
      output[[paste0("seed_", s)]] <- list(words = used_words, iters = sum(used_words != word) + 1)
      if (!quiet) message(sprintf(">> Iterations (seed = %s): %s\n", s, i + 1))
    }
    attr(output, "input") <- input
    attr(output, "word") <- word
    attr(output, "iterations") <- length(seed)
    attr(output, "elapsed") <- toc("wordle_simulation", quiet = quiet)$time
    class(output) <- c("wordle_simulation", class(output))
    invisible(output)
  }
}

#' @rdname wordle
#' @param type Integer. 1 for summary and 2 for coloured results.
#' @export
print.wordle_simulation <- function(x, type = 1, ...) {
  iters_n <- unlist(lapply(x, function(x) x$iters))
  if (type == 1) {
    print(glued(
      "Seed Word: {attr(x, 'input')}
      Objective Word: {attr(x, 'word')}
      Iterations: {attr(x, 'iterations')}
        Mean to succeed: {signif(mean(iters_n), 3)}
        Max to succeed: {max(iters_n)} [seed = {which.max(iters_n)}]
      "
    ))
  }
  if (type == 2) {
    words <- lapply(x, function(x) x$words)
    for_print <- list()
    split_col <- "BLUE"
    names(split_col) <- attr(x, "word")
    for (i in seq_along(x)) {
      # Namings: word + seed + iterations
      word_split_iter <- split_col
      names(word_split_iter) <- paste(attr(x, "word"), names(x)[i], "->", iters_n[i], "iterations")
      class(word_split_iter) <- "wordle_check"
      res <- NULL
      res[[1]] <- word_split_iter
      # Word by word coloring
      list <- words[[i]]
      ## Remove the word when it's guessed when more than 1 opts
      list <- list[list != attr(x, "word")]
      results <- lapply(list, function(i) wordle_check(i, attr(x, "word"), print = FALSE))
      res <- append(res, results)
      for_print <- append(for_print, res)
    }
    txts <- unlist(lapply(for_print, function(x) paste(print(x, print = FALSE))))
    cat(txts, sep = "\n")
  }
}

# which.max(sapply(x, function(x) x$iters))
# which.min(sapply(x, function(x) x$iters))
# wordle_simulation("SAINT", "ABBEY", 9)

# SIMULATE N SOLUTIONS BY START WORD
wordle_opts <- function(input, word, dictionary = NULL, lang_dic = "en", method = 3, quiet = FALSE, ...) {
  if (is.null(dictionary)) dictionary <- wordle_dictionary(lang_dic, quiet = TRUE)
  if (!word %in% dictionary) dictionary <- c(dictionary, word)
  n_words_init <- length(dictionary)
  check <- wordle_check(input, word, dictionary, lang_dic, method, print = FALSE)
  dictionary <- discard_words(input, word, dictionary, check)
  input_coloured <- print(check, print = FALSE)
  if (!quiet) {
    message(cat(
      input_coloured, "reduced from", formatNum(n_words_init, 0),
      "to", formatNum(length(dictionary), 0)
    ))
  }
  attr(dictionary, "answer") <- word
  attr(dictionary, "last_word") <- input
  dictionary
}

discard_words <- function(input, word, dictionary, check) {
  if (input[1] != word[1]) {
    # Prepare data.frame to filtering words that match criteria
    words_lst <- lapply(dictionary, function(x) toupper(unlist(strsplit(x, ""))))
    words_df <- bind_rows(lapply(words_lst, function(x) as.data.frame(t(x)))) %>% mutate(word = dictionary)
    # Letters not present
    these <- unique(names(check[check == "RED"]))
    if (length(these) > 0) {
      are_nums <- these %in% as.character(0:9)
      if (any(are_nums)) {
        words_df <- filter(words_df, !grepl(paste0("[", paste(these[are_nums], collapse = "-"), "]"), .data$word))
      }
      are_lets <- grepl("[a-zA-Z]", these)
      if (any(are_lets)) {
        words_df <- filter(words_df, !grepl(paste(these[are_lets], collapse = "|"), .data$word))
      }
      are_symbs <- !(are_nums | are_lets)
      if (any(are_symbs)) {
        words_df <- filter(words_df, !grepl(paste(paste0("\\", these[are_symbs]), collapse = "|"), .data$word))
      }
    }
    # Letters present but not in position
    these <- names(check[check == "YELLOW"])
    words_df <- filter(words_df, grepl(paste(these, collapse = "|"), .data$word))
    these_id <- which(check == "YELLOW")
    for (i in seq_along(these)) words_df <- words_df[words_df[, these_id[i]] != these[i], ]
    # Letters present in position
    these <- names(check[check == "GREEN"])
    these_id <- which(check == "GREEN")
    for (i in seq_along(these)) words_df <- words_df[words_df[, these_id[i]] == these[i], ]
    dictionary <- words_df$word
  } else {
    dictionary <- word
  }
  dictionary
}

############ PLAY AREA ############

# # word <- sample(wordle_dictionary("en"), 1)
# word <- "ABBEY"
# wordle_check("OPENS", word)
# wordle_check("BABES", word)
# wordle_check("KEBAB", word)
# wordle_check("ABYSS", word)
# wordle_check("ABBEY", word)
#
# # Cheat using scrabble_words()
# set.seed(1) # For reproducibility
# # word <- sample(wordle_dictionary("en"), 1)
# word <- "FLESH" # Secret word
# # First iteration
# wordle_check("SAINT", word)
# scrabble_words(
#   language = "en",
#   force_n = 5,
#   tiles = "S",
#   force_exclude = "AINT"
# ) %>% pull(.data$word) %>% sample(10)
# # Second iteration
# wordle_check("fords", word)
# scrabble_words(
#   language = "en",
#   force_n = 5,
#   tiles = "S",
#   force_start = "F",
#   force_exclude = "AINTORD"
# ) %>% pull(.data$word) %>% sample(10)
# # Third iteration
# wordle_check("fesse", word)
# scrabble_words(
#   language = "en",
#   force_n = 5,
#   tiles = "ES",
#   force_start = "F__S",
#   force_exclude = "AINTORD"
# )
# # Fourth iteration
# wordle_check("flesh", word)

####################################################################
#' Automated H2O's AutoML
#'
#' This function lets the user create a robust and fast model, using
#' H2O's AutoML function. The result is a list with the best model,
#' its parameters, datasets, performance metrics, variables
#' importance, and plots. Read more about the \code{h2o_automl()} pipeline
#' \href{https://laresbernardo.github.io/lares/articles/machine-learning.html}{here}.
#'
#' For additional tutorials and examples:
#' \itemize{
#'   \item \href{https://laresbernardo.github.io/lares/articles/machine-learning.html}{Machine Learning with H2O Package}
#'   \item \href{https://laresbernardo.github.io/lares/reference/mplot_roc.html}{Understanding ROC Curves}
#' }
#'
#' @section List of algorithms:
#' \href{https://docs.h2o.ai/h2o/latest-stable/h2o-docs/automl.html}{-> Read more here}
#' \describe{
#'   \item{DRF}{Distributed Random Forest, including Random Forest (RF)
#'   and Extremely-Randomized Trees (XRT)}
#'   \item{GLM}{Generalized Linear Model}
#'   \item{XGBoost}{eXtreme Grading Boosting}
#'   \item{GBM}{Gradient Boosting Machine}
#'   \item{DeepLearning}{Fully-connected multi-layer artificial neural network}
#'   \item{StackedEnsemble}{Stacked Ensemble}
#' }
#'
#' @section Methods:
#' \describe{
#'   \item{print}{Use \code{print} method to print models stats and summary}
#'   \item{plot}{Use \code{plot} method to plot results using \code{mplot_full()}}
#' }
#'
#' @family Machine Learning
#' @inheritParams h2o::h2o.automl
#' @inheritParams mp3_get
#' @param df Dataframe. Dataframe containing all your data, including
#' the dependent variable labeled as \code{'tag'}. If you want to define
#' which variable should be used instead, use the \code{y} parameter.
#' @param y Variable or Character. Name of the dependent variable or response.
#' @param ignore Character vector. Force columns for the model to ignore
#' @param train_test Character. If needed, \code{df}'s column name with 'test'
#' and 'train' values to split
#' @param split Numeric. Value between 0 and 1 to split as train/test
#' datasets. Value is for training set. Set value to 1 to train with all
#' available data and test with same data (cross-validation will still be
#' used when training). If \code{train_test} is set, value will be overwritten
#' with its real split rate.
#' @param weight Column with observation weights. Giving some observation a
#' weight of zero is equivalent to excluding it from the dataset; giving an
#' observation a relative weight of 2 is equivalent to repeating that
#' row twice. Negative weights are not allowed.
#' @param target Value. Which is your target positive value? If
#' set to \code{'auto'}, the target with largest \code{mean(score)} will be
#' selected. Change the value to overwrite. Only used when binary
#' categorical model.
#' @param balance Boolean. Auto-balance train dataset with under-sampling?
#' @param impute Boolean. Fill \code{NA} values with MICE?
#' @param no_outliers Boolean/Numeric. Remove \code{y}'s outliers from the dataset?
#' Will remove those values that are farther than n standard deviations from
#' the dependent variable's mean (Z-score). Set to \code{TRUE} for default (3)
#' or numeric to set a different multiplier.
#' @param unique_train Boolean. Keep only unique row observations for training data?
#' @param center,scale Boolean. Using the base function scale, do you wish
#' to center and/or scale all numerical values?
#' @param thresh Integer. Threshold for selecting binary or regression
#' models: this number is the threshold of unique values we should
#' have in \code{'tag'} (more than: regression; less than: classification)
#' @param seed Integer. Set a seed for reproducibility. AutoML can only
#' guarantee reproducibility if max_models is used because max_time is
#' resource limited.
#' @param max_models,max_time Numeric. Max number of models and seconds
#' you wish for the function to iterate. Note that max_models guarantees
#' reproducibility and max_time not (because it depends entirely on your
#' machine's computational characteristics)
#' @param start_clean Boolean. Erase everything in the current h2o
#' instance before we start to train models? You may want to keep other models
#' or not. To group results into a custom common AutoML project, you may
#' use \code{project_name} argument.
#' @param exclude_algos,include_algos Vector of character strings. Algorithms
#' to skip or include during the model-building phase. Set NULL to ignore.
#' When both are defined, only \code{include_algos} will be valid.
#' @param plots Boolean. Create plots objects?
#' @param alarm Boolean. Ping (sound) when done. Requires \code{beepr}.
#' @param print Boolean. Print summary when process ends?
#' @param save Boolean. Do you wish to save/export results into your
#' working directory?
#' @param subdir Character. In which directory do you wish to save
#' the results? Working directory as default.
#' @param project Character. Your project's name
#' @param model_name Character. Optional custom name for the model. If provided,
#' this name will be used when saving the model with \code{export_results()}. If NULL
#' (default), the H2O-generated model ID will be used.
#' @param verbosity Character. Verbosity of the backend messages printed during training.
#' Must be one of NULL (live log disabled), "debug", "info", "warn", "error". Defaults to NULL.
#' @param ... Additional parameters on \code{h2o::h2o.automl}
#' @return List. Trained model, predicted scores and datasets used, performance
#' metrics, parameters, importance data.frame, seed, and plots when \code{plots=TRUE}.
#' @examples
#' \dontrun{
#' # CRAN
#' data(dft) # Titanic dataset
#' dft <- subset(dft, select = -c(Ticket, PassengerId, Cabin))
#'
#' # Classification: Binomial - 2 Classes
#' r <- h2o_automl(dft, y = Survived, max_models = 1, impute = FALSE, target = "TRUE", alarm = FALSE)
#'
#' # Let's see all the stuff we have inside:
#' lapply(r, names)
#'
#' # Classification: Multi-Categorical - 3 Classes
#' r <- h2o_automl(dft, Pclass, ignore = c("Fare", "Cabin"), max_time = 30, plots = FALSE)
#'
#' # Regression: Continuous Values
#' r <- h2o_automl(dft, y = "Fare", ignore = c("Pclass"), exclude_algos = NULL, quiet = TRUE)
#' print(r)
#'
#' # WITH PRE-DEFINED TRAIN/TEST DATAFRAMES
#' splits <- msplit(dft, size = 0.8)
#' splits$train$split <- "train"
#' splits$test$split <- "test"
#' df <- rbind(splits$train, splits$test)
#' r <- h2o_automl(df, "Survived", max_models = 1, train_test = "split")
#' }
#' @export
h2o_automl <- function(df, y = "tag",
                       ignore = NULL,
                       train_test = NA,
                       split = 0.7,
                       weight = NULL,
                       target = "auto",
                       balance = FALSE,
                       impute = FALSE,
                       no_outliers = TRUE,
                       unique_train = TRUE,
                       center = FALSE,
                       scale = FALSE,
                       thresh = 10,
                       seed = 0,
                       nfolds = 5,
                       max_models = 3,
                       max_time = 10 * 60,
                       start_clean = FALSE,
                       exclude_algos = c("StackedEnsemble", "DeepLearning"),
                       include_algos = NULL,
                       plots = TRUE,
                       alarm = TRUE,
                       quiet = FALSE,
                       print = TRUE,
                       save = FALSE,
                       subdir = NA,
                       project = "AutoML Results",
                       model_name = NULL,
                       verbosity = NULL,
                       ...) {
  try_require("h2o")
  tic(id = "h2o_automl")
  on.exit(toc(id = "h2o_automl", msg = "Process duration:", quiet = quiet))

  if (!quiet) message(paste(Sys.time(), "| Started process..."))

  quiet(h2o.init(nthreads = -1, port = 54321, ip = "127.0.0.1"))

  df <- as.data.frame(df)
  y <- gsub('"', "", as_label(enquo(y)))

  # PROCESS THE DATA
  processed <- model_preprocess(
    df,
    y = y,
    train_test = train_test,
    split = split,
    weight = weight,
    target = target,
    balance = balance,
    impute = impute,
    no_outliers = no_outliers,
    unique_train = unique_train,
    center = center,
    scale = scale,
    thresh = thresh,
    seed = seed,
    quiet = quiet
  )

  # PROCESSED DATA: TRAIN AND TEST
  df <- processed$data
  train <- df[processed$train_index, ]
  test <- df[-processed$train_index, ]
  if (nrow(test) == 0) test <- train

  # MODEL TYPE (based on inputs + thresh value)
  model_type <- processed$model_type

  # ALGORITHMS
  if (length(exclude_algos) > 0 && length(include_algos) == 0 && !quiet) {
    message(paste("- ALGORITHMS: excluded", vector2text(exclude_algos)))
  }
  if (length(include_algos) > 0 && !quiet) {
    message(paste("- ALGORITHMS: included", vector2text(include_algos)))
    exclude_algos <- NULL
  }

  # START FRESH?
  if (!quiet && !isTRUE(start_clean)) {
    message(sprintf(
      paste(
        "- CACHE: Previous models %s being erased.",
        "You may use 'start_clean' [clear] or 'project_name' [join]"
      ),
      ifelse(start_clean, "are", "are not")
    ))
  }
  if (start_clean) quiet(h2o.removeAll())

  # INFORMATIVE MSG ON FLOW's UI
  flow <- "http://localhost:54321/flow/index.html"
  if (!quiet && Sys.getenv("HOSTNAME") == "") {
    message("- UI: You may check results using H2O Flow's interactive platform: ", flow)
  }

  # RUN AUTOML
  if (!quiet) {
    message(sprintf(">>> Iterating until %s models or %s seconds...", max_models, max_time))
  }
  training <- .quiet_h2o(as.h2o(train), quiet = TRUE)
  aml <- .quiet_h2o(h2o.automl(
    x = colnames(df)[!colnames(df) %in% c("tag", ignore)],
    y = "tag",
    training_frame = training,
    weights_column = weight,
    max_runtime_secs = max_time,
    max_models = max_models,
    exclude_algos = exclude_algos,
    include_algos = include_algos,
    nfolds = nfolds,
    # project_name = project,
    seed = seed,
    verbosity = verbosity,
    ...
  ), quiet = quiet)

  if (nrow(aml@leaderboard) == 0) {
    warning("NO MODELS TRAINED. Please set max_models to at least 1 and increase max_time")
  } else {
    if (!is.nan(aml@leaderboard[1, 2])) {
      if (!quiet) {
        message(paste("- EUREKA: Succesfully generated", nrow(aml@leaderboard), "models"))
        if (print) print(head(aml@leaderboard, 3))
      }
    }
  }

  # GET RESULTS AND PERFORMANCE
  results <- h2o_results(
    aml, test, train, y,
    which = 1,
    model_type = model_type,
    target = target,
    split = split,
    plots = plots,
    project = project,
    ignore = ignore,
    seed = seed,
    quiet = quiet,
    model_name = model_name
  )

  if (save) {
    export_results(results, subdir = subdir, thresh = thresh)
    if (!quiet) message("- EXPORT: Results and model files exported succesfully!")
  }

  if (!quiet && print) print(results)

  if (alarm && !quiet) {
    try_require("beepr", stop = FALSE)
    try(beep())
  }

  attr(results, "type") <- "h2o_automl"
  results
}

#' @rdname h2o_automl
#' @aliases h2o_automl
#' @param x h2o_automl object
#' @export
plot.h2o_automl <- function(x, ...) {
  if (!inherits(x, "h2o_automl")) {
    stop("Object must be class h2o_automl")
  }
  if ("plots" %in% names(x)) {
    x$plots$dashboard
  } else {
    invisible(mplot_full(
      tag = x$scores_test$tag,
      score = x$scores_test$score,
      multis = select(x$scores_test, -.data$tag, -.data$score)
    ))
  }
}

#' @rdname h2o_automl
#' @aliases h2o_automl
#' @param importance Boolean. Print important variables?
#' @export
print.h2o_automl <- function(x, importance = TRUE, ...) {
  if (!inherits(x, "h2o_automl")) {
    stop("Object must be class h2o_automl")
  }
  aux <- list()
  selected <- which(as.vector(x$leaderboard$model_id) == x$model_name)
  n_models <- nrow(x$leaderboard)
  data_points <- nrow(x$datasets$global)
  split <- round(100 * x$split)

  if (x$type == "Classification") {
    cats <- filter(x$datasets$global, grepl("train", .data$train_test)) %>%
      .[, x$y] %>%
      unique() %>%
      nrow()
    x$type <- sprintf("%s (%s classes)", x$type, cats)
  }

  aux[["met"]] <- glued(
    "Test metrics:
{v2t({met}, sep = '\n', quotes = FALSE)}",
    met = paste(
      "  ",
      names(x$metrics$metrics), "=",
      signif(x$metrics$metrics, 5)
    )
  )

  if ("importance" %in% names(x) && importance == TRUE) {
    if (nrow(x$importance) > 0) {
      aux[["imp"]] <- glued(
        "Most important variables:
{v2t({imp}, sep = '\n', quotes = FALSE)}",
        imp = paste(
          "  ",
          x$importance %>% head(5) %>%
            mutate(label = sprintf(
              "%s (%s)",
              .data$variable,
              formatNum(100 * .data$importance, 1, pos = "%")
            )) %>%
            pull(.data$label)
        )
      )
    }
  }

  print(glued("
    Model ({selected}/{n_models}): {x$model_name}
    Dependent Variable: {x$y}
    Type: {x$type}
    Algorithm: {toupper(x$algorithm)}
    Split: {split}% training data (of {data_points} observations)
    Seed: {x$seed}

    {aux$met}

    {aux$imp}"))
}


####################################################################
#' Automated H2O's AutoML Results
#'
#' This is an auxiliary function to calculate predictions and results
#' when using the \code{h2o_automl()} function.
#'
#' @inheritParams h2o_automl
#' @param h2o_object H2O Leaderboard (H2OFrame/H2OAutoML) or Model (h2o)
#' @param test,train Dataframe. Must have the same columns
#' @param which Integer. Which model to select from leaderboard
#' @param model_type Character. Select "Classification" or "Regression"
#' @param ignore Character vector. Columns too ignore
#' @param leaderboard H2O's Leaderboard. Passed when using
#' \code{h2o_selectmodel} as it contains plain model and no leader board.
#' @return List. Trained model, predicted scores and datasets used, performance
#' metrics, parameters, importance data.frame, seed, and plots when \code{plots=TRUE}.
#' @export
h2o_results <- function(h2o_object, test, train, y = "tag", which = 1,
                        model_type, target = "auto", split = 0.7,
                        ignore = NULL, quiet = FALSE,
                        project = "ML Project", seed = 0,
                        leaderboard = list(),
                        plots = TRUE,
                        model_name = NULL,
                        ...) {
  # MODEL TYPE
  types <- c("Classification", "Regression")
  check_opts(model_type, types)
  thresh <- ifelse(model_type == types[1], 10000, 0)

  # When using h2o_select
  if ("train_test" %in% colnames(test)) {
    colnames(test)[colnames(test) == y] <- "tag"
    colnames(train)[colnames(train) == y] <- "tag"
    test <- test[, 1:(which(colnames(test) == "train_test") - 1)]
    train <- train[, 1:(which(colnames(train) == "train_test") - 1)]
    split <- round(nrow(train) / (nrow(train) + nrow(test)), 2)
  }

  # GLOBAL DATAFRAME FROM TEST AND TRAIN
  if (!all(colnames(train) %in% colnames(test))) {
    stop("All columns from train data must be present on test data as well!")
  }
  if (split == 1) {
    global <- train %>% mutate(train_test = "train_test")
  } else {
    global <- data.frame(test) %>%
      bind_rows(train) %>%
      mutate(train_test = c(rep("test", nrow(test)), rep("train", nrow(train))))
  }
  colnames(global)[colnames(global) == "tag"] <- y
  if (model_type == "Classification") {
    cats <- unique(global[, colnames(global) == y])
  } else {
    cats <- "None"
  }

  # SELECT MODEL FROM h2o_automl()
  if (any(c("H2OFrame", "H2OAutoML") %in% class(h2o_object))) {
    # Note: Best model from leaderboard is which = 1
    m <- h2o.getModel(as.vector(h2o_object@leaderboard$model_id[which]))
    if (!quiet) message(paste("SELECTED MODEL:", as.vector(m@model_id)))
  } else {
    m <- h2o_object
  }

  # VARIABLES IMPORTANCES
  # https://docs.h2o.ai/h2o/latest-stable/h2o-docs/variable-importance.html
  if (sum(grepl("Stacked", as.vector(m@model_id))) > 0) {
    stacked <- TRUE
    if (!quiet) message("- NOTE: No importance features for Stacked Ensemble Models")
  } else {
    stacked <- FALSE
  }
  if (!stacked) {
    imp <- data.frame(h2o.varimp(m)) %>%
      {
        if ("names" %in% colnames(.)) {
          rename(., "variable" = "names", "importance" = "coefficients")
        } else {
          .
        }
      } %>%
      {
        if ("percentage" %in% colnames(.)) {
          rename(., "importance" = "percentage")
        } else {
          .
        }
      }
    noimp <- if (nrow(imp) > 0) {
      dplyr::filter(imp, .data$importance < 1 / (nrow(imp) * 4)) %>%
        arrange(desc(.data$importance))
    } else {
      imp
    }
    if (nrow(noimp) > 0) {
      topn <- noimp %>%
        ungroup() %>%
        slice(1:8)
      which <- vector2text(topn$variable, quotes = FALSE)
      if (nrow(noimp) > 8) {
        which <- paste(which, "and", nrow(noimp) - 8, "other...")
      }
      if (!quiet) {
        message(paste("- NOTE: The following variables were the least important:", which))
      }
    }
  }

  # GET PREDICTIONS
  if (!quiet) message(paste0(">>> Running predictions for ", y, "..."))
  predictions <- .quiet_h2o(h2o_predict_model(global, m), quiet = TRUE)
  global <- cbind(global, predictions)
  # Change dots for space
  if (sum(grepl(" ", cats)) > 0) {
    colnames(global) <- str_replace_all(colnames(global), "\\.", " ")
  }

  # For performance metrics
  scores_test <- .get_scores(
    predictions, test,
    model_type = model_type,
    target = target,
    cats = cats
  )
  multis <- scores_test$multis
  scores <- scores_test$scores

  # # Used for train metrics
  # scores_train <- .get_scores(
  #   predictions, train,
  #   model_type = model_type,
  #   target = target, cats = cats)
  # scores_tr <- scores_train$scores
  # multis_tr <- scores_train$multis
  # scores_train <- data.frame(tag = as.vector(train$tag), scores_tr)

  # GET ALL RESULTS INTO A LIST
  results <- list()
  results[["model"]] <- m
  results[["y"]] <- y
  results[["scores_test"]] <- data.frame(tag = as.vector(test$tag), scores)
  results[["metrics"]] <- model_metrics(
    tag = results$scores_test$tag,
    score = results$scores_test$score,
    multis = multis,
    thresh = thresh,
    target = target,
    model_name = as.vector(m@model_id),
    plots = plots
  )
  cvresults <- m@model$cross_validation_metrics_summary
  if (!is.null(cvresults)) {
    results$metrics[["cv_metrics"]] <- as_tibble(
      data.frame(
        metric = rownames(cvresults),
        mutate_all(cvresults, list(~ as.numeric(as.character(.))))
      )
    )
  }
  if (model_type == "Classification") {
    if (length(cats) == 2) {
      results$metrics[["max_metrics"]] <- data.frame(
        m@model$cross_validation_metrics@metrics$max_criteria_and_metric_scores
      )
    }
    if (length(cats) > 2) {
      results$metrics[["hit_ratio"]] <- data.frame(
        m@model$cross_validation_metrics@metrics$hit_ratio_table
      )
    }
  }
  # results[["parameters"]] <- m@parameters[
  #   sapply(m@parameters, function(x) length(x) == 1)] %>%
  #   bind_rows() %>% tidyr::gather(key = "parameter")
  results[["parameters"]] <- m@parameters
  if (!stacked) results[["importance"]] <- imp
  results[["datasets"]] <- list(
    global = as_tibble(global),
    test = filter(global, grepl("test", .data$train_test))
  )
  # results[["metrics_train"]] <- model_metrics(
  #   tag = scores_train$tag,
  #   score = scores_train$score,
  #   multis = multis_tr,
  #   thresh = thresh,
  #   target = target,
  #   model_name = as.vector(m@model_id),
  #   type = "train")
  results[["scoring_history"]] <- as_tibble(m@model$scoring_history)
  results[["categoricals"]] <- list_cats(filter(global, grepl("train", .data$train_test)))
  results[["type"]] <- model_type
  results[["split"]] <- split
  if (model_type == "Classification") {
    results[["threshold"]] <- thresh
  }
  use_model_name <- if (!is.null(model_name)) model_name else as.vector(m@model_id)
  results[["model_name"]] <- use_model_name
  results[["algorithm"]] <- m@algorithm
  if (any(c("H2OFrame", "H2OAutoML") %in% class(h2o_object))) {
    results[["leaderboard"]] <- h2o_object@leaderboard
  }
  if (length(leaderboard) > 0) {
    results[["leaderboard"]] <- leaderboard
  }
  results[["project"]] <- project
  results[["y"]] <- y
  results[["ignored"]] <- ignore
  results[["seed"]] <- seed
  results[["h2o"]] <- h2o.getVersion()

  if (plots) {
    if (!quiet) message(">>> Generating plots...")
    plots <- list()
    plots[["dashboard"]] <- mplot_full(
      tag = results$scores_test$tag,
      score = results$scores_test$score,
      multis = multis,
      thresh = thresh,
      subtitle = results$project,
      model_name = results$model_name,
      plot = FALSE
    )
    plots[["metrics"]] <- results$metrics$plots
    results$metrics$plots <- NULL
    if (length(multis) > 1) {
      plots[["top_cats"]] <- mplot_topcats(
        tag = results$scores_test$tag,
        score = results$scores_test$score,
        multis = multis,
        model_name = results$model_name
      )
    }
    if (!stacked) {
      plots[["importance"]] <- mplot_importance(
        var = results$importance$variable,
        imp = results$importance$importance,
        model_name = results$model_name,
        subtitle = results$project
      )
    }
    plots <- append(plots, rev(as.list(results$metrics$plots)))
    results$plots <- plots
  }

  attr(results, "type") <- "h2o_automl"
  class(results) <- c("h2o_automl", class(results))
  results
}

.get_scores <- function(predictions,
                        traintest,
                        model_type,
                        target = "auto",
                        cats) {
  type <- deparse(substitute(traintest))

  # Train or Test data
  nrows <- nrow(traintest)
  if (type == "test") {
    scores <- predictions[1:nrows, ]
  }
  if (type == "train") {
    scores <- predictions[(nrows + 1):nrows, ]
  }

  # Selected target value
  if (target != "auto" && length(cats) == 2) {
    scores <- target_set(
      tag = as.vector(traintest$tag),
      score = scores[, 2],
      target = target,
      quiet = TRUE
    )$df
  }

  # Multis object and standard predictions output
  multis <- NA
  if (model_type == "Classification") {
    if (length(cats) == 2) {
      scores <- select_if(scores, is.numeric) %>% .[, 1]
    } else {
      colnames(scores)[1] <- "score"
      multis <- select(scores, -.data$score)
    }
  } else {
    scores <- data.frame(score = as.vector(scores))
  }
  list(scores = scores, multis = multis)
}


####################################################################
#' Select Model from h2o_automl's Leaderboard
#'
#' Select wich model from the h2o_automl function to use
#'
#' @family Machine Learning
#' @family Tools
#' @inheritParams h2o_automl
#' @param results \code{h2o_automl()} object.
#' @param which_model Integer. Which model from the leaderboard you wish to use?
#' @return H2O processed model
#' @export
h2o_selectmodel <- function(results, which_model = 1, quiet = FALSE, ...) {
  check_attr(results, attr = "type", check = "h2o_automl")

  # Select model (Best one by default)
  ntop <- nrow(results$leaderboard)
  if (which_model > ntop) {
    stop("Select a valid model ID. Range: 1 to ", ntop)
  }
  model_id <- as.vector(results$leaderboard$model_id[which_model])
  if (!quiet) message("Model selected: ", model_id)
  m <- h2o.getModel(model_id)
  d <- results$datasets

  # Calculate everything
  output <- h2o_results(
    m,
    test = d$test,
    train = filter(d$global, grepl("test", .data$train_test)),
    y = results$y,
    which = which_model,
    model_type = results$type,
    project = results$project,
    leaderboard = results$leaderboard,
    seed = results$seed,
    quiet = TRUE,
    ...
  )

  if (!quiet) print(output)
  output
}


####################################################################
#' Export h2o_automl's Results
#'
#' Export RDS, TXT, POJO, MOJO and all results from \code{h2o_automl()}.
#'
#' @family Machine Learning
#' @family Tools
#' @param results \code{h2o_automl} or \code{h2o} model
#' @param thresh Integer. Threshold for selecting binary or regression
#' models: this number is the threshold of unique values we should
#' have in 'tag' (more than: regression; less than: classification)
#' @param which Character vector. Select which file format to export:
#' Possible values: txt, csv, rds, binary, mojo, plots. You might also
#' use dev (txt, csv, rds) or production (binary, mojo) or simply don't use
#' parameter to export everything
#' @param note Character. Add a note to the txt file. Useful when lots of
#' models are trained and saved to remember which one is which one
#' @param subdir Character. In which directory do you wish to save
#' the results?
#' @param save Boolean. Do you wish to save/export results?
#' @param seed Numeric. For reproducible results and random splits.
#' @return No return value, called for side effects.
#' @export
export_results <- function(results,
                           thresh = 10,
                           which = c(
                             "txt", "csv", "rds",
                             "binary", "mojo", "plots",
                             "dev", "production"
                           ),
                           note = NA,
                           subdir = NA,
                           save = TRUE,
                           seed = 0) {
  if (save) {
    try_require("h2o")
    quiet(h2o.init(nthreads = -1, port = 54321, ip = "127.0.0.1"))

    pass <- !is.null(attr(results, "type"))
    if (!pass) results <- list(model = results)
    stopifnot(grepl("H2O", class(results$model)))
    name <- ifelse(pass, results$model_name, results$model@model_id)
    subdir <- paste0(ifelse(is.na(subdir), "", subdir), "/", name)
    if (substr(subdir, 1, 1) == "/") subdir <- substr(subdir, 2, nchar(subdir))

    # Directory to save all our results
    dir <- file.path(subdir)
    message(paste("Export directory:", dir))
    if (!dir.exists(dir)) {
      message("Creating directory: ", subdir)
      dir.create(dir, recursive = TRUE)
    }

    if ("dev" %in% which) which <- unique(c(which, "txt", "csv", "rds"))
    if ("production" %in% which) which <- unique(c(which, "binary", "mojo"))

    if ("txt" %in% which || !is.na(note)[1] && pass) {
      on.exit(set.seed(seed))
      results_txt <- list(
        "Project" = results$project,
        "Note" = note,
        "Model Type" = results$type,
        "Algorithm" = results$algorithm,
        "Model name" = name,
        "Train/Test" = table(results$datasets$global$train_test),
        "Metrics Glossary" = results$metrics$dictionary,
        "Test Metrics" = results$metrics$metrics,
        "Test Metrics by labels" = if (length(results$metrics$metrics_tags) > 1) {
          results$metrics$metrics_tags
        } else {
          "NA"
        },
        "Test's Confusion Matrix" = if (length(results$metrics$confusion_matrix) > 1) {
          results$metrics$confusion_matrix
        } else {
          NULL
        },
        "Predicted Variable" = results$y,
        "Ignored Variables" = results$ignored,
        "Variables Importance" = results$importance,
        "H2O Global Results" = results$model,
        "Leaderboard" = results$leaderboard,
        "Data examples" = data.frame(sample_n(results$datasets$global, 10)),
        "Seed" = results$seed,
        "H20 Version" = results$h2o
      )
      if (is.na(note)[1]) results_txt$Note <- NULL
      capture.output(results_txt, file = paste0(dir, "/", name, ".txt"))
      cats <- lapply(results$categoricals, data.frame)
      aux <- cats[names(cats)[!names(cats) %in% results$ignore]]
      capture.output(aux, file = paste0(dir, "/", name, "_cats.txt"))
      message(">>> Summary text files saved...")
    }

    # Export CSV with predictions and datasets
    if ("csv" %in% which && pass) {
      write.csv(results$datasets$global,
        paste0(dir, "/", name, ".csv"),
        row.names = FALSE
      )
      message(">>> CSV file exported...")
    }

    # Export Results List
    if ("rds" %in% which) {
      saveRDS(results, file = paste0(dir, "/", name, ".rds"))
      message(">>> RDS file exported...")
    }

    # Export Model as POJO && MOJO for Production
    if ("mojo" %in% which) {
      h2o.download_mojo(results$model, path = dir, get_genmodel_jar = TRUE)
      message(">>> MOJO (zip + jar files) exported...")
    }

    # if (pojo) h2o.download_pojo(results$model, path = dir)

    # Export Binary
    if ("binary" %in% which) {
      h2o.saveModel(results$model, path = dir, force = TRUE)
      message(">>> Binary file saved...")
    }

    if ("plots" %in% which && "plots" %in% names(results) && pass) {
      message(">>> Saving plots...")
      # Metrics plots
      aux <- results$plots$metrics
      for (i in seq_along(aux)) {
        export_plot(aux[[i]],
          name = names(aux)[i],
          width = 8, height = 6, res = 300,
          subdir = paste0(subdir, "/Plots"),
          quiet = TRUE
        )
      }
      # Other plots
      aux <- results$plots
      aux$metrics <- NULL
      for (i in seq_along(aux)) {
        export_plot(aux[[i]],
          name = names(aux)[i],
          width = 8, height = 6, res = 300,
          subdir = paste0(subdir, "/Plots"),
          quiet = TRUE
        )
      }
      message(">>> Plots saved...")
    }
    message(paste("Succesfully exported files:", vector2text(which)))
  }
}


####################################################################
#' Split a dataframe for training and testing sets
#'
#' This function splits automatically a dataframe into train and
#' test datasets. You can define a seed to get the same results
#' every time, but has a default value. You can prevent it from
#' printing the split counter result.
#'
#' @family Machine Learning
#' @family Tools
#' @param df Dataframe
#' @param size Numeric. Split rate value, between 0 and 1. If set to
#' 1, the train and test set will be the same.
#' @param seed Integer. Seed for random split
#' @param print Boolean. Print summary results?
#' @return List with both datasets, summary, and split rate.
#' @examples
#' data(dft) # Titanic dataset
#' splits <- msplit(dft, size = 0.7, seed = 123)
#' names(splits)
#' @export
msplit <- function(df, size = 0.7, seed = 0, print = TRUE) {
  if (size <= 0 || size > 1) stop("Set size parameter to a value >0 and <=1")

  on.exit(set.seed(seed))
  df <- data.frame(df)
  if (size == 1) train <- test <- df

  if (size < 1 && size > 0) {
    ind <- sample(seq_len(nrow(df)), size = floor(size * nrow(df)))
    train <- df[ind, ]
    test <- df[-ind, ]
  } else {
    ind <- seq_len(nrow(df))
  }

  train_size <- dim(train)
  test_size <- dim(test)
  summary <- rbind(train_size, test_size)[, 1]

  if (print == TRUE) print(summary)

  list(
    train = train,
    test = test,
    summary = summary,
    split_size = size,
    train_index = ind
  )
}


####################################################################
#' Set Target Value in Target Variable
#'
#' This function detects or forces the target value when predicting
#' a categorical binary model. This is an auxiliary function.
#'
#' @inheritParams mp3_get
#' @param tag Vector. Real known label
#' @param score Vector. Predicted value or model's result
#' @param target Value. Which is your target positive value? If
#' set to 'auto', the target with largest mean(score) will be
#' selected. Change the value to overwrite. Only used when binary
#' categorical model.
#' @return List. Contains original data.frame \code{df} and
#' \code{which} with the target variable.
#' @export
target_set <- function(tag, score, target = "auto", quiet = FALSE) {
  df <- data.frame(tag = tag, score = score)

  # Validate inputs
  if (!is.numeric(score) || is.numeric(tag)) {
    stop("Your tag must be categorical. Your score must be numerical.")
  }

  # Get mean scores for each tag
  means <- df %>%
    group_by(.data$tag) %>%
    summarise(mean = mean(.data$score))
  auto <- means$tag[means$mean == max(means$mean)]
  if (length(auto) > 1) {
    auto <- if (any(auto %in% target)) target else auto[1]
  }
  if (target == "auto") {
    target <- auto
  }
  if (!target %in% unique(df$tag)) {
    stop(paste(
      "Your target value", target, "is not valid.",
      "Possible other values:", vector2text(unique(df$tag))
    ))
  }
  if (!quiet) message(paste("Target value:", target))
  # If the forced target value is the "lower scores" value, invert scores
  if (auto != target) df$score <- df$score * (-1) + 1
  list(df = df, which = target)
}


####################################################################
#' Iterate Seeds on AutoML
#'
#' This functions lets the user iterate and search for best seed. Note that if
#' the results change a lot, you are having a high variance in your data.
#'
#' @family Machine Learning
#' @inheritParams h2o_automl
#' @param tries Integer. Number of iterations
#' @param ... Additional arguments passed to \code{h2o_automl}
#' @return data.frame with performance results by seed tried on every row.
#' @export
iter_seeds <- function(df, y, tries = 10, ...) {
  seeds <- data.frame()
  for (i in 1:tries) {
    model <- h2o_automl(df, y, seed = i, quiet = TRUE, ...)
    seeds <- rbind(seeds, cbind(seed = i, model$metrics$metrics))
    seeds <- arrange(seeds, desc(2))
    statusbar(i, tries, seeds[1, 1])
  }
  seeds
}

.quiet_h2o <- function(..., quiet = TRUE) {
  if (quiet) on.exit(h2o.no_progress())
  x <- eval(...)
  h2o.show_progress()
  x
}

####################################################################
#' Theme for ggplot2 (lares)
#'
#' Based on \code{hrbrthemes}' \code{theme_ipsum} and customized for \code{lares} usage.
#' With this team you can custom the colour and fill palettes, global colour parameters,
#' major and minor grids, legend, font and font size.
#'
#' @section Why Arial Narrow?:
#' First and foremost, Arial Narrow is generally installed by default or readily
#' available on any modern system, so it's "free"-ish; plus, it is a condensed font
#' with solid default kerning pairs and geometric numbers.
#'
#' @family Themes
#' @inheritParams cache_write
#' @param font,size Character and numeric. Base font family and base size for texts.
#' \code{Arial Narrow} is set by default when the library is loaded; you may change it
#' with \code{Sys.setenv("LARES_FONT" = "X")} or by using this parameter manually.
#' If \code{font} is "ignore" or NA, it'll be ignored.
#' @param main_colour,hard_colour,soft_colour,plot_colour,panel_colour
#' Character. Main colours for your theme.
#' @param background Character. Main colour for your background. Overwrites
#' \code{plot_colour} and \code{panel_colour}.
#' @param no_facets Boolean. Suppress facet labels?
#' @param legend Character. Legend position: \code{"top"}, \code{"right"},
#' \code{"bottom"}, or \code{"left"} You can also set to \code{FALSE} or
#' \code{"none"} to suppress legend.
#' @param grid Character or Boolean. Use \code{TRUE/FALSE} or a combination of
#' \code{X}, \code{x}, \code{Y}, \code{y} to enable/disable minor and major grids.
#' @param axis Character or Boolean. Use \code{TRUE/FALSE}, \code{x} or \code{Y}
#' to enable X and/or Y axis lines.
#' @param clean Boolean. Suppress grids and axis? Overwrites both parameters.
#' @param mg Numeric. External margins reference.
#' @param pal Integer. \code{1} for fill and colour palette,
#' \code{2} for only colour palette, \code{3} for only fill palette, \code{4} for
#' personal labels-colour palette. \code{0} for nothing.
#' @param palette Character vector. Pass a vector with HEX colour
#' codes to use a custom palette. If you pass a named vector, the name values will be
#' used as fill and the values will be used as colour.
#' @param which Character. When \code{pal = 3}, select which colours should be
#' added with the custom colours palette: fill, colour, text (fct) - first letters.
#' @return Themed ggplot2 object
#' @examples
#' \donttest{
#' data(dft)
#' library(ggplot2)
#' p <- ggplot(dft, aes(x = Pclass, y = sum(Fare), fill = Pclass)) +
#'   geom_col()
#' p + theme_lares()
#' p + theme_lares(pal = 1)
#' p + theme_lares(background = "#999999", mg = 25)
#' p + theme_lares(legend = "top", grid = "Yy")
#' p + theme_lares(clean = TRUE)
#' }
#' @export
theme_lares <- function(font = Sys.getenv("LARES_FONT"),
                        size = 12,
                        main_colour = "darkorange3",
                        hard_colour = "black",
                        soft_colour = "grey30",
                        plot_colour = "transparent",
                        panel_colour = "transparent",
                        background = "transparent",
                        no_facets = FALSE,
                        legend = NULL,
                        grid = TRUE,
                        axis = TRUE,
                        clean = FALSE,
                        mg = 9,
                        pal = 0,
                        palette = NULL,
                        which = "fc",
                        quiet = TRUE,
                        ...) {
  # Start from theme_minimal()
  ret <- theme_minimal(base_size = size)

  # Overwrite parameters with quick setting parameters
  if (clean) axis <- grid <- FALSE
  if (background != "transparent") panel_colour <- plot_colour <- background
  if (isFALSE(legend)) legend <- "none"

  # Check and set font
  font <- .font_global(font, quiet = quiet, ...)
  font <- ifelse(is.na(font), "", font)
  ret <- ret + theme(text = element_text(family = font))

  # Set some defaults
  update_geom_defaults("text", list(colour = hard_colour, family = font))
  update_geom_defaults("label", list(colour = hard_colour, family = font))
  update_geom_defaults("point", list(colour = hard_colour, alpha = 0.95))
  update_geom_defaults("line", list(colour = hard_colour, alpha = 0.95))
  update_geom_defaults("area", list(fill = main_colour, alpha = 0.95))
  update_geom_defaults("rect", list(fill = main_colour, alpha = 0.95))
  update_geom_defaults("density", list(fill = main_colour, alpha = 0.95))
  update_geom_defaults("bar", list(fill = main_colour, alpha = 0.95))
  update_geom_defaults("col", list(fill = main_colour, alpha = 0.95))
  update_geom_defaults("boxplot", list(fill = main_colour, alpha = 0.9))
  # update_geom_defaults("text_repel", list(family = font))

  # Legend
  aux <- ifelse("top" %in% legend, "right", "left")
  xj <- switch(tolower(substr(aux, 1, 1)),
    b = 0,
    l = 0,
    m = 0.5,
    c = 0.5,
    r = 1,
    t = 1
  )
  yj <- switch(tolower(substr(aux, 2, 2)),
    b = 0,
    l = 0,
    m = 0.5,
    c = 0.5,
    r = 1,
    t = 1
  )
  if (!is.null(legend)) {
    ret <- ret + theme(
      legend.title = element_text(color = soft_colour, size = size * 0.9, face = "bold"),
      legend.position = legend,
      legend.justification = c(
        ifelse(legend %in% c("top", "bottom"), 0, .5),
        ifelse(legend == "top", 0, ifelse(legend %in% "left", 1, .5))
      ),
      legend.margin = margin(-3, 0, -4, 0)
    )
  }

  if (inherits(grid, "character") || grid) {
    grid_col <- "#CCCCCC"
    ret <- ret + theme(panel.grid = element_line(color = grid_col, linewidth = 0.2))
    ret <- ret + theme(panel.grid.major = element_line(color = grid_col, linewidth = 0.1))
    ret <- ret + theme(panel.grid.minor = element_line(color = grid_col, linewidth = 0.05))
    if (inherits(grid, "character")) {
      if (regexpr("X", grid)[1] < 0) ret <- ret + theme(panel.grid.major.x = element_blank())
      if (regexpr("Y", grid)[1] < 0) ret <- ret + theme(panel.grid.major.y = element_blank())
      if (regexpr("x", grid)[1] < 0) ret <- ret + theme(panel.grid.minor.x = element_blank())
      if (regexpr("y", grid)[1] < 0) ret <- ret + theme(panel.grid.minor.y = element_blank())
    }
  } else {
    ret <- ret + theme(panel.grid = element_blank())
  }

  # Axis lines
  ret <- ret + theme(axis.line = element_blank())

  # Axis ticks
  ret <- ret + theme(
    axis.ticks = element_blank(),
    axis.ticks.x = element_blank(),
    axis.ticks.y = element_blank()
  )

  # Axis text
  ret <- ret + theme(axis.text.x = element_text(
    size = size * 0.8, margin = margin(t = 0), colour = soft_colour
  ))
  ret <- ret + theme(axis.text.y = element_text(
    size = size * 0.8, margin = margin(r = 0), colour = soft_colour
  ))

  # Axis titles
  ret <- ret + theme(axis.title = element_text(
    size = size * 0.85, family = font, colour = soft_colour
  ))
  ret <- ret + theme(axis.title.x = element_text(
    hjust = xj, size = size * 0.85, family = font, face = "bold", colour = soft_colour
  ))
  ret <- ret + theme(axis.title.y = element_text(
    hjust = yj, size = size * 0.85, colour = soft_colour, family = font, face = "bold"
  ))
  ret <- ret + theme(axis.title.y.right = element_text(
    hjust = yj, size = size * 0.85, angle = -90, colour = soft_colour, family = font, face = "bold"
  ))

  # Suppress axis
  if (axis == FALSE) {
    ret <- ret +
      theme(
        axis.text.x = element_blank(), axis.text.y = element_blank(),
        axis.title.x = element_blank(), axis.title.y = element_blank()
      )
  }
  if (is.character(axis)) {
    if (tolower(axis) == "x") {
      ret <- ret + theme(
        axis.text.y = element_blank(), axis.title.y = element_blank()
      )
    }
    if (tolower(axis) == "y") {
      ret <- ret + theme(
        axis.text.x = element_blank(), axis.title.x = element_blank()
      )
    }
  }

  # facet_grid
  ret <- ret + theme(strip.text = element_text(
    hjust = 0, size = size * 0.9, colour = soft_colour, face = "bold", family = font
  ))
  ret <- ret + theme(panel.spacing = grid::unit(0.8, "lines"))
  if (no_facets) {
    ret <- ret + theme(strip.background = element_blank(), strip.text = element_blank())
  }

  # Plot title
  ret <- ret + theme(plot.title = element_text(
    size = size * 1.25, margin = margin(b = size * 0.3),
    family = font, face = "bold", color = "black"
  ))
  # Align plot title to the very left edge (more space) [ggplot2 >= 3.3.0]
  ggc <- stringr::str_split(as.character(packageVersion("ggplot2")), "\\.")[[1]]
  if (ggc[1] >= 3 && ggc[2] >= 3) {
    ret <- ret + theme(plot.title.position = "plot")
  }

  # Plot subtitle
  ret <- ret + theme(plot.subtitle = element_text(
    hjust = 0, size = size * 1.1, colour = soft_colour,
    margin = margin(b = size * 0.5), family = font, face = "italic"
  ))

  # Caption
  ret <- ret + theme(plot.caption = element_text(
    hjust = 1, size = size * 0.85, margin = margin(t = size * 0.9),
    family = font, color = soft_colour
  ))

  # Background
  ret <- ret + theme(
    panel.background = element_rect(fill = panel_colour, colour = NA),
    plot.background = element_rect(fill = plot_colour, colour = NA)
  )

  # External margins
  ret <- ret + theme(plot.margin = margin(mg, mg, mg, mg))

  # Colour Palette
  if (!is.null(palette)) {
    if (is.null(names(palette))) {
      names(palette) <- as.vector(palette)
    }
    colours_pal <- palette
  } else {
    colours_pal <- lares_pal()$palette
  }

  # Palette with fills and colour
  if (pal == 1) {
    ret <- list(ret, scale_fill_manual(values = names(colours_pal)))
    ret <- list(ret, scale_colour_manual(values = as.vector(colours_pal)))
  }

  # Palette without fills
  if (pal == 2) {
    ret <- list(ret, scale_colour_manual(values = names(colours_pal)))
  }

  # Palette without fills
  if (pal == 3) {
    ret <- list(ret, scale_fill_manual(values = names(colours_pal)))
  }

  # Custom palette defined in colour_palettes.R and/or lares.colours.custom options
  if (pal == 4) {
    which <- tolower(which)
    if ((grepl("c", which) && grepl("t", which))) {
      stop("In your 'which' parameter, pass only 'c' OR 't', not both")
    }
    # FIX: Scale for 'fill' is already present. Adding another scale for 'fill',
    # which will replace the existing scale. (not being suppressed)
    suppressMessages({
      ret <- list(ret)
      if (grepl("f", which)) ret <- append(ret, gg_fill_customs())
      if (grepl("c", which)) ret <- append(ret, gg_colour_customs())
      if (grepl("t", which)) ret <- append(ret, gg_text_customs())
    })
  }
  ret
}

####################################################################
#' Custom fill, colour and text colours for ggplot2
#'
#' This function lets the user use pre-defined default colours.
#' Check your \code{lares_pal()$labels} scale. Feel free to use
#' \code{gg_vals()} to debug colours used in latest plot.
#'
#' Notice that when the layer defined is any of GeomPoint, GeomLine,
#' GeomText or GeomLabel, \code{gg_colour_customs()} will force
#' \code{column = "fill"} parameter.
#'
#' @family Themes
#' @param column Character. Select any of "fill" or "colour" to use on
#' your \code{lares_pal()$labels} palette.
#' @param ... Allow additional parameters not used.
#' @examples
#' library("ggplot2")
#' # Generic plot function to run examples to
#' run_plot <- function(add_fxs = TRUE) {
#'   p <- data.frame(station = c("spring", "summer", "fall", "winter"), num = 1:4) %>%
#'     ggplot(aes(x = station, y = num, fill = station)) +
#'     geom_col() +
#'     geom_text(aes(y = 0.5, label = num, colour = station), size = 6)
#'   if (add_fxs) p <- p + gg_fill_customs() + gg_colour_customs()
#'   return(p)
#' }
#' # Default colours
#' run_plot()
#' # Check last colours used
#' gg_vals("fill", "fill")
#' gg_vals("colour", "colour")
#' # Change any default colour
#' options("lares.colours.custom" = data.frame(
#'   values = c("summer", "winter"),
#'   fill = c("pink", "black"),
#'   colour = c("black", "white")
#' ))
#' run_plot()
#' # Check last colours used
#' gg_vals("fill", "fill")
#' gg_vals("colour", "colour")
#' # Reset to default colours
#' options("lares.colours.custom" = NULL)
#' # Notice you can use 'pal = 4' argument on theme_lares() too
#' run_plot(add_fxs = FALSE) + theme_lares(pal = 4)
#' @return Same as \code{scale_fill_manual} or \code{scale_colour_manual}
#' but with custom palette applied.
#' @export
gg_fill_customs <- function(column = "fill", ...) {
  scale_fill_manual(values = gg_vals("fill", column, ...), aesthetics = "fill")
}

#' @rdname gg_fill_customs
#' @export
gg_colour_customs <- function(column = "colour", ...) {
  scale_color_manual(values = gg_vals("colour", column, ...), aesthetics = "colour")
}

#' @rdname gg_fill_customs
#' @export
gg_text_customs <- function(column = "colour", ...) {
  scale_color_manual(values = gg_vals("label", column, ...), aesthetics = "colour")
}

#' @rdname gg_fill_customs
#' @param layer Character. Select any of "fill", "colour", or "label" to get the
#' layer containing the colours labels you wish to colour.
#' @param cols Data.frame. Customize colour palette with a data.frame. Must
#' contain values, fill, and colour columns.
#' @export
gg_vals <- function(layer = "fill", column = layer, cols = NULL, ...) {
  check_opts(layer, c("fill", "colour", "label"))
  check_opts(column, c("fill", "colour"))
  x <- last_plot()
  if (!is.null(cols)) {
    stopifnot(is.data.frame(cols))
    check_opts(colnames(cols), c("values", "fill", "colour"))
  } else {
    cols <- lares_pal()$labels
  }
  # Get colours present in data
  labs <- unlist(lapply(x$layers, function(y) as_label(y$mapping[[layer]])))
  labs <- c(labs, as_label(x$mapping[[layer]]))
  labs <- labs[!labs %in% c("NULL", "<uneval>")]
  cols <- cols[cols$values %in% unique(unlist(select(x$data, any_of(labs)))), ]
  # If point, line, text, label, force using fill colours
  invert <- c("GeomPoint", "GeomLine", "GeomText", "GeomLabel")
  layers_present <- unique(unlist(lapply(x$layers, function(y) class(y$geom))))
  if (any(invert %in% layers_present)) column <- "fill"
  # Final values vector
  values <- as.character(t(cols[, column])[1, ])
  names(values) <- cols$values
  values
}

.font_global <- function(font, quiet = FALSE, when_not = NA, ask_install = FALSE, ...) {
  if (is.null(font) || any(c(NA, "ignore") %in% tolower(font))) {
    ""
  } else {
    temp <- font_exists(font, quiet = quiet, ...)
    if (!any(isTRUE(temp))) {
      if (isFALSE(is.na(font[1]))) {
        if (isTRUE(font[1] != "") && !quiet) {
          if (ask_install & font[1] %in% list.files(system.file("fonts", package = "lares"))) {
            yes <- readline(sprintf("Do you want to install %s font for better results? [y/n]: ", font))
            if ("y" %in% yes) try(install_localfont(font))
          } else {
            warning(sprintf("Font(s) %s not installed, with other name, or can't be found", v2t(font)))
          }
        }
        Sys.unsetenv("LARES_FONT") # So R doesn't try again by default
        font <- when_not
      }
    } else {
      # Return first one that is found
      font <- font[head(which(temp), 1)]
    }
    font
  }
}

install_localfont <- function(
    font, dir = system.file(paste0("fonts/", gsub(" ", "_", font)), package = "lares"),
    ...) {
  try_require("extrafont")
  font_import(dir, prompt = FALSE, ...)
  # loadfonts(quiet = TRUE)
}

####################################################################
#' Split ICS File
#'
#' This function splits a large ICS calendar file into smaller files.
#'
#' @family Tools
#' @param file Character. Path to the ICS file to split.
#' @param n_splits Integer. Number of files to split into.
#' @param prefix Character. Prefix for the output filenames.
#' @param dir Character. Directory to save the output files.
#' Defaults to same directory as input file.
#' @return Invisible list. The paths of the created files.
#' @export
cal_split <- function(file, n_splits = 5, prefix = "calendar_split_", dir = NULL) {
  
  if (!file.exists(file)) stop("File not found: ", file)
  if (is.null(dir)) dir <- dirname(file)
  lines <- readLines(file, warn = FALSE)

  # 1. Extract the header (lines before the first VEVENT)
  header_end <- grep("BEGIN:VEVENT", lines)[1] - 1
  if (is.na(header_end)) {
    stop("No 'BEGIN:VEVENT' found in the file. Is this a valid ICS file?")
  }
  calendar_header <- lines[1:header_end]

  # 2. Extract the footer (usually "END:VCALENDAR")
  calendar_footer <- "END:VCALENDAR"

  # 3. Identify start and end positions for every event block
  event_starts <- grep("BEGIN:VEVENT", lines)
  event_ends <- grep("END:VEVENT", lines)

  if (length(event_starts) != length(event_ends)) {
    stop("Mismatch between 'BEGIN:VEVENT' and 'END:VEVENT' counts.")
  }

  if (length(event_starts) == 0) {
    stop("No events found to split.")
  }

  # Create a list where each element is a full VEVENT block
  events <- lapply(seq_along(event_starts), function(i) {
    lines[event_starts[i]:event_ends[i]]
  })

  # 4. Calculate how many events go into each of the n_splits files
  num_events <- length(events)
  events_per_file <- ceiling(num_events / n_splits)

  message(sprintf(
    "Splitting %s events into %s files (approx %s events/file)...",
    formatNum(num_events, 0), n_splits, events_per_file
  ))

  # 5. Split and save the files
  output_files <- c()
  for (i in 1:n_splits) {
    # Determine range of events for this specific file
    start_idx <- ((i - 1) * events_per_file) + 1
    end_idx <- min(i * events_per_file, num_events)

    if (start_idx <= num_events) {
      # Combine Header + Subset of Events + Footer
      current_content <- c(
        calendar_header,
        unlist(events[start_idx:end_idx]),
        calendar_footer
      )

      # Write to a new file
      output_filename <- paste0(prefix, i, ".ics")
      output_path <- file.path(dir, output_filename)
      writeLines(current_content, output_path)
      output_files <- c(output_files, output_path)
    }
  }

  # Return the list of created files
  invisible(output_files)
}

####################################################################
#' File Encryption and Decryption (AES-256-CBC)
#'
#' @rdname encrypt_file
#' @description This set of functions provides utilities for encrypting
#' and decrypting files using AES-256 in CBC mode, primarily for
#' handling key-value pair "secrets" files.
#' @details The encryption process pads the data, generates a random
#' Initialization Vector (IV), and encrypts the content. The IV is
#' prepended to the encrypted data in the output file.
#' Decryption involves extracting the IV, decrypting, and unpadding the data.
#' A specialized decryption function \code{read_encrypted} is provided
#' to directly parse decrypted content (assumed to be \code{key=value} pairs)
#' into a named R list. If no "=" sign found in the first line, text will be
#' imported as character vector.
#' @keywords Encryption Decryption Security
#' @family Credentials
#' @param input_file Character string. The path to the file to be encrypted.
#' @param output_file Character string. The path where the encrypted
#' file will be saved.
#' @param key A raw vector of 32 bytes (256 bits) to be used as the encryption key.
#' @return No return value, called for side effects (writes to \code{output_file}).
#' @examples
#' \dontrun{
#' # Create dummy files for demonstration
#' temp_input <- tempfile(fileext = ".txt")
#' temp_output <- tempfile(fileext = ".enc")
#' writeLines("library=lares \nversion=5.3.0 \n pass=123", temp_input)
#' readLines(temp_input)
#'
#' # Generate a random 32-byte key (and save it in a secure place)
#' raw_key <- openssl::rand_bytes(32)
#' # You can convert from hex to raw and viceversa
#' raw_to_hex(raw_key)
#' hex_to_raw(raw_to_hex(raw_key))
#'
#' # Encrypt the file
#' encrypt_file(temp_input, temp_output, raw_key)
#'
#' # Import the data from encrypted file to list or vector
#' secrets <- read_encrypted(temp_output, raw_key)
#' unlist(secrets)
#'
#' # Example using a string (JSON in this case)
#' writeLines(jsonlite::toJSON(list(a = 1, list(b = 2, c = 1:3))), temp_input)
#' encrypt_file(temp_input, temp_output, raw_key)
#' read_encrypted(temp_output, raw_key)
#'
#' # Example writing a file from a list or vector directly from R
#' my_secrets <- list(
#'   api_key = "some_secret_api_key_123",
#'   username = "data_analyst",
#'   server = "production.server.com"
#' )
#' raw_key <- write_encrypted(my_secrets, temp_output, quiet = FALSE)
#' fromJSON(read_encrypted(temp_output, key = raw_key))
#' }
#' @export
encrypt_file <- function(input_file, output_file, key) {
  try_require("openssl")
  # Check key type
  if (!"raw" %in% class(key)) key <- hex_to_raw(key)
  # Read binary file
  data <- readBin(input_file, what = "raw", n = file.info(input_file)$size)
  # Pad the data
  padded_data <- pad(data)
  # Generate random IV
  iv <- rand_bytes(16)
  # Encrypt using AES-256-CBC
  encrypted <- aes_cbc_encrypt(padded_data, key = key, iv = iv)
  # Write IV + encrypted data to output file
  writeBin(c(iv, encrypted), output_file)
}

#' Decript file to a list
#' @rdname encrypt_file
#' @export
read_encrypted <- function(input_file, key) {
  try_require("openssl")
  # Check key type
  if (!"raw" %in% class(key)) key <- hex_to_raw(key)
  # Read binary file
  encrypted_data <- readBin(input_file, what = "raw", n = file.info(input_file)$size)
  # Extract IV (first 16 bytes) and encrypted content (remaining bytes)
  iv <- encrypted_data[1:16]
  encrypted_content <- encrypted_data[-(1:16)]
  # Decrypt using AES-256-CBC
  decrypted <- aes_cbc_decrypt(encrypted_content, key = key, iv = iv)
  # Remove padding
  unpadded_data <- unpad(decrypted)
  # Convert raw decrypted data into character format
  decrypted_text <- rawToChar(unpadded_data)
  # Split decrypted content into lines
  parsed_list <- lines <- strsplit(decrypted_text, "\n")[[1]]
  # Parse lines into key-value pairs
  if (grepl("=", lines[1])) {
    key_value_pairs <- lapply(lines, function(line) {
      parts <- strsplit(line, "=")[[1]]
      list(key = parts[1], value = parts[2])
    })
    # Create a named list with values and keys
    parsed_list <- as.list(setNames(
      trimws(lapply(key_value_pairs, `[[`, "value")),
      trimws(sapply(key_value_pairs, `[[`, "key"))
    ))
  }
  return(parsed_list)
}

#' Write a vector or list into an encoded file in JSON format
#' @inheritParams mp3_get
#' @param data List, vector or any json-able object to write into encrypted file.
#' @rdname encrypt_file
#' @export
write_encrypted <- function(data, output_file = "encrypted.enc", key = NULL, quiet = FALSE) {
  try_require("openssl")
  # 1. Convert R object to JSON string
  json_string <- jsonlite::toJSON(data, auto_unbox = TRUE, pretty = FALSE)
  # 2. Convert JSON string to raw vector
  data_raw <- charToRaw(json_string)
  # 3. Generate a new 32-byte (256-bit) AES key
  if (is.null(key)) {
    key <- rand_bytes(32)
  } else {
    if ("character" %in% class(key)) key <- hex_to_raw(key)
    quiet <- TRUE
  }
  # 4. Pad the raw data
  padded_data <- pad(data_raw)
  # 5. Generate random IV
  iv <- rand_bytes(16)
  # 6. Encrypt using AES-256-CBC
  encrypted <- aes_cbc_encrypt(padded_data, key = key, iv = iv)
  # 7. Write IV + encrypted data to output file
  writeBin(c(iv, encrypted), output_file)
  # 8. Return the hexadecimal key string
  if (!quiet) message("KEY: ", raw_to_hex(key))
  return(key)
}

# Convert Hexadecimal String to Raw Vector
#' @rdname encrypt_file
#' @param x Character string or hex vector to convert.
#' @export
hex_to_raw <- function(x) {
  try_require("openssl")
  stopifnot("character" %in% class(x))
  # Split hexadecimal string into pairs of two characters
  hex_pairs <- substring(
    x, seq(1, nchar(x), by = 2),
    seq(2, nchar(x), by = 2)
  )
  raw_vector <- as.raw(strtoi(hex_pairs, base = 16))
  return(raw_vector)
}

# Convert Raw Vector to Hexadecimal String
#' @rdname encrypt_file
#' @export
raw_to_hex <- function(x) {
  stopifnot("raw" %in% class(x))
  paste0(as.character(x), collapse = "")
}

# Pad data to 16-byte AES block size
pad <- function(data) {
  pad_len <- 16 - (length(data) %% 16)
  c(data, rep(as.raw(pad_len), pad_len))
}

# Remove padding from decrypted data
unpad <- function(data) {
  pad_len <- as.integer(data[length(data)])
  data[1:(length(data) - pad_len)]
}

# Import decrypted file and convert to JSON
convert_to_list <- function(decrypted_file) {
  # Read the decrypted file
  lines <- readLines(decrypted_file)
  # Split each line into key-value pairs
  key_value_pairs <- lapply(lines, function(line) {
    parts <- strsplit(line, "=")[[1]]
    list(key = parts[1], value = parts[2])
  })
  # Create a named list or vector
  parsed_list <- setNames(
    trimws(lapply(key_value_pairs, `[[`, "value")), # Extract values
    trimws(sapply(key_value_pairs, `[[`, "key")) # Extract keys
  )
  return(parsed_list)
}

####################################################################
#' Most Relevant Features Using Lasso Regression
#'
#' Use Lasso regression to identify the most relevant variables that
#' can predict/identify another variable. You might want to compare
#' with \code{corr_var()} and/or \code{x2y()} results to compliment
#' the analysis No need to standardize, center or scale your data.
#' Tidyverse friendly.
#'
#' @family Machine Learning
#' @family Exploratory
#' @inheritParams mp3_get
#' @param df Dataframe. Any dataframe is valid as \code{ohse} will be applied to
#' process categorical values, and values will be standardize automatically.
#' @param variable Variable. Dependent variable or response.
#' @param ignore Character vector. Variables to exclude from study.
#' @param nlambdas Integer. Number of lambdas to be used in a search.
#' @param nfolds Integer. Number of folds for K-fold cross-validation (>= 2).
#' @param top Integer. Plot top n results only.
#' @param seed Numeric.
#' @param ... Additional parameters passed to \code{ohse()}.
#' @return List. Contains lasso model coefficients, performance metrics, the
#' actual model fitted and a plot.
#' @examples
#' \dontrun{
#' # CRAN
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#'
#' m <- lasso_vars(dft, Survived, ignore = c("Cabin"))
#' print(m$coef)
#' print(m$metrics)
#' plot(m$plot)
#' }
#' @export
lasso_vars <- function(df, variable,
                       ignore = NULL,
                       nlambdas = 100,
                       nfolds = 10,
                       top = 20,
                       quiet = FALSE,
                       seed = 123, ...) {
  tic("lasso_vars")
  try_require("h2o")
  quiet(h2o.init(nthreads = -1, port = 54321))
  h2o.no_progress()
  on.exit(set.seed(seed))
  var <- enquo(variable)
  df <- select(df, !!var, everything())
  if (!is.null(ignore)) ignore <- c(ignore, as_label(var))

  # Run one-hot-smart-encoding
  temp <- data.frame(as.vector(df[, 1]), ohse(df[, -1], ignore = ignore, quiet = quiet, ...))
  colnames(temp)[1] <- "main"

  temp <- temp %>%
    filter(!is.na(.data$main)) %>%
    mutate(main = as.vector(base::scale(as.numeric(.data$main), scale = FALSE, center = TRUE)))

  if (!quiet) message(">>> Searching for optimal lambda with CV...")
  lasso_logistic <- h2o.glm(
    alpha = 1,
    seed = seed,
    y = "main",
    nfolds = nfolds,
    training_frame = as.h2o(temp),
    lambda_search = TRUE,
    nlambdas = nlambdas,
    standardize = TRUE
  )
  if (!quiet) message("Found best lambda: ", signif(lasso_logistic@model$lambda_best, 5))
  if (!quiet) message(">>> Fetching most relevant variables...")
  t_lasso_model_val <- h2o.glm(
    y = "main",
    training_frame = as.h2o(temp),
    alpha = 1,
    lambda = lasso_logistic@model$lambda_best
  )
  t_pred <- h2o.predict(t_lasso_model_val, as.h2o(temp))

  # R-squared, values closer to 1 are the best (rsq)
  rsq <- model_metrics(tag = temp$main, score = as.vector(t_pred[, 1]), thresh = 1)
  rsq$metrics$bestlambda <- lasso_logistic@model$lambda_best

  t_lasso_model_coeff <- lasso_logistic@model$coefficients_table %>%
    arrange(desc(abs(.data$standardized_coefficients))) %>%
    # To avoid numbers like -0.000000000000000364
    mutate(standardized_coefficients = as.numeric(ifelse(
      .data$names == "Intercept", 0, .data$standardized_coefficients
    )))

  t_lasso_model_coeff <- mutate(
    t_lasso_model_coeff,
    abs = abs(.data$standardized_coefficients),
    prc = .data$abs / sum(.data$abs),
    coef = ifelse(.data$coefficients > 0, "positive", "negative")
  ) %>%
    filter(.data$prc > 0)
  if (!quiet) message(">>> Generating plots for ", as_label(var), "...")
  if (nrow(t_lasso_model_coeff) > top && !quiet) {
    message(paste("- Plotting only the", top, "most relevant features..."))
  }

  p <- t_lasso_model_coeff %>%
    head(top) %>%
    ggplot(aes(
      x = reorder(.data$names, .data$prc),
      y = abs(.data$standardized_coefficients)
    )) +
    geom_col(aes(fill = .data$coef)) +
    coord_flip() +
    labs(
      x = NULL, y = "Absolute Standarized Coefficient",
      title = "Most Relevant Features (Lasso Regression)",
      subtitle = sprintf(
        "Response variable: %s | RSQ = %s",
        as_label(var), round(rsq$metrics$rsq, 4)
      ),
      fill = "Coeff > 0"
    ) +
    theme_lares(legend = "none", pal = 4) +
    scale_y_percent(expand = c(0, 0), position = "right")

  toc("lasso_vars", quiet = quiet)

  list(
    coef = as_tibble(t_lasso_model_coeff),
    metrics = as_tibble(rsq$metrics),
    model = invisible(t_lasso_model_val),
    plot = p
  )
}

####################################################################
#' Get Tweets
#'
#' This function downloads tweets with personal credentials
#'
#' @family Credentials
#' @family Twitter
#' @param q Query. Check for ?rtweet::search_tweets()
#' @param n Integer. Total of tweets to return
#' @param creds Character. Credential's user (see \code{get_creds()})
#' @return data.frame with API response results.
#' @export
get_tweets <- function(q, n = 10000, creds = NA) {
  try_require("rtweet")
  c <- get_credentials(from = "twitter", dir = creds)
  token <- create_token(
    app = c$app,
    consumer_key = c$consumer_key,
    consumer_secret = c$consumer_secret,
    access_token = c$access_token,
    access_secret = c$access_secret
  )
  search_tweets(q = as.character(q), n = n, retryonratelimit = T)
}

#' Lean Analytics and Robust Exploration Sidekick
#'
#' R library for better/faster analytics, visualization, data mining, and machine learning tasks.
#'
#' @md
#' @name lares
#' @docType package
#' @author Bernardo Lares (laresbernardo@@gmail.com)
#' @importFrom dplyr %>% all_of any_of arrange as_tibble bind_cols bind_rows case_when
#' contains count desc distinct distinct_at everything filter first group_by inner_join
#' lag last left_join mutate mutate_all mutate_at mutate_if n ntile ensym across
#' one_of pull rename rename_at right_join row_number rowwise sample_n matches sym
#' select select_if slice slice_head starts_with summarise summarise_all case_match
#' summarise_if summarize summarize_all tally tibble transmute_if ungroup vars
#' @importFrom ggplot2 aes after_stat annotate coord_equal coord_flip
#' element_blank element_line element_rect element_text enquo enquos expand_limits
#' facet_grid facet_wrap geom_abline geom_area geom_bar geom_boxplot
#' geom_col geom_density geom_hline geom_jitter geom_label geom_line
#' geom_path geom_point geom_polygon geom_raster geom_rect geom_segment
#' geom_smooth geom_text geom_tile geom_vline ggplot ggsave ggtitle
#' guide_colorbar guide_colourbar guide_legend guides labs last_plot
#' margin position_dodge position_fill position_jitter position_stack
#' quo_name quos rel scale_color_brewer scale_color_identity
#' scale_color_manual scale_colour_brewer scale_colour_continuous
#' scale_colour_gradient scale_colour_identity scale_colour_manual
#' scale_fill_brewer scale_fill_continuous scale_fill_gradient
#' scale_fill_grey scale_fill_identity scale_fill_manual
#' scale_shape_identity scale_size scale_x_continuous scale_x_date
#' scale_x_discrete scale_x_reverse scale_y_continuous sec_axis
#' stat_density_2d stat_ecdf theme theme_minimal unit
#' update_geom_defaults xlab xlim ylab ylim
#' @importFrom graphics box hist plot points rasterImage rect grid legend mtext
#' @importFrom grDevices graphics.off dev.off png dev.size
# @importFrom h2o as.h2o h2o.automl h2o.accuracy h2o.getModel h2o.performance h2o.init
# h2o.removeAll h2o.download_mojo h2o.download_pojo h2o.loadModel h2o.no_progress h2o.predict
# h2o.predict_json h2o.saveModel h2o.varimp h2o.getVersion h2o.glm
# predict_contributions.H2OModel h2o.import_mojo h2o.no_progress h2o.show_progress
#' @importFrom httr GET POST oauth_endpoint oauth_app oauth1.0_token authenticate
#' stop_for_status upload_file add_headers content http_error set_config config
#' @importFrom jsonlite fromJSON toJSON flatten
#' @importFrom lubridate date day week weeks month year wday dmy_hms dmy ymd_hms ymd days
#' minute hour second %m+% %m-% floor_date ceiling_date years
#' @importFrom openxlsx addWorksheet copyWorkbook loadWorkbook read.xlsx removeWorksheet
#' getSheetNames renameWorksheet saveWorkbook sheets write.xlsx
#' @importFrom patchwork guide_area plot_layout plot_annotation wrap_plots
#' @importFrom pROC roc ci
#' @importFrom rlang as_label .data
#' @importFrom rpart rpart rpart.control
#' @importFrom rpart.plot rpart.rules rpart.plot
#' @importFrom rvest html_node html_nodes html_attrs html_attr html_table html_text
#' @importFrom stats cor quantile complete.cases na.omit sd median dist end lm predict runif
#' reorder start kmeans var xtabs as.formula prcomp p.adjust pt model.matrix qt cor.test setNames
#' @importFrom stringr fixed str_count str_length str_pad str_replace_all str_split
#' str_to_title word
#' @importFrom tidyr gather pivot_longer pivot_wider spread unite
#' @importFrom utils head tail packageVersion URLencode capture.output data download.file
#' globalVariables installed.packages write.table install.packages remove.packages object.size
#' type.convert flush.console read.table modifyList write.csv combn browseURL type.convert
#' getParseData find read.delim read.csv packageDescription
#' @importFrom yaml read_yaml
"_PACKAGE"

####################################################################
#' Install/Update Additional Recommended Libraries
#'
#' All needed libraries to use (most) lares are already a dependency.
#' There are some functions that many people won't event know exist
#' that will require other additional libraries. Also, this may be
#' used as a Docker way of installing useful libraries on an new instance.
#'
#' @param progress Boolean. Show status bar?
#' @param all Boolean. All packages? If not, only the ones not installed yet.
#' @export
install_recommended <- function(progress = TRUE, all = FALSE) {
  pkgs <- names(recommended)
  these <- if (!all) pkgs[!pkgs %in% installed.packages()[, 1]] else pkgs
  message(sprintf("Installing %s packages: %s", length(these), v2t(these)))
  for (lib in these) {
    invisible(install.packages(lib, quiet = TRUE, verbose = FALSE))
    if (progress) {
      statusbar(which(lib == these), length(these), lib, msg = "")
    }
  }
}

# Recommended additional libraries to fully take advantage of lares library
recommended <- list(
  base64enc = "base64encode",
  beepr = "beep",
  circlize = c("chordDiagram", "uh"),
  DALEX = c("explain.default", "model_performance", "model_profile", "predict_parts"),
  DBI = c("dbDriver", "dbConnect", "dbSendQuery", "fetch", "dbDisconnect"),
  devtools = c("install", "install_github", "with_proxy"),
  exifr = "read_exif",
  extrafont = c("font_import", "loadfonts"),
  factoextra = "fviz_nbclust",
  forecast = c("Arima", "auto.arima", "forecast", "accuracy"),
  gdata = "read.xls",
  ggbeeswarm = "geom_quasirandom",
  # ggforce = "geom_mark_ellipse",
  # ggrepel = "geom_label_repel",
  googleAnalyticsR = "google_analytics",
  googleAuthR = "gar_auth",
  googledrive = c("drive_auth", "drive_find", "local_drive_quiet"),
  googlesheets4 = c(
    "sheets_auth", "read_sheet", "gs4_create", "range_write",
    "gs4_auth_configure", "gs4_auth", "sheet_append"
  ),
  grDevices = "windowsFonts",
  h2o = c(
    "as.h2o", "h2o.automl", "h2o.accuracy", "h2o.getModel", "h2o.performance",
    "h2o.init", "h2o.removeAll", "h2o.download_mojo", "h2o.download_pojo",
    "h2o.loadModel", "h2o.no_progress", "h2o.predict", "h2o.predict_json",
    "h2o.saveModel", "h2o.varimp", "h2o.getVersion", "h2o.glm",
    "predict_contributions.H2OModel", "h2o.import_mojo", "h2o.no_progress",
    "h2o.show_progress"
  ),
  KeyboardSimulator = c("mouse.click", "mouse.get_cursor", "mouse.move"),
  knitr = c("kable", "knit"),
  methods = "as",
  mice = c("mice", "complete"),
  nloptr = "nloptr",
  openssl = c("rand_bytes", "aes_cbc_encrypt", "aes_cbc_decrypt"),
  optimParallel = "optimParallel",
  parallel = c("detectCores", "makeCluster", "setDefaultCluster", "stopCluster"),
  plotly = c("ggplotly", "plot_ly", "add_markers", "add_markers"),
  prophet = c(
    "prophet", "fit.prophet", "prophet_plot_components",
    "add_country_holidays", "make_future_dataframe"
  ),
  quantmod = c("getDividends", "getSymbols", "getQuote"),
  rdrop2 = c("drop_auth", "drop_dir", "drop_download", "drop_search", "drop_upload"),
  reticulate = c("import", "tuple"),
  rgdal = c("readOGR", "project"),
  Robyn = c("robyn_allocator", "robyn_immcarr", "robyn_response"),
  RPostgreSQL = "PostgreSQL",
  Rtsne = "Rtsne",
  rtweet = c("create_token", "search_tweets"),
  rmarkdown = "render",
  skimr = "skim",
  spotifyr = c("get_album", "get_spotify_access_token", "search_spotify"),
  syuzhet = "get_sentiment_dictionary",
  # sp = c("CRS", "over", "coordinates", "proj4string", "proj4string<-", "coordinates<-", "spTransform"),
  threed = c("mesh3dobj", "transform_by", "invert_matrix", "perspective_projection", "look_at_matrix"),
  tidytext = "unnest_tokens",
  tm = c(
    "content_transformer", "Corpus", "removeNumbers", "removePunctuation",
    "removeWords", "readPlain", "stopwords", "stripWhitespace", "TermDocumentMatrix",
    "tm_map", "VectorSource", "VCorpus"
  ),
  toon = "as_toon",
  udpipe = c("keywords_rake", "udpipe_annotate", "udpipe_download_model", "udpipe_load_model"),
  wordcloud = c("wordcloud", "textplot")
)

if (getRversion() >= "2.15.1") {
  globalVariables(c(as.vector(unlist(recommended)), "."))
}

####################################################################
#' Compare Variables with their Distributions
#'
#' Compare the distribution of a target variable vs another variable. This
#' function automatically splits into quantiles for numerical variables.
#' Custom and tidyverse friendly.
#'
#' @family Exploratory
#' @family Visualization
#' @param data Dataframe
#' @param ... Variables. Main (target variable) and secondary (values
#' variable) to group by (if needed).
#' @param type Integer. 1 for both plots, 2 for counter plot only, 3 for
#' percentages plot only.
#' @param ref Boolean. Show a reference line if levels = 2? Quite useful
#' when data is unbalanced (not 50/50) because a reference line is drawn.
#' @param note Character. Caption for the plot.
#' @param top Integer. Filter and plot the most n frequent for categorical values.
#' @param breaks Integer. Number of splits for numerical values.
#' @param na.rm Boolean. Ignore \code{NA}s if needed.
#' @param force Character. Force class on the values data. Choose between 'none',
#' 'character', 'numeric', 'date'
#' @param trim Integer. Trim labels until the nth character for categorical values
#' (applies for both, target and values)
#' @param clean Boolean. Use \code{cleanText()} for categorical values (applies
#' for both, target and values)
#' @param abc Boolean. Do you wish to sort by alphabetical order?
#' @param custom_colours Boolean. Use custom colours function?
#' @param plot Boolean. Return a plot? Otherwise, a table with results
#' @param chords Boolean. Use a chords plot?
#' @param save Boolean. Save the output plot in our working directory
#' @param subdir Character. Into which subdirectory do you wish to save the plot to?
#' @return Plot when \code{plot=TRUE} with two plots in one: counter distribution
#' grouped by cuts, and proportions distribution grouped by same cuts. data.frame when
#' \code{plot=FALSE} with counting, percentages, and cumulative percentages results.
#' When \code{type} argument is used, single plots will be returned.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#'
#' # Relation for categorical/categorical values
#' distr(dft, Survived, Sex)
#'
#' # Relation for categorical/numeric values
#' dft %>%
#'   distr(Survived, Fare, plot = FALSE) %>%
#'   head(10)
#' # Sort values
#' dft %>% distr(Survived, Fare, abc = TRUE)
#' # Less splits/breaks
#' dft %>% distr(Survived, Fare, abc = TRUE, breaks = 5)
#'
#' # Distribution of numerical only
#' dft[dft$Fare < 20, ] %>% distr(Fare)
#'
#' # Distribution of numerical/numerical
#' dft %>% distr(Fare, Age)
#'
#' # Select only one of the two default plots of distr()
#' dft %>% distr(Survived, Age, type = 2)
#' dft %>% distr(Survived, Age, type = 3)
#' @export
distr <- function(data, ...,
                  type = 1,
                  ref = TRUE,
                  note = NA,
                  top = 10,
                  breaks = 10,
                  na.rm = FALSE,
                  force = "none",
                  trim = 0,
                  clean = FALSE,
                  abc = FALSE,
                  custom_colours = FALSE,
                  plot = TRUE,
                  chords = FALSE,
                  save = FALSE,
                  subdir = NA) {
  # # To handle scientific notation inputs correctly
  # on.exit(options("scipen" = 999))

  data <- data.frame(data)

  vars <- enquos(...)
  var1 <- vars[[1]]
  if (length(vars) > 1) {
    var2 <- vars[[2]]
  } else {
    var2 <- NULL
  }

  # When we only have one variable, use freqs()
  if (length(vars) == 1) {
    value <- select(data, !!var1)
    variable_name <- colnames(value)
    value <- value[, 1]
    value <- .force_class(value, force)
    value <- .fxtrim(value, trim)
    value <- .fxclean(value, clean)

    df <- data.frame(value = value, dummy = 0)
    df <- .fxna_rm(df, na.rm)

    is.Date <- function(x) inherits(x, "Date")
    is.POSIXct <- function(x) inherits(x, "POSIXct")
    is.POSIXlt <- function(x) inherits(x, "POSIXlt")
    if (is.numeric(value) || is.Date(value) || is.POSIXct(value) || is.POSIXlt(value)) {
      # Continuous and date values
      if (is.numeric(value)) {
        p <- ggplot(df, aes(x = .data$value))
      } else {
        p <- ggplot(df, aes(x = date(.data$value)))
      }
      p <- p +
        geom_density(fill = "deepskyblue", alpha = 0.7, adjust = 1 / 3) +
        labs(
          y = NULL, x = NULL, fill = "Density",
          title = "Density Distribution",
          subtitle = paste("Variable:", variable_name),
          caption = paste("Obs:", formatNum(nrow(df), 0))
        ) +
        theme_lares()
      if (top != 10) {
        p <- p + xlim(0, top)
      }
    } else {
      # Discrete values
      p <- freqs(df, value, plot = TRUE, variable_name = variable_name, abc = abc, top = top)
    }
    # Return table with results?
    if (!plot) {
      df %>% freqs(value, top = top)
    } else {
      p
    }
  } else {
    # Check if secondary variable exists and fix if possible
    var <- gsub('"', "", as_label(var2))
    if (!var %in% colnames(data)) {
      msg <- paste("Not a valid input:", var, "was transformed or does not exist.")
      maybes <- colnames(data)[grepl(var, colnames(data))]
      if (length(maybes) > 0 && maybes[1] %in% colnames(data)) {
        message(paste0(
          "Maybe you meant one of: ", vector2text(maybes), ". ",
          "Automatically using '", maybes[1], "'"
        ))
        var2 <- quos(maybes[1])
        warning(msg)
      } else {
        stop(msg)
      }
    }

    targets <- select(data, !!var1)
    targets_name <- colnames(targets)
    targets <- targets[, 1]
    value <- select(data, !!var2)
    variable_name <- colnames(value)
    # Transformations
    value <- value[, 1] # do.call("c", value)
    value <- .force_class(value, force)
    value <- .fxtrim(value, trim)
    value <- .fxclean(value, clean)

    if (length(targets) != length(value)) {
      message("The targets and value vectors should be the same length.")
      stop(message(paste(
        "Currently, targets has", length(targets),
        "rows and value has", length(value)
      )))
    }

    # For num-num distributions or too many unique target variables
    if (length(unique(targets)) >= 8) {
      if (is.numeric(targets) && is.numeric(value)) {
        subtitle <- paste0(
          "Variables: ", variable_name, " vs. ", targets_name,
          ". Obs: ", formatNum(length(value), 0)
        )
        df <- data.frame(x = targets, y = value)
        df <- .fxna_rm(df, na.rm = TRUE)
        p <- df %>%
          ggplot(aes(x = .data$x, y = .data$y)) +
          stat_density_2d(aes(fill = after_stat(.data$level)), geom = "polygon") +
          labs(
            title = "2D Density Distribution",
            x = targets_name, y = variable_name,
            subtitle = subtitle
          ) +
          scale_x_comma() +
          scale_y_comma() +
          theme_lares()
        p
      } else {
        message("You should try a 'target' variable with max 8 different values.")
        message("Automatically trying a chords plot...")
        chords <- TRUE
      }
    } else if (chords) {
      df <- data.frame(value = value, targets = targets)
      output <- freqs(df, targets, value)
      if (!na.rm) {
        output <- output %>% replaceall(NA, "NA")
      }
      title <- "Frequency Chords Diagram"
      subtitle <- paste("Variables:", targets_name, "to", variable_name)
      if (!plot) {
        output
      } else {
        plot_chord(
          output$targets, output$value, output$n,
          mg = 13, title = title, subtitle = subtitle
        )
      }
    } else {
      # Only n numeric values, really numeric?
      if (is.numeric(value) && length(unique(value)) <= 8) {
        value <- .force_class(value, class = "char")
      }

      # Turn numeric variables into quantiles
      if (is.numeric(value)) {
        breaks <- ifelse(top != 10, top, breaks)
        value <- quants(value, breaks, return = "labels")
        cuts <- length(unique(value[!is.na(value)]))
        if (cuts != breaks) {
          message(paste(
            "When dividing", variable_name, "into", breaks, "quantiles,",
            cuts, "cuts/groups are possible."
          ))
        }
        top <- top + 1
      }

      # Finally, we have our data.frame
      df <- data.frame(targets = targets, value = value)
      df <- .fxna_rm(df, na.rm)

      # Captions for plots
      subtitle <- paste0(
        "Variables: ", targets_name, " vs. ", variable_name,
        ". Obs: ", formatNum(nrow(df), 0)
      )

      freqs <- df %>%
        group_by(.data$targets, .data$value) %>%
        count() %>%
        ungroup() %>%
        arrange(desc(.data$n)) %>%
        group_by(.data$value) %>%
        mutate(
          p = round(100 * .data$n / sum(.data$n), 2),
          pcum = cumsum(.data$p)
        ) %>%
        ungroup() %>%
        filter(!is.na(.data$value)) %>%
        mutate(
          row = row_number(),
          order = suppressWarnings(ifelse(
            grepl("\\(|\\)", .data$value),
            as.numeric(as.character(substr(gsub(",.*", "", .data$value), 2, 100))),
            .data$row
          ))
        )
      if (length(unique(value)) > top && !is.numeric(value)) {
        message(paste("Filtering the", top, "most frequent values. Use 'top' to overrule."))
        which <- freqs(df, .data$value) %>% slice(1:top)
        freqs <- freqs %>%
          mutate(value = ifelse(.data$value %in% which$value, as.character(.data$value), "OTHERS")) %>%
          group_by(.data$value, .data$targets) %>%
          select(-.data$row, -.data$order) %>%
          summarise(n = sum(.data$n)) %>%
          mutate(p = round(100 * n / sum(.data$n), 2)) %>%
          ungroup() %>%
          arrange(desc(.data$n)) %>%
          mutate(
            row = row_number(),
            order = row_number()
          )
      }

      # Sort values alphabetically or ascending if numeric
      if (abc) freqs <- mutate(freqs, order = rank(.data$value))

      # Counter plot
      if (type %in% c(1, 2)) {
        vadj <- ifelse(type == 1, -0.15, 0.5)
        hadj <- ifelse(type == 1, 0.5, -0.15)
        count <- ggplot(freqs, aes(
          x = reorder(as.character(.data$value), .data$order), y = .data$n,
          fill = as.character(.data$targets),
          label = formatNum(.data$n, 0), ymax = max(.data$n) * 1.1
        )) +
          geom_col(position = "dodge", colour = "transparent") +
          geom_text(
            colour = "black",
            check_overlap = TRUE,
            position = position_dodge(0.9),
            size = 3, vjust = vadj, hjust = hadj
          ) +
          labs(x = NULL, y = "Counter [#]", fill = targets_name, caption = note) +
          theme(legend.position = "top") +
          guides(colour = "none") +
          theme(axis.title.y = element_text(size = rel(0.8), angle = 90)) +
          scale_y_comma(expand = c(0, 0)) +
          theme_lares(pal = 1)
        # Give an angle to labels when more than...
        if (length(unique(value)) >= 7) {
          count <- count + theme(axis.text.x = element_text(angle = 30, hjust = 1))
        }
        # Custom colours if wanted...
        if (custom_colours) {
          count <- count + suppressWarnings(gg_fill_customs())
        }
      }

      # Proportions (%) plot
      if (type %in% c(1, 3)) {
        prop <- freqs %>%
          group_by(.data$value) %>%
          mutate(size = sum(.data$n) / sum(freqs$n)) %>%
          mutate(ptag = ifelse(p < 3, "", as.character(round(.data$p, 1)))) %>%
          ggplot(aes(
            x = reorder(.data$value, -.data$order),
            y = .data$p / 100, label = .data$ptag,
            fill = as.character(.data$targets)
          )) +
          geom_col(position = "fill", colour = "transparent") +
          geom_text(aes(size = .data$size, colour = as.character(.data$targets)),
            position = position_stack(vjust = 0.5)
          ) +
          scale_size(range = c(2.2, 3)) +
          coord_flip() +
          labs(x = "Proportions [%]", y = NULL, fill = targets_name, caption = note) +
          theme(legend.position = "top") +
          guides(colour = "none", size = "none") +
          scale_y_percent(expand = c(0, 0)) +
          theme(axis.title.y = element_text(size = rel(0.8), angle = 90)) +
          theme_lares(pal = 1)

        # Show a reference line if levels = 2; quite useful when data is unbalanced (not 50/50)
        if (length(unique(targets)) == 2 && ref) {
          distr <- df %>%
            freqs(.data$targets) %>%
            arrange(as.character(.data$targets))
          h <- signif(100 - distr$p[1], 3)
          prop <- prop +
            geom_hline(
              yintercept = h / 100, colour = "purple",
              linetype = "dotted", alpha = 0.8
            ) +
            geom_label(
              data = data.frame(x = 0),
              aes(0, h / 100, label = h, vjust = -0.05),
              size = 2.5, fill = "white", alpha = 0.8
            )
        }
        # Custom colours if wanted...
        if (custom_colours) {
          prop <- prop + suppressMessages(gg_fill_customs())
        }
      }

      # Export file name and folder
      if (save) {
        file_name <- paste0(
          "viz_distr_",
          cleanText(targets_name), ".vs.",
          cleanText(variable_name),
          case_when(type == 2 ~ "_c", type == 3 ~ "_p", TRUE ~ ""), ".png"
        )
        if (!is.na(subdir)) {
          # dir.create(file.path(getwd(), subdir), recursive = TRUE)
          file_name <- paste(subdir, file_name, sep = "/")
        }
      }

      # Plot the results and save if needed
      if (type == 1) {
        count <- count + labs(
          title = "Distribution and Proportions",
          subtitle = subtitle, caption = ""
        ) +
          theme(plot.margin = margin(10, 15, -15, 15))
        prop <- prop + guides(fill = "none") + labs(caption = note) +
          theme(plot.margin = margin(-5, 15, -15, 15))
        p <- (count / prop) + plot_layout(ncol = 1, nrow = 2)
        if (save) p <- p + ggsave(file_name, width = 10, height = 7)
      }
      if (type == 2) {
        count <- count + coord_flip() +
          labs(title = "Distribution Plot", subtitle = subtitle, caption = "")
        if (save) count <- count + ggsave(file_name, width = 8, height = 6)
        p <- count
      }
      if (type == 3) {
        prop <- prop + labs(title = "Proportions Plot", subtitle = subtitle, caption = "")
        if (save) prop <- prop + ggsave(file_name, width = 8, height = 6)
        p <- prop
      }

      if (!plot) {
        select(freqs, -.data$order, -.data$row)
      } else {
        p
      }
    }
  }
}

.force_class <- function(value, class = "none") {
  if (class != "none") {
    if (grepl("char|fact", class) && is.numeric(value)) {
      value <- as.character(value)
    }
    if (grepl("num|int", class) && !is.numeric(value)) {
      value <- as.numeric(value)
    }
    if (grepl("dat|day|time", class)) {
      value <- gsub(" .*", "", as.character(value))
      value <- lubridate::date(value)
    }
  }
  value
}

.fxtrim <- function(value, trim, targets = NA) {
  if (trim > 0) {
    if (!is.numeric(value)) {
      value <- substr(value, 1, trim)
    }
    if (!is.numeric(targets) && !is.na(targets)) {
      targets <- substr(targets, 1, trim)
    }
    message(paste("Chopping everything to", trim, "characters..."))
  }
  value
}

.fxclean <- function(value, clean = FALSE, targets = NA) {
  if (clean) {
    if (!is.numeric(value)) {
      value <- cleanText(value, spaces = FALSE)
    }
    if (!is.numeric(targets) && !is.na(targets)) {
      targets <- cleanText(targets, spaces = FALSE)
    }
  }
  value
}

.fxna_rm <- function(df, na.rm = FALSE) {
  if (na.rm) {
    df <- df[complete.cases(df), ]
  }
  df
}

# options(lifecycle_repeat_warnings = TRUE)
# library(dplyr)
# data("starwars")
# foo <- function(x, ...) {
#   temp <- enquos(...)
#   group_by(x, !!!temp[[1]]) %>% tally()
# }
# foo(starwars, sex, gender)

# Unquoting language objects with `!!!` is deprecated as of rlang 0.4.0.
# Please use `!!` instead.
#
# # Bad:
# dplyr::select(data, !!!enquo(x))
#
# # Good:
# dplyr::select(data, !!enquo(x))    # Unquote single quosure
# dplyr::select(data, !!!enquos(x))  # Splice list of quosures

####################################################################
#' SHAP values for H2O Models
#'
#' SHAP (SHapley Additive exPlanations) by Lundberg and Lee (2016) is a
#' method to explain individual predictions. SHAP is based on the game
#' theoretically optimal Shapley Values. Calculate SHAP values for
#' h2o models in which each row is an observation and each column a feature.
#' Use \code{plot} method to visualize features importance and distributions.
#'
#' @family SHAP
#' @param model \code{h2o_automl} object or \code{h2o} model.
#' @param test String or Dataframe. Leave "auto" to use \code{h2o_automl}'s
#' test dataset or pass a valid dataframe.
#' @param scores Numeric vector. If test != "auto", you must provide predicted values
#' @param y Character. If test != "auto", you must provide y variable's name
#' @param ... Additional argument for \code{predict_contributions.H2OModel}
#' @return H2OFrame with shap values for every observation and feature.
#' @examples
#' \dontrun{
#' # Train a h2o_automl model
#' model <- h2o_automl(dft, Survived,
#'   max_models = 1, target = TRUE,
#'   ignore = c("Ticket", "Cabin", "PassengerId"),
#'   quiet = TRUE
#' )
#'
#' # Calculate SHAP values
#' SHAP_values <- h2o_shap(model)
#' # Equivalent to:
#' # SHAP_values <- h2o_shap(
#' #  model = model$model,
#' #  test = model$datasets$test,
#' #  scores = model$scores_test$scores)
#'
#' # Check SHAP results
#' head(SHAP_values)
#'
#' # You must have "ggbeeswarm" library to use this auxiliary function:
#' # Plot SHAP values (feature importance)
#' plot(SHAP_values)
#'
#' # Plot some of the variables (categorical)
#' shap_var(SHAP_values, Pclass)
#'
#' # Plot some of the variables (numerical)
#' shap_var(SHAP_values, Fare)
#' }
#' @export
h2o_shap <- function(model, test = "auto", scores = "auto", y = "y", ...) {
  # When h2o_automl object and test = "auto"
  if (!is.null(attr(model, "type"))) {
    if (!is.data.frame(test)) {
      if (test == "auto") {
        test <- model$datasets$test
        scores <- model$scores_test[, 2]
        algos <- c("DRF", "GBM", "XGBOOST")
        if (!toupper(model$algorithm) %in% algos) {
          warning(paste(
            "You've passed a", model$algorithm, "model.",
            "Accepted algorithms:", v2t(algos)
          ))
        }
        auto <- TRUE
        y <- model$y
        model <- model$model
      }
    }
  } else {
    auto <- FALSE
  }

  # Calculate SHAP values
  test <- .quiet_h2o(as.h2o(test))
  shap <- .quiet_h2o(predict_contributions.H2OModel(model, test, ...))

  class(shap) <- c(class(shap), "h2o_shap")
  attr(shap, "test") <- as_tibble(test)
  attr(shap, "scores") <- scores
  if (auto) attr(shap, "y") <- y
  shap
}

####################################################################
#' @rdname h2o_shap
#' @aliases h2o_shap
#' @inheritParams mp3_get
#' @param x h2o_shap object
#' @param relevant Boolean. Keep only relevant non-trivial (>0) features
#' @param top Integer. Plot only top n values (as in importance)
#' @export
plot.h2o_shap <- function(x, relevant = TRUE, top = 15, quiet = FALSE, ...) {
  if (!inherits(x, "h2o_shap")) {
    stop("Pass a valid h2o_shap object to proceed!")
  }

  try_require("ggbeeswarm", stop = FALSE)
  if ("ggbeeswarm" %in% (.packages())) {
    fun <- function() {
      geom_quasirandom(
        groupOnX = FALSE, varwidth = TRUE, size = 0.9,
        alpha = 0.5, width = 0.2
      )
    }
  } else {
    fun <- function() geom_jitter(height = 0.2, alpha = 0.5)
  }

  scores <- attr(x, "scores")

  df <- x %>%
    as.data.frame() %>%
    select(-.data$BiasTerm) %>%
    gather("feature", "shap_value") %>%
    mutate(score = rep(scores, length(unique(.data$feature)))) %>%
    group_by(.data$feature) %>%
    mutate(
      shap_importance = mean(abs(.data$shap_value)),
      shap_mean = mean(.data$shap_value),
      varies = length(unique(.data$shap_value)) != 1
    ) %>%
    filter(.data$varies == relevant)
  features <- distinct(df, .data$feature, .data$shap_importance) %>%
    arrange(desc(.data$shap_importance)) %>%
    pull(.data$feature)
  df$feature <- factor(df$feature, levels = features)

  if (length(features) > top) {
    df <- dplyr::filter(df, .data$feature %in% features[1:top])
    subtitle <- sprintf("%s most important variables (of %s)", top, length(features))
  }

  # SHAP contribution plot
  pal <- names(lares_pal()$palette)
  p1 <- ggplot(df, aes(
    colour = .data$score,
    x = .data$shap_value,
    y = reorder(.data$feature, .data$shap_importance)
  )) +
    geom_vline(xintercept = 0, alpha = 0.5, colour = "black") +
    fun() +
    scale_colour_gradient(low = pal[3], high = pal[1]) +
    labs(
      x = "SHAP values", y = NULL,
      title = "SHAP Distribution"
    ) +
    theme_lares(legend = "none")

  # SHAP importance plot
  p2 <- df %>%
    select(.data$feature, .data$shap_importance) %>%
    distinct() %>%
    ggplot(aes(
      x = reorder(.data$feature, .data$shap_importance),
      y = .data$shap_importance
    )) +
    geom_col(fill = "black", alpha = 0.8) +
    coord_flip() +
    labs(
      x = NULL, y = "mean(|SHAP value|)",
      title = "Feature Importance"
    ) +
    theme_lares()
  if (length(features) > top) {
    p2 <- p2 + labs(caption = sprintf("%s most important features (of %s)", top, length(features)))
  }

  # Combine plots
  p <- p1 + p2
  p
}


####################################################################
#' SHAP-based dependence plots for categorical/numerical features (PDP)
#'
#' Having a \code{h2o_shap} object, plot a dependence plot for any
#' categorical or numerical feature.
#'
#' @family SHAP
#' @param x \code{h2o_shap} object
#' @param var Variable name
#' @param keep_outliers Boolean. Outliers detected with z-score and 3sd
#' may be suppress or kept in your plot. Keep them?
#' @return ggplot2 objct with shap values plotted
#' @inherit h2o_shap examples
#' @export
shap_var <- function(x, var, keep_outliers = FALSE) {
  if (!inherits(x, "h2o_shap")) {
    stop("Pass a valid h2o_shap object to proceed!")
  }

  try_require("ggbeeswarm", stop = FALSE)
  if ("ggbeeswarm" %in% (.packages())) {
    fun <- function() {
      geom_quasirandom(
        groupOnX = TRUE, varwidth = TRUE, size = 1,
        alpha = 0.6, width = 0.4
      )
    }
  } else {
    fun <- function() geom_jitter(height = 0.2, alpha = 0.5)
  }

  test <- attr(x, "test")
  scores <- attr(x, "scores")
  y <- attr(x, "y")
  var <- enquo(var)
  name <- as_label(var)

  # Gather everything up
  shap_df2 <- x %>%
    as.data.frame() %>%
    select(starts_with(name)) %>%
    mutate(
      model_result = attr(x, "scores"),
      real_value = pull(attr(x, "test")[name], 1)
    ) %>%
    tidyr::gather(starts_with(name), key = "feature", value = "shap")

  type <- if (is.numeric(pull(shap_df2, .data$real_value)[1])) "numerical" else "categorical"
  title <- glued("SHAP values for {type} feature: {as_label(var)}")

  # Outliers
  if (type == "numerical") {
    shap_df2 <- mutate(shap_df2, outlier = outlier_zscore(.data$real_value))
  } else {
    shap_df2$outlier <- FALSE
  }
  outs <- freqs(shap_df2, .data$outlier)
  outs_msg <- NULL
  if (!keep_outliers) {
    if (nrow(outs) > 1) {
      outs_msg <- glued("{outs$n[2]} outlier data points excluded (out of {nrow(shap_df2)})")
      shap_df2 <- filter(shap_df2, .data$outlier == FALSE)
    }
  }

  pal <- names(lares_pal()$palette)
  p <- shap_df2 %>%
    ggplot(aes(x = .data$real_value, y = .data$shap, colour = .data$model_result)) +
    geom_hline(yintercept = 0, alpha = 0.5) +
    fun() +
    geom_smooth(method = "loess", formula = "y ~ x", colour = "black", size = 0.6) +
    geom_smooth(method = "lm", formula = "y ~ x", colour = "black", size = 0.3) +
    scale_colour_gradient(low = pal[3], high = pal[1]) +
    labs(
      x = name, y = paste("SHAP values for", name),
      colour = "Prediction",
      title = title, caption = outs_msg,
      subtitle = paste("Predicted variable:", y)
    ) +
    theme_lares()
  p
}

# # Train a model
# model <- h2o_automl(dft, Survived, max_models = 1,
#                     target = TRUE,
#                     ignore = c("Ticket", "Cabin", "PassengerId"))
#
# # Calculate SHAP values
# SHAP_values <- h2o_shap(model)
#
# # Plot some of the variables (categorical and numerical)
# shap_var(SHAP_values, Pclass)
# shap_var(SHAP_values, Fare)
# shap_var(SHAP_values, Fare, keep_outliers = TRUE)
# shap_var(SHAP_values, Age)

.onLoad <- function(libname, pkgname) {
  Sys.setenv(
    # So user can set another font by default on theme_lares()
    "LARES_FONT" = if (Sys.getenv("LARES_FONT") != "") {
      Sys.getenv("LARES_FONT")
    } else {
      "Arial Narrow"
    },
    # Standard format for formatNum()
    "LARES_NUMFORMAT" = if (Sys.getenv("LARES_NUMFORMAT") != "") {
      Sys.getenv("LARES_NUMFORMAT")
    } else {
      2
    },
    # Status / Loading bar style for statusbar()
    "LARES_STATUSBAR" = if (Sys.getenv("LARES_STATUSBAR") != "") {
      Sys.getenv("LARES_STATUSBAR")
    } else {
      "filled"
    },
    # Scrabble and other language settings
    "LARES_LANG" = if (Sys.getenv("LARES_LANG") != "") {
      Sys.getenv("LARES_LANG")
    } else {
      "es"
    },
    # ChatGPT Default values
    "LARES_GPT_MODEL" = if (Sys.getenv("LARES_GPT_MODEL") != "") {
      Sys.getenv("LARES_GPT_MODEL")
    } else {
      "gpt-3.5-turbo"
    },
    "LARES_GPT_URL" = if (Sys.getenv("LARES_GPT_URL") != "") {
      Sys.getenv("LARES_GPT_URL")
    } else {
      "https://api.openai.com/v1/chat/completions"
    },
    "LARES_GEMINI_API" = if (Sys.getenv("LARES_GEMINI_API") != "") {
      Sys.getenv("LARES_GEMINI_API")
    } else {
      "https://generativelanguage.googleapis.com/v1beta/models/"
    }
  )
  options(
    # Additional default options for other libraries
    "getSymbols.warning4.0" = FALSE,
    "getSymbols.yahoo.warning" = FALSE,
    "h2o.use.data.table" = FALSE, # Avoid data.table warning: cannot be used without bit64
    "lares.colours" = c(
      orange = "#EA9A28",
      blue = "#00B1DA",
      purple = "#6A2C70",
      red = "#E63946",
      green = "#00C9AE",
      navy = "#125D98",
      yellow = "#F8D962",
      grey = "#8D99AE",
      pink = "#F6B8B8",
      fuchsia = "#FF2EB5",
      brown = "#884F3E",
      black = "#0c0c0c",
      white = "#F6F5F5"
    ),
    "lares.palette" = c(
      "#EA9A28" = "#000000",
      "#00B1DA" = "#000000",
      "#6A2C70" = "#FFFFFF",
      "#E63946" = "#000000",
      "#00C9AE" = "#000000",
      "#F8D962" = "#000000",
      "#D35FB7" = "#000000",
      "#125D98" = "#FFFFFF",
      "#F6B8B8" = "#000000",
      "#8D99AE" = "#000000",
      "#290452" = "#FFFFFF",
      "#7ADf90" = "#000000",
      "#2FFECC" = "#000000",
      "#0C7BDC" = "#000000",
      "#817B7B" = "#000000",
      "#F66320" = "#000000",
      "#F4A261" = "#000000",
      "#005AB5" = "#FFFFFF",
      "#9A9A9A" = "#000000",
      "#00008B" = "#FFFFFF",
      "#E1BE6A" = "#000000",
      "#40B0A6" = "#000000",
      "#056E00" = "#FFFFFF",
      "#E40000" = "#000000",
      "#FE8A71" = "#000000",
      "#8600A1" = "#FFFFFF",
      "#A52A2A" = "#FFFFFF",
      "#000000" = "#FFFFFF",
      "#E69F00" = "#000000",
      "#009E73" = "#000000",
      "#0072B2" = "#FFFFFF",
      "#D55E00" = "#000000"
    )
  )
}

####################################################################
#' Dataset columns and rows structure
#'
#' This function lets the user to check quickly the structure of a
#' dataset (data.frame). It returns multiple counters for useful metrics,
#' a plot, and a list of column names for each of the column metrics.
#'
#' @family Exploratory
#' @param df Dataframe
#' @param return Character. Return "skimr" for skim report, "numbers" for
#' stats and numbers, "names" for a list with the column names of each of
#' the class types, "plot" for a nice plot with "numbers" output, "distr"
#' for an overall summary plot showing categorical, numeric, and missing
#' values by using \code{plot_df}
#' distributions
#' @param subtitle Character. Add subtitle to plot
#' @param quiet Boolean. Keep quiet or show other options available?
#' @return Depending on \code{return} input and based on your \code{df} structure:
#' \itemize{
#'   \item \code{list} with the names of the columns classified by class
#'   \item \code{data.frame} with numbers: total values, row, columns,
#'   complete rows
#'   \item \code{plot} with visualizations
#' }
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#' df_str(dft, "names")
#' df_str(dft, "numbers", quiet = TRUE)
#' df_str(dft, "plot", quiet = TRUE)
#' @export
df_str <- function(df,
                   return = "plot",
                   subtitle = NA,
                   quiet = FALSE) {
  if (!quiet) {
    rets <- c("skimr", "numbers", "names", "distr", "plot")
    message(paste("Other available 'return' options:", v2t(rets[rets != return])))
  }

  df <- data.frame(df)

  if (return == "skimr") {
    try_require("skimr")
    output <- skim(df)
  } else if (return == "distr") {
    p <- plot_df(df)
    output <- p
  } else {
    names <- list(
      cols = colnames(df),
      nums = colnames(df)[unlist(lapply(df, is.numeric))],
      char = colnames(df)[unlist(lapply(df, is.character))],
      factor = colnames(df)[unlist(lapply(df, is.factor))],
      logic = colnames(df)[unlist(lapply(df, is.logical))]
    )
    names[["time"]] <- names$cols[!colnames(df) %in% c(
      names$nums, names$char, names$factor, names$logic
    )]
    names[["allnas"]] <- names$cols[unlist(lapply(df, function(x) all(is.na(x))))]

    if (return == "names") {
      output <- names
    } else {
      numbers <- data.frame(
        "Total Values" = nrow(df) * ncol(df),
        "Total Rows" = nrow(df),
        "Total Columns" = ncol(df),
        "Numeric Columns" = length(names$nums),
        "Character Columns" = length(names$char),
        "Factor Columns" = length(names$factor),
        "Logical Columns" = length(names$logic),
        "Time/Date Columns" = length(names$time),
        "All Missing Columns" = length(names$allnas),
        "Missing Values" = sum(is.na(df)),
        "Complete Rows" = sum(complete.cases(df)),
        "Memory Usage" = as.numeric(object.size(df))
      )
      intro2 <- data.frame(counter = t(numbers)) %>%
        mutate(
          metric = row.names(.),
          type = ifelse(grepl("Column", colnames(numbers)), "Columns",
            ifelse(grepl("Rows", colnames(numbers)), "Rows", "Values")
          ),
          p = ifelse(.data$type == "Columns", 100 * .data$counter / numbers$Total.Columns,
            ifelse(.data$type == "Rows", 100 * .data$counter / numbers$Total.Rows,
              100 * .data$counter / numbers$Total.Values
            )
          ),
          p = round(.data$p, 2),
          type = factor(.data$type, levels = c("Values", "Columns", "Rows"))
        ) %>%
        select(.data$metric, .data$counter, .data$type, .data$p)

      if (return == "numbers") {
        output <- select(intro2, -.data$type)
      } else if (return == "plot") {
        p <- intro2 %>%
          filter(!.data$metric %in% "Memory.Usage") %>%
          mutate(x = ifelse(.data$p < 75, -0.15, 1.15)) %>%
          ggplot(aes(
            x = reorder(.data$metric, as.integer(.data$counter)),
            y = .data$p, fill = .data$type,
            label = formatNum(.data$counter, 0)
          )) +
          geom_col() +
          coord_flip() +
          ylim(0, 100) +
          theme_minimal() +
          guides(fill = "none") +
          labs(
            title = "Dataset overall structure",
            x = "", y = "% of total", fill = "",
            caption = paste("Memory Usage:", formatNum(numbers$Memory.Usage / (1024 * 1024)), "Mb")
          ) +
          facet_grid(type ~ ., scales = "free", space = "free") +
          geom_text(aes(hjust = .data$x), size = 3) +
          theme_lares(pal = 1)
        if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
        output <- p
      }
    }
  }
  output
}


####################################################################
#' Plot All Numerical Features (Boxplots)
#'
#' This function filters numerical columns and plots boxplots.
#'
#' @family Exploratory
#' @param df Dataframe
#' @return Plot. Result of \code{df} numerical features.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#' plot_nums(dft)
#' @export
plot_nums <- function(df) {
  which <- df %>% select_if(is.numeric)
  if (length(which) > 0) {
    p <- gather(which) %>%
      filter(!is.na(.data$value)) %>%
      ggplot(aes(x = .data$key, y = .data$value)) +
      geom_jitter(alpha = 0.2, size = 0.8) +
      geom_boxplot(alpha = 0.8, outlier.shape = NA, width = 1) +
      facet_wrap(.data$key ~ ., scales = "free") +
      labs(title = "Numerical Features Boxplots", x = NULL, y = NULL) +
      theme_lares() +
      theme(
        axis.text.y = element_blank(),
        axis.text.x = element_text(vjust = 2, size = 8),
        panel.spacing.y = unit(-.5, "lines"),
        strip.text = element_text(size = 10, vjust = -1.3)
      ) +
      coord_flip()
    p
  } else {
    message("No numerical variables found!")
  }
}


####################################################################
#' Plot All Categorical Features (Frequencies)
#'
#' This function filters categorical columns and plots the frequency
#' for each value on every feature.
#'
#' @family Exploratory
#' @param df Dataframe
#' @return Plot. Result of \code{df} categorical features.
#' @export
plot_cats <- function(df) {
  plot <- df %>% select_if(Negate(is.numeric))
  if (length(plot) > 0) {
    plot %>% freqs(plot = TRUE) +
      labs(title = "Categorical Features Frequencies")
  } else {
    message("No categorical variables found!")
  }
}


####################################################################
#' Plot Summary of Numerical and Categorical Features
#'
#' This function plots all columns frequencies and boxplots, for
#' categorical and numerical respectively.
#'
#' @family Exploratory
#' @param df Dataframe
#' @return Plot. Result of \code{df} categorical and numerical features.
#' @export
plot_df <- function(df) {
  plots <- list()

  cats <- plot_cats(df)
  if (length(cats) != 0) {
    plots[["cats"]] <- cats +
      theme(plot.title = element_text(size = 12))
  }

  nums <- plot_nums(df)
  if (length(nums) != 0) {
    plots[["nums"]] <- nums +
      theme(plot.title = element_text(size = 12))
  }

  mis <- missingness(df, plot = TRUE, summary = FALSE)
  if (length(mis) != 0) {
    plots[["miss"]] <- mis +
      theme(plot.title = element_text(size = 12)) + guides(fill = "none")
  }

  if (length(plots) == 3) heights <- c(4 / 12, 1 / 2, 3 / 12)
  if (length(plots) == 2) heights <- c(0.5, 0.5)
  if (length(plots) == 1) heights <- NULL

  margin <- theme(plot.margin = unit(c(0.1, 0.5, 0.1, 0.5), "cm"))
  plots <- lapply(plots, "+", margin)
  wrap_plots(plots, heights = heights)
}

#' ####################################################################
#' #' Image's Objects Detection using YOLO
#' #'
#' #' Run devtools::install_github("bnosac/image",
#' #' subdir = "image.darknet", build_vignettes = TRUE)
#' #'
#' #' @param filename Character. Name of the file to analize
#' #' @param threshold Numeric. Range from 0 to 1. Detection threshold
#' #' @param type Character. Select 'faster' or 'better' for YOLO method
#' #' @export
#' object_detection <- function(filename, threshold = 0.2, type = "faster") {
#'
#'   # require(image.darknet)
#'
#'   path <- getwd()
#'   file <- paste0(path, "/", filename)
#'   message("File: ", file)
#'
#'   if (!file.exists(file)) {
#'     stop("That file doesn't exist or is not in your working directory!")
#'   }
#'
#'   if (!type %in% c("faster","better")) {
#'     stop("Please try 'faster' or 'better' on your type parameter!")
#'   }
#'
#'   if (threshold > 1 || threshold < 0) {
#'     stop("You should use a threshold value between 0 and 1")
#'   }
#'
#'   if (type == "faster") {
#'     model <- "tiny-yolo-voc.cfg"
#'     weights <- system.file(package="image.darknet", "models", "tiny-yolo-voc.weights")
#'     labels <- system.file(package="image.darknet", "include", "darknet", "data", "voc.names")
#'   }
#'
#'   if (type == "better") {
#'     weights <- file.path(system.file(package="image.darknet", "models"), "yolo.weights")
#'     if (!file.exists(weights)) {
#'       download.file(url = "http://pjreddie.com/media/files/yolo.weights", destfile = weights)
#'     }
#'     model <- "yolo.cfg"
#'     weights <- system.file(package="image.darknet", "models", "yolo.weights")
#'     labels <- system.file(package="image.darknet", "include", "darknet", "data", "coco.names")
#'   }
#'
#'   yolo <- image_darknet_model(
#'     type = 'detect',
#'     model = model,
#'     weights = weights,
#'     labels = labels)
#'
#'   x <- image_darknet_detect(
#'     file = file,
#'     object = yolo,
#'     threshold = threshold)
#'
#'   message("Succesfully exported 'predictions.png' into your working directory")
#'
#' }

####################################################################
#' Recursive Partitioning and Regression Trees
#'
#' Fit and plot a \code{rpart} model for exploratory purposes using
#' \code{rpart} and \code{rpart.plot} libraries.
#'
#' @family Exploratory
#' @family Visualization
#' @inheritParams h2o_automl
#' @inheritParams rpart::rpart
#' @inheritParams rpart::rpart.control
#' @inheritParams rpart.plot::rpart.plot
#' @param df Data frame
#' @param max Integer. Maximal depth of the tree.
#' @param min Integer. The minimum number of observations that must
#' exist in a node in order for a split to be attempted.
#' @param ohse Boolean. Auto generate One Hot Smart Encoding?
#' @param plot Boolean. Return a plot? If not, \code{rpart} object.
#' @param explain Boolean. Include a brief explanation on the bottom
#' part of the plot.
#' @param title,subtitle Character. Title and subtitle to include in plot.
#' Set to \code{NULL} to ignore.
#' @param ... Additional parameters passed to \code{rpart.plot()}.
#' @return (Invisible) list type 'tree_var' with plot (function), model,
#' predictions, performance metrics, and interpret auxiliary text.
#' @examples
#' data(dft)
#' # Regression Tree
#' tree <- tree_var(dft, Fare, subtitle = "Titanic dataset")
#' tree$plot() # tree plot
#' tree$model # rpart model object
#' tree$performance # metrics
#' # Binary Tree
#' tree_var(dft, Survived_TRUE, explain = FALSE, cex = 0.8)$plot()
#' # Multiclass tree
#' tree_var(dft[, c("Pclass", "Fare", "Age")], Pclass, ohse = FALSE)$plot()
#' @export
tree_var <- function(df, y, type = 2, max = 3, min = 20, cp = 0,
                     ohse = TRUE, plot = TRUE, explain = TRUE,
                     title = NA, subtitle = NULL, ...) {
  if (ohse) df <- ohse(df, limit = min)

  # Check if main variable exists
  target_quo <- enquo(y)
  target_txt <- quo_name(target_quo)[[1]]
  if (!target_txt %in% colnames(df)) {
    message(paste("Not a valid input:", target_txt, "was transformed or does not exist."))
    maybes <- colnames(df)[grepl(target_txt, colnames(df))]
    if (length(maybes) > 0) message(paste("Maybe you meant one of:", vector2text(maybes)))
    stop()
  }

  # Train tree
  mod <- rpart(as.formula(paste(target_txt, "~ .")),
    data = df,
    control = rpart.control(maxdepth = max, minsplit = min, cp = cp)
  )

  # Tree's performance
  aux <- df[!is.na(df[, target_txt]), ]
  p <- predict(mod, aux)
  real <- unlist(aux[, target_txt])
  if (!is.vector(p)) {
    multis <- p
    p <- colnames(p)[apply(p, 1, which.max)]
    thresh <- ncol(multis) - 1
  } else {
    multis <- NULL
  }
  performance <- model_metrics(real, p, multis, auto_n = FALSE, plots = FALSE, quiet = TRUE)
  preds <- as_tibble(bind_cols(tag = real, pred = p, multis))

  interpret <- sprintf(paste(
    "Recursive partitioning and regression tree",
    "with a minimum split of %s observations per node and up to depth %s,",
    "with %s complexity parameter (doesn't necessarily increase performance at each step).",
    "On each split or leaf, upper number is 'mean value' and lower number is",
    "'percentage of observations' per level.",
    collapse = ""
  ), min, max, cp)

  if (plot) {
    if (nrow(mod$frame) > 1) {
      bottom_mg <- ifelse(explain, 3.5, 1)
      plot_tree <- function(tree, title, subtitle, explain = TRUE) {
        font <- .font_global(Sys.getenv("LARES_FONT"), when_not = NULL, ...)
        rpart.plot(tree,
          type = type, roundint = FALSE, family = font,
          mar = c(bottom_mg, 1, 2.5, 1), fallen.leaves = FALSE, ...
        )
        if (is.na(title)) {
          mtext(
            side = 3, line = 2.5, at = -0.05, adj = 0, cex = 1.1,
            paste("Decision Tree for", target_txt), family = font
          )
        }
        mtext(side = 3, line = 1.6, at = -0.05, adj = 0, cex = 0.9, subtitle, family = font)
        if (explain) {
          mtext(
            side = 1, line = 3, at = -0.05, adj = 0, cex = 0.6,
            family = font, autoline(interpret, rel = 4.4)
          )
        }
      }
      tree.plot <- function() plot_tree(mod, title, subtitle, explain)
    } else {
      tree.plot <- noPlot("Can't grow decision tree")
    }
  } else {
    plot <- NULL
  }

  ret <- list(
    plot = tree.plot,
    model = mod,
    preds = preds,
    performance = performance,
    interpret = interpret
  )
  attr(ret, "type") <- "tree_var"
  invisible(ret)
}

#' @export
plot.tree_var <- function(x, ...) {
  if (!inherits(x, "tree_var")) {
    stop("Object must be class tree_var")
  }
  if ("plot" %in% names(x)) x$plot()
}

# data(dft)
# tree_var(dft, Survived_TRUE)$plot()
# tree <- tree_var(dft, Fare)
# tree$plot() # tree plot
# tree$model # model
# tree$performance # metrics

# tree_var(dft[,c("Pclass","Fare","Age")], Pclass, ohse = FALSE)$plot()
# dft %>%
#   mutate(weird = Fare >= 26 & Age < 15 & Pclass == 3) %>%
#   filter(!is.na(weird)) %>%
#   freqs(weird)
# x <- tree_var(dft[,c("Pclass","Fare","Age")], Pclass, ohse = FALSE)

# ROC(x$preds$tag, x$preds$pred, x$preds[,3:5])
# tag = x$preds$tag
# score = x$preds$pred
# multis = x$preds[,3:5]

####################################################################
#' Stopwatch to measure timings in R
#'
#' Start a stopwatch.
#'
#' @family Time
#' @family Tools
#' @inheritParams mp3_get
#' @param id Define ID if multiple \code{tic} & \code{toc} are being used.
#' @param start Start time. Now is default.
#' @return Invisible list. Contains tic (start time), toc (stop time),
#' elapsed time and message printed.
#' @examples
#' # Basic use (global stopwatch)
#' tic()
#' Sys.sleep(0.1)
#' toc()
#'
#' # Multiple tic tocs
#' tic(id = "two", quiet = FALSE)
#' Sys.sleep(0.2)
#' toc(id = "two")
#'
#' # Global is still working (id = 1)
#' toc(msg = "The function finished its work in")
#' @export
#' @rdname tic
tic <- function(id = 1, start = proc.time()["elapsed"], quiet = TRUE) {
  tic <- .set.tictoc("TIC", id, start = start)
  if (!quiet) message(sprintf("Tic `id = %s` start time: %s", id, Sys.time()))
  invisible(tic)
}


####################################################################
#' Stopwatch Stop
#'
#' Stop a stopwatch.
#'
#' @inheritParams cache_write
#' @param msg Character. Custom message shown
#' @param type Character. Output format for \code{time} list element.
#' Choose any of: \code{units, clock, seconds}.
#' @param signif Integer. Significant digits.
#' @return \code{toc} returns an (invisible) list containing the time-stamps
#' \code{tic} and \code{toc}, \code{time} in seconds and the message \code{msg}.
#' @export
#' @rdname tic
toc <- function(id = 1, msg = "Elapsed time:", type = "units", signif = 3, quiet = FALSE, ...) {
  check_opts(type, c("units", "clock", "seconds"))
  current <- getOption("LARES_TICTOC")
  id <- as.character(id)
  if (!id %in% names(current)) {
    stop(sprintf("You need to tic(id = '%s') before you toc(id = '%s')", id, id))
  }

  toc <- .set.tictoc("TOC", id)
  tic <- getOption("LARES_TICTOC")[[id]][["TIC"]]
  toc <- getOption("LARES_TICTOC")[[id]][["TOC"]]
  time <- as.numeric(toc - tic)

  if (type == "units") {
    x <- time
    u <- ifelse(x < 60, "s", ifelse(x < 3600, "m", ifelse(x < 86400, "h", "d")))
    d <- ifelse(x < 60, 1, ifelse(x < 3600, 60, ifelse(x < 86400, 3600, 86400)))
    out <- paste0(signif(time / d, signif), u)
  }

  if (type == "clock") {
    td <- lubridate::seconds_to_period(round(time))
    out <- sprintf("%02d:%02d:%02d", td@hour, minute(td), second(td))
  }

  if (type == "seconds") {
    out <- signif(time, signif)
  }

  msg <- sprintf("%s %s", msg, out)
  if (!quiet) message(msg)
  res <- list(tic = tic, toc = toc, time = out, msg = msg)
  invisible(res)
}

.set.tictoc <- function(which = c("TIC", "TOC"), id = 1,
                        start = proc.time()["elapsed"]) {
  aux <- getOption("LARES_TICTOC")
  if (is.null(aux)) aux <- list()
  id <- as.character(id)
  if (which[1] == "TIC") temp <- data.frame(TIC = start, TOC = NA)
  if (which[1] == "TOC") temp <- data.frame(TIC = aux[[id]]$TIC, TOC = start)
  aux[[id]] <- temp
  options("LARES_TICTOC" = aux)
  aux[id]
}

####################################################################
#' DALEX Explainer for H2O
#'
#' DALEX helper function to create an \code{explainer} object using
#' a \code{h2o} trained model.
#'
#' @family Interpretability
#' @param df Dataframe. Must contain all columns and predictions
#' @param model Model object (H2O)
#' @param y Character or Variable name. Variable's column name.
#' @param ignore Character vector. Which columns should be ignored?
#' @param ... Additional parameters to pass to \code{h2o_predict_model} or
#' \code{h2o_predict_MOJO}.
#' @return List; explainer. Containing the model, data, y, predict_function,
#' y_hat, residuals, class, label, model_info, residual_function, and weights.
#' @aliases dalex_explainer
#' @examples
#' # You must have "DALEX" library to use this auxiliary function:
#' \dontrun{
#' data(dft) # Titanic dataset
#'
#' # TRAIN A SIMPLE MODEL
#' dfm <- h2o_automl(dft,
#'   y = "Survived",
#'   ignore = c("Ticket", "PassengerId", "Cabin"),
#'   max_models = 1
#' )
#'
#' # EXPLAINER
#' explainer <- h2o_explainer(df = dfm$datasets$test, model = dfm$model, y = "Survived")
#' explainer$data <- na.omit(explainer$data)
#'
#' # CATEGORICAL EXAMPLE
#' class <- dalex_variable(explainer, vars = c("Pclass", "Sex"))
#' class$plot
#'
#' # NUMERICAL EXAMPLE
#' num <- dalex_variable(explainer, vars = c("Fare", "Age"))
#' num$plot
#'
#' # LOCAL EXAMPLE
#' local <- dalex_local(explainer, row = 1)
#' # OR YOU COULD MANUALLY INPUT THE OBSERVATION
#' local <- dalex_local(explainer, observation = explainer$data[1, ])
#' local$plot
#'
#' # xai2shiny's UI (needs to be installed from ModelOriented/xai2shiny)
#' xai2shiny(explainer, run = TRUE)
#' }
#' @export
h2o_explainer <- function(df, model, y = "tag", ignore = NULL, ...) {
  try_require("DALEX")

  df <- data.frame(df) %>%
    # No need to use prediction results
    select(-c(which(colnames(.) == "train_test"):ncol(.))) %>%
    # Exclude variables with no variance
    select(-one_of(zerovar(.)))

  y <- gsub('"', "", as_label(enquo(y)))

  if (!y %in% colnames(df)) {
    stop(paste("The y value", y, "is not in your data.frame"))
  }

  df <- df[, !(colnames(df) %in% ignore)]
  x_valid <- select(df, -one_of(y))
  y_valid <- df[y][, 1]

  if (any(grepl("H2O", class(model)))) {
    label <- model@model_id
  } else {
    label <- basename(model)
  }

  h2o_predict_fx <- function(model, newdata, ...) {
    try_require("h2o")
    # h2o_predict_model()
    if (any(grepl("H2O", class(model)))) {
      results <- .quiet_h2o(h2o_predict_model(newdata, model))
    } else {
      model <- dirname(model)
      # h2o_predict_MOJO()
      if (dir.exists(model)) {
        results <- .quiet_h2o(h2o_predict_MOJO(newdata, model, ...))
      } else {
        stop("Directory doesn't exist: ", model)
      }
    }
    results <- results[[2L]]
    results
  }

  explainer <- explain.default(
    model = model,
    data = x_valid,
    y = y_valid,
    predict_function = h2o_predict_fx,
    label = label
  )
  explainer$model_info$package <- "h2o"
  explainer
}


####################################################################
#' DALEX Local
#'
#' DALEX function for local interpretations
#'
#' @family Interpretability
#' @param explainer Object. Result from h2o_explainer function
#' @param observation Data.frame. If you want to use an observation
#' that was not in the original explainer function, add here. Else, use row
#' @param row Dataframe. Row number from the data.frame used in explainer.
#' @param type Character. The type of variable attributions.
#' Either shap, oscillations, break_down or break_down_interactions.
#' @return List. Containing observation, breakdown results, and breakdown plot.
#' @export
dalex_local <- function(explainer, observation = NA, row = 1, type = "break_down") {
  try_require("DALEX")
  tic("dalex_local")

  subtitle <- paste0("Observation #", v2t(row, quotes = FALSE))
  if (!is.data.frame(observation)) {
    observation <- explainer$data[row, ]
  } else {
    if (!all(colnames(observation) %in% colnames(explainer$data))) {
      stop(paste(
        "All columns must be present in your observation:",
        v2t(colnames(explainer$data))
      ))
    }
  }

  # BREAKDOWN
  breakdown <- predict_parts(explainer, new_observation = observation, type = type)

  p <- plot(breakdown) +
    theme_lares(legend = "none") +
    labs(subtitle = NULL, caption = subtitle)

  return <- list(observation = observation, breakdown = breakdown, plot = p)
  toc("dalex_local")
  return
}


####################################################################
#' DALEX Residuals
#'
#' DALEX function for residuals
#'
#' @family Interpretability
#' @param explainer Object. Result from h2o_explainer function
#' @return Plot. Based of \code{explainer} residual results.
#' @export
dalex_residuals <- function(explainer) {
  try_require("DALEX")
  resids <- model_performance(explainer)
  p1 <- plot(resids) + theme_lares(legend = "none")
  p2 <- plot(resids, geom = "boxplot") + theme_lares(legend = "none")
  p1 + p2 + plot_layout(nrow = 2)
}


####################################################################
#' DALEX Partial Dependency Plots (PDP)
#'
#' DALEX auxiliary function for creating Partial Dependency Plots and
#' study variable's responses vs independent vector.
#'
#' @family Interpretability
#' @inheritParams clusterKmeans
#' @param explainer Object. Result from \code{h2o_explainer} function.
#' @param vars Character vector. Which features do you wish to study?
#' @param force_class Character. If you wish to force a class on your
#' vars, which one do you need?
#' @param ... Additional parameters passed to \code{model_profile}.
#' @return List. Containing PDP results, plot and \code{vars} input.
#' @examples
#' # You must have "DALEX" library to use this auxiliary function:
#' \dontrun{
#' # Having an "explainer" object created with \code{h2o_explainer}:
#' # For numerical variables
#' dalex_variable(explainer, vars = c("Age", "Fare"))
#' # For categorical variables
#' dalex_variable(explainer, vars = c("Pclass", "Sex"))
#' }
#' @export
dalex_variable <- function(explainer, vars, force_class = NA, seed = 123, ...) {
  try_require("DALEX")
  all_vars <- colnames(explainer$data)
  if (!all(vars %in% all_vars)) {
    stop("Select any variable(s) from the following: ", v2t(all_vars))
  }

  if (!is.na(force_class)) {
    classes <- c("factor", "numeric")
    if (force_class %in% classes) {
      class(explainer$data[[vars]]) <- force_class
      message(paste("Changed class to", force_class))
    } else {
      if (!is.na(force_class)) {
        stop(paste("Try using force_class:", vector2text(classes)))
      }
    }
  }

  set.seed(seed)
  aux <- model_profile(explainer, variables = vars, ...)
  p <- plot(aux) + theme_lares(legend = "top") +
    labs(y = "Average Prediction") +
    scale_y_formatNum(signif = 2)
  list(pdp = aux, plot = p, vars = vars)
}

####################################################################
#' One Hot Smart Encoding (Dummy Variables)
#'
#' This function lets the user automatically transform a dataframe with
#' categorical columns into numerical by one hot encoding technic.
#'
#' @family Data Wrangling
#' @family Feature Engineering
#' @family One Hot Encoding
#' @inheritParams cache_write
#' @param df Dataframe
#' @param redundant Boolean. Should we keep redundant columns? i.e. If the
#' column only has two different values, should we keep both new columns?
#' Is set to \code{NULL}, only binary variables will dump redundant columns.
#' @param drop Boolean. Drop automatically some useless features?
#' @param ignore Vector or character. Which column should be ignored?
#' @param dates Boolean. Do you want the function to create more features
#' out of the date/time columns?
#' @param holidays Boolean. Include holidays as new columns?
#' @param country Character or vector. For which countries should the holidays
#' be included?
#' @param currency_pair Character. Which currency exchange do you
#' wish to get the history from? i.e, USD/COP, EUR/USD...
#' @param trim Integer. Trim names until the nth character
#' @param limit Integer. Limit one hot encoding to the n most frequent
#' values of each column. Set to \code{NA} to ignore argument.
#' @param variance Numeric. Drop columns with more than n variance.
#' Range: 0-1. For example: if a variable contains 91 unique different
#' values out of 100 observations, this column will be suppressed if
#' value is set to 0.9
#' @param other_label Character. With which text do you wish to replace
#' the filtered values with?
#' @param sep Character. Separator's string
#' @return data.frame on which all features are numerical by nature or
#' transformed with one hot encoding.
#' @examples
#' data(dft)
#' dft <- dft[, c(2, 3, 5, 9, 11)]
#'
#' ohse(dft, limit = 3) %>% head(3)
#' ohse(dft, limit = 3, redundant = NULL) %>% head(3)
#'
#' # Getting rid of columns with no (or too much) variance
#' dft$no_variance1 <- 0
#' dft$no_variance2 <- c("A", rep("B", nrow(dft) - 1))
#' dft$no_variance3 <- as.character(rnorm(nrow(dft)))
#' dft$no_variance4 <- c(rep("A", 20), round(rnorm(nrow(dft) - 20), 4))
#' ohse(dft, limit = 3) %>% head(3)
#' @export
ohse <- function(df,
                 redundant = FALSE,
                 drop = TRUE,
                 ignore = NULL,
                 dates = FALSE,
                 holidays = FALSE,
                 country = "Venezuela",
                 currency_pair = NA,
                 trim = 0,
                 limit = 10,
                 variance = 0.9,
                 other_label = "OTHER",
                 sep = "_",
                 quiet = FALSE,
                 ...) {
  if (is.vector(df)) {
    df <- data.frame(var = df)
  } else {
    df <- data.frame(df)
  }

  order <- colnames(df)

  # Dummy variables that will be filled
  no_need_to_convert <- converted <- converted_binary <- NULL

  # Leave some columns out of the logic
  ignore <- unique(ignore)
  if (!is.null(ignore)) {
    if (!quiet) message(">>> Omitting transformations for ", vector2text(ignore))
    ignored <- select(df, any_of(ignore))
    df <- select(df, -any_of(ignore))
  } else {
    ignored <- NULL
  }

  # No variance columns
  no_variance <- zerovar(df)
  if (drop) df <- select(df, !any_of(no_variance))

  # Create features out of date/time variables
  if (dates == TRUE || holidays == TRUE || !is.na(currency_pair)) {
    times <- df_str(df, return = "names", quiet = TRUE)$time
    if (length(times) <= 1) {
      df_dates <- date_feats(df,
        drop = FALSE,
        append = FALSE,
        holidays = holidays,
        country = country,
        currency_pair = currency_pair,
        quiet = quiet
      )

      if (ncol(df_dates) != ncol(df)) {
        df <- left_join(df, df_dates, by = as.character(times[1])) %>% distinct()
      }
    }
  }

  # Name and type of variables
  types <- data.frame(
    name = colnames(df),
    type = unlist(lapply(lapply(df, class), `[[`, 1))
  )

  # Iterate all columns
  for (i in seq_along(df)) {
    vector_type <- types[i, "type"]
    vector_name <- as.character(types$name[i])
    vector_levels <- length(unique(df[, c(vector_name)]))
    vector_values <- df[toString(types[i, "name"])]

    # Non numeric or date/time variables
    if (!vector_type %in% c("integer", "numeric", "POSIXct", "POSIXt", "Date")) {
      # Char columns with too much variance (unique values vs total observations)
      if (vector_levels >= variance * nrow(df)) {
        no_variance <- c(no_variance, vector_name)
      }

      vector_values <- vector_values %>%
        mutate_all(as.character) %>%
        replace(., is.na(.), "NAs")
      vector_values[, 1] <- paste0(sep, vector_values[, 1])

      # Columns with 2 possible values
      if (vector_levels == 2 && !isTRUE(redundant)) {
        which <- as.character(levels(as.factor(df[, c(vector_name)]))[2])
        df[, c(vector_name)] <- as.integer(as.factor(df[, c(vector_name)])) - 1
        converted_binary <- rbind(converted_binary, vector_name)
        df <- rename_at(df, vars(vector_name), list(~ paste0(vector_name, "_", which)))
      }

      # ONE HOT ENCODING
      if (!colnames(vector_values) %in% c(converted_binary, no_variance)) {
        if (vector_levels >= 2 && !vector_name %in% converted_binary) {
          options("na.action" = "na.pass")
          reduced <- categ_reducer(
            vector_values, !!as.name(vector_name),
            top = limit,
            other_label = paste0(sep, other_label)
          )
          dummy_matx <- data.frame(model.matrix(~ . - 1, reduced))
          colnames(dummy_matx) <- paste0(vector_name, sort(unique(reduced[, 1])))
          if (isFALSE(redundant)) dummy_matx <- dummy_matx[, 1:(ncol(dummy_matx) - 1)]
          df <- cbind(df, dummy_matx)
          converted <- rbind(converted, vector_name)
        }
      }
    }
    no_need_to_convert <- rbind(no_need_to_convert, vector_name)
  }

  # Shorten up the long names of some variables
  if (trim > 0) colnames(df) <- substr(colnames(df), 1, trim)

  # Summary of transformations
  if (!quiet) {
    total_converted <- rbind(converted, converted_binary)
    if (length(total_converted) > 1) {
      message(paste(
        ">>> One Hot Encoding applied to", length(total_converted),
        "variables:", vector2text(total_converted)
      ))
    }
    if (length(no_variance) > 0 && drop) {
      no_variance <- no_variance[no_variance %in% ignore]
      if (length(no_variance) > 0) {
        message(paste0(
          ">>> Automatically dropped ", length(no_variance),
          " columns with 0% or >=", round(variance * 100),
          "% variance: ", vector2text(no_variance)
        ))
      }
    }
  }

  # Return only useful columns
  if (drop && length(c(converted, no_variance)) > 0) {
    df <- df[, c(!colnames(df) %in% c(converted, no_variance))]
  }

  # Bind ignored untouched columns and order
  order <- order[order %in% colnames(df)]
  df <- bind_cols(df, ignored) %>% select(any_of(order), everything())
  as_tibble(df)
}


####################################################################
#' One Hot Encoding for a Vector with Comma Separated Values
#'
#' This function lets the user do one hot encoding on a variable with
#' comma separated values
#'
#' @family Data Wrangling
#' @family One Hot Encoding
#' @param df Dataframe. May contain one or more columns with comma separated
#' values which will be separated as one hot encoding
#' @param ... Variables. Which variables to split into new columns?
#' @param sep Character. Which regular expression separates the elements?
#' @param noval Character. No value text
#' @param remove Boolean. Remove original variables?
#' @return data.frame on which all features are numerical by nature or
#' transformed with one hot encoding.
#' @examples
#' df <- data.frame(
#'   id = c(1:5),
#'   x = c("AA, D", "AA,B", "B,  D", "A,D,B", NA),
#'   z = c("AA+BB+AA", "AA", "BB,  AA", NA, "BB+AA")
#' )
#' ohe_commas(df, x, remove = TRUE)
#' ohe_commas(df, z, sep = "\\+")
#' ohe_commas(df, x, z)
#' @export
ohe_commas <- function(df, ..., sep = ",", noval = "NoVal", remove = FALSE) {
  vars <- quos(...)
  var <- gsub("~", "", as.character(vars))

  df <- as.data.frame(df)

  for (i in var) {
    df$temp <- as.character(df[, i])
    # Handling missingness
    df$temp[as.character(df$temp) == "" | is.na(df$temp)] <- noval
    vals <- v2t(as.character(df$temp), quotes = FALSE)
    vals <- unique(trimws(unlist(strsplit(vals, sep))))
    # aux <- sprintf("--%s--", vals)
    l <- strsplit(df$temp, sep)
    mat <- NULL
    for (i in seq_along(vals)) {
      which <- unlist(lapply(l, function(x) any(trimws(x) %in% vals[i])))
      mat <- cbind(mat, which)
    }
    colnames(mat) <- gsub('"', "", paste(var, vals, sep = "_"))
    df$temp <- NULL
    df <- cbind(df, mat)
  }
  if (remove) df <- df[, !colnames(df) %in% var]
  as_tibble(df, .name_repair = "minimal")
}


####################################################################
#' One Hot Encoding for Date/Time Variables (Dummy Variables)
#'
#' This function lets the user automatically create new columns out
#' of a dataframe or vector with date/time variables.
#'
#' @family Data Wrangling
#' @family Feature Engineering
#' @family One Hot Encoding
#' @inheritParams mp3_get
#' @param dates Vector or dataframe. Non-date/time columns will be
#' automatically ignored/extracted.
#' @param drop Boolean. Should the original date/time columns be
#' kept in the results? Only valid when input is a dataframe.
#' @param only Character or vector. Which columns do you wish to process? If
#' non are explicitly defined, all will be processed
#' @param append Boolean. Append results to existing data.frame? If FALSE,
#' only calculated values will be returned.
#' @param holidays Boolean. Include holidays as new columns?
#' @param country Character or vector. For which countries should the holidays
#' be included?
#' @param currency_pair Character. Which currency exchange do you
#' wish to get the history from? i.e, USD/COP, EUR/USD...
#' @return data.frame with additional features calculated out of time or date vectors.
#' @examples
#' df <- data.frame(
#'   dates = sample(seq(Sys.Date() - 365, Sys.Date(), by = 1), 50),
#'   times = sample(seq(Sys.time() - 1e7, Sys.time(), by = 1), 50)
#' )
#'
#' # Input as a vector or dataframe
#' date_feats(df, drop = TRUE, quiet = TRUE) %>% head(10)
#'
#' # Holidays given a date range and country
#' \dontrun{
#' hol <- date_feats(
#'   seq(Sys.Date() - 365, Sys.Date(), by = 1),
#'   holidays = TRUE,
#'   country = "Venezuela"
#' )
#' head(hol[!is.na(hol$holiday_name), ])
#' }
#' @export
date_feats <- function(dates,
                       drop = FALSE,
                       only = NA,
                       append = FALSE,
                       holidays = FALSE,
                       country = "Venezuela",
                       currency_pair = NA,
                       quiet = FALSE) {
  results <- NULL
  original <- dates
  date_cols <- df_str(dates, return = "names", quiet = TRUE)$time
  vector <- is.null(dim(dates))

  if (length(date_cols) == 0) {
    dates
  } else {
    if (!is.na(only)) {
      date_cols <- date_cols[date_cols %in% only]
    }

    iters <- ifelse(date_cols == "df", 1, length(date_cols))[1]

    if (is.na(iters)) {
      dates
    } else {
      if (!quiet) message(paste(">>> Processing", iters, "date/time columns:", vector2text(date_cols)))

      if (!"data.frame" %in% class(dates) && iters == 1) {
        dates <- data.frame(values_date = dates)
        date_cols <- "values_date"
      }

      alldates <- NULL
      if (holidays || !is.na(currency_pair)) {
        search_dates <- dates[, date_cols, drop = FALSE]
        search_dates[] <- lapply(search_dates, function(x) gsub(" .*", "", as.character(x)))
        alldates <- as.Date(unlist(search_dates, use.names = FALSE))
        alldates <- alldates[!is.na(alldates)]
      }

      holidays_dates <- NULL
      if (holidays && length(alldates) > 0) {
        years <- sort(unique(year(alldates)))
        holidays_dates <- holidays(countries = country, years)
        colnames(holidays_dates)[1] <- "values_date"
        holidays_dates$values_date <- as.character(as.Date(holidays_dates$values_date))
        cols <- paste0("values_date_holiday_", colnames(holidays_dates)[4:ncol(holidays_dates)])
        colnames(holidays_dates)[-(1:3)] <- cols
      }

      for (col in date_cols) {
        result <- dates %>% select(!!sym(col))
        values <- result[[1]]
        result$values_date <- as.character(as.Date(values))

        result$values_date_year <- year(values)
        result$values_date_month <- month(values)
        result$values_date_day <- day(values)
        result$values_date_week <- week(values)
        result$values_date_weekday <- weekdays(values, abbreviate = TRUE)
        result$values_date_weekend <- format(values, "%u") %in% c("6", "7")
        result$values_date_year_day <- as.integer(difftime(
          values, floor_date(values, unit = "year"),
          units = "day"
        ))

        if (any(grepl("POSIX", class(values)))) {
          result$values_date_hour <- hour(values)
          result$values_date_minute <- minute(values)
          result$values_date_minutes <- as.integer(difftime(
            values, floor_date(values, unit = "day"),
            units = "mins"
          ))
          result$values_date_second <- second(values)
        }

        if (!is.null(holidays_dates)) {
          result <- result %>%
            left_join(holidays_dates, by = "values_date", relationship = "many-to-many") %>%
            mutate(values_date_holiday_county = as.character(.data$values_date_holiday_county)) %>%
            mutate(across(starts_with("values_date_holiday_"), ~ replace(., is.na(.), FALSE)))
        }

        if (!is.na(currency_pair) && length(alldates) > 0) {
          currency <- get_currency(currency_pair, from = min(alldates), to = max(alldates))
          colnames(currency) <- c("values_date", paste0("values_date_", tolower(cleanText(currency_pair))))
          currency[[1]] <- as.character(currency[[1]])
          result <- result %>% left_join(currency, by = "values_date")
        }

        prefix <- if (col == "values_date") "" else paste0(col, "_")
        colnames(result)[-1] <- gsub("^values_date_", prefix, colnames(result)[-1])

        results <- results %>%
          bind_cols(result) %>%
          select(-contains("values_date"))

        if (vector) colnames(results)[1] <- "values"
      }

      if (append) {
        results <- bind_cols(original, select(results, -any_of(colnames(original))))
      }

      if (drop) {
        results <- results[, !colnames(results) %in% date_cols, drop = FALSE]
      }

      as_tibble(results)
    }
  }
}


####################################################################
#' Holidays in your Country
#'
#' This function lets the user automatically retrieve public holiday dates
#' for any country supported by the Nager.Date API. Accepts country names
#' (e.g., "Portugal") or ISO 3166-1 alpha-2 codes (e.g., "PT").
#' Thanks to \href{https://date.nager.at}{Nager.Date}!
#'
#' @family Data Wrangling
#' @family Feature Engineering
#' @family Scrapper
#' @family One Hot Encoding
#' @inheritParams mp3_get
#' @param years Character or vector. For which year(s) do you wish to import
#' holiday dates?
#' @param countries Character or vector. For which country(ies) should the
#' holidays be imported? Accepts country names or ISO 3166-1 alpha-2 codes.
#' @param include_regions Boolean. Default FALSE. If TRUE, for countries with
#' internal subdivisions, it will provide details on which sub-state the found
#' holidays apply.
#' @return \code{data.frame} with holidays data for given \code{countries} and \code{years}.
#' @examples
#' \donttest{
#' holidays(countries = "Argentina")
#' year <- as.integer(format(Sys.Date(), format = "%Y"))
#' holidays(countries = c("Spain", "Venezuela"), years = year)
#' holidays(countries = "Germany", include_regions = TRUE)
#' holidays(countries = "PT") # Also accepts ISO country codes
#' }
#' @export
holidays <- function(countries = "Venezuela",
                     years = year(Sys.Date()),
                     quiet = FALSE,
                     include_regions = FALSE) {
  if (!haveInternet()) {
    message("No internet connection...")
    invisible(NULL)
  } else {
    results <- NULL

    # Resolve country names to ISO codes via Nager.Date API
    available <- tryCatch(
      jsonlite::fromJSON("https://date.nager.at/api/v3/AvailableCountries"),
      error = function(e) {
        stop("Failed to fetch available countries from Nager.Date API: ", e$message)
      }
    )

    country_codes <- vapply(countries, function(cntry) {
      # If already a valid 2-letter ISO code, use directly
      if (nchar(cntry) == 2 && toupper(cntry) %in% available$countryCode) {
        return(toupper(cntry))
      }
      # Match by name (case-insensitive)
      match_idx <- which(tolower(available$name) == tolower(cntry))
      if (length(match_idx) == 1) return(available$countryCode[match_idx])
      # Partial match fallback
      match_idx <- grep(cntry, available$name, ignore.case = TRUE)
      if (length(match_idx) >= 1) return(available$countryCode[match_idx[1]])
      warning(
        "Country '", cntry, "' not found in Nager.Date API. ",
        "Use a supported country name or ISO code."
      )
      NA_character_
    }, character(1))

    combs <- expand.grid(year = years, country = countries, stringsAsFactors = FALSE)
    combs$code <- country_codes[combs$country]

    for (i in seq_len(nrow(combs))) {
      if (is.na(combs$code[i])) next
      if (!quiet) {
        message(
          ">>> Extracting ", combs$country[i],
          "'s holidays for ", combs$year[i]
        )
      }
      url <- sprintf(
        "https://date.nager.at/api/v3/PublicHolidays/%s/%s",
        combs$year[i], combs$code[i]
      )
      hol_data <- tryCatch(
        jsonlite::fromJSON(url),
        error = function(e) {
          warning(
            "Failed to fetch holidays for ", combs$country[i],
            " (", combs$year[i], "): ", e$message
          )
          NULL
        }
      )
      if (is.null(hol_data) || nrow(hol_data) == 0) next

      # Flatten the types list-column into a single string per row
      hol_types <- vapply(hol_data$types, function(x) {
        paste(x, collapse = ", ")
      }, character(1))

      result <- data.frame(
        holiday = as.Date(hol_data$date),
        holiday_name = hol_data$name,
        holiday_type = hol_types,
        stringsAsFactors = FALSE
      )

      if (include_regions) {
        result$holiday_details <- vapply(hol_data$counties, function(x) {
          if (is.null(x)) NA_character_ else paste(x, collapse = ", ")
        }, character(1))
      }

      result <- result %>%
        mutate(
          national = grepl("Public", .data$holiday_type) & hol_data$global,
          observance = grepl("Observance", .data$holiday_type),
          bank = grepl("Bank", .data$holiday_type),
          nonwork = grepl("Optional", .data$holiday_type),
          season = grepl("School|Authorities", .data$holiday_type),
          hother = !grepl("Public|Observance|Bank|Optional", .data$holiday_type)
        )

      if (length(unique(countries)) > 1L) {
        result$country <- combs$country[i]
      }
      result$county <- combs$country[i]
      results <- bind_rows(results, result)
    }

    if (is.null(results) || nrow(results) == 0) {
      return(as_tibble(data.frame(
        holiday = as.Date(character()),
        holiday_name = character(),
        holiday_type = character(),
        stringsAsFactors = FALSE
      )))
    }

    results %>%
      filter(!is.na(.data$holiday)) %>%
      cleanNames() %>%
      as_tibble()
  }
}

####################################################################
#' Build a Personal Investing Portfolio Report
#'
#' \code{stocks_file()} lets the user download his personal Excel with his
#' Portfolio's data, locally or from Dropbox.
#'
#' @family Investment
#' @family Credentials
#' @inheritParams cache_write
#' @param file Character. Import an Excel file, local or from URL.
#' @param creds Character. Dropbox's credentials (see \code{get_creds()})
#' @param auto Boolean. Automatically use my local personal file? You might want
#' to set in into your .Renviron \code{LARES_PORTFOLIO=~/dir/to/your/file.xlsx} so you
#' can leave all other parameters as \code{NA} and use it every time.
#' @param sheets Character Vector. Names of each sheet containing Portfolio summary,
#' Cash, and Transactions information. Please, keep the order of these tabs.
#' @param keep_old Boolean. Include sold tickers even though not currently in portfolio?
#' @param cache Boolean. Use daily cache if available?
#' @return List with portfolio, transactions, and cash data.frames.
#' @examples
#' \dontrun{
#' # Load lares dummy portfolio XLSX
#' file <- system.file("inst/docs", "dummyPortfolio.xlsx", package = "lares")
#' df <- stocks_file(
#'   file = file,
#'   sheets = c("Portafolio", "Fondos", "Transacciones"),
#'   keep_old = FALSE
#' )
#' }
#' @export
#' @rdname stocks_report
stocks_file <- function(file = NA,
                        creds = NA,
                        auto = TRUE,
                        sheets = c("Portafolio", "Fondos", "Transacciones"),
                        keep_old = TRUE,
                        cache = TRUE,
                        quiet = FALSE,
                        ...) {
  cache_file <- c(as.character(Sys.Date()), "stocks_file")
  if (cache_exists(cache_file) && cache) {
    cache_read(cache_file, quiet = quiet)
  } else {
    processFile <- function(file, sheets = NULL, keep_old = TRUE) {
      mylist <- lapply(sheets, function(x) {
        as_tibble(read.xlsx(
          file,
          sheet = x,
          skipEmptyRows = TRUE,
          detectDates = TRUE
        ))
      })
      if (length(mylist) == 3) {
        names(mylist) <- c("port", "cash", "trans")
        mylist$port$StartDate <- try(as.Date(mylist$port$StartDate, origin = "1889-12-31", ...))
        mylist$trans$Date <- try(as.Date(mylist$trans$Date, origin = "1889-12-31", ...))
        mylist$cash$Date <- try(as.Date(mylist$cash$Date, origin = "1889-12-31", ...))
        if ("Value" %in% colnames(mylist$trans)) {
          mylist$trans <- rename(mylist$trans, Each = .data$Value, Invested = .data$Amount)
        }
        if (!keep_old) mylist$port <- try(mylist$port[mylist$port$Stocks != 0, ])
        mylist <- list("portfolio" = mylist$port, "transactions" = mylist$trans, "cash" = mylist$cash)
      }
      if (length(mylist) == 1) {
        mylist <- mylist[[1]]
      }
      mylist
    }

    # FOR PERSONAL USE
    local <- Sys.info()
    if (auto && Sys.getenv("LARES_PORTFOLIO") != "") {
      if (!quiet) message("Using BL's local file...")
      local <- Sys.getenv("LARES_PORTFOLIO")
      results <- processFile(local, sheets, keep_old)
    } else {
      # FOR EVERYONE'S USE
      if (!is.na(file)) {
        if (file.exists(file) || is_url(file)) {
          results <- processFile(file, sheets, keep_old)
        } else {
          stop("Error: that file doesn't exist or it's not in your working directory!")
        }
      } else {
        # FOR DROPBOX'S USE
        file <- paste0(tempdir(), "/Portfolio LC.xlsx")
        db_download(
          query = "Portfolio LC.xlsx",
          local_path = file,
          xlsx = FALSE, # Do not import as Excel, just download
          token_dir = creds
        )
        results <- processFile(file, sheets, keep_old = keep_old)
      }
    }
    if (length(results) == 3) {
      attr(results$portfolio, "type") <- "stocks_file_portfolio"
      attr(results$transactions, "type") <- "stocks_file_transactions"
      attr(results$cash, "type") <- "stocks_file_cash"
    }
    attr(results, "type") <- "stocks_file"
    cache_write(results, cache_file, quiet = TRUE, ...)
    results
  }
}

####################################################################
#' Download Stocks Historical and Current Values
#'
#' \code{stocks_quote()} lets the user download stocks live data.
#'
#' @family Investment
#' @family Scrapper
#' @inheritParams cache_write
#' @param symbols Character Vector. List of symbols to download historical data.
#' @return data.frame with Symbol, Type of stock, Quote time, current value,
#' Daily Change, Market, and Symbol Name.
#' @examples
#' \dontrun{
#' # Multiple quotes at the same time
#' stocks_quote(c("VTI", "VOO", "TSLA"))
#' }
#' @export
#' @rdname stocks_hist
stocks_quote <- function(symbols, ...) {
  ret <- noret <- NULL
  try_require("quantmod")
  for (i in seq_along(symbols)) {
    z <- try(data.frame(getQuote(symbols[i], ...)))
    if ("try-error" %in% class(z)) {
      invisible(ret)
    }
    if (length(z) > 0) {
      z <- data.frame(Symbol = symbols[i], z)
      ret <- bind_rows(ret, z)
    } else {
      noret <- rbind(noret, symbols[i])
    }
  }
  if (length(noret) > 0) {
    message(paste("No results for", vector2text(noret)))
  }
  if (length(ret) > 0) {
    colnames(ret) <- c(
      "Symbol", "QuoteTime", "Value", "DailyChange",
      "DailyChangeP", "Open", "High", "Low", "Volume"
    )
    ret <- as_tibble(ret) %>%
      mutate(QuoteTime = as.POSIXct(
        .data$QuoteTime,
        origin = "1970-01-01 00:00:00"
      ))
    row.names(ret) <- NULL
  }
  ret
}

####################################################################
#' Download Stocks Historical Data
#'
#' \code{stocks_hist()} lets the user download stocks historical data.
#'
#' @inheritParams stocks_file
#' @inheritParams cache_write
#' @param from,to Date. Dates for range. If not set, 1 year will be downloaded.
#' Do use more than 4 days or will be over-written.
#' @param today Boolean. Do you wish to add today's live quote? This will happen
#' only if to value is the same as today's date
#' @param tax Numeric. How much [0-99] of your dividends are gone with taxes?
#' @param parg Boolean. Personal argument. Used to personalize stuff, in this
#' case, taxes changed from A to B in given date (hard-coded)
#' @examples
#' \dontrun{
#' df <- stocks_hist(symbols = c("VTI", "META", "FIW"), from = Sys.Date() - 180)
#' print(head(df))
#' plot(df)
#' }
#' @rdname stocks_hist
#' @export
stocks_hist <- function(symbols = c("VTI", "META"),
                        from = Sys.Date() - 365,
                        to = Sys.Date(),
                        today = TRUE,
                        tax = 15,
                        parg = FALSE,
                        cache = TRUE,
                        quiet = FALSE,
                        ...) {
  cache_file <- c(as.character(Sys.Date()), "stocks_hist")
  if (cache_exists(cache_file) && cache) {
    cache_read(cache_file, quiet = quiet)
  } else {
    try_require("quantmod")
    data <- divs <- NULL

    if (!any(is.na(symbols))) {
      if (length(from) != length(symbols)) {
        from <- rep(from[1], length(symbols))
      }
      for (i in seq_along(symbols)) {
        # Daily quotes (except today)
        symbol <- as.character(symbols[i])
        if (as.Date(from[i]) > (Sys.Date() - 4)) from[i] <- Sys.Date() - 4
        start_date <- as.character(from[i])
        values <- try(suppressWarnings(data.frame(getSymbols(
          symbol,
          env = NULL, from = start_date, to = to, src = "yahoo"
        ))))
        if ("try-error" %in% class(values)) next
        values <- cbind(row.names(values), as.character(symbol), values)
        colnames(values) <- c("Date", "Symbol", "Open", "High", "Low", "Close", "Volume", "Adjusted")
        values <- mutate(values, Adjusted = rowMeans(select(values, .data$High, .data$Close), na.rm = TRUE))
        row.names(values) <- NULL
        # if (as.Date(from[i]) > (Sys.Date() - 4))
        #   values <- head(values, 1)

        # Add right now's data
        if (today && to == Sys.Date()) {
          now <- stocks_quote(symbol, ...)
          # Append to historical data / replace most recent
          if (length(now) > 0) {
            now <- data.frame(
              Date = as.character(as.Date(now$QuoteTime)),
              Symbol = symbol,
              Open = now$Value, High = now$Value,
              Low = now$Value, Close = now$Value,
              Volume = NA, Adjusted = now$Value
            )
            values <- filter(values, as.Date(.data$Date) != as.Date(now$Date))
            values <- rbind(values, now)
          }
        }
        # Append to other symbols' data
        data <- rbind(data, values)

        # Dividends (with Taxes deduction)
        d <- suppressWarnings(getDividends(
          as.character(symbol),
          from = start_date, split.adjust = FALSE
        ))
        if (isTRUE(nrow(d) > 0)) {
          div <- data.frame(
            Symbol = rep(symbol, nrow(d)),
            Date = ymd(row.names(data.frame(d))),
            Div = as.vector(d),
            DivTaxP = tax / 100
          )
          if (parg) {
            div <- mutate(div, DivTaxP = ifelse(
              .data$Date < as.Date("2020-03-03"),
              (30 / 100), # Interactive Brokers (Colombia)
              (15 / 100) # Schwab (Argentina/Veneezuela)
            ))
          }
          div <- mutate(div,
            DivReal = as.vector(d) * (1 - .data$DivTaxP),
            Symbol = as.character(.data$Symbol)
          )
          divs <- rbind(divs, div)
        }

        if (!quiet && length(symbols) > 1) {
          info <- paste(symbol, "since", start_date, "   ")
          statusbar(i, length(symbols), info)
        }
      }
    } else {
      message("You need to define which stocks to bring. Use the 'symbols=' parameter.")
    }

    results <- data %>%
      select(.data$Date, .data$Symbol, .data$Close) %>%
      rename(Value = .data$Close) %>%
      filter(.data$Value > 0) %>%
      mutate(
        Date = as.Date(.data$Date),
        Symbol = as.character(.data$Symbol)
      )
    if (length(divs) > 0) {
      results <- left_join(results, divs, by = c("Date", "Symbol"))
    }
    results <- results %>%
      replace(is.na(.), 0) %>%
      arrange(desc(.data$Date))

    results <- as_tibble(results)
    attr(results, "type") <- "stocks_hist"
    class(results) <- c("stocks_hist", class(results))
    cache_write(results, cache_file, quiet = TRUE)
    results
  }
}

#' @rdname stocks_hist
#' @aliases stocks_hist
#' @param x stocks_hist object
#' @param type Integer. Select type of plot.
#' @export
plot.stocks_hist <- function(x, type = 1, ...) {
  if (!inherits(x, "stocks_hist")) stop("Object must be class stocks_hist")
  if (type == 1) {
    group_by(x, .data$Symbol) %>%
      mutate(
        RelValue = (.data$Value - last(.data$Value)),
        Growth = .data$RelValue / last(.data$Value),
        Total = first(.data$Growth)
      ) %>%
      ungroup() %>%
      arrange(desc(.data$Total)) %>%
      mutate(
        Symbol = sprintf("%s (%s%%)", .data$Symbol, round(100 * .data$Total, 2)),
        Symbol = factor(.data$Symbol, unique(.data$Symbol))
      ) %>%
      ggplot(aes(x = .data$Date, y = .data$Growth, colour = .data$Symbol)) +
      geom_line(alpha = 0.8) +
      theme_lares(pal = 2, legend = "left") +
      scale_y_percent(position = "right") +
      labs(
        title = "Relative Growth since Origin", colour = NULL,
        subtitle = sprintf("%s - %s", min(x$Date), max(x$Date)),
        x = NULL, y = "Value Growth [%]"
      )
  }
}


####################################################################
#' Daily Stocks Dataframe
#'
#' \code{daily_stocks()} creates a dataframe with all relevant metrics and values,
#' for each ticker or symbol, for every day since inception.
#'
#' @inheritParams cache_write
#' @param hist Dataframe. Result from \code{stocks_hist()}
#' @param trans Dataframe. Result from \code{stocks_file()$transactions}
#' @param tickers Dataframe. Result from \code{stocks_file()$portfolio}
#' @param window Character. Choose any of: "1W", "1M", "6M", "1Y", "YTD", "5Y", "MAX"
#' @return data.frame. Processed at date and symbol level.
#' @export
#' @rdname stocks_report
daily_stocks <- function(hist, trans, tickers = NA, window = "MAX", ...) {
  check_attr(hist, check = "stocks_hist")
  check_attr(trans, check = "stocks_file_transactions")
  if (!is.na(tickers)[1]) {
    check_attr(tickers, check = "stocks_file_portfolio")
  }
  check_opts(window, c("1W", "1M", "6M", "1Y", "YTD", "5Y", "MAX"))

  with_div <- "DivReal" %in% colnames(hist)

  daily <- expand.grid(Date = unique(hist$Date), Symbol = unique(hist$Symbol)) %>%
    left_join(hist, c("Date", "Symbol")) %>%
    left_join(trans, by = c("Date", "Symbol")) %>%
    select(-.data$CODE, -.data$Description) %>%
    mutate(Date = as.Date(.data$Date)) %>%
    arrange(desc(.data$Date), .data$Symbol) %>%
    group_by(.data$Symbol) %>%
    tidyr::fill(.data$Value, .direction = "up") %>%
    replace(is.na(.), 0) %>%
    arrange(.data$Date) %>%
    {
      if (!with_div) mutate(., DivReal = 0) else .
    } %>%
    group_by(.data$Symbol) %>%
    mutate(
      CumQuant = cumsum(.data$Quant),
      CumInvested = cumsum(.data$Invested),
      CumCost = cumsum(.data$Cost),
      CumValue = .data$CumQuant * .data$Value,
      CumROI = 100 * (.data$CumValue / .data$CumInvested - 1),
      DivPreTax = .data$Div * .data$CumQuant,
      Dividend = .data$DivReal * .data$CumQuant,
      DifUSD = .data$CumValue - .data$Invested - lag(.data$CumValue),
      CumDivPreTax = cumsum(.data$DivPreTax),
      CumDividend = cumsum(.data$Dividend)
    ) %>%
    select(
      .data$Date, .data$Symbol, .data$Value, .data$Quant, .data$Each,
      .data$Invested, .data$Cost, .data$Dividend,
      .data$CumValue, .data$CumInvested, .data$CumROI, .data$CumQuant,
      .data$DifUSD, .data$CumDivPreTax, .data$CumDividend, .data$CumCost
    ) %>%
    group_by(.data$Date, .data$Symbol) %>%
    slice(1) %>%
    arrange(desc(.data$Date), desc(.data$CumValue)) %>%
    ungroup()

  if ("Type" %in% colnames(tickers)) {
    tickers_structure <- c("Symbol", "Type")
    if (!all(tickers_structure %in% colnames(tickers))) {
      stop(paste(
        "The structure of the 'tickers' table should be:",
        paste(shQuote(tickers_structure), collapse = ", ")
      ))
    }
    daily <- left_join(daily, select(tickers, .data$Symbol, .data$Type), "Symbol")
  } else {
    daily$Type <- "No type"
  }

  levs <- unique(daily$Symbol[daily$Date == max(daily$Date)])
  daily <- daily %>%
    mutate(Symbol = factor(.data$Symbol, levels = levs)) %>%
    .filter_window(window) %>%
    group_by(.data$Date) %>%
    mutate(
      wt_value = weighted_value(.data$Value, n = .data$Quant),
      wt_total = .data$wt_value * .data$Quant,
      wt = signif(100 * .data$wt_total / sum(.data$wt_total), 4)
    ) %>%
    ungroup()

  attr(daily, "type") <- "daily_stocks"
  daily
}

####################################################################
#' Daily Portfolio Dataframe
#'
#' \code{daily_portfolio()} creates a data.frame with all relevant
#' metrics and values, for the overall portfolio, for every day
#' since inception.
#'
#' @inheritParams daily_stocks
#' @param cash Dataframe. Result from \code{stocks_file()$cash}
#' @param cash_fix Numeric. If, for some reason, you need to fix your
#' cash amount for all reports, set the amount here
#' @return data.frame. Processed at date and portfolio level.
#' @export
#' @rdname stocks_report
daily_portfolio <- function(hist, trans, cash, cash_fix = 0, window = "MAX") {
  check_attr(hist, check = "stocks_hist")
  check_attr(trans, check = "stocks_file_transactions")
  check_attr(cash, check = "stocks_file_cash")

  temp <- expand.grid(
    Date = unique(hist$Date),
    Symbol = unique(hist$Symbol)
  ) %>%
    left_join(daily_stocks(hist, trans), c("Date", "Symbol")) %>%
    mutate(Date = as.Date(.data$Date)) %>%
    arrange(desc(.data$Date), .data$Symbol) %>%
    filter(!is.na(.data$CumQuant)) %>%
    group_by(.data$Symbol) %>%
    tidyr::fill(.data$Value, .data$CumInvested, .data$CumValue,
      .data$CumDividend, .data$CumCost,
      .direction = "up"
    ) %>%
    group_by(.data$Date) %>%
    summarise_if(is.numeric, list(~ sum(., na.rm = TRUE))) %>%
    arrange(.data$Date) %>%
    mutate(
      ROI = 100 * (.data$CumValue / .data$CumInvested - 1),
      DifUSD = .data$CumValue - .data$Invested - lag(.data$CumValue)
    ) %>%
    select(
      .data$Date, .data$CumInvested, .data$CumValue, .data$DifUSD,
      .data$ROI, .data$Invested, .data$CumCost, .data$CumDividend, .data$Dividend
    )

  days <- tibble(Date = as.Date(min(hist$Date):Sys.Date(), origin = "1970-01-01")) %>%
    left_join(temp, "Date") %>%
    tidyr::fill(.data$ROI, .data$CumInvested, .data$CumValue,
      .data$CumDividend, .data$CumCost,
      .direction = "up"
    ) %>%
    mutate(Date = as.character(.data$Date)) %>%
    left_join(mutate(cash, Date = as.character(.data$Date)) %>% select(.data$Date, .data$Cash), "Date") %>%
    mutate(Date = as.Date(.data$Date)) %>%
    replace(is.na(.), 0) %>%
    arrange(.data$Date) %>%
    mutate(
      Cash = as.numeric(ifelse(is.na(.data$Cash), 0, .data$Cash)),
      CumCash = round(
        cumsum(.data$Cash) + cumsum(.data$Dividend) + cash_fix -
          .data$CumInvested - .data$CumCost
      )
    ) %>%
    filter(.data$CumInvested != 0) %>%
    arrange(desc(.data$Date)) %>%
    ungroup()

  # One row per day only
  ret <- days %>%
    group_by(
      .data$Date, .data$CumInvested, .data$CumValue,
      .data$CumCost, .data$CumDividend, .data$ROI
    ) %>%
    summarise_if(is.numeric, sum) %>%
    arrange(desc(.data$Date)) %>%
    ungroup() %>%
    mutate(Portfolio = .data$CumCash + .data$CumValue) %>%
    .filter_window(window) %>%
    ungroup()

  attr(ret, "type") <- "daily_portfolio"
  ret
}


####################################################################
#' Calculate weighted stock values using FIFO/LIFO
#'
#' @inheritParams cache_write
#' @param value Numeric vector. Representing the values of the stock.
#' @param n Numeric vector. Representing the volume of the operation.
#' Positive for 'Buy' and negative for 'Sale'.
#' @param technique Character. Pick any of FIFO or LIFO, or NULL to skip.
#' @param n_stocks Integer. Specify the number of stocks to consider. By
#' default will sum positive values of \code{n}.
#' @param buy_only Boolean. Consider only buy (positive) values?
#' @param type Integer. 1 for returning the value, 2 for returning the
#' data.frame with the details ("df" attribute)
#' @return The calculated weighted mean value.
#' @examples
#' values <- c(10, 20, 30, 40, 50)
#' weights <- c(2, 3, -4, 5, 6)
#' mean(values)
#' weighted_value(values)
#' weighted.mean(values, weights)
#' weighted_value(values, weights, buy_only = FALSE)
#' # Using FIFO and LIFO
#' weighted_value(values, weights, "FIFO")
#' weighted_value(values, weights, "LIFO", n_stocks = 8)
#' @export
weighted_value <- function(value,
                           n = rep(1, length(value)),
                           technique = NULL,
                           n_stocks = NULL,
                           buy_only = TRUE,
                           type = 1,
                           ...) {
  check_opts(technique, c("FIFO", "LIFO"))
  stopifnot(length(value) == length(n))
  df <- data.frame(value = value, n = n, total = n)
  if (buy_only) df <- df[df$n > 0, ]
  if (is.null(n_stocks)) n_stocks <- sum(n, na.rm = TRUE)
  if (nrow(df) > 0) {
    df$id <- seq_len(nrow(df))
    if ("LIFO" %in% technique) {
      df <- df %>%
        arrange(desc(.data$id)) %>%
        mutate(cum = cumsum(.data$n), n_stocks = n_stocks) %>%
        rowwise() %>%
        mutate(
          total = min(.data$n, .data$n_stocks - .data$cum + .data$n),
          total = ifelse(.data$total < 0, 0, .data$total)
        ) %>%
        arrange(.data$id) %>%
        data.frame()
    } else if ("FIFO" %in% technique) {
      df <- df %>%
        mutate(cum = cumsum(.data$n), n_stocks = n_stocks) %>%
        rowwise() %>%
        mutate(
          total = min(.data$n, .data$n_stocks - .data$cum + .data$n),
          total = ifelse(.data$total < 0, 0, .data$total)
        ) %>%
        data.frame()
    }
    ret <- sum(df$value * df$total, na.rm = TRUE) / sum(df$total, na.rm = TRUE)
    attr(ret, "df") <- select(df, -any_of(c("id", "cum")))
    if (type == 2) {
      attr(ret, "df")
    } else {
      ret
    }
  } else {
    0
  }
}

################# PLOTTING FUNCTIONS #################

####################################################################
#' Investing Portfolio Reporting Plots
#'
#' \code{splot_summary()} plots a summary for the whole portfolio, showing
#' how much have you invested, how much has each ticker changed, etc.
#'
#' @family Investment
#' @param p Dataframe. Result from \code{daily_portfolio()}
#' @param s Dataframe. Result from \code{daily_stocks()}
#' @param save Boolean. Save plot into a local file?
#' @return ggplot object
#' @export
#' @rdname stocks_plots
splot_summary <- function(p, s, save = FALSE) {
  check_attr(p, check = "daily_portfolio")
  check_attr(s, check = "daily_stocks")

  today <- filter(p, .data$Date == max(p$Date))
  summary <- rbind(
    paste0("Portfolio: $", formatNum(today$Portfolio, 0), " | ", today$Date),
    paste0("Stocks: $", formatNum(today$CumValue, 0), " & Cash: $", formatNum(today$CumCash, 0)),
    paste0(
      "ROI: ", formatNum(today$ROI, 2), "% ($",
      formatNum(today$CumValue - today$CumInvested, 0), ")"
    ),
    paste0(
      "Dividends: $", formatNum(today$CumDividend, 0), " | Expenses: $",
      formatNum(today$CumCost, 0)
    )
  )

  today <- s %>%
    group_by(.data$Symbol) %>%
    filter(.data$Date == max(p$Date)) %>%
    filter(.data$CumQuant > 0) %>%
    ungroup()
  tops <- max(rbind(today$CumInvested, today$CumValue))
  plot <- today %>%
    mutate(
      shapeflag = ifelse(.data$CumROI < 0, 25, 24),
      box = -tops / 5.5,
      Symbol = factor(
        paste0(
          .data$Symbol, " (",
          formatNum(100 * .data$CumValue / sum(.data$CumValue)), "%)"
        ),
        levels = paste0(.data$Symbol, " (", formatNum(100 * .data$CumValue / sum(.data$CumValue)), "%)")
      ),
      DifUSD = .data$CumValue - .data$CumInvested,
      DifPer = 100 * .data$DifUSD / .data$CumInvested
    ) %>%
    ggplot(aes(x = reorder(.data$Symbol, .data$CumInvested))) +
    geom_hline(yintercept = 0, colour = "black") +
    geom_col(aes(y = .data$CumInvested, fill = .data$Symbol, group = 1)) +
    geom_col(aes(y = .data$CumInvested + .data$DifUSD, fill = .data$Symbol), alpha = 0.5) +
    geom_col(aes(y = .data$box), fill = "grey", alpha = 0.5) +
    geom_point(aes(y = .data$CumInvested + .data$DifUSD, shape = .data$shapeflag), colour = "black") +
    scale_shape_identity() +
    geom_text(
      aes(
        label = paste0("$", formatNum(.data$DifUSD, 0)),
        y = .data$CumInvested + .data$DifUSD
      ),
      size = 3, hjust = -.2, vjust = -0.2
    ) +
    geom_text(
      aes(
        label = paste0(round(.data$DifPer, 1), "%"),
        y = .data$CumInvested + .data$DifUSD
      ),
      size = 2.9, hjust = -.2, vjust = 1.2
    ) +
    geom_text(aes(label = paste0("$", formatNum(.data$CumValue, 0)), y = .data$box),
      size = 3, hjust = -.1, vjust = -0.2
    ) +
    geom_text(aes(
      label = paste0(.data$CumQuant, " @$", formatNum(.data$CumValue / .data$CumQuant, 1)),
      y = .data$box
    ), size = 2, hjust = -.1, vjust = 1.5) +
    geom_text(
      aes(
        label = paste0("$", formatNum(.data$CumInvested, 1)),
        y = 0, colour = .data$Symbol
      ),
      size = 2, hjust = 0, vjust = -0.2
    ) +
    geom_text(
      aes(
        label = paste0("@$", formatNum(.data$CumInvested / .data$CumQuant, 1)),
        y = 0, x = .data$Symbol, colour = .data$Symbol
      ),
      size = 2, hjust = 0, vjust = 1.5
    ) +
    annotate("label",
      x = length(unique(today$CumQuant)) * 0.25, y = tops * 0.6,
      label = vector2text(summary, "\n", quotes = F), size = 3.5, hjust = 0, alpha = 0.55
    ) +
    scale_y_comma(limits = c(NA, tops * 1.12), expand = c(0, 0)) +
    labs(y = NULL, x = NULL, title = "Stocks Distribution and Growth") +
    guides(fill = "none", colour = "none") +
    coord_flip() +
    theme_lares(pal = 1)

  if (save) {
    plot <- plot +
      ggsave("portf_stocks_change.png", width = 8, height = 8, dpi = 300)
  }

  plot
}


####################################################################
#' Portfolio Plots: Daily Change
#'
#' \code{splot_change()} plots each stock's change through history, since
#' inception, with weighted attributions or absolute values.
#'
#' @inheritParams splot_summary
#' @inheritParams stocks_file
#' @param rel Boolean. Relative delta values (weighted with portfolio)?
#' If not, absolute monetary delta values.
#' @param group Boolean. Group stocks by stocks type?
#' @param n_days Integer. How many days back you want to see?
#' sold entirely?
#' @export
#' @rdname stocks_plots
splot_change <- function(p, s,
                         rel = TRUE,
                         group = FALSE,
                         n_days = 365,
                         keep_old = FALSE,
                         save = FALSE) {
  check_attr(p, check = "daily_portfolio")
  check_attr(s, check = "daily_stocks")

  current_stocks <- as.character(s$Symbol[s$Date == max(s$Date) & s$CumQuant > 0])
  start_stocks <- as.character(s$Symbol[s$Date == max(s$Date) - n_days & s$CumQuant > 0])

  d <- s %>%
    filter(.data$CumQuant > 0) %>%
    arrange(.data$Date) %>%
    {
      if (!is.na(n_days)) filter(., .data$Date >= max(s$Date) - n_days) else .
    } %>%
    select(
      .data$Date, .data$Symbol, .data$Value, .data$CumQuant, .data$CumValue,
      .data$CumInvested, .data$Invested, .data$wt
    ) %>%
    {
      if (!keep_old) filter(., as.character(.data$Symbol) %in% current_stocks) else .
    } %>%
    group_by(.data$Symbol) %>%
    mutate(first_date = .data$Date == min(.data$Date)) %>%
    # run_fix = .data$Symbol %in% start_stocks) %>%
    # mutate(CumValue = .data$CumValue - .data$CumValue[.data$first_date] * as.integer(.data$run_fix)) %>%
    mutate(CumValue = .data$CumValue - .data$CumValue[.data$first_date]) %>%
    {
      if (rel) {
        mutate(., Hist = .data$CumValue / .data$CumInvested)
      } else {
        mutate(., Hist = .data$CumValue)
      }
    } %>%
    mutate(Hist = ifelse(.data$CumQuant == 0, NA, .data$Hist)) %>%
    mutate(BuySell = case_when(
      .data$Invested > 0 ~ "Bought",
      .data$Invested < 0 ~ "Sold",
    )) %>%
    mutate(BuySellShape = case_when(
      .data$BuySell == "Bought" ~ "24",
      .data$BuySell == "Sold" ~ "25"
    )) %>%
    mutate(Hist = ifelse(.data$CumQuant > 0, .data$Hist, tail(.data$Hist, 1))) %>%
    mutate(Symbol = factor(.data$Symbol, levels = unique(s$Symbol))) %>%
    filter(.data$CumQuant > 0)

  # Plot auxiliary stuff
  days <- as.integer(difftime(range(d$Date)[2], range(d$Date)[1], units = "days"))
  plot <- ggplot(d, aes(x = .data$Date, y = .data$Hist)) +
    geom_vline(xintercept = max(s$Date), linetype = "dashed", alpha = 0.7) +
    ylab(glued("Total changed [{x}]", x = ifelse(rel, "%", "$"))) +
    geom_hline(yintercept = 0, alpha = 0.8, colour = "black") +
    geom_line(aes(colour = .data$Symbol),
      alpha = 0.9, size = 0.5, na.rm = TRUE
    ) +
    geom_point(
      data = filter(d, !is.na(.data$BuySellShape)),
      aes(shape = .data$BuySellShape, colour = .data$Symbol),
      fill = "grey10", na.rm = TRUE
    ) +
    ggplot2::scale_shape_manual(values = c(24, 25), labels = c("Bought", "Sold")) +
    scale_size(range = c(0, 3.2)) +
    guides(size = "none", colour = "none", fill = "none") +
    xlim(min(d$Date), max(d$Date) + round(days * 0.06)) +
    labs(
      title = glued("Weighted Stocks Daily Change [{x}]", x = ifelse(rel, "%", "$")),
      x = NULL, colour = NULL,
      subtitle = sprintf(
        "%s delta since %s or first transaction",
        ifelse(rel, "Relative total", "Total"), min(d$Date)
      ),
      shape = "Transaction"
    ) +
    theme_lares(pal = 2, legend = "top") +
    geom_text(
      data = d[which(d$Date == max(d$Date)), ],
      aes(label = .data$Symbol, colour = .data$Symbol),
      hjust = -0.1, size = 3
    )

  if (group) plot <- plot + facet_grid(.data$Type ~ ., scales = "free", switch = "both")
  if (!is.na(n_days)) plot <- plot + labs(caption = glued("Last {n_days} days"))
  if (rel) {
    plot <- plot + scale_y_percent(position = "right")
  } else {
    plot <- plot + scale_y_formatNum(position = "right", abbr = TRUE, pos = "$")
  }
  if (save) {
    plot <- plot +
      ggsave("portf_stocks_histchange.png", width = 8, height = 5, dpi = 300)
  }
  plot
}


####################################################################
#' Portfolio Plots: Growth (Cash + Invested)
#'
#' \code{splot_growth()} plots your portfolio's growth, in cash and investment,
#' since inception.
#'
#' @inheritParams splot_summary
#' @export
#' @rdname stocks_plots
splot_growth <- function(p, save = FALSE) {
  check_attr(p, check = "daily_portfolio")
  newp <- p %>%
    ungroup() %>%
    mutate(
      CumValue = .data$CumValue - tail(.data$CumValue, 1),
      CumCash = .data$CumCash - tail(.data$CumCash, 1),
      Portfolio = .data$Portfolio - tail(.data$Portfolio, 1)
    )
  caption <- paste0(
    "Portfolio: $", formatNum(newp$Portfolio[1], 0), " (",
    as.integer(range(p$Date)[2] - range(p$Date)[1]), " days)",
    "\nInvested: $", formatNum(newp$CumValue[1], 0), " (",
    round(100 * newp$CumValue[1] / newp$Portfolio[1], 1), "%)"
  )
  aux <- rbind(
    data.frame(Date = newp$Date, Amount = newp$CumCash, Type = "Cash"),
    data.frame(Date = newp$Date, Amount = newp$CumValue, Type = "Stocks")
  ) %>%
    left_join(select(newp, .data$Date, .data$Cash) %>%
      filter(.data$Cash != 0) %>%
      mutate(Type = "Cash"), by = c("Date", "Type")) %>%
    group_by(.data$Type) %>%
    mutate(Amount = .data$Amount - tail(.data$Amount, 1)) %>%
    ungroup() %>%
    group_by(.data$Date) %>%
    mutate(Portfolio = sum(.data$Amount))
  days <- nrow(aux) / 2

  plot <- ggplot(aux, aes(x = .data$Date, y = .data$Amount)) +
    geom_area(aes(y = .data$Amount, fill = .data$Type), position = "stack", alpha = 0.7) +
    labs(title = "Daily Total Portfolio Value", y = NULL, x = NULL, fill = "") +
    scale_y_dollar(
      position = "right",
      breaks = round(seq(min(aux$Amount), max(aux$Amount), length.out = 14))
    ) +
    xlim(min(aux$Date - round(days * 0.03)), max(aux$Date) + round(days * 0.03)) +
    annotate("text",
      label = caption, x = max(newp$Date),
      y = 0.09 * max(aux$Amount),
      size = 3.3, colour = "black", hjust = 1.1
    ) +
    geom_vline(xintercept = max(newp$Date), alpha = 0.5) +
    theme_lares(pal = 1) +
    theme(legend.position = "top", legend.justification = c(0, 1)) +
    geom_hline(yintercept = 0, size = 0.3, color = "black") +
    geom_hline(
      yintercept = aux$Amount[1], alpha = 0.9,
      colour = names(lares_pal()$palette[3])
    ) +
    geom_hline(
      yintercept = max(aux$Amount), size = 0.2,
      linetype = "dashed", colour = "#3DA4AB"
    )

  points <- aux[!is.na(aux$Cash) & aux$Cash != 0, ]
  if (nrow(points) > 0) {
    plot <- plot +
      geom_point(data = points, aes(x = .data$Date, y = .data$Portfolio)) +
      geom_text(
        data = points,
        aes(
          x = .data$Date, y = .data$Portfolio, hjust = -0.6,
          label = formatNum(.data$Cash, 2, abbr = TRUE)
        ),
        size = 2.7, angle = 90
      )
  }
  if (save) {
    plot <- plot +
      ggsave("portf_total_hist.png", width = 8, height = 5, dpi = 300)
  }
  plot
}


####################################################################
#' Portfolio Plots: Dividends per Year and Quarter
#'
#' \code{stocks_plots()} plots a portfolio's historical dividends incomes
#' grouped by quarter an year.
#'
#' @inheritParams splot_summary
#' @param type Integer. Typo of plot. 1 for incomes.
#' @export
#' @rdname stocks_plots
splot_divs <- function(p, type = 1) {
  if (type == 1) {
    p %>%
      mutate(
        year = year(.data$Date),
        quarter = date_cuts(.data$Date, "Q")
      ) %>%
      group_by(.data$year, .data$quarter) %>%
      summarise(Dividend = sum(.data$Dividend)) %>%
      ungroup() %>%
      group_by(year) %>%
      mutate(DividendYear = sum(.data$Dividend)) %>%
      ungroup() %>%
      mutate(year = sprintf("%s\n%s", .data$year, formatNum(.data$DividendYear, pre = "$", abbr = TRUE))) %>%
      filter(.data$DividendYear > 0) %>%
      ggplot(aes(x = .data$quarter, y = .data$Dividend, fill = .data$quarter)) +
      geom_col() +
      facet_grid(. ~ year) +
      scale_y_dollar() +
      labs(
        title = "Dividends per Quarter and Year",
        subtitle = sprintf("Total: %s", formatNum(sum(p$Dividend), 0, pre = "$")),
        x = NULL, y = "Dividends (post-tax)"
      ) +
      theme_lares(pal = 1, grid = "Yy", legend = "none")
  }
}


####################################################################
#' Portfolio Plots: Daily ROI
#'
#' \code{splot_roi()} plots a portfolio's historical ROI since inception
#' or since last n days, with 2 moving average lines.
#'
#' @inheritParams splot_summary
#' @param n_days Integer. How many days back you want to see?
#' @param historical Boolean. Historical ROI metric? If not, ROI
#' will be calculated locally for n_days parameter
#' @param ma Numeric Vector. Select 2 values for moving averages.
#' Set to NA to turn this metric off
#' @export
#' @rdname stocks_plots
splot_roi <- function(p, n_days = 365, historical = TRUE, ma = c(12, 50), save = FALSE) {
  check_attr(p, check = "daily_portfolio")

  if (is.na(n_days)) {
    n_days <- nrow(p)
  }
  if (n_days > nrow(p)) {
    message(sprintf(
      "As there are less observations than for %s days, changed n_days all days: %s",
      n_days, nrow(p)
    ))
    n_days <- nrow(p)
  }

  n_days <- ifelse(n_days < max(ma), max(ma) + 2, n_days)
  newp <- rbind(p, mutate(tail(p, 1), Date = .data$Date - 1, ROI = 0, CumValue = 0))

  caption <- ifelse(is.na(ma)[1], "",
    paste("Showing moving averages for", ma[1], "and", ma[2], "days.")
  )

  if (!historical) newp <- mutate(newp, ROI = .data$ROI - newp$ROI[n_days])
  if (!is.na(n_days)) {
    newp <- slice(newp, 1:n_days)
    mas <- function(x, n = 5) {
      stats::filter(x, rep(1 / n, n), sides = 1)
    }
    if (n_days != nrow(p)) {
      caption <- paste(
        caption,
        "\nShowing last", n_days, "days,",
        ifelse(historical, "with historical ROI results.",
          "with relative ROI results."
        )
      )
    }
  }

  if (!is.na(ma)[1]) {
    newp <- newp %>%
      mutate(
        ma1 = mas(.data$ROI, ma[1]),
        ma2 = mas(.data$ROI, ma[2])
      )
  }

  pos <- mutate(newp, ROI = ifelse(.data$ROI >= 0, .data$ROI, 0.0001))
  neg <- mutate(newp, ROI = ifelse(.data$ROI < 0, .data$ROI, -0.0001))

  plot <- ggplot(newp, aes(x = .data$Date)) +
    geom_hline(yintercept = 0, size = 0.3, color = "black") +
    geom_hline(yintercept = newp$ROI[1], alpha = 0.9, colour = "#40A4D8") +
    geom_hline(
      yintercept = max(newp$ROI), size = 0.2,
      linetype = "dashed", colour = "#3DA4AB"
    ) +
    geom_hline(
      yintercept = min(newp$ROI), size = 0.2,
      linetype = "dashed", colour = "tomato"
    ) +
    geom_area(data = pos, aes(y = .data$ROI), fill = "#3DA4AB", alpha = 0.3) +
    geom_area(data = neg, aes(y = .data$ROI), fill = "tomato", alpha = 0.3) +
    geom_line(aes(y = .data$ROI), size = 0.5) +
    scale_x_date(date_labels = "%b%y") +
    labs(
      y = "ROI [%]", x = NULL,
      title = "Portfolio's Daily Growth (%)",
      subtitle = paste0(
        newp$Date[1], " | ROI: ",
        formatNum(newp$ROI[1], 2), "% ($",
        formatNum(newp$CumValue[1] - newp$CumInvested[1], 0),
        ") | Stocks Value: $", formatNum(newp$CumValue[1], 0),
        " | Portfolio: $", formatNum(newp$Portfolio[1], 0)
      ),
      caption = if (!is.na(ma)[1]) caption
    ) +
    theme_lares(legend = "top")

  if (!is.na(ma)[1]) {
    plot <- plot +
      geom_line(aes(y = .data$ma1),
        size = 0.7,
        colour = "#071D49", alpha = 0.9, na.rm = TRUE
      ) +
      geom_line(aes(y = .data$ma2),
        size = 0.7,
        colour = "darkorange", alpha = 0.9, na.rm = TRUE
      ) +

      if (save) {
        plot <- plot +
          ggsave("portf_daily_change.png", width = 8, height = 5, dpi = 300)
      }
  }

  plot
}


####################################################################
#' Portfolio Plots: Types of Stocks
#'
#' \code{splot_types()} lets the user plot types or categories of tickers.
#'
#' @inheritParams splot_summary
#' @export
#' @rdname stocks_plots
splot_types <- function(s, save = FALSE) {
  check_attr(s, check = "daily_stocks")

  plot <- s %>%
    filter(.data$Date == max(.data$Date), .data$CumQuant > 0) %>%
    group_by(.data$Type) %>%
    mutate(label = paste0(.data$Type, "\n", formatNum(
      100 * sum(.data$CumValue) / sum(s$CumValue[s$Date == max(s$Date)]), 0
    ), "%")) %>%
    ggplot() +
    geom_bar(aes(x = "", y = .data$CumValue, fill = .data$Symbol),
      width = 1, stat = "identity"
    ) +
    facet_grid(. ~ label, scales = "free") +
    theme_lares(pal = 1) +
    scale_y_formatNum(expand = c(0, 0), abbr = TRUE, pre = "$") +
    labs(x = NULL, y = "Total value", title = "Portfolio's Category Distribution")
  if (save) {
    plot <- plot +
      ggsave("portf_distribution.png", width = 8, height = 5, dpi = 300)
  }
  plot
}


####################################################################
#' ETF's Sectors Breakdown
#'
#' \code{etf_sector()} scraps etf.com data for sector breakdown on ETFs.
#' Use \code{splot_etf()} for visualization.
#'
#' @family Investment
#' @inheritParams stocks_hist
#' @param etf Character Vector. Which ETFs you wish to scrap?
#' @return data.frame with ETF break.down data by sector
#' @examples
#' \donttest{
#' etf_sector(etf = "VTI")
#' }
#' @export
etf_sector <- function(etf = "VTI", quiet = FALSE, cache = TRUE) {
  if ("daily_stocks" %in% attr(etf, "type")) {
    etf <- as.character(unique(etf$Symbol[etf$Date == max(etf$Date)]))
  }
  cache_file <- c(as.character(Sys.Date()), "etf_sector", etf)
  if (cache_exists(cache_file) && cache) {
    cache_read(cache_file, quiet = quiet)
  } else {
    if (!quiet) message(">>> Downloading sectors for each ETF...")
    ret <- data.frame()
    nodata <- NULL
    for (i in seq_along(etf)) {
      info <- toupper(etf[i])
      url <- sprintf("https://etfdb.com/etf/%s/", info)
      # exists <- tryCatch({!httr::http_error(url)}, error = function(err) {FALSE})
      sector <- tryCatch(
        suppressWarnings(content(GET(url))),
        error = function(err) {
          closeAllConnections()
          gc()
          nodata <- c(nodata, info)
          NULL
        }
      )
      if (is.null(sector)) next
      tables <- sector %>% html_table()
      sector_table <- lapply(tables, function(x) {
        "Sector" %in% colnames(x)
      }) %>% unlist()
      if (sum(sector_table) > 0) {
        sec <- tables[sector_table][[1]] %>%
          mutate(ETF = info) %>%
          mutate(Percentage = as.numeric(gsub("%", "", .data$Percentage)))
        ret <- rbind(ret, sec)
      } else {
        nodata <- c(nodata, info)
      }
      if (!quiet && length(etf) > 1) {
        statusbar(i, length(etf), info)
      }
    }

    attr(ret, "type") <- "etf_sector"
    if (length(unique(nodata)) > 0) {
      if (!quiet) message("Some ticks were not found as ETF: ", v2t(unique(nodata)))
    }
    if (nrow(ret) == 0) {
      if (!quiet) message("No data found for given Tickers!")
      invisible(NULL)
    } else {
      cache_write(ret, cache_file, quiet = TRUE)
      ret
    }
  }
}


####################################################################
#' Portfolio's Sector Distribution (ETFs)
#'
#' \code{splot_etf()} lets the user plot his portfolio's distribution,
#' specifically ETF's sectors.
#'
#' @family Scrapper
#' @inheritParams splot_summary
#' @inheritParams stocks_file
#' @param keep_all Boolean. Keep "Not Known / Not ETF"?
#' @export
#' @rdname stocks_plots
splot_etf <- function(s, keep_all = FALSE, cache = TRUE, save = FALSE) {
  check_attr(s, check = "daily_stocks")
  if (!"Date" %in% colnames(s)) s$Date <- Sys.Date()
  cache_file <- c(as.character(Sys.Date()), "splot_etf", s)
  if (cache_exists(cache_file) && cache) {
    etfs <- cache_read(cache_file, quiet = quiet)
  } else {
    etfs <- etf_sector(s)
  }

  if (length(etfs) > 0) {
    df <- etfs %>%
      right_join(
        select(s, .data$Symbol, .data$CumValue, .data$Date) %>%
          filter(.data$Date == max(.data$Date)) %>%
          mutate(Symbol = as.character(.data$Symbol)),
        by = c("ETF" = "Symbol")
      ) %>%
      {
        if (keep_all == FALSE) filter(., !is.na(.data$Sector)) else .
      } %>%
      mutate(
        ETF = factor(.data$ETF, levels = s$Symbol[s$Date == max(s$Date)]),
        Sector = ifelse(is.na(.data$Sector),
          "Not Known / Not ETF",
          as.character(.data$Sector)
        )
      ) %>%
      replace(., is.na(.), 100) %>%
      mutate(Value = .data$CumValue * .data$Percentage / 100) %>%
      group_by(.data$Sector) %>%
      mutate(ValueSector = sum(.data$Value)) %>%
      ungroup() %>%
      group_by(.data$Sector) %>%
      mutate(label = paste0(
        .data$Sector, " (",
        signif(100 * .data$ValueSector / sum(s$CumValue[s$Date == max(s$Date)]), 2), "%)"
      ))

    plot <- ggplot(df, aes(
      x = reorder(.data$label, .data$ValueSector),
      y = .data$Value, fill = .data$ETF
    )) +
      geom_bar(width = 1, stat = "identity") +
      coord_flip() +
      scale_y_dollar(expand = c(0, 0)) +
      theme_lares(pal = 1) +
      labs(
        x = NULL, y = "Total value [$]", fill = NULL,
        title = "Portfolio's Sector Distribution (ETFs)"
      )

    if (save) {
      plot <- plot +
        ggsave("portf_distribution_etfs.png", width = 8, height = 5, dpi = 300)
    }
    plot
  } else {
    noPlot("No data here!")
  }
}


####################################################################
#' Portfolio's Calculations and Plots
#'
#' \code{stocks_obj()} lets the user create his portfolio's calculations and
#' plots for further study.
#'
#' @inheritParams stocks_file
#' @inheritParams daily_portfolio
#' @inheritParams stocks_hist
#' @inheritParams daily_stocks
#' @param data List. Containing the following dataframes: portfolio,
#' transactions, cash. They have to follow the original xlsx format
#' @param sectors Boolean. Return sectors segmentation for ETFs?
#' @param parg Boolean. Personal argument. Used to personalize stuff, in this
#' case, taxes changed from A to B in given date (hard-coded)
#' @return List. Aggregated results and plots.
#' @export
#' @rdname stocks_report
stocks_obj <- function(data = stocks_file(),
                       cash_fix = 0,
                       tax = 30,
                       sectors = FALSE,
                       parg = FALSE,
                       window = c("1M", "YTD", "1Y", "MAX"),
                       cache = TRUE,
                       quiet = FALSE) {
  check_attr(data, check = "stocks_file")
  check_opts(window, c("1W", "1M", "6M", "1Y", "YTD", "5Y", "MAX"))

  ret <- list()
  trans <- data$transactions
  cash <- data$cash
  tickers <- data$portfolio

  message(">>> Downloading historical data for each stock...")
  ret[["quotes"]] <- hist <- stocks_hist(
    symbols = tickers$Symbol,
    from = tickers$StartDate,
    tax = tax,
    parg = parg,
    cache = cache,
    quiet = quiet
  )

  ws <- length(window)
  ret[["stocks"]] <- s <- daily_stocks(hist, trans, tickers, window = "MAX")
  ret[["portfolio"]] <- p <- daily_portfolio(hist, trans, cash, cash_fix = cash_fix, window = "MAX")

  # Fixed plots
  ret[["plots_fixed"]] <- list(
    "Positions" = splot_summary(p, s),
    "Types of Stocks" = splot_types(s),
    "ETFs by Industry" = if (sectors) splot_etf(s, cache = cache) else NULL,
    "Dividends Incomes" = if (sum(p$Dividend) > 0) splot_divs(p, type = 1) else NULL
  )

  # Relative plots (using time windows)
  message(glued(">>> Running calculations and plots for {ws} time window{ifelse(ws>1,'s','')}..."))
  plots_relative <- lapply(window, function(x) {
    # Filter the data given the window
    s <- daily_stocks(hist, trans, tickers, window = x)
    p <- daily_portfolio(hist, trans, cash, cash_fix = cash_fix, window = x)

    # Visualizations
    plots <- list()
    # plots[["ROI"]] <- splot_roi(p, n_days = 5)
    plots[["History"]] <- splot_growth(p)
    plots[["Relative Growth"]] <- splot_change(p, s, rel = TRUE, n_days = NA, keep_old = FALSE)
    plots[["Absolute Growth"]] <- splot_change(p, s, rel = FALSE, n_days = NA, keep_old = FALSE)
    plots
  })
  names(plots_relative) <- sprintf("plots_%s", window)
  ret[["plots_relative"]] <- plots_relative

  attr(ret, "type") <- "stocks_obj"
  ret
}
# x <- stocks_obj(window = c("1M","1Y","YTD"))

####################################################################
#' Portfolio's Full Report and Email
#'
#' \code{stocks_report()} lets the user create his portfolio's full report with
#' plots and send it to an email with the HTML report attached
#'
#' @inheritParams stocks_file
#' @param data Character. \code{stocks_obj()} output. If NA, automatic
#' parameters and \code{stocks_file()} defaults will be used.
#' @param dir Character. Directory for HTML report output. If set to NA,
#' current working directory will be used. If mail sent, file will be erased
#' @param mail Boolean. Do you want to send an email with the report attached?
#' If not, an HTML file will be created in dir
#' @param attachment Boolean. Create and add report as attachment if
#' \code{mail=TRUE}? If not, no report will be rendered and only tabulated
#' summaries will be included on email's body.
#' @param to Character. Email to send the report to
#' @param sectors Boolean. Return sectors segmentation for ETFs?
#' @param keep Boolean. Keep HTML file when sent by email?
#' @param creds Character. Credential's user (see \code{get_creds()}) for
#' sending mail and Dropbox interaction.
#' @return Invisible list. Aggregated results and plots.
#' @examples
#' \dontrun{
#' list <- stocks_obj()
#' stocks_report(list, dir = "~/Desktop")
#' }
#' @export
#' @rdname stocks_report
stocks_report <- function(data = NA,
                          keep_old = TRUE,
                          dir = NA,
                          mail = FALSE,
                          attachment = TRUE,
                          to = "laresbernardo@gmail.com",
                          sectors = FALSE,
                          keep = FALSE,
                          creds = NA,
                          cache = TRUE) {
  try_require("rmarkdown")
  tic("stocks_report")

  if (is.na(data)[1]) {
    df <- stocks_file(creds = creds, keep_old = keep_old, cache = cache)
    data <- stocks_obj(df, sectors = sectors, parg = is.na(creds), cache = cache)
  }

  check_attr(data, check = "stocks_obj")

  # pandoc <- Sys.getenv("RSTUDIO_PANDOC")
  # Sys.setenv(RSTUDIO_PANDOC = pandoc)
  if (is.na(dir)) dir <- getwd()

  # Summary as tables
  data$summary_df1 <- data$stocks %>%
    filter(.data$Date == max(.data$Date), .data$CumQuant > 0) %>%
    mutate(Change = sprintf(
      "%s (%s)",
      formatNum(100 * .data$DifUSD / (.data$CumValue - .data$DifUSD), 2, pos = "%", sign = TRUE),
      formatNum(.data$DifUSD / .data$CumQuant, 2, sign = TRUE)
    )) %>%
    arrange(desc(abs(.data$DifUSD / .data$CumValue))) %>%
    select(
      .data$Symbol, .data$Value, .data$Change, .data$CumQuant,
      .data$DifUSD, .data$CumValue, .data$wt, .data$CumROI
    ) %>%
    rename(
      "Abs.Change" = .data$DifUSD, "Quant" = .data$CumQuant,
      "Total Value" = .data$CumValue, "Weight [%]" = .data$wt,
      "Hist.ROI [%]" = .data$CumROI
    )
  data$summary_df2 <- data$portfolio %>%
    filter(.data$Date == max(.data$Date)) %>%
    mutate(DifP = round(100 * .data$DifUSD / (.data$CumValue - .data$DifUSD), 2)) %>%
    mutate_if(is.numeric, function(x) formatNum(x, 2, signif = 6)) %>%
    mutate_all(as.character) %>%
    select(.data$Date, one_of(sort(colnames(.)))) %>%
    tidyr::gather() %>%
    rename("Metric" = .data$key, "Value" = .data$value) %>%
    mutate("Window" = c(
      rep("Today", 2), rep("Total", 5),
      rep("Today", 4), rep("Total", 2)
    )) %>%
    arrange(.data$Window)

  # Can be more accurate with names but works for me!
  params <- list(
    summary_df1 = data$summary_df1,
    summary_df2 = data$summary_df2,
    plots_fixed = data[["plots_fixed"]],
    plots_relative = data[["plots_relative"]]
  )

  if (attachment) {
    message(">>> Rendering HTML report...")
    html_file <- "stocksReport.html"
    render(system.file("docs", "stocksReport.Rmd", package = "lares"),
      output_file = html_file,
      output_dir = dir,
      params = params,
      envir = new.env(parent = globalenv()),
      quiet = TRUE
    )
    message("HTML report created succesfully!")
  }

  if (mail) {
    try_require("knitr")
    subject <- sprintf(
      "Report: %s | %s (%s)", max(data$portfolio$Date),
      paste0(data$summary_df2$Value[data$summary_df2$Metric == "DifUSD"]),
      paste0(data$summary_df2$Value[data$summary_df2$Metric == "DifP"], "%")
    )
    html_body <- paste0(
      "<p>Stocks Summary:</p>",
      kable(data$summary_df1,
        align = "lrcrrrr", row.names = TRUE,
        digits = c(0, 2, 0, 2, 0, 0, 1), format = "html"
      ),
      "<p>Portfolio Status:</p>",
      kable(data$summary_df2, align = "lr", row.names = FALSE, format = "html")
    )

    message(">>> Sending email...")
    mail_send(
      to = to,
      subject = subject,
      html = html_body,
      attachment = if (attachment) html_file else NULL,
      creds = creds,
      quiet = FALSE
    )
    if (!keep && attachment) invisible(file.remove(html_file))
  }

  toc("stocks_report")
  invisible(data)
}

.filter_window <- function(df, window) {
  new_df <- df %>%
    arrange(desc(.data$Date)) %>%
    {
      if (window == "1W") filter(., .data$Date >= Sys.Date() - 7) else .
    } %>%
    {
      if (window == "1M") filter(., .data$Date >= Sys.Date() %m-% months(1)) else .
    } %>%
    {
      if (window == "6M") filter(., .data$Date >= Sys.Date() %m-% months(6)) else .
    } %>%
    {
      if (window == "1Y") filter(., .data$Date >= Sys.Date() %m-% years(1)) else .
    } %>%
    {
      if (window == "YTD") filter(., .data$Date >= floor_date(Sys.Date(), "year")) else .
    } %>%
    {
      if (window == "5Y") filter(., .data$Date >= Sys.Date() %m-% years(5)) else .
    }

  if ("ROI" %in% colnames(df) && window == "MAX") { # Dataframe: portfolio
    # Add a last row (first date) with no data
    new_df <- rbind(new_df, df %>% arrange(.data$Date) %>% slice(1) %>% ungroup() %>% mutate_if(
      is.numeric, function(x) 0
    ) %>% mutate(Date = .data$Date - 1))
  }
  if (!"ROI" %in% colnames(df) && window == "MAX") { # Dataframe: stocks
    new_df <- rbind(new_df, df %>% arrange(.data$Date) %>%
      group_by(.data$Symbol) %>% slice(1) %>% ungroup() %>% mutate_if(
        is.numeric, function(x) 0
      ) %>% mutate(Date = .data$Date - 1))
  }
  new_df
}

# # TESTING
# devtools::load_all()
# library(lares)
# library(tidyverse)
# library(openxlsx)
# library(quantmod)
# library(jsonlite)
# library(lubridate)
# library(rvest)
# library(httr)
# library(rdrop2)
# data <- stocks_file()
# trans <- data$transactions
# cash <- data$cash
# tickers <- data$portfolio
# q <- stocks_obj(window = c("1M","1Y","MAX"), sectors = FALSE, cache = TRUE)
# hist <- q$quotes
# p <- q$portfolio
# s <- q$stocks
# stocks_report(q, dir = "~/Desktop")

# s %>%
#   filter(Value > 0, CumQuant > 0) %>%
#   select(Symbol, Date, CumQuant, Value) %>%
#   arrange(Date) %>%
#   group_by(Symbol) %>%
#   mutate(RelValue = Value / Value[1]) %>%
#   #filter(Symbol == "TSLA") %>%
#   ggplot(aes(x = Date, y = log(RelValue), colour = Symbol)) +
#   geom_line() +
#   theme_lares(pal = 2) +
#   scale_y_percent()

####################################################################
#' Clean text strings automatically
#'
#' \code{cleanText}: Clean character strings automatically. Options to keep ASCII
#' characters only, keep certain characters, lower caps, title format, are available.
#'
#' @family Data Wrangling
#' @family Text Mining
#' @param text Character Vector
#' @param spaces Boolean. Keep spaces? If character input, spaces
#' will be transformed into passed argument.
#' @param keep Character. String (concatenated or as vector) with all characters
#' that are accepted and should be kept, in addition to alphanumeric.
#' @param lower Boolean. Transform all to lower case?
#' @param ascii Boolean. Only ASCII characters?
#' @param title Boolean. Transform to title format (upper case on first letters).
#' @return Character vector with transformed strings.
#' @examples
#' cleanText("Bernardo Lares 123")
#' cleanText("Bèrnärdo LáreS 123", lower = FALSE)
#' cleanText("Bernardo Lare$", spaces = ".", ascii = FALSE)
#' cleanText("\\@®ì÷å   %ñS  ..-X", spaces = FALSE)
#' cleanText(c("maría", "€", "núñez_a."), title = TRUE)
#' cleanText("29_Feb-92()#", keep = c("#", "_"), spaces = FALSE)
#' @export
#' @rdname clean_text
cleanText <- function(text, spaces = TRUE, keep = "", lower = TRUE,
                      ascii = TRUE, title = FALSE) {
  text <- as.character(text)
  if (ascii) {
    text <- iconv(text, from = "UTF-8", to = "ASCII//TRANSLIT", sub = "")
  }
  keep <- paste(keep, collapse = "")
  text <- gsub(paste("[^[:alnum:] ", keep, "]"), "", text)
  if (lower) text <- tolower(text)
  if (is.character(spaces)) text <- gsub(" ", spaces, text)
  if (isFALSE(spaces)) text <- gsub(" ", "", text)
  text <- trimws(gsub("[[:space:].]+", " ", text))
  if (title) text <- stringr::str_to_title(text)
  text
}


####################################################################
#' Clean column names for data.frame/tibble objects
#'
#' \code{cleanNames}: Resulting names are unique and consist only of the \code{_}
#' character, numbers, and ASCII letters. Capitalization preferences can be
#' specified using the \code{lower} parameter.
#'
#' Inspired by \code{janitor::clean_names}.
#'
#' @param df data.frame/tibble.
#' @param num Add character before only-numeric names.
#' @param ... Additional parameters passed to \code{cleanText()}.
#' @return data.frame/tibble with transformed column names.
#' @examples
#'
#' # For a data.frame directly:
#' df <- dft[1:5, 1:6] # Dummy data
#' colnames(df) <- c("ID.", "34", "x_2", "Num 123", "Nòn-äscì", "  white   Spaces  ")
#' print(df)
#' cleanNames(df)
#' cleanNames(df, lower = FALSE)
#' @export
#' @rdname clean_text
cleanNames <- function(df, num = "x", keep = "_", ...) {
  if (!is.null(df)) {
    # Initial cleanse
    cols <- cleanText(colnames(df), keep = keep, ...)
    # Simple spaces turned into "_"
    cols <- trimws(gsub("[[:space:].]+", "_", cols))
    # If only numeric, add x at the begining
    onlynum <- !grepl("[[:alpha:]]", cols)
    cols[onlynum] <- paste0(num, cols[onlynum])
    # Change column names
    df <- stats::setNames(df, cols)
    # Keep tibble if original data.frame is tibble
    if ("tbl_df" %in% class(df)) df <- as_tibble(df)
  }
  df
}

####################################################################
#' Tokenize Vectors into Words
#'
#' This function transforms texts into words, calculate frequencies,
#' supress stop words in a given language.
#'
#' @family Data Wrangling
#' @family Text Mining
#' @inheritParams mp3_get
#' @param text Character vector. Sentences or texts you wish to tokenize.
#' @param exclude Character vector. Which words do you wish to exclude?
#' @param lang Character. Language in text (used for stop words). Example:
#' "spanish" or "english". Set to \code{NA} to ignore.
#' @param min_word_freq Integer. This will discard words that appear
#' less than <int> times. Defaults to 2. Set to \code{NA} to ignore.
#' @param min_word_len Integer. This will discard words that have
#' less than <int> characters. Defaults to 5. Set to \code{NA} to ignore.
#' @param keep_spaces Boolean. If you wish to keep spaces in each line
#' to keep unique compound words, separated with spaces, set to TRUE.
#' For example, 'one two' will be set as 'one_two' and treated as
#' a single word.
#' @param lowercase,remove_numbers,remove_punct Boolean.
#' @param remove_lettt Boolean. Repeated letters (more than 3 consecutive).
#' @param laughs Boolean. Try to unify all laughs texts.
#' @param utf Boolean. Transform all characters to UTF (no accents and crazy symbols)
#' @param df Boolean. Return a dataframe with a one-hot-encoding kind of
#' results? Each word is a column and returns if word is contained.
#' @param h2o Boolean. Return \code{H2OFrame}?
#' @return data.frame. Tokenized words with counters.
#' @export
textTokenizer <- function(text,
                          exclude = NULL,
                          lang = NULL,
                          min_word_freq = 5,
                          min_word_len = 2,
                          keep_spaces = FALSE,
                          lowercase = TRUE,
                          remove_numbers = TRUE,
                          remove_punct = TRUE,
                          remove_lettt = TRUE,
                          laughs = TRUE,
                          utf = TRUE,
                          df = FALSE,
                          h2o = FALSE,
                          quiet = FALSE) {
  try_require("tm")

  text <- as.character(text)
  lang <- tolower(lang)

  # text <- as.character(c("Hooooolaa 123 4 jaja ALLÁ que bueeeno toOdO!...","seguuuimos","jajaja?"))

  # When multiple words should be kept together
  if (keep_spaces) {
    if (!quiet) message(">>> Keeping spaced multi-words")
    text <- gsub(" ", "_", text) # '_' deleted later on
  }

  ## Load the data as a corpus
  docs <- VCorpus(x = VectorSource(text), readerControl = list(reader = readPlain))

  # Convert the text to lower case
  if (lowercase) {
    if (!quiet) message(">>> Transforming to lower case")
    aux <- function(x) tolower(x)
    docs <- tm_map(docs, content_transformer(aux))
  }

  # Remove numbers
  if (remove_numbers) {
    if (!quiet) message(">>> Removing numbers")
    aux <- function(x) gsub("[0-9]", " ", x)
    docs <- tm_map(docs, content_transformer(aux))
  }

  # Remove punctuations
  if (remove_punct) {
    if (!quiet) message(">>> Removing punctuation marks")
    aux <- function(x) gsub("[[:punct:] ]+", " ", x)
    docs <- tm_map(docs, content_transformer(aux))
  }

  # Eliminate extra white spaces
  if (!quiet) message(">>> Removing extra spaces")
  aux <- function(x) gsub("\\s+", " ", x)
  docs <- tm_map(docs, content_transformer(aux))

  # Remove crazy UTF-8 symbols over letters
  if (utf) {
    if (!quiet) message(">>> Transforming UTF8 to ASCII symbols")
    aux <- function(x) gsub("[^[:alnum:] ]", "", iconv(x, from = "UTF-8", to = "ASCII//TRANSLIT"))
    docs <- tm_map(docs, content_transformer(aux))
  }

  # Repeated letters (more than 3 times)
  if (remove_lettt) {
    if (!quiet) message(">>> Removing repeated characters (more than 2)")
    aux <- function(x) gsub("([[:alpha:]])\\1{2,}", "\\1", x)
    docs <- tm_map(docs, content_transformer(aux))
    # Double vowels as well in spanish
    if ("spanish" %in% lang) {
      aux <- function(x) gsub("[aeiou]*([aeiou])\\1+", "\\1", x)
      docs <- tm_map(docs, content_transformer(aux))
    }
  }

  # Laughs replacements
  if (laughs) {
    if (!quiet) message(">>> Standarizing laughs")
    aux <- function(x) gsub("a*ja+j[ja]*|a*ha+h[ha]*|o?l+o+l+[ol]*", "(laugh)", x)
    docs <- tm_map(docs, content_transformer(aux))
  }

  # Remove your own stop words
  if (!is.null(exclude)) {
    if (!quiet) message(">>> Removing stopwords")
    docs <- tm_map(docs, removeWords, c("https", "http", "que", as.character(exclude)))
  }

  # Remove stopwords (common stopwords)
  if (!is.null(lang)) {
    if (!quiet) message(">>> Removing language stopwords")
    docs <- tm_map(docs, removeWords, stopwords(lang))
  }

  ## Build a term-document matrix
  dtm <- TermDocumentMatrix(docs)
  m <- as.matrix(dtm)
  v <- sort(rowSums(m), decreasing = TRUE)
  d <- data.frame(word = names(v), freq = v)
  rownames(d) <- NULL

  if (!is.na(min_word_freq)) {
    if (!quiet) message(">>> Removing words with frequency lower than ", min_word_freq)
    d <- filter(d, .data$freq >= min_word_freq)
  }

  if (!is.na(min_word_len)) {
    if (!quiet) message(">>> Removing words with length lower than ", min_word_len)
    d <- filter(d, nchar(.data$freq) >= min_word_len)
  }

  if (df) {
    texts <- cleanText(unique(text))
    if (length(texts) != length(text)) {
      if (!quiet) message("Returning unique texts results...")
    }
    toksdf <- NULL
    for (i in seq_len(nrow(d))) {
      word <- as.character(d$word[i])
      vector <- grepl(word, texts)
      toksdf <- cbind(toksdf, vector)
      colnames(toksdf)[colnames(toksdf) == "vector"] <- word
      statusbar(i, nrow(d), word)
    }
    toksdf <- data.frame(texts = texts, toksdf)
    as_tibble(toksdf)
  } else {
    as_tibble(d)
  }
}


####################################################################
#' Create features out of text
#'
#' This function creates a data.frame with features based on a text vector
#'
#' @family Data Wrangling
#' @family Text Mining
#' @param text Character vector
#' @param auto Boolean. Auto create some useful parameters?
#' @param contains Character vector. Which columns do you wish to add
#' with a contains (counter) string validator?
#' @param prc Boolean. Also add percentage of each column compared with length?
#' @return data.frame with additional features based on \code{text}.
#' @examples
#' textFeats("Bernardo Lares")
#' textFeats("Bernardo Lares 123!", prc = TRUE)
#' textFeats("I'm 100% Lares...", contains = c("Lares", "lares"))
#' textFeats(c("GREAT library!!", "Have you tried this 2?", "Happy faces :D :-)"))
#' @export
textFeats <- function(text, auto = TRUE, contains = NA, prc = FALSE) {
  ret <- data.frame(text = text)


  if (auto) {
    ret <- ret %>%
      mutate(
        length = str_length(text),
        ncap = str_count(text, "[A-Z]"),
        # ncap_len = round(ncap / length, 4),
        nvoc = str_count(toupper(text), "A|E|I|O|U"),
        # nvoc_len = round(nvoc / length, 4),
        nexcl = str_count(text, fixed("!")),
        nquest = str_count(text, fixed("?")),
        nats = str_count(text, fixed("@")),
        npunct = str_count(text, "[[:punct:]]"),
        ndig = str_count(text, "[[0-9]]"),
        nword = 1 + str_count(text, "\\ "),
        nsymb = str_count(text, "&|@|#|\\$|%|\\*|\\^"),
        nsmile = str_count(text, "((?::|;|=)(?:-)?(?:\\)|D|P))")
      )
  }


  # Custom columns with contains argument
  if (!is.na(contains[1])) {
    df <- NULL
    for (i in seq_along(contains)) {
      word <- as.character(contains[i])
      vector <- str_count(text, fixed(word))
      df <- cbind(df, vector)
      colnames(df)[colnames(df) == "vector"] <- word
    }
    ret <- data.frame(ret, df)
  }

  if (prc) {
    ret <- ret %>%
      rowwise() %>%
      mutate_if(is.numeric, list(pct = ~ round(. / length, 5))) %>%
      ungroup()
  }

  as_tibble(ret)
}

####################################################################
#' Wordcloud Plot
#'
#' Study the distribution of a target variable vs another variable. This
#' function is quite similar to the funModeling's corrplot function.
#'
#' @family Text Mining
#' @param text Character vector
#' @param lang Character. Language in text (used for stop words)
#' @param exclude Character vector. Which word do you wish to exclude?
#' @param seed Numeric. Seed for re-producible plots
#' @param keep_spaces Boolean. If you wish to keep spaces in each line
#' to keep unique compount words, separated with spaces, set to TRUE.
#' For example, 'LA ALAMEDA' will be set as 'LA_ALAMEDA' and treated as
#' a single word.
#' @param min Integer. Words with less frequency will not be plotted
#' @param pal Character vector. Which colours do you wish to use
#' @param print Boolean. Plot results as textcloud?
#' @return wordcloud plot object
#' @export
textCloud <- function(text, lang = "english", exclude = NULL, seed = 0,
                      keep_spaces = FALSE, min = 2, pal = NA, print = TRUE) {
  try_require("wordcloud")
  on.exit(set.seed(seed))

  d <- textTokenizer(text, lang, exclude, keep_spaces)
  if (print) message(paste0(capture.output(head(d, 10)), collapse = "\n"))

  pal <- if (is.na(pal)) rev(names(lares_pal()$palette)[1:6])
  wordcloud(
    words = d$word, freq = d$freq,
    scale = c(3.5, .7),
    min.freq = min,
    max.words = 200,
    random.order = FALSE,
    rot.per = 0.2,
    colors = pal
  )
}


####################################################################
#' Sentiment Breakdown on Text
#'
#' This function searches for relevant words in a given text and adds
#' sentiments labels (joy, anticipation, surprise, positive, trust,
#' anger, sadness, fear, negative, disgust) for each of them, using NRC.
#' Then, makes a summary for all words and plot results.
#'
#' @family Text Mining
#' @param text Character vector
#' @param lang Character. Language in text (used for stop words)
#' @param exclude Character vector. Which word do you wish to exclude?
#' @param append_file Character. Add a dictionary to append. This file
#' must contain at least two columns, first with words and second with
#' the sentiment (consider sentiments on description).
#' @param append_words Dataframe. Same as append_file but appending
#' data frame with word and sentiment directly
#' @param plot Boolean. Plot results summary?
#' @param subtitle Character. Add subtitle to the plot
#' @return List. Contains data.frame with words and sentiments, summary and plot.
#' @export
sentimentBreakdown <- function(text, lang = "spanish",
                               exclude = c("maduro", "que"),
                               append_file = NA,
                               append_words = NA,
                               plot = TRUE,
                               subtitle = NA) {
  try_require("syuzhet")
  ret <- list()

  dictionary <- get_sentiment_dictionary("nrc", language = lang) %>%
    select(-.data$lang, -.data$value) %>%
    filter(!.data$word %in% exclude) %>%
    rbind(data.frame(
      word = c("(laugh)", "(laugh)"),
      sentiment = c("positive", "joy")
    ))

  if (!is.na(append_file)) {
    new_words <- read.file(normalizePath(append_file), current_wd = FALSE)[, 1:2]
    colnames(new_words) <- c("word", "sentiment")
    new_words$word <- cleanText(new_words$word)
    new_words$sentiment <- cleanText(new_words$sentiment)
    dictionary <- rbind(dictionary, new_words)
  }

  if (!is.na(append_words)) {
    if (ncol(append_words) >= 2) {
      new_words <- append_words[, 1:2]
      colnames(new_words) <- c("word", "sentiment")
      new_words$word <- cleanText(new_words$word)
      new_words$sentiment <- cleanText(new_words$sentiment)
      dictionary <- rbind(dictionary, new_words)
    } else {
      message("Be sure to have 'word' and 'sentiment' columns in append_words!")
    }
  }

  ret[["words"]] <- textTokenizer(text, lang = lang, exclude = exclude)

  ret[["result"]] <- ret$words %>%
    inner_join(dictionary, "word") %>%
    distinct(.data$word, .data$sentiment, .keep_all = TRUE)

  ret[["summary"]] <- ret$result %>%
    group_by(.data$sentiment) %>%
    summarise(freq = sum(.data$freq)) %>%
    mutate(freq = 100 * .data$freq / sum(.data$freq)) %>%
    arrange(desc(.data$freq))

  if (plot) {
    p <- ggplot(
      ret$summary, aes(
        x = reorder(.data$sentiment, .data$freq),
        y = .data$freq
      )
    ) +
      geom_col(aes(fill = .data$sentiment)) +
      theme_lares(pal = 4) +
      coord_flip() +
      guides(fill = "none") +
      labs(x = NULL, y = "Intensity", title = "Sentiment Breakdown")
    if (!is.na(subtitle)) p <- p + labs(subtitle = autoline(subtitle))
    ret[["plot"]] <- p

    # wrd <- spread(ret$result, sentiment, freq, fill = 0)
    # rownames(wrd) <- wrd$word
    # wrd$word <- NULL
    # pal <- lares_pal()$labels %>%
    #   filter(values %in% ret$result$sentiment) %>% arrange(values)
    # comparison.cloud(wrd, max.words = 100,
    #                  scale = c(3.5,.5),
    #                  title.size = NULL,
    #                  random.order = FALSE,
    #                  colors = as.character(pal$fill),
    #                  match.colors = TRUE,
    #                  title.bg.colors = "transparent")
  }
  ret
}


####################################################################
#' Keyword/Topic identification using RAKE
#'
#' RAKE is a basic algorithm which tries to identify keywords in text.
#' Based on \code{udpipe} library, model models, and keywords_rake function.
#'
#' @family Text Mining
#' @param text Character vector
#' @param file Character. Name of \code{udpipe} model previously downloaded
#' for a specific language
#' @param lang Character. If file does not exist, this language will be
#' downloaded from \code{udpipe}'s models.
#' @return data.frame with topics for each \code{text} input.
#' @export
topics_rake <- function(text, file = "english-ewt-ud-2.4-190531.udpipe", lang = "english") {
  try_require("udpipe")
  if (file.exists(file)) {
    message(">>> Loading previously downloaded model...")
    model <- udpipe_load_model(file)
  } else {
    message(">>> Downloading new language model...")
    model <- udpipe_download_model(tolower(lang))
    if (model$download_failed) {
      stop(model$download_message)
    }
  }
  message(">>> Annotating text...")
  doc <- udpipe_annotate(model, cleanText(text))
  aux <- as.data.frame(doc)
  message(">>> Calculating RAKE values...")
  topics <- keywords_rake(
    x = aux, term = "lemma", group = "doc_id",
    relevant = aux$upos %in% c("NOUN", "ADJ")
  )
  topics
}


####################################################################
#' Build N-grams and keep most frequent
#'
#' Build out n-grams for multiple text inputs and keep the n most frequent
#' combinations.
#'
#' @family Text Mining
#' @inheritParams remove_stopwords
#' @param ngram Integer vector. Number of continuous n items in text.
#' @param top Integer. Keep n most frequent ngrams only.
#' @param ... Additional parameters passed to \code{remove_stopwords}.
#' @return data.frame with ngrams and counters, sorted by frequency.
#' @examples
#' # You must have "tidytext" library to use this auxiliary function:
#' \dontrun{
#' women <- read.csv("https://bit.ly/3mXJOOi")
#' x <- women$description
#' ngrams(x, ngram = c(2, 3), top = 3)
#' ngrams(x, ngram = 2, top = 6, stop_words = c("a", "is", "of", "the"))
#' }
#' @export
ngrams <- function(text, ngram = c(2, 3), top = 10, stop_words = NULL, ...) {
  try_require("tidytext")
  x_counts <- lapply(ngram, function(i) {
    data.frame(msg = data.frame(text)[, 1]) %>%
      {
        if (!is.null(stop_words)[1]) {
          mutate(., msg = remove_stopwords(.data$msg, stop_words, ...))
        } else {
          .
        }
      } %>%
      unnest_tokens("comb", .data$msg, token = "ngrams", n = i) %>%
      filter(!is.na(.data$comb)) %>%
      freqs(.data$comb) %>%
      head(top) %>%
      mutate(ngram = i)
  })
  bind_rows(x_counts)
}


####################################################################
#' Remove stop-words and patterns from character vector
#'
#' Remove all stop-words and specific patterns from a character vector
#'
#' @family Text Mining
#' @param text Character vector
#' @param stop_words Character vector. Words to exclude from text. Example:
#' if you want to exclude "a", whenever that word appears it will be excluded,
#' but when the letter "a" appears in a word, it will remain.
#' @param exclude Character. Pattern to exclude using regex.
#' @param sep Character. String that separate the terms.
#' @return Character vector with removed texts.
#' @examples
#' x <- c("A brown fox jumps over a dog.", "Another brown dog.")
#' remove_stopwords(x, stop_words = c("dog", "brown", "a"), exclude = "\\.")
#' @export
remove_stopwords <- function(text, stop_words, exclude = NULL, sep = " ") {
  tok <- lapply(text, function(i) str_split(trimws(i), sep)[[1]])
  if (!is.null(exclude)[1]) {
    tok <- lapply(tok, function(i) gsub(paste(exclude, collapse = "|"), "", i))
  }
  if (!is.null(stop_words)[1]) {
    tok <- lapply(tok, function(i) i[!tolower(i) %in% unique(c("", tolower(stop_words)))])
  }
  fin <- lapply(tok, function(i) paste(i, collapse = sep))
  unlist(fin)
}

####################################################################
#' Plot Result with Nothing to Plot
#'
#' This function lets the user print a plot without plot, with a
#' customizable message. It is quite useful for Shiny renderPlot when
#' using filters and no data is returned.
#'
#' @family Visualization
#' @param message Character. What message do you wish to show?
#' @param size Numeric. Font size for \code{message} input.
#' @param ... Additional parameters passed to \code{theme_lares()}.
#' @return Empty ggplot2 object (with a \code{message} if set).
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' noPlot(message = "No plot to show!")
#' noPlot(background = "#FF5500", size = 7)
#' @export
noPlot <- function(message = "Nothing to show here!",
                   size = 4.5, ...) {
  ggplot(data.frame(), aes(x = 0, y = 0, label = message)) +
    theme_lares(clean = TRUE, ...) +
    geom_text(size = size)
}


####################################################################
#' Export ggplot2, gridExtra, or any plot object into rendered file
#'
#' Export any \code{ggplot2}, \code{gridExtra}, or any plot object
#' created with R into rendered \code{png} or \code{jpg} file.
#'
#' @family Tools
#' @param p Plot object. Plot to render and export.
#' @param name Character. File's name or suffix if vars is not \code{NA}. No need
#' to include file format on file name.
#' @param vars Vector. Variable names to identify by filename.
#' @param sep Character. Separator for \code{vars}.
#' @param format Character. One of: \code{png} or \code{jpeg}.
#' @param width,height,res Numeric. Plot's width, height, and res (for grids).
#' @param dir,subdir Character. In which directory/subdirectory do you
#' wish to save the plot? Working directory as default \code{dir}.
#' @param quiet Boolean. Display successful message with filename when saved?
#' @return No return value, called for side effects.
#' @examples
#' p <- noPlot()
#' export_plot(p, name = "noplot", width = 10, height = 8, res = 300, dir = tempdir())
#' export_plot(p, name = "noplot2", subdir = "newplots", dir = tempdir())
#' @export
export_plot <- function(p,
                        name = "plot", vars = NA, sep = ".vs.",
                        width = 8, height = 6,
                        format = "png", res = 300,
                        dir = getwd(), subdir = NA,
                        quiet = FALSE) {
  check_opts(format, c("png", "jpeg"))

  # File name
  name <- sub("\\..[^\\.]*$", "", name)
  end <- paste0(".", format)
  if (!is.na(vars)) {
    names <- v2t(cleanText(as.character(vars), spaces = FALSE), sep = sep, quotes = FALSE)
    file_name <- paste0(name, "_", names, end)
  } else {
    file_name <- paste0(name, end)
  }

  # Create directory if needed
  if (!is.na(subdir)) {
    dir <- file.path(dir, subdir)
    if (!dir.exists(dir)) dir.create(dir)
  }

  file_name <- paste(dir, file_name, sep = "/")
  export_fx <- base::get(format)
  export_fx(file_name, height = height * res, width = width * res, res = res)
  plot(p)
  dev.off()

  if (!quiet) message(paste("Plot saved as", file_name))
}


####################################################################
#' Plot timeline as Gantt Plot
#'
#' This function plots groups of observartions with timelines in a
#' Gantt Plot way. Only works if start and end are date format values.
#'
#' @family Visualization
#' @param event Vector. Event, role, label, or row.
#' @param start Vector. Start date.
#' @param end Vector. End date. Only one day be default if not defined
#' @param label Vector. Place, institution, or label.
#' @param group Vector. Academic, Work, Extracurricular... Pass as factor
#' to keep a specific order
#' @param title Character. Title for the plot
#' @param subtitle Character. Subtitle for the plot
#' @param interactive Boolean. Run with plotly?
#' @param save Boolean. Save the output plot in our working directory
#' @param subdir Character. Into which subdirectory do you wish to save the plot to?
#' @return ggplot2 object
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' cols <- c("Role", "Place", "Type", "Start", "End")
#' today <- as.character(Sys.Date())
#' cv <- data.frame(rbind(
#'   c("Marketing Science Partner", "Facebook", "Work Experience", "2019-12-09", today),
#'   c("Data Scientist Consultant", "MatrixDS", "Work Experience", "2018-09-01", today),
#'   c("R Community Contributor", "lares library", "Extra", "2018-07-18", today),
#'   c("Lead Data Scientist", "MEG", "Work Experience", "2019-01-15", "2019-12-09"),
#'   c("Head of Analytics", "Comparamejor/R5", "Work Experience", "2016-08-01", "2019-01-15"),
#'   c("Big Data & Data Science Programme", "UdC", "Academic", "2017-09-01", "2018-02-28"),
#'   c("Project Engineer", "Polytex", "Work Experience", "2016-05-15", "2016-09-01"),
#'   c("Big Data Analyst", "MEG", "Work Experience", "2016-01-01", "2016-04-30"),
#'   c("Advanced Excel Instructor", "ARTS", "Work Experience", "2015-11-01", "2016-04-30"),
#'   c("Continuous Improvement Intern", "PAVCO", "Work Experience", "2015-04-01", "2015-08-30"),
#'   c("Mechanical Design Intern", "SIGALCA", "Work Experience", "2013-07-01", "2013-09-30"),
#'   c("DJs Online Community Owner", "LaresDJ.com / SoloParaDJs", "Extra", "2010-01-05", "2020-05-20"),
#'   c("Mechanical Engineer Degree", "USB", "Academic", "2009-09-15", "2015-11-20"),
#'   c("DJ and Composer/Producer", "Legacy Discplay", "Extra", "2009-05-01", "2015-04-30")
#' ))
#' colnames(cv) <- cols
#' plot_timeline(
#'   event = cv$Role,
#'   start = cv$Start,
#'   end = cv$End,
#'   label = cv$Place,
#'   # Simple trick to re-arrange the grids
#'   group = factor(cv$Type, levels = c("Work Experience", "Academic", "Extra"))
#' )
#' @export
plot_timeline <- function(event,
                          start, end = start + 1,
                          label = NA, group = NA,
                          title = "Curriculum Vitae Timeline",
                          subtitle = "Bernardo Lares",
                          interactive = FALSE,
                          save = FALSE,
                          subdir = NA) {
  # Let's gather all the data
  df <- data.frame(
    Role = as.character(event),
    Place = as.character(label),
    Start = as.Date(as.character(start)),
    End = as.Date(as.character(end)),
    Type = group
  )

  # Duplicate data for ggplot's geom_lines
  cvlong <- data.frame(
    pos = rep(as.numeric(rownames(df)), 2),
    name = rep(as.character(df$Role), 2),
    type = rep(factor(df$Type, ordered = TRUE), 2),
    where = rep(as.character(df$Place), 2),
    value = c(df$Start, df$End),
    label_pos = rep(df$Start + floor((df$End - df$Start) / 2), 2)
  )

  # Plot timeline
  maxdate <- max(df$End)
  p <- ggplot(cvlong, aes(
    x = .data$value,
    y = reorder(.data$name, -.data$pos),
    label = .data$where,
    group = .data$pos
  )) +
    geom_vline(xintercept = maxdate, alpha = 0.2) +
    labs(title = title, subtitle = subtitle, x = NULL, y = NULL, colour = NULL) +
    theme(
      panel.background = element_rect(fill = "white", colour = NA),
      axis.ticks = element_blank(),
      panel.grid.major.x = element_line(size = 0.25, colour = "grey80")
    )
  # scale_x_date(expand = c(0, 0))

  if (!is.na(cvlong$type)[1] || length(unique(cvlong$type)) > 1) {
    p <- p + geom_line(aes(colour = .data$type), size = 7) +
      facet_grid(.data$type ~ ., scales = "free", space = "free") +
      guides(colour = "none")
  }

  p <- p +
    geom_label(aes(x = .data$label_pos), colour = "black", size = 2, alpha = 0.7) +
    theme_lares(pal = 2, legend = "none")

  # Export file name and folder for plot
  if (save) {
    file_name <- "cv_timeline.png"
    if (!is.na(subdir)) {
      # dir.create(file.path(getwd(), subdir), recursive = T)
      file_name <- paste(subdir, file_name, sep = "/")
    }
    p <- p + ggsave(file_name, width = 8, height = 6)
    message(paste("Saved plot as", file_name))
  }

  if (interactive) {
    try_require("plotly")
    p <- ggplotly(p)
  }
  p
}


####################################################################
#' Chords Plot
#'
#' This auxiliary function plots discrete and continuous values results
#'
#' @family Visualization
#' @param origin,dest Vectors. Origin and destination vectors
#' @param weight Vector. Weight for each chord.
#' @param mg Numeric. Margin adjust for plot in case of need
#' @param title Character. Title for the plot
#' @param subtitle Character. Subtitle for the plot
#' @param pal Vector. Colour pallete. Order matters.
#' @return chordDiagram object
#' @examples
#' # You must have "circlize" library to use this auxiliary function:
#' \dontrun{
#' df <- data.frame(from = c(1, 1, 2, 3, 4, 1, 6), to = c(4, 4, 4, 2, 2, NA, NA))
#' plot_chord(df$from, df$to)
#' }
#' @export
plot_chord <- function(origin, dest,
                       weight = 1, mg = 3,
                       title = "Chord Diagram",
                       subtitle = "", pal = NA) {
  try_require("circlize")

  if (length(origin) != length(dest)) {
    stop("The origin and dest vectors should have the same length!")
  }

  df <- data.frame(origin, dest, weight) %>%
    mutate(
      origin = ifelse(.data$origin == "", " ", as.character(.data$origin)),
      dest = ifelse(.data$dest == "", " ", as.character(.data$dest))
    ) %>%
    replaceall(NA, "NA")
  colnames(df) <- c("orig_reg", "dest_reg", "flow")
  uniq <- unique(c(as.character(df$orig_reg), as.character(df$dest_reg)))

  if (is.na(pal)) pal <- names(lares_pal()$palette)

  if (length(unique(origin)) > length(pal)) {
    stop("Too many chords to plot and not enough colours :(")
  }

  col <- c(
    pal[seq_along(unique(origin))],
    rep("darkgrey", length(unique(uniq)) - length(unique(origin)))
  )

  chordDiagram(
    x = df,
    grid.col = col,
    transparency = 0.2, directional = 1,
    preAllocateTracks = list(
      track.height = uh(mg, "mm"),
      track.margin = c(uh(mg, "mm"), 0)
    ),
    direction.type = c("arrows", "diffHeight"), diffHeight = -0.04,
    annotationTrack = c("grid", "axis"), annotationTrackHeight = c(0.05, 0.1),
    link.arr.type = "big.arrow", link.sort = TRUE, link.largest.ontop = TRUE
  )

  title(main = title, line = -1, sub = subtitle, font.sub = 3, family = "Arial Narrow")
  legend("bottomright",
    pch = 15, col = col, legend = unique(origin),
    bg = "transparent", box.lty = 0, cex = 0.8
  )
}

LARES_DIGITS <- 0:9
LARES_OPERATORS <- c("+", "-", "*", "/")

numble_valid <- function(x,
                         result = eval(parse(text = x)),
                         len = 7, ...) {
  temp <- unlist(strsplit(x, ""))
  if (length(temp) != len) stop(paste("Provide only", len, "elements"))
  digits_check <- temp %in% as.character(LARES_DIGITS)
  ops_check <- temp %in% as.character(LARES_OPERATORS)
  not_valid <- !(digits_check | ops_check)
  if (any(not_valid)) stop("Not valid input values: ", v2t(temp[not_valid]))
  total <- eval(parse(text = x))
  if (total != result) stop(paste("Your operation must equal to", result, "not", total))
}

numble_check <- function(input, solution, len = 7, print = TRUE) {
  numble_valid(solution, len = len)
  result <- eval(parse(text = solution))
  numble_valid(input, result = result, len = len)
  out <- pos_check(input, solution, len = len, print = FALSE)
  if (print) {
    print(out)
    cat(paste(" =", result))
  }
  invisible(out)
}

####################################################################
#' Personal Colours Palette
#'
#' Fetch customizable palettes for the library's usage. The package has
#' its own default colour-blind friendly colours but can be customized using
#' R internal options (i.e. \code{options("lares.palette" = c("#FF8303" = "#000",
#' "#40A4D8" = "#FFF", ...))}. There are 3 options you can use to customize all
#' colour palletes: "lares.palette" (vector, will be used in the same order as passed,
#' and must have a counter colour defined), "lares.colours" (vector, simple colour
#' names and their HEX codes), and "lares.colours.custom" (data.frame, containing
#' "values" to use dynamically, "fill" for main colour, and "colour" (not obligatory)
#' for counter colour).
#'
#' @family Themes
#' @param return Character. Get only what you need. Select any of:
#' "all" or "list" (list),
#' "colors" or "colours" (vector),
#' "pal" or "palette" (named vector),
#' "simple" (named vector),
#' "custom" or "personal" (data.frame)
#' @return Depending on the \code{return} input, we get a:
#' \itemize{
#'   \item \code{vector} with \code{palette} results vector
#'   \item \code{vector} with \code{palette} results vector's names
#'   \item \code{list} with \code{palette} results vector, \code{labels}
#'   results data.frame, and \code{simple} results named vector
#' }
#' @examples
#' # Simple colour-named palette
#' lares_pal("simple")
#'
#' # Raw colours and counter-colours
#' # OR simply: lares_pal("palette")
#' nice_palette <- lares_pal("colours")
#' nice_palette_ctr <- as.vector(lares_pal()$palette)
#' lapply(list(nice_palette, nice_palette_ctr), head)
#'
#' # Personal colours by name
#' df <- lares_pal("custom")
#' df[sample(nrow(df), 5), ]
#' @export
lares_pal <- function(return = "list") {
  # Generic colour and counter-colour palette
  colours_names <- getOption("lares.palette")
  if (length(unique(names(colours_names))) != length(colours_names)) {
    stop("Check your lares.palette option. Do not repeat colours.")
  }
  if (!"character" %in% class(colours_names)) {
    stop("Check your lares.palette option. Must be a character vector.")
  }

  # Specific (simple) colours by names
  simple <- c(getOption("lares.colours"), getOption("lares.colors"))
  if (length(unique(names(simple))) != length(simple)) {
    stop("Check your lares.colours option. Do not repeat colours.")
  }
  if (!"character" %in% class(simple)) {
    stop("Check your lares.colours option. Must be a character vector.")
  }

  # Personal colours list (add more with lares.colours.custom options)
  colours_list <- data.frame(rbind(
    c("m", simple[["blue"]], simple[["white"]]),
    c("f", simple[["pink"]], simple[["black"]]),
    c("male", simple[["blue"]], simple[["white"]]),
    c("female", simple[["pink"]], simple[["black"]]),
    c("hombre", simple[["blue"]], simple[["white"]]),
    c("mujer", simple[["pink"]], simple[["black"]]),
    c("true", simple[["green"]], simple[["white"]]),
    c("false", simple[["red"]], simple[["black"]]),
    c("TRUE", simple[["green"]], simple[["white"]]),
    c("FALSE", simple[["red"]], simple[["black"]]),
    c("1", simple[["green"]], simple[["white"]]),
    c("0", simple[["red"]], simple[["black"]]),
    c("-1", simple[["red"]], simple[["black"]]),
    c("X", simple[["grey"]], simple[["black"]]),
    c("good", "#59B3D2", simple[["white"]]),
    c("bad", "#E5586E", simple[["black"]]),
    c("aprueba", "#59B3D2", simple[["white"]]),
    c("rechaza", "#E5586E", simple[["black"]]),
    c("ns/nc", "#E5586E", simple[["grey"]]),
    c("positive", "#59B3D2", simple[["black"]]),
    c("negative", "#E5586E", simple[["black"]]),
    c("bueno", "#59B3D2", simple[["white"]]),
    c("malo", "#E5586E", simple[["black"]]),
    c("spring", simple[["green"]], simple[["white"]]),
    c("summer", simple[["red"]], simple[["white"]]),
    c("fall", simple[["orange"]], simple[["black"]]),
    c("winter", simple[["blue"]], simple[["black"]]),
    c("meg1", "#071D49", simple[["white"]]),
    c("meg2", "#EBB600", simple[["black"]]),
    c("meg3", "#F2F1F0", simple[["black"]]),
    c("meg4", "#9A9A9A", simple[["white"]]),
    c("r5", "#290452", simple[["white"]]),
    c("otro", simple[["grey"]], simple[["black"]]),
    c("fear", "#810806", simple[["white"]]),
    c("disgust", "#BF200E", simple[["black"]]),
    c("anger", "#FE9B13", simple[["black"]]),
    c("sadness", "#838B8B", simple[["black"]]),
    c("anticipation", "#FE8A71", simple[["black"]]),
    c("surprise", "#F7E565", simple[["black"]]),
    c("trust", "#40A4D8", simple[["black"]]),
    c("joy", "#BD116F", simple[["black"]]),
    c("fb1", "#405996", simple[["white"]]),
    c("fb2", "#7184B2", simple[["black"]]),
    c("fb3", "#AFBEE3", simple[["black"]]),
    c("facebook", "#4267B2", simple[["black"]]),
    c("instagram", "#E95950", simple[["black"]]),
    c("messenger", "#0084FF", simple[["black"]]),
    c("whatsapp", "#4DC247", simple[["black"]])
  ))
  colnames(colours_list) <- c("values", "fill", "colour")

  more_cols <- getOption("lares.colours.custom")
  if (length(more_cols) > 1) {
    if (!"data.frame" %in% class(more_cols)) {
      stop("Check your lares.colours.custom option. Must be a data.frame.")
    }
    if (!"colour" %in% colnames(more_cols)) more_cols$colour <- "#000"
    if (!all(colnames(more_cols) %in% colnames(colours_list))) {
      stop(
        "Check your lares.colours.custom option. Column names must match: ",
        v2t(colnames(colours_list))
      )
    }
    colours_list <- colours_list %>%
      filter(!.data$values %in% more_cols$values) %>%
      bind_rows(more_cols)
  }

  pal <- list(
    labels = colours_list,
    palette = rep(colours_names, 4),
    simple = simple
  )

  if (return %in% c("colors", "colours", "color", "colour", "raw", "col")) {
    pal <- names(colours_names)
  }
  if (return %in% c("pal", "palette")) {
    pal <- colours_names
  }
  if (return %in% c("personal", "custom", "labels")) {
    pal <- colours_list
  }
  if (return %in% "simple") {
    pal <- simple
  }

  structure(pal, class = "lares_pal")
  attr("type", return)
  pal
}


####################################################################
#' Plot Palette Colours
#'
#' This function plots a list of colours
#'
#' @family Themes
#' @inheritParams cache_write
#' @param fill Vector. List of colours for fills.
#' @param colour Vector. List of colours for colours.
#' @param id Vector. ID for each color.
#' @param limit Integer. Show only first n values.
#' @return Plot with \code{fill} colours and \code{colour} counter-colours
#' if provided.
#' @examples
#' # Simply pass a vector
#' pal <- lares_pal("simple")
#' plot_palette(pal)
#' # Or fill + color named vector
#' pal <- lares_pal("pal")
#' plot_palette(fill = names(pal), colour = as.vector(pal))
#' @export
plot_palette <- function(fill, colour = "black", id = NA, limit = 12, ...) {
  if (length(fill) > limit) {
    fill <- fill[1:limit]
    colour <- colour[1:limit]
    colour <- colour[!is.na(colour)]
    message(paste("Limited to", limit, "colours. Overwrite with 'limit' parameter"))
  }
  if (is.na(id[1])) id <- seq_along(fill)
  data.frame(fill = fill, colour = colour, id = id) %>%
    distinct(.keep_all = TRUE) %>%
    ggplot(aes(x = reorder(fill, -id), y = 1)) +
    geom_bar(aes(fill = fill), stat = "identity", position = "dodge") +
    geom_text(aes(colour = colour, label = id), hjust = 1.5) +
    scale_fill_identity() +
    scale_colour_identity() +
    coord_flip() +
    labs(x = NULL, y = NULL) +
    guides(fill = "none", colour = "none") +
    theme_lares(font = NA, axis = "Y", ...)
}

####################################################################
#' Correlation table
#'
#' This function correlates a whole dataframe, running one hot smart
#' encoding (\code{ohse}) to transform non-numerical features.
#' Note that it will automatically suppress columns
#' with less than 3 non missing values and warn the user.
#'
#' @family Calculus
#' @family Correlations
#' @inheritParams ohse
#' @inheritParams numericalonly
#' @param df Dataframe. It doesn't matter if it's got non-numerical
#' columns: they will be filtered.
#' @param method Character. Any of: c("pearson", "kendall", "spearman").
#' @param use Character. Method for computing covariances in the presence
#' of missing values. Check \code{stats::cor} for options.
#' @param pvalue Boolean. Returns a list, with correlations and statistical
#' significance (p-value) for each value.
#' @param padjust Character. NULL to skip or any of \code{p.adjust.methods} to
#' calculate adjust p-values for multiple comparisons using \code{p.adjust()}.
#' @param half Boolean. Return only half of the matrix? The redundant
#' symmetrical correlations will be \code{NA}.
#' @param dec Integer. Number of decimals to round correlations and p-values.
#' @param dummy Boolean. Should One Hot (Smart) Encoding (\code{ohse()})
#' be applied to categorical columns?
#' @param top Integer. Select top N most relevant variables? Filtered
#' and sorted by mean of each variable's correlations.
#' @param ... Additional parameters passed to \code{ohse}, \code{corr},
#' and/or \code{cor.test}.
#' @return data.frame. Squared dimensions (N x N) to match every
#' correlation between every \code{df} data.frame column/variable. Notice
#' that when using \code{ohse()} you may get more dimensions.
#' @examples
#' data(dft) # Titanic dataset
#' df <- dft[, 2:5]
#'
#' # Correlation matrix (without redundancy)
#' corr(df, half = TRUE)
#'
#' # Ignore specific column
#' corr(df, ignore = "Pclass")
#'
#' # Calculate p-values as well
#' corr(df, pvalue = TRUE, limit = 1)
#'
#' # Test when no more than 2 non-missing values
#' df$trash <- c(1, rep(NA, nrow(df) - 1))
#' # and another method...
#' corr(df, method = "spearman")
#' @export
corr <- function(df, method = "pearson",
                 use = "pairwise.complete.obs",
                 pvalue = FALSE,
                 padjust = NULL,
                 half = FALSE,
                 dec = 6,
                 ignore = NULL,
                 dummy = TRUE,
                 redundant = NULL,
                 logs = FALSE,
                 limit = 10,
                 top = NA,
                 ...) {
  # Ignored columns
  if (isTRUE(is.na(ignore)[1])) ignore <- NULL
  df <- select(df, -any_of(ignore))

  # One hot encoding for categorical features
  if (dummy) {
    df <- ohse(df, quiet = TRUE, limit = limit, redundant = redundant, ...)
  }

  # Select only numerical features and create log+1 for each one
  d <- numericalonly(df, logs = logs)

  # Drop columns with not enough data to calculate correlations / p-values
  miss <- missingness(d, summary = FALSE)
  if (is.data.frame(miss)) {
    toDrop <- miss %>%
      mutate(drop = nrow(d) - .data$missing < 3L) %>%
      filter(.data$drop) %>%
      pull(.data$variable)
    if (length(toDrop) > 0) {
      warning("Dropped columns with less than 3 non-missing values: ", v2t(toDrop))
      d <- select(d, -one_of(toDrop))
    }
  }

  # Avoid sd = 0 warning:
  # In cor(x, y) : the standard deviation is zeroIn cor(x, y) : the standard deviation is zero
  d <- Filter(function(x) sd(x, na.rm = TRUE) != 0, d)

  # Correlations
  rs <- suppressWarnings(cor(d, method = method, use = use))
  if (half) for (i in seq_along(rs[, 1])) rs[1:i, i] <- NA
  cor <- round(data.frame(rs), dec)
  colnames(cor) <- row.names(cor) <- colnames(d)

  # Top N
  if (!is.na(top)) {
    message(paste("Returning the top", top, "variables only..."))
    imp <- cor %>%
      summarise_all(list(~ mean(.))) %>%
      t() %>%
      data.frame(variable = row.names(.), mean = abs(.)) %>%
      arrange(desc(abs(.data$mean)))
    which <- as.vector(imp$variable[1:top])
    cor <- cor[which, which]
  }

  # Statistical significance (p-value)
  if (pvalue) {
    output <- list(cor = cor, pvalue = .cor_test_p(d, method = method, ...))
    if (!is.null(padjust)) {
      output[["pvalue_adj"]] <- matrix(
        p.adjust(output$pvalue, method = padjust),
        ncol = ncol(output$cor),
        dimnames = list(names(output$cor), names(output$cor))
      )
    }
    output
  } else {
    cor
  }
}


####################################################################
#' Correlation between variable and dataframe
#'
#' This function correlates a whole dataframe with a single feature. It
#' automatically runs \code{ohse} (one-hot-smart-encoding) so no need to input
#' only numerical values.
#'
#' @family Exploratory
#' @family Correlations
#' @inheritParams corr
#' @inheritParams mp3_get
#' @param var Variable. Name of the variable to correlate. Note that if the
#' variable \code{var} is not numerical, 1. you may define which category to select
#' from using `var_category`; 2. You may have to add \code{redundant = TRUE} to
#' enable all categories (instead of \code{n-1}).
#' @param ignore Character vector. Which columns do you wish to exclude?
#' @param trim Integer. Trim words until the nth character for
#' categorical values (applies for both, target and values)
#' @param clean Boolean. Use lares::cleanText for categorical values (applies
#' for both, target and values)
#' @param plot Boolean. Do you wish to plot the result? If set to TRUE, the
#' function will return only the plot and not the result's data
#' @param top Integer. If you want to plot the top correlations,
#' define how many
#' @param ceiling Numeric. Remove all correlations above... Range: (0-1]
#' @param max_pvalue Numeric. Filter non-significant variables. Range (0, 1]
#' @param limit Integer. Limit one hot encoding to the n most frequent
#' values of each column. Set to \code{NA} to ignore argument.
#' @param ranks Boolean. Add ranking numbers?
#' @param zeroes Do you wish to keep zeroes in correlations too?
#' @param save Boolean. Save output plot into working directory
#' @param ... Additional parameters passed to \code{corr} and \code{cor.test}
#' @return data.frame. With variables, correlation and p-value results
#' for each feature, arranged by descending absolute correlation value.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#'
#' corr_var(dft, Survived, method = "spearman", plot = FALSE, top = 10)
#'
#' # With plots, results are easier to compare:
#'
#' # Correlate Survived with everything else and show only significant results
#' dft %>% corr_var(Survived_TRUE, max_pvalue = 0.05)
#'
#' # Top 15 with less than 50% correlation and show ranks
#' dft %>% corr_var(Survived_TRUE, ceiling = .6, top = 15, ranks = TRUE)
#' @export
corr_var <- function(df, var,
                     ignore = NULL,
                     trim = 0,
                     clean = FALSE,
                     plot = TRUE,
                     top = NA,
                     ceiling = 1,
                     max_pvalue = 1,
                     limit = 10,
                     ranks = FALSE,
                     zeroes = FALSE,
                     save = FALSE,
                     quiet = FALSE,
                     ...) {
  if (isTRUE(try(is.character(var), silent = TRUE))) {
    var <- eval(substitute(var), df)
  } else {
    var <- as_label(enquo(var))
  }

  df <- select(df, -contains(paste0(var, "_log")))

  # Calculate correlations
  rs <- corr(df, half = FALSE, ignore = ignore, limit = limit, pvalue = TRUE, ...)
  if (is.data.frame(rs)) rs <- list(cor = rs, pvalue = mutate_all(rs, ~1))

  # Check if main variable exists
  if (!var %in% colnames(rs$cor)) {
    msg <- paste("Not a valid input:", var, "was transformed or does not exist.")
    maybes <- colnames(rs$cor)[grepl(var, colnames(rs$cor))]
    if (length(maybes) > 0 && maybes[1] %in% colnames(rs$cor)) {
      msg <- sprintf("%s\n  >> Automatically using '%s'", msg, maybes[1])
      var <- maybes[1]
      fixable <- TRUE
    } else {
      fixable <- FALSE
    }
    if (fixable) warning(msg) else stop(msg)
  }

  d <- as_tibble(data.frame(
    variables = colnames(rs$cor),
    corr = rs$cor[, c(var)],
    pvalue = rs$pvalue[, c(var)]
  ))
  if ("pvalue_adj" %in% names(rs)) {
    d$pvalue_adj <- rs$pvalue_adj[, c(var)]
  }
  d <- d[(d$corr < 1 & !is.na(d$corr)), ]
  d <- d[order(-abs(d$corr)), ]

  original_n <- nrow(d)

  if (!zeroes) d <- d[d$corr != 0, ]

  # Suppress non-statistical significant correlations
  if ("pvalue_adj" %in% colnames(d)) {
    d <- d %>%
      mutate(max_pvalue = as.numeric(ifelse(is.na(.data$pvalue_adj), 1, .data$pvalue_adj))) %>%
      filter(.data$pvalue_adj <= max_pvalue)
    message(">>> Filtered 'max_pvalue' using adjusted p-values")
  } else {
    d <- d %>%
      mutate(pvalue = as.numeric(ifelse(is.na(.data$pvalue), 1, .data$pvalue))) %>%
      filter(.data$pvalue <= max_pvalue)
  }
  if (nrow(d) == 0) {
    warning(sprintf(paste(
      "Check your 'max_pvalue' input (%s):",
      "might be too low given there are no correlations matching this criteria"
    ), max_pvalue))
  }

  # Limit automatically when more than 20 observations
  if (is.na(top) && nrow(d) > 20) {
    top <- 20
    if (!quiet) message(paste(">>> Reduced results to", top, "largest correlations. Set by 'top' parameter"))
  }

  if (ceiling < 1) {
    d <- d[abs(d$corr) < ceiling, ]
    if (!quiet) message(paste0("Removing all correlations greater than ", 100 * ceiling, "% (absolute)"))
  }

  d <- d[complete.cases(d), ]

  if (!is.na(top)) d <- head(d, top)

  # Shorten up the long names of some variables
  if (trim > 0) {
    d$variables <- substr(d$variables, 1, trim)
    if (!quiet) message(paste("Trimmed all name values into", trim, "characters"))
  }

  # Add ranking numbers
  if (ranks) {
    d <- mutate(d, variables = sprintf("%s. %s", row_number(), .data$variables))
  }

  class(d) <- c("corr_var", class(d))

  if (plot) {
    plot(d, var, max_pvalue = max_pvalue, top = top, limit = original_n)
  } else {
    d
  }
}

#' @param x corr_var object
#' @rdname corr_var
#' @export
plot.corr_var <- function(x, var, max_pvalue = 1, top = NA, limit = NULL, ...) {
  pal <- lares_pal("labels")
  stopifnot(is.data.frame(x))
  if (nrow(x) == 0) {
    noPlot("No data available")
  } else {
    p <- ungroup(x) %>%
      filter(!.data$variables %in% c("pvalue", "pvalue_adj")) %>%
      mutate(
        pos = ifelse(.data$corr > 0, pal$fill[pal$values == "positive"],
          pal$fill[pal$values == "negative"]
        ),
        hjust = ifelse(abs(.data$corr) < max(abs(.data$corr)) / 1.5, -0.1, 1.1)
      ) %>%
      ggplot(aes(
        x = reorder(.data$variables, abs(.data$corr)),
        y = abs(.data$corr), fill = .data$pos,
        label = sub("^(-)?0[.]", "\\1.", signif(.data$corr, 3))
      )) +
      geom_hline(yintercept = 0, alpha = 0.5) +
      geom_col(colour = "transparent") +
      coord_flip() +
      geom_text(aes(hjust = .data$hjust), size = 3, colour = "black") +
      guides(fill = "none") +
      labs(title = paste("Correlations of", var), x = NULL, y = NULL) +
      scale_fill_identity() +
      scale_y_continuous(
        expand = c(0, 0), position = "right",
        labels = function(x) sub("^(-)?0[.]", "\\1.", x)
      ) +
      theme_lares(...)

    if (is.null(limit)) limit <- 100
    if (!is.na(top) && top < limit) {
      p <- p + labs(subtitle = paste(top, "largest correlation variables (original & dummy)"))
    }

    if (max_pvalue < 1) {
      p <- p + labs(caption = paste("Correlations with p-value <", max_pvalue))
    }
    p
  }
}


####################################################################
#' Ranked cross-correlation across all variables
#'
#' This function creates a correlation full study and returns a rank
#' of the highest correlation variables obtained in a cross-table.
#'
#' For a detailed tutorial on using this function, see:
#' \href{https://www.r-bloggers.com/2019/08/find-insights-with-ranked-cross-correlations/}{Find Insights with Ranked Cross-Correlations}
#'
#' @family Calculus
#' @family Correlations
#' @inheritParams corr
#' @param df data.frame
#' @param plot Boolean. Show and return a plot?
#' @param max_pvalue Numeric. Filter non-significant variables. Range (0, 1]
#' @param type Integer. Plot type. 1 is for overall rank. 2 is for local rank.
#' @param max Numeric. Maximum correlation permitted (from 0 to 1)
#' @param top Integer. Return top n results only. Only valid when type = 1. Set
#' value to NA to use all cross-correlations
#' @param local Integer. Label top n local correlations. Only valid when type = 2
#' @param contains Character vector. Filter cross-correlations
#' with variables that contains certain strings (using any value if vector used).
#' @param grid Boolean. Separate into grids?
#' @param rm.na Boolean. Remove NAs?
#' @param quiet Boolean. Keep quiet? If not, show messages
#' @param ... Additional parameters passed to \code{corr}
#' @return Depending on input \code{plot}, we get correlation and p-value results for
#' every combination of features, arranged by descending absolute correlation value,
#' with a data.frame \code{plot = FALSE} or plot \code{plot = TRUE}.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#'
#' # Only data with no plot
#' corr_cross(dft, plot = FALSE, top = 10)
#'
#' # Show only most relevant results filtered by pvalue
#' corr_cross(dft, rm.na = TRUE, max_pvalue = 0.05, top = 15)
#'
#' # Cross-Correlation for certain variables
#' corr_cross(dft, contains = c("Survived", "Fare"))
#'
#' # Cross-Correlation max values per category
#' corr_cross(dft, type = 2, top = NA)
#' @export
corr_cross <- function(df, plot = TRUE,
                       pvalue = TRUE, max_pvalue = 1,
                       type = 1, max = 1, top = 20, local = 1,
                       ignore = NULL, contains = NA, grid = TRUE,
                       rm.na = FALSE, quiet = FALSE,
                       ...) {
  check_opts(type, 1:2)

  cor <- corr(df, half = TRUE, ignore = ignore, pvalue = pvalue, ...)

  cluster <- "cluster_" %in% contains
  if (!is.data.frame(cor)) {
    ret <- .transf(cor$cor, max = max, contains = contains, cluster = cluster, rm.na = rm.na)
    aux <- .transf(cor$pvalue, max = max, contains = contains, cluster = cluster, rm.na = rm.na)
  } else {
    ret <- aux <- .transf(cor, max = max, contains = contains, cluster = cluster, rm.na = rm.na)
    aux$corr <- 0
  }

  ret <- left_join(ret, rename(aux, pvalue = .data$corr), c("key", "mix")) %>%
    mutate(pvalue = as.numeric(ifelse(is.na(.data$pvalue), 1, .data$pvalue))) %>%
    filter(.data$pvalue <= max_pvalue)

  ret <- ret %>%
    mutate(group1 = "fill", group2 = "fill")

  for (group in colnames(df)) {
    ret <- ret %>%
      mutate(
        group1 = ifelse(.data$group1 == "fill" & grepl(group, .data$key), group, .data$group1),
        group2 = ifelse(.data$group2 == "fill" & grepl(group, .data$mix), group, .data$group2)
      )
  }
  ret <- filter(ret, .data$group1 != .data$group2)
  if (nrow(ret) > top && !is.na(top) && !quiet && type != 2) {
    message(sprintf("Returning only the top %s. You may override with the 'top' argument", top))
    ret <- slice(ret, 1:top)
  }

  ret <- ret %>%
    rowwise() %>%
    mutate(
      cat1 = gsub(paste0(.data$group1, "_"), "", .data$key),
      cat2 = gsub(paste0(.data$group2, "_"), "", .data$mix)
    ) %>%
    select(1:4, .data$group1, .data$cat1, .data$group2, .data$cat2)

  if (plot) {
    n <- ifelse(type == 1, top, local)
    n <- ifelse(n > nrow(ret), nrow(ret), n)
    subtitle <- paste(n, "most relevant")
    if (!is.na(contains)[1]) subtitle <- paste(subtitle, "containing", vector2text(contains))
    if (max < 1) subtitle <- paste0(subtitle, " (excluding +", 100 * max, "%)")
    if (rm.na) subtitle <- paste(subtitle, paste("[NAs removed]"))
    if (!is.na(contains[1])) {
      ret <- ret %>%
        # Pick the one that contains is contained in key or mix
        mutate(facet = ifelse(grepl(paste(contains, collapse = "|"), .data$key), .data$key, .data$mix)) %>%
        mutate(facet = gsub("_", "", .data$facet))
    }

    if (type == 1) {
      p <- ret %>%
        head(top) %>%
        ungroup() %>%
        mutate(
          label = paste(.data$key, "+", .data$mix),
          abs = abs(.data$corr),
          sign = ifelse(.data$corr < 0, "negative", "positive"),
          hjust = ifelse(.data$abs < max(.data$abs) / 1.5, -0.1, 1.1)
        ) %>%
        ggplot(aes(
          x = reorder(.data$label, .data$abs),
          y = .data$abs,
          label = sub("^(-)?0[.]", "\\1.", signif(.data$corr, 3))
        )) +
        geom_col(aes(fill = .data$sign)) +
        coord_flip() +
        geom_text(aes(hjust = .data$hjust), size = 3, colour = "black") +
        guides(fill = "none", colour = "none") +
        labs(
          title = "Ranked Cross-Correlations",
          subtitle = subtitle,
          x = NULL, y = NULL
        ) +
        scale_y_continuous(
          expand = c(0, 0), position = "right",
          labels = function(x) sub("^(-)?0[.]", "\\1.", x)
        ) +
        theme_lares(pal = 4, ...)

      if (!is.na(contains)[1] && grid) {
        p <- p + facet_grid(.data$facet ~ ., scales = "free", space = "free")
      }
    }

    if (type == 2) {
      ret <- rbind(
        data.frame(ret, group = ret$group1),
        data.frame(ret, group = ret$group2)
      ) %>%
        arrange(desc(abs(.data$corr)))
      aux <- position_jitter(width = 0.4, seed = 123)
      p <- ret %>%
        group_by(.data$group) %>%
        mutate(
          hjust = ifelse(.data$corr > 0, 1, 0),
          size = abs(.data$corr),
          alpha = ifelse(row_number() <= local, 2, .data$size),
          label = ifelse(row_number() <= local, paste(.data$key, "+", .data$mix), "")
        ) %>%
        ggplot(aes(
          x = .data$group, y = .data$corr,
          label = .data$label, colour = .data$group
        )) +
        geom_jitter(position = aux, alpha = 0.4) +
        geom_hline(yintercept = 0, alpha = 0.3) +
        geom_text(aes(hjust = .data$hjust), size = 2.9, position = aux, colour = "black") +
        guides(colour = "none", alpha = "none", size = "none") +
        scale_size(range = c(0.4, 2)) +
        labs(
          x = NULL, y = "Correlation",
          subtitle = subtitle,
          title = "Local Cross-Correlations"
        ) +
        scale_y_continuous(labels = function(x) sub("^(-)?0[.]", "\\1.", x)) +
        coord_flip() +
        theme_lares(pal = 2, ...)
    }
    if (max_pvalue < 1) {
      p <- p + labs(caption = paste("Correlations with p-value <", max_pvalue))
    }
    p
  } else {
    ret
  }
}

.transf <- function(x, max = 1, contains = NA, cluster = FALSE, rm.na = FALSE) {
  aux <- gather(data.frame(x), "key", "value")
  ret <- mutate(aux, mix = rep(unique(aux$key), length = nrow(aux))) %>%
    mutate(p1 = cumsum(!duplicated(.data$mix))) %>%
    group_by(.data$mix) %>%
    mutate(p2 = row_number()) %>%
    ungroup() %>%
    filter(.data$p2 != .data$p1) %>%
    mutate(rel = abs(.data$value)) %>%
    filter(.data$rel < max) %>%
    arrange(desc(.data$rel)) %>%
    {
      if (!is.na(contains[1])) {
        filter(., grepl(
          paste(contains, collapse = "|"),
          paste(.data$key, .data$mix)
        ))
      } else {
        .
      }
    } %>%
    {
      if (rm.na) filter(., !grepl("_NAs", .data$mix)) else .
    } %>%
    filter(!grepl("_OTHER", .data$key)) %>%
    rename(corr = .data$value) %>%
    mutate(value = paste(.data$key, .data$mix)) %>%
    select(.data$key, .data$mix, .data$corr)
  ret
}


# https://stackoverflow.com/questions/60512043/r-creating-a-p-value-matrix-with-missing-values
.cor_test_p <- function(mat, method = "pearson", exact = TRUE, ...) {
  mat <- as.matrix(mat)
  n <- ncol(mat)
  p.mat <- matrix(NA, n, n)
  diag(p.mat) <- 0
  for (i in 1:(n - 1)) {
    for (j in (i + 1):n) {
      error <- try(tmp <- suppressWarnings(cor.test(
        mat[, i], mat[, j],
        method = method, exact = exact, ...
      )), silent = TRUE)
      if (inherits(error, "try-error")) {
        p.mat[i, j] <- NA
      } else {
        p.mat[i, j] <- p.mat[j, i] <- tmp$p.value
      }
    }
  }
  colnames(p.mat) <- rownames(p.mat) <- colnames(mat)
  p.mat
}

####################################################################
#' New Line Feed for Long Strings (Wrapper)
#'
#' Add a break or new line without breaking words. Automatically,
#' the function can detect your plot's width and will dynamically
#' set an auto width. You can adjust the relation (rel) parameter
#' for different fonts and sizes until perfect harmony found.
#' Quite similar to \code{stringr::str_wrap} but, if the text vector
#' is a factor, the levels will be kept in order and transformed.
#'
#' @family Tools
#' @param text Character or factor vector.
#' @param top Integer. How many characters aprox. should be on each line?
#' @param rel Numeric. Relation of pixels and characters per line
#' @return Character. String (vector) including some \code{\\n} within.
#' @examples
#' cat(autoline("This is a long text that may not fit into a single line", 8))
#'
#' text <- factor(c("First value", "Second value", "First value"),
#'   levels = c("First value", "Second value")
#' )
#' autoline(text, 1)
#'
#' path <- file.path(R.home("doc"), "THANKS")
#' text <- paste(readLines(path), collapse = " ")
#' cat(autoline(text))
#' @export
autoline <- function(text, top = "auto", rel = 9) {
  # Auto-maximum
  if (top == "auto") top <- round(dev.size("px")[1] / rel)

  # Keep factors
  is_factor <- is.factor(text)
  if (is_factor) {
    levs <- levels(text)
    text <- as.character(text)
  }

  ret <- stringr::str_wrap(text, top)

  if (is_factor) {
    aux <- data.frame(ret, text) %>%
      mutate(text = factor(text, levels = levs)) %>%
      arrange(text) %>%
      unique()
    ret <- factor(ret, levels = aux$ret)
  }
  ret
}


####################################################################
#' Format a string text as markdown/HTML
#'
#' Format any character string to HTML or markdown format. We
#' recommend using this format with the \code{ggtext::geom_richtext}
#' function to format text in \code{ggplot2} objects.
#'
#' @family Tools
#' @param text Character. Strings to format.
#' @param color Character. Hex colour code.
#' @param size Numeric. Text size.
#' @param bold Boolean. Should the text be bold?
#' @return String with format characters included.
#' @examples
#' formatHTML("Text test", color = "#000000")
#' formatHTML(c(123, 456), color = "orange", size = 120, bold = TRUE)
#'
#' # If you want to use it with \code{ggtext}:
#' \dontrun{
#' col1 <- "grey"
#' col2 <- "orange"
#' pt <- data.frame(
#'   label = paste0(
#'     formatHTML(123, color = col2, size = 120, bold = TRUE), "<br/>",
#'     formatHTML("of children had a", col1), "<br/>",
#'     formatHTML("traditional stay-at-home mom", color = col2, bold = TRUE), "<br/>",
#'     formatHTML(paste0("in 2012, compared to ", 321, " in 1970"), color = col1)
#'   )
#' )
#' ggplot(pt, aes(x = 0, y = 0)) +
#'   ggtext::geom_richtext(
#'     aes(label = label),
#'     hjust = 0,
#'     label.color = NA,
#'     lineheight = 1.5
#'   ) +
#'   xlim(0, 0.01) +
#'   theme_void()
#' }
#' @export
#' @rdname format_string
formatHTML <- function(text, color = "black", size = 20, bold = FALSE) {
  opening_span <- paste0("<span style='font-size:", size, "px; color:", color, "'>")
  if (bold) text <- paste0("**", text, "**")
  closing_span <- "</span>"
  text <- paste(text, collapse = "<br/>")
  ret <- paste0(opening_span, text, closing_span)
  ret
}


####################################################################
#' Interpolate a string [glue wrapper]
#'
#' Format and interpolate a string using a \code{glue} wrapper. Allows
#' simple operations, \code{NULL} values as input, and interactions with
#' internal (created within \code{glued}) and external (environment) objects.
#'
#' @family Tools
#' @inheritParams stringr::str_glue
#' @param empty_lines Character. Set to \code{"keep"} to keep or
#' \code{"drop"} to drop empty lines.
#' @return Same as input but transformed (glued).
#' @examples
#' name <- "Bernardo"
#' age <- 29
#' anniversary <- as.Date("2016-04-30")
#' glued("
#'   My name is {name},
#'   my age next year will be {age + 1},
#'   and I got married on {format(anniversary, '%A, %B %d, %Y')}.")
#'
#' # Single braces can be inserted by doubling them
#' glued("My name is {name}, not {{name}}.")
#'
#' # You can also used named arguments
#' glued(
#'   "Her name is {name}, ",
#'   "and her age next year will be {age + 1}.",
#'   name = "Maru",
#'   age = 6
#' )
#'
#' # And run operations with memories (beware!)
#' glued("My name, {name}, has {n <- nchar(name); n} characters.
#'        If we multiply by ten, we'll have {10 * n} characters!")
#'
#' # If you pass a vector, the operation will be repeated for each element
#' glued("Here's the value #{1:3}")
#' @export
glued <- function(..., .sep = "", empty_lines = "keep", .envir = parent.frame()) {
  null_transformer <- function(text, envir) {
    out <- eval(parse(text = text, keep.source = FALSE), envir)
    if (is.null(out)) out <- ""
    out
  }
  output <- stringr::str_glue(
    ...,
    .sep = .sep,
    .transformer = null_transformer,
    .envir = .envir
  )
  if (empty_lines == "drop") {
    lines <- stringr::str_split(output, "\n")[[1]]
    output <- glued(paste(lines[trimws(lines) != ""], collapse = "\n"))
  }
  output
}


####################################################################
#' Pattern Matching for Any or All Multiple Matches
#'
#' This function returns a boolean vector of the same length as `x`,
#' each element of which is the result of applying the `type` of matches
#' to the corresponding element of `x`, using regular expressions.
#'
#' @family Tools
#' @inheritParams base::grep
#' @param x Character vector. Text where matches are sought, or an object
#' which can be coerced by as.character to a character vector.
#' Long vectors are supported.
#' @param type Character. Type of match. Choose one of:
#' \code{any}, \code{all}
#' @param ... Additional arguments to pass to \code{grepl}
#' @return Boolean of same length as \code{x}
#' @examples
#' x <- c(123, 876, 18761)
#' patterns <- c(1, 2)
#' grepm(patterns, x, type = "any")
#' grepm(patterns, x, type = "all")
#' @export
grepm <- function(pattern, x, type = "all", ...) {
  lapply(x, function(a) lapply(pattern, function(i) grepl(i, a, ...))) %>%
    lapply(get(type)) %>%
    unlist() %>%
    suppressWarnings()
}

####################################################################
#' Print Coloured Messages
#'
#' @family Tools
#' @param txt Character. Text to print or transform.
#' @param colour Character. Any of: grey, red, green, yellow, blue, or purple.
#' @param bold Boolean. Set bold text?
#' @param cat Boolean. Print with cat? If not, raw string
#' @return Depends on \code{cat}: NULL if TRUE or character string if FALSE.
#' @examples
#' opts <- c("GREY", "RED", "GREEN", "YELLOW", "BLUE", "PURPLE")
#' for (colour in opts) formatColoured(paste("Colour:", colour, "\n"), colour)
#' formatColoured("my bold coloured text", bold = TRUE, cat = TRUE)
#' @export
formatColoured <- function(txt, colour = c("yellow", "blue", "grey"), bold = FALSE, cat = TRUE) {
  colour <- toupper(colour)[1]
  opts <- c("GREY", "RED", "GREEN", "YELLOW", "BLUE", "PURPLE", "CYAN", "WHITE")
  check_opts(colour, opts)
  if (colour == opts[1]) code <- 30
  if (colour == opts[2]) code <- 31
  if (colour == opts[3]) code <- 32
  if (colour == opts[4]) code <- 33
  if (colour == opts[5]) code <- 34
  if (colour == opts[6]) code <- 35
  if (colour == opts[7]) code <- 36
  if (colour == opts[8]) code <- 37
  out <- paste0("\033[", ifelse(!bold, 0, 1), ";", code, "m", txt, "\033[0m")
  if (cat) cat(out) else out
}

####################################################################
#' Test the Truth of R Expressions and Warn
#'
#' If the expression in ... is not \code{TRUE}, \code{warning} is called,
#' producing a warning message indicating the expression which was not true.
#'
#' @family Tools
#' @param ... any R expression, which should evaluate to TRUE
#' @examples
#' warnifnot(TRUE)
#' warnifnot(FALSE)
#' warnifnot(1 + 1 == 3)
#' @export
warnifnot <- function(...) if (!isTRUE(...)) warning(paste(deparse(...), "is not TRUE"))

####################################################################
#' Send Slack Message (Webhook)
#'
#' This function send a Slack message using its Webhooks.
#'
#' For more help, you can follow the
#' \href{https://api.slack.com/messaging/webhooks#posting_with_webhooks}{Sending messages using Incoming Webhooks}
#' original documentarion.
#'
#' @family API
#' @family Credentials
#' @param text,title,pretext Character. Content on you Slack message.
#' @param hook Character. Web hook URL. Ths value will be overwritten by
#' creds if correctly used.
#' @param creds Character. Credential's dir (see \code{get_creds()}). Set
#' hook URL into the "slack" list in your YML file. Will use first value.
#' @return Invisible POST response
#' @examples
#' \dontrun{
#' slackSend(text = "This is a message", title = "TEST", pretext = Sys.info()["user"])
#' }
#' @export
slackSend <- function(text, title = "", pretext = "",
                      hook = NA,
                      creds = NA) {
  if (is.na(text)) {
    stop("You must write something for message to be sent!")
  }

  if (is.na(hook)) {
    c <- get_credentials(from = "slack", dir = creds)
    hook <- c[[1]]
  }

  aux <- paste(
    '{"attachments": [{',
    '"title": "', title, '",',
    '"pretext": "', text, '",',
    '"text": "', pretext, '",
    "color": "#ee0000"',
    "}]}",
    sep = ""
  )

  invisible(POST(hook, body = aux))
}

####################################################################
#' Cache Save and Load (Write and Read)
#'
#' This function lets the user save and load a cache of any R object to
#' improve timings and UX.
#'
#' @family Cache
#' @inheritParams mp3_get
#' @param data Object
#' @param base Character vector. Unique name for your cache file. You can pass
#' a character vector with multiple elements that will be concatenated.
#' All cache files with start with \code{lares_cache_*} automatically to
#' quickly detect these cache files.
#' @param cache_dir Character. Where do you want to save you cache files?
#' By default they'll be stored on \code{tempdir()} but you can change it
#' using this parameter or setting a global option called \code{"LARES_CACHE_DIR"}.
#' @param ask Boolean. If cache exists, when reading: (interactive) ask the user
#' if the cache should be used to proceed or ignored; when writing, (interactive)
#' ask the user if the cache should be overwritten. Note that you can only ask for
#' one cache file at a time because vectors are concatenated.
#' @param overwrite Boolean. Set to overwrite existing cache file. When reading,
#' this parameter answers to ask prompt instead.
#' @param ... Additional parameters.
#' @return \code{cache_write}. No return value, called for side effects.
#' @examples
#' x <- list(a = 1, b = 2:4)
#' base <- c(as.character(Sys.Date()), "A", "B")
#' cache_write(x, base)
#' cache_read(base, ask = FALSE)
#' cache_exists(base)
#' cache_clear()
#'
#' # Use cache_pipe() for automatic use (saves and loads if exists)
#' cache_pipe({
#'   x <- 1
#'   x + 1
#' })
#' cache_pipe({
#'   x <- 1
#'   x + 1
#' })
#' @export
cache_write <- function(data,
                        base = "temp",
                        cache_dir = getOption("LARES_CACHE_DIR"),
                        ask = FALSE,
                        overwrite = TRUE,
                        quiet = FALSE,
                        ...) {
  if (is.null(cache_dir)) {
    cache_dir <- tempdir()
  }
  base <- paste(base, collapse = ".")
  if (left(base, 12) != "lares_cache_") base <- paste0("lares_cache_", base)
  if (right(base, 4) == ".RDS") base <- gsub("\\.RDS", "", base)
  file <- sprintf("%s/%s.RDS", cache_dir, base)
  if (nchar(file) >= 252) {
    stop("Your file name can't contain more than 250 characters.")
  }
  if (dir.exists(cache_dir)) {
    if (cache_exists(file) && ask) {
      message("> Cache found: ", base)
      answer <- readline("Press ENTER to rewrite cache or type [i] to ignore: ")
    } else {
      answer <- "use"
    }
    if (answer != "i" && !isFALSE(overwrite)) {
      saveRDS(data, file = file)
      if (!quiet) message("> Cache saved succesfully: ", base)
    } else {
      if (!quiet) message("> Skipped writing cache for: ", base)
      invisible(NULL)
    }
  } else {
    stop("Not a valid directory: ", cache_dir)
  }
}

#' @rdname cache_write
#' @return \code{cache_read}. R object. Data from cache file or NULL if no cache found.
#' @export
cache_read <- function(base,
                       cache_dir = getOption("LARES_CACHE_DIR"),
                       ask = FALSE,
                       overwrite = TRUE,
                       quiet = FALSE,
                       ...) {
  base <- paste(base, collapse = ".")
  if (left(base, 12) != "lares_cache_") base <- paste0("lares_cache_", base)
  if (right(base, 4) == ".RDS") base <- gsub("\\.RDS", "", base)
  exists <- cache_exists(base, cache_dir)
  if (exists) {
    file <- attr(exists, "filename")
    base <- attr(exists, "base")
    if (ask) {
      message("> Cache found: ", base)
      answer <- readline("Press ENTER to use cache or type [i] to ignore: ")
    } else {
      answer <- "use"
    }
    if (answer != "i" && isTRUE(overwrite)) {
      data <- readRDS(file)
      if (!quiet) message("> Cache loaded succesfully: ", base)
      data
    }
  } else {
    if (!quiet) {
      if (!is.function(file)) {
        message("No cache found for: ", file)
      } else {
        message("No cache file found for ", base)
      }
    }
    invisible(NULL)
  }
}

#' @rdname cache_write
#' @return \code{cache_exists}. Boolean. Result of \code{base} existence.
#' @export
cache_exists <- function(base = NULL, cache_dir = getOption("LARES_CACHE_DIR"), ...) {
  if (is.null(cache_dir)) {
    cache_dir <- tempdir()
  }
  if (!is.null(base)) {
    base <- paste(base, collapse = ".")
    if (left(base, 12) != "lares_cache_") base <- paste0("lares_cache_", base)
    if (right(base, 4) == ".RDS") base <- gsub("\\.RDS", "", base)
    filename <- sprintf("%s/%s.RDS", cache_dir, base)
  } else {
    files_dir <- list.files(cache_dir)
    base <- files_dir[startsWith(files_dir, "lares_cache_")]
    filename <- sprintf("%s/%s", cache_dir, base)
  }
  exists <- any(file.exists(filename))
  attr(exists, "filename") <- filename
  attr(exists, "base") <- base
  attr(exists, "cache_dir") <- cache_dir
  exists
}

#' @rdname cache_write
#' @return \code{cache_clear}. Invisible vector containing cache file names removed.
#' @export
cache_clear <- function(cache_dir = getOption("LARES_CACHE_DIR"), quiet = FALSE, ...) {
  if (is.null(cache_dir)) cache_dir <- tempdir()
  files <- list.files(cache_dir)
  caches <- files[startsWith(files, "lares_cache_")]
  if (length(caches) > 0) {
    invisible(file.remove(sprintf("%s/%s", cache_dir, caches)))
  }
  if (!quiet) {
    if (length(caches) > 0) {
      message(paste("Removed", length(caches), "cache files succesfully!"))
    } else {
      message("No cache files available to be removed.")
    }
  }
  invisible(caches)
}

#' @rdname cache_write
#' @param read,write Boolean. Do you wish to read and or write cache?
#' @return \code{cache_pipe}. Same as \code{x} or cached result.
#' @export
cache_pipe <- function(data, base = "cache_pipe", read = TRUE, write = TRUE, ...) {
  if (isTRUE(read) && lares::cache_exists(base, ...)) {
    cache_read(base, ...)
  } else {
    if (isTRUE(write)) {
      data <- eval(data)
      cache_write(data, base, ...)
    }
    data
  }
}

####################################################################
#' Solve Sudoku Puzzles
#'
#' Solve a Sudoku puzzle, where empty values are represented by 0s
#' into a matrix object.
#'
#' @family Games
#' @param board Matrix. 9x9 matrix or vector length 81, with only digits from 0 to 9.
#' @param needed_cells,index Auxiliary parameters to auto-iterate using this same fx.
#' @param quiet Boolean. Keep quiet? If not, plot results.
#' @return Logical output answering of the input board can be solved. The
#' actual solved solution will be created as an object named \code{solved}
#' in your \code{.GlobalEnv}.
#' @examples
#' \donttest{
#' # board <- c(0,0,0,0,0,6,0,0,0,
#' #            0,9,5,7,0,0,3,0,0,
#' #            4,0,0,0,9,2,0,0,5,
#' #            7,6,4,0,0,0,0,0,3,
#' #            0,0,0,0,0,0,0,0,0,
#' #            2,0,0,0,0,0,9,7,1,
#' #            5,0,0,2,1,0,0,0,9,
#' #            0,0,7,0,0,5,4,8,0,
#' #            0,0,0,8,0,0,0,0,0)
#' # sudoku_solver(board)
#'
#' # Trivial input (everything)
#' trivial <- matrix(rep(0, 81), byrow = TRUE, ncol = 9)
#' trivial
#' sudoku_solver(trivial)
#'
#' # Wrong / Impossible to solve input
#' imp <- matrix(c(rep(1, 72), rep(0, 9)), byrow = TRUE, ncol = 9)
#' imp
#' sudoku_solver(imp)
#' }
#' @export
sudoku_solver <- function(board, needed_cells = NULL, index = 1, quiet = FALSE) {
  # Convert vector to matrix
  if (is.vector(board)) {
    ints_split <- as.integer(unlist(str_split(board, pattern = "", n = 9 * 9)))
    board <- matrix(ints_split, byrow = TRUE, ncol = 9)
  }

  if (!all(dim(board) == 9)) {
    stop("Check your input's dimensions. 9x9 digits needed.")
  }

  # Determine empty cells to fill
  if (is.null(needed_cells)) {
    needed_cells <- which(board == 0, arr.ind = TRUE)
  }

  # Base case: all cells filled
  if (index > nrow(needed_cells)) {
    if (!quiet) print(board)
    TRUE
  } else {
    row <- needed_cells[index, 1]
    col <- needed_cells[index, 2]

    # Try digits 1 through 9
    solved <- FALSE
    for (num in 1:9) {
      if (.sudoku_valid_input(board, num, row, col)) {
        board2 <- board
        board2[row, col] <- num
        solved <- sudoku_solver(board2, needed_cells, index + 1, quiet)
        if (isTRUE(solved)) break
      }
    }
    solved
  }
}

.sudoku_valid_input <- function(board, i, row, col) {
  invalid_row <- any(board[row, ] == i)
  invalid_col <- any(board[, col] == i)

  box_x <- floor((row - 1) / 3) + 1
  box_y <- floor((col - 1) / 3) + 1
  box <- board[(3 * box_x - 2):(3 * box_x), (3 * box_y - 2):(3 * box_y)]
  invalid_box <- any(box == i)

  !(invalid_row || invalid_col || invalid_box)
}

####################################################################
#' Ranked Predictive Power of Cross-Features (x2y)
#'
#' The relative reduction in error when we go from a baseline model
#' (average for continuous and most frequent for categorical features) to
#' a predictive model, can measure the strength of the relationship between
#' two features. In other words, \code{x2y} measures the ability of \code{x}
#' to predict \code{y}. We use CART (Classification And Regression Trees) models
#' to be able to 1) compare numerical and non-numerical features, 2) detect
#' non-linear relationships, and 3) because they are easy/quick to train.
#'
#' This \code{x2y} metric is based on Rama Ramakrishnan's
#' \href{https://bit.ly/3sOVbei}{post}: An Alternative to the Correlation
#' Coefficient That Works For Numeric and Categorical Variables. This analysis
#' complements our \code{lares::corr_cross()} output.
#'
#' @param df data.frame. Note that variables with no variance will be ignored.
#' @param target Character vector. If you are only interested in the \code{x2y}
#' values between particular variable(s) in \code{df}, set
#' name(s) of the variable(s) you are interested in. Keep \code{NULL}
#' to calculate for every variable (column). Check \code{target_x} and
#' \code{target_y} parameters as well.
#' @param confidence Boolean. Calculate 95\% confidence intervals estimated
#' with N \code{bootstraps}.
#' @param bootstraps Integer. If \code{confidence=TRUE}, how many bootstraps?
#' The more iterations we run the more precise the confidence internal will be.
#' @param symmetric Boolean. \code{x2y} metric is not symmetric with respect to
#' \code{x} and \code{y}. The extent to which \code{x} can predict \code{y} can
#' be different from the extent to which \code{y} can predict \code{x}. Set
#' \code{symmetric=TRUE} if you wish to average both numbers.
#' @param target_x,target_y Boolean. Force target features to be part of
#' \code{x} OR \code{y}?
#' @param plot Boolean. Return a plot? If not, only a data.frame with calculated
#' results will be returned.
#' @param top Integer. Show/plot only top N predictive cross-features. Set
#' to \code{NULL} to return all.
#' @param quiet Boolean. Keep quiet? If not, show progress bar.
#' @param ohse Boolean. Use \code{lares::ohse()} to pre-process the data?
#' @param corr Boolean. Add correlation and pvalue data to compare with? For
#' more custom studies, use \code{lares::corr_cross()} directly.
#' @param ... Additional parameters passed to \code{x2y_metric()}
#' @return Depending on \code{plot} input, a plot or a data.frame with x2y results.
#' @examples
#' \donttest{
#' data(dft) # Titanic dataset
#' x2y_results <- x2y(dft, quiet = TRUE, max_cat = 10, top = NULL)
#' head(x2y_results, 10)
#' plot(x2y_results, type = 2)
#'
#' # Confidence intervals with 10 bootstrap iterations
#' x2y(dft,
#'   target = c("Survived", "Age"),
#'   confidence = TRUE, bootstraps = 10, top = 8
#' )
#'
#' # Compare with mean absolute correlations
#' x2y(dft, "Fare", corr = TRUE, top = 6, target_x = TRUE)
#'
#' # Plot (symmetric) results
#' symm <- x2y(dft, target = "Survived", symmetric = TRUE)
#' plot(symm, type = 1)
#'
#' # Symmetry: x2y vs y2x
#' on.exit(set.seed(42))
#' x <- seq(-1, 1, 0.01)
#' y <- sqrt(1 - x^2) + rnorm(length(x), mean = 0, sd = 0.05)
#'
#' # Knowing x reduces the uncertainty about the value of y a lot more than
#' # knowing y reduces the uncertainty about the value of x. Note correlation.
#' plot(x2y_preds(x, y), corr = TRUE)
#' plot(x2y_preds(y, x), corr = TRUE)
#' }
#' @export
x2y <- function(df, target = NULL, symmetric = FALSE,
                target_x = FALSE, target_y = FALSE,
                plot = FALSE, top = 20, quiet = "auto",
                ohse = FALSE, corr = FALSE, ...) {
  if (target_x && target_y) {
    stop("Set to TRUE only target_x OR target_y, not both.")
  }
  if ((target_x || target_y) && symmetric) {
    target_x <- target_y <- FALSE
    warning("When symmetric, target_x NOR target_y will affect results.")
  }

  df <- select(df, -zerovar(df))
  if (ohse) df <- lares::ohse(df, ...)
  pairs <- combn(ncol(df), 2)
  pairs <- cbind(pairs, pairs[2:1, ])

  if (!is.null(target)) {
    check_opts(target, colnames(df))
    cols <- which(colnames(df) %in% target)
    with_cols <- apply(pairs, 2, function(x) any(cols %in% x))
    pairs <- pairs[, with_cols]
  }

  n <- dim(pairs)[2]

  confidence <- isTRUE(list(...)[["confidence"]])
  bootstraps <- list(...)[["bootstraps"]]
  bootstraps <- ifelse(confidence & !is.null(bootstraps), bootstraps, 0)
  results <- data.frame(x = names(df)[pairs[1, ]], y = names(df)[pairs[2, ]])
  x2ys <- data.frame()

  # Show progress bar when quiet = "auto" and lots of data
  if (!is.logical(quiet)) {
    quiet <- n * nrow(df) < 5e4 & bootstraps < 50
  }

  for (i in 1:n) {
    x <- pull(df, pairs[1, i])
    y <- pull(df, pairs[2, i])
    r <- x2y_metric(x, y, ...)
    x2ys <- rbind(x2ys, r)
    if (!quiet) statusbar(i, n)
  }

  results <- cbind(results, x2ys) %>%
    arrange(desc(.data$x2y), desc(.data$obs_p)) %>%
    as_tibble()

  if (symmetric) {
    results <- results %>%
      mutate(xy = apply(results[, 1:2], 1, function(x) {
        paste(sort(x), collapse = "<>")
      })) %>%
      group_by(.data$xy) %>%
      summarise_if(is.numeric, function(x) mean(x, na.rm = TRUE)) %>%
      arrange(desc(.data$x2y), desc(.data$obs_p)) %>%
      tidyr::separate(.data$xy, c("x", "y"), sep = "<>")
  }

  if (corr) results <- .x2y_addcorr(results, df)

  if (!is.null(target)) {
    if (target_x) results <- filter(results, .data$x %in% target)
    if (target_y) results <- filter(results, .data$y %in% target)
  }

  if (!is.null(top)) results <- head(results, top)
  if (confidence) attr(results, "bootstraps") <- attr(r, "bootstraps")
  attr(results, "symmetric") <- symmetric
  class(results) <- c("x2y", class(results))
  if (plot) {
    plot(results)
  } else {
    results
  }
}

.x2y_addcorr <- function(x2y, df) {
  corr_df <- corr_cross(df, plot = FALSE, quiet = TRUE) %>%
    group_by(.data$group1, .data$group2) %>%
    summarise(
      mean_abs_corr = mean(abs(.data$corr), na.rm = TRUE),
      mean_pvalue = mean(.data$pvalue, na.rm = TRUE),
      .groups = "drop"
    ) %>%
    arrange(desc(abs(.data$mean_abs_corr)))
  results <- x2y %>%
    left_join(corr_df, by = c("x" = "group1", "y" = "group2")) %>%
    left_join(corr_df, by = c("x" = "group2", "y" = "group1")) %>%
    tidyr::unite("mean_abs_corr", contains("mean_abs_corr"), na.rm = TRUE, remove = TRUE) %>%
    tidyr::unite("mean_pvalue", contains("mean_pvalue"), na.rm = TRUE, remove = TRUE) %>%
    mutate(
      mean_abs_corr = as.numeric(eval(.data$mean_abs_corr)),
      mean_pvalue = as.numeric(eval(.data$mean_pvalue))
    )
  results
}


#' @rdname x2y
#' @param x,y Vectors. Categorical or numerical vectors of same length.
#' @param max_cat Integer. Maximum number of unique \code{x} or \code{y} values
#' when categorical. Will select then most frequent values and the rest will
#' be passed as \code{""}.
#' @export
x2y_metric <- function(x, y, confidence = FALSE, bootstraps = 20, max_cat = 20) {
  results <- list()
  missing <- is.na(x) | is.na(y)
  results$obs_p <- round(100 * (1 - sum(missing) / length(x)), 2)
  x <- x[!missing]
  y <- y[!missing]
  results$x2y <- .x2y_vals(x, y, max_cat)

  if (confidence) {
    results$lower_ci <- NA
    results$upper_ci <- NA
    if (!is.na(results$x2y) && results$x2y > 0) {
      n <- length(x)
      draws <- replicate(bootstraps, .simple_boot(x, y))
      errors <- draws - results$x2y
      results$lower_ci <- results$x2y - round(
        quantile(errors, probs = 0.975, na.rm = TRUE), 2
      )
      results$upper_ci <- results$x2y - round(
        quantile(errors, probs = 0.025, na.rm = TRUE), 2
      )
    }
    attr(results, "bootstraps") <- bootstraps
  }
  class(results) <- c("x2y_metric", class(results))
  results
}

#' @rdname x2y
#' @export
plot.x2y_preds <- function(x, corr = FALSE, ...) {
  if (!inherits(x, "x2y_preds")) stop("Object must be class x2y_preds")
  p <- ggplot(x, aes(x = .data$x)) +
    geom_point(aes(y = .data$y), size = 0.5) +
    geom_line(aes(y = .data$p),
      colour = names(lares_pal()[[2]])[2],
      size = 0.8, alpha = 0.7
    ) +
    scale_color_brewer(name = NULL) +
    labs(
      title = "x's predictive power over y",
      subtitle = sprintf("x2y: %s", x2y_metric(x$x, x$y)$x2y)
    ) +
    theme_lares()
  if (corr && is.numeric(x$x) && is.numeric(x$y)) {
    p <- p + labs(caption = paste("Correlation:", signif(cor(x$x, x$y), 1))) +
      geom_smooth(aes(y = .data$y), method = "lm", formula = "y ~ x", size = 0.5)
  }
  p
}

#' @rdname x2y
#' @param type Integer. Plot type: \code{1} for tile plot,
#' \code{2} for ranked bar plot.
#' @aliases x2y
#' @export
plot.x2y <- function(x, type = 1, ...) {
  if (!inherits(x, "x2y")) stop("Object must be class x2y")
  check_opts(type, 1:2)
  if (nrow(x) > 0) {
    x$var <- paste(x$x, ifelse(isTRUE(attr(x, "symmetric")), "><", ">>"), x$y)
    x$x2y <- signif(x$x2y, 3)
    if (type == 1) {
      p <- x %>%
        filter(.data$x2y > 0) %>%
        ggplot(aes(y = reorder(.data$var, .data$x2y), x = .data$x2y / 100)) +
        geom_col(colour = "transparent") +
        geom_text(
          aes(
            label = sub("^(-)?0[.]", "\\1.", signif(.data$x2y, 3)),
            hjust = ifelse(.data$x2y > max(.data$x2y) / 5, 1.1, -0.1)
          ),
          size = 3
        ) +
        scale_x_percent(expand = c(0, 0), position = "top") +
        labs(
          title = "Ranked Predictive Power of Cross-Features (x2y)",
          subtitle = ifelse(
            isTRUE(attr(x, "symmetric")),
            "Symmetric results: mean(x2y, y2x)",
            "Non-symmetric results: x2y != y2x"
          ),
          x = NULL, y = NULL
        ) +
        theme_lares()
    }
    if (type == 2) {
      p <- x %>%
        filter(.data$x2y > 0) %>%
        group_by(.data$x) %>%
        mutate(ximp = max(.data$x2y, na.rm = TRUE)) %>%
        ungroup() %>%
        group_by(.data$y) %>%
        mutate(yimp = max(.data$x2y, na.rm = TRUE)) %>%
        ungroup() %>%
        ggplot(aes(
          x = reorder(.data$x, -.data$ximp),
          y = reorder(.data$y, -.data$yimp),
          fill = .data$x2y, label = .data$x2y
        )) +
        geom_tile() +
        geom_text(size = 3.2) +
        scale_fill_gradient(low = "white", high = names(lares::lares_pal()[[2]])[1]) +
        labs(
          title = "Cross-features Predictive Power (x2y)",
          x = NULL, y = NULL,
          subtitle = ifelse(
            isTRUE(attr(x, "symmetric")),
            "Symmetric results: mean(x2y, y2x)",
            "Non-symmetric results: x2y != y2x"
          )
        ) +
        theme_lares(grid = "") +
        theme(axis.text.x = element_text(angle = 45, vjust = 0.5, hjust = 1)) +
        coord_equal()
    }
    p
  }
}

#' @rdname x2y
#' @export
x2y_preds <- function(x, y, max_cat = 10) {
  # If no variance
  if (length(unique(x)) == 1 || length(unique(y)) == 1) {
    NA
  } else {
    # If x is categorical
    x <- .reduce_cats(x, max_cat)
    # If y is continuous
    if (is.numeric(y)) {
      preds <- predict(rpart(y ~ x, method = "anova"), type = "vector")
    } else {
      # If y is categorical
      y <- .reduce_cats(y, max_cat)
      preds <- predict(rpart(y ~ x, method = "class"), type = "class")
    }
    preds <- as_tibble(data.frame(x = x, y = y)) %>%
      removenarows(all = FALSE) %>%
      mutate(p = preds)
    attr(preds, "max_cat") <- max_cat
    class(preds) <- c("x2y_preds", class(preds))
    preds
  }
}

.x2y_vals <- function(x, y, ...) {
  if (length(unique(x)) == 1 || length(unique(y)) == 1) {
    NA
  } else {
    preds <- x2y_preds(x, y, ...)$p
    if (is.numeric(y)) {
      .mae_reduction(preds, y)
    } else {
      .misclass_reduction(preds, y)
    }
  }
}

.mae_reduction <- function(y_hat, y_actual) {
  model_error <- mean(abs(y_hat - y_actual))
  baseline <- mean(y_actual, na.rm = TRUE)
  baseline_error <- mean(abs(baseline - y_actual))
  result <- 1 - model_error / baseline_error
  result <- max(0.0, min(result, 1.0))
  round(100 * result, 2)
}

.misclass_reduction <- function(y_hat, y_actual) {
  tab <- table(y_hat, y_actual)
  model_error <- 1 - sum(diag(tab)) / sum(tab)
  majority_class <- names(which.max(table(y_actual)))
  baseline.preds <- rep(majority_class, length(y_actual))
  baseline_error <- mean(baseline.preds != y_actual)
  result <- 1 - model_error / baseline_error
  result <- max(0.0, min(result, 1.0))
  round(100 * result, 2)
}

.simple_boot <- function(x, y, ...) {
  ids <- sample(length(x), replace = TRUE)
  .x2y_vals(x[ids], y[ids], ...)
}

.reduce_cats <- function(x, max_cat) {
  if (!is.numeric(x) && length(unique(x)) > max_cat) {
    x <- as.character(x)
    top <- head(names(sort(table(x), decreasing = TRUE)), max_cat)
    x[!x %in% top] <- ""
  }
  x
}

####################################################################
#' Titanic Dataset
#'
#' The sinking of the Titanic is one of the most infamous shipwrecks in history.
#' On April 15, 1912, during her maiden voyage, the widely considered "unsinkable"
#' RMS Titanic sank after colliding with an iceberg. Unfortunately, there weren’t
#' enough lifeboats for everyone onboard, resulting in the death of 1502 out of
#' 2224 passengers and crew. While there was some element of luck involved in
#' surviving, it seems some groups of people were more likely to survive than others.
#' This dataset contains the details of a subset of the passengers on board
#' (891 to be exact) taken from Kaggle's Titanic
#' \href{https://www.kaggle.com/c/titanic/overview}{Train.csv}.
#'
#' @family Dataset
#' @docType data
#' @usage data(dft)
#' @return data.frame
#' @format An object of class \code{"data.frame"}
#' \describe{
#'   \item{PassengerId}{Unique ID for each passenger (1-891)}
#'   \item{Survived}{Did the passenger survive? (TRUE, FALSE)}
#'   \item{Pclass}{Ticket class, from first to third (1, 2, 3)}
#'   \item{Sex}{Gender (female, male)}
#'   \item{Age}{Age for each passenger in years (0.42-80)}
#'   \item{SibSp}{Amount of siblings / spouses aboard the Titanic (0-8)}
#'   \item{Parch}{Amount of parents / children aboard the Titanic (0-6)}
#'   \item{Ticket}{Ticket IDs}
#'   \item{Fare}{Amount paid for passenger's ticket (0-512.3292)}
#'   \item{Cabin}{width of top of diamond relative to widest point (43-95)}
#'   \item{Embarked}{Port of Embarkation (43-95)}
#' }
#' @examples
#' data(dft)
#' head(dft)
"dft"

# dft <- read.csv("~/Dropbox (Personal)/Documentos/R/R Titanic/train.csv") %>%
#   mutate(Pclass = factor(Pclass),
#          Sex = factor(Sex),
#          Survived = as.logical(Survived),
#          Ticket = as.character(Ticket)) %>%
#   select(-Name)
# save(dft, file = "data/dft.RData")


####################################################################
#' Results for AutoML Predictions
#'
#' List with categorical (2 and 3 classes) and continuous predictions,
#' generated with \code{h2o_automl()} and the \code{dft}. Note that
#' the models per se won't work to predict.
#'
#' @family Dataset
#' @docType data
#' @usage data(dfr)
#' @return List
#' @format An object of class \code{"list"} with 3 \code{"data.frame"}
#' \describe{
#'   \item{class2}{Predictions for a Binomial Classification Model}
#'   \item{class3}{Predictions for a Multi-Categorical Classification Model}
#'   \item{regr}{Predictions for a Continuous Regression Model}
#' }
#' @examples
#' data(dfr)
#' lapply(dfr, head)
"dfr"

# data(dft)
# df <- subset(dft, select = -c(Ticket, PassengerId, Cabin))
# # Classification: 2 classes
# class2 <- h2o_automl(df, y = Survived, max_models = 1, target = "TRUE", start_clean = TRUE)
# # Classification: 3 classes
# class3 <- h2o_automl(df, Pclass, ignore = "Fare", impute = TRUE, start_clean = FALSE)
# # Regression (Continuous Values)
# regr <- h2o_automl(df, y = "Fare", ignore = c("Pclass","Cabin"), exclude_algos = NULL, start_clean = FALSE)
# # Save results
# dfr <- list(class2 = class2$scores_test,
#             class3 = class3$scores_test,
#             regr = regr$scores_test)
# save(dfr, file = "data/dfr.RData", version = 2)

####################################################################
#' Load Credentials from a YML File
#'
#' Load credentials from a local YML file. You can set your \code{.Renviron}
#' and the \code{LARES_CREDS} parameter to remember (forever) the directory
#' of your credentials' file. To use it later, you may leave \code{dir = NA}.
#' You may also use this function for external (non-\code{lares}) code/use.
#'
#' @section Set the default directory:
#' The first time you use any function that has the \code{creds} parameter, if
#' the \code{dir} parameter is set to \code{NA}, this function will ask you to
#' set the directory where you save your YML local file with your credentials.
#' This will be asked once and will be set for further R sessions. Remember to
#' reset your session for this setup to start working properly.
#'
#' @section YML file format:
#' A YML file is a text file, with \code{.yml} file format. You may start from
#' the dummy YML file shared which shows the structure you must follow to set your
#' credentials file. Check it out
#' \href{https://raw.githubusercontent.com/laresbernardo/lares/master/inst/docs/config.yml}{here}
#' or find it locally using \code{system.file("docs", "config.yml", package = "lares")}.
#'
#' @family Credentials
#' @inheritParams cache_write
#' @param from Character. Family of values to import from the YML file.
#' If you don't know these names, set \code{from = NA}
#' and a warning will display all possible values, depending on your YML file.
#' @param dir Character. Credentials directory where your YML file is.
#' If used frequently, set your directory by using the \code{.Renviron} file.
#' To do so, leave \code{dir} as \code{NA} and follow the steps.
#' If \code{dir} is a list, it'll return \code{dir} (manual credentials input).
#' If \code{dir} ends with \code{.yml} or \code{.yaml}, \code{filename} will be ignored.
#' @param filename Character. YML filename with your credentials.
#' Ignored if \code{dir} contains the file path.
#' @param env Character. Environment variable name. No need to set differently
#' for any function that uses this library. Only for external use.
#' @return List. Result of reading your credential's YML file, filtered by your
#' \code{from} input if provided.
#' @examples
#' \dontrun{
#' # Load dummy config.yml file from the library
#' # Recommendation: set dir with NA (read documentation)
#' # We need the directory, not the file
#' yml <- dirname(system.file("docs", "config.yml", package = "lares"))
#'
#' # Let's see which credentials we have in our file
#' get_credentials(dir = yml)
#' # Warning message: No credentials for NA found in your YML file.
#' # Try any of the following: 'service1', 'service2', 'service3'
#'
#' # Get credentials for service2
#' get_credentials("service2", dir = yml)
#' }
#' @export
get_credentials <- function(from = NA, dir = NA,
                            filename = "config.yml",
                            env = "LARES_CREDS",
                            ...) {
  if (is.list(dir)) {
    dir
  } else {
    if (isTRUE(is.na(dir)) || isTRUE(is.null(dir))) {
      dir <- Sys.getenv(env)
    }
    if (dir == "") {
      message(sprintf(
        "Please, set your creds directory (one-time only step to set %s):", env
      ))
      dir <- readline(sprintf("Set directory where your %s file is saved: ", filename))
      if (!dir.exists(dir)) {
        stop(sprintf("Directory: %s. \nCan't find or does not exist. Please try again...", dir))
      }
      line <- sprintf("%s=%s", env, dir)
      write(line, file = "~/.Renviron", append = TRUE)
      message("ALL's SET! But, you must reset your session for it to work!")
      invisible(NULL)
    } else {
      # Handle .yml file in dir parameter
      if (grepl("\\.ya?ml$", dir)) {
        file <- dir
      } else {
        # Normalize paths to handle slash inconsistencies
        dir <- gsub("/+$", "", dir)
        filename <- gsub("^/+", "", filename)
        file <- file.path(dir, filename)
      }

      if (!file.exists(file)) {
        message(sprintf("YML file with credentials not found in %s", dir))
      } else {
        credentials <- yaml::read_yaml(file)$default
        if (is.null(credentials)) {
          trues <- names(credentials)
          warning(paste(
            "No credentials for", from, "found in your YML file.",
            "\nTry any of the following:", v2t(trues)
          ))
        }
        if (!is.na(from)) credentials <- credentials[[from]]
        credentials
      }
    }
  }
}

#' @rdname get_credentials
#' @export
get_creds <- get_credentials

####################################################################
#' Download/Import Dropbox File by File's Name
#'
#' This function lets the user download a file from Dropbox, specifying
#' its name, using a previously created token or with interactive window.
#'
#' @family Tools
#' @family Credentials
#' @family Dropbox
#' @inheritParams mp3_get
#' @param query Search string. This string is split (on spaces) into
#' individual words. Files will be used if they contain
#' all words in the search string.
#' @param local_path Character. Path to save file to. If \code{NULL}
#' (the default), saves file to working directory with same name.
#' If not, but a valid folder, file will be saved in this folder with
#' same basename as path. If not \code{NULL} and not a folder,
#' file will be saved to this path exactly.
#' @param xlsx Boolean. Is it an Excel file? Can be returned as a list
#' for each tab and not as a file if needed. Will delete downloaded file.
#' @param token_dir Character. RDS with token local directory. You may set to
#' NA if you already set your credentials (see \code{get_creds()})
#' @param token_name Character. RDS file name with your token's data.
#' @return If \code{query} returns a .xlsx file and \code{xlsx=TRUE}, will
#' return a data.frame. Else, \code{local_path} string.
#' @examples
#' \dontrun{
#' # Download a specific file
#' db_download("stocksReport.Rmd", local_path = "~/Desktop/generic.Rmd")
#' # Import an Excel file from Dropbox into a data.frame
#' df <- db_download("Portfolio LC.xlsx", xlsx = FALSE)
#' }
#' @export
db_download <- function(query,
                        local_path = NULL,
                        xlsx = TRUE,
                        token_dir = NA,
                        token_name = "token_pers.rds",
                        quiet = FALSE) {
  try_require("rdrop2")

  if (is.na(token_dir)) {
    load(paste0(Sys.getenv("LARES_CREDS"), "/", token_name))
  } else {
    token_file <- paste0(token_dir, "/", token_name)
    if (file.exists(token_dir)) {
      load(token_file)
    } else {
      token <- drop_auth()
    }
  }

  x <- drop_search(query, dtoken = token)
  if (length(x$matches) == 0) {
    stop("There were no files found in your Dropbox with: ", query)
  }
  if (length(x$matches) > 1 && !quiet) {
    warning(
      "There were multiple files found in your Dropbox with: ", query,
      "\nUsing: ", x$matches[[1]]$metadata$name
    )
  }

  if (is.null(local_path)) {
    local_path <- x$matches[[1]]$metadata$name
  }

  invisible(
    drop_download(x$matches[[1]]$metadata$path_lower,
      local_path = local_path,
      overwrite = TRUE,
      dtoken = token,
      verbose = FALSE
    )
  )
  if (!quiet) message(paste("> File", x$matches[[1]]$metadata$name, "downloaded succesfully!"))

  if (xlsx && right(local_path, 5) == ".xlsx") {
    results <- importxlsx(local_path)
    invisible(file.remove(local_path))
    if (!quiet) message("> File imported succesfully as an object!")
    results
  } else {
    invisible(local_path)
  }
}

####################################################################
#' Upload Local Files to Dropbox
#'
#' This function lets the user upload a local file to Dropbox,
#' using a previously created token or with interactive window.
#'
#' @family Tools
#' @family Credentials
#' @family Dropbox
#' @inheritParams db_download
#' @param filename String. Local file's name to upload.
#' @param dir String. Directory you wish to upload the file to.
#' @param delete_file Boolean. Delete local file after uploading?
#' @return \code{TRUE} when successfully uploads file.
#' @export
db_upload <- function(filename, dir, delete_file = FALSE,
                      token_dir = NA,
                      token_name = "token_pers.rds") {
  try_require("rdrop2")
  if (is.na(token_dir)) {
    load(paste0(Sys.getenv("LARES_CREDS"), "/", token_name))
  } else {
    token_file <- paste0(token_dir, "/", token_name)
    if (file.exists(token_dir)) {
      load(token_file)
    } else {
      token <- drop_auth()
    }
  }
  invisible(drop_upload(filename, path = dir, dtoken = token))
  if (delete_file) file.remove(filename)
  invisible(TRUE)
}

####################################################################
#' Convert Date into Year-Month, Year-Quarter or Year-Week Format
#'
#' This function lets the user convert a date into YYYY-MM, YYYY-QX,
#' or YYYY-WW format easily.
#'
#' @family Data Wrangling
#' @param date Date vector. Date to transform format.
#'
#' @return Vector with dates reformatted
#' @examples
#' year_month(Sys.Date())
#' year_quarter(Sys.Date())
#' year_week(Sys.Date())
#' @export
year_month <- function(date) {
  paste(year(date), str_pad(lubridate::month(date), 2, pad = "0"), sep = "-")
}
#' @rdname year_month
#' @export
year_quarter <- function(date) {
  paste(year(date), date_cuts(date, type = "Q"), sep = "-")
}
#' @rdname year_month
#' @export
year_week <- function(date) {
  paste(year(date), str_pad(lubridate::week(date), 2, pad = "0"), sep = "-")
}


####################################################################
#' Convert Date into Year + Cut
#'
#' This function returns categorical values for any date(s) using year
#' cuts such as bimonths, quarters, terms, and halves.
#'
#' @family Data Wrangling
#' @param date Date. Date we wish to transform
#' @param type Character. Any of the following: B (2 months),
#' Q (3 months), T (4 months), H (6 months)
#' @return Vector with date cut for each \code{date}
#' @examples
#' date_cuts(Sys.Date(), type = "Q")
#' date_cuts(Sys.Date(), type = "H")
#' @export
date_cuts <- function(date = Sys.Date(), type = "Q") {
  df <- data.frame(date = as.Date(as.character(date))) %>%
    mutate(month = month(.data$date), cut = NA)
  if (tolower(type) == "b") aux <- 2
  if (tolower(type) == "q") aux <- 3
  if (tolower(type) == "t") aux <- 4
  if (tolower(type) == "h") aux <- 6
  cuts <- c(seq(1, 12, aux), 12)
  for (i in 1:(12 / aux)) {
    df <- df %>% mutate(cut = ifelse(
      .data$month >= cuts[i] & .data$month <= cuts[i + 1],
      paste0(toupper(type), i), .data$cut
    ))
  }
  df$cut
}


####################################################################
#' Left or Right N characters of a string
#'
#' This functions lets the user extract the first or last n characters
#' of a string or vector of strings.
#'
#' @family Data Wrangling
#' @param string String or Vector.
#' @param n Integer. How many characters starting on right/left?
#' @return Character. Trimmed strings.
#' @examples
#' left("Bernardo", 3)
#' right(c("Bernardo", "Lares", "V"), 3)
#' @export
#' @rdname left_right
left <- function(string, n = 1) {
  string <- as.character(string)
  substr(string, 1, n)
}

#' @rdname left_right
#' @export
right <- function(string, n = 1) {
  string <- as.character(string)
  substr(string, nchar(string) - n + 1, nchar(string))
}


####################################################################
#' Zero Variance Columns
#'
#' This function detects which columns have the same value (whichever)
#' for each column.
#'
#' @family Data Wrangling
#' @param df Dataframe
#' @return Character vector with column names on which its values have no variance.
#' @examples
#' df <- data.frame(a = c(1, NA, 3), b = rep(NA, 3), c = rep(5, 3))
#' print(df)
#' zerovar(df)
#' @export
zerovar <- function(df) {
  out <- lapply(df, function(x) length(unique(x)))
  which <- which(out <= 1)
  names(unlist(which))
}


####################################################################
#' Normalize Vector
#'
#' This function normalizes numerical values into a specified range,
#' defaulting to the 0 to 1 range.
#'
#' @inheritParams ohse
#' @family Data Wrangling
#' @param x Numeric vector. The numbers to be normalized.
#' @param range A numeric vector of length 2 specifying the desired range
#' for normalization. Default is c(0, 1).
#' @return A numeric vector with normalized \code{x} values.
#' @examples
#' x <- c(0, 1, 4, 7.5, 10)
#' normalize(x)
#' normalize(x, range = c(-1, 1))
#' @export
normalize <- function(x, range = c(0, 1), ...) {
  if (!is.numeric(x)) {
    stop("Input must be a numeric vector!")
  }
  if (length(unique(x)) == 1) {
    rep(range[2], length(x))
  } else {
    if (!is.numeric(range) || length(range) != 2 || range[1] >= range[2]) {
      stop("Parameter 'range' must be a numeric vector of length 2 with range[1] < range[2].")
    }
    scaled_x <- (x - min(x, ...)) / (max(x, ...) - min(x, ...))
    normalized_x <- scaled_x * (range[2] - range[1]) + range[1]
    normalized_x
  }
}


####################################################################
#' Abbreviate or dis-abbreviate numerical values
#'
#' This function converts a numeric vector's values into their
#' abbreviated character equivalent, i.e. 100E6 into 100M
#' and viceversa.
#'
#' @family Data Wrangling
#' @inheritParams cache_write
#' @param x Numeric vector
#' @param n Integer. Single numeric value, specifying number of
#' significant figures to show. Range 1 to 6.
#' @param numeric Boolean. Transform abbreviated number into numeric?
#' @return Vector of character or numeric values that contain converted values
#' @examples
#' num_abbr(rnorm(10) * 1e6)
#' num_abbr(rnorm(10) * 1e6, n = 1)
#' num_abbr(c("3K", "-58.3M", NA, 1), numeric = TRUE)
#' @export
num_abbr <- function(x, n = 3, numeric = FALSE, ...) {
  if (is.null(x)) {
    x
  } else {
    if (!is.numeric(x)) {
      if (!numeric) {
        stop("Input vector 'x' needs to be numeric. To convert abbr to num, set numeric = TRUE")
      }
    }
    if (!is.numeric(n)) stop("Input 'n' needs to be numeric.")
    if (length(n) > 1) stop("Please make sure that n takes on a single value.")
    if (!n %in% 1:6) stop("Please make sure that n takes on an interger value between 1 to 6.")

    if (numeric) {
      num <- as.numeric(gsub("[A-z]+", "", x))
      abbr_value <- unlist(lapply(seq_along(x), function(i) {
        abbr <- gsub(as.character(num[i]), "", x[i])
        av <- dplyr::case_when(
          abbr == "" ~ 1,
          abbr == "K" ~ 1E3,
          abbr == "M" ~ 1E6,
          abbr == "B" ~ 1E9,
          abbr == "T" ~ 1E12,
          abbr == "Qa" ~ 1E15,
          abbr == "Qi" ~ 1E18,
          TRUE ~ NA
        )
        as.numeric(av)
      }))
      output <- num * abbr_value
      output
    } else {
      # # To handle scientific notation inputs correctly
      # on.exit(options("scipen" = getOption('scipen')))
      # options("scipen" = 999)
      # Clean up x
      negative_positions <- ifelse(x < 0, "-", "")
      x <- abs(x)
      div <- findInterval(x, c(0, 1e3, 1e6, 1e9, 1e12, 1e15, 1e18))
      # Round x with some cleaning
      x <- round(x, -nchar(round(x, 0)) + n) / 10^(3 * (div - 1))
      # Fix numbers rounded up to another digit
      # i.e. 999k -> 1000k should actually be 1M
      div <- ifelse(nchar(as.integer(x)) > 3, div + 1, div)
      x <- ifelse(nchar(as.integer(x)) > 3, x / 1e3, x)
      # Cap decimal places to 3
      x <- round(x, 3)
      # Qa = Quadrillion; Qi = Quintillion
      x <- paste0(x, c("", "K", "M", "B", "T", "Qa", "Qi")[div])
      output <- paste0(negative_positions, x)
      output[grepl("NA", output)] <- NA
      output
    }
  }
}


####################################################################
#' Reduce categorical values
#'
#' This function lets the user reduce categorical values in a vector.
#' It is tidyverse friendly for use on pipelines
#'
#' @family Data Wrangling
#' @inheritParams ohse
#' @inheritParams cache_write
#' @param df Categorical Vector
#' @param var Variable. Which variable do you wish to reduce?
#' @param nmin Integer. Number of minimum times a value is repeated
#' @param pmin Numerical. Percentage of minimum times a value is repeated
#' @param pcummax Numerical. Top cumulative percentage of most
#' repeated values
#' @param top Integer. Keep the n most frequently repeated values
#' @param pvalue_max Numeric (0-1]. Max pvalue categories
#' @param cor_var Character. If pvalue_max < 1, you must define which
#' column name will be compared with (numerical or binary).
#' @param other_label Character. With which text do you wish to replace
#' the filtered values with?
#' @return data.frame \code{df} on which \code{var} has been transformed
#' @examples
#' data(dft) # Titanic dataset
#' categ_reducer(dft, Embarked, top = 2) %>% freqs(Embarked)
#' categ_reducer(dft, Ticket, nmin = 7, other_label = "Other Ticket") %>% freqs(Ticket)
#' categ_reducer(dft, Ticket, pvalue_max = 0.05, cor_var = "Survived") %>% freqs(Ticket)
#' @export
categ_reducer <- function(df, var,
                          nmin = 0,
                          pmin = 0,
                          pcummax = 100,
                          top = NA,
                          pvalue_max = 1,
                          cor_var = "tag",
                          limit = 20,
                          other_label = "other",
                          ...) {
  var <- enquo(var)

  dff <- freqs(df, !!var)

  if (!is.na(top)) {
    tops <- dff %>% slice(1:top)
  } else {
    tops <- filter(dff, .data$n >= nmin & .data$p >= pmin & .data$p <= pcummax)
  }

  name <- as.name(names(dff[, 1]))
  vector <- df %>% select(as.character(name))
  new_vector <- ifelse(vector[[as.character(name)]] %in% tops[[as.character(name)]],
    as.character(vector[[as.character(name)]]),
    other_label
  )

  if (pvalue_max < 1) {
    best_vars <- df %>%
      mutate(cor_var_temp = unlist(df[, cor_var])) %>%
      select(ncol(.), !!var) %>%
      corr_var(.data$cor_var_temp,
        plot = FALSE,
        limit = limit, quiet = TRUE, ...
      ) %>%
      filter(.data$pvalue < pvalue_max) %>%
      rowwise() %>%
      mutate(variables = gsub(paste0(name, "_"), "", .data$variables)) %>%
      pull(.data$variables)
    new_vector[!new_vector %in% best_vars] <- other_label
  }

  df[[as.character(name)]] <- new_vector
  df
}


####################################################################
#' Convert a vector into a comma separated text
#'
#' Convert a vector into a comma separated text
#'
#' @family Data Wrangling
#' @param vector Vector. Vector with more than 1 observation.
#' @param sep Character. String text wished to insert between values.
#' @param quotes Boolean. Bring simple quotes for each observation.
#' @param force_single Boolean. Force single quotes by replacing \code{\"}.
#' @param and Character. Add 'and' or something before last observation.
#' Not boolean variable so it can be used on other languages. Note that
#' the last comma will be suppressed if \code{Sys.getenv("LARES_NUMFORMAT")}
#' is set to \code{1} and you have less than 3 values.
#' @return Vector pasting \code{vector} values into a single string
#' @examples
#' vector2text(LETTERS[1:5])
#' vector2text(c(1:5), quotes = FALSE)
#' vector2text(c(1:5), quotes = FALSE, sep = "-")
#' vector2text(c(1:5), and = "and also")
#' vector2text(c("Text", "R's"), force_single = TRUE)
#' # Shorter function with same purpose
#' v2t(LETTERS[1:5])
#' @export
vector2text <- function(vector, sep = ", ", quotes = TRUE, force_single = FALSE, and = "") {
  # Add "and" or something before the last value
  n <- length(vector)
  if (and != "") {
    # Makes no sense to keep quotes but leave the option
    quotes <- !quotes
    if (n > 1) {
      vector <- c(vector[1:(n - 1)], paste(and, vector[n]))
    }
  }

  # Paste everythign together
  output <- paste(shQuote(vector), collapse = sep)

  # Get rid of quotes
  if (!quotes) {
    output <- gsub("'", "", output)
  }

  # Get rid of the last comma when using and?
  if (and != "" && (Sys.getenv("LARES_NUMFORMAT") == 1 || n == 2)) {
    last_comma <- tail(c(gregexpr(",", output)[[1]]), 1)
    output <- paste0(
      substr(output, 1, last_comma - 1),
      substr(output, last_comma + 1, nchar(output))
    )
  }

  if (force_single) {
    output <- gsub('"', "'", output)
    output <- gsub('\"', "'", output)
  }

  output
}
#' @rdname vector2text
#' @export
v2t <- vector2text


####################################################################
#' Nicely Format Texts and Numericals
#'
#' This function lets the user format numerical values nicely
#'
#' @family Data Wrangling
#' @param x Numerical Vector
#' @param decimals Integer. Amount of decimals to display. If set to
#' \code{NULL}, then \code{getOption("digits")} will be used.
#' @param signif Integer. Rounds the values in its first argument to
#' the specified number of significant digits.
#' @param type Integer. \code{1} for International standards. \code{2}
#' for American Standards. Use \code{Sys.setenv("LARES_NUMFORMAT" = 2)}
#' to set this parameter globally.
#' @param pre,pos Character. Add string before or after number.
#' @param sign Boolean. Add \code{+} sign to positive values.
#' @param abbr Boolean. Abbreviate using num_abbr()? You can use
#' the `decimals` parameter to set abbr's \code{n}(-1) parameter.
#' @param ... Additional lazy eval parameters.
#' @return Character. String vector with reformatted continuous numbers
#' @examples
#' formatNum(1.23456, decimals = 3)
#' formatNum(1.23456, type = 1)
#' formatNum(1.23456, pre = "$", pos = "/person")
#' formatNum(123456, abbr = TRUE)
#' formatNum(c(123123, 123.123, 0.123123), signif = 2)
#' formatNum(1234567890, abbr = TRUE, signif = 3)
#' formatNum(1234567890, decimals = 0, abbr = TRUE)
#' formatNum(c(-3:3), sign = TRUE)
#' @export
#' @rdname format_string
formatNum <- function(x, decimals = 2, signif = NULL,
                      type = Sys.getenv("LARES_NUMFORMAT"),
                      pre = "", pos = "", sign = FALSE,
                      abbr = FALSE,
                      ...) {
  if (is.null(x) || !is.numeric(x)) {
    x
  } else {
    # Auxiliary function to save signs
    if (sign) signs <- ifelse(x > 0, "+", "")

    # Decimals: Round numbers
    if (is.null(decimals)) decimals <- getOption("digits")
    x <- base::round(x, digits = decimals)

    # Significant digits
    if (!is.null(signif)) {
      x <- lapply(x, function(y) if (is.na(y)) y else base::signif(y, signif))
    }

    if (abbr) {
      x <- unlist(lapply(x, function(y) if (is.na(y)) y else num_abbr(y, n = decimals + 1)))
    } else {
      x <- unlist(lapply(x, function(y) {
        if (is.na(y)) {
          y
        } else {
          if (type == 1) {
            y <- format(as.numeric(y), big.mark = ".", decimal.mark = ",", ...)
          } else {
            y <- format(as.numeric(y), big.mark = ",", decimal.mark = ".", ...)
          }
          trimws(y)
        }
      }))
    }

    if (pre == "$") x <- gsub("\\$-", "-$", x)
    if (sign) x <- paste0(signs, x)
    ret <- paste0(pre, x, pos)
    ret
  }
}


####################################################################
#' Balance Binary Data by Resampling: Under-Over Sampling
#'
#' This function lets the user balance a given data.frame by resampling
#' with a given relation rate and a binary feature.
#'
#' @family Data Wrangling
#' @inheritParams mp3_get
#' @param df Vector or Dataframe. Contains different variables in each
#' column, separated by a specific character
#' @param var Variable. Which variable should we used to re-sample dataset?
#' @param rate Numeric. How many X for every Y we need? Default: 1. If there are
#' more than 2 unique values, rate will represent percentage for number of rows
#' @param target Character. If binary, which value should be reduced? If kept in
#' \code{"auto"}, then the most frequent value will be reduced.
#' @param seed Numeric. Seed to replicate and obtain same values
#' @return data.frame. Reduced sampled data.frame following the \code{rate} of
#' appearance of a specific variable.
#' @examples
#' data(dft) # Titanic dataset
#' df <- balance_data(dft, Survived, rate = 1)
#' df <- balance_data(dft, .data$Survived, rate = 0.5, target = "TRUE")
#' @export
balance_data <- function(df, var, rate = 1, target = "auto", seed = 0, quiet = FALSE) {
  on.exit(set.seed(seed))
  if (isTRUE(try(is.character(var), silent = TRUE))) {
    var <- eval(substitute(var), df)
  } else {
    var <- as_label(enquo(var))
  }
  stopifnot(var %in% names(df))
  names(df)[names(df) == var] <- "tag"
  tags <- group_by(df, .data$tag) %>%
    summarize(n = n()) %>%
    arrange(desc(.data$n)) %>%
    pull(.data$tag) %>%
    as.character()

  if (length(tags) != 2) {
    # For numerical re-sampling:
    samp <- round(rate * nrow(df))
    balanced <- sample_n(df, samp)
    if (nrow(df) != samp) {
      if (!quiet) message(paste("Resampled from", nrow(df), "to", samp, "rows"))
    }
  } else {
    # For binary re-sampling:
    if (!quiet) {
      message(paste("Resampled from:", vector2text(
        formatNum(table(df$tag), 0),
        sep = " v ", quotes = FALSE
      )))
    }
    if (as.character(target) == "auto") {
      zero <- tags[1]
      one <- tags[2]
    } else {
      check_opts(target, tags)
      zero <- target
      one <- tags[tags != target]
    }
    if (!quiet) message(paste("Reducing size for label:", zero))
    ones <- filter(df, .data$tag == one)
    zeros <- filter(df, .data$tag == zero)
    zeros <- sample_n(zeros, round(rate * nrow(ones)))
    balanced <- rbind(ones, zeros)
    if (!quiet) {
      message(paste("New label distribution:", vector2text(
        formatNum(table(balanced$tag), 0),
        sep = " v ", quotes = FALSE
      )))
    }
  }
  balanced <- rename_at(balanced, vars("tag"), list(~ paste0(var)))
  as_tibble(balanced)
}


####################################################################
#' Replace Values With
#'
#' This function lets the user replace all specific values in a
#' vector or data.frame into another value. If replacing more than
#' one value, order matters so they will be replaced in the same
#' order that you pass them to the function. Factors will be refactored.
#'
#' @family Data Wrangling
#' @family Text Mining
#' @inheritParams mp3_get
#' @param df Data.frame or Vector
#' @param original String or Vector. Original text you wish to replace
#' @param change String or Vector. Values you wish to replace the originals with
#' @param which Character vector. Name of columns to use. Leave "all" for everything
#' @param fixclass Boolean. Try to detect logical classes after transformations (or
#' leave as default classes as character)?
#' @return data.frame with replaced values based on inputs.
#' @examples
#' df <- data.frame(
#'   one = c(1:4, NA),
#'   two = LETTERS[1:5],
#'   three = rep("A", 5),
#'   four = c(NA, "Aaa", 123, "B", "C")
#' )
#' print(df)
#'
#' replaceall(df, "A", NA)
#'
#' replaceall(df, "A", "a")
#'
#' replaceall(df, 1, "*")
#'
#' replaceall(df, NA, "NotNA")
#'
#' replaceall(df, NA, 0)
#'
#' replaceall(df, c("A", "B"), c("'A'", "'B'"))
#'
#' replaceall(df, "a", "*", which = "four")
#' @export
replaceall <- function(df, original, change, which = "all",
                       fixclass = TRUE, quiet = TRUE) {
  if (is.vector(df)) {
    vector <- TRUE
    df <- data.frame(x = df)
    dic <- data.frame(original, change)
    original <- dic[, 1]
    change <- dic[, 2]
  } else {
    vector <- FALSE
  }
  if (length(original) != length(change)) {
    stop("Vectors original and change should have the same length!")
  }
  if (length(unique(original)) != length(original)) {
    aux <- freqs(dic, original) %>%
      filter(.data$n > 1) %>%
      .$original
    stop("You have repeated original values to replace: ", vector2text(aux))
  }
  if (which[1] != "all") {
    aux <- df
    df <- select(df, one_of(which))
  }
  df <- df %>% mutate_all(as.character)
  if (sum(is.na(original)) > 0) {
    df <- df %>% replace(is.na(.), change[is.na(original)])
    change <- change[!is.na(original)]
    original <- original[!is.na(original)]
  }
  if (sum(is.null(original)) > 0) {
    df <- df %>% replace(is.null(.), change[is.null(original)])
    change <- change[!is.null(original)]
    original <- original[!is.null(original)]
  }
  if (length(original) > 0) {
    for (i in seq_along(original)) {
      if (!quiet) {
        message(paste("Transforming all", original[i], "into", change[i]))
      }
      df[] <- lapply(df, function(x) gsub(original[i], change[i], x))
    }
  }
  if (which[1] != "all") {
    df <- select(aux, -one_of(which)) %>%
      cbind(df) %>%
      select(one_of(colnames(aux)))
  }
  if (vector) {
    df <- df[, 1]
  }
  if (fixclass) {
    df <- suppressMessages(type.convert(df, numerals = "no.loss", as.is = TRUE))
  }
  if (vector) {
    as.vector(df)
  } else {
    as_tibble(df)
  }
}


####################################################################
#' Calculate cuts by quantiles
#'
#' This function lets the user quickly calculate cuts for quantiles
#' and discretize numerical values into categorical values.
#'
#' @family Data Wrangling
#' @family Calculus
#' @param values Vector. Values to calculate quantile cuts
#' @param splits Integer. How many cuts should split the values?
#' @param return Character. Return "summary" or "labels"
#' @param n Integer. Determines the number of digits used in
#' formatting the break numbers.
#' @return Factor vector or data.frame. Depending on \code{return} input:
#' \itemize{
#'   \item \code{labels} a factor ordered vector with each observation's quantile
#'   \item \code{summary} a data.frame with information on each quantile cut
#' }
#' @examples
#' data(dft) # Titanic dataset
#' quants(dft$Age, splits = 5, "summary")
#' quants(dft$Age, splits = 5, "labels")[1:10]
#' @export
quants <- function(values, splits = 10, return = "labels", n = 2) {
  if (splits > length(unique(values[!is.na(values)])) - 1) {
    stop("There are not enough observations to split the data in ", splits)
  }
  cuts <- quantile(values, probs = seq(0, 1, length = splits + 1), na.rm = TRUE)
  labels <- cut(values, unique(cuts), dig.lab = n, include.lowest = TRUE, ordered_result = TRUE)
  output <- labels
  if (return == "summary") {
    output <- data.frame(percentile = names(cuts)[-1], cut = cuts[-1]) %>%
      mutate(
        label = paste0("(", signif(lag(.data$cut), n), "-", signif(.data$cut, n), "]"),
        label = gsub("\\(NA", paste0("[", signif(min(.data$cut), n)), .data$label),
        label = factor(.data$label, levels = unique(.data$label), ordered = TRUE)
      )
  }
  output
}


####################################################################
#' Distance from specific point to line
#'
#' This function lets the user calculate the mathematical linear distance
#' Between a specific point and a line (given geometrical 3 points)
#'
#' @family Calculus
#' @param x Vector. Coordinates of the point from which we want to
#' measure the distance
#' @param a Vector. Coordinates of 1st point over the line
#' @param b Vector. Coordinates of 2st point over the line
#' @return Numeric value result
#' @examples
#' dist2d(x = c(5, 2))
#' dist2d(x = c(5, 2), a = c(0, 0), b = c(0, 1))
#' dist2d(x = c(5, 2), a = c(0, 0), b = c(1, 0))
#' @export
dist2d <- function(x, a = c(0, 0), b = c(1, 1)) {
  if (all(c(length(x), length(a), length(b)) == 2)) {
    v1 <- a - b
    v2 <- x - a
    m <- cbind(v1, v2)
    d <- abs(det(m)) / sqrt(sum(v1 * v1))
    d
  } else {
    stop("Every point most consist of 2 values (X & Y), thus all input lengths must be 2.")
  }
}


####################################################################
#' Remove/Drop Columns in which ALL or SOME values are NAs
#'
#' This function lets the user remove all columns that have some or
#' all values as NAs
#'
#' @family Data Wrangling
#' @param df Data.frame
#' @param all Boolean. Remove columns containing ONLY \code{NA} values.
#' If set to \code{FALSE}, remove columns containing at least one \code{NA}.
#' @param ignore Character vector. Column names to ignore validation.
#' @return data.frame with removed columns.
#' @export
#' @rdname filterdata
removenacols <- function(df, all = TRUE, ignore = NULL) {
  if (is.null(df)) {
    NULL
  } else {
    if (!is.data.frame(df)) stop("Input 'df' must be a valid data.frame")
    if (all) {
      not_all_nas <- colSums(is.na(df)) != nrow(df)
      keep <- (colnames(df) %in% ignore) | as.vector(not_all_nas)
      df[, keep]
    } else {
      df[, complete.cases(t(df[!colnames(df) %in% ignore]))]
    }
  }
}

####################################################################
#' Remove/Drop Rows in which ALL or SOME values are NAs
#'
#' This function lets the user remove all rows that have some or
#' all values as NAs
#'
#' @family Data Wrangling
#' @param df Data.frame
#' @param all Boolean. Remove rows which contains ONLY NA values.
#' If set to FALSE, rows which contains at least one NA will be removed
#' @return data.frame with removed rows.
#' @export
#' @rdname filterdata
removenarows <- function(df, all = TRUE) {
  if (is.null(df)) {
    NULL
  } else {
    if (!is.data.frame(df)) stop("df must be a valid data.frame")
    if (all) {
      df[rowSums(is.na(df)) != ncol(df), ]
    } else {
      df[complete.cases(df), ]
    }
  }
}


####################################################################
#' Select only numerical columns in a dataframe
#'
#' @family Data Wrangling
#' @param df Data.frame
#' @param dropnacols Boolean. Drop columns with only NA values?
#' @param logs Boolean. Calculate log(x)+1 for numerical columns?
#' @param natransform String. "mean" or 0 to impute NA values. If
#' set to NA no calculation will run.
#' @return data.frame with all numerical columns selected.
#' @examples
#' data(dft) # Titanic dataset
#' str(dft)
#' numericalonly(dft) %>% head()
#' numericalonly(dft, natransform = "mean") %>% head()
#' @export
#' @rdname filterdata
numericalonly <- function(df, dropnacols = TRUE, logs = FALSE, natransform = NA) {
  # Drop ALL NAs columns
  if (dropnacols) df <- removenacols(df, all = TRUE)

  # Which character columns may be used as numeric?
  transformable <- apply(df, 2, function(x) length(unique(x)))
  which <- names(transformable[transformable == 2])
  dfn <- data.frame(df[, colnames(df) %in% which])
  colnames(dfn) <- which
  non_numeric <- mutate_all(dfn, list(~ as.integer(as.factor(.)) - 1))
  # Which are already numeric?
  numeric <- select_if(df, is.numeric)

  # Calculate logs
  if (logs) {
    # Non binary numeric features
    whichlog <- colnames(numeric)[!colnames(numeric) %in% which]
    numeric <- numeric %>% mutate_at(vars(whichlog), list(log = ~ log(. + 1)))
    is.na(numeric) <- do.call(cbind, lapply(numeric, is.infinite))
  }

  # Join everything
  d <- cbind(numeric, non_numeric[!colnames(non_numeric) %in% colnames(numeric)])

  if (!is.na(natransform)) {
    if (natransform == 0) {
      d[is.na(d)] <- 0
    }
    if (natransform == "mean") {
      for (i in seq_along(d)) {
        if (median(d[, i], na.rm = TRUE) != 0) {
          d[is.na(d[, i]), i] <- mean(d[, i], na.rm = TRUE)
        } else {
          d[is.na(d[, i]), i] <- 0
        }
      }
    }
  }
  d
}


####################################################################
#' Replace Factor Values
#'
#' This function lets the user replace levels on a factor vector.
#'
#' @family Data Wrangling
#' @param x Factor (or Character) Vector
#' @param original String or Vector. Original text you wish to replace
#' @param change String or Vector. Values you wish to replace the originals with
#' @return Factor vector with transformed levels.
#' @examples
#' library(dplyr)
#' data(dft)
#' # Replace a single value
#' dft <- mutate(dft, Pclass = replacefactor(Pclass, original = "1", change = "First"))
#' levels(dft$Pclass)
#' # Replace multiple values
#' dft <- mutate(dft, Pclass = replacefactor(Pclass, c("2", "3"), c("Second", "Third")))
#' levels(dft$Pclass)
#' @export
replacefactor <- function(x, original, change) {
  if (length(original) != length(change)) {
    stop("You must provide same length vectors for original and change!")
  }
  for (i in seq_along(change)) {
    if (!is.factor(x)) {
      warning("Automatically turned non-factor variable into factor")
      x <- factor(x, levels = unique(as.character(x)))
    }
    levels(x)[levels(x) == original[i]] <- change[i]
  }
  x
}

####################################################################
#' Extract file raw name and type from file names
#'
#' @family Data Wrangling
#' @param filepath Character vector. File path(s) to get file raw names
#' without extension nor path OR extension without path nor raw name.
#' @examples
#' file_name("file.aux")
#' file_name("temp/file.R")
#' file_name("/temp/temp3/music.mp3")
#' @export
file_name <- function(filepath) {
  sub(pattern = "(.*)\\..*$", replacement = "\\1", basename(filepath))
}

####################################################################
#' Get file extensions without file names
#'
#' @family Data Wrangling
#' @rdname file_name
#' @examples
#' file_type("file.aux")
#' file_type("temp/file.R")
#' file_type("/temp/temp3/music.mp3")
#' @export
file_type <- function(filepath) tolower(gsub(".*\\.", "", filepath))

####################################################################
#' Download MP3 from URL
#'
#' This function downloads YouTube videos or Soundcloud or any other
#' platform supported by the youtube-dl library, and converts them into
#' high quality MP3 files. The URL can be for a single video or a whole
#' playlist. It also returns metadata into an (invisible) list.
#'
#' @section youtube-dl:
#' More info from the original developers and its code:
#' \href{https://github.com/ytdl-org/youtube-dl/}{youtube-dl's Github}
#'
#' @family Scrapper
#' @family Audio
#' @param id Character. YouTube URL or ID to search for.
#' @param mp3 Boolean. Add mp3 optimal parameters?
#' @param repo Character. Chose repository you installed youtube-dl from.
#' Any of: "youtube-dl" (latest stable version), "yt-dlp"
#' (latest dev version).
#' @param params Character. Additional parameters.
#' @param start_time,end_time Numeric. Start and end time
#' to trim the audio output in seconds.
#' @param overwrite Boolean. Overwrite original file?
#' @param open,delete Boolean. After everything's done, should the
#' file be opened? Should it be deleted?
#' @param info Boolean. Import and return metadata?
#' @param title Character. Title of the track to search for cover, metadata, and
#' name the file.
#' @param cover Boolean. Google Search its squared cover? Uses \code{title}
#' input when provided.
#' @param metadata Boolean. Use iTunes Search API (free) to extract "track" 
#' data and high-res cover art. Falls back to "spotifyr" if no results
#' are found (requires credentials and, since 2026-03-09, Spotify Premium 
#' linked to the client id). Uses \code{title} input when provided.
#' @param quiet Boolean. Keep quiet? If not, informative messages will be shown.
#' @return (Invisible) list with id's meta-data.
#' @examples
#' # You must have "youtube-dl" installed in your OS:
#' \dontrun{
#' # Download video from YouTube and convert to MP3
#' mp3_get("https://www.youtube.com/watch?v=lrlKcCdVw9Q")
#' # OR simply
#' mp3_get("lrlKcCdVw9Q")
#' # For dev version, use:
#' mp3_get("m3RX4LJh0iI", repo = "yt-dlp")
#' }
#' @export
mp3_get <- function(id,
                    mp3 = TRUE,
                    repo = "yt-dlp",
                    params = "--no-check-certificate",
                    start_time = 0,
                    end_time = NA,
                    overwrite = TRUE,
                    open = FALSE,
                    delete = open,
                    info = TRUE,
                    title = NULL,
                    cover = FALSE,
                    metadata = FALSE,
                    quiet = FALSE) {
  # Build query's parameters
  query <- "--rm-cache-dir"
  if (mp3) {
    query <- c(
      query,
      "-f bestaudio",
      "--extract-audio",
      "--audio-format mp3",
      "--audio-quality 0"
    )
  }
  if (info) query <- c(query, "--write-info-json")
  query <- c(query, '-o "%(title)s.%(ext)s"', params)
  query <- v2t(c(repo, query, id), quotes = FALSE, sep = " ")
  if (!quiet) message(v2t(c("Query:", query), quotes = FALSE, sep = " "))

  # Run youtube-dl
  ret <- tryCatch(
    {
      system(query)
    },
    error = function(err) {
      msg <- "Something went wrong. Do you have youtube-dl installed?"
      if (grepl("^darwin", R.version$os)) {
        msg <- paste(msg, "Run in Terminal: brew install youtube-dl",
          "Then restart and try again",
          sep = "\n"
        )
      }
      msg <- paste(msg, "If already installed, try updating it with:",
        "sudo pip3 install --upgrade youtube_dl",
        sep = "\n"
      )
      msg <- paste(msg, "Our you could also try latest dev version that contains new features:",
        "https://github.com/yt-dlp/yt-dlp/wiki/Installation",
        sep = "\n"
      )
      stop(msg)
    }
  )

  if (1 %in% ret) {
    invisible(NULL)
  } else {
    f <- listfiles(getwd(), recursive = FALSE) %>%
      filter(grepl("\\.info\\.json", .data$filename)) %>%
      arrange(desc(.data$mtime)) %>%
      .[1, 1]

    infox <- jsonlite::read_json(f)
    invisible(file.remove(f))
    infox[["formats"]] <- NULL

    # The actual file path created by youtube-dl
    mp3_file_original <- gsub(".info.json", ".mp3", f)
    mp3_file <- mp3_file_original

    if (!is.null(title)) {
      infox$title <- title
      safe_title_for_file <- trimws(title)
      safe_title_for_file <- gsub("\\s+", " ", safe_title_for_file)
      new_mp3_file_base <- paste0(safe_title_for_file, ".mp3")
      new_mp3_path <- file.path(dirname(mp3_file_original), new_mp3_file_base)
      if (file.exists(mp3_file_original)) {
        if (file.exists(new_mp3_path) && overwrite) {
          file.remove(new_mp3_path)
        } else if (file.exists(new_mp3_path) && !overwrite) {
          if (!quiet) warning(paste("File already exists:", basename(new_mp3_path), "- skipping rename"))
          mp3_file <- mp3_file_original
        }
        if (file.exists(mp3_file_original) && !file.exists(new_mp3_path)) {
          file.rename(mp3_file_original, new_mp3_path)
          mp3_file <- new_mp3_path
          if (!quiet) message(sprintf("File renamed to: %s", basename(mp3_file)))
        }
      } else {
        if (!quiet) warning("Could not find the original MP3 file for renaming. Proceeding with original filename.")
      }
    }

    filename <- basename(mp3_file)

    if (metadata) {
      # Clean search term: remove (lyrics), [official], etc.
      q <- gsub("\\(?lyrics?\\)?|\\(?official\\)?|\\(?video\\)?|\\(?hd\\)?|\\(?4k\\)?", "", infox$title, ignore.case = TRUE)
      q <- gsub("\\[?lyrics?\\]?|\\[?official\\]?|\\[?video\\]?|\\[?hd\\]?|\\[?4k\\]?", "", q, ignore.case = TRUE)
      q <- gsub("\\s+", " ", q)
      q <- trimws(q)
      message(">>> Adding metadata: ", q)
      
      sp <- NULL
      source <- NULL

      # 1. Try iTunes Search API (Free, no key required, highly reliable)
      tryCatch({
        url <- paste0("https://itunes.apple.com/search?term=", URLencode(q), "&media=music&entity=song&limit=1")
        itunes <- jsonlite::fromJSON(url)$results
        if (length(itunes) > 0 && nrow(itunes) > 0) {
          message(">>> Source: iTunes Search API")
          sp <- data.frame(
            name = itunes$trackName,
            artists = I(list(data.frame(name = itunes$artistName))),
            album.name = itunes$collectionName,
            album.release_date = as.character(as.Date(itunes$releaseDate)),
            track_number = itunes$trackNumber,
            album.total_tracks = itunes$trackCount,
            genre = itunes$primaryGenreName,
            cover_url = gsub("100x100bb", "600x600bb", itunes$artworkUrl100),
            stringsAsFactors = FALSE
          )
          source <- "itunes"
        }
      }, error = function(e) {
        warning("iTunes metadata collection failed: ", e$message)
      })

      # 2. Try Spotify as Fallback (requires credentials and Premium)
      if (is.null(sp)) {
        try_require("spotifyr")
        creds <- get_creds("spotify")
        if (!is.null(creds$SPOTIFY_CLIENT_ID) && !is.null(creds$SPOTIFY_CLIENT_SECRET)) {
          tryCatch({
            authorization <- get_spotify_access_token(
              client_id = creds$SPOTIFY_CLIENT_ID,
              client_secret = creds$SPOTIFY_CLIENT_SECRET
            )
            sp <- search_spotify(
              q = q,
              type = "track",
              limit = 3,
              authorization = authorization
            ) %>%
              rowwise() %>%
              mutate(matching = grepl(.data$name, q) |
                grepl(.data$album.name, q)) %>%
              arrange(desc(.data$matching)) %>%
              head(1)
            
            if (nrow(sp) > 0) {
              message(">>> Source: Spotify API")
              source <- "spotify"
            }
          }, error = function(e) {
            warning("Spotify metadata collection failed: ", e$message)
          })
        }
      }

      # 3. Process and apply metadata
      if (!is.null(sp) && nrow(sp) > 0) {
        tryCatch({
          if (source == "spotify") {
            album <- get_album(sp$album.id, authorization = authorization)
            # Fetch (first) artist's genre if no album genres available
            if (length(album$genres) == 0) {
              artist <- search_spotify(
                sp$artists[[1]]$name[1],
                type = "artist", limit = 1,
                authorization = authorization
              )
              genres <- cleanText(artist$genres[[1]][1],
                spaces = TRUE, keep = c("&", "/"), title = TRUE
              )
            } else {
              genres <- v2t(album$genres, sep = " / ")
            }
            label <- album$label
            cover_url <- sp$album.images[[1]]$url[1]
          } else {
            # iTunes/Fallback specific data
            genres <- sp$genre
            label <- NULL
            cover_url <- sp$cover_url
          }

          # Update all metadata tags
          infox$audio <- mp3_update_tags(
            mp3_file,
            title = sp$name,
            artist = v2t(sp$artists[[1]]$name, quotes = FALSE),
            album = sp$album.name,
            release_date = sp$album.release_date,
            recording_date = sp$album.release_date,
            track_num = tuple(count = sp$track_number, total = sp$album.total_tracks),
            genre = genres,
            label = label
          )

          # Handle cover art from URL
          if (!is.null(cover_url)) {
            message(">>> Adding cover...")
            tmp_img <- tempfile(fileext = ".jpg")
            httr::GET(cover_url, httr::write_disk(tmp_img, overwrite = TRUE))
            img_bytes <- readBin(tmp_img, what = "raw", n = file.info(tmp_img)$size)
            infox$audio$tag$images$set(3L, img_bytes, "image/jpeg", "Cover")
            infox$audio$tag$save()
            cover <- FALSE
          }
          infox$metadata <- sp
        }, error = function(e) {
          warning("Metadata processing failed: ", e$message)
        })
      } else {
        warning("No metadata found for: ", infox$title)
      }
    }

    if (cover && mp3 && info) {
      aux <- gsub("\\.mp3", "", infox$title)
      aux <- gsub("lyrics|lyric|official|video", "", tolower(aux))
      aux <- gsub(" ", "\\+", cleanText(aux))
      url <- glued("https://www.google.com/search?q={aux}&tbm=isch&tbs=iar%3As")
      browseURL(url)
    }

    # TRIM START AND/OR END OF AUDIO FILE
    if (any(c(start_time > 0, !is.na(end_time)))) {
      message(">>> Trimming audio file: ", filename)
      mp3_trim(
        file = mp3_file,
        start_time = start_time,
        end_time = end_time,
        overwrite = overwrite,
        quiet = quiet
      )
    }

    # Open file once everything is done
    if (open) {
      if (file.exists(mp3_file)) {
        message("Opening file: ", mp3_file)
        browseURL(mp3_file)
      } else {
        warning("Can't open file; possibly due to strange characters in title: ", mp3_file)
      }
    }
    # Delete file if delete
    if (delete) {
      if (file.exists(mp3_file)) {
        message("Deleting file: ", mp3_file)
        if (open) Sys.sleep(5)
        file.remove(mp3_file)
      } else {
        warning("Can't delete file; possibly due to strange characters in title: ", mp3_file)
      }
    }
    invisible(infox)
  }
}


####################################################################
#' Trim MP3 Audio File
#'
#' This function trims MP3 files given a start and/or end numeric
#' timestamp. Requires \code{ffmpeg} installed in your machine.
#'
#' @family Audio
#' @inheritParams mp3_get
#' @param file Character. File name to trim.
#' @param ext Character. File extension/type.
#' @export
mp3_trim <- function(file, start_time = 0, end_time = NA,
                     overwrite = FALSE, ext = "mp3",
                     quiet = FALSE) {
  start <- paste("-ss", start_time)
  end <- ifelse(!is.na(end_time), paste("-to", end_time), "")
  for (i in file) {
    f <- ifelse(endsWith(i, ext), i, sprintf("%s.%s", file_name(i), ext))
    if (!file.exists(f)) {
      message(paste("File", f, "does not exist or can't be found."))
      next
    }
    # Quality for MP3: -q:a 0 is best VBR quality (~245kbps)
    # Preservation of path and renaming
    file_trimmed <- paste0(file_name(f), "_trimmed.", ext)
    if (dirname(f) != ".") {
      file_trimmed <- file.path(dirname(f), file_trimmed)
    }
    query2 <- paste(
      "ffmpeg -hide_banner -loglevel panic -y",
      start, end, "-i",
      sprintf("'%s'", f),
      "-q:a 0",
      sprintf("'%s'", file_trimmed)
    )
    if (!quiet) message("Query: ", query2)
    system(query2)
    if (overwrite && file.exists(file_trimmed)) {
      file.remove(f)
      file.rename(file_trimmed, f)
    }
  }
}


####################################################################
#' Update MP3 Metadata Tags
#'
#' Updates the ID3 metadata tags of an MP3 file using the Python `eyeD3` library
#' via the `reticulate` package. You can modify standard fields such as
#' \code{title}, \code{artist}, \code{album}, and \code{genre}, as well as
#' additional tags passed through \code{...} if they exist in the MP3 file.
#' @details
#' The function requires Python and the \code{eyeD3} package installed. It will
#' automatically initialize ID3 tags if they do not exist. Tags provided via
#' \code{...} are checked against the existing tag names to avoid errors.
#'
#' @family Audio
#' @param filename Character. Path to the MP3 file to update.
#' @param title Character. New title of the track (optional).
#' @param artist Character. New artist name (optional).
#' @param album Character. New album name (optional).
#' @param genre Character. New genre name (optional).
#' @param ... Additional named arguments corresponding to other ID3 tags.
#' Only tags that exist in the MP3 file will be updated.
#' @examples
#' \dontrun{
#' mp3_update_tags(
#'   "song.mp3",
#'   title = "My Jazz Song",
#'   artist = "Bernardo",
#'   album = "Smooth Album",
#'   genre = "Jazz"
#' )
#' }
#' @return Invisibly returns \code{NULL}. The MP3 file is updated in-place.
#' @export
mp3_update_tags <- function(filename, title = NULL, artist = NULL, album = NULL, genre = NULL, ...) {
  if (!file.exists(filename)) {
    message("File does not exist. Check filename input and working directory....")
  } else {
    try_require("reticulate")
    eyeD3 <- NULL
    eyeD3 <- import("eyed3")
    audio <- eyeD3$load(normalizePath(filename))
    if (is.null(audio$tag)) audio$initTag()
    tag_args <- list(title = title, artist = artist, album = album, genre = genre, ...)
    valid_tags <- names(audio$tag)
    for (n in names(tag_args)) {
      if (!is.null(tag_args[[n]]) && n %in% valid_tags) {
        audio$tag[[n]] <- tag_args[[n]]
      }
    }
    audio$tag$save()
  }
  invisible(audio)
}

####################################################################
#' Robyn: Generate default hyperparameters
#'
#' Generate a list with hyperparameter default values, ready to be
#' passed to \code{Robyn::robyn_inputs()}.
#'
#' @family Robyn
#' @param channels Character vector. Paid media and organic variables names.
#' @param media_type Character vector. Must be length 1 or same as
#' \code{channels}. Pick, for every \code{channels} value,
#' what type of media it is: "online" or "offline".
#' @param adstock Character. Pick one of: "geometric" or "weibull".
#' @param date_type Character. Pick one of: "daily", "weekly", or "monthly".
#' Only valid to transform thetas when using geometric adstock. Set to "skip"
#' in case you wish to leave default weekly values.
#' @param lagged Boolean vector. Must be length 1 or same as
#' \code{channels}. Pick, for every \code{channels} value,
#' if you wish to have a lagged effect. Only valid for Weibull adstock.
#' @return list with default hyperparameters ranges.
#' @examples
#' robyn_hypsbuilder(
#'   channels = c(
#'     "branded_search_spend",
#'     "nonbranded_search_spend",
#'     "print_spend",
#'     "ooh_spend",
#'     "tv_spend",
#'     "radio_spend"
#'   ),
#'   media_type = c(
#'     "online", "online", "offline",
#'     "offline", "offline", "offline"
#'   ),
#'   adstock = "geometric",
#'   date_type = "weekly"
#' )
#' @export
robyn_hypsbuilder <- function(
    channels,
    media_type = "default",
    adstock = "geometric",
    date_type = "weekly",
    lagged = FALSE) {
  # Check inputs validity
  check_opts(media_type, c("default", "online", "offline"))
  check_opts(adstock, c("geometric", "weibull", "weibull_pdf", "weibull_cdf"))
  check_opts(date_type, c("daily", "weekly", "monthly", "skip"))
  check_opts(lagged, c(TRUE, FALSE))

  # Hyperparameters names
  hyps <- c("alphas", "gammas")
  hyps2 <- if (adstock %in% "geometric") "thetas" else c("shapes", "scales")
  all_hyps <- c(hyps, hyps2)

  # Repeat to all channels when provided 1 value
  if (length(media_type) == 1) {
    media_type <- rep(media_type, length(channels))
  }
  if (length(lagged) == 1) {
    lagged <- rep(lagged, length(channels))
  }
  if (any(lagged) && adstock %in% c("geometric", "weibull_cdf")) {
    stop("To be able to have a lagged effect you need to set 'weibull_pdf' adstock")
  }
  stopifnot(length(channels) == length(media_type))

  # Generate all combinations and data.frame
  df <- expand.grid(channels, all_hyps)
  df$media_type <- rep(media_type, length(all_hyps))
  df$lagged <- rep(lagged, length(all_hyps))

  # Apply default rules
  df <- df %>%
    mutate(low = case_when(
      .data$Var2 == "alphas" ~ 0.01,
      .data$Var2 == "gammas" ~ 0.3,
      .data$Var2 == "thetas" ~ 0.1,
      .data$Var2 == "shapes" & isTRUE(.data$lagged) ~ 2,
      .data$Var2 == "shapes" ~ 0,
      .data$Var2 == "scales" ~ 0
    )) %>%
    mutate(high = case_when(
      .data$Var2 == "alphas" & .data$media_type %in% c("online", "default") ~ 3,
      .data$Var2 == "alphas" & .data$media_type == "offline" ~ 1,
      .data$Var2 == "gammas" ~ 1,
      .data$Var2 == "thetas" ~ 0.5,
      .data$Var2 == "shapes" & isTRUE(.data$lagged) ~ 10,
      .data$Var2 == "shapes" & isFALSE(.data$lagged) ~ 1,
      .data$Var2 == "shapes" ~ 10,
      .data$Var2 == "scales" & isTRUE(.data$lagged) ~ 0.05,
      .data$Var2 == "scales" & isFALSE(.data$lagged) ~ 0.2,
      .data$Var2 == "scales" ~ 0.2
    )) %>%
    # Applies only for geometric adstock
    mutate(
      low = case_when(
        .data$Var2 == "thetas" & date_type == "daily" ~ .data$low^(1 / 7),
        .data$Var2 == "thetas" & date_type == "monthly" ~ .data$low^(4),
        TRUE ~ .data$low
      ),
      high = case_when(
        .data$Var2 == "thetas" & date_type == "daily" ~ .data$high^(1 / 7),
        .data$Var2 == "thetas" & date_type == "monthly" ~ .data$high^(4),
        TRUE ~ .data$high
      )
    ) %>%
    mutate(Var1 = factor(.data$Var1, levels = channels)) %>%
    arrange(.data$Var1)

  # Return as named list
  out <- lapply(seq_along(df$Var1), function(x) c(df$low[x], df$high[x]))
  names(out) <- paste(df$Var1, df$Var2, sep = "_")
  out <- append(out, list(train_size = c(0.5, 0.8)))
  out
}


####################################################################
#' Robyn: Model Selection by Weighted Criteria Scores
#'
#' @description
#' Consider N best models to select the right ones to study using
#' several criteria/metrics such as potential improvement on budget
#' allocator, how many non-zero coefficients there are, R squared,
#' historical performance, baseline expectation, etc.
#'
#' Read more about this functionality in Medium post:
#' \href{https://medium.com/@laresbernardo/select-the-right-mmm-candidate-based-on-your-specific-criteria-and-business-knowledge-1f583c3cb97a}{here}.
#'
#' @family Robyn
#' @inheritParams cache_write
#' @param InputCollect,OutputCollect Robyn output objects.
#' @param metrics Character vector. Which metrics do you want to consider?
#' Pick any combination from: "rsq_train" for trained R squared,
#' "performance" for ROAS or (inverse) CPA, "potential_improvement" for
#' default budget allocator improvement using \code{allocator_limits},
#' "non_zeroes" for non-zero beta coefficients, "incluster_models" for
#' amount of models per cluster, "baseline_dist" for the difference between
#' the model's baseline and \code{baseline_ref} value, "certainty" metric
#' to minimize the channels' distance to their cluster's mean performance,
#' weighted by spends if \code{spend_wt = TRUE}, "cluster_sd" metric to score
#' based on the paid channels' performance standard deviations in clusters.
#' Additionally, you can use the standard MOO errors:
#' "nrmse", "decomp.rssd", and "mape" (the lowest the error, the highest
#' the score; same for "baseline_dist" and "cluster_sd").
#' @param wt Vector. Weight for each of the normalized \code{metrics} selected,
#' to calculate the score and rank models. Must have the same order and length
#' of \code{metrics} parameter input.
#' @param baseline_ref Numeric value. Between 0 and 1. What is the baseline
#' percentage you expect? Baseline in this case are all the sales or conversions
#' from non-media channels (organic & paid). Use with "baseline_dist" metric.
#' @param top Integer. How many ranked models to star? The better the model
#' is, the more stars it will have marked.
#' @param allocator_limits Numeric vector, length 2. How flexible do you
#' want to be with the budget allocator? By default, we'll consider a
#' 0.5X and 2X range to let the budget shift across channels.
#' @param n_per_cluster Integer. How many models per cluster do you
#' want to plot? Default: 5. Keep in mind they will all be considered
#' for the calculations.
#' @param cache Use cache functionality for allocator's results?
#' @return list with resulting ranked data.frames, weights and plot.
#' @export
robyn_modelselector <- function(
    InputCollect,
    OutputCollect,
    metrics = c(
      "rsq_train", "performance",
      "potential_improvement",
      "non_zeroes",
      "incluster_models", "cluster_sd",
      "certainty", "baseline_dist"
    ),
    wt = c(2, 0.1, 0, 1, 0.1, 0, 1.5, 0),
    baseline_ref = 0,
    top = 4,
    n_per_cluster = 5,
    allocator_limits = c(0.5, 2),
    quiet = FALSE,
    cache = TRUE,
    ...) {
  if (length(OutputCollect$allSolutions) <= 1) {
    msg <- "Not enough models for model selection"
    warning(msg)
    ret <- invisible(list(
      data = tibble(solID = OutputCollect$allSolutions, score = 1),
      plot = noPlot(msg)
    ))
    class(ret) <- c("robyn_modelselector", class(ret))
    ret
  } else {
    if (length(metrics) == 1) wt <- 1
    stopifnot(length(wt) == length(metrics))
    stopifnot(length(allocator_limits) == 2)
    stopifnot(baseline_ref >= 0 && baseline_ref <= 1)

    # Available metrics
    metrics_df <- data.frame(
      metric = c(
        "rsq_train", "performance", "potential_improvement",
        "non_zeroes", "incluster_models",
        "baseline_dist", "certainty", "cluster_sd",
        "nrmse", "decomp.rssd", "mape"
      ),
      metric_name = c(
        "R^2", ifelse(InputCollect$dep_var_type == "revenue", "High ROAS", "Low CPA"),
        "Potential Boost", "Non-Zero Betas", "Models in Cluster",
        sprintf("Baseline Distance [%.0f%%]", signif(baseline_ref * 100, 2)),
        "Certainty in Cluster", "Cluster Mean Std Dev",
        "1 - NRMSE", "1 - DECOMP.RSSD", "1 - MAPE"
      )
    )
    # The following criteria are inverted because the smaller, the better
    invert_criteria <- c("nrmse", "decomp.rssd", "mape")
    check_opts(metrics, metrics_df$metric)

    # Metrics Used
    metrics_used <- filter(metrics_df, .data$metric %in% metrics) %>%
      arrange(match(.data$metric, metrics)) %>%
      left_join(data.frame(metric = metrics, wt = wt), "metric") %>%
      filter(.data$wt > 0)

    # Add Default Potential Improvement values
    sols <- sort(OutputCollect$allSolutions)
    if ("potential_improvement" %in% metrics & isTRUE(wt[which(metrics == "potential_improvement")] != 0)) {
      cache_name <- c("robyn_modelselector", length(sols))
      if (cache & cache_exists(cache_name, ...)) {
        potOpt <- cache_read(cache_name, quiet = quiet, ...)
      } else {
        if (!quiet) {
          message(sprintf(
            ">>> Calculating potential performance improvements on %s Pareto-front models...",
            length(sols)
          ))
        }
        try_require("Robyn")
        defpot <- lapply(sols, function(x) {
          if (!quiet) statusbar(which(sols == x), length(sols), x)
          suppressMessages(robyn_allocator(
            InputCollect = InputCollect,
            OutputCollect = OutputCollect,
            select_model = x,
            channel_constr_low = min(allocator_limits),
            channel_constr_up = max(allocator_limits),
            plots = FALSE,
            export = FALSE,
            quiet = TRUE,
            ...
          ))
        })
        potOpt <- data.frame(
          solID = sols,
          potential_improvement = unlist(lapply(defpot, function(x) {
            x$dt_optimOut$optmResponseUnitTotalLift[1]
          }))
        )
        if (cache) cache_write(potOpt, cache_name, quiet = quiet, ...)
      }
    } else {
      potOpt <- data.frame(solID = sols, potential_improvement = 0)
    }

    # Check pareto-front models summaries to calculate performance (ROAS/CPA)
    performance <- OutputCollect$xDecompAgg %>%
      filter(!is.na(.data$mean_spend)) %>% # get rid of organic
      filter(.data$solID %in% OutputCollect$clusters$data$solID) %>%
      group_by(.data$solID) %>%
      summarise(
        performance = sum(.data$xDecompAgg) / sum(.data$total_spend, na.rm = TRUE)
      )

    # Check pareto-front models to calculate non-zero betas in media channels
    non_zeroes_rate <- OutputCollect$allPareto$xDecompAgg %>%
      filter(.data$solID %in% OutputCollect$clusters$data$solID) %>%
      group_by(.data$solID) %>%
      summarise(
        non_zeroes = length(.data$rn[
          round(.data$coef, 6) > 0 & .data$rn %in% InputCollect$all_media
        ]) /
          length(InputCollect$all_media)
      )

    # Count models per cluster
    if (!"clusters" %in% names(OutputCollect)) {
      OutputCollect$clusters$data <- data.frame(solID = sols, cluster = "None")
      OutputCollect$clusters$clusters_means <- data.frame(cluster = "None", n = length(sols))
      OutputCollect$clusters$df_cluster_ci <- data.frame(cluster = "None", sd = 0)
    }
    clus_mean_sd <- OutputCollect$clusters$df_cluster_ci %>%
      group_by(.data$cluster) %>%
      summarise(mean_sd = mean(.data$sd, na.rm = TRUE), .groups = "drop") %>%
      mutate(
        cluster_sd = normalize(-.data$mean_sd, range = c(0.01, 1)),
        cluster = as.character(.data$cluster)
      )
    temp <- OutputCollect$clusters$data %>%
      select(.data$solID, .data$cluster) %>%
      mutate(cluster = as.character(.data$cluster)) %>%
      left_join(select(OutputCollect$clusters$clusters_means, .data$cluster, .data$n), "cluster") %>%
      left_join(clus_mean_sd, "cluster") %>%
      rename(incluster_models = "n")

    # Calculate baselines
    baselines <- OutputCollect$xDecompAgg %>%
      mutate(rn = ifelse(.data$rn %in% c(InputCollect$all_media), .data$rn, "baseline")) %>%
      group_by(.data$solID, .data$rn) %>%
      summarize(contribution = sum(.data$xDecompAgg), .groups = "drop") %>%
      group_by(.data$solID) %>%
      mutate(baseline = .data$contribution / sum(.data$contribution)) %>%
      ungroup() %>%
      filter(.data$rn == "baseline") %>%
      arrange(abs(.data$baseline)) %>%
      mutate(
        baseline_dist_real = abs(baseline_ref - .data$baseline),
        baseline_dist = normalize(-.data$baseline_dist_real, range = c(
          0, 1 - min(.data$baseline_dist_real) / max(.data$baseline_dist_real)
        ))
      ) %>%
      select(c("solID", "baseline", "baseline_dist")) %>%
      arrange(.data$baseline_dist)

    # Certainty Criteria: distance to cluster's mean weighted by spend
    certainty <- certainty_score(InputCollect, OutputCollect, ...) %>%
      select("solID", "certainty")

    # Gather everything up
    dfa <- OutputCollect$allPareto$resultHypParam %>%
      filter(.data$solID %in% OutputCollect$clusters$data$solID) %>%
      select(.data$solID, .data$rsq_train, .data$nrmse, .data$decomp.rssd, .data$mape) %>%
      left_join(performance, "solID") %>%
      left_join(non_zeroes_rate, "solID") %>%
      left_join(potOpt, "solID") %>%
      left_join(temp, "solID") %>%
      left_join(certainty, "solID") %>%
      ungroup() %>%
      left_join(baselines, "solID")

    # Helper function to calculate normalized and weighted scores
    calculate_score <- function(metric_name, data, metrics, weights, invert_criteria) {
      score <- 0
      if (metric_name %in% metrics) {
        normalized_value <- normalize(data[[metric_name]], na.rm = TRUE)
        weight <- weights[which(metrics == metric_name)]
        sign <- ifelse(metric_name %in% invert_criteria, -1, 1)
        normalized_value[is.na(normalized_value)] <- 0
        score <- normalized_value * weight * sign
      }
      score
    }

    # Calculate scores
    scores <- list()
    for (metric in metrics_df$metric) {
      scores[[metric]] <- calculate_score(metric, dfa, metrics, wt, invert_criteria)
    }
    dfa <- dfa %>%
      mutate(
        score = normalize(rowSums(bind_cols(scores))),
        aux = rank(-.data$score, ties.method = "first")
      ) %>%
      rowwise() %>%
      mutate(note = ifelse(.data$aux %in% 1:top, paste(
        rep("*", (top + 1) - .data$aux),
        collapse = ""
      ), "")) %>%
      select(-.data$aux) %>%
      arrange(desc(.data$score), desc(3), desc(4))
    if (!quiet) message("Recommended considering these models first: ", v2t(head(dfa$solID, top)))

    # Generate plot/dashboard
    caption <- metrics_used %>%
      mutate(value = .data$wt / sum(.data$wt)) %>%
      arrange(desc(.data$value)) %>%
      mutate(lab = sprintf("%s%% %s", round(100 * .data$value), .data$metric_name)) %>%
      pull(.data$lab) %>%
      paste(., collapse = " | ")
    sorting <- dfa %>%
      mutate(cluster = sprintf("%s (%s)", .data$cluster, .data$incluster_models)) %>%
      group_by(.data$cluster, .data$note) %>%
      tally() %>%
      arrange(desc(.data$note)) %>%
      pull(.data$cluster) %>%
      unique()
    pdat <- dfa %>%
      # So that inverted variables have larger relative bars (darker blue)
      mutate_at(all_of(invert_criteria), function(x) 1 - x) %>%
      mutate(
        cluster = factor(sprintf("%s (%s)", .data$cluster, .data$incluster_models), levels = sorting),
        incluster_models = .data$incluster_models / max(dfa$incluster_models, na.rm = TRUE)
      ) %>%
      tidyr::pivot_longer(any_of(metrics_used$metric)) %>%
      filter(.data$name %in% metrics_used$metric) %>%
      mutate(
        name_metrics = rep(metrics_used$metric_name, length.out = nrow(.)),
        name = factor(.data$name, levels = metrics_used$metric),
        name_metrics = factor(.data$name_metrics, levels = metrics_used$metric_name)
      ) %>%
      group_by(.data$name) %>%
      mutate(top = rank(-.data$value)) %>%
      left_join(select(dfa, .data$solID, .data$rsq_train), "solID") %>%
      group_by(.data$cluster) %>%
      slice(1:((length(metrics_used$metric)) * n_per_cluster)) %>%
      mutate(solID = paste(.data$note, .data$solID))
    p <- pdat %>%
      ggplot(aes(y = reorder(.data$solID, .data$score), x = .data$value)) +
      geom_col(aes(group = .data$name, fill = .data$top)) +
      # geom_vline(xintercept = 1, alpha = 0.5, linetype = "dotted") +
      facet_grid(.data$cluster ~ .data$name_metrics, scales = "free") +
      labs(
        y = NULL, x = "Criteria Scores (the highest the better)",
        title = "Select a model based on these metrics and rankings",
        subtitle = paste(
          "Showing up to", n_per_cluster,
          "models per cluster and ranked solutions marked with stars"
        ),
        caption = paste("Weights:", caption)
      ) +
      scale_x_percent(position = "top") +
      theme_lares(legend = FALSE)

    # Create the exported object
    ret <- invisible(list(
      data = select(dfa, "solID", "score", all_of(metrics_used$metric), everything(), -.data$note),
      metrics = metrics, weights = wt, baseline_ref = baseline_ref,
      weighted_scores = tibble(solID = dfa$solID, score = dfa$score, bind_cols(scores)),
      plot = p
    ))
    class(ret) <- c("robyn_modelselector", class(ret))
    ret
  }
}

### Certainty Criteria: distance to cluster's mean weighted by spend
#
# Formula: spend rate * mean to models' performance distance ^2 * penalization
# Channel Score = Xi = Si * (P - Pi)^2 * Penalization if outside CI
# Model Score = Mi = norm(-sum(Xi)) between 0 and 1, being 1 a perfect certainty.
#
# Where:
# Pi is model's Performance
# P is mean Performance in Cluster
# Si is % of total spend per channel
#
# So we need:
# 1) cluster's mean, low and up CI per cluster and channel
# 2) model's paid channels performance and spends
certainty_score <- function(
    InputCollect, OutputCollect,
    penalization = 2, spend_wt = TRUE, ...) {
  if (!"clusters" %in% names(OutputCollect)) {
    data.frame(
      solID = unique(OutputCollect$xDecompAgg$solID),
      certainty = 0
    )
  } else {
    clusters <- OutputCollect$clusters$df_cluster_ci
    perfs <- OutputCollect$xDecompAgg %>%
      filter(!is.na(.data$mean_spend)) %>% # get rid of organic
      filter(.data$solID %in% OutputCollect$clusters$data$solID) %>%
      mutate(performance = .data$xDecompAgg / .data$total_spend) %>%
      select("solID", "channel" = "rn", "spend" = "total_spend", "performance")
    df <- perfs %>%
      left_join(select(OutputCollect$clusters$data, "solID", "cluster"), "solID") %>%
      left_join(clusters, by = c("cluster", "channel" = "rn")) %>%
      select("solID", "channel", "cluster", "spend",
        "Pi" = "performance",
        "P" = "boot_mean", "Pmin" = "ci_low", "Pmax" = "ci_up"
      )
    res <- df %>%
      group_by(.data$cluster, .data$solID) %>%
      mutate(Si = ifelse(spend_wt == FALSE, 1, .data$spend / sum(.data$spend))) %>%
      ungroup() %>%
      mutate(
        pen = ifelse(.data$P < .data$Pmax & .data$P > .data$Pmin, 1, penalization),
        Xi = .data$Si * ((.data$P - .data$Pi)^2 * .data$pen)
      ) %>%
      group_by(.data$solID, .data$cluster) %>%
      summarize(Mi = sum(.data$Xi, na.rm = TRUE), .groups = "drop") %>%
      ungroup() %>%
      mutate(certainty = normalize(-.data$Mi, range = c(0, 1 - min(.data$Mi) / max(.data$Mi)))) %>%
      arrange(desc(.data$certainty))

    if (FALSE) {
      # Check best and worst
      GeMMMa_onepagers(InputCollect, OutputCollect, res$solID[res$certainty == 1], export = FALSE)[[2]][[2]]
      GeMMMa_onepagers(InputCollect, OutputCollect, res$solID[res$certainty == 0], export = FALSE)[[2]][[2]]

      # Viz distribution of scores per cluster
      ggplot(res, aes(x = as.character(.data$cluster), y = .data$certainty)) +
        geom_boxplot() +
        lares::theme_lares() +
        lares::scale_y_percent()
    }
    res
  }
}

#' @param x robyn_modelselector object
#' @rdname robyn_modelselector
#' @export
plot.robyn_modelselector <- function(x, ...) {
  x$plot
}

####################################################################
#' Robyn: Dynamic Performance and Contribution Report
#'
#' Given a date range, calculate specific and total performance and
#' contribution for each of your marketing and non-marketing channels.
#'
#' @family Robyn
#' @inheritParams cache_write
#' @inheritParams robyn_modelselector
#' @param start_date,end_date Date. Start and end date to filter
#' the data to be reported.
#' @param solID Character. Single ID of the model to report. If there's
#' only one available in OutputCollect, no need to define.
#' @param totals Boolean. Add total rows. This includes summary rows
#' (promotional which is paid and organic channels, baseline, grand total).
#' @param non_promo Boolean. Add non-promotional responses as well?
#' @param marginals Boolean. Include mROAS or mCPA marginal performance metric
#' as an additional column called "marginal". Calculations are based on
#' mean spend and mean response with mean carryover results,
#' between \code{start_date} and \code{end_date}.
#' @param carryovers Boolean. Add mean percentage of carryover response for
#' date range between \code{start_date} and \code{end_date} on paid channels.
#' Keep in mind organic variables also have carryover but currently not showing.
#' @param new_version Boolean. Use dev version's new function for marginal
#' calculations (if available)?
#' @return data.frame with results on ROAS/CPA, spend, response, contribution
#' per channel, with or without total rows.
#' @examples
#' \dontrun{
#' # You may load an exported model to recreate Robyn objects
#' mod <- Robyn::robyn_recreate(json_file = "your_model.json")
#' robyn_performance(mod$InputCollect, mod$OutputCollect)
#' }
#' @export
robyn_performance <- function(
    InputCollect, OutputCollect,
    start_date = NULL, end_date = NULL,
    solID = NULL, totals = TRUE, non_promo = FALSE,
    marginals = FALSE, carryovers = FALSE,
    new_version = FALSE,
    quiet = FALSE, ...) {
  dt_mod <- InputCollect$dt_mod
  df <- OutputCollect$mediaVecCollect
  if (!is.null(solID)) {
    if (length(solID) > 1) {
      stop("Select a single model using 'solID' parameter input")
    }
  } else {
    if (length(unique(df$solID)) > 1) {
      stop("Provide single model results or define 'solID' parameter input")
    } else {
      solID <- unique(df$solID)
    }
  }
  if (is.null(start_date)) {
    start_date <- as.Date(InputCollect$window_start)
  }
  if (is.null(end_date)) {
    end_date <- as.Date(InputCollect$window_end)
  }
  stopifnot(start_date <= end_date)

  # Filter data for ID, modeling window and selected date range
  dt_mod <- dt_mod %>%
    filter(
      .data$ds >= InputCollect$window_start,
      .data$ds <= InputCollect$window_end,
      .data$ds >= start_date, .data$ds <= end_date
    ) %>%
    mutate(solID = solID, type = "rawSpend") %>%
    select(c("ds", "solID", "type", InputCollect$all_media))
  df <- df[df$solID %in% solID, ] %>%
    filter(
      .data$ds >= InputCollect$window_start,
      .data$ds <= InputCollect$window_end,
      .data$ds >= start_date, .data$ds <= end_date
    ) %>%
    select(c("ds", "solID", "type", InputCollect$all_media))
  if (nrow(dt_mod) == 0 && !quiet) {
    warning(sprintf(
      "No data for model %s within modeling window (%s:%s) and date range filtered (%s:%s)",
      solID, InputCollect$window_start, InputCollect$window_end,
      start_date, end_date
    ))
    NULL
  } else {
    spends <- dt_mod %>%
      filter(.data$type == "rawSpend") %>%
      summarise_if(is.numeric, function(x) sum(x, na.rm = TRUE))
    response <- df %>%
      filter(.data$type == "decompMedia") %>%
      summarise_if(is.numeric, function(x) sum(x, na.rm = TRUE))
    metric <- ifelse(InputCollect$dep_var_type == "revenue", "ROAS", "CPA")
    if (metric == "ROAS") {
      performance <- response / spends
      total <- sum(response) / sum(spends)
    } else {
      performance <- spends / response
      total <- sum(spends) / sum(response)
    }
    ret <- dplyr::tibble(
      solID = solID,
      start_date = min(df$ds, na.rm = TRUE),
      end_date = max(df$ds, na.rm = TRUE),
      channel = InputCollect$all_media,
      type = factor(
        case_when(
          InputCollect$all_media %in% InputCollect$paid_media_spends ~ "Paid",
          InputCollect$all_media %in% InputCollect$organic_vars ~ "Organic"
        ),
        levels = c("Paid", "Organic")
      ),
      metric = ifelse(InputCollect$all_media %in% InputCollect$paid_media_spends, metric, ""),
      performance = unlist(performance),
      spend = unlist(spends),
      response = unlist(response)
    ) %>%
      arrange(.data$type, desc(.data$spend))

    # Create TOTAL row
    totals_df <- ret[1, 1:3] %>%
      mutate(
        channel = "PROMOTIONAL TOTAL",
        metric = metric,
        performance = total,
        spend = sum(spends),
        response = sum(response)
      )
    # Add marketing contribution to sales/conversions (dynamic depending on date)
    xdvc <- OutputCollect$xDecompVecCollect
    xDecompTotal <- xdvc[xdvc$solID %in% solID, ] %>%
      filter(.data$ds >= start_date, .data$ds <= end_date) %>%
      summarise_if(is.numeric, function(x) sum(x, na.rm = TRUE))
    xDecompPerc <- xDecompTotal %>%
      # Divide by prediction, not real values
      mutate_all(function(x) x / .$depVarHat) %>%
      tidyr::gather(key = "channel", value = "contribution") %>%
      mutate(response = t(xDecompTotal)[, 1]) %>%
      filter(!.data$channel %in% c("dep_var", "depVarHat"))
    mktg_contr <- filter(xDecompPerc, .data$channel %in% InputCollect$all_media)
    mksum <- sum(mktg_contr$contribution)
    mktg_contr2 <- rbind(
      data.frame(
        channel = c("PROMOTIONAL TOTAL", "BASELINE", "GRAND TOTAL"),
        contribution = c(mksum, 1 - mksum, 1)
      ),
      xDecompPerc[, 1:2]
    )
    # Create Baseline row
    resp_baseline <- (1 - mksum) * sum(ret$response) / mksum
    base_df <- ret[1, 1:3] %>% mutate(
      channel = "BASELINE",
      metric = "",
      performance = NA,
      spend = 0,
      response = resp_baseline
    )
    # Baseline (L4 - until contextual) variables
    if (non_promo) {
      baseL4_contr <- filter(xDecompPerc, !.data$channel %in% InputCollect$all_media)
      base_df <- bind_rows(baseL4_contr, base_df) %>%
        select(all_of(colnames(base_df))) %>%
        tidyr::fill(c("solID", "start_date", "end_date", "metric", "spend"), .direction = "up")
    }
    # Create TOTAL row
    grand_total <- ret[1, 1:3] %>%
      mutate(
        channel = "GRAND TOTAL",
        metric = metric,
        performance = (resp_baseline + sum(ret$response)) / sum(ret$spend),
        spend = sum(ret$spend),
        response = resp_baseline + sum(ret$response)
      )
    # Join everything together
    if (totals) ret <- bind_rows(ret, totals_df, base_df, grand_total)
    ret <- left_join(ret, mktg_contr2, "channel") %>%
      mutate(
        type = factor(
          case_when(
            .data$channel %in% InputCollect$paid_media_spends ~ "Paid",
            .data$channel %in% InputCollect$organic_vars ~ "Organic",
            .data$channel %in% InputCollect$context_vars ~ "Context",
            .data$channel %in% c("intercept", InputCollect$prophet_vars) ~ "Context"
          ),
          levels = c("Paid", "Organic", "Context", NA)
        )
      )

    # Add mROAS/mCPA
    if (marginals) {
      try_require("Robyn")
      # Experimental approach to add organic vars too
      if (packageVersion("Robyn") >= "3.12.0.9007" && new_version) {
        temp <- lapply(InputCollect$all_media, function(x) {
          robyn_marginal(
            InputCollect = InputCollect,
            OutputCollect = OutputCollect,
            select_model = solID,
            metric_name = x,
            metric_value = NULL,
            date_range = c(as.Date(start_date), as.Date(end_date)),
            marginal_unit = 1,
            plots = FALSE, quiet = TRUE
          )
        })
        marginal <- data.frame(
          channel = InputCollect$all_media,
          marginal = unlist(lapply(temp, function(x) x$marginal))
        )
      } else {
        ba_temp <- robyn_allocator(
          InputCollect = InputCollect,
          OutputCollect = OutputCollect,
          select_model = solID,
          date_range = c(as.Date(start_date), as.Date(end_date)),
          export = FALSE, quiet = TRUE, ...
        )
        marginal <- ba_temp$dt_optimOut %>%
          select(c("channels", "initResponseMargUnit")) %>%
          rename(
            "channel" = "channels",
            "marginal" = "initResponseMargUnit"
          ) %>%
          {
            if (metric == "CPA") mutate(., marginal = 1 / .data$marginal) else .
          }
      }
      ret <- left_join(ret, marginal, "channel") %>%
        dplyr::relocate("marginal", .after = "performance")
    }
    # Add carryover response percentage
    if (carryovers) {
      try_require("Robyn")
      carrov <- robyn_immcarr(
        InputCollect, OutputCollect,
        solID = solID,
        start_date = start_date, end_date = end_date, ...
      ) %>%
        filter(.data$type == "Carryover")
      mean_carryovers <- data.frame(
        channel = carrov$rn,
        carryover = signif(carrov$carryover_pct, 4)
      )
      ret <- left_join(ret, mean_carryovers, "channel") %>%
        dplyr::relocate("carryover", .after = "response")
    }
    ret
  }
}

####################################################################
#' Robyn: Marginal Performance (mROAS & mCPA) [Experimental]
#'
#' Calculate and plot marginal performance of any spend or organic variable.
#'
#' @family Robyn
#' @inheritParams cache_write
#' @inheritParams robyn_modelselector
#' @param marginal_unit Additional units to calculate the marginal performance.
#' @return list with base and marginal response results, marginal performance
#' metric and value, and plot.
#' @examples
#' \dontrun{
#' # You may load an exported model to recreate Robyn objects
#' mod <- Robyn::robyn_recreate(json_file = "your_model.json")
#' robyn_marginal(
#'   InputCollect = mod$InputCollect,
#'   OutputCollect = mod$OutputCollect,
#'   metric_name = "emails_O",
#'   metric_value = 100000,
#'   date_range = "all",
#'   marginal_unit = 10000000
#' )
#' }
#' @export
robyn_marginal <- function(..., marginal_unit = 1) {
  try_require("Robyn")
  stopifnot(packageVersion("Robyn") >= "3.12.0.9007")
  args <- args2 <- list(...)
  Response1 <- robyn_response(...)
  args$metric_value <- Response1$metric_value
  Response1 <- do.call(robyn_response, args)
  args2$metric_value <- Response1$metric_value + marginal_unit
  args2$quiet <- TRUE
  Response2 <- do.call(robyn_response, args2)
  ret <- list(Response1 = Response1, Response2 = Response2, inputs = args)
  ret$inputs$marginal_unit <- marginal_unit
  if (args$InputCollect$dep_var_type == "revenue") {
    ret$marginal_metric <- "mROAS"
    ret$marginal <- (Response2$sim_mean_response - Response1$sim_mean_response) /
      (Response2$sim_mean_spend - Response1$sim_mean_spend)
  } else {
    ret$marginal_metric <- "mCPA"
    ret$marginal <- (Response2$sim_mean_spend - Response1$sim_mean_spend) /
      (Response2$sim_mean_response - Response1$sim_mean_response)
  }
  cap <- sprintf(
    "\n%s: %s (marginal units: %s)",
    ret$marginal_metric,
    num_abbr(ret$marginal),
    num_abbr(marginal_unit)
  )
  ret$plot <- ret$Response1$plot + labs(caption = paste0(
    ret$Response1$plot$labels$caption, cap
  ))
  ret
}

#' ####################################################################
#' #' Hubspot contacts (Somos F1)
#' #'
#' #' This function brings all contacts from Somos F1's Hubspot account.
#' #' It also transforms the API's JSON result data into a relational table
#' #' or data.frame. Keep in mind that it will always bring the latest
#' #' value for each parameter (column).
#' #'
#' #' @family Credentials
#' #' @family Hubspot
#' #' @param limit Integer. Limit the amount of contacts to return.
#' #' Multiples of 100 or round up will occur.
#' #' @param creds Character. Credential's user (see \code{get_creds()})
#' #' @export
#' f1_contacts <- function(limit = 1000, creds = NA) {
#'
#'   hsdates <- function(date) {
#'     as.character(
#'       as.POSIXct(
#'         as.numeric(
#'           as.character(date)) / 1000,
#'         origin="1970-01-01"))
#'   }
#'
#'   properties <- c(
#'     # BASIC
#'     "contacts_vid",
#'     "createdate",
#'     "lastmodifieddate",
#'     "email",
#'     "product",
#'     "firstname",
#'     "secondname",
#'     "lastname",
#'     "secondlastname",
#'     "date_of_birth",
#'     "identification",
#'     "identification_date",
#'     "identification_department",
#'     "identification_type",
#'     "identification_city",
#'     "identification_country",
#'     "birth_city",
#'     "birth_department",
#'     "country_birth",
#'     "address",
#'     "home_city",
#'     "home_locality",
#'     "city",
#'     "departament",
#'     "country",
#'     "home_country",
#'     "mobilephone",
#'     # CREDITS
#'     "amount_requested",
#'     "loan_use",
#'     "debt_payees_counter",
#'     "payroll_deduction",
#'     "credit_value",
#'     "disbursed",
#'     "monthly_rate",
#'     "vehicle_brand",
#'     "vehicle_model",
#'     "vehicle_commercial_value",
#'     "vehicle_line",
#'     "vehicle_class",
#'     "vehicle_service_type",
#'     "vehicle_use",
#'     "sale_value",
#'     "credit_entity",
#'     "credit_installments",
#'     "payment_day",
#'     "dependents",
#'     "country_nationality",
#'     "civil_status",
#'     "educational_level",
#'     "profession",
#'     "contract_type",
#'     "economic_activity",
#'     "independent_type",
#'     "residence_months",
#'     "housing_type",
#'     "stratum",
#'     "sex",
#'     "violence_victim",
#'     "pep_public_resources",
#'     "pep_public_power",
#'     "pep_public_recognition",
#'     "working_continuity",
#'     "company",
#'     "company_nit",
#'     "company_relation",
#'     "company_activity",
#'     "company_position",
#'     "company_address",
#'     "company_locality",
#'     "company_city",
#'     "company_department",
#'     "company_phone",
#'     "company_phone_ext",
#'     "monthly_incomes",
#'     "monthly_outcomes",
#'     "total_assets",
#'     "total_liabilities",
#'     "reference1_firstname",
#'     "reference1_secondname",
#'     "reference1_lastname",
#'     "reference1_secondlastname",
#'     "reference1_city",
#'     "reference1_mobilephone",
#'     "reference2_firstname",
#'     "reference2_secondname",
#'     "reference2_firstlastname",
#'     "reference2_secondlastname",
#'     "family_relationship",
#'     "reference2_city",
#'     "reference2_mobilephone",
#'     "international_transactions",
#'     "country_transactions",
#'     "international_transactions_types",
#'     "international_tax_declaration",
#'     "international_tax_declaration_country",
#'     "foreign_currency",
#'     # FORM 4
#'     "file_identification",
#'     "files_labour_certificate",
#'     "files_payroll",
#'     # OTHER
#'     "lifecyclestage",
#'     "ipaddress",
#'     "ip_city",
#'     "ip_state",
#'     "ip_country",
#'     "hs_ip_timezone",
#'     "currentlyinworkflow",
#'     "hs_analytics_average_page_views",
#'     "hs_analytics_num_visits",
#'     "hs_social_last_engagement",
#'     "num_unique_conversion_events",
#'     "hs_analytics_num_event_completions",
#'     "first_conversion_event_name",
#'     "recent_conversion_event_name",
#'     "hs_analytics_num_page_views",
#'     "hs_lifecyclestage_lead_date",
#'     "hs_email_domain",
#'     "hs_analytics_source",
#'     "num_conversion_events",
#'     "hs_social_num_broadcast_clicks",
#'     "first_conversion_date",
#'     "hs_analytics_first_timestamp",
#'     "hs_analytics_first_visit_timestamp",
#'     "hs_analytics_first_touch_converting_campaign",
#'     "hs_analytics_last_timestamp",
#'     "hs_analytics_last_visit_timestamp",
#'     "hs_analytics_last_touch_converting_campaign",
#'     "hs_email_optout",
#'     "form_submissions")
#'
#'   credentials <- get_credentials("hubspot", dir = creds)
#'
#'   # ALL Contacts with properties
#'   properties_string <- paste(rbind("property=", properties, "&"), collapse="")
#'   API <- "https://api.hubapi.com/contacts/v1/lists/all/contacts/recent?"
#'   hapikey <- paste0("hapikey=", credentials$token, "&count=100")
#'   URL <- paste0(API, properties_string, hapikey)
#'   contacts <- bring_api(URL, status = TRUE)
#'   message(paste(nrow(contacts),"contacts ready..."))
#'   while (!as.logical(contacts$has_more[1]) | nrow(contacts) < limit) {
#'     vidoffset <- contacts$vid_offset[nrow(contacts)]
#'     timeoffset <- contacts$time_offset[nrow(contacts)]
#'     URLi <- paste0(URL, "&vidOffset=", vidoffset, "&timeOffset=",timeoffset)
#'     contactsi <- bring_api(URLi, status = FALSE)
#'     if (!is.na(contactsi)) {
#'       contacts <- dplyr::bind_rows(contacts, contactsi)
#'       message(paste(nrow(contacts),"contacts ready..."))
#'     } else { break }
#'   }
#'   message(paste(nrow(contacts),"total contacts imported!"))
#'
#'   contacts$has_more <- contacts$vid_offset <- NULL
#'   colnames(contacts) <- gsub("contacts_properties_|_value", "", colnames(contacts))
#'   contacts <- contacts %>%
#'     dplyr::rename(credit_value = .data$credit,
#'                   #sale_value = .data$sale,
#'                   vehicle_commercial_value = .data$vehicle_commercial) %>%
#'     mutate_at(vars(contains('timestamp')), funs(hsdates(.))) %>%
#'     mutate_at(vars(contains('date')), funs(hsdates(.))) %>%
#'     mutate_at(vars(contains('addedAt')), funs(hsdates(.))) %>%
#'     mutate(date_of_birth = as.Date(.data$date_of_birth),
#'            identification_date = as.Date(.data$identification_date))
#'
#'   contacts$last_form_filled <- ifelse(
#'     grepl("146f9e71-b10e-4f7d-b0ef-8b65b122301f", contacts$contacts_form_submissions), "datos_personales",
#'     ifelse(grepl("851c6046-201a-4df9-8dab-8f2bf6a67a23", contacts$contacts_form_submissions), "pre_solicitud",
#'            ifelse(grepl("4442aa58-bd8a-4c3f-9710-cb3f6ecb0096", contacts$contacts_form_submissions), "solicitud",
#'                   ifelse(grepl("d9377981-7def-4489-9e2d-09b4b6ab755f", contacts$contacts_form_submissions), "documentos",
#'                          ifelse(grepl("e81e2002-b382-4c9d-90a4-75e6f6615132", contacts$contacts_form_submissions), "lead_soat",
#'                                 ifelse(grepl("4533ec08-dc7d-4c9e-8b5d-f521b6e6f28c", contacts$contacts_form_submissions), "blog_suscripcion",
#'                                        ifelse(grepl("7c8b4b79-f851-4f86-a00f-45d7dec89385", contacts$contacts_form_submissions), "contacto",
#'                                               NA)))))))
#'
#'   contacts$last_form_filled_credit <- ifelse(
#'     grepl("e81e2002-b382-4c9d-90a4-75e6f6615132", contacts$contacts_form_submissions), "0_lead_soat",
#'     ifelse(grepl("146f9e71-b10e-4f7d-b0ef-8b65b122301f", contacts$contacts_form_submissions), "1_datos_personales",
#'            ifelse(grepl("851c6046-201a-4df9-8dab-8f2bf6a67a23", contacts$contacts_form_submissions), "2_pre_solicitud",
#'                   ifelse(grepl("4442aa58-bd8a-4c3f-9710-cb3f6ecb0096", contacts$contacts_form_submissions), "3_solicitud",
#'                          ifelse(grepl("d9377981-7def-4489-9e2d-09b4b6ab755f", contacts$contacts_form_submissions), "4_documentos",
#'                                 NA)))))
#'
#'   contacts$step_done <- ifelse(
#'     !is.na(contacts$file_identification) |
#'       !is.na(contacts$files_labour_certificate) |
#'       !is.na(contacts$files_payroll), "documentos",
#'     ifelse(!is.na(contacts$credit_installments), "solicitud",
#'            ifelse(!is.na(contacts$economic_activity), "pre_solicitud",
#'                   ifelse(!is.na(contacts$firstname), "datos_personales", "lead"))))
#'
#'   contacts <- contacts %>%
#'     select(-.data$contacts_identity_profiles, -.data$contacts_profile_url,
#'            -.data$contacts_profile_token, -.data$contacts_portal_id,
#'            -.data$contacts_canonical_vid, -.data$contacts_is_contact,
#'            -.data$contacts_merge_audits, -.data$contacts_form_submissions,
#'            -contains("files_payroll"), -contains("file_identification"),
#'            -contains("files_labour")) %>%
#'     select(.data$contacts_vid, .data$step_done, one_of(properties), everything()) %>%
#'     dplyr::rename(., "vid" = "contacts_vid")
#'
#'   # # One contact
#'   # API <- "https://api.hubapi.com/contacts/v1/contact/vid/1016501/profile?"
#'   # hapikey <- paste0("hapikey=", credentials$token)
#'   # URL <- paste0(API, hapikey)
#'   # get <- httr::GET(url = URL)
#'   # message(paste0("Status: ", ifelse(get$status_code == 200, "OK", "ERROR")))
#'   # char <- rawToChar(get$content)
#'   # import <- jsonlite::fromJSON(char)
#'   # last_property <- data.frame(lapply(import$properties, `[[`, 1)) %>%
#'   #   mutate_at(vars(contains('timestamp')), funs(hsdates(.))) %>%
#'   #   mutate_at(vars(contains('date')), funs(hsdates(.)))
#'   # colnames(last_property)[!colnames(last_property) %in% properties]
#'
#'   # Final format
#'   contacts <- type.convert(contacts)
#'   posixs <- vector2text(c("timestamp","addedAt","_date", "createdate", "lastmodifieddate"), sep="|", quotes = F)
#'   dates <- vector2text(c("date_of_birth","identification_date"), sep="|", quotes = F)
#'   contacts <- contacts %>%
#'     mutate_at(vars(grep(posixs, colnames(.))), funs(as.POSIXct)) %>%
#'     mutate_at(vars(grep(dates, colnames(.))), funs(as.Date))
#'
#'   return(contacts)
#'
#' }

####################################################################
#' Download and plot daily downloads of CRAN packages
#'
#' Download daily downloads stats from CRAN for any package, and plot.
#' It can also be used as an auxiliary function to plot
#' (\code{cranlogs::cran_downloads}) results.
#'
#' @param input Character vector with package names or data.frame product of
#' \code{cranlogs::cran_downloads}.
#' @param from,to Dates. Range of dates to fetch downloads metrics.
#' @param type Character. Any of: "daily" or "total".
#' @param plot Boolean. Create a plot?
#' @examples
#' \donttest{
#' cran_logs(c("lares", "dplyr"), from = "2021-05-31")
#' }
#' @return List with data.frame and plot if \code{plot=TRUE}.
#' @export
cran_logs <- function(input = "lares",
                      from = Sys.Date() - 31,
                      to = Sys.Date() - 1,
                      type = "daily",
                      plot = TRUE) {
  if (!haveInternet()) {
    message("No internet connetion...")
    invisible(NULL)
  } else {
    # Fetch the data
    if (is.vector(input)) {
      check_opts(type, c("daily", "total"))
      base <- sprintf("https://cranlogs.r-pkg.org/downloads/%s", type)
      dates <- sprintf("%s:%s", as.character(from), as.character(to))
      packages <- paste(unique(input), collapse = ",")
      url <- paste(base, dates, packages, sep = "/")
      scrap <- content(GET(url), encoding = "UTF-8")
      if (!"downloads" %in% names(scrap[[1]])) {
        warning("Site currently unavailable")
        invisible(url)
      } else {
        cran.df <- bind_rows(scrap)
        if (type == "daily") {
          cran.df <- cran.df %>%
            mutate(
              date = unlist(lapply(.data$downloads, function(x) x[[1]])),
              count = unlist(lapply(.data$downloads, function(x) x[[2]]))
            ) %>%
            select(.data$date, .data$count, .data$package) %>%
            mutate(date = as.Date(.data$date, origin = "1970-01-01")) %>%
            arrange(desc(.data$date))
        }
      }
    } else {
      check_opts(colnames(input), c("date", "count", "package"), input_name = "column names")
      cran.df <- as_tibble(input)
    }
    package <- unique(cran.df$package)
    if (nrow(cran.df) >= length(package) * 2 && plot && type == "daily") {
      input <- cran.df %>%
        arrange(.data$date) %>%
        group_by(.data$package) %>%
        mutate(
          cum = cumsum(.data$count),
          package = sprintf("%s\n(%s)", .data$package, formatNum(max(.data$cum), abbr = TRUE)),
          MN = mean(.data$count, na.rm = TRUE)
        )
      dMean <- input %>%
        group_by(.data$package) %>%
        summarise(
          MN = mean(.data$count, na.rm = TRUE),
          date = min(.data$date, na.rm = TRUE),
          MN_label = formatNum(.data$MN, 2, abbr = TRUE)
        )
      plot <- ggplot(input, aes(x = .data$date)) +
        geom_line(aes(y = .data$count), colour = lares_pal("simple")[1], alpha = 0.8) +
        theme_lares(legend.position = "none", pal = 2, panel_colour = "grey95") +
        labs(
          title = glued("CRAN: {x} downloads", x = v2t(package, and = "and")),
          subtitle = sprintf("(%s to %s)", min(input$date), max(input$date)),
          x = "Download Date", y = "Dailly Downloads",
        ) +
        facet_grid(.data$package ~ ., scales = "free") +
        geom_line(aes(y = .data$MN), linetype = "dashed") +
        geom_text(
          data = dMean, aes(x = .data$date, y = .data$MN, label = .data$MN_label),
          vjust = -0.5, hjust = 1, size = 2.7
        ) +
        scale_y_abbr() +
        expand_limits(y = 0)
      list(df = as_tibble(cran.df), plot = plot)
    } else {
      cran.df
    }
  }
}

# Auxiliary constant values
# META_GRAPH_URL <- "https://graph.intern.facebook.com"
META_GRAPH_URL <- "https://graph.facebook.com"
META_API_VER <- "v21.0"

####################################################################
#' Paginate and Process Facebook's API Results
#'
#' Process and paginate raw results from Facebook's API, result of
#' querying the API with \code{httr::GET} or by passing an API link.
#'
#' @family API
#' @family Meta
#' @inheritParams tic
#' @inheritParams cache_write
#' @param input GET's output object (response) or link (character).
#' @param paginate Boolean or integer. Run through all paginations? If set
#' to \code{FALSE}, only the first one will be processed. If set to any other
#' integer value, will process the first N paginations.
#' @param sleep Numeric value. How much should each loop wait until until running
#' the next pagination query?
#' @return data.frame with un-nested processed results or NULL if no results found.
#' @export
fb_process <- function(input, paginate = TRUE, sleep = 0, quiet = FALSE, ...) {
  if (is.character(input)) {
    if (is_url(input)) input <- httr::GET(input)
  }
  if (!"response" %in% class(input)) {
    stop("You must provide a response object (GET's output) or a valid URL")
  }
  tic(paste0("fb_process_", input$url))
  import <- content(input, encoding = "json")

  # Show and return error (and help is token expired)
  if ("error" %in% names(import)) {
    if (!quiet) message(paste("API ERROR:", import$error$message))
    import$url <- input$url
    if ("error" %in% names(import)) {
      if (grepl("expired", import$error$message) & !quiet) {
        message(paste(
          "You must be logged in to your Facebook account and refresh/get a new token:",
          "\n1. Within the Graph API Explorer, select an Application.",
          "\n2. Go to Get Token and select the Page Acces Token needed.",
          "Note that there are navigation arrows if there are lots of accounts.",
          "\n3. Copy and use the Acces Token created."
        ))
        browseURL("https://developers.facebook.com/tools/explorer")
      }
    }
    invisible(import)
  }

  # Check for no-data (user might think there was an error on GET request - warn him!)
  if (length(import[[1]]) == 0) {
    message("NO DATA: No results found for this request")
    invisible(import)
  } else {
    # Deal with GET + response vs jsonlite
    out <- fromJSON(toJSON(import))
    if (all(lapply(out, length) == 1)) {
      out
    } else {
      # First pagination
      results <- list()
      i <- 1
      results[[i]] <- list_flattener(out[[1]], first = TRUE)
      if (!quiet) show_pag_status(results, i, sleep, quiet)

      # Following iterations
      if (exists("paging", out) && as.integer(paginate) >= 1) {
        if (exists("next", out$paging)) {
          i <- i + 1
          out <- fromJSON(out$paging[["next"]])
          results[[i]] <- list_flattener(out$data, i)
          if (!quiet) show_pag_status(results, i, sleep, quiet)
          # Re-run first iteration and overwrite as everything MUST match to bind
          if (i == 2) {
            out <- fromJSON(out$paging[["previous"]])
            results[[1]] <- list_flattener(out$data, i)
            if (!quiet) show_pag_status(results, i, sleep, quiet)
          }
          while (exists("next", out$paging) && (i < as.integer(paginate) || isTRUE(paginate))) {
            i <- i + 1
            res <- try(
              {
                out <- fromJSON(out$paging[["next"]])
                results[[i]] <- list_flattener(out$data, i)
                if (!quiet) show_pag_status(results, i, sleep, quiet)
              },
              silent = TRUE
            )
            if (inherits(res, "try-error")) {
              warning(paste0("Returning partial results given last pagination (", i, ") returned error"))
              break
            }
          }
        }
      }
      done <- bind_rows(results)
      ret <- suppressMessages(type.convert(
        done,
        numerals = "no.loss", as.is = TRUE
      )) %>%
        mutate_at(vars(contains("date")), list(as.Date)) %>%
        mutate_at(vars(contains("id")), list(as.character)) %>%
        mutate_at(vars(contains("url")), list(as.character)) %>%
        mutate_at(vars(contains("name")), list(as.character)) %>%
        as_tibble()

      # Save last pagination and if it returned complete results
      attr(ret, "paging_done") <- !("next" %in% names(out[["paging"]]))
      attr(ret, "paging") <- out[["paging"]]
      toc(paste0("fb_process_", input$url), quiet = quiet, ...)
      ret
    }
  }
}

list_flattener <- function(x, i = 1, first = FALSE) {
  bind_rows(x) %>%
    mutate(get_id = paste(i - !first, row_number(), sep = "-")) %>%
    select(.data$get_id, everything()) %>%
    tidyr::unnest(everything(), names_sep = ".") %>%
    dplyr::as_tibble()
}

show_pag_status <- function(results, i = 1, sleep = 0, quiet = FALSE) {
  if (!quiet) {
    if (i == 1 & !"next" %in% names(results[["paging"]])) {
      # We can skip this message: no paginations
    } else {
      flush.console()
      total <- sum(unlist(lapply(results, nrow)))
      cat(paste(sprintf("\r>>> Pagination imported: %s | Total rows: %s", i, total)))
    }
  }
  Sys.sleep(sleep)
}


####################################################################
#' Facebook API Report Status Check
#'
#' This returns all available FB insights per day including any given
#' breakdown to the specified report level, and place into a data frame.
#' For more information on Ad Insights' API, go to the original
#' \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{documentaion}.
#'
#' @family API
#' @family Meta
#' @inheritParams fb_insights
#' @inheritParams fb_process
#' @param report_run_id Integer. Report ID to check status.
#' @param live Boolean. Run until status report is finished?
#' @param sleep Boolean. If \code{live=TRUE}, then how many seconds should
#' we wait until next check?
#' @return List with API status results.
#' @examples
#' \dontrun{
#' token <- "YOURTOKEN"
#' report_run_id <- "123456789"
#' fb_report_check(token, report_run_id, live = TRUE, quiet = FALSE)
#' }
#' @export
fb_report_check <- function(token, report_run_id, api_version = NULL,
                            live = FALSE, sleep = 10, quiet = FALSE) {
  keep_running <- TRUE
  api_version <- ifelse(is.null(api_version), META_API_VER, api_version)
  URL <- glued("{META_GRAPH_URL}/{api_version}/{report_run_id}")
  while (isTRUE(keep_running)) {
    async_s <- httr::GET(url = URL, query = list(access_token = token))
    resp <- httr::content(async_s, encoding = "json")
    keep_running <- isTRUE(resp$async_percent_completion < 100 && isTRUE(live))
    if (!quiet) {
      flush.console()
      cat(sprintf("\rStatus: %s (%.0f%%)\n", resp$async_status, resp$async_percent_completion))
    }
    if (keep_running) Sys.sleep(sleep)
  }
  resp
}


####################################################################
#' Facebook Insights API
#'
#' This returns all available FB insights per day including any given
#' breakdown to the specified report level, and place into a data frame.
#' For more information on Ad Insights' API, go to the original
#' \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{documentaion}.
#'
#' @family API
#' @family Meta
#' @inheritParams cache_write
#' @param token Character. Valid access token with sufficient privileges. Visit the
#' \href{https://developers.facebook.com/tools/explorer}{Facebook API Graph Explorer}
#' to acquire one.
#' @param which Character vector. This is the accounts, campaigns, adsets,
#' or ads IDs to be queried. Remember: if \code{report_level = "account"}, you must
#' start the ID with \code{act_}.
#' @param start_date,end_date Character. The first and last full day to report, in the
#' format \code{"YYYY-MM-DD"}.
#' @param time_increment Character. Group by months ("monthly"),
#' everything together ("all_days") or an integer per days [1-90].
#' Default: each day separately (i.e. "1").
#' @param report_level Character. One of "ad", "adset", "campaign", or "account"
#' @param ad_object Character. One of: "insights" (default), "adsets", ...
#' @param breakdowns Character Vector. One or more of breakdowns for
#' segmentation results. Set to NA for no breakdowns
#' @param fields Character, json format. Leave \code{NA} for default fields OR
#' \code{NULL} to ignore.
#' @param filtering List. Each filter will be a list containing "field",
#' "operator", and "value". Read more about the operators in the official
#' \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights}{docs}.
#' Example: \code{dplyr::tibble(field = "country", operator = "IN", value = list("PE")))}.
#' @param limit Integer. Query limit by pagination.
#' @param api_version Character. Facebook API version.
#' @param process Boolean. Process GET results to a more friendly format?
#' @param async Boolean. Run an async query. When set to \code{TRUE}, instead of making
#' a GET query, it'll run a POST query and will return a report run ID.
#' @return data.frame with un-nested processed results if \code{process=TRUE} or
#' raw API results as list when \code{process=FALSE}.
#' @examples
#' \dontrun{
#' token <- "YOURTOKEN"
#' which <- "act_20846447"
#'
#' # Platforms' Insights: all ad-sets platforms of "which" account,
#' # aggregated, for the last 30 days
#' platforms <- fb_insights(
#'   token, which,
#'   start_date = Sys.Date() - 30,
#'   time_increment = "all_days",
#'   report_level = "adset",
#'   fields = c(
#'     "account_name",
#'     "adset_id",
#'     "adset_start",
#'     "adset_end"
#'   ),
#'   breakdowns = c(
#'     "publisher_platform",
#'     "platform_position",
#'     "device_platform"
#'   )
#' )
#'
#' # Daily results for all campaigns of "which" account,
#' # with custom performance fields with no breakdowns.
#' insights_adset <- fb_insights(
#'   token, which,
#'   time_increment = "1",
#'   report_level = "campaign",
#'   fields = c(
#'     "adset_id",
#'     "reach",
#'     "frequency",
#'     "spend",
#'     "cpm",
#'     "objective",
#'     "optimization_goal"
#'   )
#' )
#' }
#' @export
fb_insights <- function(token,
                        which,
                        start_date = Sys.Date() - 7,
                        end_date = Sys.Date(),
                        time_increment = "1",
                        report_level = "campaign",
                        ad_object = "insights",
                        breakdowns = NA,
                        fields = NA,
                        filtering = NULL,
                        limit = 100,
                        api_version = NULL,
                        process = TRUE,
                        async = FALSE,
                        ...) {
  set_config(config(http_version = 0))
  check_opts(report_level, c("ad", "adset", "campaign", "account"))

  if (isTRUE(is.na(fields[1]))) {
    fields <- c(
      "campaign_name", "campaign_id",
      "objective", "adset_id",
      "adset_name", "ad_id",
      "ad_name", "impressions",
      "spend", "cpm", "ctr", "cpc",
      "reach", "clicks", "unique_clicks",
      "unique_ctr", "cost_per_unique_click"
    )
  }

  aux <- v2t(which, quotes = FALSE)
  api_version <- ifelse(is.null(api_version), META_API_VER, api_version)
  URL <- glued("{META_GRAPH_URL}/{api_version}/{aux}/{ad_object}")
  api_verb <- ifelse(async, httr::POST, httr::GET)

  # Call insights API
  import <- api_verb(
    URL,
    query = list(
      time_range = paste0('{\"since\":\"', start_date, '\",\"until\":\"', end_date, '\"}'),
      level = report_level,
      fields = if (length(fields) > 1) {
        vector2text(fields, sep = ",", quotes = FALSE)
      } else {
        fields
      },
      breakdowns = if (!is.na(breakdowns[1])) {
        vector2text(breakdowns, sep = ",", quotes = FALSE)
      } else {
        NULL
      },
      filtering = ifelse(!is.null(filtering), toJSON(filtering), ""),
      time_increment = time_increment,
      limit = as.character(limit),
      access_token = token
    ),
    encode = "json"
  )

  if (!process) {
    invisible(import)
  } else {
    fb_process(import, ...)
  }
}

####################################################################
#' Facebook Reach and Frequency API
#'
#' Create or query reach and frequency predictions using Facebook's
#' Reach and Frequency API. For more information on the API and its parameters, go to the
#' \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{original documentaion}.
#'
#' @family API
#' @family Meta
#' @inheritParams fb_insights
#' @param ad_account Character. Ad Account. Remember to start with \code{act_}. If you
#' use the \code{prediction} argument, no need to provide this parameter.
#' @param prediction Integer. Prediction ID if you already created the prediction
#' and wish to query the curve's data. As this prediction already exists, the
#' rest of arguments of this function will be ignored.
#' @param objective Character. Any of: "BRAND_AWARENESS", "LINK_CLICKS", "POST_ENGAGEMENT",
#' "MOBILE_APP_INSTALLS", "CONVERSIONS", "REACH", or "VIDEO_VIEWS".
#' @param days Integer. Amount of days for your campaign's predictions.
#' @param budget Integer. The budget in the Ad Account currency in cents.
#' @param destination_ids Integer vector. Page ID and/or Instagram Account ID.
#' @param countries Character vector. Country's acronyms.
#' @param frequency_cap Integer. Frequency cap over all the campaign duration.
#' @param prediction_mode Integer. "1" for predicting Reach by providing budget,
#' "2" is for predicting Budget given a specific Reach.
#' @param curve Boolean. Return curve data? If not, only prediction will be created.
#' @param ... Additional parameters passed to target specs.
#' @return data.frame with un-nested processed results if \code{process=TRUE} or
#' raw API results as list when \code{process=FALSE}.
#' @examples
#' \dontrun{
#' token <- "YOURTOKEN"
#' account_id <- "act_20846447"
#'
#' # BASIC 1: Create and return data for a new prediction
#' basic1 <- fb_rf(token, account_id, destination_ids = 187071108930, countries = "AR")
#'
#' # BASIC 2: Fetch data for an existing prediction ID
#' basic2 <- fb_rf(token, account_id, prediction = 6317720998974)
#'
#' # ADVANCED (Fully custom prediction)
#' advanced <- fb_rf(token, account_id,
#'   objective = "REACH",
#'   days = 28,
#'   budget = 2000000,
#'   destination_ids = c(187071108930, 1142958119078556),
#'   age_min = 15,
#'   age_max = 65,
#'   genders = 2,
#'   countries = "MX",
#'   publisher_platforms = c(
#'     "facebook",
#'     "instagram",
#'     #' audience_network',
#'     "messenger"
#'   ),
#'   # interests_ids = NA,
#'   facebook_positions = c(
#'     "feed",
#'     #' instant_article',
#'     "marketplace",
#'     "video_feeds",
#'     "story",
#'     "search",
#'     "instream_video"
#'   ),
#'   instagram_positions = c(
#'     "stream",
#'     "story",
#'     "explore"
#'   ),
#'   # audience_network_positions = c(
#'   #  'classic',
#'   #  'instream_video')
#'   messenger_positions = c(
#'     "messenger_home",
#'     "sponsored_messages",
#'     "story"
#'   ),
#'   device_platforms = c(
#'     "mobile",
#'     "desktop"
#'   )
#' )
#' }
#' @export
fb_rf <- function(token,
                  ad_account = NA,
                  prediction = NA,
                  objective = "REACH",
                  days = 28,
                  budget = 2e6,
                  destination_ids = NA,
                  countries = "MX",
                  frequency_cap = 8,
                  prediction_mode = 1,
                  curve = TRUE,
                  api_version = NULL,
                  process = TRUE,
                  ...) {
  set_config(config(http_version = 0))

  if (is.na(prediction[1])) {
    message(">>> Creating prediction...")

    # Transform target_spec from list to JSON format
    target_spec <- list(
      geo_locations = if (length(countries) > 0) {
        list(countries = countries)
      } else {
        NULL
      },
      ...
    )
    ts <- toJSON(target_spec)

    # Force some values that must be integers
    if ("age_min" %in% names(target_spec)) {
      ts <- gsub(paste0("\\[", target_spec[["age_min"]], "\\]"), target_spec[["age_min"]], ts)
    }
    if ("age_max" %in% names(target_spec)) {
      ts <- gsub(paste0("\\[", target_spec[["age_max"]], "\\]"), target_spec[["age_max"]], ts)
    }

    # Create a prediction (returns ID if successful)
    daysecs <- 60 * 60 * 24
    current_time <- as.integer(Sys.time())
    api_version <- ifelse(is.null(api_version), META_API_VER, api_version)
    prediction <- content(POST(
      glued("{META_GRAPH_URL}/{api_version}/{ad_account}/reachfrequencypredictions"),
      query = list(
        start_time = current_time + daysecs,
        end_time = min(
          (current_time + daysecs + (days) * daysecs),
          (current_time + daysecs + (89 * daysecs))
        ),
        objective = toupper(objective),
        frequency_cap = frequency_cap,
        prediction_mode = prediction_mode,
        budget = as.integer(budget),
        destination_ids = toJSON(as.character(destination_ids)),
        target_spec = ts,
        access_token = token
      )
    ), encode = "json")[[1]]

    if ("message" %in% names(prediction)) {
      message(paste(
        prediction$message,
        prediction$error_user_title,
        prediction$error_user_msg,
        sep = "\n"
      ))
      invisible(prediction)
    } else {
      message(glued("Prediction created: {prediction}"))

      if (curve) {
        if (length(prediction) > 1) {
          stop("Please, provide only 1 prediction per query")
        }
        api_version <- ifelse(is.null(api_version), META_API_VER, api_version)
        curves <- GET(
          glued("{META_GRAPH_URL}/{api_version}/{prediction}"),
          query = list(
            fields = "curve_budget_reach",
            access_token = token
          )
        )

        if (process) {
          this <- fb_process(curves, ...)
          curves <- this %>%
            select(-one_of(zerovar(.))) %>%
            mutate(
              budget = .data$budget / 100,
              frequency = .data$impression / .data$reach,
              weekly_frequency = .data$frequency / (days / 7),
              cpm = 1000 * .data$budget / .data$impression,
              prediction = prediction
            )
          curves <- curves %>%
            as.matrix() %>%
            data.frame() %>%
            as_tibble() %>%
            mutate_at(colnames(.)[-c(1, ncol(.))], as.numeric)
        }

        attr(curves, "prediction") <- prediction
        class(curves) <- c(class(curves), "fb_rf")
        curves
      } else {
        prediction
      }
    }
  }
}


####################################################################
#' Facebook Ad Accounts
#'
#' This returns all ad accounts for a FB Business Account FB.
#' For more information on Ad Insights' API, go to the
#' \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{original documentaion}
#'
#' @family API
#' @family Meta
#' @inheritParams fb_insights
#' @inheritParams fb_process
#' @param business_id Character. Business ID.
#' @param type Character vector. Values: owned, client.
#' @return data.frame with un-nested processed results fetched with API.
#' @examples
#' \dontrun{
#' # Query all accounts (owned and with permissions) of a Business ID
#' accounts <- fb_accounts(YOURTOKEN, YOURBUSINESS)
#' }
#' @export
fb_accounts <- function(token,
                        business_id = "904189322962915",
                        type = c("owned", "client"),
                        limit = 100,
                        api_version = NULL,
                        ...) {
  set_config(config(http_version = 0))
  api_version <- ifelse(is.null(api_version), META_API_VER, api_version)
  output <- NULL

  # Select which type of ad accounts
  type <- paste0(type, "_ad_accounts")

  for (i in seq_along(type)) {
    message(paste(">>> Fetching", type[i]))
    URL <- paste(META_GRAPH_URL, api_version, business_id, type[i], sep = "/")
    continue <- TRUE

    # Call insights
    import <- GET(
      URL,
      query = list(
        fields = "name,account_status,amount_spent,business_country_code",
        limit = as.character(limit),
        access_token = token
      ),
      encode = "json"
    )
    ret <- fb_process(import, quiet = TRUE, ...)

    if (inherits(ret, "data.frame")) {
      ret$type <- type[i]
      output <- bind_rows(output, ret)
    }
  }

  if (!inherits(output, "data.frame")) {
    invisible(NULL)
  } else {
    # Account status dictionary
    output <- mutate(output, account_status = case_when(
      account_status == "1" ~ "ACTIVE",
      account_status == "2" ~ "DISABLED",
      account_status == "3" ~ "UNSETTLED",
      account_status == "7" ~ "PENDING_RISK_REVIEW",
      account_status == "8" ~ "PENDING_SETTLEMENT",
      account_status == "9" ~ "IN_GRACE_PERIOD",
      account_status == "100" ~ "PENDING_CLOSURE",
      account_status == "101" ~ "CLOSED",
      account_status == "201" ~ "ANY_ACTIVE",
      account_status == "202" ~ "ANY_CLOSED"
    ))
    suppressMessages(type.convert(
      output,
      numerals = "no.loss", as.is = TRUE
    )) %>%
      arrange(desc(.data$amount_spent)) %>%
      as_tibble()
  }
}

####################################################################
#' Facebook's Long-Life User API Token
#'
#' Using a 1-hour generic user token you can generate a 60 day token.
#' You will need to have an App ID and App secret, and a valid token.
#' Generate a new valid User Token with the
#' \href{https://developers.facebook.com/tools/explorer}{API Graph}.
#'
#' More info: \href{https://developers.facebook.com/documentation/facebook-login/guides/access-tokens/get-long-lived/}{Long-Lived Access Tokens}
#'
#' @family API
#' @family Meta
#' @inheritParams fb_insights
#' @param app_id,app_secret Character. Application ID and Secret.
#' @param token Character. User token, created with
#' \href{https://developers.facebook.com/tools/explorer}{API Graph}
#' or with this same \code{fb_token()}'s token.
#' @return Character. String with token requested. If successful, it'll contain
#' an attribute called "expiration" with date and time of expiration.
#' @export
fb_token <- function(app_id, app_secret, token, api_version = NULL) {
  api_version <- ifelse(is.null(api_version), META_API_VER, api_version)
  link <- glued(paste0(
    "{META_GRAPH_URL}/", api_version, "/oauth/access_token?",
    "grant_type=fb_exchange_token&client_id=%s&client_secret=%s&fb_exchange_token=%s"
  ))
  linkurl <- sprintf(link, app_id, app_secret, token)
  ret <- content(GET(linkurl), encode = "json")
  if ("access_token" %in% names(ret)) {
    expiration <- Sys.time() + ret$expires_in
    message(sprintf("Your token was created succesfully. It will expire on %s", expiration))
    output <- ret$access_token
    attr(output, "expiration") <- expiration
    output
  } else {
    message(ret$error$message)
    ret
  }
}

####################################################################
#' Facebook Creatives API
#'
#' For more information: \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/}{Marketing API}
#'
#' @family API
#' @family Meta
#' @inheritParams fb_process
#' @inheritParams fb_insights
#' @return data.frame with un-nested processed results if \code{process=TRUE} or
#' raw API results as list when \code{process=FALSE}.
#' @examples
#' \dontrun{
#' token <- YOURTOKEN
#' account <- act_ADACCOUNT
#'
#' # Query all creatives for "which" (account in this case)
#' creatives <- fb_creatives(token, account)
#' }
#' @export
fb_creatives <- function(token, which,
                         api_version = NULL,
                         process = TRUE,
                         ...) {
  set_config(config(http_version = 0))
  api_version <- ifelse(is.null(api_version), META_API_VER, api_version)

  fields <- c(
    "account_id", "object_type", "name", "status", "campaign_id",
    "call_to_action_type", "image_url", "thumbnail_url"
  )
  link <- glued(paste0(
    "{META_GRAPH_URL}/%s/%s/adcreatives?",
    "grant_type=fb_exchange_token",
    "&fields=%s",
    "&access_token=%s"
  ))
  linkurl <- sprintf(
    link, api_version, which,
    vector2text(fields, sep = ",", quotes = FALSE),
    token
  )
  import <- GET(linkurl)
  if (!process) {
    import
  } else {
    fb_process(import, ...)
  }
}

####################################################################
#' Facebook Ads API
#'
#' This returns all available FB ads for any account, campaign, or ad set id.
#' For more information on Ad' API, go to the
#' \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/reference/adgroup/}{original documentaion}
#'
#' This function was based on FBinsightsR.
#'
#' @family API
#' @family Meta
#' @inheritParams fb_insights
#' @inheritParams fb_process
#' @return data.frame with un-nested processed results if \code{process=TRUE} or
#' raw API results as list when \code{process=FALSE}.
#' @examples
#' \dontrun{
#' token <- YOURTOKEN
#' account <- act_ADACCOUNT
#'
#' # Query all ads for "which" (account) with results in the last 10 days
#' ads <- fb_ads(token, account, start_date = Sys.Date() - 10)
#' }
#' @export
fb_ads <- function(token,
                   which,
                   start_date = Sys.Date() - 31,
                   end_date = Sys.Date(),
                   fields = NA,
                   api_version = NULL,
                   process = TRUE,
                   ...) {
  set_config(config(http_version = 0))
  api_version <- ifelse(is.null(api_version), META_API_VER, api_version)

  if (is.na(fields[1])) {
    fields <- paste(
      "account_id, campaign, campaign_id, objective, adset_id, id,",
      "name, source_ad, cpm, spend, recommendations"
    )
  }

  import <- GET(
    glued("{META_GRAPH_URL}/{api_version}/{which}/ads"),
    query = list(
      time_range = paste0('{\"since\":\"', start_date, '\",\"until\":\"', end_date, '\"}'),
      fields = if (length(fields) > 1) {
        vector2text(fields, sep = ",", quotes = FALSE)
      } else {
        fields
      },
      access_token = token
    ),
    encode = "json"
  )

  if (!process) {
    import
  } else {
    fb_process(import, ...)
  }
}

####################################################################
#' Progressive Status Bar (Loading)
#'
#' This function lets the user view a progressbar for a 'for' loop.
#'
#' @family Tools
#' @inheritParams h2o_automl
#' @param run Iterator. for loop or an integer with the current loop number.
#' Start with 1 preferibly
#' @param max.run Number. Maximum number of loops
#' @param label String. With additionaly information to be printed
#' at the end of the line. The default is \code{run}.
#' @param msg Character. Finish message.
#' @param type Character. Loading type style: equal, domino, sword, filled.
#' @param start_time POSIXct. Start time to consider. If NA, then
#' when first iteration starts will be set as start time. Useful
#' for when first iteration is showed as done but started a few
#' seconds/minutes ago.
#' @param multiples Integer. Only print when multiples of N (to avoid)
#' wasting resources on fast and lots of iterations.
#' @return No return value, called for side effects.
#' @examples
#' for (i in 1:9) {
#'   statusbar(i, 9, multiples = 2)
#'   Sys.sleep(0.3)
#' }
#' @export
statusbar <- function(run = 1, max.run = 100, label = run, msg = "",
                      type = Sys.getenv("LARES_STATUSBAR"),
                      start_time = NA, multiples = 1, alarm = FALSE) {
  if (run == 1 && is.na(start_time)) tic("startclock")
  if (!is.na(start_time)) tic("startclock", start = start_time)

  if (multiples > 1) {
    if (!run %% multiples == 0 && run != max.run) {
      NULL
    }
  } else {
    if (length(run) > 1 && !is.numeric(run)) {
      stop("Parameter 'run' must be a numerical value!")
    }
    if (length(max.run) == 0 && !is.numeric(run)) {
      stop("Parameter 'max.run' needs to be greater than 0!")
    }

    width.set <- getOption("width") * 0.5
    width.labs <- getOption("width") - width.set - 15

    if (length(max.run) > 1) {
      percent <- which(run == max.run) / length(max.run)
    } else {
      percent <- run / max.run
    }

    if (type == "domino") syms <- list(first = "|", middle = "/", last = "_")
    if (type == "equal") syms <- list(first = " ", middle = "=", last = "=")
    if (type == "sword") syms <- list(first = " ", middle = ">", last = ":")
    if (type == "filled") syms <- list(first = " ", middle = "\u2588", last = "\u2588")

    percent.step <- trunc(percent * width.set, 5)
    part_done <- paste0(rep(syms$last, percent.step), collapse = "")
    part_middle <- if (percent == 1) syms$last else syms$middle
    part_left <- paste0(rep(syms$first, width.set - percent.step), collapse = "")
    perc <- signif(percent * 100, 3)
    if (percent == 1 && msg != "") label <- msg
    parts <- ifelse(label != "", paste("|", label), label)
    parts <- stringr::str_pad(parts, width = width.labs, pad = " ", side = "right")

    now <- toc("startclock", quiet = TRUE, type = "clock")$time
    progress <- glued("{now} [{part_done}{part_middle}{part_left}] {perc}% {parts}")
    if (percent == 1) progress <- paste(progress, "\n")
    cat("\r", progress)
    flush.console()

    if (alarm) {
      try_require("beepr", stop = FALSE)
      try(beep())
    }
  }
}

####################################################################
#' Download Historical Currency Exchange Rate
#'
#' This function lets the user download historical currency exchange
#' rate between two currencies.
#'
#' @family Currency
#' @inheritParams cache_write
#' @param currency_pair Character. Which currency exchange do you
#' wish to get the history from? i.e, USD/COP, EUR/USD...
#' @param from Date. From date
#' @param to Date. To date
#' @param fill Boolean. Fill weekends and non-quoted dates with
#' previous values?
#' @return data.frame. Result of fetching online data for \code{currency_pair}
#' grouped by date.
#' @examples
#' \donttest{
#' # For today (or any one single date)
#' get_currency("USD/COP", from = Sys.Date())
#' # For multiple dates
#' get_currency("EUR/USD", from = Sys.Date() - 7, fill = TRUE)
#' }
#' @export
get_currency <- function(currency_pair,
                         from = Sys.Date() - 99,
                         to = Sys.Date(),
                         fill = FALSE, ...) {
  if (!haveInternet()) {
    message("No internet connetion...")
    invisible(NULL)
  } else {
    try_require("quantmod")

    string <- paste0(toupper(cleanText(currency_pair)), "=X")

    if (is.na(from) || is.na(to)) {
      stop("You must insert a valid date")
    }

    from <- as.Date(from)
    to <- as.Date(to)

    if (from == to) to <- from + 1
    if (to > Sys.Date()) to <- Sys.Date()

    x <- try(data.frame(suppressWarnings(getSymbols(
      string,
      env = NULL,
      from = from,
      to = to,
      ...
    ))))
    if ("try-error" %in% class(x)) {
      warning(x)
      x
    }
    dates <- as.Date(gsub("\\.", "\\-", gsub("X", "", rownames(x))))
    rate <- data.frame(date = dates, rate = x[, 1])
    # Sometimes, the last date is repeated
    if (tail(rate$date, 1) == tail(rate$date, 2)[1] && nrow(rate) > 1) {
      rate <- rate[-nrow(rate), ]
    }

    if (fill) {
      rate <- data.frame(date = as.character(
        as.Date(as.Date(from):Sys.Date(), origin = "1970-01-01")
      )) %>%
        left_join(rate %>% mutate(date = as.character(date)), "date") %>%
        tidyr::fill(rate, .direction = "down") %>%
        tidyr::fill(rate, .direction = "up") %>%
        mutate(date = as.Date(date)) %>%
        filter(date >= as.Date(from))
    }
    rate
  }
}

#' ####################################################################
#' #' Get Google's Geodata given the Addresses
#' #'
#' #' This function lets the user obtain Google's Geodata on given addresses
#' #'
#' #' @family Geographical
#' #' @family Credentials
#' #' @family Google
#' #' @param address Character Vector. Addresses you wish to query
#' #' @param country Character. Default Country if known
#' #' @param index Character Vector. If you wish to keep an id on each
#' #' address, set this values
#' #' @param creds Character. Credential's user (see \code{get_creds()})
#' #' @param which Character. Which of the APIs should be used?
#' #' @export
#' geoAddress <- function(address, country = "Argentina", index = NA, creds = NA, which = "api_01") {
#'
#'   try_require("rlist")
#'
#'   message("API Documentation: https://developers.google.com/maps/documentation/geocoding/usage-and-billing")
#'   c <- get_credentials(from = "google_api", dir = creds)
#'   message("API Geocoding user: ", as.character(c[grepl(paste0("user_",right(which, 2)), names(c))]))
#'
#'   getGeoDetails <- function(address){
#'
#'     c <- get_credentials(from = "google_api", dir = creds)
#'     url <- "https://maps.google.com/maps/api/geocode/json?address="
#'     api <- as.character(c[grepl(which, names(c))]) # api_01 is my personal free API
#'     ctry <- paste0("&components=country:", country)
#'     url <- URLencode(paste(url, address, ctry, "&key=", api, sep = ""))
#'     x <- fromJSON(url, simplifyVector = FALSE)
#'
#'     # Return Na's if we didn't get a match:
#'     if (x$status != "OK") return(message(paste(x$status, x$error_message)))
#'
#'     # Else, extract what we need from the Google server reply into a dataframe:
#'
#'     # Note: We bring the first and most probable result or prefered country only
#'     if (length(x$results) > 1) {
#'       country_pref <- data.frame(
#'         is_country = as.integer(as.character(lapply(
#'           x$results, function(x) grep(paste(",",country), x)))))
#'       country_pref$row <- seq_len(nrow(country_pref))
#'       country_pref <- country_pref[!is.na(country_pref$is_country),]
#'       which_list <- min(country_pref$row)
#'       if (nrow(country_pref) == 0) which_list <- 1
#'     } else {
#'       which_list <- 1
#'     }
#'
#'     # Address_components (varies a lot depending on results)
#'     values <- data.frame(list.cbind(x$results[[which_list]]$address_components))[1,]
#'     names <- unlist(do.call(rbind, list.cbind(x$results[[which_list]]$address_components)[3,])[,1])
#'     if (length(colnames(values)) > 1) colnames(values) <- names
#'
#'     out <- data.frame(
#'       status = x$status,
#'       lon = x$results[[which_list]]$geometry$location$lng,
#'       lat = x$results[[which_list]]$geometry$location$lat,
#'       accuracy = x$results[[which_list]]$types[[1]],
#'       address_type = paste(x$results[[which_list]]$types, collapse = ','),
#'       address_formatted = x$results[[which_list]]$formatted_address,
#'       neighborhood = ifelse("neighborhood" %in% colnames(values),
#'                             toupper(as.character(values$neighborhood)), NA),
#'       postal_code = ifelse("postal_code" %in% colnames(values),
#'                            as.integer(as.character(values$postal_code)), NA),
#'       political = ifelse("political" %in% colnames(values),
#'                          toupper(as.character(values$political)), NA),
#'       country = ifelse("country" %in% colnames(values),
#'                        toupper(as.character(values$country)), NA))
#'     return(out)
#'   }
#'
#'   output <- NULL
#'
#'   for (i in seq_along(address)) {
#'     message(paste("Working on address", i, "of", length(address)))
#'     result <- getGeoDetails(address[i])
#'     if (length(result) > 1) {
#'       result$index <- i
#'       result$search <- address[i]
#'       if (!is.na(index)) result$id <- index[i]
#'       output <- rbind(output, result)
#'       message(paste("Found:", result$address_formatted))
#'     }
#'     done <- output
#'   }
#'   return(done)
#' }
#'
#'
#' ####################################################################
#' #' Get Colombia's Stratum given the Coordinates
#' #'
#' #' This function lets the user obtain Colombia's stratum given the coordinates of an address
#' #'
#' #' @family Geographical
#' #' @family Scrapper
#' #' @param lon Numeric Vector. Longitudes
#' #' @param lat Numeric Vector. Latitudes
#' #' @param label Character Vector. If you wish to keep an id on each address, set this values
#' #' @export
#' geoStratum <- function(lon, lat, label = NA) {
#'   # Manzanas: http://serviciosgis.catastrobogota.gov.co/arcgis/rest/services/social/estrato/MapServer?f=jsapi
#'   url <- URLencode(paste0(
#'     "http://serviciosgis.catastrobogota.gov.co/arcgis/rest/services/social/estrato/MapServer/0/query?",
#'     "where=1%3D1&text=&objectIds=&time=&distance=15&",
#'     "geometry=%7B%22x%22%3A+", lon, "%2C+%22y%22%3A+", lat,
#'     "%7D&geometryType=esriGeometryPoint&inSR=4686&spatialRel=esriSpatialRelIntersects&relationParam=&outFields=ESTRATO&",
#'     "returnGeometry=false&returnTrueCurves=false&maxAllowableOffset=&geometryPrecision=&outSR=&returnIdsOnly=false&",
#'     "returnCountOnly=false&orderByFields=&groupByFieldsForStatistics=&outStatistics=&returnZ=false&returnM=false&",
#'     "gdbVersion=&returnDistinctValues=false&resultOffset=&resultRecordCount=&",
#'     "returnExtentsOnly=false&datumTransformation=&parameterValues=&rangeValues=&f=pjson"))
#'   x <- fromJSON(url, simplifyVector = FALSE)
#'   if (length(x$features) > 0) {
#'     out <- data.frame(lon = lon, lat = lat, stratum = x$features[[1]]$attributes$ESTRATO)
#'     if (!is.na(label)) out$label <- label
#'     return(out)
#'   } else {
#'     message("Stratum not found for those coordinates!")
#'   }
#' }
#'
#'
#' ####################################################################
#' #' Check, Cross, and Plot Coordinates with Polygons
#' #'
#' #' This function checks a series of coordinates and return a join
#' #' with the information of each coordinate and its respective grid.
#' #' Note that the coords and shapes coordinates MUST have the same
#' #' lon/lat reference system for it to work succesfully.
#' #'
#' #' @family Geographical
#' #' @param coords Dataframe. Dataframe containing at least langitud
#' #' and latitud data
#' #' @param map SpatialPolygonsDataFrame or .shp directory
#' #' @param fix_coords Boolean. Transform and fix coordinates system?
#' #' @param plot Boolean. Return plot with coordinates inside the grid?
#' #' @param all Boolean. Include all coordinates in plot, i.e. only the
#' #' ones who are inside the grids?
#' #' @param alpha Numeric. Points transparency for the plot
#' #' @export
#' geoGrid <- function(coords, map, fix_coords = FALSE, plot = FALSE, all = FALSE, alpha = 0.3) {
#'
#'   try_require("methods")
#'   try_require("rgdal")
#'   try_require("sp")
#'
#'   if (!class(map)[1] == "SpatialPolygonsDataFrame") {
#'     message("Importing shapefile...")
#'     map <- readOGR(dsn = file.path(map))
#'   }
#'
#'   cols <- colnames(coords)
#'   if (sum(grepl("lon|lat", tolower(cols))) != 2) {
#'     stop("Your coords dataframe must contain longitude and latitude!")
#'   }
#'   cols[grep("lon",tolower(cols))] <- "longitude"
#'   cols[grep("lat",tolower(cols))] <- "latitude"
#'   colnames(coords) <- cols
#'   coordinates(coords) <- c("longitude", "latitude")
#'
#'   if (fix_coords) map <- spTransform(map, CRS("+proj=longlat +datum=WGS84"))
#'
#'   shapes_sample <- head(map@polygons[[2]]@Polygons[[1]]@coords)
#'   proj4string <- "+proj=utm +units=mm"
#'   project(shapes_sample, proj4string)
#'
#'   # The coords and shapes coordinates MUST have the same lon/lat reference system
#'   proj4string(coords) <- proj4string(map)
#'   inside.park <- !is.na(over(coords, as(map, "SpatialPolygons")))
#'   #inside.park <- !is.na(over(coords, as.SpatialPolygons.PolygonsList(map)))
#'
#'   # What fraction of coords are inside a shape?
#'   frac <- round(100*mean(inside.park), 2)
#'   fracmsg <- paste0("Fraction of coords inside the grid: ", frac,"%")
#'   message(fracmsg)
#'   cross <- over(coords, map)
#'   crossed <- which(!is.na(cross[,1]))
#'
#'   # Join data and return results
#'   results <- data.frame(coords, cross)
#'
#'   # Plot map and coordinates
#'   if (plot) {
#'     if (!all) {
#'       toplot <- results[crossed,]
#'     } else {
#'       toplot <- results
#'     }
#'     plot <- ggplot() +
#'       geom_point(data = toplot, aes(x = .data$longitude, y = .data$latitude),
#'                  colour = "deepskyblue2", alpha = alpha) +
#'       geom_polygon(data = map, aes(x = .data$long, y = .data$lat, group = .data$group),
#'                    colour = "black", fill = "white", alpha = 0)  +
#'       labs(title = "Coordinates & Grid",
#'            subtitle = fracmsg,
#'            x = "Longitude", y = "Latitude") +
#'       theme_bw()
#'     results <- list(results = results, plot = plot)
#'   }
#'   return(results)
#' }
#'
#'
#' ####################################################################
#' #' Plot Map or Shapefile
#' #'
#' #' This function returns a simple plot from a SpatialPolygonsDataFrame
#' #' imported or a .shp file
#' #'
#' #' @family Geographical
#' #' @param map SpatialPolygonsDataFrame or .shp directory
#' #' @param fix_coords Boolean. Transform and fix coordinates system?
#' #' @param title Character. Title for the plot
#' #' @param subtitle Character. Subtitle for the plot
#' #' @export
#' geoMap <- function(map, fix_coords = FALSE, title = NA, subtitle = NA) {
#'
#'   try_require("rgdal")
#'   try_require("sp")
#'
#'   if (!class(map)[1] == "SpatialPolygonsDataFrame") {
#'     message("Importing shapefile...")
#'     map <- readOGR(dsn = file.path(map))
#'   }
#'   if (fix_coords) {
#'     message("Fixing coordinates format...")
#'     map <- spTransform(map, CRS("+proj=longlat +datum=WGS84"))
#'   }
#'   plot <- ggplot() +
#'     geom_polygon(data = map, aes(x = .data$long, y = .data$lat, group = .data$group),
#'                  colour = "black", fill = "white", alpha = 0.1) +
#'     labs(x = "Latitude", y = "Longitude") +
#'     theme_bw()
#'
#'   if (!is.na(title)) plot <- plot + labs(title = title)
#'   if (!is.na(subtitle)) plot <- plot + labs(subtitle = subtitle)
#'
#'   return(plot)
#' }
#'
#'
#' ####################################################################
#' #' Convert from degrees to numeric coordinates [Deprecated]
#' #'
#' #' This function converts degrees (DMS) coordinates into numerical.
#' #' Note that the sign (S or W) should be assigned manually if needed.
#' #'
#' #' @family Calculus
#' #' @param coord Character vector. Coordinate in format c("DD MM SS")
#' #' @param sep Character. Separator
#' #' @examples
#' #' deg2num("11 12 12")
#' #' deg2num("00-00-00", sep = "-")
#' #' @export
#' deg2num <- function(coord, sep = " ") {
#'   z <- str_split_fixed(as.character(coord), sep, 3)
#'   z <- as.numeric(as.character(unlist(z)))
#'   num <- z[1] + z[2]/60 + z[3]/3600
#'   return(num)
#' }

####################################################################
#' Visualize Survey Results
#'
#' This function lets the user plot a survey's result.
#'
#' @family Visualization
#' @param answers Dataframe. Answers. Each row a different person.
#' Each column a different answer.
#' @param ignore Numeric Vector. Which columns are NOT answers?
#' @param title Character. Title for your plot
#' @param subtitle Character. Subtitle for your plot.
#' @return ggplot2 object
#' @export
plot_survey <- function(answers, ignore = 1, title = NA, subtitle = NA) {
  if (length(ignore) > 0) {
    answers <- answers[-ignore]
  }
  p <- answers %>%
    gather() %>%
    freqs(.data$key, .data$value) %>%
    ggplot(aes(
      x = .data$key, y = .data$n, fill = .data$value,
      group = .data$key, label = .data$p
    )) +
    geom_bar(stat = "identity") +
    theme_minimal() +
    geom_text(position = position_stack(vjust = .5), colour = "white", size = 2.7) +
    coord_flip() +
    labs(
      title = "Survey Results",
      subtitle = "Answers for the n question",
      x = "", y = "[%]", fill = "",
      caption = paste("Obs.:", nrow(answers))
    )
  if (!is.na(title)) {
    p <- p + labs(title = title)
  }
  if (!is.na(subtitle)) {
    p <- p + labs(subtitle = subtitle)
  }
  p
}

####################################################################
#' Gemini API Interaction with R
#'
#' This function lets the user interact with Google's Gemini LLM Model using
#' its API, and returns the rendered reply.
#'
#' @family API
#' @family Gemini
#' @family LLM
#' @inheritParams gpt_ask
#' @inheritParams cache_write
#' @return (Invisible) list. Content returned from API POST and processed.
#' @examples
#' \dontrun{
#' api_key <- get_credentials()$gemini$api_key
#' # Open question:
#' gemini_ask("Can you write an R function to plot a dummy histogram?", api_key)
#' # Image question
#' image <- "man/figures/automl_map.png"
#' gemini_image("Can you explain this flow with more detail?", image, api_key)
#' }
#' @export
gemini_ask <- function(ask,
                       secret_key = get_creds("gemini")$api_key,
                       url = Sys.getenv("LARES_GEMINI_API"),
                       temperature = 0.5, max_tokens = 1024,
                       quiet = FALSE, ...) {
  model_query <- "gemini-pro:generateContent"
  response <- POST(
    url = paste0(url, model_query),
    query = list(key = secret_key),
    httr::content_type_json(), encode = "json", body = list(
      contents = list(parts = list(list(text = ask))),
      generationConfig = list(
        temperature = temperature,
        maxOutputTokens = max_tokens
      )
    )
  )
  this <- content(response)
  if ("error" %in% names(this)) {
    message(this$error$message)
  } else {
    candidates <- this$candidates
    if (!quiet) cat(unlist(lapply(candidates, function(candidate) candidate$content$parts)))
  }
  invisible(this)
}


#' @param image Character. Data to be encoded/decoded. It can be a raw vector,
#' text connection or file name.
#' @rdname gemini_ask
#' @export
gemini_image <- function(ask, image,
                         secret_key = get_creds("gemini")$api_key,
                         url = Sys.getenv("LARES_GEMINI_API"),
                         temperature = 0.5, max_tokens = 1024,
                         quiet = FALSE, ...) {
  try_require("base64enc")
  model_query <- "gemini-pro-vision:generateContent"
  response <- POST(
    url = paste0(url, model_query),
    query = list(key = secret_key),
    httr::content_type_json(), encode = "json", body = list(
      contents = list(parts = list(list(text = ask), list(inline_data = list(
        mime_type = "image/png", data = base64encode(image)
      )))),
      generationConfig = list(
        temperature = temperature,
        maxOutputTokens = max_tokens
      )
    )
  )
  this <- content(response)
  if ("error" %in% names(this)) {
    message(this$error$message)
  } else {
    candidates <- this$candidates
    if (!quiet) cat(unlist(lapply(candidates, function(candidate) candidate$content$parts)))
  }
  invisible(this)
}

####################################################################
#' Send Emails with Attachments (POST)
#'
#' This function lets the user send Emails with Attachments using MailGun's
#' API service.
#'
#' @family Tools
#' @family Credentials
#' @inheritParams cache_write
#' @inheritParams mp3_get
#' @param from,to,cc,bcc Character. Emails
#' @param subject Character. Subject for the email.
#' @param text,html Character. Text or HTML to send in the body.
#' @param attachment Character, plot or data.frame. Will send the file,
#' plot as PNG or data.frame as CSV, respectively.
#' @param service Character. Service platform to search on \code{creds}.
#' @param creds Character. Credential's user (see \code{get_creds()}).
#' Must contain: url (POST address), api (API key).
#' @return No return value, called for side effects.
#' @examples
#' \dontrun{
#' myPlot <- noPlot("My plot")
#' mail_send(
#'   from = "BLV <myuser@mail.com>",
#'   to = "youruser@mail.com",
#'   cc = "myuser@mail.com",
#'   subject = paste("Daily report:", Sys.Date()),
#'   attachment = myPlot
#' )
#' }
#' @export
mail_send <- function(from = "RMail <laresbernardo@gmail.com>",
                      to = "laresbernardo@gmail.com",
                      cc = NULL, bcc = NULL,
                      subject = "Mail from R",
                      text = " \n",
                      html = NULL,
                      attachment = NULL,
                      service = "mailgun",
                      creds = NULL,
                      quiet = FALSE,
                      ...) {
  # MailGun documentation:
  # https://documentation.mailgun.com/en/latest/api-sending.html#sending

  credentials <- get_credentials(from = service, dir = creds, ...)
  url <- credentials$url
  api_key <- credentials$api

  the_body <- list(from = from, to = to, subject = subject)

  if (grepl("mailgun", url)) the_body[["o:tag"]] <- "R-email"
  if (!is.null(cc)) the_body[["cc"]] <- cc
  if (!is.null(bcc)) the_body[["bcc"]] <- bcc
  if (!is.null(text)) the_body[["text"]] <- text
  if (!is.null(html)) the_body[["html"]] <- html

  if (!is.null(attachment)) {
    if ("ggplot" %in% class(attachment)) {
      as <- "png"
      file <- paste0(tempdir(), "/attch.", as)
      attachment + ggsave(file)
    }
    if ("data.frame" %in% class(attachment)) {
      as <- "csv"
      file <- paste0(tempdir(), "/attch.", as)
      write.csv(attachment, file, row.names = FALSE)
    }
    if ("character" %in% class(attachment)) {
      as <- "attachment"
      file <- as.character(attachment)
    }
    the_body[["attachment"]] <- upload_file(file)
  } else {
    as <- NULL
    file <- NULL
  }

  # Send email
  req <- POST(url,
    authenticate("api", api_key),
    add_headers("Content-Type" = "multipart/form-data"),
    body = the_body
  )
  stop_for_status(req)

  # Delete temporary files created
  if (!is.null(file) && !"attachment" %in% as) file.remove(file)

  if (!quiet) {
    message(paste(
      subject, "sent to", to, ifelse(!is.null(as), paste("with", as, "file"), "")
    ))
  }
}

####################################################################
#' Queries on Google Analytics
#'
#' This function lets the user query Google Analytics with its API.
#' More about the documentation and parameters in
#' \code{googleAnalyticsR::google_analytics()} or Google Analytics' API.
#'
#' @family Credentials
#' @family Google
#' @family API
#' @param account Character. Personal named accounts
#' @param creds Character. Credential's user (see \code{get_creds()})
#' @param token_dir Character. Credential's directory (see \code{get_creds()})
#' @param metrics Character. Which metrics we wish to bring
#' @param dimensions Character. Which dimensions we wish to bring
#' @param met_filters,dim_filters A \code{filter_clause_ga4} for filtering
#' metrics/dimensions. Check \code{googleAnalyticsR::google_analytics()}.
#' @param start Date. Start date for the report
#' @param end Date. End date for the report
#' @return data.frame with the API GET request tabulated results.
#' @export
queryGA <- function(account, creds = NA, token_dir = NA,
                    metrics = "sessions",
                    dimensions = "date",
                    met_filters = NULL,
                    dim_filters = NULL,
                    start = lubridate::floor_date(Sys.Date(), "month"),
                    end = Sys.Date()) {
  try_require("googleAnalyticsR")
  try_require("googleAuthR")

  # FIRST TIME
  # options(googleAuthR.scopes.selected = "https://www.googleapis.com/auth/analytics",
  #         googleAuthR.client_id = "XXX",
  #         googleAuthR.client_secret = "XXX")
  # gar_auth()

  account <- paste("google_analytics", account, sep = "_")
  vars <- get_credentials(from = account, dir = creds)
  options(
    "googleAuthR.scopes.selected" = "https://www.googleapis.com/auth/analytics",
    "googleAuthR.client_id" = vars$client_id,
    "googleAuthR.client_secret" = vars$client_secret
  )

  # Authenticate with local file
  if (!is.na(token_dir)) {
    token <- paste0(token_dir, "/", vars$token_name)
  } else {
    token <- vars$token_name
  }
  message(paste("Token:", token))
  gar_auth(token) # From googleAuthR

  # Query on Google Analytics
  google_analytics(
    vars$ga_id,
    date_range = c(start, end),
    metrics = metrics,
    dimensions = dimensions,
    met_filters = met_filters,
    dim_filters = dim_filters
  )
}

####################################################################
#' Google Trends: Related and Time Plots
#'
#' This function creates a plot with Google Trend's related topics
#' and queries, and let the user compare different keywords.
#'
#' @family Scrapper
#' @family Google
#' @aliases trendsRelated
#' @param gtrend List. Result from \code{gtrendsR::gtrends(keyword, geo, time)}
#' @param top Integer. Filter top n results only.
#' @param title Character. Custom title for the plot.
#' @param note Character. Add a note to the plot if needed.
#' @param exclude Character vector. Which observations do you wish to exclude?
#' @return plot for Google Trend's results input \code{gtrend}.
#' @export
#' @rdname google_trends
gtrends_related <- function(gtrend, top = NA, title = NA, note = NA, exclude = NULL) {
  start <- as.Date(min(gtrend$interest_over_time$date))
  end <- as.Date(max(gtrend$interest_over_time$date))
  range <- paste0(start, " - ", end, " (", as.integer(end - start), "d)")

  t <- !is.null(gtrend$related_topics) # Related topics
  q <- !is.null(gtrend$related_queries) # Related queries

  if (!t && !q) {
    stop("No related topics nor queries found!")
  }

  fx <- function(gtrend_related) {
    which <- colnames(gtrend_related)[2]
    ptitle <- ifelse(which == "related_topics", "Related topics", "Related queries")
    colnames(gtrend_related)[2] <- "related"
    more2 <- length(unique(gtrend_related$keyword))

    related <- gtrend_related[gtrend_related$related == "top", ] %>%
      left_join(gtrend_related[gtrend_related$related == "rising", ] %>%
        select(.data$value, .data$subject), "value", "keyword") %>%
      distinct() %>%
      rename(
        rank = .data$subject.x,
        type = .data$subject.y
      ) %>%
      mutate(type_label = ifelse(grepl("+", .data$type), as.character(.data$type), "")) %>%
      mutate(rank = ifelse(.data$rank == "<1", "0.1", as.character(.data$rank))) %>%
      mutate(type = ifelse(is.na(.data$type), "Steady", "Big change")) %>%
      group_by(.data$keyword) %>%
      arrange(desc(as.numeric(.data$rank))) %>%
      mutate(rank = ifelse(is.na(.data$rank), max(.data$rank, na.rm = T), .data$rank)) %>%
      mutate(rank = ifelse(is.infinite(.data$rank), 100, as.integer(.data$rank))) %>%
      mutate(rank = as.integer(as.character(.data$rank))) %>%
      {
        if (more2 > 1) mutate(., value = paste0(.data$value, " (", .data$keyword, ")")) else .
      } %>%
      distinct()
    if (length(exclude) != 0) {
      related <- related %>%
        filter(!.data$value %in% exclude) %>%
        mutate(rank = 100 * .data$rank / max(.data$rank))
    }

    n <- ifelse(is.na(top), round(50 / more2, 0), top)
    if (!is.na(top) || more2 > 3) {
      related <- related %>%
        group_by(.data$keyword) %>%
        arrange(desc(.data$rank)) %>%
        slice(1:n)
      if (is.na(top)) {
        message("Filtering top ", n, " values. Use top to over-write if needed")
      }
    }

    if (length(exclude) != 0) {
      related <- related %>% filter(!.data$value %in% exclude)
    }

    plot <- related %>%
      ggplot(aes(
        x = reorder(.data$value, .data$rank), y = .data$rank,
        fill = .data$type, label = .data$type_label
      )) +
      geom_col() +
      coord_flip() +
      theme_lares() +
      geom_text(hjust = -0.1, size = 2.8) +
      labs(y = "Relevance | % Increase", subtitle = ptitle, x = "", fill = "") +
      theme(legend.position = "bottom") +
      scale_y_comma(
        limits = c(0, 110),
        minor_breaks = seq(0, 100, 10),
        breaks = seq(0, 100, 20)
      )
    if (more2 > 1) plot <- plot + facet_grid(.data$keyword ~ ., scales = "free")
    plot
  }

  if (t) {
    rq1 <- fx(gtrend$related_topics)
  } else {
    message("No related topics found")
  }
  if (q) {
    rq2 <- fx(gtrend$related_queries)
    if (t == FALSE) {
      rq1 <- rq2
    }
  } else {
    message("No related queries found")
  }

  if (!is.na(title)) rq1 <- rq1 + labs(title = title)

  if (!is.na(note)) {
    rq1 <- rq1 + labs(caption = note)
  } else {
    rq1 <- rq1 + labs(caption = range)
  }

  if (t && q) {
    rq1 <- rq1 + guides(fill = "none") + labs(caption = NULL)
    rq2 <- rq2 + labs(title = range) +
      theme(plot.title = element_text(size = 9))
    rq1 + rq2 + plot_layout(nrow = 1, ncol = 2)
  } else {
    if (t || q) rq1
  }
}

####################################################################
#' Google Trends: Timelines Plot
#'
#' This function creates a plot with google trend's data on timelines
#' and let the user compare different keywords.
#'
#' @return Plot for Google Trend's results input \code{gtrend}.
#' @aliases trendsTime
#' @export
#' @rdname google_trends
gtrends_time <- function(gtrend, title = NA) {
  try_require("gtrendsR")
  int1 <- plot(gtrend) + theme_lares() +
    labs(x = "", colour = "") + ylim(0, 100) +
    theme(legend.position = "top") +
    geom_hline(yintercept = 100, alpha = 0.5)

  if (length(unique(gtrend$interest_over_time$keyword)) == 1) {
    int1
  } else {
    start <- as.Date(min(gtrend$interest_over_time$date))
    end <- as.Date(max(gtrend$interest_over_time$date))
    range <- paste0(start, " - ", end, " (", as.integer(end - start), "d)")

    int2 <- gtrend$interest_over_time %>%
      mutate(hits = ifelse(.data$hits == "<1", "0.5", as.character(.data$hits))) %>%
      mutate(hits = as.numeric(as.character(.data$hits))) %>%
      group_by(.data$keyword) %>%
      mutate(legend = paste0(.data$keyword, " (", .data$geo, ")")) %>%
      ggplot(aes(x = .data$date, y = .data$hits, fill = .data$legend)) +
      geom_area(alpha = 0.9) +
      theme_lares() +
      guides(fill = "none") +
      labs(x = "", y = "Search hits", fill = "", subtitle = "Mixed hits scale")

    int3 <- gtrend$interest_over_time %>%
      mutate(hits = ifelse(.data$hits == "<1", "0.5", as.character(.data$hits))) %>%
      mutate(hits = as.numeric(as.character(.data$hits))) %>%
      group_by(.data$keyword) %>%
      mutate(hits = 100 * .data$hits / max(.data$hits)) %>%
      mutate(legend = paste0(.data$keyword, " (", .data$geo, ")")) %>%
      ggplot(aes(x = .data$date, y = .data$hits, colour = .data$legend)) +
      geom_line() +
      theme_minimal() +
      guides(colour = "none") +
      labs(
        x = "", y = "Search hits",
        subtitle = "Normalized hit scale",
        caption = range
      ) +
      geom_hline(yintercept = 100, alpha = 0.5) +
      theme_lares()

    if (!is.na(title)) int1 <- int1 + labs(title = title, subtitle = "Real hits scale")

    p <- int1 + int2 + int3 + plot_layout(nrow = 3, ncol = 1)
    p
  }
}

####################################################################
#' ChatGPT API Interaction with R
#'
#' This function lets the user ask ChatGPT via its API, and returns
#' the rendered reply. There are a couple of specific verbs (functions) with a
#' preset prompt to help fetch the data in specific formats. We also
#' store the prompts and replies in current session with their respective
#' time-stamps so user can gather historical results.
#'
#' @family API
#' @family ChatGPT
#' @family LLM
#' @inheritParams cache_write
#' @inheritParams db_download
#' @param ask Character. Redacted prompt to ask. If multiple asks are
#' requested, they will be concatenated with "+" into a single request.
#' @param secret_key Character. Secret Key. Get yours in:
#' platform.openai.com for OpenAI or makersuite.google.com/app/apikey
#' for Gemini.
#' @param url Character. Base API URL.
#' @param model Character. OpenAI model to use. This can be adjusted
#' according to the available models in the OpenAI API (such as "gpt-4").
#' @param temperature Numeric. The temperature to use for generating
#' the response. Default is 0.5. The lower the \code{temperature},
#' the more deterministic the results in the sense that the highest probable
#' next token is always picked. Increasing temperature could lead to more
#' randomness, which encourages more diverse or creative outputs. You are
#' essentially increasing the weights of the other possible tokens.
#' In terms of application, you might want to use a lower temperature value
#' for tasks like fact-based QA to encourage more factual and concise responses.
#' For poem generation or other creative tasks, it might be beneficial to
#' increase the temperature value.
#' @param max_tokens Integer. The maximum number of tokens in the response.
#' @param num_retries Integer. Number of times to retry the request in
#' case of failure. Default is 3.
#' @param pause_base Numeric. The number of seconds to wait between retries.
#' Default is 1.
#' @return (Invisible) list. Content returned from API POST and processed.
#' @examples
#' \dontrun{
#' api_key <- get_credentials()$openai$secret_key
#' # Open question:
#' gpt_ask("Can you write an R function to plot a dummy histogram?", api_key)
#'
#' ##### The following examples return dataframes:
#' # Classify each element based on categories:
#' gpt_classify(1:10, c("odd", "even"))
#'
#' # Add all tags that apply to each element based on tags:
#' gpt_tag(
#'   c("I love chocolate", "I hate chocolate", "I like Coke"),
#'   c("food", "positive", "negative", "beverage")
#' )
#'
#' # Extract specific information:
#' gpt_extract(
#'   c("My mail is 123@@test.com", "30 Main Street, Brooklyn, NY, USA", "+82 2-312-3456", "$1.5M"),
#'   c("email", "full state name", "country of phone number", "amount as number")
#' )
#'
#' # Format values
#' gpt_format(
#'   c("March 27th, 2021", "12-25-2023 3:45PM", "01.01.2000", "29 Feb 92"),
#'   format = "ISO Date getting rid of time stamps"
#' )
#'
#' # Convert temperature units
#' gpt_convert(c("50C", "300K"), "Fahrenheit")
#'
#' # Create a table with data
#' gpt_table("5 random people's address in South America, email, phone, age between 18-30")
#' gpt_table(
#'   ask = "5 largest cities, their countries, and population",
#'   cols = c("city_name", "where", "POP")
#' )
#'
#' # Translate text to any language
#' gpt_translate(
#'   rep("I love you with all my heart", 5),
#'   language = c("spanish", "chinese", "japanese", "russian", "german")
#' )
#'
#' # Now let's read the historical prompts, replies, ano more from current session
#' gpt_history()
#' }
#' @export
gpt_ask <- function(ask,
                    secret_key = get_credentials()$openai$secret_key,
                    url = Sys.getenv("LARES_GPT_URL"),
                    model = Sys.getenv("LARES_GPT_MODEL"),
                    num_retries = 3,
                    temperature = 0.5,
                    max_tokens = NULL,
                    pause_base = 1,
                    quiet = FALSE, ...) {
  ts <- Sys.time()
  if (length(ask) > 1) ask <- paste(ask, collapse = " + ")
  # Save historical questions
  cache <- bind_rows(
    data.frame(ts = ts, prompt = ask),
    cache_read("GPT_HIST_ASK", quiet = TRUE, ...)
  ) %>%
    as_tibble()
  cache_write(distinct(cache), "GPT_HIST_ASK", quiet = TRUE, ...)

  # Ask ChatGPT using their API
  response <- POST(
    url = url,
    add_headers(Authorization = paste("Bearer", secret_key)),
    httr::content_type_json(),
    encode = "json",
    times = num_retries,
    pause_base = pause_base,
    body = list(
      model = model,
      temperature = temperature,
      max_tokens = max_tokens,
      messages = list(list(
        role = "user",
        content = ask
      ))
    )
  )
  ret <- content(response)
  if ("error" %in% names(ret)) warning(ret$error$message)
  if ("message" %in% names(ret$choices[[1]]) & !quiet) {
    cat(paste(stringr::str_trim(ret$choices[[1]]$message$content), "\n"))
  }

  # Save historical answers
  cache <- rbind(
    data.frame(ts = ts, reply = toJSON(ret)),
    cache_read("GPT_HIST_REPLY", quiet = TRUE, ...)
  ) %>%
    as_tibble()
  cache_write(distinct(cache), "GPT_HIST_REPLY", quiet = TRUE, ...)
  invisible(ret)
}

#' @rdname gpt_ask
#' @export
gpt_history <- function(quiet = TRUE, ...) {
  asks <- cache_read("GPT_HIST_ASK", quiet = quiet, ...)
  replies <- cache_read("GPT_HIST_REPLY", quiet = quiet, ...)
  if (!is.null(asks)) {
    if (!is.null(replies)) {
      left_join(asks, replies, by = "ts") %>%
        select(.data$ts, .data$prompt, contains("message.content"), everything())
    } else {
      asks
    }
  } else {
    message("No historical prompts nor replies registered yet")
    invisible(NULL)
  }
}

#' @param cols Vector. Force column names for your table results.
#' @rdname gpt_ask
#' @export
gpt_table <- function(x, cols = NULL, quiet = TRUE, ...) {
  p <- gpt_prompter(instruction = x, output = "table", cols = cols, ...)
  resp <- gpt_ask(p$prompt, quiet = quiet, ...)
  gpt_markdown2df(resp)
}

#' @param x Vector. List items you wish to process in your instruction
#' @param categories,tags Vector. List of possible categories/tags to consider.
#' @rdname gpt_ask
#' @export
gpt_classify <- function(x, categories, quiet = TRUE, ...) {
  p <- gpt_prompter(
    instruction = paste(
      "For each of the inputs,",
      "classify using only the options in context"
    ),
    input = x, context = categories,
    output = "table",
    cols = c("Input", "Category"),
    ...
  )
  resp <- gpt_ask(p$prompt, quiet = quiet, ...)
  gpt_markdown2df(resp)
}

#' @rdname gpt_ask
#' @export
gpt_tag <- function(x, tags, quiet = TRUE, ...) {
  p <- gpt_prompter(
    instruction = paste(
      "For each of the inputs,",
      "provide which of the context values apply as correct tags using TRUE/FALSE"
    ),
    input = x, context = tags,
    output = "table",
    ...
  )
  resp <- gpt_ask(p$prompt, quiet = quiet, ...)
  gpt_markdown2df(resp)
}

#' @param extract,format,unit Character. Length 1 or same as x to extract/format/unit
#' information from x. For example: email, country of phone number, country, amount as number,
#' currency ISO code, ISO, Fahrenheit, etc.
#' @rdname gpt_ask
#' @export
gpt_extract <- function(x, extract, quiet = TRUE, ...) {
  stopifnot(length(extract) %in% c(1, length(x)))
  p <- gpt_prompter(
    instruction = paste(
      "For each of the inputs,",
      "extract each of the information asked in context"
    ),
    input = x, context = extract,
    output = "table",
    cols = c("Input", "Element_to_extract", "Value"),
    ...
  )
  resp <- gpt_ask(p$prompt, quiet = quiet, ...)
  gpt_markdown2df(resp)
}

#' @rdname gpt_ask
#' @export
gpt_format <- function(x, format, quiet = TRUE, ...) {
  stopifnot(length(format) %in% c(1, length(x)))
  p <- gpt_prompter(
    instruction = paste(
      "For each of the inputs,",
      "standardize and format all values to the format in context"
    ),
    input = x, context = format,
    output = "table",
    cols = c("Input", "Formatted"),
    ...
  )
  resp <- gpt_ask(p$prompt, quiet = quiet, ...)
  gpt_markdown2df(resp)
}

#' @rdname gpt_ask
#' @export
gpt_convert <- function(x, unit, quiet = TRUE, ...) {
  stopifnot(length(unit) %in% c(1, length(x)))
  p <- gpt_prompter(
    instruction = paste(
      "For each of the inputs,",
      "provide new converted values using the units in context"
    ),
    input = x, context = unit,
    output = "table",
    cols = c("Input", "Original_Unit", "Total_Value", "Converted_Value", "New_Unit"),
    ...
  )
  resp <- gpt_ask(p$prompt, quiet = quiet, ...)
  gpt_markdown2df(resp)
}

#' @param language Character. Language to translate to
#' @rdname gpt_ask
#' @export
gpt_translate <- function(x, language, quiet = TRUE, ...) {
  stopifnot(length(language) %in% c(1, length(x)))
  p <- gpt_prompter(
    instruction = paste(
      "For each of the inputs,",
      "translate to the respective languages provided in context"
    ),
    input = x, context = language,
    output = "table",
    cols = c("Input", "Language", "Translation"),
    ...
  )
  resp <- gpt_ask(p$prompt, quiet = quiet, ...)
  gpt_markdown2df(resp)
}

gpt_markdown2df <- function(resp) {
  if ("message" %in% names(resp$choices[[1]])) {
    resp <- resp$choices[[1]]$message$content
  }
  df <- try(markdown2df(resp))
  attr(df, "response") <- df
  df
}

# DEPRECATED: Now using the new gpt_prompter()
# gpt_prompter_basic <- function(type = "category", cols = c("item", type), x, y) {
#   paste(
#     "Return a structured markdown table",
#     "with", length(cols), "columns named exactly", v2t(cols, and = "and"),
#     ". Consider the following items:", v2t(x, quotes = FALSE),
#     ". For each respective item, what", type, "represent each item using:", v2t(y, quotes = FALSE),
#     ". If you don't know any item, replace with NA"
#   )
# }

### Theoretical elements to build prompts in a standard and cleaner way
# TIP: Start simple and be very specific


####################################################################
#' Structured Prompt Builder for LLM (ChatGPT)
#'
#' Build standard prompts to get the right outcomes using the four theoretical
#' elements required to build standard and cleaner asks: instruction, input, context,
#' and output. Inspired by the
#' \href{https://www.promptingguide.ai/}{Prompt Engineering Guide} free guide.
#' Remember to start simple and be very specific to get exactly what you need.
#'
#' @section Elements of a Prompt:
#' \describe{
#'   \item{Instruction}{a specific task or instruction you want the model to perform.
#'   It usually starts with a verb given its an instruction.}
#'   \item{Input}{input data, elements or question that we are interested to find a response for,}
#'   \item{Context}{external information, additional context or references to steer
#'   the model to better responses.}
#'   \item{Output}{type or format of the output.}
#' }
#'
#' @family ChatGPT
#' @family LLM
#' @param instruction,input,context,output Character or vector.
#' You do not need all the four elements for a prompt and the
#' format depends on the task at hand.
#' @param quiet Boolean. Should the written prompt be printed or not?
#' @param ... Additional parameters. You can pass \code{cols} parameter to
#' explicitly set column names.
#' @return (Invisible) list with written prompt and elements provided.
#' @examples
#' # A simple formatted table with data
#' # Note: I mostly use output = "table" and enabled an auxiliary enrichment prompt
#' (p <- gpt_prompter(instruction = "Capitals of the world", output = "table"))
#'
#' # Classify
#' p <- gpt_prompter(
#'   instruction = "For each of the inputs, classify using only the options in context",
#'   input = c("Molecule", "Elephant", "Milky Way", "Cat", "Planet Earth"),
#'   context = c("Big", "Medium", "Small"),
#'   output = "table",
#'   # This cols parameter is auxiliary
#'   cols = c("Input", "Category"),
#'   quiet = FALSE
#' )
#'
#' # Tag all categories that apply
#' p <- gpt_prompter(
#'   instruction = paste(
#'     "For each of the inputs, provide which of the",
#'     "context values apply as correct tags using TRUE/FALSE"
#'   ),
#'   input = c("I love chocolate", "I hate chocolate", "I like Coke", "Who am I?", "T-REX"),
#'   context = c("food", "positive", "negative", "beverage"),
#'   output = "table",
#'   quiet = FALSE
#' )
#'
#' # Extract information from strings
#' p <- gpt_prompter(
#'   instruction = "For each of the inputs, extract each of the information asked in context",
#'   input = c("My mail is 123@@test.com", "30 Main St, NY, USA", "+82 2-312-3456", "$1.5M"),
#'   context = c("email", "full state name", "country of phone", "full non-abbreviated number"),
#'   output = "table",
#'   cols = c("Input", "Element_to_extract", "Value"),
#'   quiet = FALSE
#' )
#'
#' # Translate to several languages
#' p <- gpt_prompter(
#'   instruction = "For each of the inputs, translate to the respective languages in context",
#'   input = rep("I love you with all my heart", 5),
#'   context = c("spanish", "chinese", "japanese", "russian", "german"),
#'   output = "table",
#'   cols = c("Input", "Language", "Translation"),
#'   quiet = FALSE
#' )
#'
#' # Format date values
#' p <- gpt_prompter(
#'   instruction = paste(
#'     "For each of the inputs,",
#'     "standardize and format all values to the format in context"
#'   ),
#'   input = c("March 27th, 2021", "12-25-2023 3:45PM", "01.01.2000", "29 Feb 92"),
#'   context = "ISO Date getting rid of time stamps",
#'   output = "table",
#'   cols = c("Input", "Formatted"),
#'   quiet = FALSE
#' )
#'
#' # Convert units
#' p <- gpt_prompter(
#'   instruction = paste(
#'     "For each of the inputs,",
#'     "provide new converted values using the units in context"
#'   ),
#'   input = c("50C", "300K", "100F", "0F", "32C", "0K"),
#'   context = "Fahrenheit",
#'   output = "table",
#'   cols = c("Input", "Original_Unit", "Total_Value", "Converted_Value", "New_Unit"),
#'   quiet = FALSE
#' )
#'
#' # Read a text and answer a question related to it
#' gpt_prompter(
#'   instruction = "read",
#'   context = "Long text here",
#'   input = "Question here"
#' )$prompt
#' @export
gpt_prompter <- function(instruction = NULL,
                         input = NULL,
                         context = NULL,
                         output = NULL,
                         quiet = TRUE,
                         ...) {
  # Generic helpers for common use cases
  if ("table" %in% output) {
    output <- paste(
      "Markdown format table. If missing elements, replace with NA.",
      "For numerical columns, only use a dot for decimals, no delimiters, no abbreviations"
    )
  }
  if ("read" %in% instruction) {
    instruction <- paste(
      "You are a pro content editor who will read the context text and",
      "reply using a single concise response to the question asked in input.",
      "If the answer is not found in context, reply you cannot find the answer in the text"
    )
  }
  if ("cols" %in% names(list(...))) {
    cols <- list(...)[["cols"]]
    if (!is.null(cols)) {
      output <- paste0(
        output, "; also, it must have ", length(cols), " columns named exactly: ", v2t(cols, and = "and")
      )
    }
  }

  # Build the list with 4 elements (and additional information passed through ...)
  elements <- list(
    instruction = if (!is.null(instruction)) paste("Instruction =", paste(instruction, collapse = "; ")) else NULL,
    input = if (!is.null(input)) paste("Input =", paste(input, collapse = "; ")) else NULL,
    context = if (!is.null(context)) paste("Context =", paste(context, collapse = "; ")) else NULL,
    output = if (!is.null(output)) paste("Output =", paste(output, collapse = "; ")) else NULL,
    ...
  )
  # Get rid of already used information
  elements["cols"] <- NULL
  # Check if something's actually being prompted
  are_null <- unlist(lapply(elements, is.null))
  if (all(are_null)) warning("No prompt provided. Set any of the elements: ", v2t(names(elements)))
  if (!quiet) cat(paste(paste(elements[!are_null], collapse = "\n"), "\n"))
  invisible(
    list(
      prompt = v2t(elements[!are_null], quotes = FALSE, sep = " ### "),
      elements = list(instruction = instruction, input = input, context = context, output = output, ...)
    )
  )
}

####################################################################
#' Check if Specific Package is Installed
#'
#' This function checks library dependencies
#'
#' @family Tools
#' @inheritParams cache_write
#' @param package Character. Name of the library
#' @param stop Boolean. Stop if not installed. If \code{FALSE} and
#' library is not available, warning will be shown.
#' @param load Boolean. Load library?
#' @param lib.loc Character vector. Location of R library trees
#' to search through, or \code{NULL}. The default value of \code{NULL}
#' corresponds to all libraries currently known to \code{.libPaths()}.
#' Non-existent library trees are silently ignored.
#' @return No return value, called for side effects.
#' @examples
#' # Check if library base is installed. If not, stop and show error
#' try_require("base", stop = TRUE)
#' # Check if library xxx is installed. If not, show warning
#' try_require("xxx", stop = FALSE)
#' @export
try_require <- function(package, stop = TRUE, load = TRUE, lib.loc = NULL, ...) {
  present <- length(find.package(package, quiet = TRUE)) > 0
  if (present && load) {
    # Be careful: ... parameter is not enabled in library()
    suppressPackageStartupMessages(library(package, character.only = TRUE, lib.loc = lib.loc))
  } else {
    if (stop) {
      stop(paste0("Package '", package, "' required. Install and try again."), call. = FALSE)
    } else {
      warning(paste0("Package '", package, "' recommended. Install for better results."), call. = FALSE)
    }
  }
}


####################################################################
#' Get Meta Data from Image Files
#'
#' This function lets the user get meta data from image files or directory.
#'
#' @family Tools
#' @param files Character vector. Files or directory which contains files.
#' @return data.frame with meta-data for each image file.
#' @export
image_metadata <- function(files) {
  try_require("exifr")

  files <- as.character(files)
  if (length(files) == 1) {
    if (dir.exists(files)) {
      files <- listfiles(files, recursive = TRUE)
    }
  }

  df <- data.frame(file = files)
  tags <- c(
    "FileName", "SourceFile",
    "CreateDate", "DateTimeOriginal", "FileModifyDate",
    "FileTypeExtension", "Megapixels",
    "ImageSize", "ImageWidth", "ImageHeight",
    "GPSLongitude", "GPSLatitude",
    "GPSLatitudeRef", "GPSLongitudeRef",
    "Rotation", "Flash", "Duration",
    "Make", "Model"
  )

  aux <- ceiling(nrow(df) / 500)
  for (i in 1:aux) {
    if (i == 1) {
      ret <- NULL
      if (aux > 2) {
        message(paste(
          "This might take a while... Analizing",
          formatNum(nrow(df), decimals = 0), "files!"
        ))
      }
    }
    from <- (i - 1) * 500 + 1
    to <- i * 500
    x <- slice(df, from:to)
    temp <- read_exif(as.character(x$file), tags = tags)
    if (nrow(temp) > 0) {
      if ("DateTimeOriginal" %in% colnames(temp)) {
        ret <- bind_rows(ret, select(temp, one_of(tags)))
      }
    }
    statusbar(i, aux, label = paste(formatNum(from, 0), "-", formatNum(to, 0)))
  }

  if (length(ret) > 0) {
    if ("DateTimeOriginal" %in% colnames(ret)) {
      df <- ret %>%
        mutate(
          DateTimeOriginal = ymd_hms(.data$DateTimeOriginal),
          CreateDate = ymd_hms(.data$CreateDate),
          FileModifyDate = ymd_hms(.data$FileModifyDate)
        )
      df <- df[, colSums(is.na(df)) < nrow(df)]
      if (aux > 10) {
        tryCatch({
          try_require("beepr", stop = FALSE)
          beep()
        })
      }
      df
    }
  } else {
    message("No images found to process...")
  }
}


####################################################################
#' List files in a directory
#'
#' This function lets the user list all files on a given directory.
#' It also lets filter files which contains a string.
#'
#' @family Tools
#' @param folder Character. Directory which contains files
#' @param recursive Boolean. Should the listing recurse into directories?
#' @param regex Character. String to use for filtering files
#' @param images Boolean. Bring only image files?
#' @return data.frame with relevant data for each file on
#' \code{folder} directory.
#' @examples
#' # All files in current directory (without recursive files)
#' df <- listfiles(recursive = TRUE)
#' head(df, 3)
#'
#' # All files in current directory (with recursive files)
#' df <- listfiles(recursive = TRUE)
#' tail(df, 3)
#'
#' # Check R files using regex
#' df <- listfiles(regex = "\\.R$")
#' @export
listfiles <- function(folder = getwd(),
                      recursive = TRUE,
                      regex = NA,
                      images = FALSE) {
  if (!file.exists(folder)) {
    stop("That directory doesn't exist; please try again!")
  }

  files <- list.files(folder, recursive = recursive)
  address <- paste0(folder, "/", files)
  info <- file.info(address)

  df <- data.frame(filename = files, address, info)
  df$size <- as.integer(df$size / 1024)
  # imgs <- "jpg|JPG|jpeg|JPEG|png|PNG|gif|GIF"

  if (!is.na(regex)) df <- df[grep(regex, df$filename), ]

  if (images) {
    df <- image_metadata(df$address)
  }

  row.names(df) <- NULL
  df$address <- NULL
  df
}


####################################################################
#' What's my IP?
#'
#' Reveal your current IP address.
#'
#' @family Tools
#' @return Character. Result of your IP address based on ipify.org
#' @examples
#' \donttest{
#' myip()
#' }
#' @export
myip <- function() {
  if (!haveInternet()) {
    message("No internet connetion...")
    invisible(NULL)
  } else {
    ipify <- "https://api.ipify.org/"
    try(content(GET(ipify), encoding = "UTF-8"))
  }
}


####################################################################
#' Quiet prints and verbose noise
#'
#' This function silences (verbose) output prints. Thanks to Hadley Wickham
#' for bringing the idea.
#'
#' @family Tools
#' @param fx Function to quiet
#' @param quiet Quiet outputs? If not, skip quietness.
#' @return Same as \code{fx} but with no messages or prints.
#' @export
quiet <- function(fx, quiet = TRUE) {
  if (!quiet) {
    fx
  } else {
    invisible(capture.output(fx))
  }
}


####################################################################
#' Internet Connection Check
#'
#' This function checks if your R session currently have Wifi or
#' Internet connection.
#'
#' @family Tools
#' @param thresh Numeric. How many seconds to consider a slow connection?
#' @param url Character. URL to test the readLines 1 command
#' @return Boolean. Result of checking if device has internet connection.
#' @export
haveInternet <- function(thresh = 3, url = "http://www.google.com") {
  start <- Sys.time()
  if (!capabilities(what = "http/ftp")) {
    FALSE
  } else {
    test <- try(suppressWarnings(readLines(url, n = 1)), silent = TRUE)
    if (as.numeric(Sys.time() - start) > thresh) {
      message("Slow internet connection but available...")
    }
    !inherits(test, "try-error")
  }
}


####################################################################
#' Check if Font is Installed
#'
#' This function checks if a font is installed in your machine.
#' To list all available fonts, set \code{font = NULL}.
#'
#' @family Tools
#' @inheritParams cache_write
#' @param font Character. Which font to check. No need to add .TFF.
#' @param font_dirs Character vector. Additional directories to check for fonts.
#' @param quiet Boolean. Keep quiet? If not, show message
#' @return Boolean result of the existing fonts check.
#' @examples
#' font_exists(font = "Arial")
#' font_exists(font = "arial")
#' font_exists(font = "")
#' font_exists(font = NULL)
#' @export
font_exists <- function(font = "Arial Narrow", font_dirs = NULL, quiet = FALSE, ...) {
  if (isTRUE(font[1] == "")) {
    FALSE
  }
  if (isTRUE(is.na(font[1]))) {
    FALSE
  }
  tryCatch(
    check_font(font, font_dirs, quiet),
    error = function(err) {
      if (!quiet) message(paste("Font issue:", err))
      FALSE
    }
  )
}

# Thanks to extrafont for the idea for this code
ttf_find_default_path <- function(font_dirs = NULL) {
  os <- R.version$os
  paths <- if (grepl("^darwin", os)) {
    c(
      "/Library/Fonts/",
      "/System/Library/Fonts",
      "/System/Library/Fonts/Supplemental",
      "~/Library/Fonts/",
      font_dirs
    )
  } else if (grepl("^linux-gnu", os)) {
    c(
      "/usr/share/fonts/",
      "/usr/local/share/fonts/",
      "/usr/X11R6/lib/X11/fonts/TrueType/",
      "~/.local/share/fonts/",
      "~/.fonts/",
      font_dirs
    )
  } else if (grepl("^freebsd", os)) {
    c(
      "/usr/local/share/fonts/truetype/",
      "/usr/local/lib/X11/fonts/",
      "~/.fonts/",
      font_dirs
    )
  } else if (grepl("^mingw", os)) {
    c(
      file.path(Sys.getenv("SystemRoot"), "Fonts"),
      file.path(Sys.getenv("LOCALAPPDATA"), "Microsoft", "Windows", "Fonts"),
      font_dirs
    )
  } else {
    stop("Unknown platform. Don't know where to look for truetype fonts. Sorry!")
  }
  paths[file.exists(paths)]
}

check_font <- function(font, font_dirs = NULL, quiet = FALSE) {
  pattern <- "\\.ttf$|\\.otf"
  fonts_path <- ttf_find_default_path(font_dirs)
  ttfiles <- list.files(fonts_path,
    pattern = pattern, recursive = TRUE,
    full.names = TRUE, ignore.case = TRUE
  )
  font_names <- basename(ttfiles)
  nice_names <- gsub(pattern, "", font_names, ignore.case = TRUE)
  if (is.null(font)) {
    nice_names
  } else {
    if (grepl("^mingw", R.version$os)) {
      try_require("grDevices")
      win_fonts <- unlist(windowsFonts())
      ret <- font %in% win_fonts
      # font_names <- c(font_names, win_fonts)
    } else {
      ret <- font %in% nice_names
    }
    if (!quiet && !all(ret)) {
      if (!is.null(font)) {
        font_names <- font_names[
          grepl(tolower(paste(font, collapse = "|")), tolower(font_names))
        ]
      }
      if (length(font_names) > 0) {
        message("Maybe you meant one of these:\n", v2t(sort(gsub("\\..*", "", font_names))))
      }
    }
    ret
  }
}


####################################################################
#' List all functions used in R script files by package
#'
#' Parses all functions called by an R script and then lists
#' them by package. Wrapper for 'getParseData'. May be of great
#' use for those developing a package to help see what
#' namespace 'importsFrom' calls will be required.
#'
#' @family Tools
#' @param filename Character. Path to an R file (or directory)
#' containing R code files.
#' @param abc Boolean. List functions alphabetically.
#' If FALSE, will list in order of frequency.
#' @param quiet Boolean. Keep quiet? If not, print messages and
#' \code{statusbar}.
#' @return data.frame. Each row is a function and columns stating
#' number of appearances, percentage, packages, and files searched.
#' @examples
#' \dontrun{
#' # Choose an R script file with functions
#' rfile <- file.choose()
#' files_functions(rfile)
#' }
#' @export
files_functions <- function(filename, abc = TRUE, quiet = FALSE) {
  if (dir.exists(filename)) {
    if (!quiet) message("Importing R files from directory")
    filename <- sprintf("%s/%s", filename, listfiles(filename, regex = "\\.R")$filename)
  }
  results <- data.frame()
  for (i in seq_along(filename)) {
    if (!file.exists(filename[i])) {
      stop("Couldn't find file ", filename[i])
    }
    tmp <- getParseData(parse(filename[i], keep.source = TRUE))
    funs <- tmp$text[which(tmp$token == "SYMBOL_FUNCTION_CALL")] %>%
      freqs(abc = abc) %>%
      rename(fun = 1)
    pkgs <- lapply(funs$fun, find)
    pkgs[unlist(lapply(pkgs, function(x) length(x) == 0))] <- NA
    pkgs <- lapply(pkgs, function(x) v2t(x, quotes = FALSE))
    funs$package <- gsub("package:", "", unlist(pkgs))
    funs$file <- filename[i]
    results <- rbind(results, funs)
    if (length(filename) > 1 && !quiet) {
      statusbar(i, length(filename), filename[i])
    }
  }
  results
}


####################################################################
#' Move files from A to B
#'
#' Move one or more files from a directory to another using R.
#'
#' @family Tools
#' @param from Character. File names and directories. All files
#' will be moved recursively.
#' @param to Character. File names for each \code{from} file or
#' directory. If directory does not exist, it will be created.
#' @return No return value, called for side effects.
#' @export
move_files <- function(from, to) {
  froms <- dirs <- NULL
  for (i in seq_along(from)) {
    fromi <- from[i]
    if (isTRUE(file.info(fromi)$isdir)) {
      fromi <- list.files(fromi, recursive = TRUE)
      fromi <- paste(from[i], fromi, sep = "/")
      dirs <- c(dirs, from[i])
    }
    froms <- c(froms, fromi)
  }
  froms <- froms[grepl("\\.", basename(froms))]
  froms <- gsub(paste0(getwd(), "/"), "", froms)

  tos <- to
  # If it is a directory
  if (length(tos) == 1) {
    if (!isTRUE(file.info(tos)$isdir)) {
      dir.create(tos, recursive = FALSE)
      message(sprintf("Directory '%s' did not exist and was created", tos))
    }
    tos <- paste(to, sub(".*?/", "", froms), sep = "/")
  }

  # Final check for all files
  if (length(froms) == 0) {
    warning(sprintf("No files to move from %s...", from))
    invisible(NULL)
  } else {
    if (length(tos) != length(froms)) {
      stop("Every 'from' must have a respective 'to' filename")
    }
    # Now move/rename all files
    for (i in seq_along(froms)) {
      todir <- dirname(tos[i])
      if (!isTRUE(file.info(todir)$isdir)) {
        dir.create(todir, recursive = FALSE)
      }
      file.rename(from = froms[i], to = basename(tos[i]))
    }
  }
}


####################################################################
#' Calculate the size of any R object or directory
#'
#' @family Tools
#' @inheritParams base::format
#' @inheritParams base::list.files
#' @param x Object
#' @param units Character. Specify which unit to use,
#' i.e. "Gb", "Mb", "Kb".
#' @examples
#' what_size(seq(1:1e3), "Kb")
#' what_size(seq(1:1e6))
#' what_size(as.character(seq(1:1e6)))
#' # what_size(path = ".")
#' @export
what_size <- function(x = NULL, units = "Mb", path = NULL, recursive = TRUE, ...) {
  if (!is.null(path)) {
    size <- dir_size(path, recursive, ...)
    format(size, units = units, ...)
  } else {
    format(object.size(x), units = units, ...)
  }
}

dir_size <- function(path = getwd(), recursive = TRUE, pattern = NULL, ...) {
  stopifnot(is.character(path))
  stopifnot(dir.exists(path))
  # files <- list.files(
  #   path = path, pattern = pattern,
  #   recursive = recursive, full.names = TRUE)
  # size_files <- 0
  # if (length(files) > 0) {
  #   vect_size <- sapply(files, function(x) file.size(x))
  #   size_files <- sum(vect_size, na.rm = TRUE)
  # }
  cmd <- paste("du -hs", path)
  size_files <- system(cmd, intern = TRUE, ignore.stdout = TRUE, ignore.stderr = TRUE)
  size_files <- gsub("\\\t.*", "", as.numeric(size_files))
  size_files <- num_abbr(size_files, numeric = TRUE)
  class(size_files) <- "object_size"
  size_files
}

####################################################################
#' Axis scales format
#'
#' The \code{_comma} ones set comma format for axis text, the \code{_percent}
#' ones set percent format for axis text, \code{_dollar} for collar currency,
#' and \code{_abbr} for abbreviated format. Lastly, use \code{_formatNum} to
#' further customize your numerical scales with \code{formatNum}.
#'
#' @param ... Arguments passed to \code{ggplot2::continuous_scale} or
#' \code{formatNum} depending on the function.
#' @return Reformatted scales on ggplot2 object
#' @examples
#' library(ggplot2)
#' df <- ggplot2::txhousing %>% removenarows(all = FALSE)
#'
#' ggplot(df, aes(x = sales, y = volume)) +
#'   geom_point() +
#'   scale_x_dollar() +
#'   scale_y_abbr()
#'
#' # Use any argument from scale_x/y_continuous
#' ggplot(df, aes(x = listings, y = log(inventory))) +
#'   geom_point() +
#'   scale_x_comma() +
#'   scale_y_percent(limits = c(0, 3))
#'
#' # Use any argument from scale_x/y_continuous AND formatNum
#' ggplot(df, aes(x = median, y = inventory)) +
#'   geom_point() +
#'   scale_x_formatNum(n.breaks = 3, pre = "@", abbr = TRUE) +
#'   scale_y_formatNum(position = "right", decimals = 0, pos = " X")
#' @export
scale_x_comma <- function(...) {
  scale_x_continuous(..., labels = function(x) {
    formatNum(x, decimals = NULL, signif = 3)
  })
}

#' @rdname scale_x_comma
#' @export
scale_y_comma <- function(...) {
  scale_y_continuous(..., labels = function(x) {
    formatNum(x, decimals = NULL, signif = 3)
  })
}

#' @rdname scale_x_comma
#' @export
scale_x_percent <- function(...) {
  scale_x_continuous(..., labels = function(x) {
    formatNum(x * 100, pos = "%", decimals = NULL, signif = 3)
  })
}

#' @rdname scale_x_comma
#' @export
scale_y_percent <- function(...) {
  scale_y_continuous(..., labels = function(x) {
    formatNum(x * 100, pos = "%", decimals = NULL, signif = 3)
  })
}

#' @rdname scale_x_comma
#' @export
scale_x_dollar <- function(...) {
  scale_x_continuous(..., labels = function(x) {
    formatNum(x, pre = "$", decimals = NULL, signif = 3)
  })
}

#' @rdname scale_x_comma
#' @export
scale_y_dollar <- function(...) {
  scale_y_continuous(..., labels = function(x) {
    formatNum(x, pre = "$", decimals = NULL, signif = 3)
  })
}

#' @rdname scale_x_comma
#' @export
scale_x_abbr <- function(...) scale_x_continuous(..., labels = num_abbr)

#' @rdname scale_x_comma
#' @export
scale_y_abbr <- function(...) scale_y_continuous(..., labels = num_abbr)

#' @rdname scale_x_comma
#' @inheritParams formatNum
#' @export
scale_x_formatNum <- function(..., decimals = 2, signif = NULL,
                              type = Sys.getenv("LARES_NUMFORMAT"),
                              pre = "", pos = "", sign = FALSE, abbr = FALSE) {
  scale_x_continuous(..., labels = function(x) {
    formatNum(x,
      decimals = decimals, signif = signif, type = type,
      pre = pre, pos = pos, sign = sign, abbr = abbr
    )
  })
}

#' @rdname scale_x_comma
#' @export
scale_y_formatNum <- function(..., decimals = 2, signif = NULL,
                              type = Sys.getenv("LARES_NUMFORMAT"),
                              pre = "", pos = "", sign = FALSE, abbr = FALSE) {
  scale_y_continuous(..., labels = function(x) {
    formatNum(x,
      decimals = decimals, signif = signif, type = type,
      pre = pre, pos = pos, sign = sign, abbr = abbr
    )
  })
}

####################################################################
#' Lower/Upper Confidence Intervals
#'
#' Calculate lower and upper confidence intervals given a mean,
#' standard deviation, sample size, and confidence level. You may
#' want to use \code{ci_var()} to calculate all values quickly.
#'
#' @family Confidence
#' @param mean Numeric. Mean: \code{mean(var, na.rm = TRUE)}
#' @param ssd Numeric. Standard deviation: \code{sd(var, na.rm = TRUE)}
#' @param n Integer. Amount of observations: \code{n()}
#' @param conf Numeric (0-1). Confidence level.
#' @return Vector with confidence limit value.
#' @examples
#' ci_lower(100, 5, 10)
#' ci_upper(100, 5, 10)
#' @export
ci_lower <- function(mean, ssd, n, conf = 0.95) {
  if (is.na(ssd)[1]) {
    NA
  } else {
    se <- ssd / sqrt(n)
    ci <- mean - qt(1 - ((1 - conf) / 2), n - 1) * se
    ci
  }
}
#' @rdname ci_lower
#' @export
ci_upper <- function(mean, ssd, n, conf = 0.95) {
  if (is.na(ssd)[1]) {
    NA
  } else {
    se <- ssd / sqrt(n)
    ci <- mean + qt(1 - ((1 - conf) / 2), n - 1) * se
    ci
  }
}

####################################################################
#' Confidence Intervals on Dataframe
#'
#' Calculate confidence intervals for a continuous numerical column on
#' a dataframe, given a confidence level. You may also group results
#' using another variable. Tidyverse friendly.
#'
#' @family Confidence
#' @param df Dataframe
#' @param var Variable name. Must be a numerical column.
#' @param group_var Variable name. Group results by another variable.
#' @param conf Numeric. Confidence level (0-1).
#' @return data.frame mean, standard deviation, counter, upper and lower CIs.
#' @examples
#' data(dft) # Titanic dataset
#' ci_var(dft, Fare)
#' ci_var(dft, Fare, Pclass)
#' ci_var(dft, Fare, Pclass, conf = 0.99)
#' @export
ci_var <- function(df, var, group_var = NULL, conf = 0.95) {
  if (isTRUE(try(is.character(var), silent = TRUE))) {
    var <- eval(substitute(var), df)
  } else {
    var <- as_label(enquo(var))
  }
  group_var <- enquo(group_var)

  if (as_label(group_var) != "NULL") {
    df <- df %>% group_by(!!group_var)
  }

  aux <- df %>%
    summarise(
      smean = mean(!!var, na.rm = TRUE),
      ssd = sd(!!var, na.rm = TRUE),
      n = n()
    ) %>%
    mutate(
      lower_ci = ci_lower(.data$smean, .data$ssd, .data$n, conf),
      upper_ci = ci_upper(.data$smean, .data$ssd, .data$n, conf)
    )

  varname <- as_label(var)
  varname <- gsub('"', "", varname)
  cols <- colnames(aux)
  colnames(aux)[cols == "smean"] <- sprintf("%s_mean", varname)
  colnames(aux)[cols == "ssd"] <- sprintf("%s_sd", varname)
  as_tibble(aux)
}

####################################################################
#' Automate Data Preprocess for Modeling
#'
#' Pre-process your data before training a model. This is the prior step
#' on the \code{h2o_automl()} function's pipeline. Enabling for
#' other use cases when wanting too  use any other framework, library,
#' or custom algorithm.
#'
#' @family Machine Learning
#' @inheritParams h2o_automl
#' @param y Character. Column name for dependent variable or response.
#' @return List. Contains original data.frame \code{df}, an index
#' to identify which observations with be part of the train dataset
#' \code{train_index}, and which model type should be \code{model_type}.
#' @examples
#' data(dft) # Titanic dataset
#'
#' model_preprocess(dft, "Survived", balance = TRUE)
#'
#' model_preprocess(dft, "Fare", split = 0.5, scale = TRUE)
#'
#' model_preprocess(dft, "Pclass", ignore = c("Fare", "Cabin"))
#'
#' model_preprocess(dft, "Pclass", quiet = TRUE)
#' @export
model_preprocess <- function(df,
                             y = "tag",
                             ignore = NULL,
                             train_test = NA,
                             split = 0.7,
                             weight = NULL,
                             target = "auto",
                             balance = FALSE,
                             impute = FALSE,
                             no_outliers = TRUE,
                             unique_train = TRUE,
                             center = FALSE,
                             scale = FALSE,
                             thresh = 10,
                             seed = 0,
                             quiet = FALSE) {
  tic(id = "model_preprocess")

  # DEPENDENT VARIABLE
  if (!y %in% colnames(df)) {
    stop(paste(
      "You should have a 'tag' column in your data.frame or select",
      "an dependent variable using the 'y' parameter."
    ))
  }
  if (!quiet) message(sprintf("- DEPENDENT VARIABLE: %s", y))

  colnames(df)[colnames(df) == y] <- "tag"
  df <- data.frame(df) %>%
    filter(!is.na(.data$tag)) %>%
    mutate_if(is.character, as.factor)

  # MODEL TYPE
  cats <- unique(df$tag)
  model_type <- ifelse(length(cats) <= as.integer(thresh), "Classification", "Regression")
  if (!quiet) message("- MODEL TYPE: ", model_type)
  # Change spaces for dots as 'multis' arguments may not match
  if (model_type == "Classification") cats <- gsub(" ", ".", cats)
  # If y variables is named as one of the categories, prediction values will be a problem
  if (model_type == "Classification" && y %in% cats) {
    stop(paste(
      "Your y parameter can't be named as any of the labels used.",
      "Please, rename", y, "into a valid column name next such as",
      paste0(y, "_labels for example.")
    ))
  }

  # If target value is not an existing target value
  if (!target %in% cats && length(cats) == 2) {
    if (!target %in% c(cats, "auto")) {
      stop(paste(
        "Your target value", target, "is not valid.",
        "Possible other values:", vector2text(cats)
      ))
    }
  }
  # When might seem numeric but is categorical
  if (model_type == "Classification" && sum(grepl("^[0-9]", cats)) > 0) {
    df <- mutate(df, tag = as.factor(as.character(
      ifelse(grepl("^[0-9]", .data$tag), paste0("n_", .data$tag), as.character(.data$tag))
    )))
  }
  # When is regression should always be numerical
  if (model_type == "Regression") {
    df$tag <- as.numeric(df$tag)
  }
  # Show a summary of our tags
  if (model_type == "Classification" && !quiet) print(freqs(df, .data$tag))
  if (model_type == "Regression" && !quiet) print(summary(df$tag))

  # MISSING VALUES
  m <- missingness(df, summary = FALSE)
  if (!is.null(m)) {
    m <- mutate(m, label = paste0(.data$variable, " (", .data$missingness, "%)"))
    if (!quiet) {
      top10 <- m %>%
        ungroup() %>%
        slice(1:10)
      which <- vector2text(top10$label, quotes = FALSE)
      if (nrow(m) > 10) {
        which <- paste(which, "and", nrow(m) - 10, "other.")
      }
      message(paste0(
        "- MISSINGS: The following variables contain missing observations: ", which,
        if (!impute && !quiet) ". Consider using the impute parameter."
      ))
    }
    if (impute) {
      if (!quiet) message(paste(">>> Imputing", sum(m$missing), "missing values..."))
      df <- impute(df, seed = seed, quiet = TRUE)
    }
  } else if (!quiet) message("- MISSINGS: No missing values in your data")

  # IGNORED VARIABLES
  if (length(ignore) > 0) {
    ignore <- ignore[ignore %in% colnames(df)]
    if (length(ignore) > 0 && !quiet) {
      message(paste("- SKIPPED: Ignored variables for training models:", vector2text(ignore)))
    }
  }

  # ONE HOT SMART ENCODING
  temp <- df[, !colnames(df) %in% c("tag", ignore)]
  nums <- df_str(temp, "names", quiet = TRUE)$nums
  if (length(nums) != ncol(temp) && !quiet) {
    message(paste(
      "- CATEGORICALS: There are", ncol(temp) - length(nums), "non-numerical features.",
      "Consider using ohse() or equivalent prior to encode categorical variables."
    ))
  }

  # ADDITIONAL TRANSFORMATIONS
  if (scale || center && length(nums) > 0) {
    new <- data.frame(lapply(df[nums], function(x) scale(x, center = center, scale = scale)))
    colnames(new) <- nums
    df[nums] <- new
    msg <- ifelse(scale & center, "scaled and centered", ifelse(scale, "scaled", "centered"))
    if (!quiet) message(paste0("- TRANSFORMATIONS: All numerical features (", length(nums), ") were ", msg))
  }

  # OUTLIERS ON DEPENDENT VARIABLE
  if (is.numeric(df$tag)) {
    thresh <- ifelse(is.numeric(no_outliers), no_outliers, 3)
    is_outlier <- outlier_zscore(df$tag, thresh = thresh)
    if (!quiet && !isTRUE(no_outliers)) {
      message(sprintf(
        "- OUTLIERS: %s (%s) of %s values are considered outliers (Z-Score: >%ssd). %s",
        formatNum(100 * sum(is_outlier) / nrow(df), 1, pos = "%"),
        formatNum(sum(is_outlier), 0), y, thresh,
        ifelse(no_outliers,
          paste(
            "Removing them from the dataset for better results.",
            "To keep them, set the 'no_outliers' parameter."
          ),
          "Consider using the 'no_outliers' parameter to remove them."
        )
      ))
    }
    if (no_outliers) df <- df[!is_outlier, ]
  }

  # SPLIT TRAIN AND TEST DATASETS
  if (is.na(train_test)) {
    if (!quiet) message(sprintf(">>> Splitting data: train = %s &&  test = %s", split, 1 - split))
    splits <- msplit(df, size = split, seed = seed, print = !quiet)
    train <- splits$train
    test <- splits$test
    train_index <- splits$train_index
  } else {
    # If we already have a default split for train and test (train_test)
    if (train_test %in% colnames(df)) {
      colnames(df)[colnames(df) == train_test] <- "train_test"
      if (all(unique(as.character(df$train_test)) %in% c("train", "test"))) {
        train <- filter(df, .data$train_test == "train")
        test <- filter(df, .data$train_test == "test")
        split <- nrow(train) / nrow(df)
        ignore <- c(ignore, train_test)
        train_index <- seq_len(nrow(train))
        if (!quiet) print(table(df$train_test))
      } else {
        stop("Your train_test column should have 'train' and 'test' values only!")
      }
    } else {
      stop(paste("There is no column named", train_test))
    }
  }

  # CHECK TRAINING DATASET
  if (unique_train) {
    train_rows <- nrow(train)
    train <- distinct(train)
    if (nrow(train) != train_rows && !quiet) {
      message(paste(
        "- REPEATED: There were", train_rows - nrow(train),
        "repeated rows which are being suppressed from the train dataset"
      ))
    }
  }
  if (nrow(train) > 10000 && !quiet) {
    message("- SAMPLE: Consider sampling or reduce the 'split' argument for faster results")
  }

  # BALANCE TRAINING SET
  if (model_type == "Classification" && balance) {
    total <- nrow(train)
    min <- freqs(train, .data$tag) %>%
      .$n %>%
      min(., na.rm = TRUE)
    train <- train %>%
      group_by(.data$tag) %>%
      sample_n(min)
    if (!quiet) {
      message(paste0(
        "- BALANCE: Training set balanced: ", min, " observations for each (", length(cats),
        ") category; using ", round(100 * nrow(train) / total, 2), "% of training data"
      ))
    }
  }

  # CHECK TRAIN/TEST VALUES
  if (model_type == "Classification") {
    if (!all(unique(df$tag) %in% unique(train$tag))) {
      warning(paste("You SHOULD train with all available tags:", vector2text(unique(df$tag))))
      temp <- df[df$tag %in% unique(train$tag), ]
      message(sprintf(paste(
        "- CHECK: excluded %s observations from your test set because",
        "those labels were not in your train set"
      ), nrow(df) - nrow(temp)))
      df <- temp
    }
    if (!all(unique(df$tag) %in% unique(test$tag))) {
      warning("You are training with tags that are not in your test set.")
    }
  }

  results <- list(data = df, train_index = train_index, model_type = model_type)
  attr(results, "type") <- "model_preprocess"
  toc(id = "model_preprocess", msg = "Pre-processed in", quiet = TRUE)
  invisible(results)
}

####################################################################
#' Get API (JSON) and Transform into data.frame
#'
#' This function lets the user bring API data as JSON format and
#' transform it into data.frame.
#'
#' @family Tools
#' @family API
#' @param url Character. API's URL to GET.
#' @param status Boolean. Display status message?
#' @return data.frame of \code{url} \code{GET} results or
#' \code{NULL} if no results returned by API.
#' @export
bring_api <- function(url, status = TRUE) {
  get <- GET(url = url)
  if (status) {
    message(paste("Status: ", ifelse(get$status_code == 200, "OK", "ERROR")))
  }
  char <- rawToChar(get$content)
  json <- fromJSON(char)

  if (length(json[[1]]) > 0) {
    import <- data.frame(json)
    import <- flatten(import)
    import <- data.frame(bind_rows(lapply(import, unlist(as.character))))
    import[import == "list()"] <- NA
    import[import == "integer(0)"] <- 0
    colnames(import) <- gsub("\\.", "_", colnames(import))
    import <- suppressMessages(type.convert(import, numerals = "no.loss", as.is = TRUE))
    import
  } else {
    invisible(NULL)
  }
}

#' ####################################################################
#' #' Hubspot Contacts
#' #'
#' #' This function brings ALL contacts with Hubspot's HAPI key
#' #'
#' #' @family Credentials
#' #' @family Hubspot
#' #' @param limit Integer. Limit how many contacts you wish to bring
#' #' @param creds Character. Credential's user (see \code{get_creds()})
#' #' @export
#' hs_contacts <- function(limit=1000, creds = NA) {
#'
#'   # require(httr)
#'   # require(jsonlite)
#'   # require(dplyr)
#'   # require(plyr)
#'   # require(rlist)
#'   # require(anytime)
#'
#'   options(warn=-1)
#'
#'   credentials <- get_credentials("hubspot", dir = creds)
#'   token <- credentials$token
#'
#'   API <- "https://api.hubapi.com/contacts/v1/lists/all/contacts/all?"
#'
#'   order <- c("createdate", "lastmodifieddate", "tipo_contacto", "lifecyclestage",
#'              # PERSONAL DATA
#'              "firstname", "lastname", "email", "fecha_de_nacimiento", "edad", "city", "country", "state",
#'              "phone", "mobilephone", "identificacion", "tipo_documento_identidad", "fecha_expedici_n_documento",
#'              "tipo_de_contrato", "ingresos", "gastos_mensuales", "fecha_ingreso_trabajo", "actividad_econ_mica",
#'              "contacto_deudor", "numero_doc_deudor",
#'              # ANALYTICS
#'              "hubspotscore", "hs_analytics_average_page_views", "num_conversion_events", "hs_analytics_num_event_completions",
#'              "hs_analytics_num_visits", "hs_analytics_last_timestamp", "hs_analytics_last_visit_timestamp",
#'              "hs_social_last_engagement", "hs_email_domain", "hs_lead_status", "hs_analytics_source")
#'
#'   endfx <- function(x) {
#'     colnames(x) <- gsub("contacts_", "", gsub("_value", "", gsub("properties_", "", gsub("profile_", "", colnames(x)))))
#'     x <- data.frame(rlist::list.cbind(lapply(x, unlist(as.character)))) %>%
#'       mutate_at(vars(contains('valid')), funs(as.logical(.))) %>%
#'       mutate_at(vars(contains('date')), funs(as.character(
#'         as.POSIXct(as.numeric(as.character(.))/1000, origin="1970-01-01") - 5*3600))) %>%
#'       mutate_at(vars(contains('timestamp')), funs(as.character(
#'         as.POSIXct(as.numeric(as.character(.))/1000, origin="1970-01-01") - 5*3600))) %>%
#'       select(-identity_profiles, -token, -url, -merge_audits, -has_more, -vid_offset, -addedAt,
#'              -portal_id, -is_contact, -form_submissions, -canonical_vid, -merged_vids) %>%
#'       select(vid, one_of(order), everything()) %>% arrange(desc(createdate))
#'     x[x == "list()"] <- NA
#'     x[x == "integer(0)"] <- 0
#'     return(x)
#'   }
#'
#'   hubspot_limit <- 100
#'   n <- ifelse(limit > hubspot_limit, ceiling(limit/hubspot_limit), 1)
#'   output <- NULL
#'
#'   for (i in 1:n) {
#'
#'     # Define Parameters
#'     offset <- ifelse(i == 1, 0, import$vid_offset[nrow(import)])
#'     limit <- ifelse(i == n, hubspot_limit - ((n*hubspot_limit) - limit), hubspot_limit)
#'     properties <- paste(rbind("property=", order,"&"), collapse="")
#'     parameters <- c(paste0("vidOffset=", offset), paste0("count=", limit), hapikey = paste0("hapikey=", token))
#'     parameters <- paste(parameters, collapse="&")
#'     URL <- paste0(API, properties, parameters)
#'     import <- bring_api(URL)
#'     imports <- select(import, -contains("_versions"), -contains("_source"), -contains("_timestamp"))
#'
#'     output <- plyr::rbind.fill(output, imports)
#'
#'     # Stop if there are no more deals
#'     if (i > 1 && import$has_more[nrow(import)] == FALSE) {
#'       output <- endfx(output)
#'       output <- mutate_at(output, vars(contains('fecha_')), funs(anytime::anydate(as.character(.))))
#'       message(paste0("All contacts (", dim(output)[1],") brought. Stopping my work now!"))
#'       return(output)
#'       break
#'     }
#'
#'     # Finished the loop
#'     if (i==n) {
#'       output <- endfx(output)
#'       output <- mutate_at(output, vars(contains('fecha_')), funs(anytime::anydate(as.character(.))))
#'       message(paste(dim(output)[1],"contacts brought"))
#'       return(output)
#'     }
#'   }
#' }
#'
#'
#' ####################################################################
#' #' Hubspot Deals
#' #'
#' #' This function brings ALL deals with Hubspot's HAPI key
#' #'
#' #' @family Credentials
#' #' @family Hubspot
#' #' @param limit Integer. Limit how many contacts you wish to bring
#' #' @param creds Character. Credential's user (see \code{get_creds()})
#' #' @export
#' hs_deals <- function(limit=10000, creds = NA) {
#'
#'   # require(httr)
#'   # require(jsonlite)
#'   # require(dplyr)
#'   # require(plyr)
#'   # require(rlist)
#'
#'   options(warn=-1)
#'
#'   credentials <- get_credentials("hubspot", dir = creds)
#'   token <- credentials$token
#'
#'   API <- "https://api.hubapi.com/deals/v1/deal/paged?"
#'
#'   order <- c("dealname", "createdate", "dealstage", "amount",
#'              "num_associated_contacts", "dealtype",
#'              "closed_lost_reason", "closedate",
#'              "utm_source", "utm_medium", "utm_campaign",
#'              # VEHICLE AND APPLICATION DATA
#'              "valor_del_veh_culo", "modelo_del_veh_culo", "marca_del_veh_culo",
#'              "vendedor_del_veh_culo", "paso_de_la_compra", "tipo_de_veh_culo", "monto_cuota_inicial",
#'              "fuente_del_veh_culo", "tipo_de_cr_dito", "web_del_veh_culo",
#'              "plazo_del_cr_dito", "ciudad_de_circulaci_n_del_veh_culo", "solicitantes_del_cr_dito",
#'              "concesionario_del_veh_culo", "periodico_revista_del_veh_culo",
#'              # VALIDATION DATA
#'              "valido_politicas", "validacion_tipo_vehiculo", "validacion_edad", "validacion_marca_vehiculo",
#'              "validacion_cuota_inicial", "validacion_ingresos_mensuales", "validacion_monto_prestamo",
#'              "validacion_sector_empresa", "validacion_tipo_documento", "validacion_modelo_vehiculo",
#'              "validacion_tipo_contrato", "validacion_fecha_ingreso", "validacion_ciudad_vehiculo",
#'              "validacion_ciudad_residencia", "validacion_consulta_buro", "validacion_intencion_compra",
#'              # ANALYTICS
#'              "hs_lastmodifieddate", "num_contacted_notes", "notes_next_activity_date",
#'              "notes_last_updated", "engagements_last_meeting_booked", "notes_last_contacted", "num_notes")
#'
#'   endfx <- function(x) {
#'     colnames(x) <- gsub("associations_", "", gsub("_value", "", gsub("properties_", "", gsub("deals_", "", colnames(x)))))
#'     x <- data.frame(rlist::list.cbind(lapply(x, unlist(as.character)))) %>%
#'       mutate_at(vars(contains('valid')), funs(as.logical(.))) %>%
#'       mutate_at(vars(contains('date')), funs(as.character(
#'         as.POSIXct(as.numeric(as.character(.))/1000, origin="1970-01-01") - 5*3600)))
#'     x[x == "list()"] <- NA
#'     x[x == "integer(0)"] <- 0
#'     x <- select(x, -portalId, -hasMore, -imports, -stateChanges, -associatedCompanyIds, -offset)
#'     dealstagesAPI <- paste0("https://api.hubapi.com/deals/v1/pipelines?hapikey=",token)
#'     get <- httr::GET(url = dealstagesAPI)
#'     dealstages <- fromJSON(rawToChar(get$content))
#'     dealstages <- dealstages$stages[[1]]
#'     dealstages <- data.frame(code = as.character(dealstages$stageId),
#'                              dealstages = dealstages$label,
#'                              stage_order = dealstages$displayOrder + 1,
#'                              probability = dealstages$probability)
#'     x <- left_join(x, dealstages, by = c("dealstage"="code")) %>% select(-dealstage) %>%
#'       mutate_at(vars(contains('date')), funs(as.POSIXct(.))) %>%
#'       mutate_at(vars(contains('valid')), funs(as.logical(.))) %>%
#'       select(dealId, associatedVids, dealstages, one_of(order), everything()) %>%
#'       arrange(desc(createdate))
#'     colnames(x) <- gsub("veh_culo", "vehiculo", colnames(x))
#'     return(x)
#'   }
#'
#'   hubspot_limit <- 250
#'   n <- ifelse(limit > hubspot_limit, ceiling(limit/hubspot_limit), 1)
#'   output <- NULL
#'
#'   for (i in 1:n) {
#'
#'     # Define Parameters
#'     offset <- ifelse(i == 1, 0, import$offset[nrow(import)])
#'     limit <- ifelse(i == n, hubspot_limit - ((n*hubspot_limit) - limit), hubspot_limit)
#'     properties <- paste(rbind("properties=", order,"&"), collapse="")
#'     parameters <- c(paste0("offset=", offset), paste0("limit=", limit), hapikey = paste0("hapikey=", token))
#'     parameters <- paste(parameters, collapse="&")
#'     URL <- paste0(API, properties, parameters)
#'     import <- bring_api(URL)
#'
#'     imports <- select(import, everything(),
#'                       -contains("_versions"), -contains("_source"), -contains("_timestamp"),
#'                       contains("utm_source_value"))
#'
#'     output <- plyr::rbind.fill(output, imports)
#'
#'     # Stop if there are no more deals
#'     if (i > 1 && import$hasMore[nrow(import)] == FALSE) {
#'       output <- endfx(output)
#'       message(paste0("All deals (", dim(output)[1],") brought. Stopping my work now!"))
#'       return(output)
#'       break
#'     }
#'
#'     # Finished the loop
#'     if (i==n) {
#'       output <- endfx(output)
#'       message(paste(dim(output)[1], "deals brought"))
#'       return(output)
#'     }
#'   }
#' }

####################################################################
#' Surnames Order Sequence
#'
#' Generate a sequence of numbers that determines the order in which
#' surnames should be listed based on the number of generations of
#' ancestors you wish to include. This sequence follows the traditional
#' Latin custom of assigning the father's surname first, followed by
#' the mother's surname. The same logic extends systematically to higher
#' generations, ensuring that the order of surnames remains consistent
#' as you move upward through the family tree.
#'
#' @param n Integer. Number of generations to include in the sequence.
#' Notice it will generate a vector with 2^(n-1) values.
#' @return Integer vector.
#' @examples
#' seq_surnames(1)
#' seq_surnames(2)
#' seq_surnames(3)
#' seq_surnames(4)
#' @export
seq_surnames <- function(n = 1) {
  stopifnot(!is.integer(n))
  ni <- n - 1 # We calculate only one and then sum 1 to the other half
  for (k in seq(2^ni)) {
    # Always start with 1
    if (k == 1) {
      vals <- NULL
      vals[1] <- 1
    } else {
      # Handle the specific case for n = 2
      if (n == 2) {
        vals[k] <- k
      } else {
        # Generate the first 4 values of seq for larger n
        if (k %in% seq(4)) {
          if (is_odd(k)) {
            vals[k] <- vals[k - 1] - 2^(ni - 1)
          } else {
            vals[k] <- vals[k - 1] + 2^(ni)
          }
        }
      }
    }
  }
  # Now sum the rest of the sequence based on vals
  if (n >= 3) {
    for (i in (n - 2):1) {
      vals <- c(vals, vals + 2^(i - 1))
    }
  }
  vals
}

####################################################################
#' Update the library (dev or CRAN version)
#'
#' This auxiliary function lets the user update \code{lares} to latest
#' CRAN or developer version.
#'
#' @family Tools
#' @param force Boolean. Force install.
#' @param dev Boolean. Developer version (Github)? If not, CRAN version.
#' @param all Boolean. Install other recommended libraries? Kinda Docker install!
#' @param local Boolean. Install package with local files? (or Github repo).
#' @param fb Boolean. From FB instance? Personal internal use.
#' @return No return value, called for side effects.
#' @examples
#' \dontrun{
#' # Update to CRAN version
#' updateLares(dev = FALSE)
#'
#' # Update to development version from GitHub
#' updateLares(dev = TRUE)
#'
#' # Install with all recommended packages
#' updateLares(all = TRUE)
#' }
#' @export
updateLares <- function(force = FALSE, dev = TRUE, all = FALSE, local = FALSE, fb = FALSE) {
  try_require("devtools")

  tic(id = "updateLares")
  message(paste(Sys.time(), "| Started update..."))

  # Auxiliary proxy fx when on devservers
  auxfx <- if (fb) with_proxy else function(x) x

  if (local) {
    install("~/Dropbox (Personal)/Documentos/R/Github/lares")
  } else {
    if (!dev) {
      auxfx(install.packages("lares"))
    } else {
      auxfx(install_github("laresbernardo/lares", force = force))
    }
  }

  if (all) auxfx(install_recommended())

  toc(id = "updateLares", msg = paste(Sys.time(), "|"))
}

####################################################################
#' Cross-MMM Budget Optimization across Channels (fast)
#'
#' Given a list of recreated Robyn models, this function optimizes budget
#' allocation across MMM with respective constraints by maximizing
#' response across all channels. This method assumes each model is
#' independent, that can be compared given its spends were cleanly and
#' properly split, they modeled the same metric (revenue or conversion)
#' and units (currency or type of conversion), and date granularity.
#' Recommended to have same channels granularity across markets
#' to simplify results readings and application.
#'
#' This approach is faster and cleaner compared with previous proposal
#' using \code{robyn_xmodels()}.
#'
#' @param models Lists. Recreated Robyn models with \code{robyn_recreate()}.
#' @param initial_budgets Numeric vector. Default will use the total spends
#' per model for the specified or total date range.
#' Must be length 1 or same as \code{models}.
#' @param start_dates,end_dates Character vector. Start and end dates for each
#' specific model. You can specify a single date and will be used in all models.
#' Default empty value will assume you want all available data and date range.
#' Must be length 1 or same as \code{models}.
#' @param channel_constr_low,channel_constr_up Numeric vector.
#' Relative lower and upper constraints per channel compared with mean
#' spend during the time period defined.
#' If mean was zero for date range, historical mean spend value will be used.
#' Must have length 1 to replicate for all channels or same length
#' (and order )as \code{paid_media_spends}.
#' @return List. Contains optimized allocation results and plots.
#' @examples
#' \dontrun{
#' # Calculate cross-brand optimal allocation
#' res <- robyn_xchannels(
#'   models,
#'   start_dates = "2023-01-01",
#'   end_dates = "2023-12-01"
#' )
#' }
#' @rdname robyn_crossmmm
#' @export
robyn_xchannels <- function(
    models,
    initial_budgets = NULL,
    start_dates = NULL,
    end_dates = NULL,
    channel_constr_low = 0.5,
    channel_constr_up = 2,
    quiet = FALSE,
    ...) {
  try_require("Robyn")
  try_require("nloptr")

  # Check all models are same type: metric and interval type
  stopifnot(length(unique(unlist(lapply(models, function(x) x$InputCollect$dep_var_type)))) == 1)
  stopifnot(length(unique(unlist(lapply(models, function(x) x$InputCollect$intervalType)))) == 1)

  # Identify channels per model
  for (i in seq_along(models)) {
    models[[i]]$OutputCollect$selectIDname <- names(models)[i]
  }
  all_channels <- unlist(
    lapply(models, function(x) {
      paste(x$OutputCollect$selectIDname, x$InputCollect$paid_media_spends, sep = "..")
    }),
    use.names = FALSE
  )

  # Setup same constraints if all not provided
  if (length(channel_constr_low) == 1) channel_constr_low <- rep(channel_constr_low, length(all_channels))
  if (length(channel_constr_up) == 1) channel_constr_up <- rep(channel_constr_up, length(all_channels))
  if (length(channel_constr_low) != length(all_channels) || length(channel_constr_up) != length(all_channels)) {
    stop(paste(
      "Inputs 'channel_constr_low' and 'channel_constr_up' must be",
      "the same length as the total number of channels across all models"
    ))
  }
  channel_constraints <- dplyr::tibble(
    channel = all_channels,
    channel_constr_low = channel_constr_low,
    channel_constr_up = channel_constr_up
  )

  # Initial state calculations
  history <- lapply(models, function(x) {
    robyn_performance(
      x$InputCollect, x$OutputCollect,
      start_date = NULL, end_date = NULL
    )
  })
  history <- lapply(history, function(x) {
    mutate(x, channel = paste(names(models)[i], .data$channel, sep = ".."))
  })

  # Extract and check start and end dates
  start_dates <- extract_dates(start_dates, history, "start_date", quiet)
  end_dates <- extract_dates(end_dates, history, "end_date", quiet)

  # Extract initial budgets, periods, and carryover for the date range
  init_budget <- init_carryover <- init_periods <- initial <- NULL
  for (i in seq_along(models)) {
    p <- robyn_performance(
      models[[i]]$InputCollect, models[[i]]$OutputCollect,
      start_date = start_dates[i], end_date = end_dates[i],
      carryovers = TRUE
    )
    paid_ch <- filter(p, .data$channel %in% models[[i]]$InputCollect$paid_media_spends)
    init_budget <- c(init_budget, sum(paid_ch$spend))
    paid_ch$periods <- ps <- models[[i]]$InputCollect$dt_mod %>%
      filter(.data$ds <= end_dates[i], .data$ds >= start_dates[i]) %>%
      nrow()
    init_periods <- c(init_periods, ps)
    initial <- bind_rows(initial, mutate(paid_ch, channel = paste(
      names(models)[i], .data$channel,
      sep = ".."
    )))
  }
  initial <- arrange(initial, factor(.data$channel, levels = all_channels))

  if (is.null(initial_budgets)) {
    names(init_budget) <- names(models)
    initial_budgets <- init_budget
  } else {
    if (length(initial_budgets) == 1) {
      initial_budgets <- rep(initial_budgets, length(models))
    }
  }

  # Spend values based on date range set
  x0 <- initSpendUnit <- initial$spend / initial$periods

  # Calculate historical carryover
  hist_carrs <- lapply(models, function(x) hist_carryover_calc(x))
  hist_carryover <- unlist(lapply(hist_carrs, function(x) lapply(x, mean)))

  # Total budget across brands
  total_budget_unit <- sum(x0)
  total_budget <- sum(initial_budgets)
  stopifnot(length(models) == length(initial_budgets))

  # Channels with no spend
  zero_spend_channel <- names(x0)[which(x0 == 0)]
  # If no spend within window as initial spend, use historical
  if (length(zero_spend_channel) > 0) {
    if (!quiet) {
      message("Media variables with 0 spending during date range: ", v2t(zero_spend_channel))
    }
    histSpendAllUnit <- bind_rows(lapply(history, function(x) {
      filter(x, .data$channel %in% all_channels)
    })) %>% pull(.data$spend)
    x0[zero_spend_channel] <- histSpendAllUnit[zero_spend_channel]
    # x0[zero_spend_channel] <- 1
  }

  ## Set local data & params values
  hypsregex <- "alphas|gammas|thetas|scales|shapes"
  dt_bestCoef <- bind_rows(lapply(models, function(x) {
    filter(
      x$OutputCollect$xDecompAgg,
      .data$solID == x$OutputCollect$selectID,
      .data$rn %in% x$InputCollect$paid_media_spends
    )
  })) %>%
    mutate(rn = paste(.data$solID, .data$rn, sep = ".."))
  dt_hyppar <- bind_cols(lapply(models, function(x) {
    filter(x$OutputCollect$resultHypParam, .data$solID == x$OutputCollect$selectID) %>%
      select(matches(hypsregex)) %>%
      select(matches(paste(x$InputCollect$paid_media_spends, collapse = "|"))) %>%
      dplyr::rename_with(~ paste(x$OutputCollect$selectIDname, ., sep = ".."))
  }))

  ## Get hill parameters for each channel
  hills <- list(
    alphas = unlist(lapply(models, function(x) {
      x$OutputCollect$resultHypParam %>%
        select(matches(paste(x$InputCollect$paid_media_spends, collapse = "|"))) %>%
        select(any_of(paste0(x$InputCollect$paid_media_spends, "_alphas")))
    })),
    inflexions = unlist(lapply(models, inflextion_points)),
    coefs_sorted = dt_bestCoef$coef
  )
  names(hills$coefs_sorted) <- all_channels

  ## Exclude 0 coef and 0 constraint channels for the optimization
  skip_these <- (channel_constr_low == 0 & channel_constr_up == 0)
  zero_constraint_channel <- all_channels[skip_these]
  if (length(zero_constraint_channel) > 0 && !quiet) {
    message(
      "Excluded variables (constrained to 0): ",
      paste(zero_constraint_channel, collapse = ", ")
    )
  }
  if (any(hills$coefs_sorted == 0)) {
    zero_coef_channel <- all_channels[hills$coefs_sorted == 0]
    if (!quiet) {
      message(
        "Excluded variables (coefficients are 0): ",
        paste(zero_coef_channel, collapse = ", ")
      )
    }
  } else {
    zero_coef_channel <- NULL
  }

  # Gather all values that will be used internally on optim (nloptr)
  no_impact <- c(zero_coef_channel, zero_constraint_channel)
  channel_for_allocation_ids <- which(!all_channels %in% no_impact)

  x0 <- x0[channel_for_allocation_ids]
  channel_constr_low <- channel_constr_low[channel_for_allocation_ids]
  channel_constr_up <- channel_constr_up[channel_for_allocation_ids]

  coefs_eval <- hills$coefs_sorted[channel_for_allocation_ids]
  alphas_eval <- hills$alphas[channel_for_allocation_ids]
  inflexions_eval <- hills$inflexions[channel_for_allocation_ids]
  hist_carryover_eval <- hist_carryover[channel_for_allocation_ids]

  # So we can implicitly use these values within eval_f()
  eval_list <- list(
    coefs_eval = coefs_eval,
    alphas_eval = alphas_eval,
    inflexions_eval = inflexions_eval,
    total_budget = NULL,
    total_budget_unit = total_budget_unit,
    hist_carryover_eval = hist_carryover_eval,
    target_value = NULL,
    target_value_ext = NULL,
    dep_var_type = models[[1]]$InputCollect$dep_var_type
  )
  options("ROBYN_TEMP" = eval_list)

  ## Bounded optimization
  nlsMod <- nloptr(
    x0 = x0 * channel_constr_low,
    eval_f = eval_f,
    eval_g_eq = eval_g_eq,
    eval_g_ineq = NULL,
    lb = x0 * channel_constr_low,
    ub = x0 * channel_constr_up,
    opts = list(
      "algorithm" = "NLOPT_LD_AUGLAG",
      "xtol_rel" = 1.0e-10,
      "maxeval" = 100000,
      "local_opts" = list(
        "algorithm" = "NLOPT_LD_SLSQP",
        "xtol_rel" = 1.0e-10
      )
    ),
    target_value = NULL
  )

  ## Get marginal response
  optmSpendUnit <- nlsMod$solution
  optmResponseUnit <- -eval_f(optmSpendUnit)[["objective.channel"]]
  optmResponseMargUnit <- mapply(
    fx_objective,
    x = optmSpendUnit + 1,
    coeff = coefs_eval,
    alpha = alphas_eval,
    inflexion = inflexions_eval,
    x_hist_carryover = hist_carryover_eval,
    get_sum = FALSE,
    SIMPLIFY = TRUE
  ) - optmResponseUnit

  # Clean auxiliary method
  .Options$ROBYN_TEMP <- NULL

  ## Collect channels outputs
  res <- channel_constraints[channel_for_allocation_ids, ] %>%
    mutate(
      model = gsub("\\.\\..*", "", .data$channel),
      channel = gsub(".*\\.\\.", "", .data$channel),
      start_date = initial$start_date[channel_for_allocation_ids],
      end_date = initial$end_date[channel_for_allocation_ids],
      periods = initial$periods[channel_for_allocation_ids],
      # Spend calculations
      initSpend = initial$spend[channel_for_allocation_ids],
      initSpendUnit = initial$spend[channel_for_allocation_ids] / .data$periods,
      initSpendDistr = .data$initSpend / sum(.data$initSpend),
      optmSpendUnit = optmSpendUnit,
      optmSpendDistr = optmSpendUnit / sum(optmSpendUnit),
      optmSpendChg = optmSpendUnit / initSpendUnit,
      # Response calcs
      initResponse = initial$response[channel_for_allocation_ids],
      initResponseUnit = initial$response[channel_for_allocation_ids] / .data$periods,
      optmResponseUnit = optmResponseUnit,
      optmResponseMargUnit = optmResponseMargUnit,
      # Performance metrics
      initROAS = .data$initResponse / .data$initSpend,
      optmROAS = .data$optmResponseUnit / .data$optmSpendUnit,
      initCPA = .data$initSpend / .data$initResponse,
      optmCPA = .data$optmSpendUnit / .data$optmResponseUnit
    ) %>%
    select("model", everything(), -contains("constr"))

  # Build models summary
  mod <- res %>%
    group_by(.data$model, .data$periods, .data$start_date, .data$end_date) %>%
    summarize(
      initSpend = sum(.data$initSpend),
      initSpendUnit = sum(.data$initSpendUnit),
      initSpendDistr = 0,
      optmSpendUnit = sum(.data$optmSpendUnit),
      optmSpendDistr = 0,
      optmSpendChg = 0,
      initResponseUnit = sum(.data$initResponseUnit),
      optmResponseUnit = sum(.data$optmResponseUnit),
      optmRespChange = 0,
      incrementalUnit = 0
    ) %>%
    ungroup() %>%
    mutate(
      initSpendDistr = .data$initSpend / sum(.data$initSpend),
      optmSpendDistr = .data$optmSpendUnit / sum(.data$initSpendUnit),
      optmSpendChg = .data$optmSpendUnit / .data$initSpendUnit,
      optmRespChange = .data$optmResponseUnit / .data$initResponseUnit,
      incrementalUnit = .data$optmResponseUnit - .data$initResponseUnit,
      incrementalTotal = .data$incrementalUnit * .data$periods
    )

  # Build total summary
  tot <- dplyr::tibble(
    models = length(models),
    incrUnit = sum(mod$incrementalUnit),
    incrTotal = sum(mod$incrementalTotal),
    initSpendUnit = sum(mod$initSpendUnit),
    optmSpendUnit = sum(optmSpendUnit),
    optmSpendChg = sum(optmSpendUnit) / sum(mod$initSpendUnit),
    initROAS = sum(res$initResponse) / sum(res$initSpend),
    optmROAS = sum(res$optmResponseUnit) / sum(res$optmSpendUnit)
  )

  output <- list(total = tot, models = mod, channels = res)
  class(output) <- c("robyn_crossmmm2", class(output))
  output
}

eval_f <- function(X, target_value) {
  eval_list <- getOption("ROBYN_TEMP")
  coefs_eval <- eval_list[["coefs_eval"]]
  alphas_eval <- eval_list[["alphas_eval"]]
  inflexions_eval <- eval_list[["inflexions_eval"]]
  hist_carryover_eval <- eval_list[["hist_carryover_eval"]]

  objective <- -sum(mapply(
    fx_objective,
    x = X,
    coeff = coefs_eval,
    alpha = alphas_eval,
    inflexion = inflexions_eval,
    x_hist_carryover = hist_carryover_eval,
    SIMPLIFY = TRUE
  ))

  gradient <- c(mapply(
    fx_gradient,
    x = X,
    coeff = coefs_eval,
    alpha = alphas_eval,
    inflexion = inflexions_eval,
    x_hist_carryover = hist_carryover_eval,
    SIMPLIFY = TRUE
  ))

  objective.channel <- mapply(
    fx_objective.chanel,
    x = X,
    coeff = coefs_eval,
    alpha = alphas_eval,
    inflexion = inflexions_eval,
    x_hist_carryover = hist_carryover_eval,
    SIMPLIFY = TRUE
  )
  list(objective = objective, gradient = gradient, objective.channel = objective.channel)
}

fx_objective <- function(x, coeff, alpha, inflexion, x_hist_carryover, get_sum = TRUE) {
  xAdstocked <- x + mean(x_hist_carryover)
  if (get_sum) {
    xOut <- coeff * sum((1 + inflexion**alpha / xAdstocked**alpha)**-1)
  } else {
    xOut <- coeff * ((1 + inflexion**alpha / xAdstocked**alpha)**-1)
  }
  xOut
}

fx_gradient <- function(x, coeff, alpha, inflexion, x_hist_carryover) {
  xAdstocked <- x + mean(x_hist_carryover)
  -coeff * sum((alpha * (inflexion**alpha) * (xAdstocked**(alpha - 1))) / (xAdstocked**alpha + inflexion**alpha)**2)
}

fx_objective.chanel <- function(x, coeff, alpha, inflexion, x_hist_carryover) {
  xAdstocked <- x + mean(x_hist_carryover)
  -coeff * sum((1 + inflexion**alpha / xAdstocked**alpha)**-1)
}

eval_g_eq <- function(X, target_value) {
  eval_list <- getOption("ROBYN_TEMP")
  constr <- sum(X) - eval_list$total_budget_unit
  grad <- rep(1, length(X))
  list(
    "constraints" = constr,
    "jacobian" = grad
  )
}

extract_dates <- function(dates, history, date_col, quiet) {
  if (is.null(dates)) {
    dates <- as.Date(
      unlist(lapply(history, function(x) {
        unlist(pull(
          filter(x, .data$channel == "PROMOTIONAL TOTAL"),
          .data[[date_col]]
        ))
      })),
      origin = "1970-01-01"
    )
    if (!quiet) {
      message("Extracted ", date_col, ": ", v2t(sprintf("%s (%s)", dates, names(dates))))
    }
  } else {
    if (length(dates) == 1) dates <- rep(dates, length(history))
    if (length(dates) != length(history)) {
      stop(sprintf(
        "Inputs length for '%s' must be either 1 or same as models (%s)",
        date_col, length(history)
      ))
    }
  }
  dates
}

inflextion_points <- function(model) {
  chnAdstocked <- filter(
    model$OutputCollect$mediaVecCollect,
    .data$type == "adstockedMedia",
    .data$solID == model$OutputCollect$selectID
  ) %>%
    select(all_of(model$InputCollect$paid_media_spends)) %>%
    slice(model$InputCollect$rollingWindowStartWhich:model$InputCollect$rollingWindowEndWhich)
  gammas <- model$OutputCollect$resultHypParam %>%
    select(all_of(paste0(model$InputCollect$paid_media_spends, "_gammas"))) %>%
    unlist()
  inflexions <- unlist(lapply(seq_along(chnAdstocked), function(i) {
    c(range(chnAdstocked[, i]) %*% c(1 - gammas[i], gammas[i]))
  }))
  names(inflexions) <- names(gammas)
  inflexions
}

hist_carryover_calc <- function(model) {
  window_loc <- model$InputCollect$rollingWindowStartWhich:
  model$InputCollect$rollingWindowEndWhich
  ics <- lapply(model$InputCollect$paid_media_vars, function(x) {
    robyn_response(
      metric_name = x,
      dt_hyppar = model$OutputCollect$resultHypParam,
      dt_coef = model$OutputCollect$xDecompAgg,
      InputCollect = model$InputCollect,
      OutputCollect = model$OutputCollect,
      quiet = TRUE,
      is_allocator = TRUE
    )$input_carryover[window_loc]
  })
  names(ics) <- model$InputCollect$paid_media_vars
  dplyr::bind_cols(ics)[, model$InputCollect$paid_media_vars]
}

####################################################################
#' Frequencies Calculations and Plot
#'
#' This function lets the user group, count, calculate percentages
#' and cumulatives. It also plots results if needed. Tidyverse friendly.
#'
#' @family Frequency
#' @family Exploratory
#' @family Visualization
#' @inheritParams mp3_get
#' @param df Data.frame
#' @param ... Variables. Variables you wish to process. Order matters.
#' If no variables are passed, the whole data.frame will be considered
#' @param wt Variable, numeric. Weights.
#' @param rel Boolean. Relative percentages (or absolute)?
#' @param results Boolean. Return results in a dataframe?
#' @param variable_name Character. Overwrite the main variable's name
#' @param plot Boolean. Do you want to see a plot? Three variables tops.
#' @param rm.na Boolean. Remove NA values in the plot? (not filtered for
#' numerical output; use na.omit() or filter() if needed)
#' @param title Character. Overwrite plot's title with.
#' @param subtitle Character. Overwrite plot's subtitle with.
#' @param top Integer. Filter and plot the most n frequent for
#' categorical values. Set to NA to return all values
#' @param abc Boolean. Do you wish to sort by alphabetical order?
#' @param save Boolean. Save the output plot in our working directory
#' @param subdir Character. Into which subdirectory do you wish to
#' save the plot to?
#' @return Plot when \code{plot=TRUE} and data.frame with grouped frequency results
#' when \code{plot=FALSE}.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#'
#' # How many survived?
#' dft %>% freqs(Survived)
#'
#' # How many survived per Class?
#' dft %>% freqs(Pclass, Survived, abc = TRUE)
#'
#' # How many survived per Class with relative percentages?
#' dft %>% freqs(Pclass, Survived, abc = TRUE, rel = TRUE)
#'
#' # Using a weighted feature
#' dft %>% freqs(Pclass, Survived, wt = Fare / 100)
#'
#' ### Let's check the results with plots:
#'
#' # How many survived and see plot?
#' dft %>% freqs(Survived, plot = TRUE)
#'
#' # How many survived per class?
#' dft %>% freqs(Survived, Pclass, plot = TRUE)
#'
#' # Per class, how many survived?
#' dft %>% freqs(Pclass, Survived, plot = TRUE)
#'
#' # Per sex and class, how many survived?
#' dft %>% freqs(Sex, Pclass, Survived, plot = TRUE)
#'
#' # Frequency of tickets + Survived
#' dft %>% freqs(Survived, Ticket, plot = TRUE)
#'
#' # Frequency of tickets: top 10 only and order them alphabetically
#' dft %>% freqs(Ticket, plot = TRUE, top = 10, abc = TRUE)
#' @export
freqs <- function(df, ..., wt = NULL,
                  rel = FALSE,
                  results = TRUE,
                  variable_name = NA,
                  plot = FALSE, rm.na = FALSE,
                  title = NA, subtitle = NA,
                  top = 20, abc = FALSE,
                  save = FALSE, subdir = NA,
                  quiet = FALSE) {
  vars <- quos(...)
  weight <- enquo(wt)

  if (is.vector(df)) {
    values <- df <- data.frame(values = df)
    vars <- quos(values)
  }

  df <- as.data.frame(df)

  # Probably an error but might be useful for the user instead
  # When it's a vector, frequencies; when it's a dataframe, global frequencies
  if (length(vars) == 0) {
    freqs_df(df, plot = plot, save = save, subdir = subdir)
  } else {
    output <- df %>%
      group_by(!!!vars) %>%
      tally(wt = !!weight) %>%
      arrange(desc(.data$n)) %>%
      {
        if (!rel) ungroup(.) else .
      } %>%
      mutate(p = round(100 * .data$n / sum(.data$n), 2))

    # Sort values alphabetically or ascending if numeric
    if (abc) {
      output <- output %>%
        arrange(!!!vars, desc(n)) %>%
        mutate(order = row_number())
    } else {
      output <- output %>%
        arrange(desc(.data$n)) %>%
        mutate(order = row_number())
    }

    # Add cumulative percentage (given the abc logic)
    output <- mutate(output, pcum = cumsum(.data$p))

    # No plot
    if (!plot && !save) {
      if (results) {
        output
      } else {
        invisible(NULL)
      }
    } else {
      if (ncol(output) - 4 >= 4) {
        stop(
          paste(
            "Sorry, but trying to plot more than 3 features is as complex as it sounds.",
            "You should try another method to understand your analysis!"
          )
        )
      }
      obs_total <- sum(output$n)
      obs <- sprintf("Total Obs.: %s", formatNum(obs_total, 0))
      weight <- gsub('"', "", as_label(weight))
      weight_text <- ifelse((weight != "NULL"),
        sprintf("(weighted by %s)", weight), ""
      )

      # Use only the most n frequent values/combinations only
      values <- unique(output[, length(vars)])
      if (nrow(values) > top) {
        if (!is.na(top)) {
          output <- output %>%
            arrange(desc(.data$n)) %>%
            slice(1:top)
          if (!quiet) {
            message(
              sprintf(
                "Slicing the top %s (out of %s) values; use 'top' parameter to overrule.",
                top, nrow(values)
              )
            )
          }
          note <- sprintf("[%s most frequent]", top)
          obs <- sprintf("Obs.: %s (out of %s)", formatNum(sum(output$n), 0), formatNum(obs_total, 0))
        }
      } else {
        note <- ""
      }

      if (abc) note <- gsub("most frequent", "A-Z sorted values", note)

      plot <- ungroup(output)

      if (rm.na) plot <- plot[complete.cases(plot), ]

      # Create some dynamic aesthetics
      plot$labels <- paste0(formatNum(plot$n, decimals = 0), " (", signif(plot$p, 4), "%)")
      plot$label_colours <- ifelse(plot$p > mean(range(plot$p)) * 0.9, "white", "black")
      lim <- 0.35
      plot$label_hjust <- ifelse(
        plot$n < min(plot$n) + diff(range(plot$n)) * lim, -0.1, 1.05
      )
      plot$label_colours <- ifelse(
        plot$label_colours == "white" & plot$label_hjust < lim, "black", plot$label_colours
      )
      variable <- colnames(plot)[1]

      # When one feature
      if (ncol(output) - 3 == 2) {
        type <- 1
        colnames(plot)[1] <- "names"
        p <- ggplot(plot, aes(
          x = reorder(.data$names, -.data$order),
          y = .data$n, label = .data$labels, fill = .data$p
        ))
      }
      # When two features
      else if (ncol(output) - 3 == 3) {
        type <- 2
        facet_name <- colnames(plot)[2]
        colnames(plot)[1] <- "facet"
        colnames(plot)[2] <- "names"
        plot$facet <- suppressWarnings(replacefactor(plot$facet, NA, "NA"))
        p <- plot %>%
          ggplot(aes(
            x = reorder_within(.data$names, -.data$order, .data$facet),
            y = .data$n, label = .data$labels, fill = .data$p
          )) +
          scale_x_reordered()
      }
      # When three features
      else if (ncol(output) - 3 == 4) {
        type <- 3
        facet_name1 <- colnames(plot)[2]
        facet_name2 <- colnames(plot)[3]
        colnames(plot)[1] <- "facet2"
        colnames(plot)[2] <- "facet1"
        colnames(plot)[3] <- "names"
        plot$facet2 <- suppressWarnings(replacefactor(plot$facet2, NA, "NA"))
        plot$facet1 <- suppressWarnings(replacefactor(plot$facet1, NA, "NA"))
        p <- plot %>%
          ggplot(aes(
            x = reorder_within(.data$names, .data$n, .data$facet2),
            y = .data$n, label = .data$labels, fill = .data$p
          )) +
          scale_x_reordered()
      }

      # Plot base
      p <- p + geom_col(alpha = 0.95, width = 0.8, colour = "transparent") +
        geom_text(aes(hjust = .data$label_hjust, colour = .data$label_colours), size = 2.8) +
        coord_flip() + guides(colour = "none") +
        labs(
          x = NULL, y = NULL, fill = NULL,
          title = ifelse(is.na(title), "Frequencies and Percentages", title),
          subtitle = ifelse(is.na(subtitle),
            paste(
              "Variable:", ifelse(!is.na(variable_name), variable_name, variable),
              note, weight_text
            ),
            subtitle
          ), caption = obs
        ) +
        scale_fill_gradient(low = "lightskyblue2", high = "navy") +
        scale_color_identity() +
        theme_lares(which = "c", legend = "none", grid = "Xx") +
        scale_y_comma(position = "right", expand = c(0, 0), limits = c(0, 1.03 * max(output$n)))

      # When two features
      if (type == 2) {
        p <- p +
          facet_grid(as.character(.data$facet) ~ ., scales = "free", space = "free") +
          labs(
            subtitle = ifelse(is.na(subtitle),
              paste("Variables:", facet_name, "grouped by", variable, "\n", note, weight_text),
              subtitle
            ),
            caption = obs
          )
      }

      # When three or more features
      else if (type >= 3) {
        if (length(unique(facet_name2)) > 3) {
          freqs_plot(df, ..., top = top, rm.na = rm.na, title = title, subtitle = subtitle)
        } else {
          p <- p + facet_grid(as.character(.data$facet2) ~ as.character(.data$facet1), scales = "free") +
            labs(
              title = ifelse(is.na(title), "Frequencies and Percentages:", title),
              subtitle = ifelse(
                is.na(subtitle),
                paste(
                  "Variables:", facet_name2, "grouped by", facet_name1, "[x] and",
                  variable, "[y]", "\n", note, weight_text
                ),
                subtitle
              ),
              caption = obs
            )
        }
        # Export file name and folder for plot
        if (save) export_plot(p, "viz_freqs", vars, subdir = subdir)
      }
      p
    }
  }
}

reorder_within <- function(x, by, within, fun = mean, sep = "___", ...) {
  new_x <- paste(x, within, sep = sep)
  stats::reorder(new_x, by, FUN = fun)
}

scale_x_reordered <- function(..., sep = "___") {
  reg <- paste0(sep, ".+$")
  ggplot2::scale_x_discrete(labels = function(x) gsub(reg, "", x), ...)
}


####################################################################
#' Plot for All Frequencies on Dataframe
#'
#' This function lets the user analize data by visualizing the
#' frequency of each value of each column from a whole data frame.
#'
#' @family Frequency
#' @family Exploratory
#' @family Visualization
#' @inheritParams mp3_get
#' @param df Data.frame
#' @param max Numeric. Top variance threshold. Range: (0-1].
#' These variables will be excluded
#' @param min Numeric. Minimum variance threshold. Range: [0-1).
#' These values will be grouped into a high frequency (HF) value
#' @param novar Boolean. Remove no variance columns?
#' @param plot Boolean. Do you want to see a plot? Three variables tops
#' @param top Integer. Plot most relevant (less categories) variables
#' @param save Boolean. Save the output plot in our working directory
#' @param subdir Character. Into which subdirectory do you wish to
#' save the plot to?
#' @return Plot when \code{plot=TRUE} and data.frame with grouped frequency results
#' when \code{plot=FALSE}.
#' @examples
#' data(dft) # Titanic dataset
#' freqs_df(dft)
#' freqs_df(dft, plot = TRUE)
#' @export
freqs_df <- function(df,
                     max = 0.9, min = 0.0, novar = TRUE,
                     plot = FALSE, top = 30,
                     quiet = FALSE,
                     save = FALSE,
                     subdir = NA) {
  out <- NULL
  if (is.vector(df)) {
    temp <- data.frame(values = df)
    freqs(temp, .data$values)
  } else {
    df <- df[!unlist(lapply(df, is.list))]
    names <- lapply(df, function(x) length(unique(x)))
    unique <- as.data.frame(names)
    colnames(unique) <- names(names)
    which <- names(sort(-rank(unique, ties.method = "first")))

    # Too much variance
    no <- names(unique)[unique > nrow(df) * max]
    if (length(no) > 0) {
      if (!quiet) {
        message(paste(
          length(no), "variables with more than", max,
          "variance exluded:", vector2text(no)
        ))
      }
      which <- which[!which %in% no]
    }

    # No variance at all
    if (novar) {
      no <- zerovar(df)
      if (length(no) > 0) {
        if (!quiet) {
          message(paste(
            length(no), "variables with no variance exluded:",
            vector2text(no)
          ))
        }
        which <- which[!which %in% no]
      }
    }

    # Too many columns
    if (length(which) > top) {
      no <- which[(top + 1):length(which)]
      if (!quiet) {
        message(paste(
          "Using the", top, "variables with less distinct categories.",
          length(no), "variables excluded:", vector2text(no)
        ))
      }
      which <- which[1:top]
    }

    if (length(which) > 0) {
      out <- lapply(which, function(iter) {
        counter <- table(df[[iter]], useNA = "ifany")
        data.frame(
          value = names(counter),
          count = as.vector(counter),
          col = iter,
          stringsAsFactors = FALSE
        ) %>% arrange(desc(.data$count))
      }) %>% bind_rows()
      out <- out %>%
        mutate(p = signif(100 * .data$count / nrow(df), 3)) %>%
        mutate(value = ifelse(.data$p > min * 100, as.character(.data$value), "(HF)")) %>%
        group_by(.data$col, .data$value) %>%
        summarise(p = sum(.data$p), count = sum(.data$count)) %>%
        arrange(desc(.data$count)) %>%
        ungroup()
    } else {
      warning("No relevant information to display regarding your data.frame!")
      out <- invisible(NULL)
    }

    if (plot & !is.null(out)) {
      out <- out %>%
        mutate(value = ifelse(is.na(.data$value), "NA", as.character(.data$value))) %>%
        mutate(col = factor(.data$col, levels = which)) %>%
        mutate(label = ifelse(.data$p > 8, as.character(.data$value), "")) %>%
        group_by(.data$col) %>%
        mutate(alpha = log(.data$count)) %>%
        mutate(alpha = as.numeric(ifelse(.data$value == "NA", 0, .data$alpha))) %>%
        mutate(alpha = as.numeric(ifelse(is.infinite(.data$alpha), 0, .data$alpha)))

      p <- ggplot(out, aes(
        x = .data$col, y = .data$count, fill = .data$col,
        label = .data$label, colour = .data$col
      )) +
        geom_col(aes(alpha = .data$alpha),
          position = "fill",
          colour = "80000000", width = 0.95
        ) +
        geom_text(position = position_fill(vjust = .5), size = 3) +
        coord_flip() +
        labs(x = NULL, y = NULL, title = "Overall Values Frequencies") +
        scale_y_percent(expand = c(0, 0)) +
        guides(fill = "none", colour = "none", alpha = "none") +
        theme_lares(pal = 1) +
        theme(
          panel.background = element_blank(),
          panel.grid.major = element_blank(),
          panel.grid.minor = element_blank()
        ) +
        geom_hline(
          yintercept = c(0.25, 0.5, 0.75), linetype = "dashed",
          color = "black", linewidth = 0.5, alpha = 0.3
        )
      if (save) export_plot(p, "viz_freqs_df", subdir = subdir)
      p
    } else {
      if (!is.null(out)) {
        out <- select(out, .data$col, .data$value, .data$count, .data$p) %>%
          rename(n = .data$count, variable = .data$col) %>%
          group_by(.data$variable) %>%
          mutate(pcum = signif(cumsum(.data$p), 3))
      }
      out
    }
  }
}


####################################################################
#' Combined Frequencies Plot for Categorical Features
#'
#' Plot frequencies of multiple categories within a data.frame in
#' a new fancy way. Tidyverse friendly, based on \code{lares::freqs()},
#' no limits on amount of features to evaluate.
#'
#' @family Frequency
#' @family Exploratory
#' @family Visualization
#' @inheritParams freqs
#' @return Plot. Result of the frequency of combined variables.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#'
#' x <- freqs_plot(dft, Pclass, Survived)
#' x$data
#' plot(x)
#'
#' freqs_plot(dft, Pclass, Survived, Sex, Embarked)
#'
#' freqs_plot(dft, Pclass, Survived, Sex, Embarked, top = 15)
#' @export
freqs_plot <- function(df, ..., top = 10, rm.na = FALSE, abc = FALSE,
                       title = NA, subtitle = NA, quiet = FALSE) {
  vars <- quos(...)

  if (length(vars) == 0) {
    freqs(df)
  } else {
    aux <- df %>%
      freqs(!!!vars, rm.na = rm.na) %>%
      mutate_if(is.factor, as.character) %>%
      mutate_at(seq_along(vars), as.character) %>%
      mutate(order = ifelse(.data$order > top, "...", .data$order))
    if ("..." %in% aux$order & !quiet) {
      message(paste(
        "Showing", top, "most frequent values. Tail of",
        nrow(aux) - top,
        "other values grouped into one"
      ))
    }
    aux <- aux %>%
      mutate(across(
        seq_len(ncol(aux) - 4),
        ~ ifelse(order == "...", "Tail", .)
      ))
    aux <- aux %>%
      group_by(!!!vars, .data$order) %>%
      summarise_all(sum) %>%
      ungroup()
    vars_names <- colnames(aux)[1:(ncol(aux) - 4)]

    labels <- aux %>%
      mutate_all(as.character) %>%
      tidyr::gather("name", "value", 1:(ncol(aux) - 4)) %>%
      mutate(label = sprintf("%s: %s", .data$name, .data$value)) %>%
      mutate(
        n = as.integer(.data$n),
        pcum = as.numeric(.data$pcum),
        p = as.numeric(.data$p)
      ) %>%
      ungroup() %>%
      arrange(desc(.data$pcum))

    if (is.na(title)) {
      title <- "Absolute Frequencies"
    }
    if (is.na(subtitle)) {
      subtitle <- paste("Grouped by", vector2text(vars_names, quotes = FALSE))
    }
    mg <- 5

    p1 <- aux %>%
      mutate(aux = ifelse(.data$order == "...", "grey", "black")) %>%
      mutate(order = paste0(.data$order, "\n", signif(as.numeric(.data$p), 2), "%")) %>%
      ggplot(aes(x = reorder(.data$order, .data$pcum), y = .data$n, label = .data$p)) +
      geom_col(aes(fill = .data$aux)) +
      scale_fill_manual(values = c("black", "grey55")) +
      labs(x = NULL, y = NULL, title = title, subtitle = subtitle) +
      scale_y_comma() +
      guides(fill = "none") +
      theme_lares() +
      theme(plot.margin = margin(mg, mg, 0, mg))

    p2 <- labels %>%
      {
        if (abc) mutate(., n = -rank(.data$label)) else .
      } %>%
      mutate(label = ifelse(.data$order == "...", " Mixed Tail", .data$label)) %>%
      ggplot(aes(
        x = reorder(.data$order, .data$pcum),
        y = reorder(.data$label, .data$n),
        group = .data$pcum
      )) +
      geom_point(size = 4) +
      labs(
        x = NULL, y = NULL,
        caption = paste("Total observations:", formatNum(nrow(df), 0))
      ) +
      guides(colour = "none") +
      theme_lares(which = "XY") +
      theme(
        axis.text.x = element_blank(),
        plot.margin = margin(0, mg, mg, mg)
      )
    if (length(vars) > 1) {
      p2 <- p2 + geom_path()
    }

    p <- (p1 / p2)
    attr(p, "freqs_labels") <- labels
    p
  }
}


####################################################################
#' Frequencies on Lists and UpSet Plot
#'
#' Visualize frequency of elements on a list, list vector, or vector with
#' comma separated values. Detect which combinations and elements are
#' the most frequent and how much they represent of your total observations.
#' This is similar to the \href{http://vcg.github.io/upset/}{UpSet Plots}
#' which may be used as an alternative to Venn diagrams.
#'
#' @family Frequency
#' @family Exploratory
#' @family Visualization
#' @inheritParams freqs
#' @param var Variable. Variables you wish to process.
#' @param wt Variable, numeric. Select a numeric column to use
#' in the colour scale, used as sum, mean... of those values for each
#' of the combinations.
#' @param fx Character. Set operation: mean, sum
#' @param rm.na Boolean. Remove NA value from \code{wt}?
#' @param min_elements Integer. Exclude combinations with less than n elements
#' @param limit,limit_x,limit_y Integer. Show top n combinations (x) and/or
#' elements (y). The rest will be grouped into a single element.
#' Set argument to 0 to ignore. \code{limit_x}/\code{limit_y} answer to
#' \code{limit}'s argument.
#' @param tail Boolean. Show tail grouped into "..." on the plots?
#' @param size Numeric. Text base size
#' @param unique Boolean. a,b = b,a?
#' @param plot Boolean. Plot viz? Will be generated anyways in the output object
#' @return List. data.frame with the data results, elements and combinations.
#' @examples
#' \dontrun{
#' df <- dplyr::starwars
#' head(df[, c(1, 4, 5, 12)], 10)
#'
#' # Characters per movies combinations in a list column
#' head(df$films, 2)
#' freqs_list(df, films)
#'
#' # Skin colours in a comma-separated column
#' head(df$skin_color)
#' x <- freqs_list(df, skin_color, min_elements = 2, limit = 5, plot = FALSE)
#' # Inside "x" we'll have:
#' names(x)
#'
#' # Using the 'wt' argument to add a continuous value metric
#' # into an already one-hot encoded columns dataset (and hide tail)
#' csv <- "https://raw.githubusercontent.com/hms-dbmi/UpSetR/master/inst/extdata/movies.csv"
#' movies <- read.csv(csv, sep = ";")
#' head(movies)
#' freqs_list(movies,
#'   wt = AvgRating, min_elements = 2, tail = FALSE,
#'   title = "Movies\nMixed Genres\nRanking"
#' )
#' # So, please: no more Comedy+SciFi and more Drama+Horror films (based on ~50 movies)!
#' }
#' @export
freqs_list <- function(df,
                       var = NULL,
                       wt = NULL, fx = "mean",
                       rm.na = FALSE,
                       min_elements = 1,
                       limit = 10,
                       limit_x = NA,
                       limit_y = NA,
                       tail = TRUE,
                       size = 10,
                       unique = TRUE,
                       abc = FALSE,
                       title = "",
                       plot = TRUE) {
  dff <- df
  var_str <- gsub('"', "", as_label(enquo(var)))
  if (var_str != "NULL") {
    check_opts(var_str, colnames(df))
    colnames(df)[colnames(df) == var_str] <- "which"
  } else {
    # If user has an already one hot encoded data.frame
    duos <- data.frame(
      count = unlist(lapply(df, function(x) length(unique(x))))
    ) %>%
      filter(.data$count == 2)
    if (nrow(duos) == 0) {
      stop("Please, define a valid 'var' column or make sure to have binary/logical columns.")
    }
    duos <- rownames(duos)
    message(paste(">>> Binary columns detected:", v2t(duos)))
    which <- unlist(lapply(df[, duos], function(x) x == 1))
    result <- vapply(seq_len(nrow(df[, duos])), function(i) {
      v2t(colnames(df[, duos])[which[i, ]], quotes = FALSE)
    }, FUN.VALUE = character(1))
    df$which <- result
  }

  # Weighted column
  wt_str <- gsub('"', "", as_label(enquo(wt)))
  if (wt_str == "NULL") wt_str <- NULL
  if (length(wt_str) > 0) {
    if (wt_str %in% colnames(df)) {
      # message(paste(">>> Colour weight:", fx, wt_str))
      colnames(df)[colnames(df) == wt_str] <- "wt"
      weights <- as.numeric(df$wt)
    } else {
      stop(sprintf("'%s' is not a valid weight column", wt_str))
    }
  } else {
    weights <- 0
  }

  # If list then convert
  if (is.list(df$which)) {
    df$which <- unlist(lapply(df$which, function(x) paste(x, collapse = ",")))
  }

  # check if it's a valid column
  if (!any(grepl(",", df$which))) {
    stop(paste(
      "Your inputs may not be valid: no valid binary columns OR",
      "'var' does NOT contain commas nor is a list vector."
    ))
  }

  # Order values so a,b = b,a
  if (unique) {
    values <- strsplit(unlist(df$which), ",")
    values <- lapply(values, sort)
    values <- unlist(lapply(values, function(x) paste(x, collapse = ",")))
  } else {
    values <- df$which
  }

  # One hot encoding
  vals <- data.frame(var = values, wt = weights) %>%
    filter(.data$var != "") %>%
    {
      if (rm.na) filter(., !is.na(.data$wt)) else .
    } %>%
    group_by(.data$var) %>%
    summarise(
      n = n(),
      wt = ifelse(
        fx == "mean", mean(.data$wt, na.rm = TRUE),
        ifelse(fx == "sum", sum(.data$wt, na.rm = TRUE), 1)
      ),
      .groups = "drop"
    ) %>%
    arrange(desc(.data$n)) %>%
    mutate(p = 100 * .data$n / sum(.data$n), order = row_number()) %>%
    data.frame() %>%
    ohe_commas(var) %>%
    as_tibble(.name_repair = "unique") %>%
    # Remove combinations with less than "min_elements" elements
    mutate(combs = rowSums(.[unlist(lapply(., is.logical))], na.rm = TRUE)) %>%
    filter(.data$combs >= min_elements) %>%
    select(-.data$combs)

  # Set limits
  if (limit == 0) limit <- nrow(df)
  if (is.na(limit_x)) limit_x <- nrow(vals)
  if (is.na(limit_y)) limit_y <- nrow(vals)

  elements <- vals %>%
    tidyr::gather(.) %>%
    mutate(n = rep(vals$n, ncol(vals))) %>%
    mutate(wt = rep(vals$wt, ncol(vals))) %>%
    filter(.data$value == TRUE) %>%
    group_by(.data$key) %>%
    summarise(
      n = sum(.data$n),
      wt = ifelse(
        fx == "mean", mean(.data$wt, na.rm = TRUE),
        ifelse(fx == "sum", sum(.data$wt, na.rm = TRUE), 1)
      ),
      .groups = "drop"
    ) %>%
    {
      if (abc) arrange(., .data$key) else arrange(., desc(.data$n))
    } %>%
    mutate(p = 100 * .data$n / nrow(df), order = row_number()) %>%
    mutate(key = as.factor(.data$key)) %>%
    mutate(label = sprintf("%s (%s)", .data$key, formatNum(.data$p, 0, pos = "%"))) %>%
    mutate(label = ifelse(.data$order > min(limit, limit_y), "...", .data$label)) %>%
    mutate(label = as.factor(.data$label))

  tgthr <- vals %>%
    select(contains("var_")) %>%
    tidyr::gather("var") %>%
    mutate(var = factor(.data$var, levels = rev(elements$key))) %>%
    group_by(.data$var) %>%
    mutate(id = row_number(), order = row_number()) %>%
    mutate(label = sprintf("#%s", .data$id)) %>%
    filter(.data$value == TRUE) %>%
    mutate(label = ifelse(.data$id > min(limit, limit_x), "...", .data$label)) %>%
    ungroup() %>%
    mutate(var = ifelse(
      as.character(.data$var) %in% as.character(elements$key[elements$label != "..."]),
      as.character(.data$var), "..."
    )) %>%
    mutate(var = factor(.data$var, levels = rev(c(as.character(elements$key), "..."))))

  # Amount of data considered
  caption <- sprintf(
    "Observations: %s | Elements: %s",
    formatNum(sum(vals$n), 0),
    formatNum(sum(elements$n), 0)
  )
  if (min_elements > 1) {
    caption <- v2t(
      c(caption, sprintf(
        "Excluding combinations with less than %s elements", min_elements
      )),
      quotes = FALSE, sep = "\n"
    )
  }
  if (!tail) caption <- paste(caption, "Tail combinations suppressed from plot", sep = "\n")

  # Scatter plot: combinations
  p1 <- tgthr %>%
    ggplot(aes(
      x = reorder(.data$label, .data$order),
      y = var, group = .data$label
    )) +
    geom_point(size = 3.5) +
    geom_path() +
    theme_lares(size = size, mg = -1, grid = "XY") +
    labs(x = "Combination rank", y = NULL, caption = caption) +
    theme(
      axis.text.y = element_blank(),
      plot.margin = margin(0, 0, 0, 0)
    ) +
    scale_x_discrete(position = "top")

  # Bar plot: elements
  p2 <- elements %>%
    mutate(label = gsub("var_", "", .data$label)) %>%
    ggplot(aes(
      x = reorder(.data$label, -.data$order),
      y = -.data$n, fill = .data$wt
    )) +
    coord_flip() +
    geom_col() +
    labs(y = "Element Size", x = NULL) +
    theme_lares(size = size, mg = -1, grid = "X") +
    theme(plot.margin = margin(0, 0, 0, 0)) +
    scale_y_continuous(
      position = "right",
      labels = function(x) formatNum(abs(x), 0)
    ) +
    guides(fill = "none")

  # Bar plot: combinations
  p3 <- vals %>%
    mutate(var = ifelse(row_number() > min(limit, limit_x), "...", .data$var)) %>%
    {
      if (!tail) mutate(., n = ifelse(.data$var == "...", 1e-8, .data$n)) else .
    } %>%
    ggplot(aes(
      x = reorder(.data$var, .data$order),
      y = .data$n, fill = .data$wt
    )) +
    geom_col() + # ggfittext::geom_bar_text(size = 9) +
    theme_lares(size = size, mg = -1, grid = "Y") +
    scale_y_continuous(labels = function(x) formatNum(abs(x), 0)) +
    labs(
      x = NULL, y = "Combination Size",
      fill = if (length(wt_str) > 0) {
        paste(str_to_title(fx), wt_str, sep = "\n")
      } else {
        NULL
      }
    ) +
    theme(
      axis.text.x = element_blank(),
      plot.margin = margin(0, 0, 0, 0)
    ) +
    scale_fill_continuous(guide = guide_colourbar(direction = "horizontal"))
  if (length(wt_str) == 0) p3 <- p3 + guides(fill = "none")

  # Join all plots
  layout <- "
  ABBB
  CBBB
  DEEE
  DEEE"
  p <- noPlot(title, size = 3.5) + p3 + guide_area() +
    p2 + p1 + plot_layout(design = layout) +
    plot_layout(guides = "collect")
  if (plot) plot(p)

  # Prepare dataframe outputs
  elements <- mutate(elements, key = gsub("var_", "", .data$key)) %>%
    rename("element" = .data$key) %>%
    select(-.data$label)
  vals <- mutate(vals, var = gsub("var_", "", .data$var)) %>%
    rename("combination" = .data$var)
  colnames(vals) <- gsub("var_", "", colnames(vals))

  results <- list(
    plot = p,
    data = rename(df, "combination" = "which"),
    elements = elements,
    combinations = vals
  )

  invisible(results)
}

####################################################################
#' Reduce Dimensionality with PCA
#'
#' Principal component analysis or (PCA) is a method we can use to
#' reduce high-dimensional data to a low-dimensional space. In other words,
#' we cannot accurately visualize high-dimensional datasets because
#' we cannot visualize anything above 3 features. The main purpose behind
#' PCA is to transform datasets with more than 3 features (high-dimensional)
#' into typically a 2/3 column dataset. Despite the reduction into a
#' lower-dimensional space we still can retain most of the variance or
#' information from our original dataset.
#'
#' @family Dimensionality
#' @family Clusters
#' @inheritParams clusterKmeans
#' @param n Integer. Number of dimensions to reduce to.
#' @param plot Boolean. Create plots?
#' @param ... Additional parameters passed to \code{stats::prcomp}
#' @return List with reduced dataframe and possible plots.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data("iris")
#' df <- subset(iris, select = c(-Species))
#' df$id <- seq_len(nrow(df))
#' reduce_pca(df, n = 3, ignore = "id")
#' @export
reduce_pca <- function(df, n = NULL, ignore = NULL,
                       comb = c(1, 2), quiet = FALSE,
                       plot = TRUE, ...) {
  if (isTRUE(is.na(ignore)[1])) ignore <- NULL
  ignore <- unique(ignore)
  df <- .prepare_reduce(df, ignore = ignore, ...)

  if (sum(is.na(df)) > 0) {
    if (!quiet) message("Replacing NA values with column's means...")
    df <- mutate_all(df, ~ ifelse(is.na(.x), mean(.x, na.rm = TRUE), .x))
  }

  PCA <- list()
  temp <- select(df, -any_of(ignore))
  pca <- prcomp(temp, ...)
  PCA$pca_explained <- round(100 * pca$sdev^2 / sum(pca$sdev^2), 4)
  PCA$pcadf <- data.frame(pca$x)[, PCA$pca_explained > 0.01]
  PCA$pcadf <- as_tibble(cbind(select(as.data.frame(df), any_of(ignore)), PCA$pcadf))
  if (!is.null(n)) PCA$pcadf <- PCA$pcadf[, 1:n]

  if (plot) {
    # How much variance is explained by n PCA features?
    pca_explained <- PCA$pca_explained
    PCA$plot_explained <- data.frame(id = seq_along(pca_explained)) %>%
      mutate(
        PC = factor(paste0("PC", .data$id),
          levels = paste0("PC", seq_along(pca_explained))
        ),
        amount = pca_explained
      ) %>%
      mutate(aux = cumsum(.data$amount)) %>%
      ggplot(aes(x = .data$id, y = .data$aux)) +
      geom_col(alpha = 0.9) +
      geom_path() +
      geom_point() +
      labs(
        title = "Principal Component Analysis",
        subtitle = "Percentage of Variation Explained by Components",
        y = "Cumulative variation explained [%]", x = "PC(i)"
      ) +
      theme_lares()

    # Two most relevant PCAs xy plot
    temp <- PCA$pcadf[, comb]
    colnames(temp) <- c("PC1", "PC2")
    explained <- formatNum(pca_explained, 1, pos = "%")
    PCA$plot <- ggplot(PCA$pcadf, aes(x = .data$PC1, y = .data$PC2)) +
      geom_point() +
      labs(
        title = "Dimensions reduction with PCA",
        x = sprintf("PCA Dimension %s", comb[1]),
        y = sprintf("PCA Dimension %s", comb[2]),
        caption = sprintf(
          "Explaining %s of the variance with 2 PCA:\nPC%s (%s), PC%s (%s)",
          formatNum(sum(pca_explained[1:2]), 1, pos = "%"),
          comb[1], explained[comb[1]], comb[2], explained[comb[2]]
        )
      ) +
      theme_lares(pal = 2)
  }
  PCA
}


####################################################################
#' Reduce Dimensionality with t-SNE
#'
#' t-SNE takes high-dimensional data and reduces it to a low-dimensional
#' graph (1-3 dimensions). Unlike PCA, t-SNE can reduce dimensions with
#' non-linear relationships. PCA attempts to draw the best fitting line
#' through the distribution. T-SNE calculates a similarity measure
#' based on the distance between points instead of trying to maximize variance.
#'
#' @family Dimensionality
#' @family Clusters
#' @inheritParams reduce_pca
#' @param ... Additional parameters passed to \code{Rtsne::Rtsne}
#' @return List with reduced dataframe and possible plots.
#' @examples
#' \dontrun{
#' data("iris")
#' df <- subset(iris, select = c(-Species))
#' df$id <- seq_len(nrow(df))
#' reduce_tsne(df, ignore = "id", max_iter = 800, perplexity = 20)
#' }
#' @export
reduce_tsne <- function(df, n = 2, ignore = NULL,
                        quiet = FALSE,
                        plot = TRUE, ...) {
  try_require("Rtsne")
  if (isTRUE(is.na(ignore)[1])) ignore <- NULL
  ignore <- unique(ignore)
  df <- .prepare_reduce(df, ignore = ignore, ...)

  tSNE <- list()

  temp <- select(df, -any_of(ignore))
  tSNE$tsne <- Rtsne(temp, dims = n, verbose = FALSE, ...)

  tSNE$tsne$df <- as_tibble(
    data.frame(
      select(df, any_of(ignore)),
      tSNE$tsne$Y,
      cost = tSNE$tsne$costs
    )
  )

  tSNE$tsne$Y <- tSNE$tsne$costs <- tSNE$tsne$N <- NULL

  if (plot && n >= 2) {
    tSNE$tsne$plot <- tSNE$tsne$df %>%
      ggplot(aes(x = .data$X1, y = .data$X2)) +
      geom_point() +
      labs(
        title = "Dimensions reduction with t-SNE",
        colour = "Cluster",
        x = "t-SNE Dimension 1", y = "t-SNE Dimension 2"
      ) +
      theme_lares(pal = 2)
  }
  tSNE$tsne
}

.prepare_reduce <- function(df, ignore = NULL, quiet = FALSE, norm = TRUE, ...) {
  df <- select(df, -any_of(zerovar(df)))
  df <- ohse(df, quiet = TRUE, ignore = ignore, ...)
  temp <- which(!colnames(df) %in% ignore)
  new_df <- distinct_at(df, temp, .keep_all = TRUE)
  df <- new_df
  if (sum(is.na(df)) > 0) {
    if (!quiet) message(">>> Replacing NA values with column's means...")
    df <- mutate_all(df, ~ ifelse(is.na(.x), mean(.x, na.rm = TRUE), .x))
  }
  if (norm) {
    df <- df %>%
      transmute_if(is.numeric, normalize) %>%
      replace(., is.na(.), 0)
  }
  df
}

####################################################################
#' Calculate predictions of h2o Models
#'
#' \code{h2o_predict_MOJO} lets the user predict using the h2o .zip file
#' containing the MOJO files. Note that it works with the files
#' generated when using the function export_results()
#'
#' @family Machine Learning
#' @family H2O
#' @param df Dataframe. Data to pass to the model.
#' @param model_path Character. Relative path of directory
#' where your zip model file is. If multiple zip files are found,
#' first one found will be used.
#' @param method Character. One of "mojo" or "json".
#' @param batch Integer. Run n batches at a time for "json" method.
#' @return data.frame with predicted results.
#' @export
#' @rdname h2o_predict
h2o_predict_MOJO <- function(df, model_path, method = "mojo", batch = 300) {
  try_require("h2o")
  quiet(h2o.init(nthreads = -1, port = 54321))

  files <- list.files(model_path)
  file <- files[endsWith(files, ".zip")][1]
  zip <- paste0(model_path, "/", file)
  # message(">>> MOJO file: ", zip)

  if (method == "mojo") {
    df <- .quiet_h2o(as.h2o(df))
    mojo_model <- .quiet_h2o(h2o.import_mojo(zip))
    output <- .quiet_h2o(as.data.frame(h2o.predict(mojo_model, df)))
  }

  if (method == "json") {
    df <- as.data.frame(df)
    df <- mutate_if(df, is.logical, as.character)
    aux <- ceiling(nrow(df) / batch)
    df$aux <- rep(1:aux, each = batch)[seq_len(nrow(df))]
    output <- NULL
    for (i in 1:aux) {
      dfi <- select(df[df$aux == i, ], -.data$aux)
      json <- toJSON(dfi)
      size <- nchar(json)
      if (size > 250000) {
        stop(paste(
          "JSON batch is too long. Please, try with a smaller 'batch' parameter.",
          "Suggested size:", round(batch * 235000 / size)
        ))
      }
      res <- h2o.predict_json(zip, json, labels = TRUE)
      if ("error" %in% names(res)) {
        message("\nERROR: There was an issue with one of the inputs to predict...")
        writeLines(res$error)
        break
      }
      output <- rbind(output, res)
      if (aux > 1) statusbar(i, aux, i * batch)
    }
    if ("classProbabilities" %in% names(output)) {
      aux <- .flatten_list(output$classProbabilities, quiet = TRUE)
      colnames(aux) <- output$responseDomainValues[[1]]
      output <- cbind(output[, c(1, 2)], aux)
    }
  }
  as_tibble(output)
}


####################################################################
#' H2O Predict using Binary file
#'
#' \code{h2o_predict_binary} lets the user predict using the h2o binary file.
#' Note that it works with the files generated when using the
#' function export_results(). Recommendation: use the
#' h2o_predict_MOJO() function when possible - it let's you change
#' h2o's version without problem.
#'
#' @param df Dataframe. Data to insert into the model.
#' @param model_path Character. Relative model path directory or zip file.
#' @param sample Integer. How many rows should the function predict?
#' @return vector with predicted results.
#' @export
#' @rdname h2o_predict
h2o_predict_binary <- function(df, model_path, sample = NA) {
  message("Use of h2o_predict_MOJO instead highly recommended!")
  try_require("h2o")
  quiet(h2o.init(nthreads = -1, port = 54321))

  if (!right(model_path, 4) == ".zip") {
    binary <- paste(model_path, gsub(".*-", "", model_path), sep = "/")
  } else {
    binary <- model_path
  }

  model <- h2o.loadModel(binary)

  if (!is.na(sample)) df <- df[1:sample, ]

  score_binary <- as.vector(predict(model, as.h2o(df))[, 3])
  score_binary
}

####################################################################
#' H2O Predict using H2O Model Object
#'
#' \code{h2o_predict_model} lets the user get scores from a H2O Model Object.
#'
#' @param df Dataframe/Vector. Data to insert into the model.
#' @param model h2o model Object
#' @return data.frame with predicted results.
#' @export
#' @rdname h2o_predict
h2o_predict_model <- function(df, model) {
  as.data.frame(predict(model, as.h2o(df)))
}


####################################################################
#' H2O Predict using API Service
#'
#' \code{h2o_predict_API} lets the user get the score from an API service
#'
#' @param df Dataframe/Vector. Data to insert into the model.
#' @param api Character. API URL.
#' @param exclude Character. Name of the variables to exclude.
#' @return vector with predicted results.
#' @export
#' @rdname h2o_predict
h2o_predict_API <- function(df, api, exclude = "tag") {
  post <- function(df, api) {
    df <- df %>%
      removenacols() %>%
      select(-one_of(exclude))
    x <- POST(
      api,
      add_headers("Content-Type" = "application/json"),
      body = as.list(df),
      encode = "json"
    )
    content(x)$probabilityToOne
  }
  batch <- NULL
  for (i in seq_len(nrow(df))) {
    x <- df[i, ]
    score <- post(x, api)
    batch <- rbind(batch, score)
  }
  as.vector(batch)
}


.flatten_list <- function(x, quiet = FALSE) {
  n <- length(x)
  for (i in 1:n) {
    if (i == 1) ret <- NULL
    values <- unlist(x[[i]])
    aux <- data.frame(t(values))
    ret <- suppressWarnings(bind_rows(ret, aux))
    if (n > 500 && !quiet) statusbar(i, n, i)
    if (i == n) ret <- as_tibble(ret)
  }
  ret
}

####################################################################
#' Density plot for discrete and continuous values
#'
#' This function plots discrete and continuous values results
#'
#' @family ML Visualization
#' @param tag Vector. Real known label
#' @param score Vector. Predicted value or model's result
#' @param thresh Integer. Threshold for selecting binary or regression
#' models: this number is the threshold of unique values we should
#' have in 'tag' (more than: regression; less than: classification)
#' @param model_name Character. Model's name
#' @param subtitle Character. Subtitle to show in plot
#' @param save Boolean. Save output plot into working directory
#' @param subdir Character. Sub directory on which you wish to save the plot
#' @param file_name Character. File name as you wish to save the plot
#' @return Plot with distribution and performance results.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr[c(1, 3)], head)
#'
#' # Plot for binomial results
#' mplot_density(dfr$class2$tag, dfr$class2$scores, subtitle = "Titanic Survived Model")
#'
#' # Plot for regression results
#' mplot_density(dfr$regr$tag, dfr$regr$score, model_name = "Titanic Fare Model")
#' @export
mplot_density <- function(tag,
                          score,
                          thresh = 6,
                          model_name = NA,
                          subtitle = NA,
                          save = FALSE,
                          subdir = NA,
                          file_name = "viz_distribution.png") {
  if (length(tag) != length(score)) {
    message("The tag and score vectors should be the same length.")
    stop(message(paste("Currently, tag has", length(tag), "rows and score has", length(score))))
  }

  if (length(unique(tag)) <= thresh) {
    out <- data.frame(tag = as.character(tag), score = score)

    if (is.numeric(out$score)) {
      if (max(out$score) <= 1) {
        out$score <- score * 100
      }
    }

    p1 <- ggplot(out) +
      geom_density(
        aes(
          x = as.numeric(.data$score),
          group = .data$tag,
          fill = as.character(.data$tag)
        ),
        alpha = 0.6, adjust = 0.25, size = 0
      ) +
      labs(
        title = "Classification Model Results",
        y = "Density by tag", x = "Score", fill = NULL
      ) +
      theme_lares(pal = 1, legend = "bottom") +
      theme(
        legend.title = element_blank(),
        legend.key.size = unit(.2, "cm")
      )

    p2 <- ggplot(out) +
      geom_density(aes(x = .data$score),
        size = 0, alpha = 0.9, adjust = 0.25, fill = "deepskyblue"
      ) +
      labs(x = NULL, y = "Density") +
      theme_lares()

    p3 <- ggplot(out, aes(x = as.numeric(.data$score), colour = as.character(.data$tag))) +
      stat_ecdf(size = 1) +
      ylab("Cumulative") +
      labs(x = NULL) +
      guides(colour = "none") +
      theme_lares(pal = 2)

    p1 <- p1 + theme(plot.margin = margin(10, 5, 5, 5))
    p2 <- p2 + theme(plot.margin = margin(0, 0, 5, 5))
    p3 <- p3 + theme(plot.margin = margin(0, 5, 5, 0))

    if (!is.na(subtitle)) p1 <- p1 + labs(subtitle = subtitle)
    if (!is.na(model_name)) p1 <- p1 + labs(caption = model_name)

    if (!is.na(subdir)) {
      # dir.create(file.path(getwd(), subdir), recursive = TRUE)
      file_name <- paste(subdir, file_name, sep = "/")
    }

    p <- (p1 / (p2 | p3)) + plot_layout(heights = c(1.2, 1))
  } else {
    df <- data.frame(
      rbind(
        cbind(values = tag, type = "Real"),
        cbind(values = score, type = "Model")
      )
    )
    df$values <- as.numeric(as.character(df$values))

    p <- ggplot(df) +
      geom_density(aes(x = .data$values, fill = as.character(.data$type)),
        alpha = 0.6, adjust = 0.25
      ) +
      labs(y = "Density", x = "Continuous values", fill = NULL) +
      guides(colour = "none") +
      theme_lares(pal = 1, legend = "top")

    if (!is.na(model_name)) p <- p + labs(caption = model_name)
    if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
  }
  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)
  p
}


####################################################################
#' Variables Importances Plot
#'
#' This function plots Variable Importances
#'
#' @family ML Visualization
#' @inheritParams mplot_density
#' @param var Vector. Variable or column's names
#' @param imp Vector. Importance of said variables. Must have same length as var
#' @param colours If positive and negative contribution is known
#' @param limit Integer. Limit how many variables you wish to plot
#' @return Plot with ranked importance variables results.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' df <- data.frame(
#'   variable = LETTERS[1:6],
#'   importance = c(4, 6, 6.7, 3, 4.8, 6.2) / 100,
#'   positive = c(TRUE, TRUE, FALSE, TRUE, FALSE, FALSE)
#' )
#' head(df)
#'
#' mplot_importance(
#'   var = df$variable,
#'   imp = df$importance,
#'   model_name = "Random values model"
#' )
#'
#' # Add a colour for categories
#' mplot_importance(
#'   var = df$variable,
#'   imp = df$importance,
#'   colours = df$positive,
#'   limit = 4
#' )
#' @export
mplot_importance <- function(var,
                             imp,
                             colours = NA,
                             limit = 15,
                             model_name = NA,
                             subtitle = NA,
                             save = FALSE,
                             subdir = NA,
                             file_name = "viz_importance.png") {
  if (is.null(imp)) {
    NULL
  } else {
    if (length(var) != length(imp)) {
      message("The variables and importance values vectors should be the same length.")
      stop(message(paste("Currently, there are", length(var), "variables and", length(imp), "importance values!")))
    }

    if (is.na(colours[1])) {
      colours <- "deepskyblue"
    }

    out <- data.frame(var = var, imp = 100 * imp, Type = colours)

    if (length(var) < limit) limit <- length(var)

    output <- out[1:limit, ]

    p <- ggplot(
      output,
      aes(
        x = reorder(.data$var, .data$imp), y = .data$imp,
        label = formatNum(.data$imp, 1)
      )
    ) +
      geom_col(aes(fill = .data$Type), width = 0.08, colour = "transparent") +
      geom_point(aes(colour = .data$Type), size = 6.2) +
      coord_flip() +
      geom_text(hjust = 0.5, size = 2.1, inherit.aes = TRUE, colour = "white") +
      labs(
        title = paste0("Most Relevant Variables (top ", limit, " of ", length(var), ")"),
        x = NULL, y = NULL
      ) +
      scale_y_continuous(
        position = "right", expand = c(0, 0),
        limits = c(0, 1.03 * max(output$imp))
      ) +
      theme_lares()

    if (length(unique(output$Type)) == 1) {
      p <- p + geom_col(fill = colours, width = 0.2, colour = "transparent") +
        geom_point(colour = colours, size = 6) +
        guides(fill = "none", colour = "none") +
        geom_text(hjust = 0.5, size = 2, inherit.aes = TRUE, colour = "white")
    }

    if (!is.na(model_name)) p <- p + labs(caption = model_name)
    if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
    if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)
    p
  }
}


####################################################################
#' ROC Curve Plot
#'
#' This function plots ROC Curves with AUC values with 95\% confidence
#' range. It also works for multi-categorical models.
#'
#' @family ML Visualization
#' @param tag Vector. Real known label.
#' @param score Vector. Predicted value or model's result.
#' @param multis Data.frame. Containing columns with each category probability
#' or score (only used when more than 2 categories coexist).
#' @param sample Integer. Number of samples to use for rendering plot.
#' @param model_name Character. Model's name
#' @param subtitle Character. Subtitle to show in plot
#' @param interval Numeric. Interval for breaks in plot
#' @param squared Boolean. Keep proportions?
#' @param plotly Boolean. Use plotly for plot's output for an interactive plot
#' @param save Boolean. Save output plot into working directory
#' @param subdir Character. Sub directory on which you wish to save the plot
#' @param file_name Character. File name as you wish to save the plot
#' @return Plot with ROC curve and AUC performance results.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr[c(1, 2)], head)
#'
#' # ROC Curve for Binomial Model
#' mplot_roc(dfr$class2$tag, dfr$class2$scores,
#'   model_name = "Titanic Survived Model"
#' )
#'
#' # ROC Curves for Multi-Categorical Model
#' mplot_roc(dfr$class3$tag, dfr$class3$score,
#'   multis = subset(dfr$class3, select = -c(tag, score)),
#'   squared = FALSE,
#'   model_name = "Titanic Class Model"
#' )
#' @export
mplot_roc <- function(tag,
                      score,
                      multis = NA,
                      sample = 1000,
                      model_name = NA,
                      subtitle = NA,
                      interval = 0.2,
                      squared = TRUE,
                      plotly = FALSE,
                      save = FALSE,
                      subdir = NA,
                      file_name = "viz_roc.png") {
  if (is.na(multis)[1]) {
    rocs <- ROC(tag, score)
    ci <- rocs$ci
  } else {
    rocs <- ROC(tag, score, multis)
    ci <- rocs$ci["mean"]
  }
  coords <- rocs$roc

  if (sample < min(table(coords$label))) {
    coords <- coords %>%
      group_by(.data$label) %>%
      sample_n(sample)
    message("ROC Curve Plot rendered with sampled data...")
  }

  scale <- function(x) sprintf("%.1f", x)
  p <- ggplot(coords, aes(x = .data$fpr, y = .data$tpr, group = .data$label)) +
    geom_line(colour = "deepskyblue", size = 0.8) +
    geom_point(aes(colour = .data$label), size = 0.7, alpha = 0.8) +
    geom_segment(aes(x = 0, y = 1, xend = 1, yend = 0), alpha = 0.2, linetype = "dotted") +
    scale_x_reverse(
      name = "1 - Specificity [False Positive Rate]", limits = c(1, 0),
      breaks = seq(0, 1, interval), expand = c(0.001, 0.001),
      labels = scale
    ) +
    scale_y_continuous(
      name = "Sensitivity [True Positive Rate]", limits = c(0, 1),
      breaks = seq(0, 1, interval), expand = c(0.001, 0.001),
      labels = scale
    ) +
    theme(axis.ticks = element_line(color = "grey80")) +
    labs(title = "ROC Curve: AUC", colour = NULL) +
    guides(colour = guide_legend(ncol = 3)) +
    annotate("text",
      x = 0.25, y = 0.10, size = 4.2,
      label = paste("AUC =", round(100 * ci[2, ], 2))
    ) +
    annotate("text",
      x = 0.25, y = 0.05, size = 2.8,
      label = paste0(
        "95% CI: ", round(100 * ci[1, ], 2), "-",
        round(100 * ci[3, ], 2)
      )
    ) +
    theme_lares(plot_colour = "white", pal = 2, legend = "bottom")

  if (squared) p <- p + coord_equal()
  if (is.na(multis)[1]) p <- p + guides(colour = "none")
  if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
  if (!is.na(model_name)) p <- p + labs(caption = model_name)

  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)

  if (plotly) {
    try_require("plotly")
    p <- ggplotly(p)
  }
  p
}


####################################################################
#' Cuts by quantiles for score plot
#'
#' This function cuts by quantiles any score or prediction.
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @param splits Integer. Numer of separations to plot
#' @param table Boolean. Do you wish to return a table with results?
#' @return Plot with performance results by cuts.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' head(dfr$class2)
#'
#' # Data
#' mplot_cuts(dfr$class2$scores, splits = 5, table = TRUE)
#'
#' # Plot
#' mplot_cuts(dfr$class2$scores, model_name = "Titanic Survived Model")
#' @export
mplot_cuts <- function(score,
                       splits = 10,
                       model_name = NA,
                       subtitle = NA,
                       table = FALSE,
                       save = FALSE,
                       subdir = NA,
                       file_name = "viz_ncuts.png") {
  if (splits > 25) stop("You should try with less splits!")

  deciles <- quantile(score,
    probs = seq((1 / splits), 1, length = splits),
    names = TRUE
  )
  deciles <- data.frame(
    range = row.names(as.data.frame(deciles)),
    cuts = as.vector(signif(deciles, 6))
  )
  rownames(deciles) <- NULL

  p <- deciles %>%
    # mutate(label_colours = ifelse(cuts*100 < 50, "1", "m")) %>%
    ggplot(aes(x = reorder(.data$range, .data$cuts), y = .data$cuts * 100)) +
    geom_col(fill = "deepskyblue", colour = "transparent") +
    xlab("Cumulative volume") +
    ylab("Score") +
    geom_text(
      aes(
        label = round(100 * .data$cuts, 1),
        vjust = ifelse(.data$cuts * 100 < 50, -0.3, 1.3)
      ),
      size = 3, colour = "black", inherit.aes = TRUE, check_overlap = TRUE
    ) +
    guides(colour = "none") +
    labs(title = sprintf("Score cuts (%s quantiles)", splits)) +
    theme_lares()

  if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
  if (!is.na(model_name)) p <- p + labs(caption = model_name)

  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)

  if (table) {
    deciles
  } else {
    p
  }
}


####################################################################
#' Cuts by quantiles on absolute and percentual errors plot
#'
#' This function cuts by quantiles on absolute and percentual errors
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @param splits Integer. Number of separations to plot
#' @param title Character. Title to show in plot
#' @return Plot with error results by cuts.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' head(dfr$regr)
#' mplot_cuts_error(dfr$regr$tag, dfr$regr$score,
#'   model_name = "Titanic Fare Model"
#' )
#' @export
mplot_cuts_error <- function(tag,
                             score,
                             splits = 10,
                             title = NA,
                             model_name = NA,
                             save = FALSE,
                             subdir = NA,
                             file_name = "viz_ncuts_error.png") {
  if (splits > 25) stop("You should try with less splits!")

  if (length(tag) != length(score)) {
    message("The tag and score vectors should be the same length.")
    stop(message(paste("Currently, tag has", length(tag), "rows and score has", length(score))))
  }

  df <- data.frame(tag = tag, score = score) %>%
    mutate(
      real_error = .data$tag - .data$score,
      abs_error = abs(.data$real_error),
      p_error = 100 * .data$real_error / .data$tag
    ) %>%
    filter(abs(.data$p_error) <= 150)

  # Useful function
  quantsfx <- function(values, splits = 10, just = 0.3) {
    cuts <- quantile(values,
      probs = seq((1 / splits), 1, length = splits),
      names = TRUE
    )
    cuts <- data.frame(deciles = names(cuts), cut = cuts)
    thresh <- max(cuts$cut) / 2
    cuts$gg_pos <- ifelse(cuts$cut > thresh, 1 + just, -just)
    cuts$colour <- ifelse(cuts$gg_pos < 0, "f", "m")
    row.names(cuts) <- NULL
    cuts
  }

  # First: absolute errors
  deciles_abs <- quantsfx(df$abs_error, splits = splits, just = 0.3)
  p_abs <- ggplot(deciles_abs, aes(
    x = reorder(.data$deciles, .data$cut),
    y = .data$cut, label = signif(.data$cut, 3)
  )) +
    geom_col(fill = "deepskyblue", colour = "transparent") +
    labs(x = NULL, y = "Absolute [#]") +
    geom_text(aes(vjust = .data$gg_pos, colour = .data$colour),
      size = 2.7, inherit.aes = TRUE, check_overlap = TRUE
    ) +
    labs(subtitle = paste("Cuts and distribution by absolute error")) +
    scale_y_comma() +
    guides(colour = "none") +
    theme_lares(pal = 4, plot_colour = "white")

  # Second: percentual errors
  deciles_perabs <- quantsfx(abs(df$p_error), splits = splits, just = 0.3)
  p_per <- ggplot(
    deciles_perabs,
    aes(
      x = reorder(.data$deciles, .data$cut),
      y = .data$cut, label = signif(.data$cut, 3)
    )
  ) +
    geom_col(fill = "deepskyblue", colour = "transparent") +
    labs(x = NULL, y = "Absolute [%]") +
    geom_text(aes(vjust = .data$gg_pos, colour = .data$colour),
      size = 2.7, inherit.aes = TRUE, check_overlap = TRUE
    ) +
    labs(subtitle = paste("Cuts and distribution by absolute percentage error")) +
    scale_y_comma() +
    guides(colour = "none") +
    theme_lares(pal = 4, plot_colour = "white")

  # Third: errors distribution
  pd_error <- ggplot(df) +
    geom_density(aes(x = .data$p_error),
      fill = "deepskyblue", alpha = 0.7
    ) +
    labs(x = NULL, y = "Density [%]") +
    geom_vline(xintercept = 0, alpha = 0.5, colour = "navy", linetype = "dotted") +
    theme_lares(plot_colour = "white")

  if (!is.na(title)) p_abs <- p_abs + labs(title = title)
  if (!is.na(model_name)) pd_error <- pd_error + labs(caption = model_name)

  p <- (p_abs + p_per + pd_error) +
    plot_layout(ncol = 1, heights = c(1.8, 1.8, 1))

  if (save) {
    if (!is.na(subdir)) {
      # dir.create(file.path(getwd(), subdir), recursive = TRUE)
      file_name <- paste(subdir, file_name, sep = "/")
    }
    png(file_name, height = 1800, width = 1800, res = 300)
    plot(p)
    dev.off()
  }
  p
}


####################################################################
#' Split and compare quantiles plot
#'
#' This function lets us split and compare quantiles on a given prediction to
#' compare different categorical values vs scores grouped by equal sized buckets.
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @param splits Integer. Number of separations to plot
#' @return Plot with distribution and performance results by splits.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr, head)
#'
#' # For categorical (binary) values
#' mplot_splits(dfr$class2$tag, dfr$class2$scores,
#'   splits = 4,
#'   model_name = "Titanic Survived Model"
#' )
#'
#' # For categorical (+2) values
#' mplot_splits(dfr$class3$tag, dfr$class2$scores,
#'   model_name = "Titanic Class Model"
#' )
#'
#' # For continuous values
#' mplot_splits(dfr$regr$tag, dfr$regr$score,
#'   splits = 4,
#'   model_name = "Titanic Fare Model"
#' )
#' @export
mplot_splits <- function(tag,
                         score,
                         splits = 5,
                         subtitle = NA,
                         model_name = NA,
                         save = FALSE,
                         subdir = NA,
                         file_name = "viz_splits.png") {
  if (length(tag) != length(score)) {
    message("The tag and score vectors should be the same length.")
    stop(message(paste("Currently, tag has", length(tag), "rows and score has", length(score))))
  }

  if (splits > 10) stop("You should try with less splits!")

  df <- data.frame(tag, score)
  npersplit <- round(nrow(df) / splits)

  # For continuous tag values
  if (length(unique(tag)) > 6) {
    names <- df %>%
      mutate(
        tag = as.numeric(.data$tag),
        quantile = ntile(.data$tag, splits)
      ) %>%
      group_by(.data$quantile) %>%
      summarise(
        n = n(),
        max_score = round(max(.data$tag), 1),
        min_score = round(min(.data$tag), 1)
      ) %>%
      mutate(quantile_tag = paste0(quantile, " (", .data$min_score, "-", .data$max_score, ")"))
    df <- df %>%
      mutate(quantile = ntile(.data$tag, splits)) %>%
      left_join(names, by = "quantile") %>%
      mutate(tag = .data$quantile_tag) %>%
      select(-.data$quantile, -.data$n, -.data$max_score, -.data$min_score)
  } else {
    # For categorical tag values
    names <- df %>%
      mutate(quantile = ntile(.data$score, splits)) %>%
      group_by(.data$quantile) %>%
      summarise(
        n = n(),
        max_score = signif(max(.data$score), 2),
        min_score = signif(min(.data$score), 2)
      ) %>%
      mutate(quantile_tag = paste0(
        .data$quantile, " (",
        round(100 * .data$min_score, 1), "-",
        round(100 * .data$max_score, 1), ")"
      ))
  }

  p <- df %>%
    mutate(quantile = ntile(.data$score, splits)) %>%
    group_by(.data$quantile, .data$tag) %>%
    tally() %>%
    ungroup() %>%
    group_by(.data$tag) %>%
    arrange(desc(.data$quantile)) %>%
    mutate(
      p = round(100 * .data$n / sum(.data$n), 2),
      cum = cumsum(100 * .data$n / sum(.data$n))
    ) %>%
    left_join(names, by = "quantile") %>%
    ggplot(aes(
      x = as.character(.data$tag), y = .data$p, label = as.character(.data$p),
      fill = reorder(as.character(.data$quantile_tag), .data$quantile)
    )) +
    geom_col(position = "stack", colour = "transparent") +
    geom_text(size = 3, position = position_stack(vjust = 0.5), check_overlap = TRUE) +
    xlab("Tag") +
    ylab("Total Percentage by Tag") +
    guides(fill = guide_legend(title = paste0("~", npersplit, " p/split"))) +
    labs(title = "Split Groups") +
    scale_fill_brewer(palette = "Spectral") +
    theme_lares()

  if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
  if (!is.na(model_name)) p <- p + labs(caption = model_name)

  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)

  p
}


####################################################################
#' Model Metrics and Performance Plots
#'
#' This function generates plots of the metrics of a predictive model.
#' This is an auxiliary function used in \code{model_metrics()} when
#' the parameter \code{plot} is set to \code{TRUE}.
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @param results Object. Results object from h2o_automl function
#' @return Plot with \code{results} performance.
#' @export
mplot_metrics <- function(results,
                          subtitle = NA,
                          model_name = NA,
                          save = FALSE,
                          subdir = NA,
                          file_name = "viz_metrics.png") {
  scoring_history <- results$model@model$scoring_history

  # Determine x-axis variable (trees or iterations)
  if ("number_of_trees" %in% names(scoring_history)) {
    x_var <- "number_of_trees"
    x_label <- "# of trees"
  } else if ("iteration" %in% names(scoring_history)) {
    x_var <- "iteration"
    x_label <- "# of iterations"
  } else if ("iterations" %in% names(scoring_history)) {
    x_var <- "iterations"
    x_label <- "# of iterations"
  } else {
    warning("No trees or iterations found in scoring history")
    return(NULL)
  }

  # Safely extract metrics, handling missing validation data
  plots_data <- data.frame(
    trees = scoring_history[[x_var]],
    train_ll = scoring_history$training_logloss,
    train_auc = scoring_history$training_auc
  )

  # Add validation metrics if they exist
  if (!is.null(scoring_history$validation_logloss)) {
    plots_data$test_ll <- scoring_history$validation_logloss
  }
  if (!is.null(scoring_history$validation_auc)) {
    plots_data$test_auc <- scoring_history$validation_auc
  }

  ll <- ggplot(plots_data) +
    geom_hline(yintercept = 0.69315, alpha = 0.5, linetype = "dotted") +
    geom_line(aes(x = .data$trees, y = .data$train_ll, colour = "Train"), size = 0.5) +
    labs(
      title = paste("Logarithmic Loss vs", x_label),
      colour = "Dataset", x = x_label, y = "LogLoss"
    ) +
    scale_colour_brewer(palette = "Set1") +
    geom_text(
      aes(
        x = .data$trees, y = .data$train_ll, colour = "Train",
        label = round(.data$train_ll, 2)
      ),
      check_overlap = TRUE, nudge_y = 0.03, size = 3
    )

  if ("test_ll" %in% names(plots_data)) {
    ll <- ll +
      geom_line(aes(x = .data$trees, y = .data$test_ll, colour = "Test"), size = 1) +
      geom_text(
        aes(
          x = .data$trees, y = .data$test_ll, colour = "Test",
          label = round(.data$test_ll, 2)
        ),
        check_overlap = TRUE, nudge_y = 0.03, size = 3
      )
  }

  ll <- ll +
    theme_lares(pal = 1) +
    theme(
      strip.text.x = element_blank(),
      strip.background = element_rect(colour = "white", fill = "white"),
      legend.position = c(0.1, 0.05)
    )

  au <- ggplot(plots_data) +
    geom_line(aes(x = .data$trees, y = .data$train_auc * 100, colour = "Train"), size = 0.5) +
    labs(
      title = "Area Under the Curve vs Number of Trees",
      colour = "Dataset", x = "# of trees", y = "AUC"
    ) +
    scale_colour_brewer(palette = "Set1") +
    geom_text(
      aes(
        x = .data$trees, y = .data$train_auc * 100, colour = "Train",
        label = round(.data$train_auc * 100, 2)
      ),
      check_overlap = TRUE, nudge_y = 0.03, size = 3
    )

  if ("test_auc" %in% names(plots_data)) {
    au <- au +
      geom_line(aes(x = .data$trees, y = .data$test_auc * 100, colour = "Test"), size = 1) +
      geom_text(
        aes(
          x = .data$trees, y = .data$test_auc * 100, colour = "Test",
          label = round(.data$test_auc * 100, 2)
        ),
        check_overlap = TRUE, nudge_y = 0.03, size = 3
      )
  }

  au <- au +
    theme_lares(pal = 1) +
    theme(
      strip.text.x = element_blank(),
      strip.background = element_rect(colour = "white", fill = "white"),
      legend.position = c(0.9, 0.05)
    )
  p <- (ll / au) + plot_layout(ncol = 1)

  if (!is.na(subtitle)) ll <- ll + labs(subtitle = subtitle)
  if (!is.na(model_name)) au <- au + labs(caption = model_name)

  if (save) {
    if (!is.na(subdir)) {
      # dir.create(file.path(getwd(), subdir), recursive = TRUE)
      file_name <- paste(subdir, file_name, sep = "/")
    }

    png(file_name, height = 1800, width = 2100, res = 300)
    plot(p)
    dev.off()
  }

  plot(p)
}


####################################################################
#' Linear Regression Results Plot
#'
#' This function plots a Linear Regression Result
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @return Plot with linear distribution and performance results.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr, head)
#' mplot_lineal(dfr$regr$tag, dfr$regr$score, model_name = "Titanic Fare Model")
#' @export
mplot_lineal <- function(tag,
                         score,
                         subtitle = NA,
                         model_name = NA,
                         save = FALSE,
                         subdir = NA,
                         file_name = "viz_lineal.png") {
  if (length(tag) != length(score)) {
    message("The tag and score vectors should be the same length.")
    stop(message(paste("Currently, tag has", length(tag), "rows and score has", length(score))))
  }

  results <- data.frame(tag = tag, score = score, dist = 0)
  results <- results[complete.cases(results), ]

  for (i in seq_len(nrow(results))) {
    results$dist[i] <- dist2d(c(results$tag[i], results$score[i]), c(0, 0), c(1, 1))
  }

  fit <- lm(results$score ~ results$tag)
  labels <- paste(
    paste("Adj R2 = ", signif(summary(fit)$adj.r.squared, 4)),
    # paste("Pval =", signif(summary(fit)$coef[2,4], 3)),
    paste("RMSE =", signif(rmse(results$tag, results$score), 4)),
    paste("MAE =", signif(mae(results$tag, results$score), 4)),
    sep = "\n"
  )

  p <- ggplot(results, aes(x = .data$tag, y = .data$score, colour = .data$dist)) +
    geom_point() +
    labs(
      title = "Regression Model Results",
      x = "Real value", y = "Predicted value",
      colour = "Deviation"
    ) +
    annotate("text", x = Inf, y = -Inf, hjust = 1, vjust = -0.05, label = labels, size = 2.8) +
    scale_x_comma() +
    scale_y_comma() +
    scale_colour_continuous(labels = function(x) formatNum(x, decimals = NULL, signif = 3)) +
    theme(legend.justification = c(0, 1), legend.position = c(0, 1)) +
    guides(colour = guide_colorbar(barwidth = 0.9, barheight = 4.5)) +
    theme_lares()

  # Draw reference lines for correlation
  intercept <- summary(fit)$coefficients[1]
  slope <- summary(fit)$coefficients[2]
  p <- p +
    geom_abline(
      slope = 1,
      intercept = 0,
      alpha = 0.3,
      colour = "grey70",
      size = 0.6
    ) +
    geom_abline(
      slope = slope,
      intercept = intercept,
      alpha = 0.5,
      colour = "orange",
      size = 0.6
    )

  if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
  if (!is.na(model_name)) p <- p + labs(caption = model_name)
  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)
  p
}


####################################################################
#' MPLOTS Score Full Report Plots
#'
#' This function plots a whole dashboard with a model's results. It will automatically
#' detect if it's a categorical or regression's model by checking how many different
#' unique values the dependent variable (tag) has.
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @param splits Integer. Number of separations to plot
#' @param thresh Integer. Threshold for selecting binary or regression
#' models: this number is the threshold of unique values we should
#' have in 'tag' (more than: regression; less than: classification)
#' @param plot Boolean. Plot results? If not, plot grid object returned
#' @return Multiple plots gathered into one, showing \code{tag} vs
#' \code{score} performance results.
#' @examples
#' \donttest{
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr, head)
#'
#' # Dasboard for Binomial Model
#' mplot_full(dfr$class2$tag, dfr$class2$scores,
#'   model_name = "Titanic Survived Model"
#' )
#'
#' # Dasboard for Multi-Categorical Model
#' mplot_full(dfr$class3$tag, dfr$class3$score,
#'   multis = subset(dfr$class3, select = -c(tag, score)),
#'   model_name = "Titanic Class Model"
#' )
#'
#' # Dasboard for Regression Model
#' mplot_full(dfr$regr$tag, dfr$regr$score,
#'   model_name = "Titanic Fare Model"
#' )
#' }
#' @export
mplot_full <- function(tag,
                       score,
                       multis = NA,
                       splits = 8,
                       thresh = 6,
                       subtitle = NA,
                       model_name = NA,
                       plot = TRUE,
                       save = FALSE,
                       subdir = NA,
                       file_name = "viz_full.png") {
  if (length(tag) != length(score)) {
    message("The tag and score vectors should be the same length.")
    stop(message(paste("Currently, tag has", length(tag), "rows and score has", length(score))))
  }


  # Categorical Binomial Models
  if (length(unique(tag)) == 2 && is.numeric(score)) {
    p1 <- mplot_density(tag = tag, score = score, subtitle = subtitle, model_name = model_name)
    p2 <- mplot_splits(tag = tag, score = score, splits = splits) +
      theme(plot.margin = margin(10, 8, 5, 0))
    p3 <- mplot_roc(tag = tag, score = score, squared = FALSE) +
      theme(plot.margin = margin(0, 8, 8, 0))
    p4 <- mplot_cuts(score = score) +
      theme(plot.margin = margin(-3, 0, 5, 8))

    p <- (wrap_plots(p1) + p2 + p4 + p3) +
      plot_layout(
        ncol = 2, nrow = 2,
        heights = c(1.4, 1),
        widths = c(1.5, 1)
      )
  }

  # Multi-Categorical Models
  if (length(unique(tag)) > 2 && length(unique(tag)) <= thresh) {
    m <- model_metrics(tag, score, multis, thresh = thresh)
    p1 <- m$plots$conf_matrix +
      labs(
        title = "Confusion Matrix",
        caption = if (!is.na(model_name)) model_name
      )
    p2 <- m$plots$ROC
    p <- (p1 + p2) + plot_layout(ncol = 2, nrow = 1)
  }

  # Regression Continuous Models
  if (is.numeric(tag) && is.numeric(score) && length(unique(tag)) > thresh) {
    p1 <- mplot_lineal(tag = tag, score = score, subtitle = subtitle, model_name = model_name) +
      theme_lares(plot_colour = "white")
    p2 <- mplot_density(tag = tag, score = score)
    p3 <- mplot_cuts_error(tag = tag, score = score, splits = splits)
    p <- ((p1 / p2) | p3) + plot_layout(widths = c(1, 1.4))
  }

  if (save) export_plot(p, file_name, subdir = subdir, width = 15, height = 10)

  if (plot) plot(p) else p
}


####################################################################
#' Confussion Matrix Plot
#'
#' This function plots a confussion matrix.
#'
#' You may use \code{conf_mat()} to get calculate values.
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @inheritParams conf_mat
#' @param abc Boolean. Arrange columns and rows alphabetically?
#' @param squared Boolean. Force plot to be squared?
#' @param top Integer. Plot only the most n frequent variables.
#' Set to \code{NA} to plot all.
#' @return Plot with confusion matrix results.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr, head)
#'
#' # Plot for Binomial Model
#' mplot_conf(dfr$class2$tag, dfr$class2$scores,
#'   model_name = "Titanic Survived Model"
#' )
#'
#' # Plot for Multi-Categorical Model
#' mplot_conf(dfr$class3$tag, dfr$class3$score,
#'   model_name = "Titanic Class Model"
#' )
#' @export
mplot_conf <- function(tag, score, thresh = 0.5, abc = TRUE,
                       squared = FALSE, diagonal = TRUE, top = 20,
                       subtitle = NA, model_name = NULL,
                       save = FALSE, subdir = NA,
                       file_name = "viz_conf_mat.png") {
  df <- data.frame(tag, score)

  # About tags
  # Keep only most frequent tags?
  if (!is.na(top)) {
    if (length(unique(df$tag)) > top) {
      tops <- freqs(df, tag) %>%
        pull(1) %>%
        head(top)
      df <- df %>% mutate(
        tag = ifelse(.data$tag %in% tops, as.character(.data$tag), "Others"),
        score = ifelse(.data$score %in% tops, as.character(.data$score), "Others")
      )
    }
  }
  values <- df %>%
    group_by(.data$tag) %>%
    tally() %>%
    {
      if (abc) arrange(., .data$tag) else arrange(., desc(.data$n))
    } %>%
    mutate(label = sprintf("%s \n(%s)", .data$tag, formatNum(.data$n, 0)))
  labels <- values$tag
  df <- mutate(df, tag = factor(.data$tag, levels = labels))

  if (!diagonal) df <- filter(df, .data$tag != .data$score)

  # About scores
  if (is.numeric(df$score) && length(unique(tag)) == 2) {
    means <- df %>%
      group_by(.data$tag) %>%
      summarise(mean = mean(.data$score))
    target <- means$tag[means$mean == max(means$mean)]
    other <- means$tag[means$mean == min(means$mean)]
    df <- df %>% mutate(pred = ifelse(
      .data$score >= thresh, as.character(target), as.character(other)
    ))
  } else {
    df <- df %>% mutate(pred = .data$score)
  }

  # Frequencies
  plot_cf <- df %>%
    freqs(.data$tag, .data$pred) %>%
    ungroup() %>%
    group_by(.data$tag) %>%
    mutate(aux = 100 * .data$n / sum(.data$n)) %>%
    mutate(label = paste0(
      formatNum(.data$n, 0), "\n",
      # formatNum(.data$p, 1), "%T\n",
      "(", formatNum(.data$aux, 1), "%)"
    ))
  trues <- sum(plot_cf$n[as.character(plot_cf$tag) == as.character(plot_cf$pred)])
  total <- sum(plot_cf$n)
  acc <- formatNum(100 * (trues / total), 2, pos = "%")
  obs <- formatNum(nrow(df), 0)
  metrics <- sprintf("%s observations | AAC %s", obs, acc)

  p <- ggplot(plot_cf, aes(
    y = as.numeric(factor(.data$tag, levels = rev(labels))),
    x = as.numeric(factor(.data$pred, levels = labels)),
    fill = .data$n, size = .data$aux, label = .data$label
  )) +
    geom_tile() +
    scale_fill_gradient(low = "white", high = "orange") +
    geom_text(lineheight = .8) +
    scale_size(range = c(2.9, 3.4)) +
    guides(fill = "none", size = "none", colour = "none") +
    labs(
      x = "Predicted values", y = "Real values",
      title = paste("Confusion Matrix", ifelse(
        thresh != 0.5, paste("with Threshold =", thresh), ""
      ), ifelse(
        diagonal == FALSE, paste("without diagonal values"), ""
      )),
      subtitle = metrics, caption = model_name
    ) +
    theme_lares() +
    theme(
      axis.text.x = element_text(angle = 30, hjust = 0),
      axis.title.x = element_text(hjust = 0.5),
      axis.text.x.bottom = element_blank(),
      axis.ticks.x.bottom = element_blank(),
      axis.text.y.right = element_blank(),
      axis.ticks.y.right = element_blank(),
      panel.background = element_blank(),
      panel.border = element_blank(),
      panel.grid.minor = element_blank(),
      strip.background = element_blank()
    ) +
    theme(axis.text.x = element_text(angle = 30, hjust = 0)) +
    scale_x_continuous(
      breaks = seq_along(labels),
      labels = labels,
      position = "bottom",
      sec.axis = sec_axis(~.,
        breaks = seq_along(labels),
        labels = labels
      )
    ) +
    scale_y_continuous(
      breaks = seq_along(labels),
      labels = labels,
      position = "right",
      sec.axis = sec_axis(~.,
        breaks = seq_along(labels),
        labels = rev(values$label)
      )
    )

  if (!is.na(subtitle)) p <- p + labs(subtitle = subtitle)
  if (squared) p <- p + coord_equal()
  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)
  p
}


####################################################################
#' Cumulative Gain Plot
#'
#' The cumulative gains plot, often named ‘gains plot’, helps us
#' answer the question: When we apply the model and select the best
#' X deciles, what % of the actual target class observations can we
#' expect to target? The cumulative gains chart shows the percentage
#' of the overall number of cases in a given category "gained" by
#' targeting a percentage of the total number of cases.
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @inheritParams mp3_get
#' @param target Value. Which is your target positive value? If
#' set to 'auto', the target with largest mean(score) will be
#' selected. Change the value to overwrite. Only works for binary classes
#' @param splits Integer. Numer of quantiles to split the data
#' @param highlight Character or Integer. Which split should be used
#' for the automatic conclussion in the plot? Set to "auto" for
#' best value, "none" to turn off or the number of split.
#' @param caption Character. Caption to show in plot
#' @return Plot with gain and performance results by cuts.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr, head)
#'
#' # Plot for Binomial Model
#' mplot_gain(dfr$class2$tag, dfr$class2$scores,
#'   caption = "Titanic Survived Model",
#'   target = "FALSE"
#' )
#' mplot_gain(dfr$class2$tag, dfr$class2$scores,
#'   caption = "Titanic Survived Model",
#'   target = "TRUE"
#' )
#'
#' # Plot for Multi-Categorical Model
#' mplot_gain(dfr$class3$tag, dfr$class3$score,
#'   multis = subset(dfr$class3, select = -c(tag, score)),
#'   caption = "Titanic Class Model"
#' )
#' @export
mplot_gain <- function(tag, score, multis = NA, target = "auto",
                       splits = 10, highlight = "auto",
                       caption = NA, save = FALSE, subdir = NA,
                       file_name = "viz_gain.png", quiet = FALSE) {
  if (is.na(multis)[1]) {
    gains <- gain_lift(tag, score, target, splits, quiet = quiet)
    aux <- data.frame(x = c(0, gains$percentile), y = c(0, gains$optimal))
    df <- mutate(gains, percentile = as.numeric(.data$percentile))
    p <- ggplot(df, aes(x = .data$percentile)) +
      # Range area
      geom_polygon(data = aux, aes(.data$x, .data$y), alpha = 0.1) +
      # Model line
      geom_line(aes(y = .data$gain), colour = "darkorange", size = 1.2) +
      geom_label(data = df[df$gain != 100, ], aes(
        y = .data$gain,
        label = round(.data$gain)
      ), alpha = 0.9) +
      guides(colour = "none") +
      scale_y_continuous(breaks = seq(0, 100, 10)) +
      scale_x_continuous(minor_breaks = NULL, breaks = seq(0, splits, 1)) +
      labs(
        title = paste("Cumulative Gains for", gains$value[1]),
        linetype = NULL,
        y = "Cumulative gains [%]",
        x = paste0("Percentiles [", splits, "]")
      ) +
      theme_lares(pal = 2) +
      theme(legend.position = c(0.88, 0.2))

    if (highlight == "auto") {
      highlight <- gains %>%
        arrange(desc(.data$lift)) %>%
        slice(1) %>%
        .$percentile
    }
    if (highlight %in% gains$percentile && highlight != "none") {
      highlight <- as.integer(highlight)
      note <- paste0(
        "If we select the top ",
        round(highlight * 100 / splits), "% cases with highest probabilities,\n",
        round(gains$gain[gains$percentile == highlight]), "% of all target class will be picked ",
        "(", round(gains$lift[gains$percentile == highlight]), "% better than random)"
      )
      p <- p + labs(subtitle = note)
    } else {
      message("That highlight value is not a percentile. Try any integer from 1 to ", splits)
    }
  } else {
    df <- data.frame(tag = tag, score = score, multis)
    out <- aux <- NULL
    for (i in 1:(ncol(df) - 2)) {
      g <- gain_lift(df$tag, df[, 2 + i], splits = splits, quiet = quiet) %>%
        mutate(label = colnames(df)[2 + i])
      x <- data.frame(
        x = as.factor(c(0, g$percentile)),
        y = c(0, g$optimal),
        label = as.character(g$label[1])
      )
      out <- rbind(out, g)
      aux <- rbind(aux, x)
    }
    p <- out %>%
      mutate(percentile = factor(.data$percentile, levels = unique(out$percentile))) %>%
      ggplot(aes(x = .data$percentile, group = .data$label)) +
      # # Random line
      # geom_line(aes(y = random, linetype = "Random"), colour = "black") +
      # # Optimal line
      # geom_line(aes(y = optimal, colour = label, linetype = "Optimal"), size = 0.4) +
      # Possible area
      geom_polygon(data = aux, aes(x = .data$x, y = .data$y, group = .data$label), alpha = 0.1) +
      # Model line
      geom_line(aes(y = .data$gain, colour = .data$label), size = 1) +
      geom_label(aes(y = .data$gain, label = round(.data$gain)), alpha = 0.8) +
      guides(colour = "none") +
      theme_lares(pal = 2) +
      labs(
        title = "Cumulative Gains for Multiple Labels",
        subtitle = paste(
          "If we select the top nth percentile with highest probabilities,",
          "\nhow much of that specific target class will be picked?"
        ),
        x = paste0("Percentiles [", splits, "]"),
        y = "Cumulative Gains [%]",
        linetype = "Reference", colour = "Label"
      ) +
      scale_y_continuous(minor_breaks = seq(0, 100, 10), breaks = seq(0, 100, 20)) +
      facet_grid(label ~ .)
  }

  aux <- freqs(data.frame(tag = tag), .data$tag)
  text <- paste(
    "Assuming rate of",
    vector2text(round(aux$p), sep = "/", quotes = FALSE), "for",
    vector2text(aux$tag, sep = "/", quotes = FALSE)
  )
  if (!is.na(caption)) caption <- paste(text, caption, sep = "\n") else caption <- text
  p <- p + labs(caption = caption)
  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)
  p
}


####################################################################
#' Cumulative Response Plot
#'
#' The response gains plot helps us answer the question: When we
#' apply the model and select up until ntile X, what is the expected
#' % of target class observations in the selection?
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @inheritParams mplot_gain
#' @return Plot with cumulative response and performance results by cuts.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' lapply(dfr, head)
#'
#' # Plot for Binomial Model
#' mplot_response(dfr$class2$tag, dfr$class2$scores,
#'   caption = "Titanic Survived Model",
#'   target = "TRUE"
#' )
#' mplot_response(dfr$class2$tag, dfr$class2$scores,
#'   caption = "Titanic Survived Model",
#'   target = "FALSE"
#' )
#'
#' # Plot for Multi-Categorical Model
#' mplot_response(dfr$class3$tag, dfr$class3$score,
#'   multis = subset(dfr$class3, select = -c(tag, score)),
#'   caption = "Titanic Class Model"
#' )
#' @export
mplot_response <- function(tag, score, multis = NA, target = "auto",
                           splits = 10, highlight = "auto",
                           caption = NA, save = FALSE, subdir = NA,
                           file_name = "viz_response.png", quiet = FALSE) {
  if (is.na(multis)[1]) {
    gains <- gain_lift(tag, score, target, splits, quiet = quiet) %>%
      mutate(
        percentile = as.numeric(.data$percentile),
        cum_response = 100 * cumsum(.data$target) / cumsum(.data$total)
      )
    target <- gains$target[1]
    rand <- 100 * sum(gains$target) / sum(gains$total)
    gains <- gains %>% mutate(cum_response_lift = 100 * .data$cum_response / rand - 100)

    p <- gains %>%
      ggplot(aes(x = .data$percentile)) +
      theme_lares(pal = 2) +
      geom_hline(yintercept = rand, colour = "black", linetype = "dashed") +
      geom_line(aes(y = .data$cum_response, colour = "x"), size = 1.2) +
      geom_label(aes(y = .data$cum_response, label = round(.data$cum_response)), alpha = 0.9) +
      geom_text(
        label = paste0("Random: ", round(rand, 1), "%"),
        y = rand, x = 1, vjust = 1.2, hjust = 0, alpha = 0.2
      ) +
      scale_y_continuous(limits = c(0, 100), breaks = seq(0, 100, 10)) +
      scale_x_continuous(
        minor_breaks = NULL,
        breaks = seq(0, splits, 1)
      ) +
      labs(
        title = paste("Cumulative Response for", gains$value[1]),
        linetype = NULL,
        y = "Cumulative response [%]",
        x = paste0("Percentiles [", splits, "]")
      ) +
      theme(legend.position = c(0.88, 0.2)) +
      guides(colour = "none")

    if (highlight == "auto") {
      highlight <- gains %>%
        arrange(desc(.data$lift)) %>%
        slice(1) %>%
        .$percentile
    }
    if (highlight %in% gains$percentile && highlight != "none") {
      highlight <- as.integer(highlight)
      note <- paste0(
        "If we select the top ",
        round(highlight * 100 / splits), "% cases with highest probabilities,\n",
        round(gains$cum_response[gains$percentile == highlight]), "% belong to the target class ",
        "(", round(gains$cum_response_lift[gains$percentile == highlight]), "% better than random)"
      )
      p <- p + labs(subtitle = note)
    } else {
      message("That highlight value is not a percentile. Try any integer from 1 to ", splits)
    }
  } else {
    df <- data.frame(tag = tag, score = score, multis)
    for (i in 1:(ncol(df) - 2)) {
      if (i == 1) out <- NULL
      g <- gain_lift(df$tag, df[, 2 + i], target, splits, quiet = quiet) %>%
        mutate(label = colnames(df)[2 + i])
      out <- rbind(out, g)
    }
    p <- out %>%
      mutate(percentile = factor(.data$percentile, levels = unique(out$percentile))) %>%
      group_by(.data$label) %>%
      mutate(
        rand = 100 * sum(.data$target) / sum(.data$total),
        cum_response = 100 * cumsum(.data$target) / cumsum(.data$total)
      ) %>%
      ggplot(aes(group = .data$label, x = .data$percentile)) +
      geom_hline(aes(yintercept = .data$rand), colour = "black", linetype = "dashed") +
      geom_line(aes(y = .data$cum_response, colour = .data$label), size = 1.2) +
      geom_label(aes(y = .data$cum_response, label = round(.data$cum_response)), alpha = 0.9) +
      geom_text(aes(label = paste0("Random: ", round(.data$rand, 1), "%"), y = .data$rand),
        x = 1, vjust = 1.2, hjust = 0, alpha = 0.2
      ) +
      labs(
        title = "Cumulative Response Plot", linetype = NULL,
        y = "Cumulative response [%]",
        x = paste0("Percentiles [", splits, "]")
      ) +
      theme(legend.position = c(0.88, 0.2)) +
      theme_lares(pal = 2) +
      labs(
        title = "Cumulative Response for Multiple Labels",
        subtitle = paste(
          "If we select the top nth percentile with highest probabilities,",
          "\nhow many observations belong to that specific target class?"
        ),
        x = paste0("Percentiles [", max(as.numeric(out$percentile)), "]"),
        y = "Cumulative Response [%]",
        linetype = "Reference", colour = "Label"
      ) +
      facet_grid(.data$label ~ .) +
      guides(colour = "none") +
      scale_y_continuous(limits = c(0, 100), breaks = seq(0, 100, 20))
  }

  aux <- freqs(data.frame(tag = tag), tag)
  text <- paste(
    "Assuming rate of",
    vector2text(round(aux$p), sep = "/", quotes = FALSE), "for",
    vector2text(aux$tag, sep = "/", quotes = FALSE)
  )
  if (!is.na(caption)) caption <- paste(text, caption, sep = "\n") else caption <- text
  p <- p + labs(caption = caption)
  if (save) export_plot(p, file_name, subdir = subdir, width = 6, height = 6)
  p
}


####################################################################
#' Top Hit Ratios for Multi-Classification Models
#'
#' Calculate and plot a multi-class model's predictions accuracy
#' based on top N predictions and distribution of probabilities.
#'
#' @family ML Visualization
#' @inheritParams mplot_roc
#' @return Plot with performance results over most frequent categories.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dfr) # Results for AutoML Predictions
#' mplot_topcats(dfr$class3$tag, dfr$class3$score,
#'   multis = subset(dfr$class3, select = -c(tag, score)),
#'   model_name = "Titanic Class Model"
#' )
#' @export
mplot_topcats <- function(tag, score, multis, model_name = NA) {
  df <- data.frame(tag, score, multis)
  cats <- unique(tag)

  DF <- df %>%
    mutate(label = row_number()) %>%
    tidyr::gather("name", "value", -.data$tag, -.data$score, -.data$label) %>%
    group_by(.data$label) %>%
    arrange(.data$label, desc(.data$value)) %>%
    mutate(
      rank = row_number(),
      cumprob = cumsum(.data$value)
    )

  p1 <- DF %>%
    mutate(
      correct_Top1 = .data$tag %in% .data$name[1],
      correct_Top2 = .data$tag %in% .data$name[1:2],
      correct_Top3 = .data$tag %in% .data$name[1:3],
      correct_Top4 = .data$tag %in% .data$name[1:4],
      correct_Top5 = .data$tag %in% .data$name[1:5]
    ) %>%
    ungroup() %>%
    select(.data$label, .data$tag, .data$score, starts_with("correct_top")) %>%
    distinct() %>%
    select(starts_with("correct_top")) %>%
    tidyr::gather() %>%
    group_by(.data$key) %>%
    summarize(total = n(), value = sum(.data$value), .groups = "drop") %>%
    mutate(correct = .data$value / max(.data$total)) %>%
    mutate(key = gsub("correct_", "", .data$key)) %>%
    head(length(cats)) %>%
    ggplot(aes(
      x = .data$key, y = .data$correct,
      label = formatNum(100 * .data$correct, 0, pos = "%")
    )) +
    geom_col() +
    geom_label() +
    scale_y_percent(limits = c(0, 1)) +
    labs(
      title = "Hit Ratios (Accuracy)",
      x = "Top N Predicted Categories",
      y = "Hit Ratio [%]"
    ) +
    theme_lares()
  if (!is.na(model_name)) {
    p1 <- p1 + labs(caption = model_name)
  }

  p2 <- DF %>%
    filter(.data$rank <= 5) %>%
    ggplot(aes(x = as.character(.data$rank), y = .data$value)) +
    # geom_jitter(alpha = 0.1) +
    geom_boxplot() + # outlier.shape = NA, alpha = 0.8
    labs(
      caption = paste(
        "Observations:", formatNum(nrow(df), 0), "|",
        "Unique labels:", formatNum(length(unique(DF$score)), 0)
      ),
      title = "How certain is the model?",
      x = "Top N Predicted Category",
      y = "Prediction (Certainty) [%]"
    ) +
    scale_y_percent(limits = c(0, 1)) +
    theme_lares()
  p <- p1 + p2
  p
}

####################################################################
#' Outliers: Winsorize
#'
#' Winsorizing a vector means that a predefined quantum of the smallest
#' and/or the largest values are replaced by less extreme values.
#' Thereby the substitute values are the most extreme retained values.
#'
#' @family Outliers
#' @param x Numeric vector. Distribution to be winsorized.
#' @param thresh Numeric vector. Lower and upper quantiles thresholds.
#' Set values within [0,1].
#' @param na.rm Boolean. Should \code{NA} be omitted to calculate the quantiles?
#' Note that \code{NA} in \code{x} are preserved and left unchanged anyway.
#' @return Numeric vector transformed.
#' @export
winsorize <- function(x, thresh = c(0.05, 0.95), na.rm = FALSE) {
  if (length(thresh) != 2 || any(abs(thresh) > 1)) {
    stop("thresh: pass a valid numeric vector of length 2 and values within [0, 1]")
  }
  cut_point_bottom <- quantile(x, thresh[1], na.rm = na.rm)
  cut_point_top <- quantile(x, thresh[2], na.rm = na.rm)
  j <- which(x <= cut_point_bottom)
  x[j] <- cut_point_bottom
  i <- which(x >= cut_point_top)
  x[i] <- cut_point_top
  attr(x, "thresh") <- thresh
  x
}

####################################################################
#' Outliers: Z-score method
#'
#' Z-score, also called a standard score, of an observation is a
#' distance from the population center measured in number of normalization
#' units. The default choice for center is sample mean and for
#' normalization unit is standard deviation. Values are considered
#' outliers based on z-score if its absolute value of
#' default z-score is higher then the threshold (popular choice is 3).
#'
#' @family Outliers
#' @param x Numeric. Distribution
#' @param thresh Numeric. Z-Score threshold for n standard deviations.
#' @param mad Boolean. Use median absolute deviation instead?
#' @return data.frame. Each row is an \code{x} observation with its
#' respective std/mean or mad/med calculations depending on \code{mad} input.
#' @export
outlier_zscore <- function(x, thresh = 3, mad = FALSE) {
  if (mad == FALSE) {
    calc <- mean(x, na.rm = TRUE)
    std <- sd(x, na.rm = TRUE)
    z <- thresh * std
    ret <- !abs(x - calc) <= z
    attr(ret, "values") <- data.frame(z = z, std = std, mean = calc)
  } else {
    calc <- median(x, na.rm = TRUE)
    mad <- mad(x, na.rm = TRUE)
    z <- thresh * mad
    ret <- !abs(x - calc) <= z
    attr(ret, "values") <- data.frame(z = z, mad = mad, med = calc)
  }
  ret
}

####################################################################
#' Outliers: Z-score method plot
#'
#' Test several Z-score thresholds to visualize outliers. Tidyverse
#' friendly.
#'
#' @family Outliers
#' @param df Dataframe.
#' @param var Numeric variable.
#' @param group Categorical variable. Grouping variable.
#' @param thresh Numeric vector. Z-Score threshold for n standard deviations.
#' @param top Integer. Show only n most frequent categorical values when
#' using the \code{group} argument.
#' @return ggplot2 object
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data(dft) # Titanic dataset
#' outlier_zscore_plot(dft, Fare)
#' p <- outlier_zscore_plot(dft, Fare, Pclass, thresh = c(3, 5))
#' plot(p)
#' attr(p, "z_values")
#' head(attr(p, "z_labels"))
#' @export
outlier_zscore_plot <- function(df, var, group = NULL,
                                thresh = c(2, 3, 5),
                                top = 5) {
  var <- enquo(var)
  group <- enquo(group)
  grouped <- !rlang::quo_is_null(group)

  df <- filter(df, !is.na(!!var))
  original_n <- filtered_n <- nrow(df)

  if (grouped) {
    if (!is.na(top)) {
      f <- freqs(df, !!group) %>% head(top)
      use <- unlist(f[, as_label(group)])
      df <- filter(df, !!group %in% use)
      filtered_n <- nrow(df)
    }
    df <- group_by(df, !!group)
  }

  df <- df %>%
    mutate(
      outlier_std = sd(!!var),
      outlier_mean = mean(!!var),
      outlier_group = ifelse(grouped, !!group, "Across all observations")
    )

  zs <- ref <- centre <- NULL
  for (i in thresh) {
    aux <- df %>%
      group_by(.data$outlier_group) %>%
      mutate(outlier = outlier_zscore(!!var, i))
    df[, paste0("Z-", i)] <- aux$outlier
    zs <- rbind(zs, c(i, unique(aux$outlier_std) * i))
  }

  df <- ungroup(df)
  temp <- select(df, starts_with("outlier_"), !!var, one_of(paste0("Z-", thresh))) %>%
    replaceall(c(TRUE, FALSE), c("purple", "#F79747"))

  subtitle <- sprintf("Variable: %s", as_label(var))
  if (grouped) subtitle <- paste(subtitle, "| Grouped by:", as_label(group))

  p <- ggplot(temp) +
    facet_grid(outlier_group ~ ., scales = "free") +
    lapply(thresh, function(t) {
      lapply(select(temp, one_of(paste0("Z-", t))), function(x) {
        geom_jitter(aes(
          x = paste0(
            "Z-", str_pad(t, nchar(max(thresh)), pad = "0"),
            sprintf("\n(%s)", sum(x == "purple"))
          ),
          colour = x, y = !!var
        ), width = .2, alpha = 0.7)
      })
    }) +
    scale_color_identity() +
    geom_hline(aes(yintercept = .data$outlier_mean), linetype = "dashed", alpha = 0.7) +
    labs(
      title = "Outliers thresholds test (Z-Scores)",
      subtitle = subtitle, x = NULL
    ) +
    theme_lares()
  if (original_n != filtered_n) {
    p <- p + labs(caption = sprintf(
      "Showing only the %s most frequent grouping categories", top
    ))
  }

  attr(p, "z_labels") <- select(df, !!var, one_of(paste0("Z-", thresh)), starts_with("outlier_"))
  attr(p, "z_values") <- zs
  p
}

####################################################################
#' Outliers: Tukey’s fences
#'
#' Tukey’s fences is a technique used in box plots. The non-outlier
#' range is defined with \code{[Q1-k(Q3-Q1), Q3+k(Q3-Q1)]}, where Q1 and Q3
#' are the lower and upper quartiles respectively, k - some non-negative
#' constant (popular choice is 1.5). A value is an outlier based
#' on Tukey’s fences when its value does not lie in non-outlier range.
#'
#' @family Outliers
#' @param x Numeric. Distribution
#' @param k Positive Numeric. K-multiplier.
#' @return Boolean vector detecting outliers.
#' @export
outlier_tukey <- function(x, k = 1.5) {
  quar <- quantile(x, probs = c(0.25, 0.75), na.rm = TRUE)
  iqr <- diff(quar)
  !((quar[1] - k * iqr <= x) & (x <= quar[2] + k * iqr))
}
#' @rdname outlier_tukey
#' @export
outlier_turkey <- outlier_tukey

#' ####################################################################
#' #' Download typeform data
#' #'
#' #' This function lets the user download surveys results from a
#' #' Typeform account.
#' #'
#' #' @family Typeform
#' #' @family Credentials
#' #' @param account Character. Which of the accounts in config.yml
#' #' do you wish to use?
#' #' @param form Character. Define which form you wish to import. Should
#' #' be the same as the form's name at the online platform
#' #' @param creds Character. Credential's user (see \code{get_creds()})
#' #' @export
#' typeform_download <- function(account = "tabunga", form, creds = NA) {
#'
#'   try_require("rtypeform")
#'
#'   accounts <- c("tabunga")
#'
#'
#'   if (account %in% accounts) {
#'     cr <- get_credentials(from = "typeform", dir = creds)
#'     if (account == "tabunga") {
#'       api <- cr$api_tabunga
#'     }
#'
#'     typeforms <- get_typeforms(api)
#'
#'     if (form %in% typeforms$content$name) {
#'       uid = typeforms$content$uid[typeforms$content$name==form]
#'       x = get_questionnaire(uid, api)
#'     } else {
#'       message(paste("Not a valid 'form' name. Try any of the following:\n",
#'                     paste(shQuote(typeforms$content$name), collapse = "\n ")))
#'     }
#'   }
#'
#'   #Some useful transformations
#'   colnames <- data.frame(id = colnames(x$completed)) %>%
#'     left_join(x$questions, by = "id") %>%
#'     mutate(question = ifelse(!is.na(question), question, id))
#'   colnames(x$completed) <- colnames$question
#'
#'   return(x)
#'
#' }
#'
#' # x <- typeform_download(form="Tabunga Créditos")

####################################################################
#' Calculate and Visualize Missingness
#'
#' This function lets the user calculate the percentage of NAs or
#' missingness in a data.frame. It also plots the results if needed.
#'
#' @family Exploratory
#' @family Missing Values
#' @param df Dataframe. Dataframe to study
#' @param plot Boolean. Do you wish to plot results?
#' @param full Boolean. Return all variables (or only with missings)?
#' @param subtitle Character. Subtitle to show in plot
#' @param summary Boolean. Show numerical summary text?
#' @return data.frame with each variable, number of missing values and percentage.
#' If \code{plot=TRUE}, a plot with the same information reflected.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#'
#' # Dummy data
#' df <- data.frame(
#'   A = c(1:5),
#'   B = c(NA, NA, 1, 1, 1),
#'   C = rep(NA, 5),
#'   D = c(NA, LETTERS[1:4])
#' )
#'
#' # Missing values summary
#' missingness(df)
#'
#' # Visual results
#'
#' missingness(df, plot = TRUE)
#'
#' # Show all variables (including those with no missing values)
#' missingness(df, plot = TRUE, full = TRUE)
#' @export
missingness <- function(df, plot = FALSE, full = FALSE,
                        subtitle = NA, summary = TRUE) {
  if (sum(is.na(df)) == 0) {
    if (summary) message("No missing values found!")
    invisible(NULL)
  } else {
    m <- df %>%
      ungroup() %>%
      summarize_all(.funs = ~ sum(is.na(.))) %>%
      gather() %>%
      arrange(desc(.data$value)) %>%
      filter(.data$value > 0) %>%
      mutate(missing = round(100 * .data$value / nrow(df), 2))
    colnames(m) <- c("variable", "missing", "missingness")

    if (plot) {
      obs <- nrow(df) * ncol(df)
      miss <- sum(m$missing)
      missp <- 100 * miss / obs
      note <- sprintf(
        "%s cols x %s rows | Values: %s | Missing: %s (%s)",
        formatNum(ncol(df), 0),
        formatNum(nrow(df), 0),
        formatNum(obs, 0),
        formatNum(miss, 0),
        formatNum(missp, 1, pos = "%")
      )


      p <- is.na(df) %>%
        data.frame() %>%
        tidyr::gather() %>%
        {
          if (!full) {
            filter(., .data$key %in% m$variable)
          } else {
            .
          }
        } %>%
        mutate(type = ifelse(.data$key %in% m$variable, "with", "without")) %>%
        group_by(.data$key) %>%
        mutate(row_num = row_number()) %>%
        mutate(perc = round(100 * sum(.data$value) / nrow(df), 2)) %>%
        mutate(label = ifelse(.data$type == "with", paste0(.data$key, " | ", .data$perc, "%"), .data$key)) %>%
        arrange(.data$value) %>%
        ggplot(aes(x = reorder(.data$label, .data$perc), y = .data$row_num, fill = .data$value)) +
        geom_raster() +
        coord_flip() +
        {
          if (full) {
            facet_grid(.data$type ~ ., space = "free", scales = "free")
          }
        } +
        {
          if (summary) {
            scale_y_comma(note, expand = c(0, 0))
          } else {
            scale_y_comma(NULL, expand = c(0, 0))
          }
        } +
        scale_fill_grey(name = NULL, labels = c("Present", "Missing"), expand = c(0, 0)) +
        labs(title = "Missing values", x = "", subtitle = if (!is.na(subtitle)) subtitle) +
        theme_lares(legend = "top") +
        theme(axis.text.y = element_text(size = 8))
      p
    } else {
      m
    }
  }
}


####################################################################
#' Impute Missing Values (using MICE)
#'
#' This function uses the MICE methodology to impute missing values.
#'
#' @family Data Wrangling
#' @family Machine Learning
#' @family Missing Values
#' @inheritParams mp3_get
#' @param df Dataframe. Dataframe to transform.
#' @param m Integer. Number of multiple imputations.
#' @param iters Integer. Number of iterations.
#' @param seed Integer. Set a seed for reproducibility.
#' @return data.frame with imputed values.
#' @export
impute <- function(df, m = 5, iters = 5, seed = 0, quiet = FALSE) {
  try_require("mice")
  on.exit(set.seed(seed))
  aux <- mice(df, seed = seed, m = m, maxit = iters, printFlag = !quiet)
  complete(aux)
}

####################################################################
#' Prevent Computer from Sleeping by Simulating Mouse Activity
#'
#' Keeps the computer awake by moving the mouse cursor and clicking
#' periodically with a human-like behavior. It automatically detects
#' the operating system (Windows or macOS) and runs the appropriate
#' underlying logic.
#'
#' The cursor initially moves to the upper-left corner (0,0), then
#' performs small random movements and clicks periodically to avoid
#' system sleep. The function exits if the user moves the mouse away
#' from the corner or when the specified off-time is reached.
#'
#' Requires \code{cliclick} installed on macOS (install via Homebrew:
#' \code{brew install cliclick}), and the \code{KeyboardSimulator}
#' package on Windows.
#'
#' @family Tools
#' @param sec_range Numeric vector of length 2. Range (in seconds) of
#' random intervals between mouse movements and clicks. Default is \code{c(20, 60)}.
#' @param off_time Numeric. Decimal hour (24h format) to stop the function
#' automatically, e.g. 18.5 means 18:30 (6:30 PM). Default is \code{18.5}.
#' @param click Logical. If \code{TRUE}, clicks the mouse. Default is \code{TRUE}.
#' @param quiet Logical. If \code{TRUE}, suppresses progress messages.
#' Default is \code{FALSE}.
#' @return Invisibly returns \code{NULL} when the function exits.
#' @examples
#' \dontrun{
#' # Keep the computer awake until 8:00 PM, with mouse actions every 30 to 60 seconds
#' dont_sleep(sec_range = c(30, 60), off_time = 20)
#' # Summary of amount of daily time running function
#' dont_sleep_time()
#' }
#' @export
dont_sleep <- function(sec_range = c(20, 60), off_time = 18.5, click = TRUE, quiet = FALSE) {
  os <- Sys.info()[["sysname"]]
  if (os == "Darwin") {
    stay_awake_mac(sec_range, off_time, quiet, click)
  } else if (os == "Windows") {
    stay_awake_windows(sec_range, off_time, quiet, click)
  } else {
    stop("Unsupported OS: ", os)
  }
}

#' @rdname dont_sleep
#' @export
dont_sleep_time <- function(quiet = FALSE) {
  msg <- NULL
  if (length(ls(.dont_sleep_tracker)) == 0) {
    if (!quiet) message("No time tracked yet in this session.")
  } else {
    if (!quiet) cat("Time spent running dont_sleep() per day in current session:\n")
    for (day in ls(.dont_sleep_tracker)) {
      seconds <- get(day, envir = .dont_sleep_tracker)
      hours <- floor(seconds / 3600)
      minutes <- floor((seconds %% 3600) / 60)
      secs <- round(seconds %% 60)
      msgi <- sprintf("  %s | %02d:%02d:%02d\n", day, hours, minutes, secs)
      if (!quiet) cat(msgi)
      msg <- c(msg, trimws(msgi))
    }
    msg <- data.frame(msg = msg) %>%
      mutate(
        date = as.Date(gsub(" \\| .*", "", .data$msg)),
        sleep_time = lubridate::hms(gsub(".* \\| ", "", .data$msg), roll = TRUE)
      ) %>%
      select(-1)
  }
  invisible(msg)
}

.dont_sleep_tracker <- new.env(parent = emptyenv())

.track_dont_sleep <- function(start_time) {
  end_time <- Sys.time()
  start_date <- as.Date(start_time)
  end_date <- as.Date(end_time)
  if (start_date == end_date) {
    # Same day, assign normally
    elapsed <- as.numeric(difftime(end_time, start_time, units = "secs"))
    if (!exists(as.character(start_date), envir = .dont_sleep_tracker)) {
      assign(as.character(start_date), elapsed, envir = .dont_sleep_tracker)
    } else {
      previous <- get(as.character(start_date), envir = .dont_sleep_tracker)
      assign(as.character(start_date), previous + elapsed, envir = .dont_sleep_tracker)
    }
  } else {
    # Crosses midnight, split time by day
    midnight <- as.POSIXct(paste0(start_date + 1, " 00:00:00"), tz = attr(start_time, "tzone"))
    first_day_secs <- as.numeric(difftime(midnight, start_time, units = "secs"))
    second_day_secs <- as.numeric(difftime(end_time, midnight, units = "secs"))
    # Assign first day
    if (!exists(as.character(start_date), envir = .dont_sleep_tracker)) {
      assign(as.character(start_date), first_day_secs, envir = .dont_sleep_tracker)
    } else {
      previous <- get(as.character(start_date), envir = .dont_sleep_tracker)
      assign(as.character(start_date), previous + first_day_secs, envir = .dont_sleep_tracker)
    }
    # Assign second day
    if (!exists(as.character(end_date), envir = .dont_sleep_tracker)) {
      assign(as.character(end_date), second_day_secs, envir = .dont_sleep_tracker)
    } else {
      previous <- get(as.character(end_date), envir = .dont_sleep_tracker)
      assign(as.character(end_date), previous + second_day_secs, envir = .dont_sleep_tracker)
    }
  }
}

.stay_awake_core <- function(
  get_mouse_position,
  move_mouse,
  click_mouse,
  move_to_origin,
  sec_range = c(20, 60),
  off_time = 18.5,
  quiet = FALSE,
  click = TRUE,
  label = "stay_awake"
) {
  move_to_origin()
  message(sprintf("Mouse moved to (0,0). Starting loop... %s clicks.", ifelse(click, "Enabled", "Disabled")))

  start_time <- Sys.time()
  last_time <- start_time
  tic(paste(label, start_time))
  timeout <- moved <- FALSE

  # Ensure time is tracked even if user stops manually
  on.exit(
    {
      .track_dont_sleep(last_time)
      if (!quiet) {
        lap <- toc(paste(label, start_time), quiet = TRUE)
        now_fmt <- format(start_time, format = "%Y-%m-%d %H:%M:%S")
        cat(sprintf("\rStarted %s ago (%s)    \n", lap$time, now_fmt))
        if (moved) message("Mouse moved by user. Exiting at ", now)
        if (timeout) message("Off-time triggered at ", format(now_posix, "%Y-%m-%d %H:%M:%S"))
      }
    },
    add = TRUE
  )

  repeat {
    now <- Sys.time()
    pos <- get_mouse_position()
    if (length(pos) != 2 || any(pos != c(0, 0))) {
      moved <- TRUE
      break
    }

    lap <- toc(paste(label, start_time), quiet = TRUE)
    now_fmt <- format(start_time, format = "%Y-%m-%d %H:%M:%S")
    if (!quiet) cat(sprintf("\rStarted %s ago (%s)    ", lap$time, now_fmt))
    .track_dont_sleep(last_time)
    last_time <- now

    dx <- sample(-20:20, 1)
    dy <- sample(-20:20, 1)
    move_mouse(dx, dy)
    move_to_origin()
    if (click) click_mouse()

    now_posix <- as.POSIXlt(now)
    if ((now_posix$hour + now_posix$min / 60 + now_posix$sec / 3600) >= off_time) {
      timeout <- TRUE
      break
    }

    if (length(sec_range) == 1) sec_range <- rep(sec_range, 2)
    Sys.sleep(runif(1, sec_range[1], sec_range[2]))
  }

  invisible(NULL)
}


# macOS wrapper
stay_awake_mac <- function(sec_range = c(20, 60), off_time = 18.5, click = TRUE, quiet = FALSE) {
  if (Sys.info()[["sysname"]] != "Darwin") {
    stop("This function is for macOS only.")
  }
  if (system("which cliclick", intern = TRUE) == "") {
    stop("Install cliclick first: brew install cliclick")
  }

  get_mouse_position <- function() {
    out <- system2("cliclick", args = "p:", stdout = TRUE)
    as.numeric(unlist(regmatches(out, gregexpr("[0-9]+", out))))
  }
  move_mouse <- function(dx, dy) {
    dx_str <- if (dx >= 0) paste0("+", dx) else as.character(dx)
    dy_str <- if (dy >= 0) paste0("+", dy) else as.character(dy)
    system2("cliclick", args = sprintf("m:%s,%s", dx_str, dy_str))
  }
  move_to_origin <- function() system2("cliclick", args = "m:0,0")
  click_mouse <- function() system2("cliclick", args = "c:0,0")

  .stay_awake_core(get_mouse_position, move_mouse, click_mouse, move_to_origin,
    sec_range, off_time, quiet, click,
    label = "stay_awake_mac"
  )
}

# Windows wrapper
stay_awake_windows <- function(sec_range = c(20, 60), off_time = 18.5, quiet = FALSE, click = TRUE) {
  if (Sys.info()[["sysname"]] != "Windows") stop("This function is for Windows only.")
  try_require("KeyboardSimulator")

  get_mouse_position <- function() mouse.get_cursor()
  move_mouse <- function(dx, dy) mouse.move(dx, dy, duration = 0)
  move_to_origin <- function() mouse.move(0, 0, duration = 0)
  click_mouse <- function() mouse.click(button = "left")

  .stay_awake_core(get_mouse_position, move_mouse, click_mouse, move_to_origin,
    sec_range, off_time, quiet, click,
    label = "stay_awake_windows"
  )
}

####################################################################
#' Automated K-Means Clustering + PCA/t-SNE
#'
#' This function lets the user cluster a whole data.frame automatically.
#' As you might know, the goal of kmeans is to group data points into
#' distinct non-overlapping subgroups. If needed, one hot encoding will
#' be applied to categorical values automatically with this function.
#' For consideration: Scale/standardize the data when applying kmeans.
#' Also, kmeans assumes spherical shapes of clusters and does not work well
#' when clusters are in different shapes such as elliptical clusters.
#'
#' @family Clusters
#' @inheritParams stats::kmeans
#' @inheritParams mp3_get
#' @param df Dataframe
#' @param k Integer. Number of clusters
#' @param wss_var Numeric. Used to pick automatic \code{k} value,
#' when \code{k} is \code{NULL} based on WSS variance while considering
#' \code{limit} clusters. Values between (0, 1). Default value could be
#' 0.05 to consider convergence.
#' @param limit Integer. How many clusters should be considered?
#' @param drop_na Boolean. Should NA rows be removed?
#' @param ignore Character vector. Names of columns to ignore.
#' @param ohse Boolean. Do you wish to automatically run one hot
#' encoding to non-numerical columns?
#' @param norm Boolean. Should the data be normalized?
#' @param dim_red Character. Select dimensionality reduction technique.
#' Pass any of: \code{c("PCA", "tSNE", "all", "none")}.
#' @param comb Vector. Which columns do you wish to plot? Select which
#' two variables by name or column position.
#' @param seed Numeric. Seed for reproducibility
#' @param ... Additional parameters to pass sub-functions.
#' @return List. If no \code{k} is provided, contains \code{nclusters} and
#' \code{nclusters_plot} to determine optimal \code{k} given their WSS (Within
#' Groups Sum of Squares). If \code{k} is provided, additionally we get:
#' \itemize{
#'   \item \code{df} data.frame with original \code{df} plus \code{cluster} column
#'   \item \code{clusters} integer which is the same as \code{k}
#'   \item \code{fit} kmeans object used to fit clusters
#'   \item \code{means} data.frame with means and counts for each cluster
#'   \item \code{correlations} plot with correlations grouped by clusters
#'   \item \code{PCA} list with PCA results (when \code{dim_red="PCA"})
#'   \item \code{tSNE} list with t-SNE results (when \code{dim_red="tSNE"})
#' }
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data("iris")
#' df <- subset(iris, select = c(-Species))
#'
#' # If dataset has +5 columns, feel free to reduce dimenstionalities
#' # with reduce_pca() or reduce_tsne() first
#'
#' # Find optimal k
#' check_k <- clusterKmeans(df, limit = 10)
#' check_k$nclusters_plot
#' # Or pick k automatically based on WSS variance
#' check_k <- clusterKmeans(df, wss_var = 0.05, limit = 10)
#' # You can also use our other functions:
#' # clusterOptimalK(df) and clusterVisualK(df)
#'
#' # Run with selected k
#' clusters <- clusterKmeans(df, k = 3)
#' names(clusters)
#'
#' # Cross-Correlations for each cluster
#' plot(clusters$correlations)
#'
#' # PCA Results (when dim_red = "PCA")
#' plot(clusters$PCA$plot_explained)
#' plot(clusters$PCA$plot)
#' @export
clusterKmeans <- function(df, k = NULL, wss_var = 0, limit = 15, drop_na = TRUE,
                          ignore = NULL, ohse = TRUE, norm = TRUE,
                          algorithm = c("Hartigan-Wong", "Lloyd", "Forgy", "MacQueen"),
                          dim_red = "PCA", comb = c(1, 2), seed = 123,
                          quiet = FALSE, ...) {
  on.exit(set.seed(seed))
  check_opts(dim_red, c("PCA", "tSNE", "all", "none"))
  if ("all" %in% dim_red) dim_red <- c("PCA", "tSNE")

  if (isTRUE(is.na(ignore)[1])) ignore <- NULL
  ignore <- unique(ignore)
  df <- .prepare_cluster(df,
    drop_na = drop_na, ohse = ohse,
    norm = norm, quiet = quiet,
    ignore = ignore, ...
  )

  # Ignore some columns
  if (!is.null(ignore)) {
    order <- colnames(df)
    aux <- select(df, any_of(ignore))
    df <- select(df, -any_of(ignore))
    if (!quiet) message(paste("Ignored features:", v2t(ignore)))
  }

  results <- list()

  # Determine number of clusters (n) using WSS methodology
  wss <- sum(apply(df, 2, var)) * (nrow(df) - 1)
  limit <- min(nrow(df) - 1, limit)
  for (i in 2:limit) {
    wss[i] <- suppressWarnings(
      sum(kmeans(df, centers = i, algorithm = algorithm)$withinss)
    )
  }
  nclusters <- data.frame(n = 1:limit, wss = wss)
  nclusters_plot <- ggplot(nclusters, aes(x = .data$n, y = .data$wss)) +
    geom_line() +
    geom_point() +
    labs(
      title = "Total Number of Clusters",
      subtitle = "HINT: Where does the curve level?",
      x = "Number of Clusters",
      y = "Within Groups Sum of Squares"
    ) +
    scale_y_comma() +
    theme_lares(...)
  results[["nclusters"]] <- nclusters
  results[["nclusters_plot"]] <- nclusters_plot

  # Auto K selected by less X WSS variance (convergence)
  if (wss_var > 0 & is.null(k)) {
    k <- nclusters %>%
      mutate(
        pareto = .data$wss / .data$wss[1],
        dif = lag(.data$pareto) - .data$pareto
      ) %>%
      filter(.data$dif > wss_var) %>%
      pull(.data$n) %>%
      max(.)
    if (!quiet) {
      message(sprintf(
        ">> Auto selected k = %s (clusters) based on minimum WSS variance of %s%%",
        k, wss_var * 100
      ))
    }
  }

  # If n is already selected
  if (!is.null(k)) {
    if (!is.null(ignore)) {
      results[["df"]] <- bind_cols(df, aux) %>% select(any_of(order), everything())
    } else {
      results[["df"]] <- df
    }
    yintercept <- nclusters$wss[nclusters$n == k]
    nclusters_plot <- nclusters_plot +
      geom_hline(aes(yintercept = yintercept), colour = "red") +
      labs(subtitle = paste("Number of clusters selected:", k))
    results[["clusters"]] <- k
    results[["nclusters_plot"]] <- nclusters_plot

    # K-Means Cluster Analysis
    fit <- kmeans(df, k, algorithm = algorithm, iter.max = limit)
    results[["fit"]] <- fit
    # Append cluster assignment
    df <- data.frame(results[["df"]], cluster = as.factor(fit$cluster))
    results[["df"]] <- df
    # Get cluster means
    results[["means"]] <- df %>%
      group_by(.data$cluster) %>%
      summarise_if(is.numeric, list(mean)) %>%
      mutate(n = as.integer(table(df$cluster)))

    # Correlations
    results[["correlations"]] <- corr_cross(
      df,
      contains = "cluster_", quiet = TRUE, ignore = ignore
    ) +
      labs(subtitle = "Most relevant correlations grouped by cluster")

    # Dim reduction: PCA
    if ("PCA" %in% dim_red) {
      PCA <- reduce_pca(df, ignore = c(ignore, "cluster"), comb = comb, quiet = quiet, ...)
      PCA$plot <- PCA$plot +
        geom_point(aes(colour = df$cluster)) +
        labs(colour = "Cluster", title = "Clusters with Principal Component Analysis")
      results[["PCA"]] <- PCA
    }

    # Dim reduction: t-SNE
    if ("tSNE" %in% dim_red) {
      tsne <- reduce_tsne(df, ignore = c(ignore, "cluster"), quiet = quiet, ...)
      tsne$plot <- tsne$plot +
        geom_point(aes(colour = df$cluster)) +
        labs(colour = "Cluster", title = "Clusters with t-SNE")
      results[["tSNE"]] <- tsne
    }
  }
  results
}


####################################################################
#' Visualize K-Means Clusters for Several K
#'
#' Visualize cluster data for assorted values of k.
#'
#' @family Clusters
#' @inheritParams clusterKmeans
#' @param ks Integer vector. Which k should be tested?
#' @param ... Additional parameters passed to \code{clusterKmeans}
#' @return List. Plot and data.frame results of clustering \code{df}
#' data.frame into \code{ks} integer clusters.
#' @examples
#' Sys.unsetenv("LARES_FONT") # Temporal
#' data("iris")
#' df <- subset(iris, select = c(-Species))
#' df <- df[sample(nrow(df)), ]
#'
#' # Calculate and plot
#' result <- clusterVisualK(df, ks = 2:4)
#' plot(result$plot)
#'
#' # You can use the data generated as well
#' lapply(result$data, function(x) head(x$cluster, 10))
#' @export
clusterVisualK <- function(df, ks = 2:6, ...) {
  clus_dat <- function(df, k, n = length(ks), ...) {
    clusters <- clusterKmeans(df, k, quiet = TRUE, ...)
    pca <- clusters$PCA$pcadf %>%
      mutate(cluster = clusters$df$cluster, k = k)
    explained <<- clusters$PCA$pca_explained[1:2]
    pca
  }
  clus_plot <- function(clus_dat, ...) {
    clus_dat %>%
      ggplot(aes(x = .data$PC1, y = .data$PC2, colour = as.character(.data$cluster))) +
      geom_point() +
      guides(colour = "none") +
      labs(subtitle = glued("{clus_dat$k[1]} clusters")) +
      theme_lares(pal = 2, ...)
  }
  dats <- lapply(ks, function(x) clus_dat(df, x, ...))
  plots <- lapply(dats, function(x) clus_plot(x, ...))

  total <- formatNum(sum(explained), 1, pos = "%")
  explained <- formatNum(explained, 1, pos = "%")
  subtitle <- sprintf(
    "Explaining %s of the variance with 2 PCA:\nPC1 (%s), PC2 (%s)",
    total, explained[1], explained[2]
  )

  wrapped <- wrap_plots(plots) +
    plot_annotation(
      title = "Kmeans Clustering across potential number of clusters",
      subtitle = subtitle
    )

  ret <- list(plot = wrapped, data = dats)
  invisible(ret)
}

####################################################################
#' Visualize K-Means Clusters for Several K Methods
#'
#' Visualize cluster data for assorted values of k and methods such as
#' WSS, Silhouette and Gap Statistic. See \code{factoextra::fviz_nbclust}
#' for more.
#'
#' @family Clusters
#' @inheritParams clusterKmeans
#' @param method Character vector.
#' @param ... Additional parameters passed to \code{factoextra::fviz_nbclust}
#' @return Plot. Optimal number of clusters of \code{df} data.frame given a
#' selected \code{method}.
#' @examples
#' # You must have "factoextra" library to use this auxiliary function:
#' \dontrun{
#' data("iris")
#' df <- subset(iris, select = c(-Species))
#' # Calculate and plot optimal k clusters
#' clusterOptimalK(df)
#' }
#' @export
clusterOptimalK <- function(df, method = c("wss", "silhouette", "gap_stat"),
                            drop_na = TRUE, ohse = TRUE, norm = TRUE,
                            quiet = TRUE, ...) {
  try_require("factoextra")
  df <- .prepare_cluster(df, drop_na = drop_na, ohse = ohse, norm = norm, quiet = quiet)
  lapply(method, function(x) {
    fviz_nbclust(df, kmeans, method = x, ...) + theme_lares(pal = 2, ...)
  })
}

.prepare_cluster <- function(df, drop_na = TRUE, ohse = TRUE,
                             norm = TRUE, quiet = FALSE, ignore = NULL, ...) {
  # Leave some columns out of the logic
  if (!is.null(ignore)) {
    ignored <- select(df, any_of(ignore))
    df <- select(df, -any_of(ignore))
  } else {
    ignored <- NULL
  }

  # There should be no NAs
  df <- removenacols(df, all = TRUE)
  if (sum(is.na(df)) > 0) {
    if (drop_na) {
      df <- removenarows(df, all = FALSE)
      if (!quiet) {
        message(
          "Automatically removed rows with NA. To overwrite: fix NAs and set drop_na = FALSE"
        )
      }
    } else {
      stop(paste("There should be no NAs in your dataframe!",
        "You can manually fix it or set drop_na to TRUE to remove these rows.",
        sep = "\n"
      ))
    }
    if (nrow(df) == 0) {
      stop("There are no observations without NA values. Please, check your dataset")
    }
  }

  # Only numerical values
  nums <- df_str(df, return = "names", quiet = TRUE)$nums
  if (isTRUE(ohse) && length(nums) != ncol(df)) {
    df <- ohse(df, dates = TRUE, limit = 8, ignore = ignore, ...)
  } else {
    df <- data.frame(df) %>% select_if(is.numeric)
  }

  # Data should be normalized
  if (norm) {
    df <- df %>%
      transmute_if(is.numeric, list(normalize)) %>%
      replace(., is.na(.), 0)
  }

  # No duplicates
  df <- bind_cols(ignored, df)
  temp <- which(!colnames(df) %in% ignore)
  new_df <- distinct_at(df, temp, .keep_all = TRUE)
  if (nrow(new_df) != nrow(df)) {
    if (!quiet) {
      message(paste(">>> Removed duplicate obserations:", nrow(df) - nrow(new_df)))
    }
  }
  new_df
}

####################################################################
#' Scrap data based on IP address
#'
#' This function lets the user scrap https://db-ip.com/ given
#' IP address(es) to get their associated address type, ASN, ISP,
#' organization, country, state or region, county, city, ZIP postal code,
#' weather station, coordinates, Timezone, local time, languages, and currency.
#'
#' @family Tools
#' @family Scrapper
#' @param ip Vector. Vector with all IP's we wish to search.
#' @param quiet Boolean. Do not show the loading \code{statusbar}?
#' @return data.frame. Each row is an unique \code{ip} address,
#' and columns will bee created for all the additional information found.
#' @examples
#' \donttest{
#' ip_data("163.114.132.0")
#' ip_data(ip = c(myip(), "201.244.197.199"), quiet = TRUE)
#' }
#' @export
ip_data <- function(ip = myip(), quiet = FALSE) {
  if (!haveInternet()) {
    message("No internet connetion...")
    invisible(NULL)
  } else {
    ip <- ip[!is.na(ip)]
    ip <- ip[is_ip(ip)]
    ip <- unique(ip)
    output <- data.frame()
    for (i in seq_along(ip)) {
      url <- paste0("https://db-ip.com/", ip[i])
      scrap <- content(GET(url)) %>% html_table()
      clean <- bind_rows(scrap[[1]], scrap[[3]])
      row <- data.frame(t(clean[, 2]))
      colnames(row) <- clean$X1
      row <- data.frame(id = ip[i], row)
      output <- bind_rows(output, row)
      if (length(ip) > 1 && !quiet) statusbar(i, length(ip), ip[i])
    }
    output <- cleanNames(output)
    row.names(output) <- NULL
    output
  }
}


####################################################################
#' Convert Python JSON string to R vector (data.frame with 1 row)
#'
#' This function lets the user transform a JSON string into vector
#' (data.frame with 1 row). You can also pass a Python's dictionary.
#' For any other JSON transformation, \code{jsonlite} is recommended.
#'
#' @family Tools
#' @param json Character. JSON string.
#' @return List, data.frame, or vector. Depends on the \code{json} string.
#' @examples
#' json2vector('{"id": 1, "nodata": null, "gender": "M"}')
#' @export
json2vector <- function(json) {
  string <- paste0("[", gsub('"', "\"", json), "]")
  string <- gsub("'", '"', string)
  string <- gsub("None", "null", string)
  string <- gsub("True", "true", string)
  string <- gsub("False", "false", string)
  vector <- fromJSON(string)
  data.frame(t(unlist(vector)))
}


####################################################################
#' Import Excel File with All Its Tabs
#'
#' This function lets the user import an Excel file's tabs into a list
#'
#' @family Tools
#' @param file String. Local Excel file name
#' @return List or data.frame. If single tab is found, a data.frame; if
#' multiple tabs are found on file, a list of data.frames.
#' @export
importxlsx <- function(file) {
  sheets <- getSheetNames(file)
  mylist <- list()
  for (i in seq_along(sheets)) {
    sheet <- read.xlsx(file,
      sheet = i,
      skipEmptyRows = TRUE,
      skipEmptyCols = TRUE,
      startRow = 1,
      detectDates = TRUE
    )
    mylist[[sheets[i]]] <- sheet
  }
  if (length(sheets) == 1) {
    mylist <- mylist[[1]][[1]]
  }
  mylist
}


####################################################################
#' Read Files Quickly (Auto-detected)
#'
#' This function lets the user import csv, xlsx, xls, sav files.
#'
#' @family Tools
#' @param filename Character. File name to import.
#' @param current_wd Boolean. Use current working directory before
#' the file's name? Use this param to NOT get absolute root directory.
#' @param sheet Character. Name or index of the sheet to read data
#' from if file is xlsx or xls.
#' @param quiet Boolean. Quiet summary message?
#' @return List or data.frame, depending on \code{filename}'s data.
#' @export
read.file <- function(filename, current_wd = TRUE, sheet = 1, quiet = FALSE) {
  if (current_wd) filename <- paste0(getwd(), "/", filename)
  if (!file.exists(filename)) {
    stop("That file doesn't exist.. try with another!")
  } else {
    filetype <- gsub("\\.", "", right(filename, 4))

    if (filetype == "csv") {
      results <- data.frame(read.csv(filename))
    }
    if (filetype == "xlsx") {
      results <- read.xlsx(filename, sheet)
    }
    if (filetype == "xls") {
      try_require("gdata")
      results <- read.xls(filename, read.xls)
    }
    # if (filetype == "sav") {
    #   try_require("foreign")
    #   results <- quiet(read.spss(filename, to.data.frame = T))
    # }
    # if (filetype == "dta") {
    #   # Stata version 5-12 .dta file
    #   # results <- foreign::read.dta(filename)
    #   # Stata version 13 .dta file
    #   try_require("readstata13")
    #   results <- read.dta13(filename)
    # }
    if (filetype == "dat") {
      results <- read.table(filename, header = TRUE)
    }

    if (!quiet) {
      message(paste(
        "Imported", filetype, "file with",
        formatNum(nrow(results), 0), "rows x",
        formatNum(ncol(results), 0), "columns, succesfully!"
      ))
    }
  }
  if (nrow(results) == 0) warning("There is no data in that file...")
  results
}


####################################################################
#' Bind Files into Dataframe
#'
#' This function imports and binds multiple files into a single
#' data.frame. Files must be inserted with absolute roots files names.
#'
#' @family Tools
#' @param files Character vector. Files names.
#' @return data.frame with data joined from all \code{files} passed.
#' @export
bind_files <- function(files) {
  alldat <- data.frame()
  for (i in seq_along(files)) {
    file <- files[i]
    dfi <- read.file(file, current_wd = FALSE)
    alldat <- bind_rows(alldat, dfi)
    statusbar(i, length(files))
  }
  as_tibble(alldat)
}


####################################################################
#' List categorical values for data.frame
#'
#' Make a list with all categorical values and
#'
#' @family Tools
#' @param df data.frame
#' @param ... Variables to segment counters
#' @param abc Boolean. Sort alphabetically?
#' @return List. Length same as number of categorical columns, each with a
#' frequency data.frame using \code{freqs()}.
#' @examples
#' data(dft) # Titanic dataset
#' df <- dft[, 1:5]
#' head(df)
#' list_cats(df)
#' @export
list_cats <- function(df, ..., abc = TRUE) {
  is.categorical <- function(x) is.character(x) | is.factor(x)
  category <- which(unlist(lapply(df, is.categorical)))
  ret <- list()
  for (i in seq_along(category)) {
    which <- as.character(names(category)[i])
    df[, which] <- as.character(df[, which])
    aux <- freqs(df, cats = base::get(which), ...) %>%
      select(-.data$pcum, -.data$order)
    if (abc) aux <- arrange(aux, .data$cats)
    colnames(aux)[1] <- which
    ret[[which]] <- aux
  }
  ret
}


####################################################################
#' Spread list column into new columns
#'
#' Spread an existing list column into new columns on a data.frame. Note
#' that every element on every observation must have a name for the function
#' to do its work. Original column will be automatically suppressed but you
#' can set the \code{replace} argument to avoid it.
#'
#' @param df Dataframe
#' @param col Variable name.
#' @param str Character. Start column names with. If set to \code{NULL},
#' original name of column will be used.
#' @param replace Boolean. Replace original values (delete column)
#' @return data.frame. Result of un-nesting named or un-named list columns.
#' @examples
#' df <- dplyr::starwars
#' # Un-named list columns
#' spread_list(df, films, replace = FALSE) %>%
#'   dplyr::select(name, dplyr::starts_with("films")) %>%
#'   head(8)
#' # Named (and un-nammed) list columns
#' df <- dplyr::tibble(id = 1:3, platform = list(
#'   list("fb" = 1, "ig" = 2),
#'   list("fb" = 3),
#'   list()
#' ))
#' spread_list(df, platform, str = "ptf_")
#' @export
spread_list <- function(df, col, str = NULL, replace = TRUE) {
  var <- enquo(col)
  col <- gsub('"', "", as_label(var))
  cols <- colnames(df)
  if (!"list" %in% unlist(lapply(df[, cols == col], class))) {
    warning("The variable provided is not a list variable")
    df
  } else {
    # If all values are NA, no need to proceed but to change into vector with NAs
    if (all(unlist(lapply(df[, cols == col], is.na)))) {
      df[, cols == col] <- rep(NA, nrow(df))
      df
    } else {
      # Automatic naming based on original
      if (is.null(str)) str <- sprintf("%s_", col)
      # Add character NAs name to those observations with no data, thus no names
      nonames <- rowwise(df) %>%
        mutate(len = length(names(!!var))) %>%
        pull(.data$len) == 0
      # Non-named list columns
      if (sum(nonames) == nrow(df)) {
        binded <- select(df, !!var) %>%
          mutate(temp_cross_id = row_number()) %>%
          tidyr::unnest_longer(!!var) %>%
          mutate(key = TRUE) %>%
          tidyr::spread(key = !!var, value = .data$key) %>%
          replace(is.na(.), FALSE)
      } else {
        # Named list columns
        binded <- lapply(df[!nonames, cols == col], bind_rows) %>%
          bind_rows() %>%
          mutate(temp_cross_id = which(!nonames))
        if (is.numeric(binded[[1]])) binded <- binded %>% replace(is.na(.), 0)
      }

      if (is.na(str)) str <- paste0(col, "_")
      ids <- which(colnames(binded) == "temp_cross_id")
      colnames(binded)[-ids] <- paste0(str, colnames(binded))[-ids]
      done <- df %>%
        mutate(temp_cross_id = row_number()) %>%
        left_join(binded, "temp_cross_id") %>%
        select(-.data$temp_cross_id)

      original <- which(cols == col)
      done <- done %>% select(1:original, starts_with(str), (original + 1):ncol(done))
      if (replace) done <- done[, -original]
      as_tibble(done)
    }
  }
}


####################################################################
#' Convert markdown string tables to data.frame
#'
#' @family Tools
#' @param text Character. Markdown text representing a table.
#' @param autoformat Boolean. Automatically format numerical,
#' logical and date values to their classes?
#' @examples
#' txt <- "| Item | Value |\n|------|-------|\n| 50C  | 122F  |\n| 300K | 80.33F |"
#' markdown2df(txt)
#' @export
markdown2df <- function(text, autoformat = TRUE) {
  df <- removenacols(read.table(
    text = text, sep = "|", header = TRUE, strip.white = TRUE, quote = "\""
  ))
  # Get rid of potential first row with all values set as --- or :---
  if (all(stringr::str_split(df[1, 1], "-")[[1]] == "")) df <- df[-1, ]
  if (substr(df[1, 1], 1, 4) == ":---") df <- df[-1, ]
  rownames(df) <- NULL
  df <- as_tibble(df)
  if (autoformat) {
    df <- df %>%
      chr2num() %>%
      chr2logical() %>%
      chr2date()
  }
  df
}

####################################################################
#' Check character values for date/numeric/logical and change datatype
#'
#' Automatically check a vector, data.frame or list for numeric, logical,
#' date content and change their datatype. Note that factors are skipped in
#' case the user requires character numeric values to be kept as they are.
#'
#' @family Tools
#' @param data Vector, data.frame or list
#' @examples
#' str(chr2num(c("1", "2", "3")))
#' df <- data.frame(A = c("1", "3"), B = c("A", "B"), c = c(pi, pi * 2))
#' str(chr2num(df))
#' lst <- list(A = c("1", "2", "3"), B = c("A", "B", "3"), C = pi, D = 3L)
#' str(chr2num(lst))
#' lst2 <- list(layer1 = ":D", layer2 = lst)
#' str(chr2num(lst2))
#' str(chr2logical(c(NA, "true", FALSE)))
#' @export
chr2num <- function(data) {
  pattern <- "^-?\\d+(\\.\\d+)?$"
  if (is.list(data)) {
    char_elements <- sapply(data, is.character)
    num_elements <- sapply(data, function(element) {
      all(grepl(pattern, element[!is.na(element)]))
    })
    elements_to_convert <- char_elements & num_elements
    data[elements_to_convert] <- lapply(data[elements_to_convert], as.numeric)
  } else {
    if (is.character(data) && all(grepl(pattern, data))) {
      data <- as.numeric(data)
    }
  }
  data
}
#' @rdname chr2num
#' @export
chr2logical <- function(data) {
  if (is.list(data)) {
    char_elements <- sapply(data, is.character)
    log_elements <- sapply(data, function(element) {
      all(toupper(element) %in% c("TRUE", "FALSE", NA))
    })
    elements_to_convert <- char_elements & log_elements
    data[elements_to_convert] <- lapply(toupper(data[elements_to_convert]), as.logical)
  } else {
    if (is.character(data) && all(toupper(data) %in% c("TRUE", "FALSE", NA))) {
      data <- as.logical(toupper(data))
    }
  }
  data
}
#' @rdname chr2num
#' @export
chr2date <- function(data) {
  pattern <- "^[0-9-]+$"
  if (is.list(data)) {
    char_elements <- sapply(data, is.character)
    log_elements <- sapply(data, function(element) {
      all(grepl(pattern, element) | is.na(element))
    })
    elements_to_convert <- char_elements & log_elements
    data[elements_to_convert] <- lapply(
      data[elements_to_convert], function(x) as.Date(x, origin = "1970-01-01")
    )
  } else {
    if (is.character(data) && all(grepl(pattern, data) | is.na(data))) {
      data <- as.Date(data, origin = "1970-01-01")
    }
  }
  data
}

####################################################################
#' Calculate Character Reduction from R Object to TOON Format
#'
#' Calculates the percentage reduction in the number of characters when
#' converting an R object to its TOON string representation compared to
#' its JSON representation (or the string itself if the input is a single string).
#'
#' @param obj Any R object (list, vector, data frame, or string).
#' @param ... Additional arguments passed to jsonlite::toJSON()
#' @return A numeric value representing the character reduction ratio (0 to 1).
#' @export
toon_reduction <- function(obj, ...) {
  toon_str <- as_toon(obj)
  toon_nchar <- nchar(toon_str)
  if (length(obj) == 1 && is.character(obj) && !is.na(obj)) {
    original_str <- obj
  } else {
    original_str <- jsonlite::toJSON(obj, ...)
  }
  original_nchar <- nchar(original_str)
  if (original_nchar == 0) {
    return(0)
  }
  reduction_amount <- original_nchar - toon_nchar
  reduction_ratio <- reduction_amount / original_nchar
  return(reduction_ratio)
}

####################################################################
#' Validate inputs (attributions, options, ...)
#'
#' This function validates if inputs match all/any of your options
#' and return error/message with possible options to use. Similar
#' to \code{match.arg()} but more flexible.
#'
#' @family Checks
#' @param inputs Vector character. Check options.
#' @param opts Vector character. Valid options.
#' @param input_name Character. Custom your message and change "input"
#' for any other string. For example: "column names".
#' @param type Character. Options: "all", "any."
#' @param not Character. Options: "stop", "message", "print", "return".
#' @param quiet Boolean. Keep quiet? If not, returns logical value.
#' @return Boolean. Result of \code{inputs} in \code{opts} (options).
#' Depending on \code{type} and/or \code{stop} arguments,
#' errors or messages will be shown.
#' @examples
#' opts <- c("A", "B", "C")
#' # Let's check the "all" logic
#' check_opts(inputs = c("A", "B"), opts, quiet = FALSE)
#' check_opts(inputs = c("X"), opts, not = "message", quiet = FALSE)
#' check_opts(inputs = c("A", "X"), opts, input_name = "value", not = "warning")
#' # Now let's check the "any" logic
#' check_opts(inputs = c("A", "X"), opts, type = "any")
#' check_opts(inputs = c("X"), opts, type = "any", not = "message")
#' check_opts(inputs = c("A", NA), opts, type = "any")
#' # Final trick: just ignore results
#' check_opts(inputs = "X", opts, not = "invisible")
#' @export
#' @rdname checks
check_opts <- function(inputs, opts,
                       input_name = "input",
                       type = "all", not = "stop",
                       quiet = TRUE) {
  aux <- base::get(type)
  not <- base::get(not)
  isit <- aux(inputs %in% opts)
  if (!isit) {
    if (type == "all") {
      inputs <- inputs[which(!inputs %in% opts)]
    }
    not(paste(
      "Your", input_name, formatColoured(v2t(inputs), cat = FALSE),
      ifelse(length(inputs) > 1, "are", "is"),
      "NOT valid;", toupper(type), "inputs should match",
      "these options:", formatColoured(v2t(opts), colour = "green", cat = FALSE)
    ))
  }
  if (!quiet) isit
}

####################################################################
#' Validate attributions
#'
#' This function checks if an object has a specific attribute and
#' stops if not.
#'
#' @family Checks
#' @param object Object of any kind
#' @param attr Character. Attribute to check
#' @param check Character. Attribute value
#' @param stop Boolean. Stop if doesn't check?
#' @return No return value, called for side effects.
#' @examples
#' test <- data.frame()
#' attributes(test)
#' check_attr(test, "class", "data.frame")
#' # check_attr(test, "class", "not.data.frame")
#' @export
#' @rdname checks
check_attr <- function(object, attr = "type", check = NULL, stop = TRUE) {
  if (is.null(check)) stop("You must set a valid check input")
  if (is.null(attr)) {
    attrs <- attributes(object)
    if (length(attrs) > 0) {
      stop(
        "Your attr input should be any of: ",
        formatColoured(v2t(attrs), colour = "green", cat = FALSE)
      )
    } else {
      stop("Your objct has no attributes to be checked")
    }
  }
  aux <- attr(object, attr)
  if (is.null(aux)) aux <- "Noclass"
  if (aux != check) {
    msg <- paste("Your object must be", attr, check)
    if (stop) stop(msg) else message(msg)
  }
}

####################################################################
#' Check if input is_* or are_*
#'
#' Check whether a value or vector is or is not following a set
#' of rules. For example: is an URL, is an ID vector, are non-variant or
#' constant values, are binary values... Notice that \code{is_} will return
#' the result for each observation and \code{are_} for the whole vector.
#'
#' @family Checks
#' @param x Vector
#' @param ... Additional parameters passed to \code{grepl()}
#' @return \code{is_url}. Boolean. Result of checking if \code{x} is a valid URL string.
#' @examples
#' is_url(c("google.com", "http://google.com"))
#'
#' is_ip(c("163.114.132.0", "7.114.132", "0.0.0.0", "1.1.1.1."))
#'
#' are_id(1:10)
#' are_id(LETTERS[1:10])
#'
#' are_constant(rep(1, 10))
#' are_constant(1:10)
#'
#' are_binary(c("A", "B", "A"))
#'
#' is_even(1:5)
#' is_odd(1:5)
#' is_odd(c(0, 1.5, 2.5, NA, Inf, NULL))
#' @rdname checks
#' @export
is_url <- function(x, ...) {
  grepl("(http|https)://[a-zA-Z0-9./?=_%:-]*", x, ...)
}

#' @rdname checks
#' @return \code{is_ip}. Boolean. Result of checking if \code{x} is a valid IP string.
#' @export
is_ip <- function(x, ...) {
  regex <- paste0(
    "^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.)",
    "{3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
  )
  grepl(regex, x, ...)
}

#' @rdname checks
#' @return \code{are_id}. Boolean. Result of checking if \code{x} is a potential ID vector
#' @export
are_id <- function(x) {
  is.character(x) && length(unique(x)) == length(x)
}

#' @rdname checks
#' @return \code{are_constant}. Boolean. Result of checking if \code{x} is a constant vector
#' @export
are_constant <- function(x) {
  length(unique(x)) == 1
}

#' @rdname checks
#' @return \code{are_binary}. Boolean. Result of checking if \code{x} is a binary vector
#' @export
are_binary <- function(x) {
  length(unique(x)) == 2
}

#' @rdname checks
#' @export
is_even <- function(x) x %% 2 == 0
#' @rdname checks
#' @export
is_odd <- function(x) x %% 2 == 1

####################################################################
#' ARIMA Forecast
#'
#' This function automates the ARIMA iterations and modeling for
#' time forecasting. For the moment, units can only be days.
#'
#' The ARIMA method is appropriate only for a time series that is
#' stationary (i.e., its mean, variance, and autocorrelation should
#' be approximately constant through time) and it is recommended
#' that there are at least 50 observations in the input data.
#'
#'  The model consists of two parts, an autoregressive (AR) part
#'  and a moving average (MA) part. The AR part involves regressing
#'  the variable on its own lagged (i.e., past) values. The MA part
#'  involves modeling the error term as a linear combination of error
#'  terms occurring contemporaneously and at various times in the past.
#'
#' One thing to keep in mind when we think about ARIMA models is
#' given by the great power to capture very complex patters of
#' temporal correlation (Cochrane, 1997: 25)
#'
#' @family Forecast
#' @param time POSIX. Vector with date values
#' @param values Numeric. Vector with numerical values
#' @param n_future Integer. How many steps do you wish to forecast?
#' @param ARMA Integer. How many days should the model look back for ARMA?
#' Between 5 and 10 days recommmended. If set to 0 then it will forecast
#' until the end of max date's month; if set to -1, until the end of
#' max date's following month
#' @param ARMA_min Integer. How many days should the model look back for ARMA?
#' Between 5 and 10 days recommmended. If set to 0 then it will forecast
#' until the end of max date's month; if set to -1, until the end of
#' max date's following month
#' @param AR Integer. Force AR value if known
#' @param MA Integer. Force MA value if known
#' @param wd_excluded Character vector. Which weekdays are excluded in
#' your training set. If there are, please define know which ones. Example:
#' c('Sunday','Thursday'). If set to 'auto' then it will detect automatically
#' which weekdays have no data and forcast without these days.
#' @param plot Boolean. If you wish to plot your results
#' @param plot_days Integer. How many days back you wish to plot?
#' @param project Character. Name of your forecast project
#' @return List. Containing the trained model, forecast accuracy results,
#' data.frame for forecast (test) and train, and if \code{plot=TRUE}, a plot.
#' @export
forecast_arima <- function(time, values, n_future = 30,
                           ARMA = 8, ARMA_min = 5,
                           AR = NA, MA = NA,
                           wd_excluded = NA,
                           plot = TRUE, plot_days = 90, project = NA) {
  try_require("forecast")

  # ARIMA doesn't use zeroes!
  time <- time[!values == 0]
  values <- values[!values == 0]

  if (length(time) < 50) {
    message("It is recommended that there are at least 50 observations in the input data")
  }

  if (Sys.Date() %in% time) {
    message("It is recommended that you do NOT use today's data for training your data")
  }

  if (n_future == -1) {
    n_future <- ceiling_date(Sys.Date(), "month") + months(1) - Sys.Date()
  }
  if (n_future == 0) {
    n_future <- ceiling_date(Sys.Date(), "month") - Sys.Date()
  }

  # Which AR and MA values minimize our AIC
  if (is.na(AR) && is.na(MA)) {
    arma <- c(ARMA_min:ARMA)
    aic <- expand.grid(AR = arma, MA = arma, cals = 0)
    message("Iterating for best AR / MA combinations; there are ", nrow(aic), "!")
    # if (length(time) > 1000) { method <- "ML" } else { method <- "CSS" }
    for (i in seq_len(nrow(aic))) {
      Tmodel <- Arima(values, order = c(aic$AR[i], 1, aic$MA[i]), method = "ML")
      aic$cals[i] <- Tmodel$aic
    }
    AR <- aic$AR[which.min(aic$cals)]
    MA <- aic$MA[which.min(aic$cals)]
    message("Best combination:", AR, "and", MA)
    aic_ARIMA <- min(aic$cals)
  }

  model <- Arima(values, order = c(AR, 1, MA), method = "ML")
  train <- data.frame(time, values,
    pred = model$fitted,
    resid = model$residuals
  )

  # Forecast
  future_dates <- seq.Date(max(time) + 1, max(time) %m+% days(n_future), by = 1)
  if (!is.na(wd_excluded)) {
    if (wd_excluded == "auto") {
      weekdays <- data.frame(table(weekdays(time)))
      weekdays_real <- c(weekdays(seq.Date(Sys.Date(), Sys.Date() + 6, by = 1)))
      wd_excluded <- weekdays_real[!weekdays_real %in% weekdays$Var1]
      message("Automatically excluding ", vector2text(wd_excluded))
    }
    exclude <- vector2text(wd_excluded, quotes = FALSE)
    future_dates <- future_dates[!weekdays(future_dates) %in% wd_excluded]
    n_future <- length(future_dates)
  }
  f <- forecast(model, h = n_future)
  test <- data.frame(time = future_dates, pred = f$mean, data.frame(f)[, -1])

  # Outut list with all results
  output <- list(
    model = model,
    metrics = accuracy(model),
    forecast = test,
    train = train
  )

  # Plot results
  if (plot == TRUE) {
    if (nrow(train) > plot_days) {
      train <- train[(nrow(train) - plot_days):nrow(train), ]
    }

    plotdata <- data.frame(
      rbind(
        data.frame(date = train$time, values = train$values, type = "Real"),
        data.frame(date = train$time, values = train$pred, type = "Model"),
        data.frame(date = test$time, values = test$pred, type = "Forecast")
      )
    )
    rects <- data.frame(start = min(future_dates), end = max(future_dates))

    output$plot <- ggplot(plotdata, aes(.data$date)) +
      geom_smooth(aes(y = .data$values), method = "loess", alpha = 0.5) +
      geom_line(aes(y = .data$values, colour = .data$type)) +
      labs(x = "Date", y = "Counter", colour = "") +
      theme_minimal() +
      theme(
        legend.position = "top",
        axis.text.x = element_text(angle = 60, hjust = 1)
      ) +
      scale_x_date(date_breaks = "1 month", date_labels = "%b-%Y") +
      ggtitle("Real & Fitted Model vs Forecast (ARIMA)",
        subtitle = paste(
          "AIC", signif(output$model$aic, 4), "|",
          "MAE", signif(output$metrics[3], 3), "|",
          "RMSE", signif(output$metrics[2], 3), "|",
          "ARIMA:", AR, "- 1 -", MA
        )
      ) +
      scale_color_manual(values = c("orange", "navy", "purple")) +
      geom_rect(
        data = rects, inherit.aes = FALSE,
        aes(
          xmin = .data$start, xmax = .data$end,
          ymin = min(plotdata$values),
          ymax = max(plotdata$values)
        ),
        color = "transparent", fill = "grey", alpha = 0.25
      )
    if (!is.na(project)) {
      output$plot <- output$plot + labs(caption = project)
    }
    plot(output$plot)
  }
  output
}


####################################################################
#' Facebook's Prophet Forecast
#'
#' Prophet is Facebook's procedure for forecasting time series data
#' based on an additive model where non-linear trends are fit with
#' yearly, weekly, and daily seasonality, plus holiday effects. It
#' works best with time series that have strong seasonal effects and
#' several seasons of historical data. Prophet is robust to missing
#' data and shifts in the trend, and typically handles outliers well.
#'
#' Official documentation: \url{https://github.com/facebook/prophet}
#'
#' @family Forecast
#' @param df Data frame. Must contain date/time column and values column,
#' in that order.
#' @param n_future Integer. How many steps do you wish to forecast?
#' @param country Character. Country code for holidays.
#' @param trend.param Numeric. Flexibility of trend component. Default is 0.05,
#' and as this value becomes larger, the trend component will be more flexible.
#' @param logged Boolean. Convert values into logs?
#' @param pout Numeric. Get rid of pout \% of outliers.
#' @param project Character. Name of your forecast project for plot title
#' @return List. Containing the forecast results, the prophet model, and a plot.
#' @export
prophesize <- function(df, n_future = 60, country = NULL,
                       trend.param = 0.05, logged = FALSE, pout = 0.03,
                       project = "Prophet Forecast") {
  try_require("prophet")

  df <- data.frame(df[, c(1, 2)])
  metric <- colnames(df)[2]
  colnames(df) <- c("ds", "y")
  df <- arrange(df, .data$ds)
  if (logged) df$y <- log(df$y)

  # Outliers
  df <- df[!rank(-df$y) %in% 1:round(nrow(df) * pout), ]

  # Run prophet functions
  m <- prophet(
    yearly.seasonality = TRUE, daily.seasonality = FALSE,
    changepoint.prior.scale = trend.param
  )
  if (!is.null(country)) {
    m <- add_country_holidays(m, country_name = country)
  }
  m <- fit.prophet(m, df)
  future <- make_future_dataframe(m, periods = n_future)

  forecast <- predict(m, future)
  forecast$y <- forecast$trend + forecast$additive_terms

  p <- plot(m, forecast) + theme_lares() +
    labs(
      y = metric, x = "Dates",
      title = project,
      subtitle = paste("Forecast results for the next", n_future, "days")
    ) +
    scale_y_comma()

  plots2 <- prophet_plot_components(m, forecast, render_plot = FALSE, weekly_start = 1)
  plots2 <- lapply(plots2, function(x) x + theme_lares())
  plot2 <- wrap_plots(plots2, ncol = 1) +
    plot_annotation(title = "Forecast components", theme = theme_lares())

  list(result = forecast, model = m, plot = p, components = plot2)
}

####################################################################
#' Scrabble: Dictionaries
#'
#' Download words from 4 different languages: English, Spanish,
#' German, and French. Words will be save into the \code{temp} directory.
#' This is an auxiliary function. You may want to use \code{scrabble_words}
#' directly if you are searching for the highest score words!
#'
#' @family Games
#' @param lang_dic Character. Any of "en","es","de","fr". Set to NULL
#' if you wish to skip this step (and use \code{words} parameter in
#' \code{scrabble_words} instead).
#' @return data.frame with words and language columns.
#' @export
#' @rdname scrabble
scrabble_dictionary <- function(lang_dic, quiet = FALSE) {
  if (is.null(lang_dic)) {
    invisible(NULL)
  } else {
    if (length(lang_dic) != 1) {
      stop("Select only 1 language at a time...")
    }
    check_opts(lang_dic, c("en", "es", "de", "fr"))
    if (cache_exists(lang_dic)) {
      words <- cache_read(lang_dic, quiet = quiet)
      if (!quiet) {
        message(sprintf(
          ">>> Loaded %s '%s' words",
          formatNum(nrow(words), 0), lang_dic
        ))
      }
      words
    } else {
      if (!haveInternet()) {
        message("No internet connetion...")
        invisible(NULL)
      } else {
        message(sprintf(
          ">>> Downloading '%s' words. Source: %s",
          lang_dic, "github.com/lorenbrichter/Words"
        ))
        url <- sprintf(
          "https://raw.githubusercontent.com/lorenbrichter/Words/master/Words/%s.txt",
          lang_dic
        )
        words <- read.table(url, col.names = "words")
        words$language <- lang_dic
        cache_write(words, lang_dic, quiet = quiet)
        if (!quiet) message(sprintf(">>> Saved (%s words) into cache", formatNum(nrow(words), 0)))
        words
      }
    }
  }
}


####################################################################
#' Scrabble: Word Scores
#'
#' Get score for any word or list of words. You may set manually depending
#' on the rules and languages you are playing with. Check the examples
#' for Spanish and English values when I played Words With Friends.
#'
#' @param words Character vector. Words to score
#' @param scores.df Dataframe. Must contain two columns: "tiles" with every
#' letter of the alphabet and "scores" for each letter's score.
#' @return data.frame with word, scores, and length values for each \code{word}.
#' @examples
#' \donttest{
#' if (haveInternet()) {
#'   # For Spanish words (default)
#'   es_scores <- scrabble_points("es")
#'   # Custom scores for each letter
#'   cu_scores <- data.frame(
#'     tiles = tolower(LETTERS),
#'     scores = c(
#'       1, 1, 1, 1, 1, 1, 1, 5, 1, 1, 5, 2, 4, 2, 1,
#'       4, 10, 1, 1, 1, 2, 5, 4, 8, 3, 10
#'     )
#'   )
#'
#'   # Score values for each set of rules
#'   words <- c("Bernardo", "Whiskey", "R is great")
#'   scrabble_score(words, es_scores)
#'   scrabble_score(words, cu_scores)
#' }
#' }
#' @export
#' @rdname scrabble
scrabble_score <- function(words, scores.df) {
  scores <- data.frame(
    tiles = tolower(scores.df$tiles),
    scores = as.integer(scores.df$scores)
  )
  counter <- lapply(words, function(word) {
    lapply(scores$tiles, function(letter) {
      str_count(word, letter)
    })
  })
  scores <- unlist(lapply(counter, function(x) sum(unlist(x) * scores$scores)))
  data.frame(word = words, scores) %>%
    mutate(length = nchar(word)) %>%
    arrange(desc(scores))
}


####################################################################
#' Scrabble: Tiles Points
#'
#' Dataframe for every letter and points given a language.
#'
#' @param lang Character. Any of "en","es" or "chars". Set to NULL
#' if you wish to skip this step (and use \code{words} parameter in
#' \code{scrabble_words()} instead). The "chars" parameter will
#' score the number of characters a word has.
#' @return data.frame with tiles and scores for each alphabet letter.
#' @examples
#' scrabble_points("es")
#' scrabble_points("en")
#' # Not yet available
#' scrabble_points("fr")
#' @export
#' @rdname scrabble
scrabble_points <- function(lang) {
  if (is.null(lang)) {
    message(">>> Skipping points schema...")
    invisible(NULL)
  } else {
    if (!lang %in% c("en", "es", "chars", "unique")) {
      message("There are no points structure for this language/system yet")
      invisible(NULL)
    } else {
      if (lang == "es") {
        scores <- data.frame(
          tiles = c(
            tolower(LETTERS)[1:14], intToUtf8(241),
            tolower(LETTERS)[15:length(LETTERS)]
          ),
          scores = c(
            1, 3, 2, 2, 1, 4, 3, 4, 1, 8, 10, 1, 3, 1,
            8, 1, 3, 5, 1, 1, 1, 2, 4, 10, 10, 5, 10
          )
        )
      }
      if (lang == "en") {
        scores <- data.frame(
          tiles = tolower(LETTERS),
          scores = c(
            1, 4, 4, 2, 1, 4, 3, 3, 1, 10, 5, 2, 4,
            2, 1, 4, 10, 1, 1, 1, 2, 5, 4, 8, 3, 10
          )
        )
      }
      if (lang %in% c("chars", "unique")) {
        scores <- data.frame(
          tiles = tolower(LETTERS)
        ) %>% mutate(scores = 1)
      }
      message(sprintf(">>> Points system: '%s'", lang))
      scores
    }
  }
}


####################################################################
#' Pattern Matching for Letters considering Blanks
#'
#' Match pattern of letters considering blanks within each element of a
#' character vector, allowing counted characters between and around
#' each letter. Used as an auxiliary function for the Scrabble family
#' of functions.
#'
#' @param x Character vector
#' @param pattern Character. Character string containing a
#' semi-regular expression which uses the following logic:
#' "a_b" means any character that contains "a" followed by
#' something followed by "b", anywhere in the string.
#' @param blank Character. String to use between letters.
#' @return Boolean check for each value on \code{x}.
#' @examples
#' x <- c("aaaa", "bbbb", "baba", "aabb", "a", "ab")
#' grepl_letters(x, "ab")
#' grepl_letters(x, "_ab")
#' grepl_letters(x, "a_a")
#' grepl_letters(x, "c")
#' @export
grepl_letters <- function(x, pattern, blank = "_") {
  # If no blank tile, use simple grepl
  if (!grepl(blank, pattern, fixed = TRUE)) {
    grepl(pattern, x, fixed = TRUE)
  } else {
    tiles <- strsplit(tolower(pattern), "")[[1]]
    n_tiles <- length(tiles)
    sapply(x, function(word) {
      chars <- strsplit(tolower(word), "")[[1]]
      n_chars <- length(chars)
      if (n_tiles > n_chars) {
        FALSE
      } else {
        possible_positions <- n_chars - n_tiles + 1
        any(sapply(0:(possible_positions - 1), function(offset) {
          slice <- chars[(1 + offset):(offset + n_tiles)]
          all(tiles == blank | tiles == slice)
        }))
      }
    })
  }
}

####################################################################
#' Scrabble: Highest score words finder
#'
#' Find highest score words given a set of letters, rules, and
#' language to win at Scrabble! You just have to find the best
#' place to post your tiles.
#'
#' @param tiles Character. The letters you wish to consider.
#' @param free Integer. How many free blank tiles you have?
#' @param force_start,force_end Character. Force words to start or end with
#' a pattern of letters and position. Examples: "S" or "SO" or "__S_O"...
#' If the string contains tiles that were not specified in \code{tiles}, they
#' will automatically be included.
#' @param force_str Character vector. Force words to contain strings.
#' If the string contains tiles that were not specified in \code{tiles}, they
#' will automatically be included.
#' @param force_exclude,exclude_here Character vector. Exclude words containing
#' these tiles (and positions). Not very relevant on Scrabble but for Wordle.
#' @param force_n,force_max Integer. Force words to be n or max n characters
#' long. Leave 0 to ignore parameter.
#' @param pattern Character string. Custom regex patterns you'd like to match.
#' @param repeated Boolean. By default, no replacement allowed. When activated,
#' a single tile can be repeated and won't be "used and discarded".
#' @param scores,language Character. Any of "en","es","de","fr".
#' If scores is not any of those languages, must be a data.frame that
#' contains two columns: "tiles" with every letter of the alphabet and
#' "scores" for each letter's score. If you wish
#' to overwrite or complement this dictionaries other words you can set to
#' \code{"none"} and/or use the \code{words} parameter.
#' You might also want to set this parameter globally with
#' \code{Sys.setenv("LARES_LANG" = "en")} and forget about it!
#' @param words Character vector. Use if you wish to manually add words.
#' @param quiet Boolean. Do not print words as they are being searched.
#' @param print Boolean. Print how many words are left by step.
#' @return data.frame with matching words found, sorted by higher points.
#' @examples
#' \donttest{
#' if (haveInternet()) {
#'   # Automatic use of languages and scores
#'   Sys.setenv("LARES_LANG" = "es")
#'   scrabble_words(
#'     tiles = "hola",
#'     free = 2,
#'     force_start = "h",
#'     force_n = 4,
#'     force_str = "_o_a",
#'     exclude_here = "__z|j"
#'   )
#'
#'   wordle <- c("board", "tempo", "shoes", "hoard")
#'   scrabble_words(
#'     language = NULL,
#'     words = wordle,
#'     force_n = 5,
#'     force_str = "O_R"
#'   )
#'
#'   # Words considered for a language (you can custom it too!)
#'   es_words <- scrabble_dictionary("es")
#' }
#' }
#' @export
#' @rdname scrabble
scrabble_words <- function(tiles = "",
                           free = 0,
                           force_start = "",
                           force_end = "",
                           force_str = "",
                           force_exclude = "",
                           exclude_here = "",
                           force_n = 0,
                           force_max = 0,
                           pattern = "",
                           repeated = FALSE,
                           language = Sys.getenv("LARES_LANG"),
                           scores = language,
                           words = NULL,
                           quiet = FALSE,
                           print = TRUE) {
  ### POINTS

  tiles <- paste(tiles, collapse = "")
  if (is.data.frame(scores)) {
    if (!all(colnames(scores) %in% c("tiles", "scores"))) {
      stop("Please, provide a valid scores data.frame with 'tiles' and 'scores' columns")
    }
  } else {
    scores.df <- scrabble_points(scores)
  }

  ### TILES

  message(">>> Setting up tiles...")
  # Split letters
  tiles <- tolower(unlist(strsplit(tiles, "")))
  # Add free letters/tiles
  if (free > 0) tiles <- c(tiles, rep("_", free))
  # Add logical tiles when using force_ arguments
  tiles <- .add_letters(force_start, tiles)
  tiles <- .add_letters(force_end, tiles)
  tiles <- .add_letters(force_str, tiles)
  # Force N letters (complete free)
  if (force_n > 0 && length(tiles) < force_n) {
    tiles <- c(tiles, rep("_", times = (force_n - length(tiles))))
  }
  ntiles <- as.integer(length(tiles))
  # Exclude these tiles
  force_not <- unique(tolower(unlist(strsplit(force_exclude, ""))))

  message("Tiles: ", v2t(toupper(tiles)))

  ### WORDS

  # Consolidate dictionary
  dictionary <- scrabble_dictionary(language)[, 1]
  nwords <- length(dictionary)
  if (!is.null(words)) {
    dictionary <- unique(tolower(c(words, dictionary)))
    unique_new <- length(dictionary) - nwords
    message(sprintf(">>> Added %s custom words %s", formatNum(length(words), 0), ifelse(
      unique_new != length(words), sprintf("(%s new)", unique_new), ""
    )))
  }
  words <- tolower(dictionary)
  # Words can't have more letters than inputs by default
  if (!repeated) words <- words[nchar(words) <= ntiles]
  .temp_print(length(words))
  # Exclude specific tiles (Wordle)
  if (length(force_not) > 0) words <- words[!grepl(paste(force_not, collapse = "|"), words)]
  .temp_print(length(words))
  # You may want to force their lengths
  if (force_n > 0) words <- words[nchar(words) == force_n]
  .temp_print(length(words))
  if (force_max > 0 && !repeated) words <- words[nchar(words) <= force_max]
  .temp_print(length(words))
  # Words can't have different letters than inputs
  words <- words[.all_tiles_present(words, tiles, free = 0, repeated = repeated)]
  .temp_print(length(words))
  # Force start/end strings
  words <- .force_words(words, force_start)
  .temp_print(length(words))
  words <- .force_words(.reverse(words), .reverse(force_end), rev = TRUE)
  .temp_print(length(words))
  # Force strings that must be contained
  if (force_str[1] != "") {
    for (str in force_str) {
      words <- words[grepl_letters(words, pattern = tolower(str), blank = "_")]
      .temp_print(length(words))
    }
  }
  # Force custom patterns that must be contained
  if (pattern[1] != "") {
    for (str in pattern) {
      words <- words[grep(tolower(str), words, perl = TRUE)]
      .temp_print(length(words))
    }
  }
  # Exclude letters from positions (Wordle)
  if (exclude_here[1] != "") {
    for (eh in exclude_here) {
      pos_tiles <- str_split_merge(tolower(eh))
      for (i in seq_along(pos_tiles)) {
        these <- str_split(pos_tiles, "\\|")[i][[1]]
        if (!any(these %in% letters)) next
        located <- stringr::str_locate_all(words, pos_tiles[i])
        these <- !unlist(lapply(located, function(x) sum(x[, 1] == i) > 0))
        words <- words[these]
        .temp_print(length(words))
      }
    }
  }

  .temp_print(length(words), last = TRUE)
  if (length(words) > 0) {
    if ("unique" %in% scores) {
      this <- lapply(words, function(x) sum(!!str_count(x, letters)))
      done <- data.frame(word = words, scores = unlist(this)) %>%
        mutate(length = str_length(.data$word)) %>%
        arrange(desc(.data$scores), desc(.data$length))
    } else {
      done <- scrabble_score(words, scores.df)
    }
    if (sum(done$scores) == 0) done$scores <- NULL
    as_tibble(done)
  } else {
    message("No words found with set criteria")
    NULL
  }
}

.temp_print <- function(x, print = TRUE, last = FALSE) {
  if (print) if (!last) formatColoured(paste(x, "> ")) else formatColoured(paste(x, "\n"))
}

# Tile used, tile that must be skipped on next iterations (when repeated is TRUE)
.all_tiles_present <- function(words, tiles, free = 0, repeated = FALSE) {
  free <- free + sum(tiles == "_")
  fx <- ifelse(repeated, gsub, sub)
  for (x in tiles) words <- fx(x, "", words)
  nchar(words) <= free
}

.force_words <- function(words, pattern, rev = FALSE, invert = FALSE) {
  forced <- tolower(unlist(strsplit(pattern, "")))
  forced_which <- which(forced != "_")
  for (i in forced_which) {
    these <- substr(words, i, i) == forced[i]
    if (invert) these <- !these
    words <- words[these]
  }
  if (rev) words <- .reverse(words)
  words
}

.reverse <- function(words) {
  original_class <- class(words)
  words <- as.character(words)
  splits <- lapply(words, function(x) strsplit(x, ""))
  reversed <- lapply(splits, function(x) rev(x[[1]]))
  words <- unlist(lapply(reversed, function(x) paste(x, collapse = "")))
  class(words) <- original_class
  words
}

.add_letters <- function(str, tiles) {
  if (str[1] != "") {
    str_tiles <- tolower(unlist(strsplit(str, "")))
    # Get rid of non alpha-numeric values
    str_tiles <- str_tiles[grepl("[[:alnum:]]", str_tiles)]
    which <- !str_tiles %in% c(tiles, "_")
    if (any(which)) {
      new <- str_tiles[which]
      tiles <- c(tiles, new)
      message(sprintf(
        "%s %s not in your tiles: now included",
        v2t(toupper(new), and = "and"),
        ifelse(length(new) > 1, "were", "was")
      ))
    }
  }
  tiles
}

# x <- "AB|CDE|F|GHIJK"
str_split_merge <- function(x, sep = "|") {
  temp <- stringr::str_split(x, "")[[1]]
  seps <- which(temp == "|")
  for (i in seq_along(temp)[-length(temp)]) {
    if (!i %in% c(seps, seps - 1)) temp[i] <- paste0(temp[i], ",")
  }
  stringr::str_split(paste(temp, collapse = ""), ",")[[1]]
}

# devtools::load_all()
# library(dplyr)
# library(stringr)
# Sys.setenv("LARES_LANG" = "en")
# words <- scrabble_dictionary("en")$words
#
# # Play and win!
# scrabble_words(tiles = "abcdef",
#                free = 0,
#                force_start = "be",
#                force_end = "do",
#                force_str = "n",
#                force_n = 0,
#                force_max = 0,
#                quiet = FALSE)
# # Fix this case (not using the tiles provided)
# scrabble_words(tiles = "bernardo",
#                free = 0,
#                quiet = FALSE)

# scrabble_words(
#   language = "en", # SOARE
#   tiles = "O",
#   force_start = "S",
#   force_end = "",
#   force_n = 5,
#   force_exclude = "ARE",
#   exclude_here = "_O__S"
# ) %>% pull(word)

####################################################################
#' Cross-MMM Budget Optimization across Models
#'
#' Given a list of recreated Robyn models, this function optimizes budget
#' allocation across MMM with respective constraints by maximizing
#' incremental revenue/conversions. This method assumes each model is
#' independent, that can be compared given its spends were cleanly and
#' properly split, they modeled the same metric (revenue or conversion)
#' and units (currency or type of conversion), and date granularity.
#' For best results, ensure channels have similar granularity across
#' markets to simplify interpretation and application of the outputs.
#'
#' @inheritParams mp3_get
#' @param budget_constr_low,budget_constr_up Numeric vector. Relative minimum
#' and maximum budgets to consider based on \code{initial_budgets}.
#' By default it'll consider 50% and 150% budget constraints.
#' Must be length 1 or same as \code{models}.
#' @param cores Integer. How many cores to use for parallel computations?
#' Set to 1 to not use this option.
#' Default will the minimum between 10 cores and all available cores - 1.
#' @param ... Additional parameters to be passed to internal functions.
#' @return Invisible vector with results by letter.
#' @examples
#' # You must have Robyn installed and some models stored as JSON files
#' \dontrun{
#' # Import and recreate the models
#' files <- c("BrandA.json", "BrandB.json", "BrandC.json", "BrandN.json")
#' models <- lapply(files, function(x) Robyn::robyn_recreate(x))
#' names(models) <- gsub("\\.json", "", files)
#'
#' # Calculate cross-brand optimal allocation
#' res <- robyn_xmodels(
#'   models,
#'   cores = 10,
#'   start_dates = "2023-01-01",
#'   end_dates = "2023-12-01"
#' )
#' print(res)
#' res$summary
#' }
#' @export
#' @rdname robyn_crossmmm
robyn_xmodels <- function(
    models, initial_budgets = NULL,
    start_dates = NULL,
    end_dates = NULL,
    budget_constr_low = 0.5,
    budget_constr_up = 1.5,
    channel_constr_low = budget_constr_low,
    channel_constr_up = budget_constr_up,
    cores = NULL,
    quiet = FALSE,
    ...) {
  # Initial state calculations
  perfs <- lapply(models, function(x) {
    robyn_performance(
      x$InputCollect, x$OutputCollect,
      start_date = NULL, end_date = NULL
    ) %>%
      filter(.data$channel == "PROMOTIONAL TOTAL")
  })

  # Extract start and end dates when not provided
  extract_dates <- function(dates, perfs, date_col, quiet) {
    if (is.null(dates)) {
      dates <- as.Date(
        unlist(lapply(perfs, function(x) unlist(pull(x, .data[[date_col]])))),
        origin = "1970-01-01"
      )
      if (!quiet) {
        message("Extracted ", date_col, ": ", v2t(sprintf("%s (%s)", dates, names(dates))))
      }
    } else {
      if (length(dates) == 1) dates <- rep(dates, length(models))
    }
    dates
  }
  start_dates <- extract_dates(start_dates, perfs, "start_date", quiet)
  end_dates <- extract_dates(end_dates, perfs, "end_date", quiet)

  # Extract initial budgets for the date range
  if (is.null(initial_budgets)) {
    for (i in seq_along(models)) {
      p <- robyn_performance(
        models[[i]]$InputCollect, models[[i]]$OutputCollect,
        start_date = start_dates[i], end_date = end_dates[i]
      ) %>%
        filter(.data$channel == "PROMOTIONAL TOTAL") %>%
        pull(.data$spend)
      initial_budgets <- c(initial_budgets, p)
    }
    names(initial_budgets) <- names(models)
    if (!quiet) {
      message("Extracted budgets: ", v2t(
        sprintf("%s (%s)", num_abbr(initial_budgets), names(initial_budgets))
      ))
    }
  } else {
    if (length(initial_budgets) == 1) initial_budgets <- rep(initial_budgets, length(models))
  }

  # Total budget across brands
  total_budget <- sum(initial_budgets)

  # Run dims checks
  stopifnot(length(models) == length(start_dates))
  stopifnot(length(models) == length(end_dates))
  stopifnot(length(models) == length(initial_budgets))

  # Function to calculate total incremental sales given a budget distribution
  calculate_isales <- function(models, budgets, isales = TRUE,
                               start_dates = NULL, end_dates = NULL,
                               budget_constr_low = 0.5,
                               budget_constr_up = 1.5,
                               channel_constr_low = budget_constr_low,
                               channel_constr_up = budget_constr_up,
                               quiet = FALSE, ...) {
    try_require("Robyn")
    stopifnot(length(models) == length(budgets))
    bas <- list()
    for (i in seq_along(budgets)) {
      bas[[names(models)[i]]] <- robyn_allocator(
        InputCollect = models[[i]]$InputCollect,
        OutputCollect = models[[i]]$OutputCollect,
        total_budget = budgets[i],
        date_range = c(start_dates[i], end_dates[i]),
        scenario = "max_response",
        channel_constr_low = channel_constr_low,
        channel_constr_up = channel_constr_up,
        plots = FALSE,
        export = FALSE,
        quiet = TRUE,
        ...
      )
    }
    if (isales) {
      unlist(lapply(bas, function(x) {
        x$dt_optimOut$optmResponseTotal[1] - x$dt_optimOut$initResponseTotal[1]
      }))
    } else {
      bas
    }
  }

  # Function to be optimized (minimized)
  objective_function <- function(budgets, total_budget, models,
                                 start_dates = NULL, end_dates = NULL,
                                 budget_constr_low = 0.5,
                                 budget_constr_up = 1.5,
                                 channel_constr_low = budget_constr_low,
                                 channel_constr_up = budget_constr_up,
                                 cores = 10, quiet = FALSE,
                                 ...) {
    # Value to minimize: penalty (tends to zero) - incremental total sales (max)
    isales <- calculate_isales(
      models, budgets,
      start_dates = start_dates,
      end_dates = end_dates,
      budget_constr_low = budget_constr_low,
      budget_constr_up = budget_constr_up,
      channel_constr_low = channel_constr_low,
      channel_constr_up = channel_constr_up, ...
    )
    # Calculate the penalty for not using the total budget
    budget_penalty <- abs(sum(budgets) - total_budget)
    sum(isales) - budget_penalty
  }

  # Optimize budgets across brands
  if (TRUE) {
    tic("robyn_crossbrand")
    if (cores > 1) {
      try_require("parallel")
      try_require("optimParallel")
      if (is.null(cores)) cores <- min(10, detectCores() - 1)
      cl <- makeCluster(cores)
      setDefaultCluster(cl)
      optimfx <- optimParallel
    } else {
      optimfx <- stats::optim
      cl <- NULL
    }
    if (!quiet) {
      message(sprintf(
        ">>> Started cross-brand optimization: %s brands using %s cores...",
        length(models), cores
      ))
    }

    # Interpolate 75% value between limits as initial budget seed
    mult <- budget_constr_low + 0.75 * (budget_constr_up - budget_constr_low)

    result <- tryCatch(
      {
        optimfx(
          parallel = list(cl = cl, loginfo = TRUE),
          fn = objective_function,
          par = initial_budgets * mult,
          lower = initial_budgets * budget_constr_low,
          upper = initial_budgets * budget_constr_up,
          total_budget = total_budget,
          models = models,
          start_dates = start_dates,
          end_dates = end_dates,
          budget_constr_low = budget_constr_low,
          budget_constr_up = budget_constr_up,
          channel_constr_low = channel_constr_low,
          channel_constr_up = channel_constr_up,
          method = "L-BFGS-B",
          control = list(fnscale = -1),
          ...
        )
      },
      error = function(e) {
        if (cores > 1) suppressWarnings(stopCluster(cl))
        stop(e)
      }
    )
    if (cores > 1) suppressWarnings(stopCluster(cl))
    if (!quiet) message(result$message)
    result$elapsed <- toc("robyn_crossbrand")$time
  }

  # Calculate allocation results with optima
  allocations <- calculate_isales(
    models, result$par,
    isales = FALSE,
    start_dates = start_dates,
    end_dates = end_dates,
    budget_constr_low = budget_constr_low,
    budget_constr_up = budget_constr_up,
    channel_constr_low = channel_constr_low,
    channel_constr_up = channel_constr_up,
    ...
  )

  df <- dplyr::tibble(
    model = names(allocations),
    start_date = unlist(start_dates),
    end_date = unlist(end_dates),
    budget = initial_budgets,
    response = unlist(lapply(allocations, function(x) {
      x$dt_optimOut$initResponseTotal[1]
    })),
    mresp_init = unlist(lapply(allocations, function(x) {
      sum(x$dt_optimOut$initResponseMargUnit)
    })),
    optbudget = unlist(lapply(allocations, function(x) {
      x$dt_optimOut$optmSpendTotal[1]
    })),
    optresponse = unlist(lapply(allocations, function(x) {
      x$dt_optimOut$optmResponseTotal[1]
    })),
    mresp_optm = unlist(lapply(allocations, function(x) {
      sum(x$dt_optimOut$optmResponseMargUnit)
    }))
  ) %>%
    mutate(
      budgetpct = round(.data$optbudget / .data$budget, 2),
      # initroas = round(response / budget, 2),
      # optroas = round(optresponse / optbudget, 2),
      incremental = .data$optresponse - .data$response,
      iperc = round(100 * .data$incremental / .data$response, 2)
    )

  # Final outputs
  out <- list(
    inputs = list(
      initial_budgets = initial_budgets,
      total_budget = total_budget,
      budget_constr_low = budget_constr_low,
      channel_constr_up = channel_constr_up,
      channel_constr_low = channel_constr_low,
      channel_constr_up = channel_constr_up,
      cores = cores
    ),
    optimization = result,
    allocations = allocations,
    incremental = sum(df$incremental),
    summary = df,
    mresp_init = sum(unlist(lapply(allocations, function(x) {
      sum(x$dt_optimOut$initResponseMargUnit)
    }))),
    mresp_optm = sum(unlist(lapply(allocations, function(x) {
      sum(x$dt_optimOut$optmResponseMargUnit)
    }))),
    other_arguments = list(...)
  )
  class(out) <- c("robyn_crossmmm", class(out))
  out
}

#' @rdname robyn_crossmmm
#' @param x Object to print: robyn_crossmmm() output.
#' @export
print.robyn_crossmmm <- function(x, ...) {
  cat(glued(
    "
Initial State:
  Models: {paste(names(x$allocations), collapse = ', ')}
  Date Ranges: {date_ranges}
  Budgets: {paste(num_abbr(x$inputs$initial_budgets), collapse = ', ')}
  Distribution: {pct_string(x$inputs$initial_budgets / x$inputs$total_budget)}
  Total Budget: {num_abbr(x$inputs$total_budget, 3)}
  Marginal Response: {signif(mresp_init, 4)}

Iterations:
  Time Elapsed: {x$optimization$elapsed}
  Total Iterations: {x$optimization$counts[[1]]}
  Convergence: {x$optimization$convergence}
  Cores: {x$inputs$cores}

Optimized State:
  Budget Increase: {num_abbr(bi)} ({formatNum(100 * bi / x$inputs$total_budget, abbr = TRUE, sign = TRUE, pos = '%')})
  Incremental: {num_abbr(x$incremental)}
  Optimal Budgets: {paste(num_abbr(x$optimization$par), collapse = ', ')}
  Optimal Distribution: {pct_string(x$optimization$par / x$inputs$total_budget)}
  Compared with Initial: {pct_string(x$optimization$par / x$inputs$initial_budgets)}
  Optimal Marginal Response: {signif(mresp_optm, 4)}
 ",
    pct_string = function(x) {
      paste(paste0(num_abbr(100 * x), "%"), collapse = ", ")
    },
    date_ranges = if (all(length(unique(x$summary$start_date))) == 1 && all(length(unique(x$summary$end_date))) == 1) {
      paste(x$summary$start_date[1], x$summary$end_date[1], sep = ":")
    } else {
      paste(paste(x$summary$start_date, x$summary$end_date, sep = ":"), collapse = ", ")
    },
    mresp_init = sum(unlist(lapply(x$allocations, function(x) {
      sum(x$dt_optimOut$initResponseMargUnit)
    }))),
    mresp_optm = sum(unlist(lapply(x$allocations, function(x) {
      sum(x$dt_optimOut$optmResponseMargUnit)
    }))),
    bi = round(sum(x$summary$optbudget) - x$inputs$total_budget)
  ))
}

####################################################################
#' PostgreSQL Queries on Database (Read)
#'
#' This function lets the user query a PostgreSQL database. Previously
#' was called \code{queryDummy} but was replaced and deprecated for a
#' more general function by using the \code{from} parameter.
#'
#' @family Credentials
#' @family Database
#' @param query Character. SQL Query
#' @param from Character. Credential's user (see \code{get_creds()})
#' @param creds Character. Credential's directory (see \code{get_creds()})
#' @return data.frame. Result of fetching the \code{query} data.
#' @examples
#' \dontrun{
#' # Query a PostgreSQL database
#' query <- "SELECT * FROM my_table LIMIT 10"
#' result <- queryDB(query, from = "my_database")
#' head(result)
#' }
#' @export
queryDB <- function(query, from, creds = NA) {
  try_require("RPostgreSQL")
  tic(id = "queryDummy")

  dw <- get_credentials(from = from, dir = creds)
  drv <- PostgreSQL()
  con <- dbConnect(drv,
    host = dw$server, dbname = dw$database,
    port = dw$port, user = dw$uid, password = dw$pwd
  )

  q <- dbSendQuery(con, query)
  q <- fetch(q, n = -1)
  dbDisconnect(con)

  toc(id = "queryDummy", msg = "Query duration:")
  q
}

####################################################################
#' Weighted Cross Tabulation
#'
#' A cross-tabulation function with output similar to STATA, tidy
#' friendly, with weighting possibility.
#'
#' @family Exploratory
#' @param df Data.frame.
#' @param ... Variables. Dependent and independent variables.
#' @param wt Variable, numeric. Weights.
#' @param prow,pcol,pall Boolean. Calculate percent values for rows, columns,
#' or the whole table, respectively.
#' @param decimals Integer. How many decimals should be returned?
#' @param rm.na Boolean. Remove NA values?
#' @param total Boolean. Return total values column?
#' @param order Boolean. Sort columns and rows by frequencies? Else, will
#' be sorted alphabetically
#' @return data.frame. Result of crossing the variables provided in \code{...} and
#' counting how many observations (rows) fall into each criteria.
#' @examples
#' data(dft) # Titanic dataset
#' crosstab(dft, Survived, Pclass, total = FALSE)
#' # Show values in percentages
#' crosstab(dft, Pclass, Survived, prow = TRUE)
#' crosstab(dft, Pclass, Survived, pall = TRUE)
#' # Weighted by another variable
#' crosstab(dft, Survived, Pclass, wt = Fare, prow = TRUE)
#' @export
crosstab <- function(df, ..., wt = NULL,
                     prow = FALSE, pcol = FALSE, pall = FALSE,
                     decimals = 2, rm.na = FALSE, total = TRUE,
                     order = TRUE) {
  if (pcol) {
    warning("Just change the order of your variables to get this results")
    invisible(NULL)
  } else {
    vars <- quos(...)
    wt <- enquo(wt)

    if (!length(vars) %in% c(2, 3)) {
      message("Use freqs() instead!")
      df %>% freqs(!!!vars)
    } else {
      names <- gsub("~", "", as.character(vars))
      names <- gsub("\\.data\\$", "", names)
      xname <- names[1]
      yname <- names[2]

      df <- df %>% mutate_at(vars(1:2), as.character)

      first <- df %>%
        freqs(!!vars[[1]], wt = !!wt, rm.na = rm.na) %>%
        select(!!vars[[1]]) %>%
        unlist()

      second <- df %>%
        freqs(!!vars[[2]], wt = !!wt, rm.na = rm.na) %>%
        select(!!vars[[2]]) %>%
        unlist()

      aux <- df %>%
        freqs(!!!vars, wt = !!wt, rm.na = rm.na) %>%
        replace(is.na(.), "")
      colnames(aux)[1:2] <- c("A", "B")

      # Order by frequency
      if (order) {
        aux <- aux %>%
          replaceall(NA, "NA", c("A", "B")) %>%
          mutate(
            A = factor(.data$A, levels = first),
            B = factor(.data$B, levels = second)
          )
      }

      ret <- aux %>%
        select(.data$A, .data$B, .data$n) %>%
        tidyr::spread(.data$B, .data$n, drop = FALSE) %>%
        select(.data$A, one_of(levels(aux$B))) %>%
        arrange()
      colnames(ret)[1] <- sprintf("%s x %s", xname, yname)

      # Create totals
      ret <- ret %>% mutate(total = rowSums(select_if(., is.numeric), na.rm = TRUE))
      if (pcol || prow) {
        ret <- ret %>% mutate_if(is.numeric, list(~ round(100 * . / sum(., na.rm = TRUE), decimals)))
      }
      if (pall) {
        numericals <- ret[, -c(1, ncol(ret))]
        all <- sum(numericals, na.rm = TRUE)
        ret <- ret %>% mutate_if(is.numeric, list(~ round(100 * . / all, decimals)))
      }
      ret <- arrange(ret, desc(.data$total))
      if (!total) ret <- ret %>% select(-.data$total)
      as_tibble(ret)
    }
  }
}

####################################################################
#' Generative Art: Sphere XmodY
#'
#' @family rtistry
#' @param eye,pal,var Parameters to change aesthetics and calculations
#' @return ggplot object
#' @export
rtistry_sphere <- function(eye = c(100, 0, 0),
                           pal = "auto",
                           var = 3) {
  try_require("threed")

  if (length(eye) != 3) stop("eye must be length 3")
  if (pal[1] == "auto") pal <- c("#264653", "#2a9d8f")
  if (length(pal) <= 1) stop("pal must be length >= 2")

  mesh3dobj$sphere %>%
    transform_by(invert_matrix(
      look_at_matrix(eye = eye, at = c(0, 0, 0))
    )) %>%
    perspective_projection() %>%
    as.data.frame() %>%
    mutate(aux = .data$element_id %% var == 1) %>%
    rowwise() %>%
    mutate(color = ifelse(.data$aux, pal[1], NA)) %>%
    # select(element_id, x, y, color) %>%
    ggplot(aes(.data$x, .data$y, group = .data$element_id)) +
    geom_polygon(aes(fill = .data$color), colour = "transparent") +
    # scale_size_manual(values = c('TRUE' = 0.2, 'FALSE' = 0.5)) +
    scale_fill_identity() +
    coord_equal() +
    theme_lares(clean = TRUE, legend = FALSE, background = pal[2])
}

# mesh3dobj$dodecahedron %>%
#   transform_by(invert_matrix(
#     look_at_matrix(eye = c(10,10,10), at = c(0, 0, 0))
#   )) %>%
#   perspective_projection() %>%
#   as.data.frame() %>%
#   mutate(aux = .data$element_id %% 3 == 1) %>%
#   rowwise() %>%
#   filter(aux) %>%
#   mutate(color = ifelse(.data$aux, sample(lares::lares_pal("col")[1:9])[1], NA)) %>%
#   ggplot(aes(group = .data$element_id)) +
#   geom_polygon(aes(-x, y, fill = rev(color)), color = "black", size = 0.1) +
#   coord_equal() + scale_fill_identity() +
#   theme_lares(clean = TRUE, legend = FALSE)

####################################################################
#' Google Sheets Reading and Writing (API v4)
#'
#' Read and write data from Google Sheets knowing the file's title.
#' You may use a single value from a cell or a data.frame from a cell range.
#'
#' @family Scrapper
#' @family Google
#' @param title Character. Title of Google Drive file. Uses regular
#' expressions so you may fetch with patterns instead of names.
#' @param sheet Character. Working sheet to import
#' @param range Character. A cell range to read from
#' @param drop_nas Boolean. Remove columns and rows that contain only NAs?
#' @param json Character. JSON filename with service auth
#' @param email,api_key Character. If you have multiple pre-authorized
#' accounts in your machine, you may non-interactively select
#' which one you wish to use by email and/or api_key.
#' @param server Boolean. Force interacting auth process?
#' @param ... Additional parameters passed to \code{read_sheet()}.
#' @return For reading, data.frame with the results of your Google Sheets file
#' based on its \code{title}, specificially the \code{sheet} and \code{range}
#' requested. For writing, no return value.
#' @aliases readGS4
#' @export
#' @rdname google_sheets
readGS <- function(title, sheet = "Hoja 1", range = NULL, drop_nas = TRUE,
                   json = NULL, email = NULL, api_key = NULL, server = FALSE, ...) {
  files <- filesGD(title = title, server = server, json = json, email = email)
  if (nrow(files) > 0) {
    message(sprintf("Using: %s (%s)", files$name[1], files$id[1]))
    df <- read_sheet(files$id[1], sheet = sheet, range = range, ...)
    if (drop_nas && isTRUE(ncol(df) > 0) && isTRUE(nrow(df) > 0)) {
      df <- df %>%
        removenacols() %>%
        removenarows()
    }
    if (length(df) > 0 && nrow(df) == 0) {
      df <- names(df)
    }
    df
  }
}

#' @param data Object (value, vector, data.frame, list).
#' @param reformat Boolean. Reformat the affected cells?
#' @param append Boolean.
#' @aliases writeGS4
#' @export
#' @rdname google_sheets
writeGS <- function(data, title, sheet = "Hoja 1", range = "A1",
                    reformat = FALSE, append = FALSE,
                    json = NULL, email = NULL, api_key = NULL, server = FALSE, ...) {
  files <- filesGD(title = title, server = server, json = json, email = email)

  if (nrow(files) > 0) {
    if (is.vector(data) && !is.list(data)) {
      data <- data.frame(data)
      col_names <- FALSE
    } else {
      col_names <- TRUE
    }

    if (nrow(files) == 0) {
      message("Google Sheet filename not found: created one for you!")
      gs4_create(title, sheets = data, col_names = col_names)
      invisible(NULL)
    } else {
      message(sprintf("Using: %s (%s)", files$name[1], files$id[1]))
      if (append) {
        invisible(sheet_append(files$id[1], sheet = sheet, data = data))
      } else {
        invisible(
          range_write(files$id[1],
            sheet = sheet, data = data,
            range = range, reformat = reformat,
            col_names = col_names, ...
          )
        )
      }
    }
  }
}


####################################################################
#' Google Drive Files (API v4)
#'
#' Authenticate and find Google Drive files and IDs by name.
#'
#' @family Scrapper
#' @family Google
#' @inheritParams readGS
#' @return Vector with found file names based on \code{title} on Google Drive.
#' @export
filesGD <- function(title, server = FALSE, json = NULL, api_key = NULL, email = NULL) {
  try_require("googledrive")
  try_require("googlesheets4")
  options("gargle_oob_default" = server)

  if (!is.null(json)) {
    if (file.exists(json)) {
      drive_auth(path = json)
      gs4_auth(path = json)
    } else {
      stop("No credentials found on ", json)
    }
  } else {
    if (!is.null(api_key)) {
      gs4_auth_configure(api_key = api_key)
    }
    if (!is.null(email)) {
      drive_auth(email = email)
      gs4_auth(email = email)
    }
  }

  try(local_drive_quiet())
  files <- drive_find(
    pattern = title,
    n_max = 199,
    type = "spreadsheet"
  )

  nfiles <- nrow(files)
  if (nfiles != 1) message(glued("{nfiles} files found with pattern: {title}"))
  files
}

####################################################################
#' Maze Solver, inspired by Micromouse competitions
#'
#' Modified recursive depth-first search (DFS) algorithm to solve mazes.
#' It explores the maze by recursively moving to adjacent cells until it finds a
#' path from the starting point to the destination. Contains options to
#' maximize paths by trying to turn less, allowing diagonal turns, prioritizing
#' turns that chooses next step pointing towards the end point, and a grid
#' search combining parameters to find best route.
#'
#' @family Games
#' @inheritParams corr_cross
#' @param maze Matrix. Using 0 for open space and 1 for walls.
#' @param start,end Integer vector, length 2. Start and end coordinates.
#' @param inertia Boolean. When enabled, algorithm will check for new
#' directions only when impossible to continue in a straight line.
#' @param aim Boolean. When enabled, algorithm will try first the directions
#' closer to the \code{end} point, ranked and sorted by shorter distances.
#' @param diagonal Boolean. When enabled, algorithm will have 8 degrees of
#' freedom to move, if not, only 4 (up, down, left, right).
#' @param random Boolean. When enabled, algorithm will pick next direction
#' randomly.
#' @param timeout Numeric. How many seconds set for timeout to force
#' algorithm to stop trying new paths?
#' @param quiet Boolean. Keep quiet? If not, print results
#' @param seed Numeric. Seed to replicate random results.
#' @return List with data.frame containing solved solution, data.frame with
#' path coordinates and directions, steps counter and turns counter.
#' @examples
#' micromouse <- matrix(c(
#'   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
#'   1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
#'   1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1,
#'   1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
#'   1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1,
#'   1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1,
#'   1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1,
#'   1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1,
#'   1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1,
#'   1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1,
#'   1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
#'   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
#' ), nrow = 12, byrow = TRUE)
#' maze_solve(micromouse, start = c(2, 2), end = c(7, 7))
#' @export
maze_solve <- function(
    maze, start = c(1, 1), end = dim(maze),
    inertia = FALSE, aim = TRUE, diagonal = TRUE, random = FALSE,
    timeout = 4,
    quiet = FALSE,
    seed = NULL,
    ...) {
  # Inputs validation
  if (!is.matrix(maze)) {
    stop("Input 'maze' must be a matrix or a data.frame")
  }
  if (!all(start %in% which(maze == 0, arr.ind = TRUE)) ||
    !all(end %in% which(maze == 0, arr.ind = TRUE)) ||
    any(start < 1) || any(end > dim(maze))) {
    stop("Invalid start or end point")
  }
  if (maze[start[1], start[1]] != 0) stop("Starting point must be a 0")
  if (maze[end[1], end[1]] != 0) stop("Ending point must be a 0")
  if (is.null(seed)) seed <- round(1000 * runif(1))
  set.seed(seed)
  if (!random) seed <- NA

  # Initialize data frame to store path coordinates
  tic("maze_solve_timeout")
  result <- maze_solve_recursive(maze, start, end, aim, inertia, diagonal, random, timeout)

  # Process results
  if (!isFALSE(result)) {
    colnames(result$path_coords) <- c("row", "col")
    result$path_coords <- rbind(result$path_coords, end)
    result$path_coords$direction <- c(sapply(1:(nrow(result$path_coords) - 1), function(i) {
      calculate_direction(
        result$path_coords$row[i + 1] - result$path_coords$row[i],
        result$path_coords$col[i + 1] - result$path_coords$col[i]
      )
    }), "X")

    colnames(result$maze) <- seq_len(ncol(result$maze))
    result$maze <- as.data.frame(result$maze)
    result$maze[result$maze == 0] <- " "
    result$maze[result$maze == 1] <- "[]"
    for (x in seq_along(result$path_coords$direction)) {
      result$maze[result$path_coords$row[x], result$path_coords$col[x]] <- result$path_coords$direction[x]
    }
    result$steps_counter <- nrow(result$path_coords)
    result$turns_counter <- count_direction_changes(result$path_coords) + 1
  } else {
    result <- list(maze = NULL)
  }
  result$start <- paste(start, collapse = ",")
  result$end <- paste(end, collapse = ",")
  result$inertia <- inertia
  result$aim <- aim
  result$random <- random
  result$seed <- seed

  class(result) <- c("maze_solve", class(result))
  if (!quiet) print(result)
  invisible(result)
}

#' @rdname maze_solve
#' @param x maze_solve object
#' @export
print.maze_solve <- function(x, ...) {
  cat(sprintf(
    "Setup: Inertia (%s) | Aim (%s) | Random (%s)%s\n",
    x$inertia, x$aim, x$random,
    ifelse(x$random, sprintf(" | Seed (%s)", x$seed), "")
  ))
  if (!is.null(x$maze)) {
    if (isTRUE(x$coords_inv)) {
      cat("  [Inverted start and end points]\n")
    }
    cat("  Total steps: ", x$steps_counter, "\n")
    cat("  Total turns: ", x$turns_counter, "\n\n")
    print(x$maze)
  } else {
    cat("No solution found for this maze\n")
  }
}

# Solve Maze recursive function with diagonal movement
maze_solve_recursive <- function(
    maze, current, end,
    aim = TRUE,
    inertia = TRUE,
    diagonal = FALSE,
    random = FALSE,
    timeout = 10,
    path_coords = data.frame(row = integer(0), col = integer(0)),
    prev_direction = NULL) {
  # Check for timeout or goal reached
  toci <- toc("maze_solve_timeout", quiet = TRUE)
  timeout_reached <- any(toci$toc - toci$tic > timeout)

  if (timeout_reached) {
    FALSE
  } else if (identical(current, end)) {
    list(maze = maze, path_coords = path_coords)
  } else {
    row <- current[1]
    col <- current[2]
    maze[row, col] <- "X"
    path_coords <- rbind(path_coords, c(row, col))

    target <- if (aim) end else c(row, col)
    positions <- rank_positions(row, col, target[1], target[2], diagonal)
    if (random) positions <- positions[sample(nrow(positions)), ]

    if (!is.null(prev_direction) && isTRUE(inertia)) {
      first_point <- linear_extrapolation(prev_direction[1], prev_direction[2], row, col)
      if (all(!is.na(first_point))) {
        skip <- which(positions$x == first_point[1] & positions$y == first_point[2])
        if (length(skip) > 0) {
          positions <- rbind(first_point, positions[-skip, -3])
        }
      }
    }

    results <- lapply(seq_len(nrow(positions)), function(i) {
      next_row <- positions$x[i]
      next_col <- positions$y[i]

      if (next_row > 0 && next_row <= nrow(maze) &&
        next_col > 0 && next_col <= ncol(maze) &&
        maze[next_row, next_col] == 0) {
        maze_solve_recursive(
          maze, c(next_row, next_col), end,
          aim, inertia, diagonal, random,
          timeout, path_coords,
          prev_direction = c(row, col)
        )
      } else {
        FALSE
      }
    })
    match_found <- Filter(function(res) !is.logical(res), results)
    if (length(match_found) > 0) match_found[[1]] else FALSE
  }
}

#' @rdname maze_solve
#' @export
maze_gridsearch <- function(
    maze,
    start = c(2, 2),
    end = round(dim(maze) / 2),
    quiet = TRUE,
    seed = 123, ...) {
  results <- list()
  for (a in c(TRUE, FALSE)) {
    for (b in c(TRUE, FALSE)) {
      for (c in c(TRUE, FALSE)) {
        for (d in c(TRUE, FALSE)) {
          this <- maze_solve(
            maze,
            start = start, end = end,
            inertia = a, diagonal = b, random = c, aim = d,
            quiet = quiet, seed = seed, ...
          )
          if (!is.logical(this)) {
            this$coords_inv <- FALSE
            results <- append(results, list(this))
          }
          this <- maze_solve(
            maze,
            start = end, end = start,
            inertia = a, diagonal = b, random = c, aim = d,
            quiet = quiet, seed = seed, ...
          )
          if (!is.logical(this)) {
            this$coords_inv <- TRUE
            results <- append(results, list(this))
          }
        }
      }
    }
  }
  counters <- dplyr::bind_rows(lapply(results, function(y) {
    y[unlist(lapply(y, function(x) !is.data.frame(x) && !is.null(x)))]
  }))
  counters <- data.frame(id = seq_along(results), counters) %>%
    arrange(.data$steps_counter, .data$turns_counter) %>%
    select(.data$id, contains("counter"), everything(), .data$start, .data$end) %>%
    data.frame()
  list(solutions = results, results = counters)
}

# Function to calculate the number of direction changes
count_direction_changes <- function(path_coords) {
  num_changes <- 0
  if (!is.null(path_coords)) {
    # Calculate the direction for each step
    directions <- atan2(diff(path_coords$row), diff(path_coords$col)) * (180 / pi)
    # Adjust angles to be between 0 and 360 degrees
    directions <- (directions + 360) %% 360
    # Calculate the absolute difference between consecutive directions
    direction_diff <- abs(diff(directions))
    # Count the number of direction changes greater than a threshold (e.g., 45 degrees)
    num_changes <- sum(direction_diff > 30)
  }
  num_changes
}

# Function to calculate the direction of point in path
calculate_direction <- function(row, col) {
  if (is.na(row) || is.na(col) || length(row) == 0 || length(col) == 0) {
    "\u003F" # ?
  }
  if (row == 0 && col == 1) {
    "\u2192" # Right
  } else if (row == 0 && col == -1) {
    "\u2190" # Left
  } else if (row == 1 && col == 0) {
    "\u2193" # Down
  } else if (row == -1 && col == 0) {
    "\u2191" # Up
  } else if (row == 1 && col == 1) {
    "\u2198" # Diagonal down-right
  } else if (row == -1 && col == 1) {
    "\u2197" # Diagonal up-right
  } else if (row == 1 && col == -1) {
    "\u2199" # Diagonal down-left
  } else if (row == -1 && col == -1) {
    "\u2196" # Diagonal up-left
  } else if (row == 0 && col == 0) {
    "\u2022" # Center (no movement)
  } else {
    "Invalid direction"
  }
}

# Order of directions based on minimum distance
rank_positions <- function(x1, y1, x2, y2, diagonal = TRUE) {
  if (diagonal) {
    positions <- data.frame(
      x = c(x1 - 1, x1, x1 + 1, x1 - 1, x1 + 1, x1 - 1, x1, x1 + 1),
      y = c(y1 + 1, y1 + 1, y1 + 1, y1, y1, y1 - 1, y1 - 1, y1 - 1)
    )
  } else {
    positions <- data.frame(
      x = c(x1, x1 + 1, x1, x1 - 1),
      y = c(y1 + 1, y1, y1 - 1, y1)
    )
  }
  positions$d <- sqrt((positions$x - x2)^2 + (positions$y - y2)^2)
  ranked_positions <- positions[order(positions$d), ]
  ranked_positions
}

# Function to perform linear extrapolation and return the next integer point
linear_extrapolation <- function(x1, y1, x2, y2) {
  # Calculate the slope
  slope <- (y2 - y1) / (x2 - x1)
  # Extrapolate to the next integer x value
  next_x <- x2 + 1 # Assuming you want to extrapolate to the next integer x value
  # Calculate the corresponding y value
  next_y <- y2 + slope * (next_x - x2)
  c(round(next_x), round((next_y)))
}

# micromouse <- matrix(c(
#   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
#   1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
#   1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1,
#   1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
#   1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1,
#   1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1,
#   1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1,
#   1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1,
#   1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1,
#   1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1,
#   1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
#   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
# ), nrow = 12, byrow = TRUE)

# temp1 <- maze_gridsearch(micromouse, end = c(7, 7))
# counters <- data.frame(
#   steps_counter = unlist(lapply(temp1, function(x) x$steps_counter)),
#   turns_counter = unlist(lapply(temp1, function(x) x$turns_counter)))
# counters[order(counters$steps_counter), ]
# temp1[[6]]

---
title: "Data Wrangling & Visualization"
author: "Bernardo Lares"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Wrangling & Visualization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  warning = FALSE,
  message = FALSE
)
library(dplyr)
```

## Install and Load

Install `lares` from CRAN or get the development version from GitHub. Then, load the package:

```{r}
library(lares)
```

## Quick Start with Built-in Data

We'll use the Titanic dataset included in `lares`:

```{r}
data(dft)
head(dft, 3)
```

## Frequency Analysis

### Basic Frequencies

The `freqs()` function provides quick frequency tables with percentages and cumulative values:

```{r}
# How many survived?
freqs(dft, Survived)
```

### Multi-variable Frequencies

```{r}
# Survival by passenger class
freqs(dft, Pclass, Survived)
```

### Visual Frequencies

```{r fig.width=7, fig.height=4}
# Visualize survival by class
freqs(dft, Pclass, Survived, plot = TRUE)
```

### Dataframe-wide Frequencies

Analyze all variables at once:

```{r fig.width=7, fig.height=5}
freqs_df(dft, plot = TRUE, top = 10)
```

## Correlation Analysis

### Correlation Matrix

Get correlations between all variables (automatically handles categorical variables):

```{r}
# Correlation matrix of numeric variables
cors <- corr(dft[, 2:5], method = "pearson")
head(cors, 3)
```

### Correlate One Variable with All Others

```{r fig.width=7, fig.height=5}
# Which variables correlate most with Survival?
corr_var(dft, Survived, top = 10)
```

### Cross-Correlations

Find the strongest correlations across the entire dataset:

```{r fig.width=7, fig.height=5}
# Top cross-correlations
corr_cross(dft[, 2:6], top = 8)
```

## Data Transformation

### Categorical Reduction

Reduce categories in high-cardinality variables:

```{r}
# Reduce ticket categories (keep top 5, group rest as "other")
dft_reduced <- categ_reducer(dft, Ticket, top = 5)
freqs(dft_reduced, Ticket, top = 10)
```

### Normalization

Normalize numeric variables to [0, 1] range:

```{r}
# Normalize age
dft$Age_norm <- normalize(dft$Age)
head(dft[, c("Age", "Age_norm")], 5)
```

### One-Hot Encoding

Convert categorical variables to binary columns:

```{r}
# One-hot encode passenger class
dft_encoded <- ohse(dft[, c("Pclass", "Survived")], limit = 5)
colnames(dft_encoded)
```

## Date Manipulation

Create date features for time series analysis:

```{r}
# Create sample dates
dates <- seq(as.Date("2024-01-01"), as.Date("2024-12-31"), by = "day")

# Extract year-month
ym <- year_month(dates[1:5])
ym

# Extract year-quarter
yq <- year_quarter(dates[1:5])
yq

# Cut dates into quarters
quarters <- date_cuts(dates[c(1, 100, 200, 300)], type = "Q")
quarters
```

## Visualization with theme_lares

### Custom ggplot2 Theme

`lares` includes a clean, professional theme:

```{r fig.width=7, fig.height=4}
library(ggplot2)

ggplot(dft, aes(x = Age, y = Fare * 1000, color = Survived)) +
  geom_point(alpha = 0.6) +
  labs(title = "Age vs Fare by Survival") +
  # Customize theme with several available options
  theme_lares(legend = "top", grid = "Yy", pal = 2, background = "#f2f2f2") +
  # Customize axis scales to look nicer
  scale_y_abbr()
```

### Distribution Plots

Visualize distributions quickly:

```{r fig.width=7, fig.height=5}
# Analyze Fare distribution
distr(dft, Fare, breaks = 20)
```

### Number Formatting

Format numbers for better readability:

```{r}
# Format large numbers
formatNum(c(1234567, 987654.321), decimals = 2)

# Abbreviate numbers
num_abbr(c(1500, 2500000, 1.5e9))

# Convert abbreviations back to numbers
num_abbr(c("1.5K", "2.5M", "1.5B"), numeric = TRUE)
```

### Custom Scales

Use lares scales for better axis formatting:

```{r fig.width=7, fig.height=4}
df_summary <- dft %>%
  group_by(Pclass) %>%
  summarize(avg_fare = mean(Fare, na.rm = TRUE), .groups = "drop")

ggplot(df_summary, aes(x = factor(Pclass), y = avg_fare)) +
  geom_col(fill = "#00B1DA") +
  labs(title = "Average Fare by Class", x = "Class", y = NULL) +
  scale_y_dollar() + # Format as currency
  theme_lares()
```

## Text and Vector Utilities

### Vector to Text

Convert vectors to readable text:

```{r}
# Simple comma-separated
vector2text(c("apple", "banana", "cherry"))

# With "and" before last item
vector2text(c("red", "green", "blue"), and = "and")

# Shorter alias
v2t(LETTERS[1:5])
```

## Putting It All Together

Here's a complete analysis workflow:

```{r fig.width=7, fig.height=5}
library(dplyr)

# 1. Load and prepare data
data(dft)

# 2. Clean and transform
dft_clean <- dft %>%
  mutate(Age_Group = cut(Age,
    breaks = c(0, 18, 35, 60, 100),
    labels = c("Child", "Young", "Adult", "Senior")
  ))

# 3. Analyze frequencies
freqs(dft_clean, Age_Group, Survived, plot = TRUE)

# 4. Check correlations
corr_var(dft_clean, Survived_TRUE, top = 8, max_pvalue = 0.05)
```

## Further Reading

### Package Resources
- **Package documentation:** [https://laresbernardo.github.io/lares/](https://laresbernardo.github.io/lares/)
- **GitHub repository:** [https://github.com/laresbernardo/lares](https://github.com/laresbernardo/lares)
- **Report issues:** [https://github.com/laresbernardo/lares/issues](https://github.com/laresbernardo/lares/issues)

### Blog Posts & Tutorials
- **Find Insights with Ranked Cross-Correlations:** [DataScience+](https://laresbernardo.github.io/lares/reference/corr_cross.html)
- **Visualize Monthly Income Distribution and Spend Curve:** [DataScience+](https://laresbernardo.github.io/lares/reference/distr.html)
- **All lares articles:** [Author page on DataScience+](https://laresbernardo.github.io/lares/articles/)

## Next Steps

- Explore machine learning with `h2o_automl()` (see Machine Learning vignette)
- Learn about API integrations with ChatGPT and Gemini (see API Integrations vignette)
- Check individual function documentation: `?freqs`, `?corr`, `?theme_lares`

---
title: "API Integrations"
author: "Bernardo Lares"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{API Integrations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  warning = FALSE,
  message = FALSE,
  eval = FALSE # API examples require API keys
)
```

## Introduction

`lares` provides convenient wrappers for popular APIs, making it easy to integrate AI services, financial data, and more into your R workflows.

## Credential Management

### Setup Credentials Securely

`lares` uses a YAML configuration file to store credentials:

```{r eval=TRUE}
library(lares)
```

Create a `config.yml` file:

```yaml
default:
  openai:
    secret_key: "sk-your-openai-key-here"
  
  gemini:
    api_key: "your-gemini-key-here"
  
  database:
    server: "localhost"
    database: "mydb"
    uid: "user"
    pwd: "password"
```

Set the credentials directory (one-time setup):

```{r}
# lares will prompt you to set the directory
creds <- get_credentials("openai")

# Or specify manually
creds <- get_credentials("openai", dir = "~/my_credentials")
```

## ChatGPT Integration

### Basic Usage

```{r}
# Simple question
response <- gpt_ask("Explain linear regression in 2 sentences")
```

### Structured Prompts

Use `gpt_prompter()` to build better prompts:

```{r}
# Create a structured prompt
prompt <- gpt_prompter(
  instruction = "Classify these items",
  input = c("Dog", "Rose", "Car", "Oak"),
  context = c("Animal", "Plant", "Object"),
  output = "table",
  cols = c("Item", "Category")
)

# Send to ChatGPT
result <- gpt_ask(prompt$prompt)
```

### Specialized Functions

#### Classification

```{r}
# Classify text into categories
items <- c("Python tutorial", "R package", "Java course", "Statistics book")
categories <- c("Programming", "Statistics", "Other")

result <- gpt_classify(items, categories)
print(result)
```

#### Data Extraction

```{r}
# Extract specific information
texts <- c(
  "My email is john@example.com",
  "Call me at +1-555-0123",
  "Visit us at 123 Main St, NYC"
)

extractions <- c("email", "phone number", "city")

result <- gpt_extract(texts, extractions)
print(result)
```

#### Translation

```{r}
# Translate text
text <- rep("Hello, how are you?", 3)
languages <- c("Spanish", "French", "German")

result <- gpt_translate(text, languages)
print(result)
```

#### Formatting

```{r}
# Standardize date formats
dates <- c("March 15, 2024", "15/03/2024", "2024-03-15")
format_spec <- "ISO 8601 (YYYY-MM-DD)"

result <- gpt_format(dates, format = format_spec)
print(result)
```

#### Tagging

```{r}
# Tag items with multiple labels
items <- c("Machine learning tutorial", "Data visualization guide")
tags <- c("AI", "Statistics", "Programming", "Visualization")

result <- gpt_tag(items, tags)
print(result)
```

### View Conversation History

```{r}
# See all previous prompts and responses from this session
history <- gpt_history()
head(history)
```

### Advanced Configuration

```{r}
# Customize model and parameters
response <- gpt_ask(
  "Write a haiku about data science",
  model = "gpt-4",
  temperature = 0.9, # More creative (0-2)
  num_retries = 3
)
```

## Google Gemini Integration

### Basic Usage

```{r}
# Text generation
response <- gemini_ask(
  "Explain the Central Limit Theorem",
  api_key = get_credentials("gemini")$api_key
)
```

### Image Analysis

```{r}
# Analyze an image
response <- gemini_image(
  prompt = "Describe this plot in detail",
  image_path = "path/to/plot.png",
  api_key = get_credentials("gemini")$api_key
)
```

## Financial Data: Stocks

### Get Stock Historical Data

```{r eval=FALSE}
# Fetch Apple stock data
aapl <- stocks_hist(
  symbols = "AAPL",
  from = Sys.Date() - 90, # Last 90 days
  to = Sys.Date()
)

head(aapl, 3)
```

### Multiple Stocks

```{r eval=FALSE}
# Compare multiple stocks
tech_stocks <- stocks_hist(
  symbols = c("AAPL", "GOOGL", "MSFT"),
  from = "2024-01-01"
)

head(tech_stocks, 3)
```

### Stock Quotes

```{r eval=FALSE}
# Real-time quotes
quote <- stocks_quote("AAPL")
print(quote)
```

### Portfolio Analysis

```{r}
# Track a portfolio
portfolio <- daily_portfolio(
  symbols = c("AAPL", "GOOGL", "TSLA"),
  shares = c(10, 5, 8),
  from = "2024-01-01"
)

# Visualize performance
splot_summary(portfolio)
```

### Stock Visualizations

```{r}
# Growth over time
splot_growth(tech_stocks, symbols = c("AAPL", "GOOGL"))

# Returns on investment
splot_roi(tech_stocks, investment = 10000)

# Sector performance
splot_types(tech_stocks)
```

## Google Sheets Integration

### Read from Google Sheets

```{r}
# Read a Google Sheet
data <- readGS(
  title = "My Spreadsheet",
  sheet = "Sheet1",
  creds = get_credentials("google")
)
```

### Write to Google Sheets

```{r}
# Write data to Google Sheets
writeGS(
  data = mtcars,
  title = "My Data",
  sheet = "Cars",
  creds = get_credentials("google")
)
```

## Database Queries

### PostgreSQL

```{r}
# Query a database
query <- "SELECT * FROM users WHERE active = TRUE LIMIT 10"
result <- queryDB(
  query = query,
  from = "my_database_creds"
)
```

## Google Trends

### Search Trends Over Time

```{r}
# Get trend data
trends <- gtrends_time(
  keyword = "machine learning",
  from = "2023-01-01",
  to = "2024-01-01"
)

head(trends)
```

### Related Queries

```{r}
# Find related search terms
related <- gtrends_related(
  keyword = "data science",
  categories = c("top", "rising")
)
```

## Best Practices

### 1. Secure Your Credentials

```{r eval=FALSE}
# ❌ DON'T hardcode
# api_key <- "sk-1234567890"

# ✅ DO use credential management
api_key <- get_credentials("openai")$secret_key
```

### 2. Handle API Errors

```{r}
# Wrap in try-catch
result <- tryCatch(
  {
    gpt_ask("Your question here")
  },
  error = function(e) {
    message("API error: ", e$message)
    NULL
  }
)
```

### 3. Rate Limiting

```{r}
# Add delays for bulk operations
items <- c("item1", "item2", "item3")

results <- lapply(items, function(item) {
  result <- gpt_ask(item)
  Sys.sleep(1) # 1 second delay
  result
})
```

### 4. Cache Results

```{r eval=FALSE}
# Cache expensive API calls
result <- cache_pipe(
  {
    # This only runs once, then cached and loaded on subsequent calls
    gpt_ask("Expensive query here")
  },
  base = "my_gpt_query"
)
```

### 5. Monitor Costs

```{r}
# Use cheaper models for simple tasks
simple_task <- gpt_ask(
  "Summarize: The meeting is at 3pm",
  model = "gpt-3.5-turbo" # Cheaper than gpt-4
)

# Use expensive models only when needed
complex_task <- gpt_ask(
  "Analyze this complex dataset: ...",
  model = "gpt-4"
)
```

## Environment Variables

Set global defaults via environment variables:

```{r eval=TRUE}
# In .Renviron file:
# LARES_GPT_MODEL=gpt-4
# LARES_GPT_URL=https://api.openai.com/v1/chat/completions
# LARES_GEMINI_API=https://generativelanguage.googleapis.com/v1beta/models/

# Check current settings
Sys.getenv(c("LARES_GPT_MODEL", "LARES_GEMINI_API"))
```

## Troubleshooting

### API Key Issues

```{r}
# Verify credentials are loaded
creds <- get_credentials("openai")
if (is.null(creds$secret_key)) {
  stop("OpenAI API key not found!")
}
```

### Network Issues

```{r eval=TRUE}
# Check internet connection
if (haveInternet()) {
  message("Connected to internet")
} else {
  stop("No internet connection")
}
```

### Rate Limit Errors

If you hit rate limits:
1. Add `Sys.sleep()` between calls
2. Use batch processing
3. Upgrade your API plan

## Complete Example: Data Analysis with AI

```{r}
# 1. Load data
data(dft)

# 2. Get AI summary of data structure
prompt <- sprintf(
  "Summarize this dataset structure: %d rows, columns: %s",
  nrow(dft),
  paste(colnames(dft), collapse = ", ")
)
summary <- gpt_ask(prompt)

# 3. Get stock data
stocks <- stocks_hist("AAPL", from = Sys.Date() - 30)

# 4. Cache the analysis
analysis <- cache_pipe(
  {
    gpt_ask(sprintf(
      "Analyze this stock: Recent high: $%.2f, Low: $%.2f",
      max(stocks$High, na.rm = TRUE),
      min(stocks$Low, na.rm = TRUE)
    ))
  },
  base = "aapl_analysis"
)

print(analysis)
```

## Further Reading

### API Documentation
- **OpenAI API Docs:** [https://platform.openai.com/docs](https://platform.openai.com/docs)
- **Gemini API Docs:** [https://ai.google.dev/docs](https://ai.google.dev/docs)
- **lares Functions:** `?gpt_ask`, `?gemini_ask`, `?stocks_hist`, `?get_credentials`

### Blog Posts & Tutorials  
- **All lares articles:** [Author page on R-bloggers](https://www.r-bloggers.com/author/bernardo-lares/)
- **Package website:** [https://laresbernardo.github.io/lares/](https://laresbernardo.github.io/lares/)

## Next Steps

- Explore data wrangling (see Data Wrangling vignette)
- Learn machine learning (see Machine Learning vignette)
- Review credential management: `?get_credentials`

---
title: "Robyn Enhancers"
author: "Bernardo Lares"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Robyn Enhancers}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE,
    cache = TRUE
)
library(lares)
library(dplyr)
library(ggplot2)
```

## Introduction

The `lares` package provides a suite of auxiliary functions to enhance your experience with [Robyn](https://github.com/facebookexperimental/Robyn), a Marketing Mix Modeling (MMM) package developed by Meta's Marketing Science Team. These functions help with hyperparameter building, model selection, performance analysis, and cross-model budget allocation.

## Hyperparameter Builder

The function `robyn_hypsbuilder()` automates the creation of hyperparameter lists for Robyn. It's particularly useful when you have many media channels and want to set reasonable default ranges.

```{r eval=TRUE}
# Define your input variables
paid_media_spends <- c("tv_spend", "digital_spend")
organic_vars <- c("newsletter")

# Generate hyperparameters
hyperparameters <- robyn_hypsbuilder(
    channels = c(paid_media_spends, organic_vars),
    media_type = "default",
    adstock = "weibull_pdf", # or "weibull_cdf", "weibull_pdf"
    date_type = "weekly",
    lagged = FALSE
)

# The output is a list ready to be passed to robyn_inputs()
print(hyperparameters)
```

## Model Selection

The function `robyn_modelselector()` provides a visual and metric-based approach to select the best models from a Robyn run. It ranks models based on their performance metrics and helps you identify the best candidates to evaluate. For a deeper dive into this topic, check out this [Medium post](https://medium.com/@laresbernardo/select-the-right-mmm-candidate-based-on-your-specific-criteria-and-business-knowledge-1f583c3cb97a).

```{r eval=FALSE}
# Assuming you have a Robyn OutputCollect object
# OutputCollect <- robyn_run(...)

# Select best models
best_models <- robyn_modelselector(
    InputCollect = InputCollect,
    OutputCollect = OutputCollect,
    metrics = c(
        "rsq_train", "performance", "potential_improvement", "non_zeroes",
        "incluster_models", "cluster_sd", "certainty", "baseline_dist"
    ),
    wt = c(2, 0.1, 0, 1, 0.1, 0, 1.5, 0)
)

# View the results
print(best_models)

# Plot the clusters
plot(best_models)
```

## Model Performance

The function `robyn_performance()` calculates detailed performance metrics for a specific model or a set of models. It breaks down contribution, ROAS or CPA for each channel, etc.

```{r eval=FALSE}
# Get performance for a specific model and date range
performance <- robyn_performance(
    InputCollect = InputCollect,
    OutputCollect = OutputCollect,
    select_model = "1_123_4",
    # Calculate for specific date range
    start_date = NULL, end_date = NULL,
    non_promo = TRUE, # Add non-promo decomposition as well
    marginals = TRUE, # Add mROAS or mCPA for spend levels
    carryovers = FALSE # Add mean percentage of carryover
)

# View performance table
print(performance)

# Plot performance
plot(performance)
```

## Marginal Performance (Experimental)

The function `robyn_marginal()` allows you to explore the marginal returns of your media channels (mROAS or mCPA).

```{r eval=FALSE}
# Calculate marginal response
marginal <- robyn_marginal(
    InputCollect = InputCollect,
    OutputCollect = OutputCollect,
    select_model = "1_123_4",
    metric_name = "tv_spend",
    metric_value = 100000
)
```

## Cross-Model Budget Allocation (Experimental)

If you have multiple Robyn models (e.g., for different countries or brands), `robyn_xmodels()` and `robyn_xchannels()` help you allocate budget across all of them simultaneously.

```{r eval=FALSE}
# Load multiple models
model_paths <- c("model_us.json", "model_uk.json", "model_de.json")
models <- lapply(model_paths, Robyn::robyn_recreate)
names(models) <- c("US", "UK", "DE")

# Allocate budget across models
cross_allocation <- robyn_xmodels(
    models = models,
    total_budget = 5000000,
    scenario = "max_response"
)

# View results
print(cross_allocation)
```

## Conclusion

These functions streamline the post-modeling analysis in Robyn, making it easier to interpret results, select the best models, and make data-driven budget allocation decisions.

## Additional Reading

For more information on Robyn and MMM, check out the following resources:

-   [Robyn GitHub Repository](https://github.com/facebookexperimental/Robyn)
-   [Robyn: Analyst's Guide to MMM](https://facebookexperimental.github.io/Robyn/docs/analysts-guide-to-MMM/)
-   [Select the right MMM candidate based on your specific criteria and business knowledge](https://medium.com/@laresbernardo/select-the-right-mmm-candidate-based-on-your-specific-criteria-and-business-knowledge-1f583c3cb97a)
-   [Stop building MMMs without business-backed hypotheses](https://medium.com/@laresbernardo/stop-building-mmms-without-business-backed-hypotheses-0e8a9fa9a375)
-   [Hitting ROAS Target Using Robyn’s Budget Allocator](https://medium.com/@gufengzhou/hitting-roas-target-using-robyns-budget-allocator-274ace3add4f)
-   [The Convergence of Marginal ROAS in the Budget Allocation in Robyn](https://medium.com/@gufengzhou/the-convergence-of-marginal-roas-in-the-budget-allocation-in-robyn-5d407aebf021)

---
title: "Portfolio Performance Report"
output: html_document
date: "`r format(Sys.time(), '%d %B, %Y')`"
params:
  summary_df1: NA
  summary_df2: NA
  plots_fixed: NA
  plots_relative: NA
---

<style type="text/css">
h1.title { font-size: 28px; color: DarkOrange;}
h1 { font-size: 24px;}
h2 { font-size: 20px;}
h3 { font-size: 18px;}
body { font-size: 16px;}
</style>

```{r echo = FALSE, message = FALSE}
# Some global stuff
suppressPackageStartupMessages(library(dplyr))
suppressPackageStartupMessages(library(ggplot2))
suppressPackageStartupMessages(library(knitr))
#params <- readRDS("params.RDS")
```

The following Portfolio Report was automatically generated by the *lares* library:

## All time performance

```{r fig.width = 8, fig.height = 5, echo = FALSE}
ggplot2:::plot.ggplot(params$plots_fixed$`Positions`)
```

## Dynamic time periods report {.tabset}

```{r, results = 'asis', fig.width = 10, fig.height = 5, echo = FALSE}
ts <- gsub("plots_", "", names(params$plots_relative))
plot_names <- names(params$plots_relative[[1]])
for (p in seq_along(plot_names)){
  cat(paste0("### ", plot_names[p], "{.tabset}\n"))
  for (i in seq_along(ts)){
      cat("####", ts[i], '<br>', '\n')
      ggplot2:::plot.ggplot(params$plots_relative[[i]][[p]])
      cat('\n', '<br>', '\n\n')
  }
  cat('\n', '<br>', '\n\n')
}
```

## Tabulated data {.tabset}

```{r, results = 'asis', fig.width = 8, fig.height = 5, echo = FALSE}
tables <- list(
  "Stocks Summary" = kable(params$summary_df1, align = "lrcrrrr", row.names = TRUE, digits = c(0, 2, 0, 2, 0, 0, 1), format.args = list(big.mark = ",")),
  "Portfolio Summary" = kable(params$summary_df2, align = "lcc", row.names = FALSE))
for (i in seq_along(tables)){
  cat("###", names(tables)[i], '<br>', '\n')
  print(tables[[i]])
  cat('\n', '<br>', '\n\n')
}
```

## Overall portfolio information {.tabset}

```{r, results = 'asis', fig.width = 8, fig.height = 5, echo = FALSE}
other <- names(params$plots_fixed)
# Exclude summary which has already been plotted
for (i in seq_along(other)[-1]){
  p <- params$plots_fixed[[i]]
  if (!is.null(p)) {
    cat("###", other[i], '<br>', '\n')
    ggplot2:::plot.ggplot(p)
    cat('\n', '<br>', '\n\n')
  }
}
```

> Report generated timestamp: `r Sys.time()`

---
title: "Games & Puzzles"  
author: "Bernardo Lares"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Games & Puzzles}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  warning = FALSE,
  message = FALSE
)

# Enable ANSI color output in HTML vignettes
Sys.setenv(CLICOLOR_FORCE = "1")
options(crayon.enabled = TRUE)
library(fansi)
old_hooks <- fansi::set_knit_hooks(knitr::knit_hooks,
  which = c("output", "message", "error")
)
```

## Introduction

Beyond data analysis and machine learning, `lares` includes a collection of fun game solvers and helpers! Challenge yourself with Wordle, Scrabble, Sudoku, mazes...

```{r}
library(lares)
```

## Wordle

Play or solve Wordle puzzles with validation, hints, and simulations.

### Basic Wordle Validation

```{r, eval=TRUE}
# Check your guess against a word
wordle_check("OPENS", "ABBEY")
wordle_check("BABES", "ABBEY")
wordle_check("ABBEY", "ABBEY")
```

### Get Wordle Hints

Use `scrabble_words()` to find possible words:

```{r}
# After OPENS: O is not in word, P/E/N/S are not in positions 2/3/4/5
hints <- scrabble_words(
  tiles = "abcdefghijklmrtuvwxyz", # Available letters
  exclude_here = list("2" = "p", "3" = "e", "4" = "n", "5" = "s"),
  force_exclude = c("o"),
  force_n = 5, # 5-letter words
  language = "en"
)
head(hints, 10)
```

### Wordle Simulation

Simulate solving a Wordle puzzle:

```{r}
# Simulate solving with different starting words
simulation <- wordle_simulation(
  input = "SAINT",
  word = "ABBEY",
  seed = 123
)
print(simulation)
```

### Wordle Dictionary

Access word lists for different languages:

```{r eval=FALSE}
# Get English 5-letter words
en_words <- wordle_dictionary("en")
head(en_words, 20)

# Spanish words
es_words <- wordle_dictionary("es")
```

## Scrabble

Maximize your Scrabble score with word finders and calculators!

### Find Highest-Scoring Words

```{r}
# Find best words from your tiles
scrabble_words(
  tiles = "aeiourtn",
  force_max = 8, # Max 8 letters
  language = "en",
  scores = "en"
)
```

### With Board Constraints

```{r}
# Must contain specific letters
scrabble_words(
  tiles = "bernardo",
  force_str = "arn",
  force_max = 7,
  language = "en"
)
```

### Calculate Word Scores

```{r}
# Get point values for each letter
en_scores <- scrabble_points("en")
print(en_scores)

# Calculate scores for words
words <- c("QUEEN", "QUIZ", "HELLO")
scrabble_score(words, en_scores)
```

### Multi-Language Support

```{r eval=FALSE}
# Spanish Scrabble
scrabble_words(
  tiles = "españa",
  language = "es",
  scores = "es"
)

# French Scrabble
scrabble_words(
  tiles = "bonjour",
  language = "fr",
  scores = "fr"
)
```

## Sudoku

Solve Sudoku puzzles automatically!

### Simple Sudoku

```{r}
# Easy puzzle (0 represents empty cells)
trivial <- matrix(c(
  0, 9, 0, 7, 0, 0, 8, 6, 0,
  0, 3, 1, 0, 0, 5, 0, 2, 0,
  8, 0, 6, 0, 0, 0, 0, 0, 0,
  0, 0, 7, 0, 5, 0, 0, 0, 6,
  0, 0, 0, 3, 0, 7, 0, 0, 0,
  5, 0, 0, 0, 1, 0, 7, 0, 0,
  0, 0, 0, 0, 0, 0, 1, 0, 9,
  0, 2, 0, 6, 0, 0, 3, 5, 0,
  0, 5, 4, 0, 0, 8, 0, 7, 0
), nrow = 9, byrow = TRUE)

solution <- sudoku_solver(trivial)
print(solution)
```

### Complex Sudoku

```{r eval=FALSE}
# Harder puzzle
difficult <- matrix(c(
  5, 3, 0, 0, 7, 0, 0, 0, 0,
  6, 0, 0, 1, 9, 5, 0, 0, 0,
  0, 9, 8, 0, 0, 0, 0, 6, 0,
  8, 0, 0, 0, 6, 0, 0, 0, 3,
  4, 0, 0, 8, 0, 3, 0, 0, 1,
  7, 0, 0, 0, 2, 0, 0, 0, 6,
  0, 6, 0, 0, 0, 0, 2, 8, 0,
  0, 0, 0, 4, 1, 9, 0, 0, 5,
  0, 0, 0, 0, 8, 0, 0, 7, 9
), nrow = 9, byrow = TRUE)

sudoku_solver(difficult)
```

## Maze Solver

Solve mazes using depth-first search algorithms!

### Basic Maze

```{r}
# Create a simple maze (0 = path, 1 = wall)
simple_maze <- matrix(c(
  0, 1, 0, 0, 0,
  0, 1, 0, 1, 0,
  0, 0, 0, 1, 0,
  1, 1, 0, 0, 0,
  0, 0, 0, 1, 0
), nrow = 5, byrow = TRUE)

solution <- maze_solve(
  simple_maze,
  start = c(1, 1),
  end = c(5, 5)
)
print(solution)
```

### Micromouse Competition Maze

```{r eval=FALSE}
# Classic Micromouse-style maze
micromouse <- matrix(c(
  1, 1, 1, 1, 1, 1, 1, 1,
  1, 0, 0, 0, 0, 0, 0, 1,
  1, 0, 1, 1, 1, 1, 0, 1,
  1, 0, 1, 0, 0, 0, 0, 1,
  1, 0, 1, 0, 1, 1, 0, 1,
  1, 0, 0, 0, 0, 1, 0, 1,
  1, 0, 1, 1, 0, 0, 0, 1,
  1, 1, 1, 1, 1, 1, 1, 1
), nrow = 8, byrow = TRUE)

maze_solve(
  micromouse,
  start = c(2, 2),
  end = c(7, 7),
  diagonal = FALSE
)
```

### Advanced Options

```{r eval=FALSE}
# With diagonal movement and aiming toward goal
maze_solve(
  micromouse,
  start = c(2, 2),
  end = c(7, 7),
  diagonal = TRUE, # Allow diagonal moves
  aim = TRUE, # Prefer directions toward goal
  inertia = TRUE, # Prefer continuing in same direction
  timeout = 5 # Max 5 seconds
)
```

## Combining Games & Data Science

Use game functions for:

**Text Analysis:**
```{r eval=FALSE}
# Find anagrams in your dataset
words <- c("listen", "silent", "hello")
scrabble_words(tiles = "listen", language = "en")
```

**Algorithm Teaching:**
```{r eval=FALSE}
# Demonstrate recursion with maze solving
maze_solve(simple_maze, start = c(1, 1), end = c(5, 5))
```

**Pattern Recognition:**
```{r eval=FALSE}
# Wordle simulations for optimal starting words
seeds <- 1:100
results <- lapply(seeds, function(s) {
  wordle_simulation("SAINT", "ABBEY", seed = s, quiet = TRUE)
})
```

## Further Reading

### Package Resources
- **Package documentation:** [https://laresbernardo.github.io/lares/](https://laresbernardo.github.io/lares/)
- **GitHub repository:** [https://github.com/laresbernardo/lares](https://github.com/laresbernardo/lares)
- **Report issues:** [https://github.com/laresbernardo/lares/issues](https://github.com/laresbernardo/lares/issues)

### Game References
- **Wordle:** The viral word game by Josh Wardle
- **Sudoku:** Japanese number puzzle
- **Scrabble:** Classic word board game
- **Micromouse:** Autonomous maze-solving competition

## Next Steps

- Explore data analysis features (see Data Wrangling vignette)
- Learn machine learning (see Machine Learning vignette)
- Try API integrations (see API Integrations vignette)
- Check game function documentation: `?wordle_check`, `?scrabble_words`, `?sudoku_solver`, `?maze_solve`

---
title: "Machine Learning"
author: "Bernardo Lares"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Machine Learning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  warning = FALSE,
  message = FALSE,
  eval = TRUE, # Show actual results like old vignette
  cache = FALSE, # Disabled for CRAN to avoid massive tarball sizes
  cache.lazy = FALSE
)
```

## Introduction

The `lares` package provides a streamlined interface to h2o's AutoML for automated machine learning. This vignette demonstrates how to build, evaluate, and interpret models with minimal code.

## Setup

Install and load required packages:

```{r eval=TRUE}
library(lares)
library(dplyr)
```

h2o must be installed separately:

```{r}
# Install h2o (run once)
# install.packages("h2o")
library(h2o)

# Initialize h2o quietly for vignette
Sys.unsetenv("http_proxy")
Sys.unsetenv("https_proxy")
h2o.init(nthreads = -1, max_mem_size = "2G", ip = "127.0.0.1")
h2o.no_progress() # Disable progress bars
```

## Pipeline

<img src="../man/figures/automl_map.png" alt="h2o_automl workflow" width="100%" />

In short, these are the steps that happen on `h2o_automl`'s backend:

1. **Input Processing**: The function receives a dataframe `df` and the dependent variable `y` to predict. Set `seed` for reproducibility.

2. **Model Type Detection**: Automatically decides between classification (categorical) or regression (continuous) based on `y`'s class and unique values (controlled by `thresh` parameter).

3. **Data Splitting**: Splits data into test and train datasets. Control the proportion with `split` parameter. Replicate this with `msplit()`.

4. **Preprocessing**: 
   - Center and scale numerical values
   - Remove outliers with `no_outliers`
   - Impute missing values with MICE (`impute = TRUE`)
   - Balance training data for classification (`balance = TRUE`)
   - Replicate with `model_preprocess()`

5. **Model Training**: Runs `h2o::h2o.automl()` to train multiple models and generate a leaderboard sorted by performance. Customize with:
   - `max_models` or `max_time`
   - `nfolds` for k-fold cross-validation
   - `exclude_algos` and `include_algos`

6. **Model Selection**: Selects the best model based on performance metric (change with `stopping_metric`). Use `h2o_selectmodel()` to choose an alternative.

7. **Performance Evaluation**: Calculates metrics and plots using test predictions (unseen data). Replicate with `model_metrics()`.

8. **Results**: Returns a list with inputs, leaderboard, best model, metrics, and plots. Export with `export_results()`.

## Quick Start: Binary Classification

Let's build a model to predict Titanic survival:

```{r}
data(dft)

# Train an AutoML model
# Binary classification
model <- h2o_automl(
  df = dft,
  y = "Survived",
  target = "TRUE",
  ignore = c("Ticket", "Cabin", "PassengerId"),
  max_models = 10,
  max_time = 120,
  impute = FALSE
)

# View results
print(model)
```

That's it! `h2o_automl()` handles:

- Train/test split
- One-hot encoding of categorical variables
- Model training with multiple algorithms
- Hyperparameter tuning
- Model selection

## Understanding the Output

The model object contains:

```{r}
names(model)
```

Key components:
- `model`: Best h2o model
- `metrics`: Performance metrics
- `importance`: Variable importance
- `datasets`: Train/test data used
- `parameters`: Configuration used

## Model Performance

### Metrics

View detailed metrics:

```{r}
# All metrics
model$metrics

# Specific metrics
model$metrics$AUC
model$metrics$Accuracy
model$metrics$Logloss
```

### Confusion Matrix

```{r}
# Confusion matrix plot
mplot_conf(
  tag = model$scores_test$tag,
  score = model$scores_test$score,
  subtitle = sprintf("AUC: %.3f", model$metrics$metrics$AUC)
)
```

### ROC Curve

```{r}
# ROC curve
mplot_roc(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)
```

### Gain and Lift Charts

```{r}
# Gain and Lift charts for binary classification
mplot_gain(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)
```

## Variable Importance

See which features matter most:

```{r}
# Variable importance dataframe
head(model$importance, 15)

# Plot top 15 important variables
top15 <- head(model$importance, 15)
mplot_importance(
  var = top15$variable,
  imp = top15$importance
)
```

## Model Interpretation with SHAP

SHAP values explain individual predictions:

```{r, eval=FALSE}
# Calculate SHAP values (computationally expensive)
shap <- h2o_shap(model)

# Plot SHAP summary
plot(shap)
```

## Advanced: Customizing AutoML

### Preprocessing Options

```{r}
model <- h2o_automl(
  df = dft,
  y = "Survived",
  # Ignore specific columns
  ignore = c("Ticket", "Cabin", "PassengerId"),
  # Use only specific algorithms (exclude_algos also available)
  include_algos = c("GBM", "DRF"), # Gradient Boosting & Random Forest
  # Data split
  split = 0.7,
  # Handle imbalanced data
  balance = TRUE,
  # Remove outliers (Z-score > 3)
  no_outliers = TRUE,
  # Impute missing values (requires mice package if TRUE)
  impute = FALSE,
  # Keep only unique training rows
  unique_train = TRUE,
  # Reproducible results
  seed = 123
)
```

## Multi-Class Classification

Predict passenger class (3 categories):

```{r}
model_multiclass <- h2o_automl(
  df = dft,
  y = "Pclass",
  ignore = c("Cabin", "PassengerId"),
  max_models = 10,
  max_time = 60
)

# Multi-class metrics
model_multiclass$metrics

# Confusion matrix for multi-class
mplot_conf(
  tag = model_multiclass$scores_test$tag,
  score = model_multiclass$scores_test$score
)
```

## Regression Example

Predict fare prices:

```{r}
model_regression <- h2o_automl(
  df = dft,
  y = "Fare",
  ignore = c("Cabin", "PassengerId"),
  max_models = 10,
  exclude_algos = NULL
)

# Regression metrics
model_regression$metrics
```

## Using Pre-Split Data

If you have predefined train/test splits:

```{r}
# Create splits
splits <- msplit(dft, size = 0.8, seed = 123)
splits$train$split <- "train"
splits$test$split <- "test"

# Combine
df_split <- rbind(splits$train, splits$test)

# Train using split column
model <- h2o_automl(
  df = df_split,
  y = "Survived",
  train_test = "split",
  max_models = 5
)
```

## Making Predictions

### On New Data

```{r}
# New data (same structure as training)
new_data <- dft[1:10, ]

# Predict
predictions <- h2o_predict_model(new_data, model$model)
head(predictions)
```

### Binary Model Predictions

```{r}
# Get probabilities
predictions <- h2o_predict_model(new_data, model$model)
head(predictions)
```

## Model Comparison

### Full Visualization Suite

```{r}
# Complete model evaluation plots
mplot_full(
  tag = model$scores_test$tag,
  score = model$scores_test$score,
  subtitle = model$model@algorithm
)
```

### Metrics Comparison

```{r}
# Model performance over trees
mplot_metrics(model)
```

## Saving and Loading Models

### Export Results

```{r}
# Save model and plots
export_results(model, subdir = "models", thresh = 0.5)
```

This creates:
- Model file (.rds)
- MOJO file (for production)
- Performance plots
- Metrics summary

### Load Saved Model

```{r, eval=FALSE}
# Load model
loaded_model <- readRDS("models/Titanic_Model/Titanic_Model.rds")

# Make predictions with MOJO (production-ready)
predictions <- h2o_predict_MOJO(
  model_path = "models/Titanic_Model",
  df = dft[1:10, ]
)
```

## Best Practices

### 1. Start Simple

```{r}
# Quick prototype
model <- h2o_automl(dft, "Survived", max_models = 3, max_time = 30)
```

### 2. Iterate and Refine

```{r}
# Refine based on results
model <- h2o_automl(
  dft, "Survived",
  max_models = 20,
  no_outliers = TRUE,
  balance = TRUE,
  ignore = c("PassengerId", "Name", "Ticket", "Cabin"),
  model_name = "Titanic_Model"
)
```

### 3. Validate Thoroughly

```{r}
# Check multiple metrics
model$metrics

# Visual inspection
mplot_full(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)

# Variable importance
mplot_importance(
  var = model$importance$variable,
  imp = model$importance$importance
)
```

### Score Distribution

```{r}
# Density plot
mplot_density(
  tag = model$scores_test$tag,
  score = model$scores_test$score
)
```

### 4. Document Your Process

```{r}
# Save everything
export_results(model, subdir = "my_project", thresh = 0.5)
```

## Troubleshooting

### h2o Initialization Issues

```{r, eval=FALSE}
# Manually initialize h2o with more memory
h2o::h2o.init(max_mem_size = "8G", nthreads = -1)
```

### Clean h2o Environment

```{r}
# Remove all models
h2o::h2o.removeAll()

# Shutdown h2o
h2o::h2o.shutdown(prompt = FALSE)
```

### Check h2o Flow UI

```{r}
# Open h2o's web interface
# Navigate to: http://localhost:54321/flow/index.html
```

## Further Reading

### Package & ML Resources
- **h2o AutoML Documentation:** [https://docs.h2o.ai/h2o/latest-stable/h2o-docs/automl.html](https://docs.h2o.ai/h2o/latest-stable/h2o-docs/automl.html)
- **lares ML functions:** `?h2o_automl`, `?h2o_explainer`, `?mplot_full`
- **SHAP explanations:** [https://github.com/shap/shap](https://github.com/shap/shap)


## Next Steps

- Explore data wrangling features (see Data Wrangling vignette)
- Learn about API integrations (see API Integrations vignette)
- Review individual plotting functions: `?mplot_conf`, `?mplot_roc`, `?mplot_importance`

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shapley.R
\name{shap_var}
\alias{shap_var}
\title{SHAP-based dependence plots for categorical/numerical features (PDP)}
\usage{
shap_var(x, var, keep_outliers = FALSE)
}
\arguments{
\item{x}{\code{h2o_shap} object}

\item{var}{Variable name}

\item{keep_outliers}{Boolean. Outliers detected with z-score and 3sd
may be suppress or kept in your plot. Keep them?}
}
\value{
ggplot2 objct with shap values plotted
}
\description{
Having a \code{h2o_shap} object, plot a dependence plot for any
categorical or numerical feature.
}
\examples{
\dontrun{
# Train a h2o_automl model
model <- h2o_automl(dft, Survived,
  max_models = 1, target = TRUE,
  ignore = c("Ticket", "Cabin", "PassengerId"),
  quiet = TRUE
)

# Calculate SHAP values
SHAP_values <- h2o_shap(model)
# Equivalent to:
# SHAP_values <- h2o_shap(
#  model = model$model,
#  test = model$datasets$test,
#  scores = model$scores_test$scores)

# Check SHAP results
head(SHAP_values)

# You must have "ggbeeswarm" library to use this auxiliary function:
# Plot SHAP values (feature importance)
plot(SHAP_values)

# Plot some of the variables (categorical)
shap_var(SHAP_values, Pclass)

# Plot some of the variables (numerical)
shap_var(SHAP_values, Fare)
}
}
\seealso{
Other SHAP: 
\code{\link{h2o_shap}()}
}
\concept{SHAP}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frequencies.R
\name{freqs_list}
\alias{freqs_list}
\title{Frequencies on Lists and UpSet Plot}
\usage{
freqs_list(
  df,
  var = NULL,
  wt = NULL,
  fx = "mean",
  rm.na = FALSE,
  min_elements = 1,
  limit = 10,
  limit_x = NA,
  limit_y = NA,
  tail = TRUE,
  size = 10,
  unique = TRUE,
  abc = FALSE,
  title = "",
  plot = TRUE
)
}
\arguments{
\item{df}{Data.frame}

\item{var}{Variable. Variables you wish to process.}

\item{wt}{Variable, numeric. Select a numeric column to use
in the colour scale, used as sum, mean... of those values for each
of the combinations.}

\item{fx}{Character. Set operation: mean, sum}

\item{rm.na}{Boolean. Remove NA value from \code{wt}?}

\item{min_elements}{Integer. Exclude combinations with less than n elements}

\item{limit, limit_x, limit_y}{Integer. Show top n combinations (x) and/or
elements (y). The rest will be grouped into a single element.
Set argument to 0 to ignore. \code{limit_x}/\code{limit_y} answer to
\code{limit}'s argument.}

\item{tail}{Boolean. Show tail grouped into "..." on the plots?}

\item{size}{Numeric. Text base size}

\item{unique}{Boolean. a,b = b,a?}

\item{abc}{Boolean. Do you wish to sort by alphabetical order?}

\item{title}{Character. Overwrite plot's title with.}

\item{plot}{Boolean. Plot viz? Will be generated anyways in the output object}
}
\value{
List. data.frame with the data results, elements and combinations.
}
\description{
Visualize frequency of elements on a list, list vector, or vector with
comma separated values. Detect which combinations and elements are
the most frequent and how much they represent of your total observations.
This is similar to the \href{http://vcg.github.io/upset/}{UpSet Plots}
which may be used as an alternative to Venn diagrams.
}
\examples{
\dontrun{
df <- dplyr::starwars
head(df[, c(1, 4, 5, 12)], 10)

# Characters per movies combinations in a list column
head(df$films, 2)
freqs_list(df, films)

# Skin colours in a comma-separated column
head(df$skin_color)
x <- freqs_list(df, skin_color, min_elements = 2, limit = 5, plot = FALSE)
# Inside "x" we'll have:
names(x)

# Using the 'wt' argument to add a continuous value metric
# into an already one-hot encoded columns dataset (and hide tail)
csv <- "https://raw.githubusercontent.com/hms-dbmi/UpSetR/master/inst/extdata/movies.csv"
movies <- read.csv(csv, sep = ";")
head(movies)
freqs_list(movies,
  wt = AvgRating, min_elements = 2, tail = FALSE,
  title = "Movies\nMixed Genres\nRanking"
)
# So, please: no more Comedy+SciFi and more Drama+Horror films (based on ~50 movies)!
}
}
\seealso{
Other Frequency: 
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_plot}()}

Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}

Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}
\concept{Frequency}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correlations.R
\name{corr_var}
\alias{corr_var}
\alias{plot.corr_var}
\title{Correlation between variable and dataframe}
\usage{
corr_var(
  df,
  var,
  ignore = NULL,
  trim = 0,
  clean = FALSE,
  plot = TRUE,
  top = NA,
  ceiling = 1,
  max_pvalue = 1,
  limit = 10,
  ranks = FALSE,
  zeroes = FALSE,
  save = FALSE,
  quiet = FALSE,
  ...
)

\method{plot}{corr_var}(x, var, max_pvalue = 1, top = NA, limit = NULL, ...)
}
\arguments{
\item{df}{Dataframe. It doesn't matter if it's got non-numerical
columns: they will be filtered.}

\item{var}{Variable. Name of the variable to correlate. Note that if the
variable \code{var} is not numerical, 1. you may define which category to select
from using `var_category`; 2. You may have to add \code{redundant = TRUE} to
enable all categories (instead of \code{n-1}).}

\item{ignore}{Character vector. Which columns do you wish to exclude?}

\item{trim}{Integer. Trim words until the nth character for
categorical values (applies for both, target and values)}

\item{clean}{Boolean. Use lares::cleanText for categorical values (applies
for both, target and values)}

\item{plot}{Boolean. Do you wish to plot the result? If set to TRUE, the
function will return only the plot and not the result's data}

\item{top}{Integer. If you want to plot the top correlations,
define how many}

\item{ceiling}{Numeric. Remove all correlations above... Range: (0-1]}

\item{max_pvalue}{Numeric. Filter non-significant variables. Range (0, 1]}

\item{limit}{Integer. Limit one hot encoding to the n most frequent
values of each column. Set to \code{NA} to ignore argument.}

\item{ranks}{Boolean. Add ranking numbers?}

\item{zeroes}{Do you wish to keep zeroes in correlations too?}

\item{save}{Boolean. Save output plot into working directory}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters passed to \code{corr} and \code{cor.test}}

\item{x}{corr_var object}
}
\value{
data.frame. With variables, correlation and p-value results
for each feature, arranged by descending absolute correlation value.
}
\description{
This function correlates a whole dataframe with a single feature. It
automatically runs \code{ohse} (one-hot-smart-encoding) so no need to input
only numerical values.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset

corr_var(dft, Survived, method = "spearman", plot = FALSE, top = 10)

# With plots, results are easier to compare:

# Correlate Survived with everything else and show only significant results
dft \%>\% corr_var(Survived_TRUE, max_pvalue = 0.05)

# Top 15 with less than 50\% correlation and show ranks
dft \%>\% corr_var(Survived_TRUE, ceiling = .6, top = 15, ranks = TRUE)
}
\seealso{
Other Exploratory: 
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}

Other Correlations: 
\code{\link{corr}()},
\code{\link{corr_cross}()}
}
\concept{Correlations}
\concept{Exploratory}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/google_analytics.R
\name{queryGA}
\alias{queryGA}
\title{Queries on Google Analytics}
\usage{
queryGA(
  account,
  creds = NA,
  token_dir = NA,
  metrics = "sessions",
  dimensions = "date",
  met_filters = NULL,
  dim_filters = NULL,
  start = lubridate::floor_date(Sys.Date(), "month"),
  end = Sys.Date()
)
}
\arguments{
\item{account}{Character. Personal named accounts}

\item{creds}{Character. Credential's user (see \code{get_creds()})}

\item{token_dir}{Character. Credential's directory (see \code{get_creds()})}

\item{metrics}{Character. Which metrics we wish to bring}

\item{dimensions}{Character. Which dimensions we wish to bring}

\item{met_filters, dim_filters}{A \code{filter_clause_ga4} for filtering
metrics/dimensions. Check \code{googleAnalyticsR::google_analytics()}.}

\item{start}{Date. Start date for the report}

\item{end}{Date. End date for the report}
}
\value{
data.frame with the API GET request tabulated results.
}
\description{
This function lets the user query Google Analytics with its API.
More about the documentation and parameters in
\code{googleAnalyticsR::google_analytics()} or Google Analytics' API.
}
\seealso{
Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}

Other Google: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{readGS}()}

Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{slackSend}()}
}
\concept{API}
\concept{Credentials}
\concept{Google}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lasso.R
\name{lasso_vars}
\alias{lasso_vars}
\title{Most Relevant Features Using Lasso Regression}
\usage{
lasso_vars(
  df,
  variable,
  ignore = NULL,
  nlambdas = 100,
  nfolds = 10,
  top = 20,
  quiet = FALSE,
  seed = 123,
  ...
)
}
\arguments{
\item{df}{Dataframe. Any dataframe is valid as \code{ohse} will be applied to
process categorical values, and values will be standardize automatically.}

\item{variable}{Variable. Dependent variable or response.}

\item{ignore}{Character vector. Variables to exclude from study.}

\item{nlambdas}{Integer. Number of lambdas to be used in a search.}

\item{nfolds}{Integer. Number of folds for K-fold cross-validation (>= 2).}

\item{top}{Integer. Plot top n results only.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{seed}{Numeric.}

\item{...}{Additional parameters passed to \code{ohse()}.}
}
\value{
List. Contains lasso model coefficients, performance metrics, the
actual model fitted and a plot.
}
\description{
Use Lasso regression to identify the most relevant variables that
can predict/identify another variable. You might want to compare
with \code{corr_var()} and/or \code{x2y()} results to compliment
the analysis No need to standardize, center or scale your data.
Tidyverse friendly.
}
\examples{
\dontrun{
# CRAN
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset

m <- lasso_vars(dft, Survived, ignore = c("Cabin"))
print(m$coef)
print(m$metrics)
plot(m$plot)
}
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}
\concept{Machine Learning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{removenacols}
\alias{removenacols}
\alias{removenarows}
\alias{numericalonly}
\title{Remove/Drop Columns in which ALL or SOME values are NAs}
\usage{
removenacols(df, all = TRUE, ignore = NULL)

removenarows(df, all = TRUE)

numericalonly(df, dropnacols = TRUE, logs = FALSE, natransform = NA)
}
\arguments{
\item{df}{Data.frame}

\item{all}{Boolean. Remove rows which contains ONLY NA values.
If set to FALSE, rows which contains at least one NA will be removed}

\item{ignore}{Character vector. Column names to ignore validation.}

\item{dropnacols}{Boolean. Drop columns with only NA values?}

\item{logs}{Boolean. Calculate log(x)+1 for numerical columns?}

\item{natransform}{String. "mean" or 0 to impute NA values. If
set to NA no calculation will run.}
}
\value{
data.frame with removed columns.

data.frame with removed rows.

data.frame with all numerical columns selected.
}
\description{
This function lets the user remove all columns that have some or
all values as NAs

This function lets the user remove all rows that have some or
all values as NAs
}
\examples{
data(dft) # Titanic dataset
str(dft)
numericalonly(dft) \%>\% head()
numericalonly(dft, natransform = "mean") \%>\% head()
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_text.R
\name{grepm}
\alias{grepm}
\title{Pattern Matching for Any or All Multiple Matches}
\usage{
grepm(pattern, x, type = "all", ...)
}
\arguments{
\item{pattern}{character string containing a \link[base]{regular expression}
    (or character string for \code{fixed = TRUE}) to be matched
    in the given character vector.  Coerced by
    \code{\link[base]{as.character}} to a character string if possible.  If a
    character vector of length 2 or more is supplied, the first element
    is used with a warning.  Missing values are allowed except for
    \code{regexpr}, \code{gregexpr} and \code{regexec}.}

\item{x}{Character vector. Text where matches are sought, or an object
which can be coerced by as.character to a character vector.
Long vectors are supported.}

\item{type}{Character. Type of match. Choose one of:
\code{any}, \code{all}}

\item{...}{Additional arguments to pass to \code{grepl}}
}
\value{
Boolean of same length as \code{x}
}
\description{
This function returns a boolean vector of the same length as `x`,
each element of which is the result of applying the `type` of matches
to the corresponding element of `x`, using regular expressions.
}
\examples{
x <- c(123, 876, 18761)
patterns <- c(1, 2)
grepm(patterns, x, type = "any")
grepm(patterns, x, type = "all")
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{quants}
\alias{quants}
\title{Calculate cuts by quantiles}
\usage{
quants(values, splits = 10, return = "labels", n = 2)
}
\arguments{
\item{values}{Vector. Values to calculate quantile cuts}

\item{splits}{Integer. How many cuts should split the values?}

\item{return}{Character. Return "summary" or "labels"}

\item{n}{Integer. Determines the number of digits used in
formatting the break numbers.}
}
\value{
Factor vector or data.frame. Depending on \code{return} input:
\itemize{
  \item \code{labels} a factor ordered vector with each observation's quantile
  \item \code{summary} a data.frame with information on each quantile cut
}
}
\description{
This function lets the user quickly calculate cuts for quantiles
and discretize numerical values into categorical values.
}
\examples{
data(dft) # Titanic dataset
quants(dft$Age, splits = 5, "summary")
quants(dft$Age, splits = 5, "labels")[1:10]
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Calculus: 
\code{\link{corr}()},
\code{\link{corr_cross}()},
\code{\link{dist2d}()},
\code{\link{model_metrics}()}
}
\concept{Calculus}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{topics_rake}
\alias{topics_rake}
\title{Keyword/Topic identification using RAKE}
\usage{
topics_rake(text, file = "english-ewt-ud-2.4-190531.udpipe", lang = "english")
}
\arguments{
\item{text}{Character vector}

\item{file}{Character. Name of \code{udpipe} model previously downloaded
for a specific language}

\item{lang}{Character. If file does not exist, this language will be
downloaded from \code{udpipe}'s models.}
}
\value{
data.frame with topics for each \code{text} input.
}
\description{
RAKE is a basic algorithm which tries to identify keywords in text.
Based on \code{udpipe} library, model models, and keywords_rake function.
}
\seealso{
Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{ngrams}()},
\code{\link{remove_stopwords}()},
\code{\link{replaceall}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textCloud}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()}
}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecasting.R
\name{forecast_arima}
\alias{forecast_arima}
\title{ARIMA Forecast}
\usage{
forecast_arima(
  time,
  values,
  n_future = 30,
  ARMA = 8,
  ARMA_min = 5,
  AR = NA,
  MA = NA,
  wd_excluded = NA,
  plot = TRUE,
  plot_days = 90,
  project = NA
)
}
\arguments{
\item{time}{POSIX. Vector with date values}

\item{values}{Numeric. Vector with numerical values}

\item{n_future}{Integer. How many steps do you wish to forecast?}

\item{ARMA}{Integer. How many days should the model look back for ARMA?
Between 5 and 10 days recommmended. If set to 0 then it will forecast
until the end of max date's month; if set to -1, until the end of
max date's following month}

\item{ARMA_min}{Integer. How many days should the model look back for ARMA?
Between 5 and 10 days recommmended. If set to 0 then it will forecast
until the end of max date's month; if set to -1, until the end of
max date's following month}

\item{AR}{Integer. Force AR value if known}

\item{MA}{Integer. Force MA value if known}

\item{wd_excluded}{Character vector. Which weekdays are excluded in
your training set. If there are, please define know which ones. Example:
c('Sunday','Thursday'). If set to 'auto' then it will detect automatically
which weekdays have no data and forcast without these days.}

\item{plot}{Boolean. If you wish to plot your results}

\item{plot_days}{Integer. How many days back you wish to plot?}

\item{project}{Character. Name of your forecast project}
}
\value{
List. Containing the trained model, forecast accuracy results,
data.frame for forecast (test) and train, and if \code{plot=TRUE}, a plot.
}
\description{
This function automates the ARIMA iterations and modeling for
time forecasting. For the moment, units can only be days.
}
\details{
The ARIMA method is appropriate only for a time series that is
stationary (i.e., its mean, variance, and autocorrelation should
be approximately constant through time) and it is recommended
that there are at least 50 observations in the input data.

 The model consists of two parts, an autoregressive (AR) part
 and a moving average (MA) part. The AR part involves regressing
 the variable on its own lagged (i.e., past) values. The MA part
 involves modeling the error term as a linear combination of error
 terms occurring contemporaneously and at various times in the past.

One thing to keep in mind when we think about ARIMA models is
given by the great power to capture very complex patters of
temporal correlation (Cochrane, 1997: 25)
}
\seealso{
Other Forecast: 
\code{\link{prophesize}()}
}
\concept{Forecast}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/robyn.R
\name{robyn_hypsbuilder}
\alias{robyn_hypsbuilder}
\title{Robyn: Generate default hyperparameters}
\usage{
robyn_hypsbuilder(
  channels,
  media_type = "default",
  adstock = "geometric",
  date_type = "weekly",
  lagged = FALSE
)
}
\arguments{
\item{channels}{Character vector. Paid media and organic variables names.}

\item{media_type}{Character vector. Must be length 1 or same as
\code{channels}. Pick, for every \code{channels} value,
what type of media it is: "online" or "offline".}

\item{adstock}{Character. Pick one of: "geometric" or "weibull".}

\item{date_type}{Character. Pick one of: "daily", "weekly", or "monthly".
Only valid to transform thetas when using geometric adstock. Set to "skip"
in case you wish to leave default weekly values.}

\item{lagged}{Boolean vector. Must be length 1 or same as
\code{channels}. Pick, for every \code{channels} value,
if you wish to have a lagged effect. Only valid for Weibull adstock.}
}
\value{
list with default hyperparameters ranges.
}
\description{
Generate a list with hyperparameter default values, ready to be
passed to \code{Robyn::robyn_inputs()}.
}
\examples{
robyn_hypsbuilder(
  channels = c(
    "branded_search_spend",
    "nonbranded_search_spend",
    "print_spend",
    "ooh_spend",
    "tv_spend",
    "radio_spend"
  ),
  media_type = c(
    "online", "online", "offline",
    "offline", "offline", "offline"
  ),
  adstock = "geometric",
  date_type = "weekly"
)
}
\seealso{
Other Robyn: 
\code{\link{robyn_marginal}()},
\code{\link{robyn_modelselector}()},
\code{\link{robyn_performance}()}
}
\concept{Robyn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/other_plots.R
\name{export_plot}
\alias{export_plot}
\title{Export ggplot2, gridExtra, or any plot object into rendered file}
\usage{
export_plot(
  p,
  name = "plot",
  vars = NA,
  sep = ".vs.",
  width = 8,
  height = 6,
  format = "png",
  res = 300,
  dir = getwd(),
  subdir = NA,
  quiet = FALSE
)
}
\arguments{
\item{p}{Plot object. Plot to render and export.}

\item{name}{Character. File's name or suffix if vars is not \code{NA}. No need
to include file format on file name.}

\item{vars}{Vector. Variable names to identify by filename.}

\item{sep}{Character. Separator for \code{vars}.}

\item{width, height, res}{Numeric. Plot's width, height, and res (for grids).}

\item{format}{Character. One of: \code{png} or \code{jpeg}.}

\item{dir, subdir}{Character. In which directory/subdirectory do you
wish to save the plot? Working directory as default \code{dir}.}

\item{quiet}{Boolean. Display successful message with filename when saved?}
}
\value{
No return value, called for side effects.
}
\description{
Export any \code{ggplot2}, \code{gridExtra}, or any plot object
created with R into rendered \code{png} or \code{jpg} file.
}
\examples{
p <- noPlot()
export_plot(p, name = "noplot", width = 10, height = 8, res = 300, dir = tempdir())
export_plot(p, name = "noplot2", subdir = "newplots", dir = tempdir())
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dim_reduction.R
\name{reduce_tsne}
\alias{reduce_tsne}
\title{Reduce Dimensionality with t-SNE}
\usage{
reduce_tsne(df, n = 2, ignore = NULL, quiet = FALSE, plot = TRUE, ...)
}
\arguments{
\item{df}{Dataframe}

\item{n}{Integer. Number of dimensions to reduce to.}

\item{ignore}{Character vector. Names of columns to ignore.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{plot}{Boolean. Create plots?}

\item{...}{Additional parameters passed to \code{Rtsne::Rtsne}}
}
\value{
List with reduced dataframe and possible plots.
}
\description{
t-SNE takes high-dimensional data and reduces it to a low-dimensional
graph (1-3 dimensions). Unlike PCA, t-SNE can reduce dimensions with
non-linear relationships. PCA attempts to draw the best fitting line
through the distribution. T-SNE calculates a similarity measure
based on the distance between points instead of trying to maximize variance.
}
\examples{
\dontrun{
data("iris")
df <- subset(iris, select = c(-Species))
df$id <- seq_len(nrow(df))
reduce_tsne(df, ignore = "id", max_iter = 800, perplexity = 20)
}
}
\seealso{
Other Dimensionality: 
\code{\link{reduce_pca}()}

Other Clusters: 
\code{\link{clusterKmeans}()},
\code{\link{clusterOptimalK}()},
\code{\link{clusterVisualK}()},
\code{\link{reduce_pca}()}
}
\concept{Clusters}
\concept{Dimensionality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{dfr}
\alias{dfr}
\title{Results for AutoML Predictions}
\format{
An object of class \code{"list"} with 3 \code{"data.frame"}
\describe{
  \item{class2}{Predictions for a Binomial Classification Model}
  \item{class3}{Predictions for a Multi-Categorical Classification Model}
  \item{regr}{Predictions for a Continuous Regression Model}
}
}
\usage{
data(dfr)
}
\value{
List
}
\description{
List with categorical (2 and 3 classes) and continuous predictions,
generated with \code{h2o_automl()} and the \code{dft}. Note that
the models per se won't work to predict.
}
\examples{
data(dfr)
lapply(dfr, head)
}
\seealso{
Other Dataset: 
\code{\link{dft}}
}
\concept{Dataset}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_text.R
\name{formatColoured}
\alias{formatColoured}
\title{Print Coloured Messages}
\usage{
formatColoured(
  txt,
  colour = c("yellow", "blue", "grey"),
  bold = FALSE,
  cat = TRUE
)
}
\arguments{
\item{txt}{Character. Text to print or transform.}

\item{colour}{Character. Any of: grey, red, green, yellow, blue, or purple.}

\item{bold}{Boolean. Set bold text?}

\item{cat}{Boolean. Print with cat? If not, raw string}
}
\value{
Depends on \code{cat}: NULL if TRUE or character string if FALSE.
}
\description{
Print Coloured Messages
}
\examples{
opts <- c("GREY", "RED", "GREEN", "YELLOW", "BLUE", "PURPLE")
for (colour in opts) formatColoured(paste("Colour:", colour, "\n"), colour)
formatColoured("my bold coloured text", bold = TRUE, cat = TRUE)
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crosstab.R
\name{crosstab}
\alias{crosstab}
\title{Weighted Cross Tabulation}
\usage{
crosstab(
  df,
  ...,
  wt = NULL,
  prow = FALSE,
  pcol = FALSE,
  pall = FALSE,
  decimals = 2,
  rm.na = FALSE,
  total = TRUE,
  order = TRUE
)
}
\arguments{
\item{df}{Data.frame.}

\item{...}{Variables. Dependent and independent variables.}

\item{wt}{Variable, numeric. Weights.}

\item{prow, pcol, pall}{Boolean. Calculate percent values for rows, columns,
or the whole table, respectively.}

\item{decimals}{Integer. How many decimals should be returned?}

\item{rm.na}{Boolean. Remove NA values?}

\item{total}{Boolean. Return total values column?}

\item{order}{Boolean. Sort columns and rows by frequencies? Else, will
be sorted alphabetically}
}
\value{
data.frame. Result of crossing the variables provided in \code{...} and
counting how many observations (rows) fall into each criteria.
}
\description{
A cross-tabulation function with output similar to STATA, tidy
friendly, with weighting possibility.
}
\examples{
data(dft) # Titanic dataset
crosstab(dft, Survived, Pclass, total = FALSE)
# Show values in percentages
crosstab(dft, Pclass, Survived, prow = TRUE)
crosstab(dft, Pclass, Survived, pall = TRUE)
# Weighted by another variable
crosstab(dft, Survived, Pclass, wt = Fare, prow = TRUE)
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/other_plots.R
\name{plot_timeline}
\alias{plot_timeline}
\title{Plot timeline as Gantt Plot}
\usage{
plot_timeline(
  event,
  start,
  end = start + 1,
  label = NA,
  group = NA,
  title = "Curriculum Vitae Timeline",
  subtitle = "Bernardo Lares",
  interactive = FALSE,
  save = FALSE,
  subdir = NA
)
}
\arguments{
\item{event}{Vector. Event, role, label, or row.}

\item{start}{Vector. Start date.}

\item{end}{Vector. End date. Only one day be default if not defined}

\item{label}{Vector. Place, institution, or label.}

\item{group}{Vector. Academic, Work, Extracurricular... Pass as factor
to keep a specific order}

\item{title}{Character. Title for the plot}

\item{subtitle}{Character. Subtitle for the plot}

\item{interactive}{Boolean. Run with plotly?}

\item{save}{Boolean. Save the output plot in our working directory}

\item{subdir}{Character. Into which subdirectory do you wish to save the plot to?}
}
\value{
ggplot2 object
}
\description{
This function plots groups of observartions with timelines in a
Gantt Plot way. Only works if start and end are date format values.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
cols <- c("Role", "Place", "Type", "Start", "End")
today <- as.character(Sys.Date())
cv <- data.frame(rbind(
  c("Marketing Science Partner", "Facebook", "Work Experience", "2019-12-09", today),
  c("Data Scientist Consultant", "MatrixDS", "Work Experience", "2018-09-01", today),
  c("R Community Contributor", "lares library", "Extra", "2018-07-18", today),
  c("Lead Data Scientist", "MEG", "Work Experience", "2019-01-15", "2019-12-09"),
  c("Head of Analytics", "Comparamejor/R5", "Work Experience", "2016-08-01", "2019-01-15"),
  c("Big Data & Data Science Programme", "UdC", "Academic", "2017-09-01", "2018-02-28"),
  c("Project Engineer", "Polytex", "Work Experience", "2016-05-15", "2016-09-01"),
  c("Big Data Analyst", "MEG", "Work Experience", "2016-01-01", "2016-04-30"),
  c("Advanced Excel Instructor", "ARTS", "Work Experience", "2015-11-01", "2016-04-30"),
  c("Continuous Improvement Intern", "PAVCO", "Work Experience", "2015-04-01", "2015-08-30"),
  c("Mechanical Design Intern", "SIGALCA", "Work Experience", "2013-07-01", "2013-09-30"),
  c("DJs Online Community Owner", "LaresDJ.com / SoloParaDJs", "Extra", "2010-01-05", "2020-05-20"),
  c("Mechanical Engineer Degree", "USB", "Academic", "2009-09-15", "2015-11-20"),
  c("DJ and Composer/Producer", "Legacy Discplay", "Extra", "2009-05-01", "2015-04-30")
))
colnames(cv) <- cols
plot_timeline(
  event = cv$Role,
  start = cv$Start,
  end = cv$End,
  label = cv$Place,
  # Simple trick to re-arrange the grids
  group = factor(cv$Type, levels = c("Work Experience", "Academic", "Extra"))
)
}
\seealso{
Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{tree_var}()}
}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{num_abbr}
\alias{num_abbr}
\title{Abbreviate or dis-abbreviate numerical values}
\usage{
num_abbr(x, n = 3, numeric = FALSE, ...)
}
\arguments{
\item{x}{Numeric vector}

\item{n}{Integer. Single numeric value, specifying number of
significant figures to show. Range 1 to 6.}

\item{numeric}{Boolean. Transform abbreviated number into numeric?}

\item{...}{Additional parameters.}
}
\value{
Vector of character or numeric values that contain converted values
}
\description{
This function converts a numeric vector's values into their
abbreviated character equivalent, i.e. 100E6 into 100M
and viceversa.
}
\examples{
num_abbr(rnorm(10) * 1e6)
num_abbr(rnorm(10) * 1e6, n = 1)
num_abbr(c("3K", "-58.3M", NA, 1), numeric = TRUE)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wordle.R
\name{wordle_check}
\alias{wordle_check}
\alias{print.wordle_check}
\alias{wordle_dictionary}
\alias{wordle_simulation}
\alias{print.wordle_simulation}
\title{Wordle Game Validation}
\usage{
wordle_check(
  input,
  word,
  dictionary = NULL,
  lang_dic = "en",
  method = 3,
  print = TRUE
)

\method{print}{wordle_check}(x, print = TRUE, ...)

wordle_dictionary(lang_dic = "en", method = 3, quiet = TRUE)

wordle_simulation(input, word, seed = NULL, quiet = FALSE, ...)

\method{print}{wordle_simulation}(x, type = 1, ...)
}
\arguments{
\item{input}{Character. Word to validate (5-letters)}

\item{word}{Character. Word actually answer (5-letters).}

\item{dictionary}{Character vector. List of valid words. If set to NULL
then will use modified \code{scrabble_dictionary()} to fetch 5 letter words.
Use \code{lang_dic} param to set language.}

\item{lang_dic}{Character. Any of: "en", "es". Only used when \code{dictionary}
parameter is NULL. Requires internet connection the first time. Uses cache.}

\item{method}{Integer. 1 for \code{scrabble_dictionary()}, 3 for scrapping
the words taken straight from the game's source code.}

\item{print}{Boolean. Print validation results?}

\item{x}{Object to print}

\item{...}{Additional parameters.}

\item{quiet}{Boolean. Do not print words as they are being searched.}

\item{seed}{Numeric. For reproducibility. Accepts more than one: will
run as many seeds there are.}

\item{type}{Integer. 1 for summary and 2 for coloured results.}
}
\value{
Invisible vector with results by letter.
}
\description{
Given and input and a word, validate each letter based on Wordle's
rules: correct letter in correct placement (green), correct letter in wrong
placement (yellow), letter is not present (red).
}
\examples{
word <- "ABBEY"
# Or pick a random one:
# word <- sample(wordle_dictionary("en"), 1)
wordle_check("OPENS", word)
wordle_check("BABES", word)
wordle_check("KEBAB", word, print = FALSE)
wordle_check("ABYSS", word)
wordle_check("ABBEY", word)
# Feel free to use scrabble_words() for hints
\donttest{
x <- wordle_simulation(input = "SAINT", word = "ABBEY", seed = 1:3)
print(x)
}
}
\seealso{
Other Games: 
\code{\link{maze_solve}()},
\code{\link{scrabble_dictionary}()},
\code{\link{sudoku_solver}()}
}
\concept{Games}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_metrics.R
\name{loglossBinary}
\alias{loglossBinary}
\title{Logarithmic Loss Function for Binary Models}
\usage{
loglossBinary(tag, score, eps = 0.001)
}
\arguments{
\item{tag}{Vector. Real known label}

\item{score}{Vector. Predicted value or model's result}

\item{eps}{Numeric. Epsilon value}
}
\description{
This function calculates log loss/cross-entropy loss for binary
models. NOTE: when result is 0.69315, the classification is neutral;
it assigns equal probability to both classes.
}
\seealso{
Other Model metrics: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{errors}()},
\code{\link{gain_lift}()},
\code{\link{model_metrics}()}
}
\concept{Model metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_roc}
\alias{mplot_roc}
\title{ROC Curve Plot}
\usage{
mplot_roc(
  tag,
  score,
  multis = NA,
  sample = 1000,
  model_name = NA,
  subtitle = NA,
  interval = 0.2,
  squared = TRUE,
  plotly = FALSE,
  save = FALSE,
  subdir = NA,
  file_name = "viz_roc.png"
)
}
\arguments{
\item{tag}{Vector. Real known label.}

\item{score}{Vector. Predicted value or model's result.}

\item{multis}{Data.frame. Containing columns with each category probability
or score (only used when more than 2 categories coexist).}

\item{sample}{Integer. Number of samples to use for rendering plot.}

\item{model_name}{Character. Model's name}

\item{subtitle}{Character. Subtitle to show in plot}

\item{interval}{Numeric. Interval for breaks in plot}

\item{squared}{Boolean. Keep proportions?}

\item{plotly}{Boolean. Use plotly for plot's output for an interactive plot}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}
}
\value{
Plot with ROC curve and AUC performance results.
}
\description{
This function plots ROC Curves with AUC values with 95\% confidence
range. It also works for multi-categorical models.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dfr) # Results for AutoML Predictions
lapply(dfr[c(1, 2)], head)

# ROC Curve for Binomial Model
mplot_roc(dfr$class2$tag, dfr$class2$scores,
  model_name = "Titanic Survived Model"
)

# ROC Curves for Multi-Categorical Model
mplot_roc(dfr$class3$tag, dfr$class3$score,
  multis = subset(dfr$class3, select = -c(tag, score)),
  squared = FALSE,
  model_name = "Titanic Class Model"
)
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_response}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_rf}
\alias{fb_rf}
\title{Facebook Reach and Frequency API}
\usage{
fb_rf(
  token,
  ad_account = NA,
  prediction = NA,
  objective = "REACH",
  days = 28,
  budget = 2e+06,
  destination_ids = NA,
  countries = "MX",
  frequency_cap = 8,
  prediction_mode = 1,
  curve = TRUE,
  api_version = NULL,
  process = TRUE,
  ...
)
}
\arguments{
\item{token}{Character. Valid access token with sufficient privileges. Visit the
\href{https://developers.facebook.com/tools/explorer}{Facebook API Graph Explorer}
to acquire one.}

\item{ad_account}{Character. Ad Account. Remember to start with \code{act_}. If you
use the \code{prediction} argument, no need to provide this parameter.}

\item{prediction}{Integer. Prediction ID if you already created the prediction
and wish to query the curve's data. As this prediction already exists, the
rest of arguments of this function will be ignored.}

\item{objective}{Character. Any of: "BRAND_AWARENESS", "LINK_CLICKS", "POST_ENGAGEMENT",
"MOBILE_APP_INSTALLS", "CONVERSIONS", "REACH", or "VIDEO_VIEWS".}

\item{days}{Integer. Amount of days for your campaign's predictions.}

\item{budget}{Integer. The budget in the Ad Account currency in cents.}

\item{destination_ids}{Integer vector. Page ID and/or Instagram Account ID.}

\item{countries}{Character vector. Country's acronyms.}

\item{frequency_cap}{Integer. Frequency cap over all the campaign duration.}

\item{prediction_mode}{Integer. "1" for predicting Reach by providing budget,
"2" is for predicting Budget given a specific Reach.}

\item{curve}{Boolean. Return curve data? If not, only prediction will be created.}

\item{api_version}{Character. Facebook API version.}

\item{process}{Boolean. Process GET results to a more friendly format?}

\item{...}{Additional parameters passed to target specs.}
}
\value{
data.frame with un-nested processed results if \code{process=TRUE} or
raw API results as list when \code{process=FALSE}.
}
\description{
Create or query reach and frequency predictions using Facebook's
Reach and Frequency API. For more information on the API and its parameters, go to the
\href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{original documentaion}.
}
\examples{
\dontrun{
token <- "YOURTOKEN"
account_id <- "act_20846447"

# BASIC 1: Create and return data for a new prediction
basic1 <- fb_rf(token, account_id, destination_ids = 187071108930, countries = "AR")

# BASIC 2: Fetch data for an existing prediction ID
basic2 <- fb_rf(token, account_id, prediction = 6317720998974)

# ADVANCED (Fully custom prediction)
advanced <- fb_rf(token, account_id,
  objective = "REACH",
  days = 28,
  budget = 2000000,
  destination_ids = c(187071108930, 1142958119078556),
  age_min = 15,
  age_max = 65,
  genders = 2,
  countries = "MX",
  publisher_platforms = c(
    "facebook",
    "instagram",
    #' audience_network',
    "messenger"
  ),
  # interests_ids = NA,
  facebook_positions = c(
    "feed",
    #' instant_article',
    "marketplace",
    "video_feeds",
    "story",
    "search",
    "instream_video"
  ),
  instagram_positions = c(
    "stream",
    "story",
    "explore"
  ),
  # audience_network_positions = c(
  #  'classic',
  #  'instream_video')
  messenger_positions = c(
    "messenger_home",
    "sponsored_messages",
    "story"
  ),
  device_platforms = c(
    "mobile",
    "desktop"
  )
)
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_token}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_insights}
\alias{fb_insights}
\title{Facebook Insights API}
\usage{
fb_insights(
  token,
  which,
  start_date = Sys.Date() - 7,
  end_date = Sys.Date(),
  time_increment = "1",
  report_level = "campaign",
  ad_object = "insights",
  breakdowns = NA,
  fields = NA,
  filtering = NULL,
  limit = 100,
  api_version = NULL,
  process = TRUE,
  async = FALSE,
  ...
)
}
\arguments{
\item{token}{Character. Valid access token with sufficient privileges. Visit the
\href{https://developers.facebook.com/tools/explorer}{Facebook API Graph Explorer}
to acquire one.}

\item{which}{Character vector. This is the accounts, campaigns, adsets,
or ads IDs to be queried. Remember: if \code{report_level = "account"}, you must
start the ID with \code{act_}.}

\item{start_date, end_date}{Character. The first and last full day to report, in the
format \code{"YYYY-MM-DD"}.}

\item{time_increment}{Character. Group by months ("monthly"),
everything together ("all_days") or an integer per days [1-90].
Default: each day separately (i.e. "1").}

\item{report_level}{Character. One of "ad", "adset", "campaign", or "account"}

\item{ad_object}{Character. One of: "insights" (default), "adsets", ...}

\item{breakdowns}{Character Vector. One or more of breakdowns for
segmentation results. Set to NA for no breakdowns}

\item{fields}{Character, json format. Leave \code{NA} for default fields OR
\code{NULL} to ignore.}

\item{filtering}{List. Each filter will be a list containing "field",
"operator", and "value". Read more about the operators in the official
\href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights}{docs}.
Example: \code{dplyr::tibble(field = "country", operator = "IN", value = list("PE")))}.}

\item{limit}{Integer. Query limit by pagination.}

\item{api_version}{Character. Facebook API version.}

\item{process}{Boolean. Process GET results to a more friendly format?}

\item{async}{Boolean. Run an async query. When set to \code{TRUE}, instead of making
a GET query, it'll run a POST query and will return a report run ID.}

\item{...}{Additional parameters.}
}
\value{
data.frame with un-nested processed results if \code{process=TRUE} or
raw API results as list when \code{process=FALSE}.
}
\description{
This returns all available FB insights per day including any given
breakdown to the specified report level, and place into a data frame.
For more information on Ad Insights' API, go to the original
\href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{documentaion}.
}
\examples{
\dontrun{
token <- "YOURTOKEN"
which <- "act_20846447"

# Platforms' Insights: all ad-sets platforms of "which" account,
# aggregated, for the last 30 days
platforms <- fb_insights(
  token, which,
  start_date = Sys.Date() - 30,
  time_increment = "all_days",
  report_level = "adset",
  fields = c(
    "account_name",
    "adset_id",
    "adset_start",
    "adset_end"
  ),
  breakdowns = c(
    "publisher_platform",
    "platform_position",
    "device_platform"
  )
)

# Daily results for all campaigns of "which" account,
# with custom performance fields with no breakdowns.
insights_adset <- fb_insights(
  token, which,
  time_increment = "1",
  report_level = "campaign",
  fields = c(
    "adset_id",
    "reach",
    "frequency",
    "spend",
    "cpm",
    "objective",
    "optimization_goal"
  )
)
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_cuts_error}
\alias{mplot_cuts_error}
\title{Cuts by quantiles on absolute and percentual errors plot}
\usage{
mplot_cuts_error(
  tag,
  score,
  splits = 10,
  title = NA,
  model_name = NA,
  save = FALSE,
  subdir = NA,
  file_name = "viz_ncuts_error.png"
)
}
\arguments{
\item{tag}{Vector. Real known label.}

\item{score}{Vector. Predicted value or model's result.}

\item{splits}{Integer. Number of separations to plot}

\item{title}{Character. Title to show in plot}

\item{model_name}{Character. Model's name}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}
}
\value{
Plot with error results by cuts.
}
\description{
This function cuts by quantiles on absolute and percentual errors
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dfr) # Results for AutoML Predictions
head(dfr$regr)
mplot_cuts_error(dfr$regr$tag, dfr$regr$score,
  model_name = "Titanic Fare Model"
)
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_response}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outliers.R
\name{outlier_zscore}
\alias{outlier_zscore}
\title{Outliers: Z-score method}
\usage{
outlier_zscore(x, thresh = 3, mad = FALSE)
}
\arguments{
\item{x}{Numeric. Distribution}

\item{thresh}{Numeric. Z-Score threshold for n standard deviations.}

\item{mad}{Boolean. Use median absolute deviation instead?}
}
\value{
data.frame. Each row is an \code{x} observation with its
respective std/mean or mad/med calculations depending on \code{mad} input.
}
\description{
Z-score, also called a standard score, of an observation is a
distance from the population center measured in number of normalization
units. The default choice for center is sample mean and for
normalization unit is standard deviation. Values are considered
outliers based on z-score if its absolute value of
default z-score is higher then the threshold (popular choice is 3).
}
\seealso{
Other Outliers: 
\code{\link{outlier_tukey}()},
\code{\link{outlier_zscore_plot}()},
\code{\link{winsorize}()}
}
\concept{Outliers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{chr2num}
\alias{chr2num}
\alias{chr2logical}
\alias{chr2date}
\title{Check character values for date/numeric/logical and change datatype}
\usage{
chr2num(data)

chr2logical(data)

chr2date(data)
}
\arguments{
\item{data}{Vector, data.frame or list}
}
\description{
Automatically check a vector, data.frame or list for numeric, logical,
date content and change their datatype. Note that factors are skipped in
case the user requires character numeric values to be kept as they are.
}
\examples{
str(chr2num(c("1", "2", "3")))
df <- data.frame(A = c("1", "3"), B = c("A", "B"), c = c(pi, pi * 2))
str(chr2num(df))
lst <- list(A = c("1", "2", "3"), B = c("A", "B", "3"), C = pi, D = 3L)
str(chr2num(lst))
lst2 <- list(layer1 = ":D", layer2 = lst)
str(chr2num(lst2))
str(chr2logical(c(NA, "true", FALSE)))
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/x2y.R
\name{x2y}
\alias{x2y}
\alias{x2y_metric}
\alias{plot.x2y_preds}
\alias{plot.x2y}
\alias{x2y_preds}
\title{Ranked Predictive Power of Cross-Features (x2y)}
\usage{
x2y(
  df,
  target = NULL,
  symmetric = FALSE,
  target_x = FALSE,
  target_y = FALSE,
  plot = FALSE,
  top = 20,
  quiet = "auto",
  ohse = FALSE,
  corr = FALSE,
  ...
)

x2y_metric(x, y, confidence = FALSE, bootstraps = 20, max_cat = 20)

\method{plot}{x2y_preds}(x, corr = FALSE, ...)

\method{plot}{x2y}(x, type = 1, ...)

x2y_preds(x, y, max_cat = 10)
}
\arguments{
\item{df}{data.frame. Note that variables with no variance will be ignored.}

\item{target}{Character vector. If you are only interested in the \code{x2y}
values between particular variable(s) in \code{df}, set
name(s) of the variable(s) you are interested in. Keep \code{NULL}
to calculate for every variable (column). Check \code{target_x} and
\code{target_y} parameters as well.}

\item{symmetric}{Boolean. \code{x2y} metric is not symmetric with respect to
\code{x} and \code{y}. The extent to which \code{x} can predict \code{y} can
be different from the extent to which \code{y} can predict \code{x}. Set
\code{symmetric=TRUE} if you wish to average both numbers.}

\item{target_x, target_y}{Boolean. Force target features to be part of
\code{x} OR \code{y}?}

\item{plot}{Boolean. Return a plot? If not, only a data.frame with calculated
results will be returned.}

\item{top}{Integer. Show/plot only top N predictive cross-features. Set
to \code{NULL} to return all.}

\item{quiet}{Boolean. Keep quiet? If not, show progress bar.}

\item{ohse}{Boolean. Use \code{lares::ohse()} to pre-process the data?}

\item{corr}{Boolean. Add correlation and pvalue data to compare with? For
more custom studies, use \code{lares::corr_cross()} directly.}

\item{...}{Additional parameters passed to \code{x2y_metric()}}

\item{x, y}{Vectors. Categorical or numerical vectors of same length.}

\item{confidence}{Boolean. Calculate 95\% confidence intervals estimated
with N \code{bootstraps}.}

\item{bootstraps}{Integer. If \code{confidence=TRUE}, how many bootstraps?
The more iterations we run the more precise the confidence internal will be.}

\item{max_cat}{Integer. Maximum number of unique \code{x} or \code{y} values
when categorical. Will select then most frequent values and the rest will
be passed as \code{""}.}

\item{type}{Integer. Plot type: \code{1} for tile plot,
\code{2} for ranked bar plot.}
}
\value{
Depending on \code{plot} input, a plot or a data.frame with x2y results.
}
\description{
The relative reduction in error when we go from a baseline model
(average for continuous and most frequent for categorical features) to
a predictive model, can measure the strength of the relationship between
two features. In other words, \code{x2y} measures the ability of \code{x}
to predict \code{y}. We use CART (Classification And Regression Trees) models
to be able to 1) compare numerical and non-numerical features, 2) detect
non-linear relationships, and 3) because they are easy/quick to train.
}
\details{
This \code{x2y} metric is based on Rama Ramakrishnan's
\href{https://bit.ly/3sOVbei}{post}: An Alternative to the Correlation
Coefficient That Works For Numeric and Categorical Variables. This analysis
complements our \code{lares::corr_cross()} output.
}
\examples{
\donttest{
data(dft) # Titanic dataset
x2y_results <- x2y(dft, quiet = TRUE, max_cat = 10, top = NULL)
head(x2y_results, 10)
plot(x2y_results, type = 2)

# Confidence intervals with 10 bootstrap iterations
x2y(dft,
  target = c("Survived", "Age"),
  confidence = TRUE, bootstraps = 10, top = 8
)

# Compare with mean absolute correlations
x2y(dft, "Fare", corr = TRUE, top = 6, target_x = TRUE)

# Plot (symmetric) results
symm <- x2y(dft, target = "Survived", symmetric = TRUE)
plot(symm, type = 1)

# Symmetry: x2y vs y2x
on.exit(set.seed(42))
x <- seq(-1, 1, 0.01)
y <- sqrt(1 - x^2) + rnorm(length(x), mean = 0, sd = 0.05)

# Knowing x reduces the uncertainty about the value of y a lot more than
# knowing y reduces the uncertainty about the value of x. Note correlation.
plot(x2y_preds(x, y), corr = TRUE)
plot(x2y_preds(y, x), corr = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scrabble.R
\name{grepl_letters}
\alias{grepl_letters}
\title{Pattern Matching for Letters considering Blanks}
\usage{
grepl_letters(x, pattern, blank = "_")
}
\arguments{
\item{x}{Character vector}

\item{pattern}{Character. Character string containing a
semi-regular expression which uses the following logic:
"a_b" means any character that contains "a" followed by
something followed by "b", anywhere in the string.}

\item{blank}{Character. String to use between letters.}
}
\value{
Boolean check for each value on \code{x}.
}
\description{
Match pattern of letters considering blanks within each element of a
character vector, allowing counted characters between and around
each letter. Used as an auxiliary function for the Scrabble family
of functions.
}
\examples{
x <- c("aaaa", "bbbb", "baba", "aabb", "a", "ab")
grepl_letters(x, "ab")
grepl_letters(x, "_ab")
grepl_letters(x, "a_a")
grepl_letters(x, "c")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audio.R
\name{mp3_trim}
\alias{mp3_trim}
\title{Trim MP3 Audio File}
\usage{
mp3_trim(
  file,
  start_time = 0,
  end_time = NA,
  overwrite = FALSE,
  ext = "mp3",
  quiet = FALSE
)
}
\arguments{
\item{file}{Character. File name to trim.}

\item{start_time, end_time}{Numeric. Start and end time
to trim the audio output in seconds.}

\item{overwrite}{Boolean. Overwrite original file?}

\item{ext}{Character. File extension/type.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\description{
This function trims MP3 files given a start and/or end numeric
timestamp. Requires \code{ffmpeg} installed in your machine.
}
\seealso{
Other Audio: 
\code{\link{mp3_get}()},
\code{\link{mp3_update_tags}()}
}
\concept{Audio}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frequencies.R
\name{freqs_df}
\alias{freqs_df}
\title{Plot for All Frequencies on Dataframe}
\usage{
freqs_df(
  df,
  max = 0.9,
  min = 0,
  novar = TRUE,
  plot = FALSE,
  top = 30,
  quiet = FALSE,
  save = FALSE,
  subdir = NA
)
}
\arguments{
\item{df}{Data.frame}

\item{max}{Numeric. Top variance threshold. Range: (0-1].
These variables will be excluded}

\item{min}{Numeric. Minimum variance threshold. Range: [0-1).
These values will be grouped into a high frequency (HF) value}

\item{novar}{Boolean. Remove no variance columns?}

\item{plot}{Boolean. Do you want to see a plot? Three variables tops}

\item{top}{Integer. Plot most relevant (less categories) variables}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{save}{Boolean. Save the output plot in our working directory}

\item{subdir}{Character. Into which subdirectory do you wish to
save the plot to?}
}
\value{
Plot when \code{plot=TRUE} and data.frame with grouped frequency results
when \code{plot=FALSE}.
}
\description{
This function lets the user analize data by visualizing the
frequency of each value of each column from a whole data frame.
}
\examples{
data(dft) # Titanic dataset
freqs_df(dft)
freqs_df(dft, plot = TRUE)
}
\seealso{
Other Frequency: 
\code{\link{freqs}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()}

Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}

Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}
\concept{Frequency}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dalex_explainers.R
\name{dalex_residuals}
\alias{dalex_residuals}
\title{DALEX Residuals}
\usage{
dalex_residuals(explainer)
}
\arguments{
\item{explainer}{Object. Result from h2o_explainer function}
}
\value{
Plot. Based of \code{explainer} residual results.
}
\description{
DALEX function for residuals
}
\seealso{
Other Interpretability: 
\code{\link{dalex_local}()},
\code{\link{dalex_variable}()},
\code{\link{h2o_explainer}()}
}
\concept{Interpretability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_ads}
\alias{fb_ads}
\title{Facebook Ads API}
\usage{
fb_ads(
  token,
  which,
  start_date = Sys.Date() - 31,
  end_date = Sys.Date(),
  fields = NA,
  api_version = NULL,
  process = TRUE,
  ...
)
}
\arguments{
\item{token}{Character. Valid access token with sufficient privileges. Visit the
\href{https://developers.facebook.com/tools/explorer}{Facebook API Graph Explorer}
to acquire one.}

\item{which}{Character vector. This is the accounts, campaigns, adsets,
or ads IDs to be queried. Remember: if \code{report_level = "account"}, you must
start the ID with \code{act_}.}

\item{start_date, end_date}{Character. The first and last full day to report, in the
format \code{"YYYY-MM-DD"}.}

\item{fields}{Character, json format. Leave \code{NA} for default fields OR
\code{NULL} to ignore.}

\item{api_version}{Character. Facebook API version.}

\item{process}{Boolean. Process GET results to a more friendly format?}

\item{...}{Additional parameters.}
}
\value{
data.frame with un-nested processed results if \code{process=TRUE} or
raw API results as list when \code{process=FALSE}.
}
\description{
This returns all available FB ads for any account, campaign, or ad set id.
For more information on Ad' API, go to the
\href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/reference/adgroup/}{original documentaion}
}
\details{
This function was based on FBinsightsR.
}
\examples{
\dontrun{
token <- YOURTOKEN
account <- act_ADACCOUNT

# Query all ads for "which" (account) with results in the last 10 days
ads <- fb_ads(token, account, start_date = Sys.Date() - 10)
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_accounts}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correlations.R
\name{corr}
\alias{corr}
\title{Correlation table}
\usage{
corr(
  df,
  method = "pearson",
  use = "pairwise.complete.obs",
  pvalue = FALSE,
  padjust = NULL,
  half = FALSE,
  dec = 6,
  ignore = NULL,
  dummy = TRUE,
  redundant = NULL,
  logs = FALSE,
  limit = 10,
  top = NA,
  ...
)
}
\arguments{
\item{df}{Dataframe. It doesn't matter if it's got non-numerical
columns: they will be filtered.}

\item{method}{Character. Any of: c("pearson", "kendall", "spearman").}

\item{use}{Character. Method for computing covariances in the presence
of missing values. Check \code{stats::cor} for options.}

\item{pvalue}{Boolean. Returns a list, with correlations and statistical
significance (p-value) for each value.}

\item{padjust}{Character. NULL to skip or any of \code{p.adjust.methods} to
calculate adjust p-values for multiple comparisons using \code{p.adjust()}.}

\item{half}{Boolean. Return only half of the matrix? The redundant
symmetrical correlations will be \code{NA}.}

\item{dec}{Integer. Number of decimals to round correlations and p-values.}

\item{ignore}{Vector or character. Which column should be ignored?}

\item{dummy}{Boolean. Should One Hot (Smart) Encoding (\code{ohse()})
be applied to categorical columns?}

\item{redundant}{Boolean. Should we keep redundant columns? i.e. If the
column only has two different values, should we keep both new columns?
Is set to \code{NULL}, only binary variables will dump redundant columns.}

\item{logs}{Boolean. Calculate log(x)+1 for numerical columns?}

\item{limit}{Integer. Limit one hot encoding to the n most frequent
values of each column. Set to \code{NA} to ignore argument.}

\item{top}{Integer. Select top N most relevant variables? Filtered
and sorted by mean of each variable's correlations.}

\item{...}{Additional parameters passed to \code{ohse}, \code{corr},
and/or \code{cor.test}.}
}
\value{
data.frame. Squared dimensions (N x N) to match every
correlation between every \code{df} data.frame column/variable. Notice
that when using \code{ohse()} you may get more dimensions.
}
\description{
This function correlates a whole dataframe, running one hot smart
encoding (\code{ohse}) to transform non-numerical features.
Note that it will automatically suppress columns
with less than 3 non missing values and warn the user.
}
\examples{
data(dft) # Titanic dataset
df <- dft[, 2:5]

# Correlation matrix (without redundancy)
corr(df, half = TRUE)

# Ignore specific column
corr(df, ignore = "Pclass")

# Calculate p-values as well
corr(df, pvalue = TRUE, limit = 1)

# Test when no more than 2 non-missing values
df$trash <- c(1, rep(NA, nrow(df) - 1))
# and another method...
corr(df, method = "spearman")
}
\seealso{
Other Calculus: 
\code{\link{corr_cross}()},
\code{\link{dist2d}()},
\code{\link{model_metrics}()},
\code{\link{quants}()}

Other Correlations: 
\code{\link{corr_cross}()},
\code{\link{corr_var}()}
}
\concept{Calculus}
\concept{Correlations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_report_check}
\alias{fb_report_check}
\title{Facebook API Report Status Check}
\usage{
fb_report_check(
  token,
  report_run_id,
  api_version = NULL,
  live = FALSE,
  sleep = 10,
  quiet = FALSE
)
}
\arguments{
\item{token}{Character. Valid access token with sufficient privileges. Visit the
\href{https://developers.facebook.com/tools/explorer}{Facebook API Graph Explorer}
to acquire one.}

\item{report_run_id}{Integer. Report ID to check status.}

\item{api_version}{Character. Facebook API version.}

\item{live}{Boolean. Run until status report is finished?}

\item{sleep}{Boolean. If \code{live=TRUE}, then how many seconds should
we wait until next check?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
List with API status results.
}
\description{
This returns all available FB insights per day including any given
breakdown to the specified report level, and place into a data frame.
For more information on Ad Insights' API, go to the original
\href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{documentaion}.
}
\examples{
\dontrun{
token <- "YOURTOKEN"
report_run_id <- "123456789"
fb_report_check(token, report_run_id, live = TRUE, quiet = FALSE)
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stocks.R
\name{splot_summary}
\alias{splot_summary}
\alias{splot_change}
\alias{splot_growth}
\alias{splot_divs}
\alias{splot_roi}
\alias{splot_types}
\alias{splot_etf}
\title{Investing Portfolio Reporting Plots}
\usage{
splot_summary(p, s, save = FALSE)

splot_change(
  p,
  s,
  rel = TRUE,
  group = FALSE,
  n_days = 365,
  keep_old = FALSE,
  save = FALSE
)

splot_growth(p, save = FALSE)

splot_divs(p, type = 1)

splot_roi(p, n_days = 365, historical = TRUE, ma = c(12, 50), save = FALSE)

splot_types(s, save = FALSE)

splot_etf(s, keep_all = FALSE, cache = TRUE, save = FALSE)
}
\arguments{
\item{p}{Dataframe. Result from \code{daily_portfolio()}}

\item{s}{Dataframe. Result from \code{daily_stocks()}}

\item{save}{Boolean. Save plot into a local file?}

\item{rel}{Boolean. Relative delta values (weighted with portfolio)?
If not, absolute monetary delta values.}

\item{group}{Boolean. Group stocks by stocks type?}

\item{n_days}{Integer. How many days back you want to see?}

\item{keep_old}{Boolean. Include sold tickers even though not currently in portfolio?}

\item{type}{Integer. Typo of plot. 1 for incomes.}

\item{historical}{Boolean. Historical ROI metric? If not, ROI
will be calculated locally for n_days parameter}

\item{ma}{Numeric Vector. Select 2 values for moving averages.
Set to NA to turn this metric off}

\item{keep_all}{Boolean. Keep "Not Known / Not ETF"?}

\item{cache}{Boolean. Use daily cache if available?}
}
\value{
ggplot object
}
\description{
\code{splot_summary()} plots a summary for the whole portfolio, showing
how much have you invested, how much has each ticker changed, etc.

\code{splot_change()} plots each stock's change through history, since
inception, with weighted attributions or absolute values.

\code{splot_growth()} plots your portfolio's growth, in cash and investment,
since inception.

\code{stocks_plots()} plots a portfolio's historical dividends incomes
grouped by quarter an year.

\code{splot_roi()} plots a portfolio's historical ROI since inception
or since last n days, with 2 moving average lines.

\code{splot_types()} lets the user plot types or categories of tickers.

\code{splot_etf()} lets the user plot his portfolio's distribution,
specifically ETF's sectors.
}
\seealso{
Other Investment: 
\code{\link{etf_sector}()},
\code{\link{stocks_file}()},
\code{\link{stocks_quote}()}

Other Scrapper: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{holidays}()},
\code{\link{ip_data}()},
\code{\link{mp3_get}()},
\code{\link{readGS}()},
\code{\link{stocks_quote}()}
}
\concept{Investment}
\concept{Scrapper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dont_sleep.R
\name{dont_sleep}
\alias{dont_sleep}
\alias{dont_sleep_time}
\title{Prevent Computer from Sleeping by Simulating Mouse Activity}
\usage{
dont_sleep(sec_range = c(20, 60), off_time = 18.5, click = TRUE, quiet = FALSE)

dont_sleep_time(quiet = FALSE)
}
\arguments{
\item{sec_range}{Numeric vector of length 2. Range (in seconds) of
random intervals between mouse movements and clicks. Default is \code{c(20, 60)}.}

\item{off_time}{Numeric. Decimal hour (24h format) to stop the function
automatically, e.g. 18.5 means 18:30 (6:30 PM). Default is \code{18.5}.}

\item{click}{Logical. If \code{TRUE}, clicks the mouse. Default is \code{TRUE}.}

\item{quiet}{Logical. If \code{TRUE}, suppresses progress messages.
Default is \code{FALSE}.}
}
\value{
Invisibly returns \code{NULL} when the function exits.
}
\description{
Keeps the computer awake by moving the mouse cursor and clicking
periodically with a human-like behavior. It automatically detects
the operating system (Windows or macOS) and runs the appropriate
underlying logic.
}
\details{
The cursor initially moves to the upper-left corner (0,0), then
performs small random movements and clicks periodically to avoid
system sleep. The function exits if the user moves the mouse away
from the corner or when the specified off-time is reached.

Requires \code{cliclick} installed on macOS (install via Homebrew:
\code{brew install cliclick}), and the \code{KeyboardSimulator}
package on Windows.
}
\examples{
\dontrun{
# Keep the computer awake until 8:00 PM, with mouse actions every 30 to 60 seconds
dont_sleep(sec_range = c(30, 60), off_time = 20)
# Summary of amount of daily time running function
dont_sleep_time()
}
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confidence.R
\name{ci_var}
\alias{ci_var}
\title{Confidence Intervals on Dataframe}
\usage{
ci_var(df, var, group_var = NULL, conf = 0.95)
}
\arguments{
\item{df}{Dataframe}

\item{var}{Variable name. Must be a numerical column.}

\item{group_var}{Variable name. Group results by another variable.}

\item{conf}{Numeric. Confidence level (0-1).}
}
\value{
data.frame mean, standard deviation, counter, upper and lower CIs.
}
\description{
Calculate confidence intervals for a continuous numerical column on
a dataframe, given a confidence level. You may also group results
using another variable. Tidyverse friendly.
}
\examples{
data(dft) # Titanic dataset
ci_var(dft, Fare)
ci_var(dft, Fare, Pclass)
ci_var(dft, Fare, Pclass, conf = 0.99)
}
\seealso{
Other Confidence: 
\code{\link{ci_lower}()}
}
\concept{Confidence}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_metrics.R
\name{ROC}
\alias{ROC}
\title{AUC and ROC Curves Data}
\usage{
ROC(tag, score, multis = NA)
}
\arguments{
\item{tag}{Vector. Real known label}

\item{score}{Vector. Predicted value or model's result}

\item{multis}{Data.frame. Containing columns with each category score
(only used when more than 2 categories coexist)}
}
\value{
List with ROC's results, area under the curve (AUC) and their CI.
}
\description{
This function calculates ROC Curves and AUC values with 95\% confidence
range. It also works for multi-categorical models.
}
\section{Plot Results}{

To plot results, use the \code{mplot_roc()} function.
}

\examples{
data(dfr) # Results for AutoML Predictions
lapply(dfr[c(1, 2)], head)

# ROC Data for Binomial Model
roc1 <- ROC(dfr$class2$tag, dfr$class2$scores)
lapply(roc1, head)

# ROC Data for Multi-Categorical Model
roc2 <- ROC(dfr$class3$tag, dfr$class3$score,
  multis = subset(dfr$class3, select = -c(tag, score))
)
lapply(roc2, head)
}
\seealso{
Other Machine Learning: 
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Model metrics: 
\code{\link{conf_mat}()},
\code{\link{errors}()},
\code{\link{gain_lift}()},
\code{\link{loglossBinary}()},
\code{\link{model_metrics}()}
}
\concept{Machine Learning}
\concept{Model metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{file_name}
\alias{file_name}
\alias{file_type}
\title{Extract file raw name and type from file names}
\usage{
file_name(filepath)

file_type(filepath)
}
\arguments{
\item{filepath}{Character vector. File path(s) to get file raw names
without extension nor path OR extension without path nor raw name.}
}
\description{
Extract file raw name and type from file names

Get file extensions without file names
}
\examples{
file_name("file.aux")
file_name("temp/file.R")
file_name("/temp/temp3/music.mp3")
file_type("file.aux")
file_type("temp/file.R")
file_type("/temp/temp3/music.mp3")
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_conf}
\alias{mplot_conf}
\title{Confussion Matrix Plot}
\usage{
mplot_conf(
  tag,
  score,
  thresh = 0.5,
  abc = TRUE,
  squared = FALSE,
  diagonal = TRUE,
  top = 20,
  subtitle = NA,
  model_name = NULL,
  save = FALSE,
  subdir = NA,
  file_name = "viz_conf_mat.png"
)
}
\arguments{
\item{tag}{Vector. Real known label.}

\item{score}{Vector. Predicted value or model's result.}

\item{thresh}{Integer. Threshold for selecting binary or regression
models: this number is the threshold of unique values we should
have in \code{'tag'} (more than: regression; less than: classification)}

\item{abc}{Boolean. Arrange columns and rows alphabetically?}

\item{squared}{Boolean. Force plot to be squared?}

\item{diagonal}{Boolean. \code{FALSE} to convert diagonal numbers to
zeroes. Ideal to detect must confusing categories.}

\item{top}{Integer. Plot only the most n frequent variables.
Set to \code{NA} to plot all.}

\item{subtitle}{Character. Subtitle to show in plot}

\item{model_name}{Character. Model's name}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}
}
\value{
Plot with confusion matrix results.
}
\description{
This function plots a confussion matrix.
}
\details{
You may use \code{conf_mat()} to get calculate values.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dfr) # Results for AutoML Predictions
lapply(dfr, head)

# Plot for Binomial Model
mplot_conf(dfr$class2$tag, dfr$class2$scores,
  model_name = "Titanic Survived Model"
)

# Plot for Multi-Categorical Model
mplot_conf(dfr$class3$tag, dfr$class3$score,
  model_name = "Titanic Class Model"
)
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_cuts}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_response}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_write}
\alias{cache_write}
\alias{cache_read}
\alias{cache_exists}
\alias{cache_clear}
\alias{cache_pipe}
\title{Cache Save and Load (Write and Read)}
\usage{
cache_write(
  data,
  base = "temp",
  cache_dir = getOption("LARES_CACHE_DIR"),
  ask = FALSE,
  overwrite = TRUE,
  quiet = FALSE,
  ...
)

cache_read(
  base,
  cache_dir = getOption("LARES_CACHE_DIR"),
  ask = FALSE,
  overwrite = TRUE,
  quiet = FALSE,
  ...
)

cache_exists(base = NULL, cache_dir = getOption("LARES_CACHE_DIR"), ...)

cache_clear(cache_dir = getOption("LARES_CACHE_DIR"), quiet = FALSE, ...)

cache_pipe(data, base = "cache_pipe", read = TRUE, write = TRUE, ...)
}
\arguments{
\item{data}{Object}

\item{base}{Character vector. Unique name for your cache file. You can pass
a character vector with multiple elements that will be concatenated.
All cache files with start with \code{lares_cache_*} automatically to
quickly detect these cache files.}

\item{cache_dir}{Character. Where do you want to save you cache files?
By default they'll be stored on \code{tempdir()} but you can change it
using this parameter or setting a global option called \code{"LARES_CACHE_DIR"}.}

\item{ask}{Boolean. If cache exists, when reading: (interactive) ask the user
if the cache should be used to proceed or ignored; when writing, (interactive)
ask the user if the cache should be overwritten. Note that you can only ask for
one cache file at a time because vectors are concatenated.}

\item{overwrite}{Boolean. Set to overwrite existing cache file. When reading,
this parameter answers to ask prompt instead.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}

\item{read, write}{Boolean. Do you wish to read and or write cache?}
}
\value{
\code{cache_write}. No return value, called for side effects.

\code{cache_read}. R object. Data from cache file or NULL if no cache found.

\code{cache_exists}. Boolean. Result of \code{base} existence.

\code{cache_clear}. Invisible vector containing cache file names removed.

\code{cache_pipe}. Same as \code{x} or cached result.
}
\description{
This function lets the user save and load a cache of any R object to
improve timings and UX.
}
\examples{
x <- list(a = 1, b = 2:4)
base <- c(as.character(Sys.Date()), "A", "B")
cache_write(x, base)
cache_read(base, ask = FALSE)
cache_exists(base)
cache_clear()

# Use cache_pipe() for automatic use (saves and loads if exists)
cache_pipe({
  x <- 1
  x + 1
})
cache_pipe({
  x <- 1
  x + 1
})
}
\concept{Cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_lares.R
\name{gg_fill_customs}
\alias{gg_fill_customs}
\alias{gg_colour_customs}
\alias{gg_text_customs}
\alias{gg_vals}
\title{Custom fill, colour and text colours for ggplot2}
\usage{
gg_fill_customs(column = "fill", ...)

gg_colour_customs(column = "colour", ...)

gg_text_customs(column = "colour", ...)

gg_vals(layer = "fill", column = layer, cols = NULL, ...)
}
\arguments{
\item{column}{Character. Select any of "fill" or "colour" to use on
your \code{lares_pal()$labels} palette.}

\item{...}{Allow additional parameters not used.}

\item{layer}{Character. Select any of "fill", "colour", or "label" to get the
layer containing the colours labels you wish to colour.}

\item{cols}{Data.frame. Customize colour palette with a data.frame. Must
contain values, fill, and colour columns.}
}
\value{
Same as \code{scale_fill_manual} or \code{scale_colour_manual}
but with custom palette applied.
}
\description{
This function lets the user use pre-defined default colours.
Check your \code{lares_pal()$labels} scale. Feel free to use
\code{gg_vals()} to debug colours used in latest plot.
}
\details{
Notice that when the layer defined is any of GeomPoint, GeomLine,
GeomText or GeomLabel, \code{gg_colour_customs()} will force
\code{column = "fill"} parameter.
}
\examples{
library("ggplot2")
# Generic plot function to run examples to
run_plot <- function(add_fxs = TRUE) {
  p <- data.frame(station = c("spring", "summer", "fall", "winter"), num = 1:4) \%>\%
    ggplot(aes(x = station, y = num, fill = station)) +
    geom_col() +
    geom_text(aes(y = 0.5, label = num, colour = station), size = 6)
  if (add_fxs) p <- p + gg_fill_customs() + gg_colour_customs()
  return(p)
}
# Default colours
run_plot()
# Check last colours used
gg_vals("fill", "fill")
gg_vals("colour", "colour")
# Change any default colour
options("lares.colours.custom" = data.frame(
  values = c("summer", "winter"),
  fill = c("pink", "black"),
  colour = c("black", "white")
))
run_plot()
# Check last colours used
gg_vals("fill", "fill")
gg_vals("colour", "colour")
# Reset to default colours
options("lares.colours.custom" = NULL)
# Notice you can use 'pal = 4' argument on theme_lares() too
run_plot(add_fxs = FALSE) + theme_lares(pal = 4)
}
\seealso{
Other Themes: 
\code{\link{lares_pal}()},
\code{\link{plot_palette}()},
\code{\link{theme_lares}()}
}
\concept{Themes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dalex_explainers.R
\name{dalex_local}
\alias{dalex_local}
\title{DALEX Local}
\usage{
dalex_local(explainer, observation = NA, row = 1, type = "break_down")
}
\arguments{
\item{explainer}{Object. Result from h2o_explainer function}

\item{observation}{Data.frame. If you want to use an observation
that was not in the original explainer function, add here. Else, use row}

\item{row}{Dataframe. Row number from the data.frame used in explainer.}

\item{type}{Character. The type of variable attributions.
Either shap, oscillations, break_down or break_down_interactions.}
}
\value{
List. Containing observation, breakdown results, and breakdown plot.
}
\description{
DALEX function for local interpretations
}
\seealso{
Other Interpretability: 
\code{\link{dalex_residuals}()},
\code{\link{dalex_variable}()},
\code{\link{h2o_explainer}()}
}
\concept{Interpretability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_preprocessing.R
\name{model_preprocess}
\alias{model_preprocess}
\title{Automate Data Preprocess for Modeling}
\usage{
model_preprocess(
  df,
  y = "tag",
  ignore = NULL,
  train_test = NA,
  split = 0.7,
  weight = NULL,
  target = "auto",
  balance = FALSE,
  impute = FALSE,
  no_outliers = TRUE,
  unique_train = TRUE,
  center = FALSE,
  scale = FALSE,
  thresh = 10,
  seed = 0,
  quiet = FALSE
)
}
\arguments{
\item{df}{Dataframe. Dataframe containing all your data, including
the dependent variable labeled as \code{'tag'}. If you want to define
which variable should be used instead, use the \code{y} parameter.}

\item{y}{Character. Column name for dependent variable or response.}

\item{ignore}{Character vector. Force columns for the model to ignore}

\item{train_test}{Character. If needed, \code{df}'s column name with 'test'
and 'train' values to split}

\item{split}{Numeric. Value between 0 and 1 to split as train/test
datasets. Value is for training set. Set value to 1 to train with all
available data and test with same data (cross-validation will still be
used when training). If \code{train_test} is set, value will be overwritten
with its real split rate.}

\item{weight}{Column with observation weights. Giving some observation a
weight of zero is equivalent to excluding it from the dataset; giving an
observation a relative weight of 2 is equivalent to repeating that
row twice. Negative weights are not allowed.}

\item{target}{Value. Which is your target positive value? If
set to \code{'auto'}, the target with largest \code{mean(score)} will be
selected. Change the value to overwrite. Only used when binary
categorical model.}

\item{balance}{Boolean. Auto-balance train dataset with under-sampling?}

\item{impute}{Boolean. Fill \code{NA} values with MICE?}

\item{no_outliers}{Boolean/Numeric. Remove \code{y}'s outliers from the dataset?
Will remove those values that are farther than n standard deviations from
the dependent variable's mean (Z-score). Set to \code{TRUE} for default (3)
or numeric to set a different multiplier.}

\item{unique_train}{Boolean. Keep only unique row observations for training data?}

\item{center, scale}{Boolean. Using the base function scale, do you wish
to center and/or scale all numerical values?}

\item{thresh}{Integer. Threshold for selecting binary or regression
models: this number is the threshold of unique values we should
have in \code{'tag'} (more than: regression; less than: classification)}

\item{seed}{Integer. Set a seed for reproducibility. AutoML can only
guarantee reproducibility if max_models is used because max_time is
resource limited.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
List. Contains original data.frame \code{df}, an index
to identify which observations with be part of the train dataset
\code{train_index}, and which model type should be \code{model_type}.
}
\description{
Pre-process your data before training a model. This is the prior step
on the \code{h2o_automl()} function's pipeline. Enabling for
other use cases when wanting too  use any other framework, library,
or custom algorithm.
}
\examples{
data(dft) # Titanic dataset

model_preprocess(dft, "Survived", balance = TRUE)

model_preprocess(dft, "Fare", split = 0.5, scale = TRUE)

model_preprocess(dft, "Pclass", ignore = c("Fare", "Cabin"))

model_preprocess(dft, "Pclass", quiet = TRUE)
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{msplit}()}
}
\concept{Machine Learning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{textCloud}
\alias{textCloud}
\title{Wordcloud Plot}
\usage{
textCloud(
  text,
  lang = "english",
  exclude = NULL,
  seed = 0,
  keep_spaces = FALSE,
  min = 2,
  pal = NA,
  print = TRUE
)
}
\arguments{
\item{text}{Character vector}

\item{lang}{Character. Language in text (used for stop words)}

\item{exclude}{Character vector. Which word do you wish to exclude?}

\item{seed}{Numeric. Seed for re-producible plots}

\item{keep_spaces}{Boolean. If you wish to keep spaces in each line
to keep unique compount words, separated with spaces, set to TRUE.
For example, 'LA ALAMEDA' will be set as 'LA_ALAMEDA' and treated as
a single word.}

\item{min}{Integer. Words with less frequency will not be plotted}

\item{pal}{Character vector. Which colours do you wish to use}

\item{print}{Boolean. Plot results as textcloud?}
}
\value{
wordcloud plot object
}
\description{
Study the distribution of a target variable vs another variable. This
function is quite similar to the funModeling's corrplot function.
}
\seealso{
Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{ngrams}()},
\code{\link{remove_stopwords}()},
\code{\link{replaceall}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{topics_rake}()}
}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{zerovar}
\alias{zerovar}
\title{Zero Variance Columns}
\usage{
zerovar(df)
}
\arguments{
\item{df}{Dataframe}
}
\value{
Character vector with column names on which its values have no variance.
}
\description{
This function detects which columns have the same value (whichever)
for each column.
}
\examples{
df <- data.frame(a = c(1, NA, 3), b = rep(NA, 3), c = rep(5, 3))
print(df)
zerovar(df)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/other_plots.R
\name{plot_chord}
\alias{plot_chord}
\title{Chords Plot}
\usage{
plot_chord(
  origin,
  dest,
  weight = 1,
  mg = 3,
  title = "Chord Diagram",
  subtitle = "",
  pal = NA
)
}
\arguments{
\item{origin, dest}{Vectors. Origin and destination vectors}

\item{weight}{Vector. Weight for each chord.}

\item{mg}{Numeric. Margin adjust for plot in case of need}

\item{title}{Character. Title for the plot}

\item{subtitle}{Character. Subtitle for the plot}

\item{pal}{Vector. Colour pallete. Order matters.}
}
\value{
chordDiagram object
}
\description{
This auxiliary function plots discrete and continuous values results
}
\examples{
# You must have "circlize" library to use this auxiliary function:
\dontrun{
df <- data.frame(from = c(1, 1, 2, 3, 4, 1, 6), to = c(4, 4, 4, 2, 2, NA, NA))
plot_chord(df$from, df$to)
}
}
\seealso{
Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lares.R
\name{install_recommended}
\alias{install_recommended}
\title{Install/Update Additional Recommended Libraries}
\usage{
install_recommended(progress = TRUE, all = FALSE)
}
\arguments{
\item{progress}{Boolean. Show status bar?}

\item{all}{Boolean. All packages? If not, only the ones not installed yet.}
}
\description{
All needed libraries to use (most) lares are already a dependency.
There are some functions that many people won't event know exist
that will require other additional libraries. Also, this may be
used as a Docker way of installing useful libraries on an new instance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_metrics.R
\name{model_metrics}
\alias{model_metrics}
\title{Model Metrics and Performance}
\usage{
model_metrics(
  tag,
  score,
  multis = NA,
  abc = TRUE,
  thresh = 10,
  auto_n = TRUE,
  thresh_cm = 0.5,
  target = "auto",
  type = "test",
  model_name = NA,
  plots = TRUE,
  quiet = FALSE,
  subtitle = NA
)
}
\arguments{
\item{tag}{Vector. Real known label}

\item{score}{Vector. Predicted value or model's result}

\item{multis}{Data.frame. Containing columns with each category score
(only used when more than 2 categories coexist)}

\item{abc}{Boolean. Arrange columns and rows alphabetically
when categorical values?}

\item{thresh}{Integer. Threshold for selecting binary or regression
models: this number is the threshold of unique values we should
have in \code{'tag'} (more than: regression; less than: classification)}

\item{auto_n}{Add \code{n_} before digits when it's categorical and
not numerical, even though seems numerical?}

\item{thresh_cm}{Numeric. Value to splits the results for the
confusion matrix. Range of values: (0-1)}

\item{target}{Value. Which is your target positive value? If
set to \code{'auto'}, the target with largest \code{mean(score)} will be
selected. Change the value to overwrite. Only used when binary
categorical model.}

\item{type}{Character. One of: "train", "test".}

\item{model_name}{Character. Model's name for reference.}

\item{plots}{Boolean. Create plots objects?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{subtitle}{Character. Subtitle for plots}
}
\value{
List. Multiple performance metrics that vary depending on
the type of model (classification or regression). If \code{plot=TRUE},
multiple plots are also returned.
}
\description{
This function lets the user get a confusion matrix and accuracy, and
for for binary classification models: AUC, Precision, Sensitivity, and
Specificity, given the expected (tags) values and predicted values (scores).
}
\examples{
data(dfr) # Results for AutoML Predictions
lapply(dfr, head)

# Metrics for Binomial Model
met1 <- model_metrics(dfr$class2$tag, dfr$class2$scores,
  model_name = "Titanic Survived Model",
  plots = FALSE
)
print(met1)

# Metrics for Multi-Categorical Model
met2 <- model_metrics(dfr$class3$tag, dfr$class3$score,
  multis = subset(dfr$class3, select = -c(tag, score)),
  model_name = "Titanic Class Model",
  plots = FALSE
)
print(met2)

# Metrics for Regression Model
met3 <- model_metrics(dfr$regr$tag, dfr$regr$score,
  model_name = "Titanic Fare Model",
  plots = FALSE
)
print(met3)
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Model metrics: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{errors}()},
\code{\link{gain_lift}()},
\code{\link{loglossBinary}()}

Other Calculus: 
\code{\link{corr}()},
\code{\link{corr_cross}()},
\code{\link{dist2d}()},
\code{\link{quants}()}
}
\concept{Calculus}
\concept{Machine Learning}
\concept{Model metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stocks.R
\name{etf_sector}
\alias{etf_sector}
\title{ETF's Sectors Breakdown}
\usage{
etf_sector(etf = "VTI", quiet = FALSE, cache = TRUE)
}
\arguments{
\item{etf}{Character Vector. Which ETFs you wish to scrap?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{cache}{Boolean. Use daily cache if available?}
}
\value{
data.frame with ETF break.down data by sector
}
\description{
\code{etf_sector()} scraps etf.com data for sector breakdown on ETFs.
Use \code{splot_etf()} for visualization.
}
\examples{
\donttest{
etf_sector(etf = "VTI")
}
}
\seealso{
Other Investment: 
\code{\link{splot_summary}()},
\code{\link{stocks_file}()},
\code{\link{stocks_quote}()}
}
\concept{Investment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataframe_str.R
\name{df_str}
\alias{df_str}
\title{Dataset columns and rows structure}
\usage{
df_str(df, return = "plot", subtitle = NA, quiet = FALSE)
}
\arguments{
\item{df}{Dataframe}

\item{return}{Character. Return "skimr" for skim report, "numbers" for
stats and numbers, "names" for a list with the column names of each of
the class types, "plot" for a nice plot with "numbers" output, "distr"
for an overall summary plot showing categorical, numeric, and missing
values by using \code{plot_df}
distributions}

\item{subtitle}{Character. Add subtitle to plot}

\item{quiet}{Boolean. Keep quiet or show other options available?}
}
\value{
Depending on \code{return} input and based on your \code{df} structure:
\itemize{
  \item \code{list} with the names of the columns classified by class
  \item \code{data.frame} with numbers: total values, row, columns,
  complete rows
  \item \code{plot} with visualizations
}
}
\description{
This function lets the user to check quickly the structure of a
dataset (data.frame). It returns multiple counters for useful metrics,
a plot, and a list of column names for each of the column metrics.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset
df_str(dft, "names")
df_str(dft, "numbers", quiet = TRUE)
df_str(dft, "plot", quiet = TRUE)
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correlations.R
\name{corr_cross}
\alias{corr_cross}
\title{Ranked cross-correlation across all variables}
\usage{
corr_cross(
  df,
  plot = TRUE,
  pvalue = TRUE,
  max_pvalue = 1,
  type = 1,
  max = 1,
  top = 20,
  local = 1,
  ignore = NULL,
  contains = NA,
  grid = TRUE,
  rm.na = FALSE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{data.frame}

\item{plot}{Boolean. Show and return a plot?}

\item{pvalue}{Boolean. Returns a list, with correlations and statistical
significance (p-value) for each value.}

\item{max_pvalue}{Numeric. Filter non-significant variables. Range (0, 1]}

\item{type}{Integer. Plot type. 1 is for overall rank. 2 is for local rank.}

\item{max}{Numeric. Maximum correlation permitted (from 0 to 1)}

\item{top}{Integer. Return top n results only. Only valid when type = 1. Set
value to NA to use all cross-correlations}

\item{local}{Integer. Label top n local correlations. Only valid when type = 2}

\item{ignore}{Vector or character. Which column should be ignored?}

\item{contains}{Character vector. Filter cross-correlations
with variables that contains certain strings (using any value if vector used).}

\item{grid}{Boolean. Separate into grids?}

\item{rm.na}{Boolean. Remove NAs?}

\item{quiet}{Boolean. Keep quiet? If not, show messages}

\item{...}{Additional parameters passed to \code{corr}}
}
\value{
Depending on input \code{plot}, we get correlation and p-value results for
every combination of features, arranged by descending absolute correlation value,
with a data.frame \code{plot = FALSE} or plot \code{plot = TRUE}.
}
\description{
This function creates a correlation full study and returns a rank
of the highest correlation variables obtained in a cross-table.
}
\details{
For a detailed tutorial on using this function, see:
\href{https://www.r-bloggers.com/2019/08/find-insights-with-ranked-cross-correlations/}{Find Insights with Ranked Cross-Correlations}
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset

# Only data with no plot
corr_cross(dft, plot = FALSE, top = 10)

# Show only most relevant results filtered by pvalue
corr_cross(dft, rm.na = TRUE, max_pvalue = 0.05, top = 15)

# Cross-Correlation for certain variables
corr_cross(dft, contains = c("Survived", "Fare"))

# Cross-Correlation max values per category
corr_cross(dft, type = 2, top = NA)
}
\seealso{
Other Calculus: 
\code{\link{corr}()},
\code{\link{dist2d}()},
\code{\link{model_metrics}()},
\code{\link{quants}()}

Other Correlations: 
\code{\link{corr}()},
\code{\link{corr_var}()}
}
\concept{Calculus}
\concept{Correlations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gemini.R
\name{gemini_ask}
\alias{gemini_ask}
\alias{gemini_image}
\title{Gemini API Interaction with R}
\usage{
gemini_ask(
  ask,
  secret_key = get_creds("gemini")$api_key,
  url = Sys.getenv("LARES_GEMINI_API"),
  temperature = 0.5,
  max_tokens = 1024,
  quiet = FALSE,
  ...
)

gemini_image(
  ask,
  image,
  secret_key = get_creds("gemini")$api_key,
  url = Sys.getenv("LARES_GEMINI_API"),
  temperature = 0.5,
  max_tokens = 1024,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{ask}{Character. Redacted prompt to ask. If multiple asks are
requested, they will be concatenated with "+" into a single request.}

\item{secret_key}{Character. Secret Key. Get yours in:
platform.openai.com for OpenAI or makersuite.google.com/app/apikey
for Gemini.}

\item{url}{Character. Base API URL.}

\item{temperature}{Numeric. The temperature to use for generating
the response. Default is 0.5. The lower the \code{temperature},
the more deterministic the results in the sense that the highest probable
next token is always picked. Increasing temperature could lead to more
randomness, which encourages more diverse or creative outputs. You are
essentially increasing the weights of the other possible tokens.
In terms of application, you might want to use a lower temperature value
for tasks like fact-based QA to encourage more factual and concise responses.
For poem generation or other creative tasks, it might be beneficial to
increase the temperature value.}

\item{max_tokens}{Integer. The maximum number of tokens in the response.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}

\item{image}{Character. Data to be encoded/decoded. It can be a raw vector,
text connection or file name.}
}
\value{
(Invisible) list. Content returned from API POST and processed.
}
\description{
This function lets the user interact with Google's Gemini LLM Model using
its API, and returns the rendered reply.
}
\examples{
\dontrun{
api_key <- get_credentials()$gemini$api_key
# Open question:
gemini_ask("Can you write an R function to plot a dummy histogram?", api_key)
# Image question
image <- "man/figures/automl_map.png"
gemini_image("Can you explain this flow with more detail?", image, api_key)
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other LLM: 
\code{\link{gpt_ask}()},
\code{\link{gpt_prompter}()}
}
\concept{API}
\concept{Gemini}
\concept{LLM}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outliers.R
\name{outlier_tukey}
\alias{outlier_tukey}
\alias{outlier_turkey}
\title{Outliers: Tukey’s fences}
\usage{
outlier_tukey(x, k = 1.5)

outlier_turkey(x, k = 1.5)
}
\arguments{
\item{x}{Numeric. Distribution}

\item{k}{Positive Numeric. K-multiplier.}
}
\value{
Boolean vector detecting outliers.
}
\description{
Tukey’s fences is a technique used in box plots. The non-outlier
range is defined with \code{[Q1-k(Q3-Q1), Q3+k(Q3-Q1)]}, where Q1 and Q3
are the lower and upper quartiles respectively, k - some non-negative
constant (popular choice is 1.5). A value is an outlier based
on Tukey’s fences when its value does not lie in non-outlier range.
}
\seealso{
Other Outliers: 
\code{\link{outlier_zscore}()},
\code{\link{outlier_zscore_plot}()},
\code{\link{winsorize}()}
}
\concept{Outliers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_text.R, R/wrangling.R
\name{formatHTML}
\alias{formatHTML}
\alias{formatNum}
\title{Format a string text as markdown/HTML}
\usage{
formatHTML(text, color = "black", size = 20, bold = FALSE)

formatNum(
  x,
  decimals = 2,
  signif = NULL,
  type = Sys.getenv("LARES_NUMFORMAT"),
  pre = "",
  pos = "",
  sign = FALSE,
  abbr = FALSE,
  ...
)
}
\arguments{
\item{text}{Character. Strings to format.}

\item{color}{Character. Hex colour code.}

\item{size}{Numeric. Text size.}

\item{bold}{Boolean. Should the text be bold?}

\item{x}{Numerical Vector}

\item{decimals}{Integer. Amount of decimals to display. If set to
\code{NULL}, then \code{getOption("digits")} will be used.}

\item{signif}{Integer. Rounds the values in its first argument to
the specified number of significant digits.}

\item{type}{Integer. \code{1} for International standards. \code{2}
for American Standards. Use \code{Sys.setenv("LARES_NUMFORMAT" = 2)}
to set this parameter globally.}

\item{pre, pos}{Character. Add string before or after number.}

\item{sign}{Boolean. Add \code{+} sign to positive values.}

\item{abbr}{Boolean. Abbreviate using num_abbr()? You can use
the `decimals` parameter to set abbr's \code{n}(-1) parameter.}

\item{...}{Additional lazy eval parameters.}
}
\value{
String with format characters included.

Character. String vector with reformatted continuous numbers
}
\description{
Format any character string to HTML or markdown format. We
recommend using this format with the \code{ggtext::geom_richtext}
function to format text in \code{ggplot2} objects.

This function lets the user format numerical values nicely
}
\examples{
formatHTML("Text test", color = "#000000")
formatHTML(c(123, 456), color = "orange", size = 120, bold = TRUE)

# If you want to use it with \code{ggtext}:
\dontrun{
col1 <- "grey"
col2 <- "orange"
pt <- data.frame(
  label = paste0(
    formatHTML(123, color = col2, size = 120, bold = TRUE), "<br/>",
    formatHTML("of children had a", col1), "<br/>",
    formatHTML("traditional stay-at-home mom", color = col2, bold = TRUE), "<br/>",
    formatHTML(paste0("in 2012, compared to ", 321, " in 1970"), color = col1)
  )
)
ggplot(pt, aes(x = 0, y = 0)) +
  ggtext::geom_richtext(
    aes(label = label),
    hjust = 0,
    label.color = NA,
    lineheight = 1.5
  ) +
  xlim(0, 0.01) +
  theme_void()
}
formatNum(1.23456, decimals = 3)
formatNum(1.23456, type = 1)
formatNum(1.23456, pre = "$", pos = "/person")
formatNum(123456, abbr = TRUE)
formatNum(c(123123, 123.123, 0.123123), signif = 2)
formatNum(1234567890, abbr = TRUE, signif = 3)
formatNum(1234567890, decimals = 0, abbr = TRUE)
formatNum(c(-3:3), sign = TRUE)
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}

Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/other_functions.R
\name{lares-exports}
\alias{lares-exports}
\alias{\%>\%}
\title{Pipe operator}
\description{
Pipe operator
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusters.R
\name{clusterVisualK}
\alias{clusterVisualK}
\title{Visualize K-Means Clusters for Several K}
\usage{
clusterVisualK(df, ks = 2:6, ...)
}
\arguments{
\item{df}{Dataframe}

\item{ks}{Integer vector. Which k should be tested?}

\item{...}{Additional parameters passed to \code{clusterKmeans}}
}
\value{
List. Plot and data.frame results of clustering \code{df}
data.frame into \code{ks} integer clusters.
}
\description{
Visualize cluster data for assorted values of k.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data("iris")
df <- subset(iris, select = c(-Species))
df <- df[sample(nrow(df)), ]

# Calculate and plot
result <- clusterVisualK(df, ks = 2:4)
plot(result$plot)

# You can use the data generated as well
lapply(result$data, function(x) head(x$cluster, 10))
}
\seealso{
Other Clusters: 
\code{\link{clusterKmeans}()},
\code{\link{clusterOptimalK}()},
\code{\link{reduce_pca}()},
\code{\link{reduce_tsne}()}
}
\concept{Clusters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_functions.R
\name{msplit}
\alias{msplit}
\title{Split a dataframe for training and testing sets}
\usage{
msplit(df, size = 0.7, seed = 0, print = TRUE)
}
\arguments{
\item{df}{Dataframe}

\item{size}{Numeric. Split rate value, between 0 and 1. If set to
1, the train and test set will be the same.}

\item{seed}{Integer. Seed for random split}

\item{print}{Boolean. Print summary results?}
}
\value{
List with both datasets, summary, and split rate.
}
\description{
This function splits automatically a dataframe into train and
test datasets. You can define a seed to get the same results
every time, but has a default value. You can prevent it from
printing the split counter result.
}
\examples{
data(dft) # Titanic dataset
splits <- msplit(dft, size = 0.7, seed = 123)
names(splits)
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()}

Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Machine Learning}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twitter.R
\name{get_tweets}
\alias{get_tweets}
\title{Get Tweets}
\usage{
get_tweets(q, n = 10000, creds = NA)
}
\arguments{
\item{q}{Query. Check for ?rtweet::search_tweets()}

\item{n}{Integer. Total of tweets to return}

\item{creds}{Character. Credential's user (see \code{get_creds()})}
}
\value{
data.frame with API response results.
}
\description{
This function downloads tweets with personal credentials
}
\seealso{
Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}
}
\concept{Credentials}
\concept{Twitter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_lares.R
\name{theme_lares}
\alias{theme_lares}
\title{Theme for ggplot2 (lares)}
\usage{
theme_lares(
  font = Sys.getenv("LARES_FONT"),
  size = 12,
  main_colour = "darkorange3",
  hard_colour = "black",
  soft_colour = "grey30",
  plot_colour = "transparent",
  panel_colour = "transparent",
  background = "transparent",
  no_facets = FALSE,
  legend = NULL,
  grid = TRUE,
  axis = TRUE,
  clean = FALSE,
  mg = 9,
  pal = 0,
  palette = NULL,
  which = "fc",
  quiet = TRUE,
  ...
)
}
\arguments{
\item{font, size}{Character and numeric. Base font family and base size for texts.
\code{Arial Narrow} is set by default when the library is loaded; you may change it
with \code{Sys.setenv("LARES_FONT" = "X")} or by using this parameter manually.
If \code{font} is "ignore" or NA, it'll be ignored.}

\item{main_colour, hard_colour, soft_colour, plot_colour, panel_colour}{Character. Main colours for your theme.}

\item{background}{Character. Main colour for your background. Overwrites
\code{plot_colour} and \code{panel_colour}.}

\item{no_facets}{Boolean. Suppress facet labels?}

\item{legend}{Character. Legend position: \code{"top"}, \code{"right"},
\code{"bottom"}, or \code{"left"} You can also set to \code{FALSE} or
\code{"none"} to suppress legend.}

\item{grid}{Character or Boolean. Use \code{TRUE/FALSE} or a combination of
\code{X}, \code{x}, \code{Y}, \code{y} to enable/disable minor and major grids.}

\item{axis}{Character or Boolean. Use \code{TRUE/FALSE}, \code{x} or \code{Y}
to enable X and/or Y axis lines.}

\item{clean}{Boolean. Suppress grids and axis? Overwrites both parameters.}

\item{mg}{Numeric. External margins reference.}

\item{pal}{Integer. \code{1} for fill and colour palette,
\code{2} for only colour palette, \code{3} for only fill palette, \code{4} for
personal labels-colour palette. \code{0} for nothing.}

\item{palette}{Character vector. Pass a vector with HEX colour
codes to use a custom palette. If you pass a named vector, the name values will be
used as fill and the values will be used as colour.}

\item{which}{Character. When \code{pal = 3}, select which colours should be
added with the custom colours palette: fill, colour, text (fct) - first letters.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}
}
\value{
Themed ggplot2 object
}
\description{
Based on \code{hrbrthemes}' \code{theme_ipsum} and customized for \code{lares} usage.
With this team you can custom the colour and fill palettes, global colour parameters,
major and minor grids, legend, font and font size.
}
\section{Why Arial Narrow?}{

First and foremost, Arial Narrow is generally installed by default or readily
available on any modern system, so it's "free"-ish; plus, it is a condensed font
with solid default kerning pairs and geometric numbers.
}

\examples{
\donttest{
data(dft)
library(ggplot2)
p <- ggplot(dft, aes(x = Pclass, y = sum(Fare), fill = Pclass)) +
  geom_col()
p + theme_lares()
p + theme_lares(pal = 1)
p + theme_lares(background = "#999999", mg = 25)
p + theme_lares(legend = "top", grid = "Yy")
p + theme_lares(clean = TRUE)
}
}
\seealso{
Other Themes: 
\code{\link{gg_fill_customs}()},
\code{\link{lares_pal}()},
\code{\link{plot_palette}()}
}
\concept{Themes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confidence.R
\name{ci_lower}
\alias{ci_lower}
\alias{ci_upper}
\title{Lower/Upper Confidence Intervals}
\usage{
ci_lower(mean, ssd, n, conf = 0.95)

ci_upper(mean, ssd, n, conf = 0.95)
}
\arguments{
\item{mean}{Numeric. Mean: \code{mean(var, na.rm = TRUE)}}

\item{ssd}{Numeric. Standard deviation: \code{sd(var, na.rm = TRUE)}}

\item{n}{Integer. Amount of observations: \code{n()}}

\item{conf}{Numeric (0-1). Confidence level.}
}
\value{
Vector with confidence limit value.
}
\description{
Calculate lower and upper confidence intervals given a mean,
standard deviation, sample size, and confidence level. You may
want to use \code{ci_var()} to calculate all values quickly.
}
\examples{
ci_lower(100, 5, 10)
ci_upper(100, 5, 10)
}
\seealso{
Other Confidence: 
\code{\link{ci_var}()}
}
\concept{Confidence}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{replacefactor}
\alias{replacefactor}
\title{Replace Factor Values}
\usage{
replacefactor(x, original, change)
}
\arguments{
\item{x}{Factor (or Character) Vector}

\item{original}{String or Vector. Original text you wish to replace}

\item{change}{String or Vector. Values you wish to replace the originals with}
}
\value{
Factor vector with transformed levels.
}
\description{
This function lets the user replace levels on a factor vector.
}
\examples{
library(dplyr)
data(dft)
# Replace a single value
dft <- mutate(dft, Pclass = replacefactor(Pclass, original = "1", change = "First"))
levels(dft$Pclass)
# Replace multiple values
dft <- mutate(dft, Pclass = replacefactor(Pclass, c("2", "3"), c("Second", "Third")))
levels(dft$Pclass)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_system.R
\name{what_size}
\alias{what_size}
\title{Calculate the size of any R object or directory}
\usage{
what_size(x = NULL, units = "Mb", path = NULL, recursive = TRUE, ...)
}
\arguments{
\item{x}{Object}

\item{units}{Character. Specify which unit to use,
i.e. "Gb", "Mb", "Kb".}

\item{path}{a character vector of full path names; the default
    corresponds to the working directory, \code{\link[base]{getwd}()}.  Tilde
    expansion (see \code{\link[base]{path.expand}}) is performed.  Missing
    values will be ignored.  Elements with a marked encoding will
    be converted to the native encoding (and if that fails, considered
    non-existent).}

\item{recursive}{logical.  Should the listing recurse into directories?}

\item{...}{further arguments passed to or from other methods.}
}
\description{
Calculate the size of any R object or directory
}
\examples{
what_size(seq(1:1e3), "Kb")
what_size(seq(1:1e6))
what_size(as.character(seq(1:1e6)))
# what_size(path = ".")
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_system.R
\name{files_functions}
\alias{files_functions}
\title{List all functions used in R script files by package}
\usage{
files_functions(filename, abc = TRUE, quiet = FALSE)
}
\arguments{
\item{filename}{Character. Path to an R file (or directory)
containing R code files.}

\item{abc}{Boolean. List functions alphabetically.
If FALSE, will list in order of frequency.}

\item{quiet}{Boolean. Keep quiet? If not, print messages and
\code{statusbar}.}
}
\value{
data.frame. Each row is a function and columns stating
number of appearances, percentage, packages, and files searched.
}
\description{
Parses all functions called by an R script and then lists
them by package. Wrapper for 'getParseData'. May be of great
use for those developing a package to help see what
namespace 'importsFrom' calls will be required.
}
\examples{
\dontrun{
# Choose an R script file with functions
rfile <- file.choose()
files_functions(rfile)
}
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missings.R
\name{missingness}
\alias{missingness}
\title{Calculate and Visualize Missingness}
\usage{
missingness(df, plot = FALSE, full = FALSE, subtitle = NA, summary = TRUE)
}
\arguments{
\item{df}{Dataframe. Dataframe to study}

\item{plot}{Boolean. Do you wish to plot results?}

\item{full}{Boolean. Return all variables (or only with missings)?}

\item{subtitle}{Character. Subtitle to show in plot}

\item{summary}{Boolean. Show numerical summary text?}
}
\value{
data.frame with each variable, number of missing values and percentage.
If \code{plot=TRUE}, a plot with the same information reflected.
}
\description{
This function lets the user calculate the percentage of NAs or
missingness in a data.frame. It also plots the results if needed.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal

# Dummy data
df <- data.frame(
  A = c(1:5),
  B = c(NA, NA, 1, 1, 1),
  C = rep(NA, 5),
  D = c(NA, LETTERS[1:4])
)

# Missing values summary
missingness(df)

# Visual results

missingness(df, plot = TRUE)

# Show all variables (including those with no missing values)
missingness(df, plot = TRUE, full = TRUE)
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}

Other Missing Values: 
\code{\link{impute}()}
}
\concept{Exploratory}
\concept{Missing Values}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/surveys.R
\name{plot_survey}
\alias{plot_survey}
\title{Visualize Survey Results}
\usage{
plot_survey(answers, ignore = 1, title = NA, subtitle = NA)
}
\arguments{
\item{answers}{Dataframe. Answers. Each row a different person.
Each column a different answer.}

\item{ignore}{Numeric Vector. Which columns are NOT answers?}

\item{title}{Character. Title for your plot}

\item{subtitle}{Character. Subtitle for your plot.}
}
\value{
ggplot2 object
}
\description{
This function lets the user plot a survey's result.
}
\seealso{
Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{ngrams}
\alias{ngrams}
\title{Build N-grams and keep most frequent}
\usage{
ngrams(text, ngram = c(2, 3), top = 10, stop_words = NULL, ...)
}
\arguments{
\item{text}{Character vector}

\item{ngram}{Integer vector. Number of continuous n items in text.}

\item{top}{Integer. Keep n most frequent ngrams only.}

\item{stop_words}{Character vector. Words to exclude from text. Example:
if you want to exclude "a", whenever that word appears it will be excluded,
but when the letter "a" appears in a word, it will remain.}

\item{...}{Additional parameters passed to \code{remove_stopwords}.}
}
\value{
data.frame with ngrams and counters, sorted by frequency.
}
\description{
Build out n-grams for multiple text inputs and keep the n most frequent
combinations.
}
\examples{
# You must have "tidytext" library to use this auxiliary function:
\dontrun{
women <- read.csv("https://bit.ly/3mXJOOi")
x <- women$description
ngrams(x, ngram = c(2, 3), top = 3)
ngrams(x, ngram = 2, top = 6, stop_words = c("a", "is", "of", "the"))
}
}
\seealso{
Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{remove_stopwords}()},
\code{\link{replaceall}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textCloud}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{topics_rake}()}
}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_system.R
\name{try_require}
\alias{try_require}
\title{Check if Specific Package is Installed}
\usage{
try_require(package, stop = TRUE, load = TRUE, lib.loc = NULL, ...)
}
\arguments{
\item{package}{Character. Name of the library}

\item{stop}{Boolean. Stop if not installed. If \code{FALSE} and
library is not available, warning will be shown.}

\item{load}{Boolean. Load library?}

\item{lib.loc}{Character vector. Location of R library trees
to search through, or \code{NULL}. The default value of \code{NULL}
corresponds to all libraries currently known to \code{.libPaths()}.
Non-existent library trees are silently ignored.}

\item{...}{Additional parameters.}
}
\value{
No return value, called for side effects.
}
\description{
This function checks library dependencies
}
\examples{
# Check if library base is installed. If not, stop and show error
try_require("base", stop = TRUE)
# Check if library xxx is installed. If not, show warning
try_require("xxx", stop = FALSE)
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/family.R
\name{seq_surnames}
\alias{seq_surnames}
\title{Surnames Order Sequence}
\usage{
seq_surnames(n = 1)
}
\arguments{
\item{n}{Integer. Number of generations to include in the sequence.
Notice it will generate a vector with 2^(n-1) values.}
}
\value{
Integer vector.
}
\description{
Generate a sequence of numbers that determines the order in which
surnames should be listed based on the number of generations of
ancestors you wish to include. This sequence follows the traditional
Latin custom of assigning the father's surname first, followed by
the mother's surname. The same logic extends systematically to higher
generations, ensuring that the order of surnames remains consistent
as you move upward through the family tree.
}
\examples{
seq_surnames(1)
seq_surnames(2)
seq_surnames(3)
seq_surnames(4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_functions.R
\name{h2o_selectmodel}
\alias{h2o_selectmodel}
\title{Select Model from h2o_automl's Leaderboard}
\usage{
h2o_selectmodel(results, which_model = 1, quiet = FALSE, ...)
}
\arguments{
\item{results}{\code{h2o_automl()} object.}

\item{which_model}{Integer. Which model from the leaderboard you wish to use?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters on \code{h2o::h2o.automl}}
}
\value{
H2O processed model
}
\description{
Select wich model from the h2o_automl function to use
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Machine Learning}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statusbar.R
\name{statusbar}
\alias{statusbar}
\title{Progressive Status Bar (Loading)}
\usage{
statusbar(
  run = 1,
  max.run = 100,
  label = run,
  msg = "",
  type = Sys.getenv("LARES_STATUSBAR"),
  start_time = NA,
  multiples = 1,
  alarm = FALSE
)
}
\arguments{
\item{run}{Iterator. for loop or an integer with the current loop number.
Start with 1 preferibly}

\item{max.run}{Number. Maximum number of loops}

\item{label}{String. With additionaly information to be printed
at the end of the line. The default is \code{run}.}

\item{msg}{Character. Finish message.}

\item{type}{Character. Loading type style: equal, domino, sword, filled.}

\item{start_time}{POSIXct. Start time to consider. If NA, then
when first iteration starts will be set as start time. Useful
for when first iteration is showed as done but started a few
seconds/minutes ago.}

\item{multiples}{Integer. Only print when multiples of N (to avoid)
wasting resources on fast and lots of iterations.}

\item{alarm}{Boolean. Ping (sound) when done. Requires \code{beepr}.}
}
\value{
No return value, called for side effects.
}
\description{
This function lets the user view a progressbar for a 'for' loop.
}
\examples{
for (i in 1:9) {
  statusbar(i, 9, multiples = 2)
  Sys.sleep(0.3)
}
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frequencies.R
\name{freqs}
\alias{freqs}
\title{Frequencies Calculations and Plot}
\usage{
freqs(
  df,
  ...,
  wt = NULL,
  rel = FALSE,
  results = TRUE,
  variable_name = NA,
  plot = FALSE,
  rm.na = FALSE,
  title = NA,
  subtitle = NA,
  top = 20,
  abc = FALSE,
  save = FALSE,
  subdir = NA,
  quiet = FALSE
)
}
\arguments{
\item{df}{Data.frame}

\item{...}{Variables. Variables you wish to process. Order matters.
If no variables are passed, the whole data.frame will be considered}

\item{wt}{Variable, numeric. Weights.}

\item{rel}{Boolean. Relative percentages (or absolute)?}

\item{results}{Boolean. Return results in a dataframe?}

\item{variable_name}{Character. Overwrite the main variable's name}

\item{plot}{Boolean. Do you want to see a plot? Three variables tops.}

\item{rm.na}{Boolean. Remove NA values in the plot? (not filtered for
numerical output; use na.omit() or filter() if needed)}

\item{title}{Character. Overwrite plot's title with.}

\item{subtitle}{Character. Overwrite plot's subtitle with.}

\item{top}{Integer. Filter and plot the most n frequent for
categorical values. Set to NA to return all values}

\item{abc}{Boolean. Do you wish to sort by alphabetical order?}

\item{save}{Boolean. Save the output plot in our working directory}

\item{subdir}{Character. Into which subdirectory do you wish to
save the plot to?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
Plot when \code{plot=TRUE} and data.frame with grouped frequency results
when \code{plot=FALSE}.
}
\description{
This function lets the user group, count, calculate percentages
and cumulatives. It also plots results if needed. Tidyverse friendly.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset

# How many survived?
dft \%>\% freqs(Survived)

# How many survived per Class?
dft \%>\% freqs(Pclass, Survived, abc = TRUE)

# How many survived per Class with relative percentages?
dft \%>\% freqs(Pclass, Survived, abc = TRUE, rel = TRUE)

# Using a weighted feature
dft \%>\% freqs(Pclass, Survived, wt = Fare / 100)

### Let's check the results with plots:

# How many survived and see plot?
dft \%>\% freqs(Survived, plot = TRUE)

# How many survived per class?
dft \%>\% freqs(Survived, Pclass, plot = TRUE)

# Per class, how many survived?
dft \%>\% freqs(Pclass, Survived, plot = TRUE)

# Per sex and class, how many survived?
dft \%>\% freqs(Sex, Pclass, Survived, plot = TRUE)

# Frequency of tickets + Survived
dft \%>\% freqs(Survived, Ticket, plot = TRUE)

# Frequency of tickets: top 10 only and order them alphabetically
dft \%>\% freqs(Ticket, plot = TRUE, top = 10, abc = TRUE)
}
\seealso{
Other Frequency: 
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()}

Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}

Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}
\concept{Frequency}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dim_reduction.R
\name{reduce_pca}
\alias{reduce_pca}
\title{Reduce Dimensionality with PCA}
\usage{
reduce_pca(
  df,
  n = NULL,
  ignore = NULL,
  comb = c(1, 2),
  quiet = FALSE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{df}{Dataframe}

\item{n}{Integer. Number of dimensions to reduce to.}

\item{ignore}{Character vector. Names of columns to ignore.}

\item{comb}{Vector. Which columns do you wish to plot? Select which
two variables by name or column position.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{plot}{Boolean. Create plots?}

\item{...}{Additional parameters passed to \code{stats::prcomp}}
}
\value{
List with reduced dataframe and possible plots.
}
\description{
Principal component analysis or (PCA) is a method we can use to
reduce high-dimensional data to a low-dimensional space. In other words,
we cannot accurately visualize high-dimensional datasets because
we cannot visualize anything above 3 features. The main purpose behind
PCA is to transform datasets with more than 3 features (high-dimensional)
into typically a 2/3 column dataset. Despite the reduction into a
lower-dimensional space we still can retain most of the variance or
information from our original dataset.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data("iris")
df <- subset(iris, select = c(-Species))
df$id <- seq_len(nrow(df))
reduce_pca(df, n = 3, ignore = "id")
}
\seealso{
Other Dimensionality: 
\code{\link{reduce_tsne}()}

Other Clusters: 
\code{\link{clusterKmeans}()},
\code{\link{clusterOptimalK}()},
\code{\link{clusterVisualK}()},
\code{\link{reduce_tsne}()}
}
\concept{Clusters}
\concept{Dimensionality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/other_plots.R
\name{noPlot}
\alias{noPlot}
\title{Plot Result with Nothing to Plot}
\usage{
noPlot(message = "Nothing to show here!", size = 4.5, ...)
}
\arguments{
\item{message}{Character. What message do you wish to show?}

\item{size}{Numeric. Font size for \code{message} input.}

\item{...}{Additional parameters passed to \code{theme_lares()}.}
}
\value{
Empty ggplot2 object (with a \code{message} if set).
}
\description{
This function lets the user print a plot without plot, with a
customizable message. It is quite useful for Shiny renderPlot when
using filters and no data is returned.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
noPlot(message = "No plot to show!")
noPlot(background = "#FF5500", size = 7)
}
\seealso{
Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_functions.R
\name{iter_seeds}
\alias{iter_seeds}
\title{Iterate Seeds on AutoML}
\usage{
iter_seeds(df, y, tries = 10, ...)
}
\arguments{
\item{df}{Dataframe. Dataframe containing all your data, including
the dependent variable labeled as \code{'tag'}. If you want to define
which variable should be used instead, use the \code{y} parameter.}

\item{y}{Variable or Character. Name of the dependent variable or response.}

\item{tries}{Integer. Number of iterations}

\item{...}{Additional arguments passed to \code{h2o_automl}}
}
\value{
data.frame with performance results by seed tried on every row.
}
\description{
This functions lets the user iterate and search for best seed. Note that if
the results change a lot, you are having a high variance in your data.
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}
}
\concept{Machine Learning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{normalize}
\alias{normalize}
\title{Normalize Vector}
\usage{
normalize(x, range = c(0, 1), ...)
}
\arguments{
\item{x}{Numeric vector. The numbers to be normalized.}

\item{range}{A numeric vector of length 2 specifying the desired range
for normalization. Default is c(0, 1).}

\item{...}{Additional parameters.}
}
\value{
A numeric vector with normalized \code{x} values.
}
\description{
This function normalizes numerical values into a specified range,
defaulting to the 0 to 1 range.
}
\examples{
x <- c(0, 1, 4, 7.5, 10)
normalize(x)
normalize(x, range = c(-1, 1))
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_metrics}
\alias{mplot_metrics}
\title{Model Metrics and Performance Plots}
\usage{
mplot_metrics(
  results,
  subtitle = NA,
  model_name = NA,
  save = FALSE,
  subdir = NA,
  file_name = "viz_metrics.png"
)
}
\arguments{
\item{results}{Object. Results object from h2o_automl function}

\item{subtitle}{Character. Subtitle to show in plot}

\item{model_name}{Character. Model's name}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}
}
\value{
Plot with \code{results} performance.
}
\description{
This function generates plots of the metrics of a predictive model.
This is an auxiliary function used in \code{model_metrics()} when
the parameter \code{plot} is set to \code{TRUE}.
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_response}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{cleanText}
\alias{cleanText}
\alias{cleanNames}
\title{Clean text strings automatically}
\usage{
cleanText(
  text,
  spaces = TRUE,
  keep = "",
  lower = TRUE,
  ascii = TRUE,
  title = FALSE
)

cleanNames(df, num = "x", keep = "_", ...)
}
\arguments{
\item{text}{Character Vector}

\item{spaces}{Boolean. Keep spaces? If character input, spaces
will be transformed into passed argument.}

\item{keep}{Character. String (concatenated or as vector) with all characters
that are accepted and should be kept, in addition to alphanumeric.}

\item{lower}{Boolean. Transform all to lower case?}

\item{ascii}{Boolean. Only ASCII characters?}

\item{title}{Boolean. Transform to title format (upper case on first letters).}

\item{df}{data.frame/tibble.}

\item{num}{Add character before only-numeric names.}

\item{...}{Additional parameters passed to \code{cleanText()}.}
}
\value{
Character vector with transformed strings.

data.frame/tibble with transformed column names.
}
\description{
\code{cleanText}: Clean character strings automatically. Options to keep ASCII
characters only, keep certain characters, lower caps, title format, are available.

\code{cleanNames}: Resulting names are unique and consist only of the \code{_}
character, numbers, and ASCII letters. Capitalization preferences can be
specified using the \code{lower} parameter.
}
\details{
Inspired by \code{janitor::clean_names}.
}
\examples{
cleanText("Bernardo Lares 123")
cleanText("Bèrnärdo LáreS 123", lower = FALSE)
cleanText("Bernardo Lare$", spaces = ".", ascii = FALSE)
cleanText("\\\\@®ì÷å   \%ñS  ..-X", spaces = FALSE)
cleanText(c("maría", "€", "núñez_a."), title = TRUE)
cleanText("29_Feb-92()#", keep = c("#", "_"), spaces = FALSE)

# For a data.frame directly:
df <- dft[1:5, 1:6] # Dummy data
colnames(df) <- c("ID.", "34", "x_2", "Num 123", "Nòn-äscì", "  white   Spaces  ")
print(df)
cleanNames(df)
cleanNames(df, lower = FALSE)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Text Mining: 
\code{\link{ngrams}()},
\code{\link{remove_stopwords}()},
\code{\link{replaceall}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textCloud}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{topics_rake}()}
}
\concept{Data Wrangling}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tictoc.R
\name{tic}
\alias{tic}
\alias{toc}
\title{Stopwatch to measure timings in R}
\usage{
tic(id = 1, start = proc.time()["elapsed"], quiet = TRUE)

toc(
  id = 1,
  msg = "Elapsed time:",
  type = "units",
  signif = 3,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{id}{Define ID if multiple \code{tic} & \code{toc} are being used.}

\item{start}{Start time. Now is default.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{msg}{Character. Custom message shown}

\item{type}{Character. Output format for \code{time} list element.
Choose any of: \code{units, clock, seconds}.}

\item{signif}{Integer. Significant digits.}

\item{...}{Additional parameters.}
}
\value{
Invisible list. Contains tic (start time), toc (stop time),
elapsed time and message printed.

\code{toc} returns an (invisible) list containing the time-stamps
\code{tic} and \code{toc}, \code{time} in seconds and the message \code{msg}.
}
\description{
Start a stopwatch.

Stop a stopwatch.
}
\examples{
# Basic use (global stopwatch)
tic()
Sys.sleep(0.1)
toc()

# Multiple tic tocs
tic(id = "two", quiet = FALSE)
Sys.sleep(0.2)
toc(id = "two")

# Global is still working (id = 1)
toc(msg = "The function finished its work in")
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Time}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{sentimentBreakdown}
\alias{sentimentBreakdown}
\title{Sentiment Breakdown on Text}
\usage{
sentimentBreakdown(
  text,
  lang = "spanish",
  exclude = c("maduro", "que"),
  append_file = NA,
  append_words = NA,
  plot = TRUE,
  subtitle = NA
)
}
\arguments{
\item{text}{Character vector}

\item{lang}{Character. Language in text (used for stop words)}

\item{exclude}{Character vector. Which word do you wish to exclude?}

\item{append_file}{Character. Add a dictionary to append. This file
must contain at least two columns, first with words and second with
the sentiment (consider sentiments on description).}

\item{append_words}{Dataframe. Same as append_file but appending
data frame with word and sentiment directly}

\item{plot}{Boolean. Plot results summary?}

\item{subtitle}{Character. Add subtitle to the plot}
}
\value{
List. Contains data.frame with words and sentiments, summary and plot.
}
\description{
This function searches for relevant words in a given text and adds
sentiments labels (joy, anticipation, surprise, positive, trust,
anger, sadness, fear, negative, disgust) for each of them, using NRC.
Then, makes a summary for all words and plot results.
}
\seealso{
Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{ngrams}()},
\code{\link{remove_stopwords}()},
\code{\link{replaceall}()},
\code{\link{textCloud}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{topics_rake}()}
}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_importance}
\alias{mplot_importance}
\title{Variables Importances Plot}
\usage{
mplot_importance(
  var,
  imp,
  colours = NA,
  limit = 15,
  model_name = NA,
  subtitle = NA,
  save = FALSE,
  subdir = NA,
  file_name = "viz_importance.png"
)
}
\arguments{
\item{var}{Vector. Variable or column's names}

\item{imp}{Vector. Importance of said variables. Must have same length as var}

\item{colours}{If positive and negative contribution is known}

\item{limit}{Integer. Limit how many variables you wish to plot}

\item{model_name}{Character. Model's name}

\item{subtitle}{Character. Subtitle to show in plot}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}
}
\value{
Plot with ranked importance variables results.
}
\description{
This function plots Variable Importances
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
df <- data.frame(
  variable = LETTERS[1:6],
  importance = c(4, 6, 6.7, 3, 4.8, 6.2) / 100,
  positive = c(TRUE, TRUE, FALSE, TRUE, FALSE, FALSE)
)
head(df)

mplot_importance(
  var = df$variable,
  imp = df$importance,
  model_name = "Random values model"
)

# Add a colour for categories
mplot_importance(
  var = df$variable,
  imp = df$importance,
  colours = df$positive,
  limit = 4
)
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_response}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colour_palettes.R
\name{lares_pal}
\alias{lares_pal}
\title{Personal Colours Palette}
\usage{
lares_pal(return = "list")
}
\arguments{
\item{return}{Character. Get only what you need. Select any of:
"all" or "list" (list),
"colors" or "colours" (vector),
"pal" or "palette" (named vector),
"simple" (named vector),
"custom" or "personal" (data.frame)}
}
\value{
Depending on the \code{return} input, we get a:
\itemize{
  \item \code{vector} with \code{palette} results vector
  \item \code{vector} with \code{palette} results vector's names
  \item \code{list} with \code{palette} results vector, \code{labels}
  results data.frame, and \code{simple} results named vector
}
}
\description{
Fetch customizable palettes for the library's usage. The package has
its own default colour-blind friendly colours but can be customized using
R internal options (i.e. \code{options("lares.palette" = c("#FF8303" = "#000",
"#40A4D8" = "#FFF", ...))}. There are 3 options you can use to customize all
colour palletes: "lares.palette" (vector, will be used in the same order as passed,
and must have a counter colour defined), "lares.colours" (vector, simple colour
names and their HEX codes), and "lares.colours.custom" (data.frame, containing
"values" to use dynamically, "fill" for main colour, and "colour" (not obligatory)
for counter colour).
}
\examples{
# Simple colour-named palette
lares_pal("simple")

# Raw colours and counter-colours
# OR simply: lares_pal("palette")
nice_palette <- lares_pal("colours")
nice_palette_ctr <- as.vector(lares_pal()$palette)
lapply(list(nice_palette, nice_palette_ctr), head)

# Personal colours by name
df <- lares_pal("custom")
df[sample(nrow(df), 5), ]
}
\seealso{
Other Themes: 
\code{\link{gg_fill_customs}()},
\code{\link{plot_palette}()},
\code{\link{theme_lares}()}
}
\concept{Themes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{read.file}
\alias{read.file}
\title{Read Files Quickly (Auto-detected)}
\usage{
read.file(filename, current_wd = TRUE, sheet = 1, quiet = FALSE)
}
\arguments{
\item{filename}{Character. File name to import.}

\item{current_wd}{Boolean. Use current working directory before
the file's name? Use this param to NOT get absolute root directory.}

\item{sheet}{Character. Name or index of the sheet to read data
from if file is xlsx or xls.}

\item{quiet}{Boolean. Quiet summary message?}
}
\value{
List or data.frame, depending on \code{filename}'s data.
}
\description{
This function lets the user import csv, xlsx, xls, sav files.
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/querieSQL.R
\name{queryDB}
\alias{queryDB}
\title{PostgreSQL Queries on Database (Read)}
\usage{
queryDB(query, from, creds = NA)
}
\arguments{
\item{query}{Character. SQL Query}

\item{from}{Character. Credential's user (see \code{get_creds()})}

\item{creds}{Character. Credential's directory (see \code{get_creds()})}
}
\value{
data.frame. Result of fetching the \code{query} data.
}
\description{
This function lets the user query a PostgreSQL database. Previously
was called \code{queryDummy} but was replaced and deprecated for a
more general function by using the \code{from} parameter.
}
\examples{
\dontrun{
# Query a PostgreSQL database
query <- "SELECT * FROM my_table LIMIT 10"
result <- queryDB(query, from = "my_database")
head(result)
}
}
\seealso{
Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}
}
\concept{Credentials}
\concept{Database}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_process}
\alias{fb_process}
\title{Paginate and Process Facebook's API Results}
\usage{
fb_process(input, paginate = TRUE, sleep = 0, quiet = FALSE, ...)
}
\arguments{
\item{input}{GET's output object (response) or link (character).}

\item{paginate}{Boolean or integer. Run through all paginations? If set
to \code{FALSE}, only the first one will be processed. If set to any other
integer value, will process the first N paginations.}

\item{sleep}{Numeric value. How much should each loop wait until until running
the next pagination query?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}
}
\value{
data.frame with un-nested processed results or NULL if no results found.
}
\description{
Process and paginate raw results from Facebook's API, result of
querying the API with \code{httr::GET} or by passing an API link.
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_full}
\alias{mplot_full}
\title{MPLOTS Score Full Report Plots}
\usage{
mplot_full(
  tag,
  score,
  multis = NA,
  splits = 8,
  thresh = 6,
  subtitle = NA,
  model_name = NA,
  plot = TRUE,
  save = FALSE,
  subdir = NA,
  file_name = "viz_full.png"
)
}
\arguments{
\item{tag}{Vector. Real known label.}

\item{score}{Vector. Predicted value or model's result.}

\item{multis}{Data.frame. Containing columns with each category probability
or score (only used when more than 2 categories coexist).}

\item{splits}{Integer. Number of separations to plot}

\item{thresh}{Integer. Threshold for selecting binary or regression
models: this number is the threshold of unique values we should
have in 'tag' (more than: regression; less than: classification)}

\item{subtitle}{Character. Subtitle to show in plot}

\item{model_name}{Character. Model's name}

\item{plot}{Boolean. Plot results? If not, plot grid object returned}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}
}
\value{
Multiple plots gathered into one, showing \code{tag} vs
\code{score} performance results.
}
\description{
This function plots a whole dashboard with a model's results. It will automatically
detect if it's a categorical or regression's model by checking how many different
unique values the dependent variable (tag) has.
}
\examples{
\donttest{
Sys.unsetenv("LARES_FONT") # Temporal
data(dfr) # Results for AutoML Predictions
lapply(dfr, head)

# Dasboard for Binomial Model
mplot_full(dfr$class2$tag, dfr$class2$scores,
  model_name = "Titanic Survived Model"
)

# Dasboard for Multi-Categorical Model
mplot_full(dfr$class3$tag, dfr$class3$score,
  multis = subset(dfr$class3, select = -c(tag, score)),
  model_name = "Titanic Class Model"
)

# Dasboard for Regression Model
mplot_full(dfr$regr$tag, dfr$regr$score,
  model_name = "Titanic Fare Model"
)
}
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_response}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dropbox.R
\name{db_download}
\alias{db_download}
\title{Download/Import Dropbox File by File's Name}
\usage{
db_download(
  query,
  local_path = NULL,
  xlsx = TRUE,
  token_dir = NA,
  token_name = "token_pers.rds",
  quiet = FALSE
)
}
\arguments{
\item{query}{Search string. This string is split (on spaces) into
individual words. Files will be used if they contain
all words in the search string.}

\item{local_path}{Character. Path to save file to. If \code{NULL}
(the default), saves file to working directory with same name.
If not, but a valid folder, file will be saved in this folder with
same basename as path. If not \code{NULL} and not a folder,
file will be saved to this path exactly.}

\item{xlsx}{Boolean. Is it an Excel file? Can be returned as a list
for each tab and not as a file if needed. Will delete downloaded file.}

\item{token_dir}{Character. RDS with token local directory. You may set to
NA if you already set your credentials (see \code{get_creds()})}

\item{token_name}{Character. RDS file name with your token's data.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
If \code{query} returns a .xlsx file and \code{xlsx=TRUE}, will
return a data.frame. Else, \code{local_path} string.
}
\description{
This function lets the user download a file from Dropbox, specifying
its name, using a previously created token or with interactive window.
}
\examples{
\dontrun{
# Download a specific file
db_download("stocksReport.Rmd", local_path = "~/Desktop/generic.Rmd")
# Import an Excel file from Dropbox into a data.frame
df <- db_download("Portfolio LC.xlsx", xlsx = FALSE)
}
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}

Other Credentials: 
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}

Other Dropbox: 
\code{\link{db_upload}()}
}
\concept{Credentials}
\concept{Dropbox}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_opts}
\alias{check_opts}
\alias{check_attr}
\alias{is_url}
\alias{is_ip}
\alias{are_id}
\alias{are_constant}
\alias{are_binary}
\alias{is_even}
\alias{is_odd}
\title{Validate inputs (attributions, options, ...)}
\usage{
check_opts(
  inputs,
  opts,
  input_name = "input",
  type = "all",
  not = "stop",
  quiet = TRUE
)

check_attr(object, attr = "type", check = NULL, stop = TRUE)

is_url(x, ...)

is_ip(x, ...)

are_id(x)

are_constant(x)

are_binary(x)

is_even(x)

is_odd(x)
}
\arguments{
\item{inputs}{Vector character. Check options.}

\item{opts}{Vector character. Valid options.}

\item{input_name}{Character. Custom your message and change "input"
for any other string. For example: "column names".}

\item{type}{Character. Options: "all", "any."}

\item{not}{Character. Options: "stop", "message", "print", "return".}

\item{quiet}{Boolean. Keep quiet? If not, returns logical value.}

\item{object}{Object of any kind}

\item{attr}{Character. Attribute to check}

\item{check}{Character. Attribute value}

\item{stop}{Boolean. Stop if doesn't check?}

\item{x}{Vector}

\item{...}{Additional parameters passed to \code{grepl()}}
}
\value{
Boolean. Result of \code{inputs} in \code{opts} (options).
Depending on \code{type} and/or \code{stop} arguments,
errors or messages will be shown.

No return value, called for side effects.

\code{is_url}. Boolean. Result of checking if \code{x} is a valid URL string.

\code{is_ip}. Boolean. Result of checking if \code{x} is a valid IP string.

\code{are_id}. Boolean. Result of checking if \code{x} is a potential ID vector

\code{are_constant}. Boolean. Result of checking if \code{x} is a constant vector

\code{are_binary}. Boolean. Result of checking if \code{x} is a binary vector
}
\description{
This function validates if inputs match all/any of your options
and return error/message with possible options to use. Similar
to \code{match.arg()} but more flexible.

This function checks if an object has a specific attribute and
stops if not.

Check whether a value or vector is or is not following a set
of rules. For example: is an URL, is an ID vector, are non-variant or
constant values, are binary values... Notice that \code{is_} will return
the result for each observation and \code{are_} for the whole vector.
}
\examples{
opts <- c("A", "B", "C")
# Let's check the "all" logic
check_opts(inputs = c("A", "B"), opts, quiet = FALSE)
check_opts(inputs = c("X"), opts, not = "message", quiet = FALSE)
check_opts(inputs = c("A", "X"), opts, input_name = "value", not = "warning")
# Now let's check the "any" logic
check_opts(inputs = c("A", "X"), opts, type = "any")
check_opts(inputs = c("X"), opts, type = "any", not = "message")
check_opts(inputs = c("A", NA), opts, type = "any")
# Final trick: just ignore results
check_opts(inputs = "X", opts, not = "invisible")
test <- data.frame()
attributes(test)
check_attr(test, "class", "data.frame")
# check_attr(test, "class", "not.data.frame")
is_url(c("google.com", "http://google.com"))

is_ip(c("163.114.132.0", "7.114.132", "0.0.0.0", "1.1.1.1."))

are_id(1:10)
are_id(LETTERS[1:10])

are_constant(rep(1, 10))
are_constant(1:10)

are_binary(c("A", "B", "A"))

is_even(1:5)
is_odd(1:5)
is_odd(c(0, 1.5, 2.5, NA, Inf, NULL))
}
\concept{Checks}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_cuts}
\alias{mplot_cuts}
\title{Cuts by quantiles for score plot}
\usage{
mplot_cuts(
  score,
  splits = 10,
  model_name = NA,
  subtitle = NA,
  table = FALSE,
  save = FALSE,
  subdir = NA,
  file_name = "viz_ncuts.png"
)
}
\arguments{
\item{score}{Vector. Predicted value or model's result.}

\item{splits}{Integer. Numer of separations to plot}

\item{model_name}{Character. Model's name}

\item{subtitle}{Character. Subtitle to show in plot}

\item{table}{Boolean. Do you wish to return a table with results?}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}
}
\value{
Plot with performance results by cuts.
}
\description{
This function cuts by quantiles any score or prediction.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dfr) # Results for AutoML Predictions
head(dfr$class2)

# Data
mplot_cuts(dfr$class2$scores, splits = 5, table = TRUE)

# Plot
mplot_cuts(dfr$class2$scores, model_name = "Titanic Survived Model")
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_response}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/currency.R
\name{get_currency}
\alias{get_currency}
\title{Download Historical Currency Exchange Rate}
\usage{
get_currency(
  currency_pair,
  from = Sys.Date() - 99,
  to = Sys.Date(),
  fill = FALSE,
  ...
)
}
\arguments{
\item{currency_pair}{Character. Which currency exchange do you
wish to get the history from? i.e, USD/COP, EUR/USD...}

\item{from}{Date. From date}

\item{to}{Date. To date}

\item{fill}{Boolean. Fill weekends and non-quoted dates with
previous values?}

\item{...}{Additional parameters.}
}
\value{
data.frame. Result of fetching online data for \code{currency_pair}
grouped by date.
}
\description{
This function lets the user download historical currency exchange
rate between two currencies.
}
\examples{
\donttest{
# For today (or any one single date)
get_currency("USD/COP", from = Sys.Date())
# For multiple dates
get_currency("EUR/USD", from = Sys.Date() - 7, fill = TRUE)
}
}
\concept{Currency}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{year_month}
\alias{year_month}
\alias{year_quarter}
\alias{year_week}
\title{Convert Date into Year-Month, Year-Quarter or Year-Week Format}
\usage{
year_month(date)

year_quarter(date)

year_week(date)
}
\arguments{
\item{date}{Date vector. Date to transform format.}
}
\value{
Vector with dates reformatted
}
\description{
This function lets the user convert a date into YYYY-MM, YYYY-QX,
or YYYY-WW format easily.
}
\examples{
year_month(Sys.Date())
year_quarter(Sys.Date())
year_week(Sys.Date())
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audio.R
\name{mp3_get}
\alias{mp3_get}
\title{Download MP3 from URL}
\usage{
mp3_get(
  id,
  mp3 = TRUE,
  repo = "yt-dlp",
  params = "--no-check-certificate",
  start_time = 0,
  end_time = NA,
  overwrite = TRUE,
  open = FALSE,
  delete = open,
  info = TRUE,
  title = NULL,
  cover = FALSE,
  metadata = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{id}{Character. YouTube URL or ID to search for.}

\item{mp3}{Boolean. Add mp3 optimal parameters?}

\item{repo}{Character. Chose repository you installed youtube-dl from.
Any of: "youtube-dl" (latest stable version), "yt-dlp"
(latest dev version).}

\item{params}{Character. Additional parameters.}

\item{start_time, end_time}{Numeric. Start and end time
to trim the audio output in seconds.}

\item{overwrite}{Boolean. Overwrite original file?}

\item{open, delete}{Boolean. After everything's done, should the
file be opened? Should it be deleted?}

\item{info}{Boolean. Import and return metadata?}

\item{title}{Character. Title of the track to search for cover, metadata, and
name the file.}

\item{cover}{Boolean. Google Search its squared cover? Uses \code{title}
input when provided.}

\item{metadata}{Boolean. Use iTunes Search API (free) to extract "track" 
data and high-res cover art. Falls back to "spotifyr" if no results
are found (requires credentials and, since 2026-03-09, Spotify Premium 
linked to the client id). Uses \code{title} input when provided.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
(Invisible) list with id's meta-data.
}
\description{
This function downloads YouTube videos or Soundcloud or any other
platform supported by the youtube-dl library, and converts them into
high quality MP3 files. The URL can be for a single video or a whole
playlist. It also returns metadata into an (invisible) list.
}
\section{youtube-dl}{

More info from the original developers and its code:
\href{https://github.com/ytdl-org/youtube-dl/}{youtube-dl's Github}
}

\examples{
# You must have "youtube-dl" installed in your OS:
\dontrun{
# Download video from YouTube and convert to MP3
mp3_get("https://www.youtube.com/watch?v=lrlKcCdVw9Q")
# OR simply
mp3_get("lrlKcCdVw9Q")
# For dev version, use:
mp3_get("m3RX4LJh0iI", repo = "yt-dlp")
}
}
\seealso{
Other Scrapper: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{holidays}()},
\code{\link{ip_data}()},
\code{\link{readGS}()},
\code{\link{splot_summary}()},
\code{\link{stocks_quote}()}

Other Audio: 
\code{\link{mp3_trim}()},
\code{\link{mp3_update_tags}()}
}
\concept{Audio}
\concept{Scrapper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataframe_str.R
\name{plot_nums}
\alias{plot_nums}
\title{Plot All Numerical Features (Boxplots)}
\usage{
plot_nums(df)
}
\arguments{
\item{df}{Dataframe}
}
\value{
Plot. Result of \code{df} numerical features.
}
\description{
This function filters numerical columns and plots boxplots.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset
plot_nums(dft)
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stocks.R
\name{weighted_value}
\alias{weighted_value}
\title{Calculate weighted stock values using FIFO/LIFO}
\usage{
weighted_value(
  value,
  n = rep(1, length(value)),
  technique = NULL,
  n_stocks = NULL,
  buy_only = TRUE,
  type = 1,
  ...
)
}
\arguments{
\item{value}{Numeric vector. Representing the values of the stock.}

\item{n}{Numeric vector. Representing the volume of the operation.
Positive for 'Buy' and negative for 'Sale'.}

\item{technique}{Character. Pick any of FIFO or LIFO, or NULL to skip.}

\item{n_stocks}{Integer. Specify the number of stocks to consider. By
default will sum positive values of \code{n}.}

\item{buy_only}{Boolean. Consider only buy (positive) values?}

\item{type}{Integer. 1 for returning the value, 2 for returning the
data.frame with the details ("df" attribute)}

\item{...}{Additional parameters.}
}
\value{
The calculated weighted mean value.
}
\description{
Calculate weighted stock values using FIFO/LIFO
}
\examples{
values <- c(10, 20, 30, 40, 50)
weights <- c(2, 3, -4, 5, 6)
mean(values)
weighted_value(values)
weighted.mean(values, weights)
weighted_value(values, weights, buy_only = FALSE)
# Using FIFO and LIFO
weighted_value(values, weights, "FIFO")
weighted_value(values, weights, "LIFO", n_stocks = 8)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lares.R
\docType{package}
\name{lares}
\alias{lares-package}
\alias{lares}
\title{Lean Analytics and Robust Exploration Sidekick}
\description{
R library for better/faster analytics, visualization, data mining, and machine learning tasks.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/laresbernardo/lares}
  \item \url{https://laresbernardo.github.io/lares/}
  \item Report bugs at \url{https://github.com/laresbernardo/lares/issues}
}

}
\author{
Bernardo Lares (laresbernardo@gmail.com)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{remove_stopwords}
\alias{remove_stopwords}
\title{Remove stop-words and patterns from character vector}
\usage{
remove_stopwords(text, stop_words, exclude = NULL, sep = " ")
}
\arguments{
\item{text}{Character vector}

\item{stop_words}{Character vector. Words to exclude from text. Example:
if you want to exclude "a", whenever that word appears it will be excluded,
but when the letter "a" appears in a word, it will remain.}

\item{exclude}{Character. Pattern to exclude using regex.}

\item{sep}{Character. String that separate the terms.}
}
\value{
Character vector with removed texts.
}
\description{
Remove all stop-words and specific patterns from a character vector
}
\examples{
x <- c("A brown fox jumps over a dog.", "Another brown dog.")
remove_stopwords(x, stop_words = c("dog", "brown", "a"), exclude = "\\\\.")
}
\seealso{
Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{ngrams}()},
\code{\link{replaceall}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textCloud}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{topics_rake}()}
}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outliers.R
\name{winsorize}
\alias{winsorize}
\title{Outliers: Winsorize}
\usage{
winsorize(x, thresh = c(0.05, 0.95), na.rm = FALSE)
}
\arguments{
\item{x}{Numeric vector. Distribution to be winsorized.}

\item{thresh}{Numeric vector. Lower and upper quantiles thresholds.
Set values within [0,1].}

\item{na.rm}{Boolean. Should \code{NA} be omitted to calculate the quantiles?
Note that \code{NA} in \code{x} are preserved and left unchanged anyway.}
}
\value{
Numeric vector transformed.
}
\description{
Winsorizing a vector means that a predefined quantum of the smallest
and/or the largest values are replaced by less extreme values.
Thereby the substitute values are the most extreme retained values.
}
\seealso{
Other Outliers: 
\code{\link{outlier_tukey}()},
\code{\link{outlier_zscore}()},
\code{\link{outlier_zscore_plot}()}
}
\concept{Outliers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/google_sheets.R
\name{readGS}
\alias{readGS}
\alias{readGS4}
\alias{writeGS}
\alias{writeGS4}
\title{Google Sheets Reading and Writing (API v4)}
\usage{
readGS(
  title,
  sheet = "Hoja 1",
  range = NULL,
  drop_nas = TRUE,
  json = NULL,
  email = NULL,
  api_key = NULL,
  server = FALSE,
  ...
)

writeGS(
  data,
  title,
  sheet = "Hoja 1",
  range = "A1",
  reformat = FALSE,
  append = FALSE,
  json = NULL,
  email = NULL,
  api_key = NULL,
  server = FALSE,
  ...
)
}
\arguments{
\item{title}{Character. Title of Google Drive file. Uses regular
expressions so you may fetch with patterns instead of names.}

\item{sheet}{Character. Working sheet to import}

\item{range}{Character. A cell range to read from}

\item{drop_nas}{Boolean. Remove columns and rows that contain only NAs?}

\item{json}{Character. JSON filename with service auth}

\item{email, api_key}{Character. If you have multiple pre-authorized
accounts in your machine, you may non-interactively select
which one you wish to use by email and/or api_key.}

\item{server}{Boolean. Force interacting auth process?}

\item{...}{Additional parameters passed to \code{read_sheet()}.}

\item{data}{Object (value, vector, data.frame, list).}

\item{reformat}{Boolean. Reformat the affected cells?}

\item{append}{Boolean.}
}
\value{
For reading, data.frame with the results of your Google Sheets file
based on its \code{title}, specificially the \code{sheet} and \code{range}
requested. For writing, no return value.
}
\description{
Read and write data from Google Sheets knowing the file's title.
You may use a single value from a cell or a data.frame from a cell range.
}
\seealso{
Other Scrapper: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{holidays}()},
\code{\link{ip_data}()},
\code{\link{mp3_get}()},
\code{\link{splot_summary}()},
\code{\link{stocks_quote}()}

Other Google: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{queryGA}()}
}
\concept{Google}
\concept{Scrapper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_metrics.R
\name{gain_lift}
\alias{gain_lift}
\title{Cumulative Gain, Lift and Response}
\usage{
gain_lift(
  tag,
  score,
  target = "auto",
  splits = 10,
  plot = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{tag}{Vector. Real known label}

\item{score}{Vector. Predicted value or model's result}

\item{target}{Value. Which is your target positive value? If
set to 'auto', the target with largest mean(score) will be
selected. Change the value to overwrite. Only used when binary
categorical model.}

\item{splits}{Integer. Number of percentiles to split the data}

\item{plot}{Boolean. Plot results? Uses \code{mplot_gain()}}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
data.frame when \code{plot=FALSE} or plot when \code{plot=TRUE}.
}
\description{
This function calculates cumulative gain, lift, and response
values for a predictive score of a specific target. You can use the
\code{mplot_gain()} function to create a plot.
}
\examples{
data(dfr) # Results for AutoML Predictions
head(dfr$class2)

# Results for Binomial Model
gain_lift(dfr$class2$tag, dfr$class2$scores, target = "FALSE")
gain_lift(dfr$class2$tag, dfr$class2$scores, target = "TRUE", splits = 5)
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Model metrics: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{errors}()},
\code{\link{loglossBinary}()},
\code{\link{model_metrics}()}
}
\concept{Machine Learning}
\concept{Model metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataframe_str.R
\name{plot_cats}
\alias{plot_cats}
\title{Plot All Categorical Features (Frequencies)}
\usage{
plot_cats(df)
}
\arguments{
\item{df}{Dataframe}
}
\value{
Plot. Result of \code{df} categorical features.
}
\description{
This function filters categorical columns and plots the frequency
for each value on every feature.
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{date_cuts}
\alias{date_cuts}
\title{Convert Date into Year + Cut}
\usage{
date_cuts(date = Sys.Date(), type = "Q")
}
\arguments{
\item{date}{Date. Date we wish to transform}

\item{type}{Character. Any of the following: B (2 months),
Q (3 months), T (4 months), H (6 months)}
}
\value{
Vector with date cut for each \code{date}
}
\description{
This function returns categorical values for any date(s) using year
cuts such as bimonths, quarters, terms, and halves.
}
\examples{
date_cuts(Sys.Date(), type = "Q")
date_cuts(Sys.Date(), type = "H")
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/google_sheets.R
\name{filesGD}
\alias{filesGD}
\title{Google Drive Files (API v4)}
\usage{
filesGD(title, server = FALSE, json = NULL, api_key = NULL, email = NULL)
}
\arguments{
\item{title}{Character. Title of Google Drive file. Uses regular
expressions so you may fetch with patterns instead of names.}

\item{server}{Boolean. Force interacting auth process?}

\item{json}{Character. JSON filename with service auth}

\item{email, api_key}{Character. If you have multiple pre-authorized
accounts in your machine, you may non-interactively select
which one you wish to use by email and/or api_key.}
}
\value{
Vector with found file names based on \code{title} on Google Drive.
}
\description{
Authenticate and find Google Drive files and IDs by name.
}
\seealso{
Other Scrapper: 
\code{\link{gtrends_related}()},
\code{\link{holidays}()},
\code{\link{ip_data}()},
\code{\link{mp3_get}()},
\code{\link{readGS}()},
\code{\link{splot_summary}()},
\code{\link{stocks_quote}()}

Other Google: 
\code{\link{gtrends_related}()},
\code{\link{queryGA}()},
\code{\link{readGS}()}
}
\concept{Google}
\concept{Scrapper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtistry.R
\name{rtistry_sphere}
\alias{rtistry_sphere}
\title{Generative Art: Sphere XmodY}
\usage{
rtistry_sphere(eye = c(100, 0, 0), pal = "auto", var = 3)
}
\arguments{
\item{eye, pal, var}{Parameters to change aesthetics and calculations}
}
\value{
ggplot object
}
\description{
Generative Art: Sphere XmodY
}
\concept{rtistry}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_response}
\alias{mplot_response}
\title{Cumulative Response Plot}
\usage{
mplot_response(
  tag,
  score,
  multis = NA,
  target = "auto",
  splits = 10,
  highlight = "auto",
  caption = NA,
  save = FALSE,
  subdir = NA,
  file_name = "viz_response.png",
  quiet = FALSE
)
}
\arguments{
\item{tag}{Vector. Real known label.}

\item{score}{Vector. Predicted value or model's result.}

\item{multis}{Data.frame. Containing columns with each category probability
or score (only used when more than 2 categories coexist).}

\item{target}{Value. Which is your target positive value? If
set to 'auto', the target with largest mean(score) will be
selected. Change the value to overwrite. Only works for binary classes}

\item{splits}{Integer. Numer of quantiles to split the data}

\item{highlight}{Character or Integer. Which split should be used
for the automatic conclussion in the plot? Set to "auto" for
best value, "none" to turn off or the number of split.}

\item{caption}{Character. Caption to show in plot}

\item{save}{Boolean. Save output plot into working directory}

\item{subdir}{Character. Sub directory on which you wish to save the plot}

\item{file_name}{Character. File name as you wish to save the plot}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
Plot with cumulative response and performance results by cuts.
}
\description{
The response gains plot helps us answer the question: When we
apply the model and select up until ntile X, what is the expected
% of target class observations in the selection?
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dfr) # Results for AutoML Predictions
lapply(dfr, head)

# Plot for Binomial Model
mplot_response(dfr$class2$tag, dfr$class2$scores,
  caption = "Titanic Survived Model",
  target = "TRUE"
)
mplot_response(dfr$class2$tag, dfr$class2$scores,
  caption = "Titanic Survived Model",
  target = "FALSE"
)

# Plot for Multi-Categorical Model
mplot_response(dfr$class3$tag, dfr$class3$score,
  multis = subset(dfr$class3, select = -c(tag, score)),
  caption = "Titanic Class Model"
)
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()},
\code{\link{mplot_topcats}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataframe_str.R
\name{plot_df}
\alias{plot_df}
\title{Plot Summary of Numerical and Categorical Features}
\usage{
plot_df(df)
}
\arguments{
\item{df}{Dataframe}
}
\value{
Plot. Result of \code{df} categorical and numerical features.
}
\description{
This function plots all columns frequencies and boxplots, for
categorical and numerical respectively.
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/robyn.R
\name{robyn_modelselector}
\alias{robyn_modelselector}
\alias{plot.robyn_modelselector}
\title{Robyn: Model Selection by Weighted Criteria Scores}
\usage{
robyn_modelselector(
  InputCollect,
  OutputCollect,
  metrics = c("rsq_train", "performance", "potential_improvement", "non_zeroes",
    "incluster_models", "cluster_sd", "certainty", "baseline_dist"),
  wt = c(2, 0.1, 0, 1, 0.1, 0, 1.5, 0),
  baseline_ref = 0,
  top = 4,
  n_per_cluster = 5,
  allocator_limits = c(0.5, 2),
  quiet = FALSE,
  cache = TRUE,
  ...
)

\method{plot}{robyn_modelselector}(x, ...)
}
\arguments{
\item{InputCollect, OutputCollect}{Robyn output objects.}

\item{metrics}{Character vector. Which metrics do you want to consider?
Pick any combination from: "rsq_train" for trained R squared,
"performance" for ROAS or (inverse) CPA, "potential_improvement" for
default budget allocator improvement using \code{allocator_limits},
"non_zeroes" for non-zero beta coefficients, "incluster_models" for
amount of models per cluster, "baseline_dist" for the difference between
the model's baseline and \code{baseline_ref} value, "certainty" metric
to minimize the channels' distance to their cluster's mean performance,
weighted by spends if \code{spend_wt = TRUE}, "cluster_sd" metric to score
based on the paid channels' performance standard deviations in clusters.
Additionally, you can use the standard MOO errors:
"nrmse", "decomp.rssd", and "mape" (the lowest the error, the highest
the score; same for "baseline_dist" and "cluster_sd").}

\item{wt}{Vector. Weight for each of the normalized \code{metrics} selected,
to calculate the score and rank models. Must have the same order and length
of \code{metrics} parameter input.}

\item{baseline_ref}{Numeric value. Between 0 and 1. What is the baseline
percentage you expect? Baseline in this case are all the sales or conversions
from non-media channels (organic & paid). Use with "baseline_dist" metric.}

\item{top}{Integer. How many ranked models to star? The better the model
is, the more stars it will have marked.}

\item{n_per_cluster}{Integer. How many models per cluster do you
want to plot? Default: 5. Keep in mind they will all be considered
for the calculations.}

\item{allocator_limits}{Numeric vector, length 2. How flexible do you
want to be with the budget allocator? By default, we'll consider a
0.5X and 2X range to let the budget shift across channels.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{cache}{Use cache functionality for allocator's results?}

\item{...}{Additional parameters.}

\item{x}{robyn_modelselector object}
}
\value{
list with resulting ranked data.frames, weights and plot.
}
\description{
Consider N best models to select the right ones to study using
several criteria/metrics such as potential improvement on budget
allocator, how many non-zero coefficients there are, R squared,
historical performance, baseline expectation, etc.

Read more about this functionality in Medium post:
\href{https://medium.com/@laresbernardo/select-the-right-mmm-candidate-based-on-your-specific-criteria-and-business-knowledge-1f583c3cb97a}{here}.
}
\seealso{
Other Robyn: 
\code{\link{robyn_hypsbuilder}()},
\code{\link{robyn_marginal}()},
\code{\link{robyn_performance}()}
}
\concept{Robyn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scrabble.R
\name{scrabble_dictionary}
\alias{scrabble_dictionary}
\alias{scrabble_score}
\alias{scrabble_points}
\alias{scrabble_words}
\title{Scrabble: Dictionaries}
\usage{
scrabble_dictionary(lang_dic, quiet = FALSE)

scrabble_score(words, scores.df)

scrabble_points(lang)

scrabble_words(
  tiles = "",
  free = 0,
  force_start = "",
  force_end = "",
  force_str = "",
  force_exclude = "",
  exclude_here = "",
  force_n = 0,
  force_max = 0,
  pattern = "",
  repeated = FALSE,
  language = Sys.getenv("LARES_LANG"),
  scores = language,
  words = NULL,
  quiet = FALSE,
  print = TRUE
)
}
\arguments{
\item{lang_dic}{Character. Any of "en","es","de","fr". Set to NULL
if you wish to skip this step (and use \code{words} parameter in
\code{scrabble_words} instead).}

\item{quiet}{Boolean. Do not print words as they are being searched.}

\item{words}{Character vector. Use if you wish to manually add words.}

\item{scores.df}{Dataframe. Must contain two columns: "tiles" with every
letter of the alphabet and "scores" for each letter's score.}

\item{lang}{Character. Any of "en","es" or "chars". Set to NULL
if you wish to skip this step (and use \code{words} parameter in
\code{scrabble_words()} instead). The "chars" parameter will
score the number of characters a word has.}

\item{tiles}{Character. The letters you wish to consider.}

\item{free}{Integer. How many free blank tiles you have?}

\item{force_start, force_end}{Character. Force words to start or end with
a pattern of letters and position. Examples: "S" or "SO" or "__S_O"...
If the string contains tiles that were not specified in \code{tiles}, they
will automatically be included.}

\item{force_str}{Character vector. Force words to contain strings.
If the string contains tiles that were not specified in \code{tiles}, they
will automatically be included.}

\item{force_exclude, exclude_here}{Character vector. Exclude words containing
these tiles (and positions). Not very relevant on Scrabble but for Wordle.}

\item{force_n, force_max}{Integer. Force words to be n or max n characters
long. Leave 0 to ignore parameter.}

\item{pattern}{Character string. Custom regex patterns you'd like to match.}

\item{repeated}{Boolean. By default, no replacement allowed. When activated,
a single tile can be repeated and won't be "used and discarded".}

\item{scores, language}{Character. Any of "en","es","de","fr".
If scores is not any of those languages, must be a data.frame that
contains two columns: "tiles" with every letter of the alphabet and
"scores" for each letter's score. If you wish
to overwrite or complement this dictionaries other words you can set to
\code{"none"} and/or use the \code{words} parameter.
You might also want to set this parameter globally with
\code{Sys.setenv("LARES_LANG" = "en")} and forget about it!}

\item{print}{Boolean. Print how many words are left by step.}
}
\value{
data.frame with words and language columns.

data.frame with word, scores, and length values for each \code{word}.

data.frame with tiles and scores for each alphabet letter.

data.frame with matching words found, sorted by higher points.
}
\description{
Download words from 4 different languages: English, Spanish,
German, and French. Words will be save into the \code{temp} directory.
This is an auxiliary function. You may want to use \code{scrabble_words}
directly if you are searching for the highest score words!

Get score for any word or list of words. You may set manually depending
on the rules and languages you are playing with. Check the examples
for Spanish and English values when I played Words With Friends.

Dataframe for every letter and points given a language.

Find highest score words given a set of letters, rules, and
language to win at Scrabble! You just have to find the best
place to post your tiles.
}
\examples{
\donttest{
if (haveInternet()) {
  # For Spanish words (default)
  es_scores <- scrabble_points("es")
  # Custom scores for each letter
  cu_scores <- data.frame(
    tiles = tolower(LETTERS),
    scores = c(
      1, 1, 1, 1, 1, 1, 1, 5, 1, 1, 5, 2, 4, 2, 1,
      4, 10, 1, 1, 1, 2, 5, 4, 8, 3, 10
    )
  )

  # Score values for each set of rules
  words <- c("Bernardo", "Whiskey", "R is great")
  scrabble_score(words, es_scores)
  scrabble_score(words, cu_scores)
}
}
scrabble_points("es")
scrabble_points("en")
# Not yet available
scrabble_points("fr")
\donttest{
if (haveInternet()) {
  # Automatic use of languages and scores
  Sys.setenv("LARES_LANG" = "es")
  scrabble_words(
    tiles = "hola",
    free = 2,
    force_start = "h",
    force_n = 4,
    force_str = "_o_a",
    exclude_here = "__z|j"
  )

  wordle <- c("board", "tempo", "shoes", "hoard")
  scrabble_words(
    language = NULL,
    words = wordle,
    force_n = 5,
    force_str = "O_R"
  )

  # Words considered for a language (you can custom it too!)
  es_words <- scrabble_dictionary("es")
}
}
}
\seealso{
Other Games: 
\code{\link{maze_solve}()},
\code{\link{sudoku_solver}()},
\code{\link{wordle_check}()}
}
\concept{Games}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update.R
\name{updateLares}
\alias{updateLares}
\title{Update the library (dev or CRAN version)}
\usage{
updateLares(force = FALSE, dev = TRUE, all = FALSE, local = FALSE, fb = FALSE)
}
\arguments{
\item{force}{Boolean. Force install.}

\item{dev}{Boolean. Developer version (Github)? If not, CRAN version.}

\item{all}{Boolean. Install other recommended libraries? Kinda Docker install!}

\item{local}{Boolean. Install package with local files? (or Github repo).}

\item{fb}{Boolean. From FB instance? Personal internal use.}
}
\value{
No return value, called for side effects.
}
\description{
This auxiliary function lets the user update \code{lares} to latest
CRAN or developer version.
}
\examples{
\dontrun{
# Update to CRAN version
updateLares(dev = FALSE)

# Update to development version from GitHub
updateLares(dev = TRUE)

# Install with all recommended packages
updateLares(all = TRUE)
}
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stocks.R
\name{stocks_file}
\alias{stocks_file}
\alias{daily_stocks}
\alias{daily_portfolio}
\alias{stocks_obj}
\alias{stocks_report}
\title{Build a Personal Investing Portfolio Report}
\usage{
stocks_file(
  file = NA,
  creds = NA,
  auto = TRUE,
  sheets = c("Portafolio", "Fondos", "Transacciones"),
  keep_old = TRUE,
  cache = TRUE,
  quiet = FALSE,
  ...
)

daily_stocks(hist, trans, tickers = NA, window = "MAX", ...)

daily_portfolio(hist, trans, cash, cash_fix = 0, window = "MAX")

stocks_obj(
  data = stocks_file(),
  cash_fix = 0,
  tax = 30,
  sectors = FALSE,
  parg = FALSE,
  window = c("1M", "YTD", "1Y", "MAX"),
  cache = TRUE,
  quiet = FALSE
)

stocks_report(
  data = NA,
  keep_old = TRUE,
  dir = NA,
  mail = FALSE,
  attachment = TRUE,
  to = "laresbernardo@gmail.com",
  sectors = FALSE,
  keep = FALSE,
  creds = NA,
  cache = TRUE
)
}
\arguments{
\item{file}{Character. Import an Excel file, local or from URL.}

\item{creds}{Character. Credential's user (see \code{get_creds()}) for
sending mail and Dropbox interaction.}

\item{auto}{Boolean. Automatically use my local personal file? You might want
to set in into your .Renviron \code{LARES_PORTFOLIO=~/dir/to/your/file.xlsx} so you
can leave all other parameters as \code{NA} and use it every time.}

\item{sheets}{Character Vector. Names of each sheet containing Portfolio summary,
Cash, and Transactions information. Please, keep the order of these tabs.}

\item{keep_old}{Boolean. Include sold tickers even though not currently in portfolio?}

\item{cache}{Boolean. Use daily cache if available?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}

\item{hist}{Dataframe. Result from \code{stocks_hist()}}

\item{trans}{Dataframe. Result from \code{stocks_file()$transactions}}

\item{tickers}{Dataframe. Result from \code{stocks_file()$portfolio}}

\item{window}{Character. Choose any of: "1W", "1M", "6M", "1Y", "YTD", "5Y", "MAX"}

\item{cash}{Dataframe. Result from \code{stocks_file()$cash}}

\item{cash_fix}{Numeric. If, for some reason, you need to fix your
cash amount for all reports, set the amount here}

\item{data}{Character. \code{stocks_obj()} output. If NA, automatic
parameters and \code{stocks_file()} defaults will be used.}

\item{tax}{Numeric. How much [0-99] of your dividends are gone with taxes?}

\item{sectors}{Boolean. Return sectors segmentation for ETFs?}

\item{parg}{Boolean. Personal argument. Used to personalize stuff, in this
case, taxes changed from A to B in given date (hard-coded)}

\item{dir}{Character. Directory for HTML report output. If set to NA,
current working directory will be used. If mail sent, file will be erased}

\item{mail}{Boolean. Do you want to send an email with the report attached?
If not, an HTML file will be created in dir}

\item{attachment}{Boolean. Create and add report as attachment if
\code{mail=TRUE}? If not, no report will be rendered and only tabulated
summaries will be included on email's body.}

\item{to}{Character. Email to send the report to}

\item{keep}{Boolean. Keep HTML file when sent by email?}
}
\value{
List with portfolio, transactions, and cash data.frames.

data.frame. Processed at date and symbol level.

data.frame. Processed at date and portfolio level.

List. Aggregated results and plots.

Invisible list. Aggregated results and plots.
}
\description{
\code{stocks_file()} lets the user download his personal Excel with his
Portfolio's data, locally or from Dropbox.

\code{daily_stocks()} creates a dataframe with all relevant metrics and values,
for each ticker or symbol, for every day since inception.

\code{daily_portfolio()} creates a data.frame with all relevant
metrics and values, for the overall portfolio, for every day
since inception.

\code{stocks_obj()} lets the user create his portfolio's calculations and
plots for further study.

\code{stocks_report()} lets the user create his portfolio's full report with
plots and send it to an email with the HTML report attached
}
\examples{
\dontrun{
# Load lares dummy portfolio XLSX
file <- system.file("inst/docs", "dummyPortfolio.xlsx", package = "lares")
df <- stocks_file(
  file = file,
  sheets = c("Portafolio", "Fondos", "Transacciones"),
  keep_old = FALSE
)
}
\dontrun{
list <- stocks_obj()
stocks_report(list, dir = "~/Desktop")
}
}
\seealso{
Other Investment: 
\code{\link{etf_sector}()},
\code{\link{splot_summary}()},
\code{\link{stocks_quote}()}

Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}
}
\concept{Credentials}
\concept{Investment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusters.R
\name{clusterOptimalK}
\alias{clusterOptimalK}
\title{Visualize K-Means Clusters for Several K Methods}
\usage{
clusterOptimalK(
  df,
  method = c("wss", "silhouette", "gap_stat"),
  drop_na = TRUE,
  ohse = TRUE,
  norm = TRUE,
  quiet = TRUE,
  ...
)
}
\arguments{
\item{df}{Dataframe}

\item{method}{Character vector.}

\item{drop_na}{Boolean. Should NA rows be removed?}

\item{ohse}{Boolean. Do you wish to automatically run one hot
encoding to non-numerical columns?}

\item{norm}{Boolean. Should the data be normalized?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters passed to \code{factoextra::fviz_nbclust}}
}
\value{
Plot. Optimal number of clusters of \code{df} data.frame given a
selected \code{method}.
}
\description{
Visualize cluster data for assorted values of k and methods such as
WSS, Silhouette and Gap Statistic. See \code{factoextra::fviz_nbclust}
for more.
}
\examples{
# You must have "factoextra" library to use this auxiliary function:
\dontrun{
data("iris")
df <- subset(iris, select = c(-Species))
# Calculate and plot optimal k clusters
clusterOptimalK(df)
}
}
\seealso{
Other Clusters: 
\code{\link{clusterKmeans}()},
\code{\link{clusterVisualK}()},
\code{\link{reduce_pca}()},
\code{\link{reduce_tsne}()}
}
\concept{Clusters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/encryption.R
\name{encrypt_file}
\alias{encrypt_file}
\alias{read_encrypted}
\alias{write_encrypted}
\alias{hex_to_raw}
\alias{raw_to_hex}
\title{File Encryption and Decryption (AES-256-CBC)}
\usage{
encrypt_file(input_file, output_file, key)

read_encrypted(input_file, key)

write_encrypted(data, output_file = "encrypted.enc", key = NULL, quiet = FALSE)

hex_to_raw(x)

raw_to_hex(x)
}
\arguments{
\item{input_file}{Character string. The path to the file to be encrypted.}

\item{output_file}{Character string. The path where the encrypted
file will be saved.}

\item{key}{A raw vector of 32 bytes (256 bits) to be used as the encryption key.}

\item{data}{List, vector or any json-able object to write into encrypted file.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{x}{Character string or hex vector to convert.}
}
\value{
No return value, called for side effects (writes to \code{output_file}).
}
\description{
This set of functions provides utilities for encrypting
and decrypting files using AES-256 in CBC mode, primarily for
handling key-value pair "secrets" files.
}
\details{
The encryption process pads the data, generates a random
Initialization Vector (IV), and encrypts the content. The IV is
prepended to the encrypted data in the output file.
Decryption involves extracting the IV, decrypting, and unpadding the data.
A specialized decryption function \code{read_encrypted} is provided
to directly parse decrypted content (assumed to be \code{key=value} pairs)
into a named R list. If no "=" sign found in the first line, text will be
imported as character vector.
}
\examples{
\dontrun{
# Create dummy files for demonstration
temp_input <- tempfile(fileext = ".txt")
temp_output <- tempfile(fileext = ".enc")
writeLines("library=lares \nversion=5.3.0 \n pass=123", temp_input)
readLines(temp_input)

# Generate a random 32-byte key (and save it in a secure place)
raw_key <- openssl::rand_bytes(32)
# You can convert from hex to raw and viceversa
raw_to_hex(raw_key)
hex_to_raw(raw_to_hex(raw_key))

# Encrypt the file
encrypt_file(temp_input, temp_output, raw_key)

# Import the data from encrypted file to list or vector
secrets <- read_encrypted(temp_output, raw_key)
unlist(secrets)

# Example using a string (JSON in this case)
writeLines(jsonlite::toJSON(list(a = 1, list(b = 2, c = 1:3))), temp_input)
encrypt_file(temp_input, temp_output, raw_key)
read_encrypted(temp_output, raw_key)

# Example writing a file from a list or vector directly from R
my_secrets <- list(
  api_key = "some_secret_api_key_123",
  username = "data_analyst",
  server = "production.server.com"
)
raw_key <- write_encrypted(my_secrets, temp_output, quiet = FALSE)
fromJSON(read_encrypted(temp_output, key = raw_key))
}
}
\seealso{
Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}
}
\concept{Credentials}
\keyword{Decryption}
\keyword{Encryption}
\keyword{Security}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{dist2d}
\alias{dist2d}
\title{Distance from specific point to line}
\usage{
dist2d(x, a = c(0, 0), b = c(1, 1))
}
\arguments{
\item{x}{Vector. Coordinates of the point from which we want to
measure the distance}

\item{a}{Vector. Coordinates of 1st point over the line}

\item{b}{Vector. Coordinates of 2st point over the line}
}
\value{
Numeric value result
}
\description{
This function lets the user calculate the mathematical linear distance
Between a specific point and a line (given geometrical 3 points)
}
\examples{
dist2d(x = c(5, 2))
dist2d(x = c(5, 2), a = c(0, 0), b = c(0, 1))
dist2d(x = c(5, 2), a = c(0, 0), b = c(1, 0))
}
\seealso{
Other Calculus: 
\code{\link{corr}()},
\code{\link{corr_cross}()},
\code{\link{model_metrics}()},
\code{\link{quants}()}
}
\concept{Calculus}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{textTokenizer}
\alias{textTokenizer}
\title{Tokenize Vectors into Words}
\usage{
textTokenizer(
  text,
  exclude = NULL,
  lang = NULL,
  min_word_freq = 5,
  min_word_len = 2,
  keep_spaces = FALSE,
  lowercase = TRUE,
  remove_numbers = TRUE,
  remove_punct = TRUE,
  remove_lettt = TRUE,
  laughs = TRUE,
  utf = TRUE,
  df = FALSE,
  h2o = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{text}{Character vector. Sentences or texts you wish to tokenize.}

\item{exclude}{Character vector. Which words do you wish to exclude?}

\item{lang}{Character. Language in text (used for stop words). Example:
"spanish" or "english". Set to \code{NA} to ignore.}

\item{min_word_freq}{Integer. This will discard words that appear
less than <int> times. Defaults to 2. Set to \code{NA} to ignore.}

\item{min_word_len}{Integer. This will discard words that have
less than <int> characters. Defaults to 5. Set to \code{NA} to ignore.}

\item{keep_spaces}{Boolean. If you wish to keep spaces in each line
to keep unique compound words, separated with spaces, set to TRUE.
For example, 'one two' will be set as 'one_two' and treated as
a single word.}

\item{lowercase, remove_numbers, remove_punct}{Boolean.}

\item{remove_lettt}{Boolean. Repeated letters (more than 3 consecutive).}

\item{laughs}{Boolean. Try to unify all laughs texts.}

\item{utf}{Boolean. Transform all characters to UTF (no accents and crazy symbols)}

\item{df}{Boolean. Return a dataframe with a one-hot-encoding kind of
results? Each word is a column and returns if word is contained.}

\item{h2o}{Boolean. Return \code{H2OFrame}?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
data.frame. Tokenized words with counters.
}
\description{
This function transforms texts into words, calculate frequencies,
supress stop words in a given language.
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{ngrams}()},
\code{\link{remove_stopwords}()},
\code{\link{replaceall}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textCloud}()},
\code{\link{textFeats}()},
\code{\link{topics_rake}()}
}
\concept{Data Wrangling}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{dft}
\alias{dft}
\title{Titanic Dataset}
\format{
An object of class \code{"data.frame"}
\describe{
  \item{PassengerId}{Unique ID for each passenger (1-891)}
  \item{Survived}{Did the passenger survive? (TRUE, FALSE)}
  \item{Pclass}{Ticket class, from first to third (1, 2, 3)}
  \item{Sex}{Gender (female, male)}
  \item{Age}{Age for each passenger in years (0.42-80)}
  \item{SibSp}{Amount of siblings / spouses aboard the Titanic (0-8)}
  \item{Parch}{Amount of parents / children aboard the Titanic (0-6)}
  \item{Ticket}{Ticket IDs}
  \item{Fare}{Amount paid for passenger's ticket (0-512.3292)}
  \item{Cabin}{width of top of diamond relative to widest point (43-95)}
  \item{Embarked}{Port of Embarkation (43-95)}
}
}
\usage{
data(dft)
}
\value{
data.frame
}
\description{
The sinking of the Titanic is one of the most infamous shipwrecks in history.
On April 15, 1912, during her maiden voyage, the widely considered "unsinkable"
RMS Titanic sank after colliding with an iceberg. Unfortunately, there weren’t
enough lifeboats for everyone onboard, resulting in the death of 1502 out of
2224 passengers and crew. While there was some element of luck involved in
surviving, it seems some groups of people were more likely to survive than others.
This dataset contains the details of a subset of the passengers on board
(891 to be exact) taken from Kaggle's Titanic
\href{https://www.kaggle.com/c/titanic/overview}{Train.csv}.
}
\examples{
data(dft)
head(dft)
}
\seealso{
Other Dataset: 
\code{\link{dfr}}
}
\concept{Dataset}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stocks.R
\name{stocks_quote}
\alias{stocks_quote}
\alias{stocks_hist}
\alias{plot.stocks_hist}
\title{Download Stocks Historical and Current Values}
\usage{
stocks_quote(symbols, ...)

stocks_hist(
  symbols = c("VTI", "META"),
  from = Sys.Date() - 365,
  to = Sys.Date(),
  today = TRUE,
  tax = 15,
  parg = FALSE,
  cache = TRUE,
  quiet = FALSE,
  ...
)

\method{plot}{stocks_hist}(x, type = 1, ...)
}
\arguments{
\item{symbols}{Character Vector. List of symbols to download historical data.}

\item{...}{Additional parameters.}

\item{from, to}{Date. Dates for range. If not set, 1 year will be downloaded.
Do use more than 4 days or will be over-written.}

\item{today}{Boolean. Do you wish to add today's live quote? This will happen
only if to value is the same as today's date}

\item{tax}{Numeric. How much [0-99] of your dividends are gone with taxes?}

\item{parg}{Boolean. Personal argument. Used to personalize stuff, in this
case, taxes changed from A to B in given date (hard-coded)}

\item{cache}{Boolean. Use daily cache if available?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{x}{stocks_hist object}

\item{type}{Integer. Select type of plot.}
}
\value{
data.frame with Symbol, Type of stock, Quote time, current value,
Daily Change, Market, and Symbol Name.
}
\description{
\code{stocks_quote()} lets the user download stocks live data.

\code{stocks_hist()} lets the user download stocks historical data.
}
\examples{
\dontrun{
# Multiple quotes at the same time
stocks_quote(c("VTI", "VOO", "TSLA"))
}
\dontrun{
df <- stocks_hist(symbols = c("VTI", "META", "FIW"), from = Sys.Date() - 180)
print(head(df))
plot(df)
}
}
\seealso{
Other Investment: 
\code{\link{etf_sector}()},
\code{\link{splot_summary}()},
\code{\link{stocks_file}()}

Other Scrapper: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{holidays}()},
\code{\link{ip_data}()},
\code{\link{mp3_get}()},
\code{\link{readGS}()},
\code{\link{splot_summary}()}
}
\concept{Investment}
\concept{Scrapper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_system.R
\name{listfiles}
\alias{listfiles}
\title{List files in a directory}
\usage{
listfiles(folder = getwd(), recursive = TRUE, regex = NA, images = FALSE)
}
\arguments{
\item{folder}{Character. Directory which contains files}

\item{recursive}{Boolean. Should the listing recurse into directories?}

\item{regex}{Character. String to use for filtering files}

\item{images}{Boolean. Bring only image files?}
}
\value{
data.frame with relevant data for each file on
\code{folder} directory.
}
\description{
This function lets the user list all files on a given directory.
It also lets filter files which contains a string.
}
\examples{
# All files in current directory (without recursive files)
df <- listfiles(recursive = TRUE)
head(df, 3)

# All files in current directory (with recursive files)
df <- listfiles(recursive = TRUE)
tail(df, 3)

# Check R files using regex
df <- listfiles(regex = "\\\\.R$")
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecasting.R
\name{prophesize}
\alias{prophesize}
\title{Facebook's Prophet Forecast}
\usage{
prophesize(
  df,
  n_future = 60,
  country = NULL,
  trend.param = 0.05,
  logged = FALSE,
  pout = 0.03,
  project = "Prophet Forecast"
)
}
\arguments{
\item{df}{Data frame. Must contain date/time column and values column,
in that order.}

\item{n_future}{Integer. How many steps do you wish to forecast?}

\item{country}{Character. Country code for holidays.}

\item{trend.param}{Numeric. Flexibility of trend component. Default is 0.05,
and as this value becomes larger, the trend component will be more flexible.}

\item{logged}{Boolean. Convert values into logs?}

\item{pout}{Numeric. Get rid of pout \% of outliers.}

\item{project}{Character. Name of your forecast project for plot title}
}
\value{
List. Containing the forecast results, the prophet model, and a plot.
}
\description{
Prophet is Facebook's procedure for forecasting time series data
based on an additive model where non-linear trends are fit with
yearly, weekly, and daily seasonality, plus holiday effects. It
works best with time series that have strong seasonal effects and
several seasons of historical data. Prophet is robust to missing
data and shifts in the trend, and typically handles outliers well.
}
\details{
Official documentation: \url{https://github.com/facebook/prophet}
}
\seealso{
Other Forecast: 
\code{\link{forecast_arima}()}
}
\concept{Forecast}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_plots.R
\name{mplot_topcats}
\alias{mplot_topcats}
\title{Top Hit Ratios for Multi-Classification Models}
\usage{
mplot_topcats(tag, score, multis, model_name = NA)
}
\arguments{
\item{tag}{Vector. Real known label.}

\item{score}{Vector. Predicted value or model's result.}

\item{multis}{Data.frame. Containing columns with each category probability
or score (only used when more than 2 categories coexist).}

\item{model_name}{Character. Model's name}
}
\value{
Plot with performance results over most frequent categories.
}
\description{
Calculate and plot a multi-class model's predictions accuracy
based on top N predictions and distribution of probabilities.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dfr) # Results for AutoML Predictions
mplot_topcats(dfr$class3$tag, dfr$class3$score,
  multis = subset(dfr$class3, select = -c(tag, score)),
  model_name = "Titanic Class Model"
)
}
\seealso{
Other ML Visualization: 
\code{\link{mplot_conf}()},
\code{\link{mplot_cuts}()},
\code{\link{mplot_cuts_error}()},
\code{\link{mplot_density}()},
\code{\link{mplot_full}()},
\code{\link{mplot_gain}()},
\code{\link{mplot_importance}()},
\code{\link{mplot_lineal}()},
\code{\link{mplot_metrics}()},
\code{\link{mplot_response}()},
\code{\link{mplot_roc}()},
\code{\link{mplot_splits}()}
}
\concept{ML Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{left}
\alias{left}
\alias{right}
\title{Left or Right N characters of a string}
\usage{
left(string, n = 1)

right(string, n = 1)
}
\arguments{
\item{string}{String or Vector.}

\item{n}{Integer. How many characters starting on right/left?}
}
\value{
Character. Trimmed strings.
}
\description{
This functions lets the user extract the first or last n characters
of a string or vector of strings.
}
\examples{
left("Bernardo", 3)
right(c("Bernardo", "Lares", "V"), 3)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dropbox.R
\name{db_upload}
\alias{db_upload}
\title{Upload Local Files to Dropbox}
\usage{
db_upload(
  filename,
  dir,
  delete_file = FALSE,
  token_dir = NA,
  token_name = "token_pers.rds"
)
}
\arguments{
\item{filename}{String. Local file's name to upload.}

\item{dir}{String. Directory you wish to upload the file to.}

\item{delete_file}{Boolean. Delete local file after uploading?}

\item{token_dir}{Character. RDS with token local directory. You may set to
NA if you already set your credentials (see \code{get_creds()})}

\item{token_name}{Character. RDS file name with your token's data.}
}
\value{
\code{TRUE} when successfully uploads file.
}
\description{
This function lets the user upload a local file to Dropbox,
using a previously created token or with interactive window.
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}

Other Credentials: 
\code{\link{db_download}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}

Other Dropbox: 
\code{\link{db_download}()}
}
\concept{Credentials}
\concept{Dropbox}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_accounts}
\alias{fb_accounts}
\title{Facebook Ad Accounts}
\usage{
fb_accounts(
  token,
  business_id = "904189322962915",
  type = c("owned", "client"),
  limit = 100,
  api_version = NULL,
  ...
)
}
\arguments{
\item{token}{Character. Valid access token with sufficient privileges. Visit the
\href{https://developers.facebook.com/tools/explorer}{Facebook API Graph Explorer}
to acquire one.}

\item{business_id}{Character. Business ID.}

\item{type}{Character vector. Values: owned, client.}

\item{limit}{Integer. Query limit by pagination.}

\item{api_version}{Character. Facebook API version.}

\item{...}{Additional parameters.}
}
\value{
data.frame with un-nested processed results fetched with API.
}
\description{
This returns all ad accounts for a FB Business Account FB.
For more information on Ad Insights' API, go to the
\href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/insights/}{original documentaion}
}
\examples{
\dontrun{
# Query all accounts (owned and with permissions) of a Business ID
accounts <- fb_accounts(YOURTOKEN, YOURBUSINESS)
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onehotencoding.R
\name{ohse}
\alias{ohse}
\title{One Hot Smart Encoding (Dummy Variables)}
\usage{
ohse(
  df,
  redundant = FALSE,
  drop = TRUE,
  ignore = NULL,
  dates = FALSE,
  holidays = FALSE,
  country = "Venezuela",
  currency_pair = NA,
  trim = 0,
  limit = 10,
  variance = 0.9,
  other_label = "OTHER",
  sep = "_",
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{Dataframe}

\item{redundant}{Boolean. Should we keep redundant columns? i.e. If the
column only has two different values, should we keep both new columns?
Is set to \code{NULL}, only binary variables will dump redundant columns.}

\item{drop}{Boolean. Drop automatically some useless features?}

\item{ignore}{Vector or character. Which column should be ignored?}

\item{dates}{Boolean. Do you want the function to create more features
out of the date/time columns?}

\item{holidays}{Boolean. Include holidays as new columns?}

\item{country}{Character or vector. For which countries should the holidays
be included?}

\item{currency_pair}{Character. Which currency exchange do you
wish to get the history from? i.e, USD/COP, EUR/USD...}

\item{trim}{Integer. Trim names until the nth character}

\item{limit}{Integer. Limit one hot encoding to the n most frequent
values of each column. Set to \code{NA} to ignore argument.}

\item{variance}{Numeric. Drop columns with more than n variance.
Range: 0-1. For example: if a variable contains 91 unique different
values out of 100 observations, this column will be suppressed if
value is set to 0.9}

\item{other_label}{Character. With which text do you wish to replace
the filtered values with?}

\item{sep}{Character. Separator's string}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}
}
\value{
data.frame on which all features are numerical by nature or
transformed with one hot encoding.
}
\description{
This function lets the user automatically transform a dataframe with
categorical columns into numerical by one hot encoding technic.
}
\examples{
data(dft)
dft <- dft[, c(2, 3, 5, 9, 11)]

ohse(dft, limit = 3) \%>\% head(3)
ohse(dft, limit = 3, redundant = NULL) \%>\% head(3)

# Getting rid of columns with no (or too much) variance
dft$no_variance1 <- 0
dft$no_variance2 <- c("A", rep("B", nrow(dft) - 1))
dft$no_variance3 <- as.character(rnorm(nrow(dft)))
dft$no_variance4 <- c(rep("A", 20), round(rnorm(nrow(dft) - 20), 4))
ohse(dft, limit = 3) \%>\% head(3)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Feature Engineering: 
\code{\link{date_feats}()},
\code{\link{holidays}()}

Other One Hot Encoding: 
\code{\link{date_feats}()},
\code{\link{holidays}()},
\code{\link{ohe_commas}()}
}
\concept{Data Wrangling}
\concept{Feature Engineering}
\concept{One Hot Encoding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{spread_list}
\alias{spread_list}
\title{Spread list column into new columns}
\usage{
spread_list(df, col, str = NULL, replace = TRUE)
}
\arguments{
\item{df}{Dataframe}

\item{col}{Variable name.}

\item{str}{Character. Start column names with. If set to \code{NULL},
original name of column will be used.}

\item{replace}{Boolean. Replace original values (delete column)}
}
\value{
data.frame. Result of un-nesting named or un-named list columns.
}
\description{
Spread an existing list column into new columns on a data.frame. Note
that every element on every observation must have a name for the function
to do its work. Original column will be automatically suppressed but you
can set the \code{replace} argument to avoid it.
}
\examples{
df <- dplyr::starwars
# Un-named list columns
spread_list(df, films, replace = FALSE) \%>\%
  dplyr::select(name, dplyr::starts_with("films")) \%>\%
  head(8)
# Named (and un-nammed) list columns
df <- dplyr::tibble(id = 1:3, platform = list(
  list("fb" = 1, "ig" = 2),
  list("fb" = 3),
  list()
))
spread_list(df, platform, str = "ptf_")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran.R
\name{cran_logs}
\alias{cran_logs}
\title{Download and plot daily downloads of CRAN packages}
\usage{
cran_logs(
  input = "lares",
  from = Sys.Date() - 31,
  to = Sys.Date() - 1,
  type = "daily",
  plot = TRUE
)
}
\arguments{
\item{input}{Character vector with package names or data.frame product of
\code{cranlogs::cran_downloads}.}

\item{from, to}{Dates. Range of dates to fetch downloads metrics.}

\item{type}{Character. Any of: "daily" or "total".}

\item{plot}{Boolean. Create a plot?}
}
\value{
List with data.frame and plot if \code{plot=TRUE}.
}
\description{
Download daily downloads stats from CRAN for any package, and plot.
It can also be used as an auxiliary function to plot
(\code{cranlogs::cran_downloads}) results.
}
\examples{
\donttest{
cran_logs(c("lares", "dplyr"), from = "2021-05-31")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chatgpt.R
\name{gpt_ask}
\alias{gpt_ask}
\alias{gpt_history}
\alias{gpt_table}
\alias{gpt_classify}
\alias{gpt_tag}
\alias{gpt_extract}
\alias{gpt_format}
\alias{gpt_convert}
\alias{gpt_translate}
\title{ChatGPT API Interaction with R}
\usage{
gpt_ask(
  ask,
  secret_key = get_credentials()$openai$secret_key,
  url = Sys.getenv("LARES_GPT_URL"),
  model = Sys.getenv("LARES_GPT_MODEL"),
  num_retries = 3,
  temperature = 0.5,
  max_tokens = NULL,
  pause_base = 1,
  quiet = FALSE,
  ...
)

gpt_history(quiet = TRUE, ...)

gpt_table(x, cols = NULL, quiet = TRUE, ...)

gpt_classify(x, categories, quiet = TRUE, ...)

gpt_tag(x, tags, quiet = TRUE, ...)

gpt_extract(x, extract, quiet = TRUE, ...)

gpt_format(x, format, quiet = TRUE, ...)

gpt_convert(x, unit, quiet = TRUE, ...)

gpt_translate(x, language, quiet = TRUE, ...)
}
\arguments{
\item{ask}{Character. Redacted prompt to ask. If multiple asks are
requested, they will be concatenated with "+" into a single request.}

\item{secret_key}{Character. Secret Key. Get yours in:
platform.openai.com for OpenAI or makersuite.google.com/app/apikey
for Gemini.}

\item{url}{Character. Base API URL.}

\item{model}{Character. OpenAI model to use. This can be adjusted
according to the available models in the OpenAI API (such as "gpt-4").}

\item{num_retries}{Integer. Number of times to retry the request in
case of failure. Default is 3.}

\item{temperature}{Numeric. The temperature to use for generating
the response. Default is 0.5. The lower the \code{temperature},
the more deterministic the results in the sense that the highest probable
next token is always picked. Increasing temperature could lead to more
randomness, which encourages more diverse or creative outputs. You are
essentially increasing the weights of the other possible tokens.
In terms of application, you might want to use a lower temperature value
for tasks like fact-based QA to encourage more factual and concise responses.
For poem generation or other creative tasks, it might be beneficial to
increase the temperature value.}

\item{max_tokens}{Integer. The maximum number of tokens in the response.}

\item{pause_base}{Numeric. The number of seconds to wait between retries.
Default is 1.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}

\item{x}{Vector. List items you wish to process in your instruction}

\item{cols}{Vector. Force column names for your table results.}

\item{categories, tags}{Vector. List of possible categories/tags to consider.}

\item{extract, format, unit}{Character. Length 1 or same as x to extract/format/unit
information from x. For example: email, country of phone number, country, amount as number,
currency ISO code, ISO, Fahrenheit, etc.}

\item{language}{Character. Language to translate to}
}
\value{
(Invisible) list. Content returned from API POST and processed.
}
\description{
This function lets the user ask ChatGPT via its API, and returns
the rendered reply. There are a couple of specific verbs (functions) with a
preset prompt to help fetch the data in specific formats. We also
store the prompts and replies in current session with their respective
time-stamps so user can gather historical results.
}
\examples{
\dontrun{
api_key <- get_credentials()$openai$secret_key
# Open question:
gpt_ask("Can you write an R function to plot a dummy histogram?", api_key)

##### The following examples return dataframes:
# Classify each element based on categories:
gpt_classify(1:10, c("odd", "even"))

# Add all tags that apply to each element based on tags:
gpt_tag(
  c("I love chocolate", "I hate chocolate", "I like Coke"),
  c("food", "positive", "negative", "beverage")
)

# Extract specific information:
gpt_extract(
  c("My mail is 123@test.com", "30 Main Street, Brooklyn, NY, USA", "+82 2-312-3456", "$1.5M"),
  c("email", "full state name", "country of phone number", "amount as number")
)

# Format values
gpt_format(
  c("March 27th, 2021", "12-25-2023 3:45PM", "01.01.2000", "29 Feb 92"),
  format = "ISO Date getting rid of time stamps"
)

# Convert temperature units
gpt_convert(c("50C", "300K"), "Fahrenheit")

# Create a table with data
gpt_table("5 random people's address in South America, email, phone, age between 18-30")
gpt_table(
  ask = "5 largest cities, their countries, and population",
  cols = c("city_name", "where", "POP")
)

# Translate text to any language
gpt_translate(
  rep("I love you with all my heart", 5),
  language = c("spanish", "chinese", "japanese", "russian", "german")
)

# Now let's read the historical prompts, replies, ano more from current session
gpt_history()
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other ChatGPT: 
\code{\link{gpt_prompter}()}

Other LLM: 
\code{\link{gemini_ask}()},
\code{\link{gpt_prompter}()}
}
\concept{API}
\concept{ChatGPT}
\concept{LLM}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sudoku.R
\name{sudoku_solver}
\alias{sudoku_solver}
\title{Solve Sudoku Puzzles}
\usage{
sudoku_solver(board, needed_cells = NULL, index = 1, quiet = FALSE)
}
\arguments{
\item{board}{Matrix. 9x9 matrix or vector length 81, with only digits from 0 to 9.}

\item{needed_cells, index}{Auxiliary parameters to auto-iterate using this same fx.}

\item{quiet}{Boolean. Keep quiet? If not, plot results.}
}
\value{
Logical output answering of the input board can be solved. The
actual solved solution will be created as an object named \code{solved}
in your \code{.GlobalEnv}.
}
\description{
Solve a Sudoku puzzle, where empty values are represented by 0s
into a matrix object.
}
\examples{
\donttest{
# board <- c(0,0,0,0,0,6,0,0,0,
#            0,9,5,7,0,0,3,0,0,
#            4,0,0,0,9,2,0,0,5,
#            7,6,4,0,0,0,0,0,3,
#            0,0,0,0,0,0,0,0,0,
#            2,0,0,0,0,0,9,7,1,
#            5,0,0,2,1,0,0,0,9,
#            0,0,7,0,0,5,4,8,0,
#            0,0,0,8,0,0,0,0,0)
# sudoku_solver(board)

# Trivial input (everything)
trivial <- matrix(rep(0, 81), byrow = TRUE, ncol = 9)
trivial
sudoku_solver(trivial)

# Wrong / Impossible to solve input
imp <- matrix(c(rep(1, 72), rep(0, 9)), byrow = TRUE, ncol = 9)
imp
sudoku_solver(imp)
}
}
\seealso{
Other Games: 
\code{\link{maze_solve}()},
\code{\link{scrabble_dictionary}()},
\code{\link{wordle_check}()}
}
\concept{Games}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_functions.R
\name{h2o_results}
\alias{h2o_results}
\title{Automated H2O's AutoML Results}
\usage{
h2o_results(
  h2o_object,
  test,
  train,
  y = "tag",
  which = 1,
  model_type,
  target = "auto",
  split = 0.7,
  ignore = NULL,
  quiet = FALSE,
  project = "ML Project",
  seed = 0,
  leaderboard = list(),
  plots = TRUE,
  model_name = NULL,
  ...
)
}
\arguments{
\item{h2o_object}{H2O Leaderboard (H2OFrame/H2OAutoML) or Model (h2o)}

\item{test, train}{Dataframe. Must have the same columns}

\item{y}{Variable or Character. Name of the dependent variable or response.}

\item{which}{Integer. Which model to select from leaderboard}

\item{model_type}{Character. Select "Classification" or "Regression"}

\item{target}{Value. Which is your target positive value? If
set to \code{'auto'}, the target with largest \code{mean(score)} will be
selected. Change the value to overwrite. Only used when binary
categorical model.}

\item{split}{Numeric. Value between 0 and 1 to split as train/test
datasets. Value is for training set. Set value to 1 to train with all
available data and test with same data (cross-validation will still be
used when training). If \code{train_test} is set, value will be overwritten
with its real split rate.}

\item{ignore}{Character vector. Columns too ignore}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{project}{Character. Your project's name}

\item{seed}{Integer. Set a seed for reproducibility. AutoML can only
guarantee reproducibility if max_models is used because max_time is
resource limited.}

\item{leaderboard}{H2O's Leaderboard. Passed when using
\code{h2o_selectmodel} as it contains plain model and no leader board.}

\item{plots}{Boolean. Create plots objects?}

\item{model_name}{Character. Optional custom name for the model. If provided,
this name will be used when saving the model with \code{export_results()}. If NULL
(default), the H2O-generated model ID will be used.}

\item{...}{Additional parameters on \code{h2o::h2o.automl}}
}
\value{
List. Trained model, predicted scores and datasets used, performance
metrics, parameters, importance data.frame, seed, and plots when \code{plots=TRUE}.
}
\description{
This is an auxiliary function to calculate predictions and results
when using the \code{h2o_automl()} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{cal_split}
\alias{cal_split}
\title{Split ICS File}
\usage{
cal_split(file, n_splits = 5, prefix = "calendar_split_", dir = NULL)
}
\arguments{
\item{file}{Character. Path to the ICS file to split.}

\item{n_splits}{Integer. Number of files to split into.}

\item{prefix}{Character. Prefix for the output filenames.}

\item{dir}{Character. Directory to save the output files.
Defaults to same directory as input file.}
}
\value{
Invisible list. The paths of the created files.
}
\description{
This function splits a large ICS calendar file into smaller files.
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frequencies.R
\name{freqs_plot}
\alias{freqs_plot}
\title{Combined Frequencies Plot for Categorical Features}
\usage{
freqs_plot(
  df,
  ...,
  top = 10,
  rm.na = FALSE,
  abc = FALSE,
  title = NA,
  subtitle = NA,
  quiet = FALSE
)
}
\arguments{
\item{df}{Data.frame}

\item{...}{Variables. Variables you wish to process. Order matters.
If no variables are passed, the whole data.frame will be considered}

\item{top}{Integer. Filter and plot the most n frequent for
categorical values. Set to NA to return all values}

\item{rm.na}{Boolean. Remove NA values in the plot? (not filtered for
numerical output; use na.omit() or filter() if needed)}

\item{abc}{Boolean. Do you wish to sort by alphabetical order?}

\item{title}{Character. Overwrite plot's title with.}

\item{subtitle}{Character. Overwrite plot's subtitle with.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
Plot. Result of the frequency of combined variables.
}
\description{
Plot frequencies of multiple categories within a data.frame in
a new fancy way. Tidyverse friendly, based on \code{lares::freqs()},
no limits on amount of features to evaluate.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset

x <- freqs_plot(dft, Pclass, Survived)
x$data
plot(x)

freqs_plot(dft, Pclass, Survived, Sex, Embarked)

freqs_plot(dft, Pclass, Survived, Sex, Embarked, top = 15)
}
\seealso{
Other Frequency: 
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()}

Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}

Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}
\concept{Frequency}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/robyn.R
\name{robyn_marginal}
\alias{robyn_marginal}
\title{Robyn: Marginal Performance (mROAS & mCPA) [Experimental]}
\usage{
robyn_marginal(..., marginal_unit = 1)
}
\arguments{
\item{...}{Additional parameters.}

\item{marginal_unit}{Additional units to calculate the marginal performance.}
}
\value{
list with base and marginal response results, marginal performance
metric and value, and plot.
}
\description{
Calculate and plot marginal performance of any spend or organic variable.
}
\examples{
\dontrun{
# You may load an exported model to recreate Robyn objects
mod <- Robyn::robyn_recreate(json_file = "your_model.json")
robyn_marginal(
  InputCollect = mod$InputCollect,
  OutputCollect = mod$OutputCollect,
  metric_name = "emails_O",
  metric_value = 100000,
  date_range = "all",
  marginal_unit = 10000000
)
}
}
\seealso{
Other Robyn: 
\code{\link{robyn_hypsbuilder}()},
\code{\link{robyn_modelselector}()},
\code{\link{robyn_performance}()}
}
\concept{Robyn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mails.R
\name{mail_send}
\alias{mail_send}
\title{Send Emails with Attachments (POST)}
\usage{
mail_send(
  from = "RMail <laresbernardo@gmail.com>",
  to = "laresbernardo@gmail.com",
  cc = NULL,
  bcc = NULL,
  subject = "Mail from R",
  text = " \\n",
  html = NULL,
  attachment = NULL,
  service = "mailgun",
  creds = NULL,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{from, to, cc, bcc}{Character. Emails}

\item{subject}{Character. Subject for the email.}

\item{text, html}{Character. Text or HTML to send in the body.}

\item{attachment}{Character, plot or data.frame. Will send the file,
plot as PNG or data.frame as CSV, respectively.}

\item{service}{Character. Service platform to search on \code{creds}.}

\item{creds}{Character. Credential's user (see \code{get_creds()}).
Must contain: url (POST address), api (API key).}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}
}
\value{
No return value, called for side effects.
}
\description{
This function lets the user send Emails with Attachments using MailGun's
API service.
}
\examples{
\dontrun{
myPlot <- noPlot("My plot")
mail_send(
  from = "BLV <myuser@mail.com>",
  to = "youruser@mail.com",
  cc = "myuser@mail.com",
  subject = paste("Daily report:", Sys.Date()),
  attachment = myPlot
)
}
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}

Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}
}
\concept{Credentials}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_text.R
\name{autoline}
\alias{autoline}
\title{New Line Feed for Long Strings (Wrapper)}
\usage{
autoline(text, top = "auto", rel = 9)
}
\arguments{
\item{text}{Character or factor vector.}

\item{top}{Integer. How many characters aprox. should be on each line?}

\item{rel}{Numeric. Relation of pixels and characters per line}
}
\value{
Character. String (vector) including some \code{\\n} within.
}
\description{
Add a break or new line without breaking words. Automatically,
the function can detect your plot's width and will dynamically
set an auto width. You can adjust the relation (rel) parameter
for different fonts and sizes until perfect harmony found.
Quite similar to \code{stringr::str_wrap} but, if the text vector
is a factor, the levels will be kept in order and transformed.
}
\examples{
cat(autoline("This is a long text that may not fit into a single line", 8))

text <- factor(c("First value", "Second value", "First value"),
  levels = c("First value", "Second value")
)
autoline(text, 1)

path <- file.path(R.home("doc"), "THANKS")
text <- paste(readLines(path), collapse = " ")
cat(autoline(text))
}
\seealso{
Other Tools: 
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_scales.R
\name{scale_x_comma}
\alias{scale_x_comma}
\alias{scale_y_comma}
\alias{scale_x_percent}
\alias{scale_y_percent}
\alias{scale_x_dollar}
\alias{scale_y_dollar}
\alias{scale_x_abbr}
\alias{scale_y_abbr}
\alias{scale_x_formatNum}
\alias{scale_y_formatNum}
\title{Axis scales format}
\usage{
scale_x_comma(...)

scale_y_comma(...)

scale_x_percent(...)

scale_y_percent(...)

scale_x_dollar(...)

scale_y_dollar(...)

scale_x_abbr(...)

scale_y_abbr(...)

scale_x_formatNum(
  ...,
  decimals = 2,
  signif = NULL,
  type = Sys.getenv("LARES_NUMFORMAT"),
  pre = "",
  pos = "",
  sign = FALSE,
  abbr = FALSE
)

scale_y_formatNum(
  ...,
  decimals = 2,
  signif = NULL,
  type = Sys.getenv("LARES_NUMFORMAT"),
  pre = "",
  pos = "",
  sign = FALSE,
  abbr = FALSE
)
}
\arguments{
\item{...}{Arguments passed to \code{ggplot2::continuous_scale} or
\code{formatNum} depending on the function.}

\item{decimals}{Integer. Amount of decimals to display. If set to
\code{NULL}, then \code{getOption("digits")} will be used.}

\item{signif}{Integer. Rounds the values in its first argument to
the specified number of significant digits.}

\item{type}{Integer. \code{1} for International standards. \code{2}
for American Standards. Use \code{Sys.setenv("LARES_NUMFORMAT" = 2)}
to set this parameter globally.}

\item{pre, pos}{Character. Add string before or after number.}

\item{sign}{Boolean. Add \code{+} sign to positive values.}

\item{abbr}{Boolean. Abbreviate using num_abbr()? You can use
the `decimals` parameter to set abbr's \code{n}(-1) parameter.}
}
\value{
Reformatted scales on ggplot2 object
}
\description{
The \code{_comma} ones set comma format for axis text, the \code{_percent}
ones set percent format for axis text, \code{_dollar} for collar currency,
and \code{_abbr} for abbreviated format. Lastly, use \code{_formatNum} to
further customize your numerical scales with \code{formatNum}.
}
\examples{
library(ggplot2)
df <- ggplot2::txhousing \%>\% removenarows(all = FALSE)

ggplot(df, aes(x = sales, y = volume)) +
  geom_point() +
  scale_x_dollar() +
  scale_y_abbr()

# Use any argument from scale_x/y_continuous
ggplot(df, aes(x = listings, y = log(inventory))) +
  geom_point() +
  scale_x_comma() +
  scale_y_percent(limits = c(0, 3))

# Use any argument from scale_x/y_continuous AND formatNum
ggplot(df, aes(x = median, y = inventory)) +
  geom_point() +
  scale_x_formatNum(n.breaks = 3, pre = "@", abbr = TRUE) +
  scale_y_formatNum(position = "right", decimals = 0, pos = " X")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_creatives}
\alias{fb_creatives}
\title{Facebook Creatives API}
\usage{
fb_creatives(token, which, api_version = NULL, process = TRUE, ...)
}
\arguments{
\item{token}{Character. Valid access token with sufficient privileges. Visit the
\href{https://developers.facebook.com/tools/explorer}{Facebook API Graph Explorer}
to acquire one.}

\item{which}{Character vector. This is the accounts, campaigns, adsets,
or ads IDs to be queried. Remember: if \code{report_level = "account"}, you must
start the ID with \code{act_}.}

\item{api_version}{Character. Facebook API version.}

\item{process}{Boolean. Process GET results to a more friendly format?}

\item{...}{Additional parameters.}
}
\value{
data.frame with un-nested processed results if \code{process=TRUE} or
raw API results as list when \code{process=FALSE}.
}
\description{
For more information: \href{https://developers.facebook.com/documentation/ads-commerce/marketing-api/}{Marketing API}
}
\examples{
\dontrun{
token <- YOURTOKEN
account <- act_ADACCOUNT

# Query all creatives for "which" (account in this case)
creatives <- fb_creatives(token, account)
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onehotencoding.R
\name{ohe_commas}
\alias{ohe_commas}
\title{One Hot Encoding for a Vector with Comma Separated Values}
\usage{
ohe_commas(df, ..., sep = ",", noval = "NoVal", remove = FALSE)
}
\arguments{
\item{df}{Dataframe. May contain one or more columns with comma separated
values which will be separated as one hot encoding}

\item{...}{Variables. Which variables to split into new columns?}

\item{sep}{Character. Which regular expression separates the elements?}

\item{noval}{Character. No value text}

\item{remove}{Boolean. Remove original variables?}
}
\value{
data.frame on which all features are numerical by nature or
transformed with one hot encoding.
}
\description{
This function lets the user do one hot encoding on a variable with
comma separated values
}
\examples{
df <- data.frame(
  id = c(1:5),
  x = c("AA, D", "AA,B", "B,  D", "A,D,B", NA),
  z = c("AA+BB+AA", "AA", "BB,  AA", NA, "BB+AA")
)
ohe_commas(df, x, remove = TRUE)
ohe_commas(df, z, sep = "\\\\+")
ohe_commas(df, x, z)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other One Hot Encoding: 
\code{\link{date_feats}()},
\code{\link{holidays}()},
\code{\link{ohse}()}
}
\concept{Data Wrangling}
\concept{One Hot Encoding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credentials.R
\name{get_credentials}
\alias{get_credentials}
\alias{get_creds}
\title{Load Credentials from a YML File}
\usage{
get_credentials(
  from = NA,
  dir = NA,
  filename = "config.yml",
  env = "LARES_CREDS",
  ...
)

get_creds(
  from = NA,
  dir = NA,
  filename = "config.yml",
  env = "LARES_CREDS",
  ...
)
}
\arguments{
\item{from}{Character. Family of values to import from the YML file.
If you don't know these names, set \code{from = NA}
and a warning will display all possible values, depending on your YML file.}

\item{dir}{Character. Credentials directory where your YML file is.
If used frequently, set your directory by using the \code{.Renviron} file.
To do so, leave \code{dir} as \code{NA} and follow the steps.
If \code{dir} is a list, it'll return \code{dir} (manual credentials input).
If \code{dir} ends with \code{.yml} or \code{.yaml}, \code{filename} will be ignored.}

\item{filename}{Character. YML filename with your credentials.
Ignored if \code{dir} contains the file path.}

\item{env}{Character. Environment variable name. No need to set differently
for any function that uses this library. Only for external use.}

\item{...}{Additional parameters.}
}
\value{
List. Result of reading your credential's YML file, filtered by your
\code{from} input if provided.
}
\description{
Load credentials from a local YML file. You can set your \code{.Renviron}
and the \code{LARES_CREDS} parameter to remember (forever) the directory
of your credentials' file. To use it later, you may leave \code{dir = NA}.
You may also use this function for external (non-\code{lares}) code/use.
}
\section{Set the default directory}{

The first time you use any function that has the \code{creds} parameter, if
the \code{dir} parameter is set to \code{NA}, this function will ask you to
set the directory where you save your YML local file with your credentials.
This will be asked once and will be set for further R sessions. Remember to
reset your session for this setup to start working properly.
}

\section{YML file format}{

A YML file is a text file, with \code{.yml} file format. You may start from
the dummy YML file shared which shows the structure you must follow to set your
credentials file. Check it out
\href{https://raw.githubusercontent.com/laresbernardo/lares/master/inst/docs/config.yml}{here}
or find it locally using \code{system.file("docs", "config.yml", package = "lares")}.
}

\examples{
\dontrun{
# Load dummy config.yml file from the library
# Recommendation: set dir with NA (read documentation)
# We need the directory, not the file
yml <- dirname(system.file("docs", "config.yml", package = "lares"))

# Let's see which credentials we have in our file
get_credentials(dir = yml)
# Warning message: No credentials for NA found in your YML file.
# Try any of the following: 'service1', 'service2', 'service3'

# Get credentials for service2
get_credentials("service2", dir = yml)
}
}
\seealso{
Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()},
\code{\link{stocks_file}()}
}
\concept{Credentials}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_text.R
\name{glued}
\alias{glued}
\title{Interpolate a string [glue wrapper]}
\usage{
glued(..., .sep = "", empty_lines = "keep", .envir = parent.frame())
}
\arguments{
\item{...}{[\code{expressions}]\cr Unnamed arguments are taken to be expression
string(s) to format. Multiple inputs are concatenated together before formatting.
Named arguments are taken to be temporary variables available for substitution.

For \code{glue_data()}, elements in \code{...} override the values in \code{.x}.}

\item{.sep}{[\code{character(1)}: \sQuote{""}]\cr Separator used to separate elements.}

\item{empty_lines}{Character. Set to \code{"keep"} to keep or
\code{"drop"} to drop empty lines.}

\item{.envir}{[\code{environment}: \code{parent.frame()}]\cr Environment to evaluate each expression in. Expressions are
evaluated from left to right. If \code{.x} is an environment, the expressions are
evaluated in that environment and \code{.envir} is ignored. If \code{NULL} is passed, it is equivalent to \code{\link[=emptyenv]{emptyenv()}}.}
}
\value{
Same as input but transformed (glued).
}
\description{
Format and interpolate a string using a \code{glue} wrapper. Allows
simple operations, \code{NULL} values as input, and interactions with
internal (created within \code{glued}) and external (environment) objects.
}
\examples{
name <- "Bernardo"
age <- 29
anniversary <- as.Date("2016-04-30")
glued("
  My name is {name},
  my age next year will be {age + 1},
  and I got married on {format(anniversary, '\%A, \%B \%d, \%Y')}.")

# Single braces can be inserted by doubling them
glued("My name is {name}, not {{name}}.")

# You can also used named arguments
glued(
  "Her name is {name}, ",
  "and her age next year will be {age + 1}.",
  name = "Maru",
  age = 6
)

# And run operations with memories (beware!)
glued("My name, {name}, has {n <- nchar(name); n} characters.
       If we multiply by ten, we'll have {10 * n} characters!")

# If you pass a vector, the operation will be repeated for each element
glued("Here's the value #{1:3}")
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onehotencoding.R
\name{holidays}
\alias{holidays}
\title{Holidays in your Country}
\usage{
holidays(
  countries = "Venezuela",
  years = year(Sys.Date()),
  quiet = FALSE,
  include_regions = FALSE
)
}
\arguments{
\item{countries}{Character or vector. For which country(ies) should the
holidays be imported? Accepts country names or ISO 3166-1 alpha-2 codes.}

\item{years}{Character or vector. For which year(s) do you wish to import
holiday dates?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{include_regions}{Boolean. Default FALSE. If TRUE, for countries with
internal subdivisions, it will provide details on which sub-state the found
holidays apply.}
}
\value{
\code{data.frame} with holidays data for given \code{countries} and \code{years}.
}
\description{
This function lets the user automatically retrieve public holiday dates
for any country supported by the Nager.Date API. Accepts country names
(e.g., "Portugal") or ISO 3166-1 alpha-2 codes (e.g., "PT").
Thanks to \href{https://date.nager.at}{Nager.Date}!
}
\examples{
\donttest{
holidays(countries = "Argentina")
year <- as.integer(format(Sys.Date(), format = "\%Y"))
holidays(countries = c("Spain", "Venezuela"), years = year)
holidays(countries = "Germany", include_regions = TRUE)
holidays(countries = "PT") # Also accepts ISO country codes
}
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Feature Engineering: 
\code{\link{date_feats}()},
\code{\link{ohse}()}

Other Scrapper: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{ip_data}()},
\code{\link{mp3_get}()},
\code{\link{readGS}()},
\code{\link{splot_summary}()},
\code{\link{stocks_quote}()}

Other One Hot Encoding: 
\code{\link{date_feats}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()}
}
\concept{Data Wrangling}
\concept{Feature Engineering}
\concept{One Hot Encoding}
\concept{Scrapper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audio.R
\name{mp3_update_tags}
\alias{mp3_update_tags}
\title{Update MP3 Metadata Tags}
\usage{
mp3_update_tags(
  filename,
  title = NULL,
  artist = NULL,
  album = NULL,
  genre = NULL,
  ...
)
}
\arguments{
\item{filename}{Character. Path to the MP3 file to update.}

\item{title}{Character. New title of the track (optional).}

\item{artist}{Character. New artist name (optional).}

\item{album}{Character. New album name (optional).}

\item{genre}{Character. New genre name (optional).}

\item{...}{Additional named arguments corresponding to other ID3 tags.
Only tags that exist in the MP3 file will be updated.}
}
\value{
Invisibly returns \code{NULL}. The MP3 file is updated in-place.
}
\description{
Updates the ID3 metadata tags of an MP3 file using the Python `eyeD3` library
via the `reticulate` package. You can modify standard fields such as
\code{title}, \code{artist}, \code{album}, and \code{genre}, as well as
additional tags passed through \code{...} if they exist in the MP3 file.
}
\details{
The function requires Python and the \code{eyeD3} package installed. It will
automatically initialize ID3 tags if they do not exist. Tags provided via
\code{...} are checked against the existing tag names to avoid errors.
}
\examples{
\dontrun{
mp3_update_tags(
  "song.mp3",
  title = "My Jazz Song",
  artist = "Bernardo",
  album = "Smooth Album",
  genre = "Jazz"
)
}
}
\seealso{
Other Audio: 
\code{\link{mp3_get}()},
\code{\link{mp3_trim}()}
}
\concept{Audio}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{distr}
\alias{distr}
\title{Compare Variables with their Distributions}
\usage{
distr(
  data,
  ...,
  type = 1,
  ref = TRUE,
  note = NA,
  top = 10,
  breaks = 10,
  na.rm = FALSE,
  force = "none",
  trim = 0,
  clean = FALSE,
  abc = FALSE,
  custom_colours = FALSE,
  plot = TRUE,
  chords = FALSE,
  save = FALSE,
  subdir = NA
)
}
\arguments{
\item{data}{Dataframe}

\item{...}{Variables. Main (target variable) and secondary (values
variable) to group by (if needed).}

\item{type}{Integer. 1 for both plots, 2 for counter plot only, 3 for
percentages plot only.}

\item{ref}{Boolean. Show a reference line if levels = 2? Quite useful
when data is unbalanced (not 50/50) because a reference line is drawn.}

\item{note}{Character. Caption for the plot.}

\item{top}{Integer. Filter and plot the most n frequent for categorical values.}

\item{breaks}{Integer. Number of splits for numerical values.}

\item{na.rm}{Boolean. Ignore \code{NA}s if needed.}

\item{force}{Character. Force class on the values data. Choose between 'none',
'character', 'numeric', 'date'}

\item{trim}{Integer. Trim labels until the nth character for categorical values
(applies for both, target and values)}

\item{clean}{Boolean. Use \code{cleanText()} for categorical values (applies
for both, target and values)}

\item{abc}{Boolean. Do you wish to sort by alphabetical order?}

\item{custom_colours}{Boolean. Use custom colours function?}

\item{plot}{Boolean. Return a plot? Otherwise, a table with results}

\item{chords}{Boolean. Use a chords plot?}

\item{save}{Boolean. Save the output plot in our working directory}

\item{subdir}{Character. Into which subdirectory do you wish to save the plot to?}
}
\value{
Plot when \code{plot=TRUE} with two plots in one: counter distribution
grouped by cuts, and proportions distribution grouped by same cuts. data.frame when
\code{plot=FALSE} with counting, percentages, and cumulative percentages results.
When \code{type} argument is used, single plots will be returned.
}
\description{
Compare the distribution of a target variable vs another variable. This
function automatically splits into quantiles for numerical variables.
Custom and tidyverse friendly.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset

# Relation for categorical/categorical values
distr(dft, Survived, Sex)

# Relation for categorical/numeric values
dft \%>\%
  distr(Survived, Fare, plot = FALSE) \%>\%
  head(10)
# Sort values
dft \%>\% distr(Survived, Fare, abc = TRUE)
# Less splits/breaks
dft \%>\% distr(Survived, Fare, abc = TRUE, breaks = 5)

# Distribution of numerical only
dft[dft$Fare < 20, ] \%>\% distr(Fare)

# Distribution of numerical/numerical
dft \%>\% distr(Fare, Age)

# Select only one of the two default plots of distr()
dft \%>\% distr(Survived, Age, type = 2)
dft \%>\% distr(Survived, Age, type = 3)
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()},
\code{\link{tree_var}()}

Other Visualization: 
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()},
\code{\link{tree_var}()}
}
\concept{Exploratory}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_functions.R
\name{target_set}
\alias{target_set}
\title{Set Target Value in Target Variable}
\usage{
target_set(tag, score, target = "auto", quiet = FALSE)
}
\arguments{
\item{tag}{Vector. Real known label}

\item{score}{Vector. Predicted value or model's result}

\item{target}{Value. Which is your target positive value? If
set to 'auto', the target with largest mean(score) will be
selected. Change the value to overwrite. Only used when binary
categorical model.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
List. Contains original data.frame \code{df} and
\code{which} with the target variable.
}
\description{
This function detects or forces the target value when predicting
a categorical binary model. This is an auxiliary function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{replaceall}
\alias{replaceall}
\title{Replace Values With}
\usage{
replaceall(df, original, change, which = "all", fixclass = TRUE, quiet = TRUE)
}
\arguments{
\item{df}{Data.frame or Vector}

\item{original}{String or Vector. Original text you wish to replace}

\item{change}{String or Vector. Values you wish to replace the originals with}

\item{which}{Character vector. Name of columns to use. Leave "all" for everything}

\item{fixclass}{Boolean. Try to detect logical classes after transformations (or
leave as default classes as character)?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
data.frame with replaced values based on inputs.
}
\description{
This function lets the user replace all specific values in a
vector or data.frame into another value. If replacing more than
one value, order matters so they will be replaced in the same
order that you pass them to the function. Factors will be refactored.
}
\examples{
df <- data.frame(
  one = c(1:4, NA),
  two = LETTERS[1:5],
  three = rep("A", 5),
  four = c(NA, "Aaa", 123, "B", "C")
)
print(df)

replaceall(df, "A", NA)

replaceall(df, "A", "a")

replaceall(df, 1, "*")

replaceall(df, NA, "NotNA")

replaceall(df, NA, 0)

replaceall(df, c("A", "B"), c("'A'", "'B'"))

replaceall(df, "a", "*", which = "four")
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{ngrams}()},
\code{\link{remove_stopwords}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textCloud}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{topics_rake}()}
}
\concept{Data Wrangling}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shapley.R
\name{h2o_shap}
\alias{h2o_shap}
\alias{plot.h2o_shap}
\title{SHAP values for H2O Models}
\usage{
h2o_shap(model, test = "auto", scores = "auto", y = "y", ...)

\method{plot}{h2o_shap}(x, relevant = TRUE, top = 15, quiet = FALSE, ...)
}
\arguments{
\item{model}{\code{h2o_automl} object or \code{h2o} model.}

\item{test}{String or Dataframe. Leave "auto" to use \code{h2o_automl}'s
test dataset or pass a valid dataframe.}

\item{scores}{Numeric vector. If test != "auto", you must provide predicted values}

\item{y}{Character. If test != "auto", you must provide y variable's name}

\item{...}{Additional argument for \code{predict_contributions.H2OModel}}

\item{x}{h2o_shap object}

\item{relevant}{Boolean. Keep only relevant non-trivial (>0) features}

\item{top}{Integer. Plot only top n values (as in importance)}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
H2OFrame with shap values for every observation and feature.
}
\description{
SHAP (SHapley Additive exPlanations) by Lundberg and Lee (2016) is a
method to explain individual predictions. SHAP is based on the game
theoretically optimal Shapley Values. Calculate SHAP values for
h2o models in which each row is an observation and each column a feature.
Use \code{plot} method to visualize features importance and distributions.
}
\examples{
\dontrun{
# Train a h2o_automl model
model <- h2o_automl(dft, Survived,
  max_models = 1, target = TRUE,
  ignore = c("Ticket", "Cabin", "PassengerId"),
  quiet = TRUE
)

# Calculate SHAP values
SHAP_values <- h2o_shap(model)
# Equivalent to:
# SHAP_values <- h2o_shap(
#  model = model$model,
#  test = model$datasets$test,
#  scores = model$scores_test$scores)

# Check SHAP results
head(SHAP_values)

# You must have "ggbeeswarm" library to use this auxiliary function:
# Plot SHAP values (feature importance)
plot(SHAP_values)

# Plot some of the variables (categorical)
shap_var(SHAP_values, Pclass)

# Plot some of the variables (numerical)
shap_var(SHAP_values, Fare)
}
}
\seealso{
Other SHAP: 
\code{\link{shap_var}()}
}
\concept{SHAP}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apis.R
\name{bring_api}
\alias{bring_api}
\title{Get API (JSON) and Transform into data.frame}
\usage{
bring_api(url, status = TRUE)
}
\arguments{
\item{url}{Character. API's URL to GET.}

\item{status}{Boolean. Display status message?}
}
\value{
data.frame of \code{url} \code{GET} results or
\code{NULL} if no results returned by API.
}
\description{
This function lets the user bring API data as JSON format and
transform it into data.frame.
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}

Other API: 
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}
}
\concept{API}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dalex_explainers.R
\name{dalex_variable}
\alias{dalex_variable}
\title{DALEX Partial Dependency Plots (PDP)}
\usage{
dalex_variable(explainer, vars, force_class = NA, seed = 123, ...)
}
\arguments{
\item{explainer}{Object. Result from \code{h2o_explainer} function.}

\item{vars}{Character vector. Which features do you wish to study?}

\item{force_class}{Character. If you wish to force a class on your
vars, which one do you need?}

\item{seed}{Numeric. Seed for reproducibility}

\item{...}{Additional parameters passed to \code{model_profile}.}
}
\value{
List. Containing PDP results, plot and \code{vars} input.
}
\description{
DALEX auxiliary function for creating Partial Dependency Plots and
study variable's responses vs independent vector.
}
\examples{
# You must have "DALEX" library to use this auxiliary function:
\dontrun{
# Having an "explainer" object created with \code{h2o_explainer}:
# For numerical variables
dalex_variable(explainer, vars = c("Age", "Fare"))
# For categorical variables
dalex_variable(explainer, vars = c("Pclass", "Sex"))
}
}
\seealso{
Other Interpretability: 
\code{\link{dalex_local}()},
\code{\link{dalex_residuals}()},
\code{\link{h2o_explainer}()}
}
\concept{Interpretability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trees.R
\name{tree_var}
\alias{tree_var}
\title{Recursive Partitioning and Regression Trees}
\usage{
tree_var(
  df,
  y,
  type = 2,
  max = 3,
  min = 20,
  cp = 0,
  ohse = TRUE,
  plot = TRUE,
  explain = TRUE,
  title = NA,
  subtitle = NULL,
  ...
)
}
\arguments{
\item{df}{Data frame}

\item{y}{Variable or Character. Name of the dependent variable or response.}

\item{type}{
    Type of plot. Possible values:

     \bold{0} Draw a split label at each split
              and a node label at each leaf.

     \bold{1} Label all nodes, not just leaves.
              Similar to \code{text.rpart}'s \code{all=TRUE}.

     \bold{2} Default.
              Like \code{1} but draw the split labels below the node labels.
              Similar to the plots in the CART book.

     \bold{3} Draw separate split labels for the left and right directions.

     \bold{4} Like \code{3} but label all nodes, not just leaves.
              Similar to \code{text.rpart}'s \code{fancy=TRUE}.
              See also \code{clip.right.labs}.\cr

     \bold{5} Show the split variable name in the interior nodes.\cr
  }

\item{max}{Integer. Maximal depth of the tree.}

\item{min}{Integer. The minimum number of observations that must
exist in a node in order for a split to be attempted.}

\item{cp}{
    complexity parameter.  Any split that does not decrease the overall
    lack of fit by a factor of \code{cp} is not attempted.  For instance,
    with \code{anova} splitting, this means that the overall R-squared must
    increase by \code{cp} at each step. The main role of this parameter
    is to save computing time by pruning off splits that are obviously
    not worthwhile. Essentially,the user informs the program that any
    split which does not improve the fit by \code{cp} will likely be
    pruned off by cross-validation, and that hence the program need
    not pursue it.
  }

\item{ohse}{Boolean. Auto generate One Hot Smart Encoding?}

\item{plot}{Boolean. Return a plot? If not, \code{rpart} object.}

\item{explain}{Boolean. Include a brief explanation on the bottom
part of the plot.}

\item{title, subtitle}{Character. Title and subtitle to include in plot.
Set to \code{NULL} to ignore.}

\item{...}{Additional parameters passed to \code{rpart.plot()}.}
}
\value{
(Invisible) list type 'tree_var' with plot (function), model,
predictions, performance metrics, and interpret auxiliary text.
}
\description{
Fit and plot a \code{rpart} model for exploratory purposes using
\code{rpart} and \code{rpart.plot} libraries.
}
\examples{
data(dft)
# Regression Tree
tree <- tree_var(dft, Fare, subtitle = "Titanic dataset")
tree$plot() # tree plot
tree$model # rpart model object
tree$performance # metrics
# Binary Tree
tree_var(dft, Survived_TRUE, explain = FALSE, cex = 0.8)$plot()
# Multiclass tree
tree_var(dft[, c("Pclass", "Fare", "Age")], Pclass, ohse = FALSE)$plot()
}
\seealso{
Other Exploratory: 
\code{\link{corr_var}()},
\code{\link{crosstab}()},
\code{\link{df_str}()},
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{lasso_vars}()},
\code{\link{missingness}()},
\code{\link{plot_cats}()},
\code{\link{plot_df}()},
\code{\link{plot_nums}()}

Other Visualization: 
\code{\link{distr}()},
\code{\link{freqs}()},
\code{\link{freqs_df}()},
\code{\link{freqs_list}()},
\code{\link{freqs_plot}()},
\code{\link{noPlot}()},
\code{\link{plot_chord}()},
\code{\link{plot_survey}()},
\code{\link{plot_timeline}()}
}
\concept{Exploratory}
\concept{Visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_predict.R
\name{h2o_predict_MOJO}
\alias{h2o_predict_MOJO}
\alias{h2o_predict_binary}
\alias{h2o_predict_model}
\alias{h2o_predict_API}
\title{Calculate predictions of h2o Models}
\usage{
h2o_predict_MOJO(df, model_path, method = "mojo", batch = 300)

h2o_predict_binary(df, model_path, sample = NA)

h2o_predict_model(df, model)

h2o_predict_API(df, api, exclude = "tag")
}
\arguments{
\item{df}{Dataframe/Vector. Data to insert into the model.}

\item{model_path}{Character. Relative model path directory or zip file.}

\item{method}{Character. One of "mojo" or "json".}

\item{batch}{Integer. Run n batches at a time for "json" method.}

\item{sample}{Integer. How many rows should the function predict?}

\item{model}{h2o model Object}

\item{api}{Character. API URL.}

\item{exclude}{Character. Name of the variables to exclude.}
}
\value{
data.frame with predicted results.

vector with predicted results.

data.frame with predicted results.

vector with predicted results.
}
\description{
\code{h2o_predict_MOJO} lets the user predict using the h2o .zip file
containing the MOJO files. Note that it works with the files
generated when using the function export_results()

\code{h2o_predict_binary} lets the user predict using the h2o binary file.
Note that it works with the files generated when using the
function export_results(). Recommendation: use the
h2o_predict_MOJO() function when possible - it let's you change
h2o's version without problem.

\code{h2o_predict_model} lets the user get scores from a H2O Model Object.

\code{h2o_predict_API} lets the user get the score from an API service
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}
}
\concept{H2O}
\concept{Machine Learning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outliers.R
\name{outlier_zscore_plot}
\alias{outlier_zscore_plot}
\title{Outliers: Z-score method plot}
\usage{
outlier_zscore_plot(df, var, group = NULL, thresh = c(2, 3, 5), top = 5)
}
\arguments{
\item{df}{Dataframe.}

\item{var}{Numeric variable.}

\item{group}{Categorical variable. Grouping variable.}

\item{thresh}{Numeric vector. Z-Score threshold for n standard deviations.}

\item{top}{Integer. Show only n most frequent categorical values when
using the \code{group} argument.}
}
\value{
ggplot2 object
}
\description{
Test several Z-score thresholds to visualize outliers. Tidyverse
friendly.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data(dft) # Titanic dataset
outlier_zscore_plot(dft, Fare)
p <- outlier_zscore_plot(dft, Fare, Pclass, thresh = c(3, 5))
plot(p)
attr(p, "z_values")
head(attr(p, "z_labels"))
}
\seealso{
Other Outliers: 
\code{\link{outlier_tukey}()},
\code{\link{outlier_zscore}()},
\code{\link{winsorize}()}
}
\concept{Outliers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_functions.R
\name{export_results}
\alias{export_results}
\title{Export h2o_automl's Results}
\usage{
export_results(
  results,
  thresh = 10,
  which = c("txt", "csv", "rds", "binary", "mojo", "plots", "dev", "production"),
  note = NA,
  subdir = NA,
  save = TRUE,
  seed = 0
)
}
\arguments{
\item{results}{\code{h2o_automl} or \code{h2o} model}

\item{thresh}{Integer. Threshold for selecting binary or regression
models: this number is the threshold of unique values we should
have in 'tag' (more than: regression; less than: classification)}

\item{which}{Character vector. Select which file format to export:
Possible values: txt, csv, rds, binary, mojo, plots. You might also
use dev (txt, csv, rds) or production (binary, mojo) or simply don't use
parameter to export everything}

\item{note}{Character. Add a note to the txt file. Useful when lots of
models are trained and saved to remember which one is which one}

\item{subdir}{Character. In which directory do you wish to save
the results?}

\item{save}{Boolean. Do you wish to save/export results?}

\item{seed}{Numeric. For reproducible results and random splits.}
}
\value{
No return value, called for side effects.
}
\description{
Export RDS, TXT, POJO, MOJO and all results from \code{h2o_automl()}.
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}
}
\concept{Machine Learning}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{ip_data}
\alias{ip_data}
\title{Scrap data based on IP address}
\usage{
ip_data(ip = myip(), quiet = FALSE)
}
\arguments{
\item{ip}{Vector. Vector with all IP's we wish to search.}

\item{quiet}{Boolean. Do not show the loading \code{statusbar}?}
}
\value{
data.frame. Each row is an unique \code{ip} address,
and columns will bee created for all the additional information found.
}
\description{
This function lets the user scrap https://db-ip.com/ given
IP address(es) to get their associated address type, ASN, ISP,
organization, country, state or region, county, city, ZIP postal code,
weather station, coordinates, Timezone, local time, languages, and currency.
}
\examples{
\donttest{
ip_data("163.114.132.0")
ip_data(ip = c(myip(), "201.244.197.199"), quiet = TRUE)
}
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{warnifnot}()},
\code{\link{what_size}()}

Other Scrapper: 
\code{\link{filesGD}()},
\code{\link{gtrends_related}()},
\code{\link{holidays}()},
\code{\link{mp3_get}()},
\code{\link{readGS}()},
\code{\link{splot_summary}()},
\code{\link{stocks_quote}()}
}
\concept{Scrapper}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_metrics.R
\name{errors}
\alias{errors}
\alias{rmse}
\alias{mae}
\alias{mse}
\alias{mape}
\alias{rsq}
\alias{rsqa}
\title{Calculate Continuous Values Errors}
\usage{
errors(tag, score)

rmse(tag, score)

mae(tag, score)

mse(tag, score)

mape(tag, score)

rsq(tag, score)

rsqa(tag, score)
}
\arguments{
\item{tag}{Vector. Real known label}

\item{score}{Vector. Predicted value or model's result}
}
\value{
data.frame or numeric values results for multiple error
metrics on continuous numerical vectors inputs.
}
\description{
This function lets the user calculate all errors and R squared
simultaneously.

This function lets the user calculate Root Mean Squared Error

This function lets the user calculate Mean Absolute Error

This function lets the user calculate Mean Squared Error

This function lets the user calculate Mean Squared Error

This function lets the user calculate R Squared

This function lets the user calculate Adjusted R Squared
}
\examples{
data(dfr) # Results for AutoML Predictions
head(dfr$regr)
df <- errors(dfr$regr$tag, dfr$regr$score)
head(df)
}
\seealso{
Other Model metrics: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{gain_lift}()},
\code{\link{loglossBinary}()},
\code{\link{model_metrics}()}
}
\concept{Model metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logo.R
\name{lares_logo}
\alias{lares_logo}
\title{Print lares R library Logo}
\usage{
lares_logo(version = TRUE)
}
\arguments{
\item{version}{Boolean. Include R and lares version?}
}
\description{
Used "ASCII Art Generator" from manytools.org to convert logo to ASCII.
\href{https://manytools.org/hacker-tools/convert-images-to-ascii-art/}{Visit}.
}
\examples{
lares_logo()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crossbrand.R, R/crossbrand2.R
\name{robyn_xmodels}
\alias{robyn_xmodels}
\alias{print.robyn_crossmmm}
\alias{robyn_xchannels}
\title{Cross-MMM Budget Optimization across Models}
\usage{
robyn_xmodels(
  models,
  initial_budgets = NULL,
  start_dates = NULL,
  end_dates = NULL,
  budget_constr_low = 0.5,
  budget_constr_up = 1.5,
  channel_constr_low = budget_constr_low,
  channel_constr_up = budget_constr_up,
  cores = NULL,
  quiet = FALSE,
  ...
)

\method{print}{robyn_crossmmm}(x, ...)

robyn_xchannels(
  models,
  initial_budgets = NULL,
  start_dates = NULL,
  end_dates = NULL,
  channel_constr_low = 0.5,
  channel_constr_up = 2,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{models}{Lists. Recreated Robyn models with \code{robyn_recreate()}.}

\item{initial_budgets}{Numeric vector. Default will use the total spends
per model for the specified or total date range.
Must be length 1 or same as \code{models}.}

\item{start_dates, end_dates}{Character vector. Start and end dates for each
specific model. You can specify a single date and will be used in all models.
Default empty value will assume you want all available data and date range.
Must be length 1 or same as \code{models}.}

\item{budget_constr_low, budget_constr_up}{Numeric vector. Relative minimum
and maximum budgets to consider based on \code{initial_budgets}.
By default it'll consider 50% and 150% budget constraints.
Must be length 1 or same as \code{models}.}

\item{channel_constr_low, channel_constr_up}{Numeric vector.
Relative lower and upper constraints per channel compared with mean
spend during the time period defined.
If mean was zero for date range, historical mean spend value will be used.
Must have length 1 to replicate for all channels or same length
(and order )as \code{paid_media_spends}.}

\item{cores}{Integer. How many cores to use for parallel computations?
Set to 1 to not use this option.
Default will the minimum between 10 cores and all available cores - 1.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters to be passed to internal functions.}

\item{x}{Object to print: robyn_crossmmm() output.}
}
\value{
Invisible vector with results by letter.

List. Contains optimized allocation results and plots.
}
\description{
Given a list of recreated Robyn models, this function optimizes budget
allocation across MMM with respective constraints by maximizing
incremental revenue/conversions. This method assumes each model is
independent, that can be compared given its spends were cleanly and
properly split, they modeled the same metric (revenue or conversion)
and units (currency or type of conversion), and date granularity.
For best results, ensure channels have similar granularity across
markets to simplify interpretation and application of the outputs.

Given a list of recreated Robyn models, this function optimizes budget
allocation across MMM with respective constraints by maximizing
response across all channels. This method assumes each model is
independent, that can be compared given its spends were cleanly and
properly split, they modeled the same metric (revenue or conversion)
and units (currency or type of conversion), and date granularity.
Recommended to have same channels granularity across markets
to simplify results readings and application.
}
\details{
This approach is faster and cleaner compared with previous proposal
using \code{robyn_xmodels()}.
}
\examples{
# You must have Robyn installed and some models stored as JSON files
\dontrun{
# Import and recreate the models
files <- c("BrandA.json", "BrandB.json", "BrandC.json", "BrandN.json")
models <- lapply(files, function(x) Robyn::robyn_recreate(x))
names(models) <- gsub("\\\\.json", "", files)

# Calculate cross-brand optimal allocation
res <- robyn_xmodels(
  models,
  cores = 10,
  start_dates = "2023-01-01",
  end_dates = "2023-12-01"
)
print(res)
res$summary
}
\dontrun{
# Calculate cross-brand optimal allocation
res <- robyn_xchannels(
  models,
  start_dates = "2023-01-01",
  end_dates = "2023-12-01"
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onehotencoding.R
\name{date_feats}
\alias{date_feats}
\title{One Hot Encoding for Date/Time Variables (Dummy Variables)}
\usage{
date_feats(
  dates,
  drop = FALSE,
  only = NA,
  append = FALSE,
  holidays = FALSE,
  country = "Venezuela",
  currency_pair = NA,
  quiet = FALSE
)
}
\arguments{
\item{dates}{Vector or dataframe. Non-date/time columns will be
automatically ignored/extracted.}

\item{drop}{Boolean. Should the original date/time columns be
kept in the results? Only valid when input is a dataframe.}

\item{only}{Character or vector. Which columns do you wish to process? If
non are explicitly defined, all will be processed}

\item{append}{Boolean. Append results to existing data.frame? If FALSE,
only calculated values will be returned.}

\item{holidays}{Boolean. Include holidays as new columns?}

\item{country}{Character or vector. For which countries should the holidays
be included?}

\item{currency_pair}{Character. Which currency exchange do you
wish to get the history from? i.e, USD/COP, EUR/USD...}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
data.frame with additional features calculated out of time or date vectors.
}
\description{
This function lets the user automatically create new columns out
of a dataframe or vector with date/time variables.
}
\examples{
df <- data.frame(
  dates = sample(seq(Sys.Date() - 365, Sys.Date(), by = 1), 50),
  times = sample(seq(Sys.time() - 1e7, Sys.time(), by = 1), 50)
)

# Input as a vector or dataframe
date_feats(df, drop = TRUE, quiet = TRUE) \%>\% head(10)

# Holidays given a date range and country
\dontrun{
hol <- date_feats(
  seq(Sys.Date() - 365, Sys.Date(), by = 1),
  holidays = TRUE,
  country = "Venezuela"
)
head(hol[!is.na(hol$holiday_name), ])
}
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Feature Engineering: 
\code{\link{holidays}()},
\code{\link{ohse}()}

Other One Hot Encoding: 
\code{\link{holidays}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()}
}
\concept{Data Wrangling}
\concept{Feature Engineering}
\concept{One Hot Encoding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/robyn.R
\name{robyn_performance}
\alias{robyn_performance}
\title{Robyn: Dynamic Performance and Contribution Report}
\usage{
robyn_performance(
  InputCollect,
  OutputCollect,
  start_date = NULL,
  end_date = NULL,
  solID = NULL,
  totals = TRUE,
  non_promo = FALSE,
  marginals = FALSE,
  carryovers = FALSE,
  new_version = FALSE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{InputCollect, OutputCollect}{Robyn output objects.}

\item{start_date, end_date}{Date. Start and end date to filter
the data to be reported.}

\item{solID}{Character. Single ID of the model to report. If there's
only one available in OutputCollect, no need to define.}

\item{totals}{Boolean. Add total rows. This includes summary rows
(promotional which is paid and organic channels, baseline, grand total).}

\item{non_promo}{Boolean. Add non-promotional responses as well?}

\item{marginals}{Boolean. Include mROAS or mCPA marginal performance metric
as an additional column called "marginal". Calculations are based on
mean spend and mean response with mean carryover results,
between \code{start_date} and \code{end_date}.}

\item{carryovers}{Boolean. Add mean percentage of carryover response for
date range between \code{start_date} and \code{end_date} on paid channels.
Keep in mind organic variables also have carryover but currently not showing.}

\item{new_version}{Boolean. Use dev version's new function for marginal
calculations (if available)?}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters.}
}
\value{
data.frame with results on ROAS/CPA, spend, response, contribution
per channel, with or without total rows.
}
\description{
Given a date range, calculate specific and total performance and
contribution for each of your marketing and non-marketing channels.
}
\examples{
\dontrun{
# You may load an exported model to recreate Robyn objects
mod <- Robyn::robyn_recreate(json_file = "your_model.json")
robyn_performance(mod$InputCollect, mod$OutputCollect)
}
}
\seealso{
Other Robyn: 
\code{\link{robyn_hypsbuilder}()},
\code{\link{robyn_marginal}()},
\code{\link{robyn_modelselector}()}
}
\concept{Robyn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusters.R
\name{clusterKmeans}
\alias{clusterKmeans}
\title{Automated K-Means Clustering + PCA/t-SNE}
\usage{
clusterKmeans(
  df,
  k = NULL,
  wss_var = 0,
  limit = 15,
  drop_na = TRUE,
  ignore = NULL,
  ohse = TRUE,
  norm = TRUE,
  algorithm = c("Hartigan-Wong", "Lloyd", "Forgy", "MacQueen"),
  dim_red = "PCA",
  comb = c(1, 2),
  seed = 123,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{Dataframe}

\item{k}{Integer. Number of clusters}

\item{wss_var}{Numeric. Used to pick automatic \code{k} value,
when \code{k} is \code{NULL} based on WSS variance while considering
\code{limit} clusters. Values between (0, 1). Default value could be
0.05 to consider convergence.}

\item{limit}{Integer. How many clusters should be considered?}

\item{drop_na}{Boolean. Should NA rows be removed?}

\item{ignore}{Character vector. Names of columns to ignore.}

\item{ohse}{Boolean. Do you wish to automatically run one hot
encoding to non-numerical columns?}

\item{norm}{Boolean. Should the data be normalized?}

\item{algorithm}{character: may be abbreviated.  Note that
    \code{"Lloyd"} and \code{"Forgy"} are alternative names for one
    algorithm.}

\item{dim_red}{Character. Select dimensionality reduction technique.
Pass any of: \code{c("PCA", "tSNE", "all", "none")}.}

\item{comb}{Vector. Which columns do you wish to plot? Select which
two variables by name or column position.}

\item{seed}{Numeric. Seed for reproducibility}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{...}{Additional parameters to pass sub-functions.}
}
\value{
List. If no \code{k} is provided, contains \code{nclusters} and
\code{nclusters_plot} to determine optimal \code{k} given their WSS (Within
Groups Sum of Squares). If \code{k} is provided, additionally we get:
\itemize{
  \item \code{df} data.frame with original \code{df} plus \code{cluster} column
  \item \code{clusters} integer which is the same as \code{k}
  \item \code{fit} kmeans object used to fit clusters
  \item \code{means} data.frame with means and counts for each cluster
  \item \code{correlations} plot with correlations grouped by clusters
  \item \code{PCA} list with PCA results (when \code{dim_red="PCA"})
  \item \code{tSNE} list with t-SNE results (when \code{dim_red="tSNE"})
}
}
\description{
This function lets the user cluster a whole data.frame automatically.
As you might know, the goal of kmeans is to group data points into
distinct non-overlapping subgroups. If needed, one hot encoding will
be applied to categorical values automatically with this function.
For consideration: Scale/standardize the data when applying kmeans.
Also, kmeans assumes spherical shapes of clusters and does not work well
when clusters are in different shapes such as elliptical clusters.
}
\examples{
Sys.unsetenv("LARES_FONT") # Temporal
data("iris")
df <- subset(iris, select = c(-Species))

# If dataset has +5 columns, feel free to reduce dimenstionalities
# with reduce_pca() or reduce_tsne() first

# Find optimal k
check_k <- clusterKmeans(df, limit = 10)
check_k$nclusters_plot
# Or pick k automatically based on WSS variance
check_k <- clusterKmeans(df, wss_var = 0.05, limit = 10)
# You can also use our other functions:
# clusterOptimalK(df) and clusterVisualK(df)

# Run with selected k
clusters <- clusterKmeans(df, k = 3)
names(clusters)

# Cross-Correlations for each cluster
plot(clusters$correlations)

# PCA Results (when dim_red = "PCA")
plot(clusters$PCA$plot_explained)
plot(clusters$PCA$plot)
}
\seealso{
Other Clusters: 
\code{\link{clusterOptimalK}()},
\code{\link{clusterVisualK}()},
\code{\link{reduce_pca}()},
\code{\link{reduce_tsne}()}
}
\concept{Clusters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maze.R
\name{maze_solve}
\alias{maze_solve}
\alias{print.maze_solve}
\alias{maze_gridsearch}
\title{Maze Solver, inspired by Micromouse competitions}
\usage{
maze_solve(
  maze,
  start = c(1, 1),
  end = dim(maze),
  inertia = FALSE,
  aim = TRUE,
  diagonal = TRUE,
  random = FALSE,
  timeout = 4,
  quiet = FALSE,
  seed = NULL,
  ...
)

\method{print}{maze_solve}(x, ...)

maze_gridsearch(
  maze,
  start = c(2, 2),
  end = round(dim(maze)/2),
  quiet = TRUE,
  seed = 123,
  ...
)
}
\arguments{
\item{maze}{Matrix. Using 0 for open space and 1 for walls.}

\item{start, end}{Integer vector, length 2. Start and end coordinates.}

\item{inertia}{Boolean. When enabled, algorithm will check for new
directions only when impossible to continue in a straight line.}

\item{aim}{Boolean. When enabled, algorithm will try first the directions
closer to the \code{end} point, ranked and sorted by shorter distances.}

\item{diagonal}{Boolean. When enabled, algorithm will have 8 degrees of
freedom to move, if not, only 4 (up, down, left, right).}

\item{random}{Boolean. When enabled, algorithm will pick next direction
randomly.}

\item{timeout}{Numeric. How many seconds set for timeout to force
algorithm to stop trying new paths?}

\item{quiet}{Boolean. Keep quiet? If not, print results}

\item{seed}{Numeric. Seed to replicate random results.}

\item{...}{Additional parameters passed to \code{corr}}

\item{x}{maze_solve object}
}
\value{
List with data.frame containing solved solution, data.frame with
path coordinates and directions, steps counter and turns counter.
}
\description{
Modified recursive depth-first search (DFS) algorithm to solve mazes.
It explores the maze by recursively moving to adjacent cells until it finds a
path from the starting point to the destination. Contains options to
maximize paths by trying to turn less, allowing diagonal turns, prioritizing
turns that chooses next step pointing towards the end point, and a grid
search combining parameters to find best route.
}
\examples{
micromouse <- matrix(c(
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1,
  1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1,
  1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1,
  1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1,
  1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1,
  1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1,
  1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
), nrow = 12, byrow = TRUE)
maze_solve(micromouse, start = c(2, 2), end = c(7, 7))
}
\seealso{
Other Games: 
\code{\link{scrabble_dictionary}()},
\code{\link{sudoku_solver}()},
\code{\link{wordle_check}()}
}
\concept{Games}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{balance_data}
\alias{balance_data}
\title{Balance Binary Data by Resampling: Under-Over Sampling}
\usage{
balance_data(df, var, rate = 1, target = "auto", seed = 0, quiet = FALSE)
}
\arguments{
\item{df}{Vector or Dataframe. Contains different variables in each
column, separated by a specific character}

\item{var}{Variable. Which variable should we used to re-sample dataset?}

\item{rate}{Numeric. How many X for every Y we need? Default: 1. If there are
more than 2 unique values, rate will represent percentage for number of rows}

\item{target}{Character. If binary, which value should be reduced? If kept in
\code{"auto"}, then the most frequent value will be reduced.}

\item{seed}{Numeric. Seed to replicate and obtain same values}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
data.frame. Reduced sampled data.frame following the \code{rate} of
appearance of a specific variable.
}
\description{
This function lets the user balance a given data.frame by resampling
with a given relation rate and a binary feature.
}
\examples{
data(dft) # Titanic dataset
df <- balance_data(dft, Survived, rate = 1)
df <- balance_data(dft, .data$Survived, rate = 0.5, target = "TRUE")
}
\seealso{
Other Data Wrangling: 
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facebook.R
\name{fb_token}
\alias{fb_token}
\title{Facebook's Long-Life User API Token}
\usage{
fb_token(app_id, app_secret, token, api_version = NULL)
}
\arguments{
\item{app_id, app_secret}{Character. Application ID and Secret.}

\item{token}{Character. User token, created with
\href{https://developers.facebook.com/tools/explorer}{API Graph}
or with this same \code{fb_token()}'s token.}

\item{api_version}{Character. Facebook API version.}
}
\value{
Character. String with token requested. If successful, it'll contain
an attribute called "expiration" with date and time of expiration.
}
\description{
Using a 1-hour generic user token you can generate a 60 day token.
You will need to have an App ID and App secret, and a valid token.
Generate a new valid User Token with the
\href{https://developers.facebook.com/tools/explorer}{API Graph}.
}
\details{
More info: \href{https://developers.facebook.com/documentation/facebook-login/guides/access-tokens/get-long-lived/}{Long-Lived Access Tokens}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()},
\code{\link{slackSend}()}

Other Meta: 
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()}
}
\concept{API}
\concept{Meta}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{categ_reducer}
\alias{categ_reducer}
\title{Reduce categorical values}
\usage{
categ_reducer(
  df,
  var,
  nmin = 0,
  pmin = 0,
  pcummax = 100,
  top = NA,
  pvalue_max = 1,
  cor_var = "tag",
  limit = 20,
  other_label = "other",
  ...
)
}
\arguments{
\item{df}{Categorical Vector}

\item{var}{Variable. Which variable do you wish to reduce?}

\item{nmin}{Integer. Number of minimum times a value is repeated}

\item{pmin}{Numerical. Percentage of minimum times a value is repeated}

\item{pcummax}{Numerical. Top cumulative percentage of most
repeated values}

\item{top}{Integer. Keep the n most frequently repeated values}

\item{pvalue_max}{Numeric (0-1]. Max pvalue categories}

\item{cor_var}{Character. If pvalue_max < 1, you must define which
column name will be compared with (numerical or binary).}

\item{limit}{Integer. Limit one hot encoding to the n most frequent
values of each column. Set to \code{NA} to ignore argument.}

\item{other_label}{Character. With which text do you wish to replace
the filtered values with?}

\item{...}{Additional parameters.}
}
\value{
data.frame \code{df} on which \code{var} has been transformed
}
\description{
This function lets the user reduce categorical values in a vector.
It is tidyverse friendly for use on pipelines
}
\examples{
data(dft) # Titanic dataset
categ_reducer(dft, Embarked, top = 2) \%>\% freqs(Embarked)
categ_reducer(dft, Ticket, nmin = 7, other_label = "Other Ticket") \%>\% freqs(Ticket)
categ_reducer(dft, Ticket, pvalue_max = 0.05, cor_var = "Survived") \%>\% freqs(Ticket)
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_text.R
\name{warnifnot}
\alias{warnifnot}
\title{Test the Truth of R Expressions and Warn}
\usage{
warnifnot(...)
}
\arguments{
\item{...}{any R expression, which should evaluate to TRUE}
}
\description{
If the expression in ... is not \code{TRUE}, \code{warning} is called,
producing a warning message indicating the expression which was not true.
}
\examples{
warnifnot(TRUE)
warnifnot(FALSE)
warnifnot(1 + 1 == 3)
}
\seealso{
Other Tools: 
\code{\link{autoline}()},
\code{\link{bind_files}()},
\code{\link{bring_api}()},
\code{\link{cal_split}()},
\code{\link{chr2num}()},
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{dont_sleep}()},
\code{\link{export_plot}()},
\code{\link{export_results}()},
\code{\link{files_functions}()},
\code{\link{font_exists}()},
\code{\link{formatColoured}()},
\code{\link{formatHTML}()},
\code{\link{glued}()},
\code{\link{grepm}()},
\code{\link{h2o_selectmodel}()},
\code{\link{haveInternet}()},
\code{\link{image_metadata}()},
\code{\link{importxlsx}()},
\code{\link{ip_data}()},
\code{\link{json2vector}()},
\code{\link{list_cats}()},
\code{\link{listfiles}()},
\code{\link{mail_send}()},
\code{\link{markdown2df}()},
\code{\link{move_files}()},
\code{\link{msplit}()},
\code{\link{myip}()},
\code{\link{quiet}()},
\code{\link{read.file}()},
\code{\link{statusbar}()},
\code{\link{tic}()},
\code{\link{try_require}()},
\code{\link{updateLares}()},
\code{\link{what_size}()}
}
\concept{Tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chatgpt.R
\name{gpt_prompter}
\alias{gpt_prompter}
\title{Structured Prompt Builder for LLM (ChatGPT)}
\usage{
gpt_prompter(
  instruction = NULL,
  input = NULL,
  context = NULL,
  output = NULL,
  quiet = TRUE,
  ...
)
}
\arguments{
\item{instruction, input, context, output}{Character or vector.
You do not need all the four elements for a prompt and the
format depends on the task at hand.}

\item{quiet}{Boolean. Should the written prompt be printed or not?}

\item{...}{Additional parameters. You can pass \code{cols} parameter to
explicitly set column names.}
}
\value{
(Invisible) list with written prompt and elements provided.
}
\description{
Build standard prompts to get the right outcomes using the four theoretical
elements required to build standard and cleaner asks: instruction, input, context,
and output. Inspired by the
\href{https://www.promptingguide.ai/}{Prompt Engineering Guide} free guide.
Remember to start simple and be very specific to get exactly what you need.
}
\section{Elements of a Prompt}{

\describe{
  \item{Instruction}{a specific task or instruction you want the model to perform.
  It usually starts with a verb given its an instruction.}
  \item{Input}{input data, elements or question that we are interested to find a response for,}
  \item{Context}{external information, additional context or references to steer
  the model to better responses.}
  \item{Output}{type or format of the output.}
}
}

\examples{
# A simple formatted table with data
# Note: I mostly use output = "table" and enabled an auxiliary enrichment prompt
(p <- gpt_prompter(instruction = "Capitals of the world", output = "table"))

# Classify
p <- gpt_prompter(
  instruction = "For each of the inputs, classify using only the options in context",
  input = c("Molecule", "Elephant", "Milky Way", "Cat", "Planet Earth"),
  context = c("Big", "Medium", "Small"),
  output = "table",
  # This cols parameter is auxiliary
  cols = c("Input", "Category"),
  quiet = FALSE
)

# Tag all categories that apply
p <- gpt_prompter(
  instruction = paste(
    "For each of the inputs, provide which of the",
    "context values apply as correct tags using TRUE/FALSE"
  ),
  input = c("I love chocolate", "I hate chocolate", "I like Coke", "Who am I?", "T-REX"),
  context = c("food", "positive", "negative", "beverage"),
  output = "table",
  quiet = FALSE
)

# Extract information from strings
p <- gpt_prompter(
  instruction = "For each of the inputs, extract each of the information asked in context",
  input = c("My mail is 123@test.com", "30 Main St, NY, USA", "+82 2-312-3456", "$1.5M"),
  context = c("email", "full state name", "country of phone", "full non-abbreviated number"),
  output = "table",
  cols = c("Input", "Element_to_extract", "Value"),
  quiet = FALSE
)

# Translate to several languages
p <- gpt_prompter(
  instruction = "For each of the inputs, translate to the respective languages in context",
  input = rep("I love you with all my heart", 5),
  context = c("spanish", "chinese", "japanese", "russian", "german"),
  output = "table",
  cols = c("Input", "Language", "Translation"),
  quiet = FALSE
)

# Format date values
p <- gpt_prompter(
  instruction = paste(
    "For each of the inputs,",
    "standardize and format all values to the format in context"
  ),
  input = c("March 27th, 2021", "12-25-2023 3:45PM", "01.01.2000", "29 Feb 92"),
  context = "ISO Date getting rid of time stamps",
  output = "table",
  cols = c("Input", "Formatted"),
  quiet = FALSE
)

# Convert units
p <- gpt_prompter(
  instruction = paste(
    "For each of the inputs,",
    "provide new converted values using the units in context"
  ),
  input = c("50C", "300K", "100F", "0F", "32C", "0K"),
  context = "Fahrenheit",
  output = "table",
  cols = c("Input", "Original_Unit", "Total_Value", "Converted_Value", "New_Unit"),
  quiet = FALSE
)

# Read a text and answer a question related to it
gpt_prompter(
  instruction = "read",
  context = "Long text here",
  input = "Question here"
)$prompt
}
\seealso{
Other ChatGPT: 
\code{\link{gpt_ask}()}

Other LLM: 
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()}
}
\concept{ChatGPT}
\concept{LLM}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrangling.R
\name{vector2text}
\alias{vector2text}
\alias{v2t}
\title{Convert a vector into a comma separated text}
\usage{
vector2text(vector, sep = ", ", quotes = TRUE, force_single = FALSE, and = "")

v2t(vector, sep = ", ", quotes = TRUE, force_single = FALSE, and = "")
}
\arguments{
\item{vector}{Vector. Vector with more than 1 observation.}

\item{sep}{Character. String text wished to insert between values.}

\item{quotes}{Boolean. Bring simple quotes for each observation.}

\item{force_single}{Boolean. Force single quotes by replacing \code{\"}.}

\item{and}{Character. Add 'and' or something before last observation.
Not boolean variable so it can be used on other languages. Note that
the last comma will be suppressed if \code{Sys.getenv("LARES_NUMFORMAT")}
is set to \code{1} and you have less than 3 values.}
}
\value{
Vector pasting \code{vector} values into a single string
}
\description{
Convert a vector into a comma separated text
}
\examples{
vector2text(LETTERS[1:5])
vector2text(c(1:5), quotes = FALSE)
vector2text(c(1:5), quotes = FALSE, sep = "-")
vector2text(c(1:5), and = "and also")
vector2text(c("Text", "R's"), force_single = TRUE)
# Shorter function with same purpose
v2t(LETTERS[1:5])
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}
}
\concept{Data Wrangling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_metrics.R
\name{conf_mat}
\alias{conf_mat}
\title{Confussion Matrix}
\usage{
conf_mat(tag, score, thresh = 0.5, sense = ">=", diagonal = TRUE, plot = FALSE)
}
\arguments{
\item{tag}{Vector. Real known label}

\item{score}{Vector. Predicted value or model's result}

\item{thresh}{Integer. Threshold for selecting binary or regression
models: this number is the threshold of unique values we should
have in \code{'tag'} (more than: regression; less than: classification)}

\item{sense}{Character. Inequation sense for threshold: <, <=, >=, >}

\item{diagonal}{Boolean. \code{FALSE} to convert diagonal numbers to
zeroes. Ideal to detect must confusing categories.}

\item{plot}{Boolean. Plot result? Uses \code{mplot_conf()}}
}
\value{
data.frame. Result of counting \code{tag} and \code{score}'s tag
given a \code{thresh}old, similar to \code{base::table()}.
}
\description{
This function calculates a Confussion Matrix using crosstab for
2 or more categories. You can either set the score and threshold
or the labels you wish to cross with.
}
\details{
You may use \code{mplot_conf()} or set \code{plot=TRUE}.
}
\examples{
data(dfr) # Results for AutoML Predictions
lapply(dfr[c(1, 2)], head)

# Results for Binomial Model
conf_mat(dfr$class2$tag, dfr$class2$scores)
conf_mat(dfr$class2$tag, dfr$class2$scores, thresh = 0.3)
conf_mat(dfr$class2$tag, dfr$class2$scores, sense = "<=")

# Results for Multi-Categorical Model
conf_mat(dfr$class3$tag, dfr$class3$score)
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Model metrics: 
\code{\link{ROC}()},
\code{\link{errors}()},
\code{\link{gain_lift}()},
\code{\link{loglossBinary}()},
\code{\link{model_metrics}()}
}
\concept{Machine Learning}
\concept{Model metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_functions.R
\name{h2o_automl}
\alias{h2o_automl}
\alias{plot.h2o_automl}
\alias{print.h2o_automl}
\title{Automated H2O's AutoML}
\usage{
h2o_automl(
  df,
  y = "tag",
  ignore = NULL,
  train_test = NA,
  split = 0.7,
  weight = NULL,
  target = "auto",
  balance = FALSE,
  impute = FALSE,
  no_outliers = TRUE,
  unique_train = TRUE,
  center = FALSE,
  scale = FALSE,
  thresh = 10,
  seed = 0,
  nfolds = 5,
  max_models = 3,
  max_time = 10 * 60,
  start_clean = FALSE,
  exclude_algos = c("StackedEnsemble", "DeepLearning"),
  include_algos = NULL,
  plots = TRUE,
  alarm = TRUE,
  quiet = FALSE,
  print = TRUE,
  save = FALSE,
  subdir = NA,
  project = "AutoML Results",
  model_name = NULL,
  verbosity = NULL,
  ...
)

\method{plot}{h2o_automl}(x, ...)

\method{print}{h2o_automl}(x, importance = TRUE, ...)
}
\arguments{
\item{df}{Dataframe. Dataframe containing all your data, including
the dependent variable labeled as \code{'tag'}. If you want to define
which variable should be used instead, use the \code{y} parameter.}

\item{y}{Variable or Character. Name of the dependent variable or response.}

\item{ignore}{Character vector. Force columns for the model to ignore}

\item{train_test}{Character. If needed, \code{df}'s column name with 'test'
and 'train' values to split}

\item{split}{Numeric. Value between 0 and 1 to split as train/test
datasets. Value is for training set. Set value to 1 to train with all
available data and test with same data (cross-validation will still be
used when training). If \code{train_test} is set, value will be overwritten
with its real split rate.}

\item{weight}{Column with observation weights. Giving some observation a
weight of zero is equivalent to excluding it from the dataset; giving an
observation a relative weight of 2 is equivalent to repeating that
row twice. Negative weights are not allowed.}

\item{target}{Value. Which is your target positive value? If
set to \code{'auto'}, the target with largest \code{mean(score)} will be
selected. Change the value to overwrite. Only used when binary
categorical model.}

\item{balance}{Boolean. Auto-balance train dataset with under-sampling?}

\item{impute}{Boolean. Fill \code{NA} values with MICE?}

\item{no_outliers}{Boolean/Numeric. Remove \code{y}'s outliers from the dataset?
Will remove those values that are farther than n standard deviations from
the dependent variable's mean (Z-score). Set to \code{TRUE} for default (3)
or numeric to set a different multiplier.}

\item{unique_train}{Boolean. Keep only unique row observations for training data?}

\item{center, scale}{Boolean. Using the base function scale, do you wish
to center and/or scale all numerical values?}

\item{thresh}{Integer. Threshold for selecting binary or regression
models: this number is the threshold of unique values we should
have in \code{'tag'} (more than: regression; less than: classification)}

\item{seed}{Integer. Set a seed for reproducibility. AutoML can only
guarantee reproducibility if max_models is used because max_time is
resource limited.}

\item{nfolds}{Number of folds for k-fold cross-validation. Must be >= 2; defaults to 5. Use 0 to disable cross-validation;
this will also disable Stacked Ensemble (thus decreasing the overall model performance).}

\item{max_models, max_time}{Numeric. Max number of models and seconds
you wish for the function to iterate. Note that max_models guarantees
reproducibility and max_time not (because it depends entirely on your
machine's computational characteristics)}

\item{start_clean}{Boolean. Erase everything in the current h2o
instance before we start to train models? You may want to keep other models
or not. To group results into a custom common AutoML project, you may
use \code{project_name} argument.}

\item{exclude_algos, include_algos}{Vector of character strings. Algorithms
to skip or include during the model-building phase. Set NULL to ignore.
When both are defined, only \code{include_algos} will be valid.}

\item{plots}{Boolean. Create plots objects?}

\item{alarm}{Boolean. Ping (sound) when done. Requires \code{beepr}.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}

\item{print}{Boolean. Print summary when process ends?}

\item{save}{Boolean. Do you wish to save/export results into your
working directory?}

\item{subdir}{Character. In which directory do you wish to save
the results? Working directory as default.}

\item{project}{Character. Your project's name}

\item{model_name}{Character. Optional custom name for the model. If provided,
this name will be used when saving the model with \code{export_results()}. If NULL
(default), the H2O-generated model ID will be used.}

\item{verbosity}{Character. Verbosity of the backend messages printed during training.
Must be one of NULL (live log disabled), "debug", "info", "warn", "error". Defaults to NULL.}

\item{...}{Additional parameters on \code{h2o::h2o.automl}}

\item{x}{h2o_automl object}

\item{importance}{Boolean. Print important variables?}
}
\value{
List. Trained model, predicted scores and datasets used, performance
metrics, parameters, importance data.frame, seed, and plots when \code{plots=TRUE}.
}
\description{
This function lets the user create a robust and fast model, using
H2O's AutoML function. The result is a list with the best model,
its parameters, datasets, performance metrics, variables
importance, and plots. Read more about the \code{h2o_automl()} pipeline
\href{https://laresbernardo.github.io/lares/articles/machine-learning.html}{here}.
}
\details{
For additional tutorials and examples:
\itemize{
  \item \href{https://laresbernardo.github.io/lares/articles/machine-learning.html}{Machine Learning with H2O Package}
  \item \href{https://laresbernardo.github.io/lares/reference/mplot_roc.html}{Understanding ROC Curves}
}
}
\section{List of algorithms}{

\href{https://docs.h2o.ai/h2o/latest-stable/h2o-docs/automl.html}{-> Read more here}
\describe{
  \item{DRF}{Distributed Random Forest, including Random Forest (RF)
  and Extremely-Randomized Trees (XRT)}
  \item{GLM}{Generalized Linear Model}
  \item{XGBoost}{eXtreme Grading Boosting}
  \item{GBM}{Gradient Boosting Machine}
  \item{DeepLearning}{Fully-connected multi-layer artificial neural network}
  \item{StackedEnsemble}{Stacked Ensemble}
}
}

\section{Methods}{

\describe{
  \item{print}{Use \code{print} method to print models stats and summary}
  \item{plot}{Use \code{plot} method to plot results using \code{mplot_full()}}
}
}

\examples{
\dontrun{
# CRAN
data(dft) # Titanic dataset
dft <- subset(dft, select = -c(Ticket, PassengerId, Cabin))

# Classification: Binomial - 2 Classes
r <- h2o_automl(dft, y = Survived, max_models = 1, impute = FALSE, target = "TRUE", alarm = FALSE)

# Let's see all the stuff we have inside:
lapply(r, names)

# Classification: Multi-Categorical - 3 Classes
r <- h2o_automl(dft, Pclass, ignore = c("Fare", "Cabin"), max_time = 30, plots = FALSE)

# Regression: Continuous Values
r <- h2o_automl(dft, y = "Fare", ignore = c("Pclass"), exclude_algos = NULL, quiet = TRUE)
print(r)

# WITH PRE-DEFINED TRAIN/TEST DATAFRAMES
splits <- msplit(dft, size = 0.8)
splits$train$split <- "train"
splits$test$split <- "test"
df <- rbind(splits$train, splits$test)
r <- h2o_automl(df, "Survived", max_models = 1, train_test = "split")
}
}
\seealso{
Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{impute}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}
}
\concept{Machine Learning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_mining.R
\name{textFeats}
\alias{textFeats}
\title{Create features out of text}
\usage{
textFeats(text, auto = TRUE, contains = NA, prc = FALSE)
}
\arguments{
\item{text}{Character vector}

\item{auto}{Boolean. Auto create some useful parameters?}

\item{contains}{Character vector. Which columns do you wish to add
with a contains (counter) string validator?}

\item{prc}{Boolean. Also add percentage of each column compared with length?}
}
\value{
data.frame with additional features based on \code{text}.
}
\description{
This function creates a data.frame with features based on a text vector
}
\examples{
textFeats("Bernardo Lares")
textFeats("Bernardo Lares 123!", prc = TRUE)
textFeats("I'm 100\% Lares...", contains = c("Lares", "lares"))
textFeats(c("GREAT library!!", "Have you tried this 2?", "Happy faces :D :-)"))
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{impute}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Text Mining: 
\code{\link{cleanText}()},
\code{\link{ngrams}()},
\code{\link{remove_stopwords}()},
\code{\link{replaceall}()},
\code{\link{sentimentBreakdown}()},
\code{\link{textCloud}()},
\code{\link{textTokenizer}()},
\code{\link{topics_rake}()}
}
\concept{Data Wrangling}
\concept{Text Mining}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{toon_reduction}
\alias{toon_reduction}
\title{Calculate Character Reduction from R Object to TOON Format}
\usage{
toon_reduction(obj, ...)
}
\arguments{
\item{obj}{Any R object (list, vector, data frame, or string).}

\item{...}{Additional arguments passed to jsonlite::toJSON()}
}
\value{
A numeric value representing the character reduction ratio (0 to 1).
}
\description{
Calculates the percentage reduction in the number of characters when
converting an R object to its TOON string representation compared to
its JSON representation (or the string itself if the input is a single string).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dalex_explainers.R
\name{h2o_explainer}
\alias{h2o_explainer}
\alias{dalex_explainer}
\title{DALEX Explainer for H2O}
\usage{
h2o_explainer(df, model, y = "tag", ignore = NULL, ...)
}
\arguments{
\item{df}{Dataframe. Must contain all columns and predictions}

\item{model}{Model object (H2O)}

\item{y}{Character or Variable name. Variable's column name.}

\item{ignore}{Character vector. Which columns should be ignored?}

\item{...}{Additional parameters to pass to \code{h2o_predict_model} or
\code{h2o_predict_MOJO}.}
}
\value{
List; explainer. Containing the model, data, y, predict_function,
y_hat, residuals, class, label, model_info, residual_function, and weights.
}
\description{
DALEX helper function to create an \code{explainer} object using
a \code{h2o} trained model.
}
\examples{
# You must have "DALEX" library to use this auxiliary function:
\dontrun{
data(dft) # Titanic dataset

# TRAIN A SIMPLE MODEL
dfm <- h2o_automl(dft,
  y = "Survived",
  ignore = c("Ticket", "PassengerId", "Cabin"),
  max_models = 1
)

# EXPLAINER
explainer <- h2o_explainer(df = dfm$datasets$test, model = dfm$model, y = "Survived")
explainer$data <- na.omit(explainer$data)

# CATEGORICAL EXAMPLE
class <- dalex_variable(explainer, vars = c("Pclass", "Sex"))
class$plot

# NUMERICAL EXAMPLE
num <- dalex_variable(explainer, vars = c("Fare", "Age"))
num$plot

# LOCAL EXAMPLE
local <- dalex_local(explainer, row = 1)
# OR YOU COULD MANUALLY INPUT THE OBSERVATION
local <- dalex_local(explainer, observation = explainer$data[1, ])
local$plot

# xai2shiny's UI (needs to be installed from ModelOriented/xai2shiny)
xai2shiny(explainer, run = TRUE)
}
}
\seealso{
Other Interpretability: 
\code{\link{dalex_local}()},
\code{\link{dalex_residuals}()},
\code{\link{dalex_variable}()}
}
\concept{Interpretability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/google_trends.R
\name{gtrends_related}
\alias{gtrends_related}
\alias{trendsRelated}
\alias{gtrends_time}
\alias{trendsTime}
\title{Google Trends: Related and Time Plots}
\usage{
gtrends_related(gtrend, top = NA, title = NA, note = NA, exclude = NULL)

gtrends_time(gtrend, title = NA)
}
\arguments{
\item{gtrend}{List. Result from \code{gtrendsR::gtrends(keyword, geo, time)}}

\item{top}{Integer. Filter top n results only.}

\item{title}{Character. Custom title for the plot.}

\item{note}{Character. Add a note to the plot if needed.}

\item{exclude}{Character vector. Which observations do you wish to exclude?}
}
\value{
plot for Google Trend's results input \code{gtrend}.

Plot for Google Trend's results input \code{gtrend}.
}
\description{
This function creates a plot with Google Trend's related topics
and queries, and let the user compare different keywords.

This function creates a plot with google trend's data on timelines
and let the user compare different keywords.
}
\seealso{
Other Scrapper: 
\code{\link{filesGD}()},
\code{\link{holidays}()},
\code{\link{ip_data}()},
\code{\link{mp3_get}()},
\code{\link{readGS}()},
\code{\link{splot_summary}()},
\code{\link{stocks_quote}()}

Other Google: 
\code{\link{filesGD}()},
\code{\link{queryGA}()},
\code{\link{readGS}()}
}
\concept{Google}
\concept{Scrapper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missings.R
\name{impute}
\alias{impute}
\title{Impute Missing Values (using MICE)}
\usage{
impute(df, m = 5, iters = 5, seed = 0, quiet = FALSE)
}
\arguments{
\item{df}{Dataframe. Dataframe to transform.}

\item{m}{Integer. Number of multiple imputations.}

\item{iters}{Integer. Number of iterations.}

\item{seed}{Integer. Set a seed for reproducibility.}

\item{quiet}{Boolean. Keep quiet? If not, informative messages will be shown.}
}
\value{
data.frame with imputed values.
}
\description{
This function uses the MICE methodology to impute missing values.
}
\seealso{
Other Data Wrangling: 
\code{\link{balance_data}()},
\code{\link{categ_reducer}()},
\code{\link{cleanText}()},
\code{\link{date_cuts}()},
\code{\link{date_feats}()},
\code{\link{file_name}()},
\code{\link{formatHTML}()},
\code{\link{holidays}()},
\code{\link{left}()},
\code{\link{normalize}()},
\code{\link{num_abbr}()},
\code{\link{ohe_commas}()},
\code{\link{ohse}()},
\code{\link{quants}()},
\code{\link{removenacols}()},
\code{\link{replaceall}()},
\code{\link{replacefactor}()},
\code{\link{textFeats}()},
\code{\link{textTokenizer}()},
\code{\link{vector2text}()},
\code{\link{year_month}()},
\code{\link{zerovar}()}

Other Machine Learning: 
\code{\link{ROC}()},
\code{\link{conf_mat}()},
\code{\link{export_results}()},
\code{\link{gain_lift}()},
\code{\link{h2o_automl}()},
\code{\link{h2o_predict_MOJO}()},
\code{\link{h2o_selectmodel}()},
\code{\link{iter_seeds}()},
\code{\link{lasso_vars}()},
\code{\link{model_metrics}()},
\code{\link{model_preprocess}()},
\code{\link{msplit}()}

Other Missing Values: 
\code{\link{missingness}()}
}
\concept{Data Wrangling}
\concept{Machine Learning}
\concept{Missing Values}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colour_palettes.R
\name{plot_palette}
\alias{plot_palette}
\title{Plot Palette Colours}
\usage{
plot_palette(fill, colour = "black", id = NA, limit = 12, ...)
}
\arguments{
\item{fill}{Vector. List of colours for fills.}

\item{colour}{Vector. List of colours for colours.}

\item{id}{Vector. ID for each color.}

\item{limit}{Integer. Show only first n values.}

\item{...}{Additional parameters.}
}
\value{
Plot with \code{fill} colours and \code{colour} counter-colours
if provided.
}
\description{
This function plots a list of colours
}
\examples{
# Simply pass a vector
pal <- lares_pal("simple")
plot_palette(pal)
# Or fill + color named vector
pal <- lares_pal("pal")
plot_palette(fill = names(pal), colour = as.vector(pal))
}
\seealso{
Other Themes: 
\code{\link{gg_fill_customs}()},
\code{\link{lares_pal}()},
\code{\link{theme_lares}()}
}
\concept{Themes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slack.R
\name{slackSend}
\alias{slackSend}
\title{Send Slack Message (Webhook)}
\usage{
slackSend(text, title = "", pretext = "", hook = NA, creds = NA)
}
\arguments{
\item{text, title, pretext}{Character. Content on you Slack message.}

\item{hook}{Character. Web hook URL. Ths value will be overwritten by
creds if correctly used.}

\item{creds}{Character. Credential's dir (see \code{get_creds()}). Set
hook URL into the "slack" list in your YML file. Will use first value.}
}
\value{
Invisible POST response
}
\description{
This function send a Slack message using its Webhooks.
}
\details{
For more help, you can follow the
\href{https://api.slack.com/messaging/webhooks#posting_with_webhooks}{Sending messages using Incoming Webhooks}
original documentarion.
}
\examples{
\dontrun{
slackSend(text = "This is a message", title = "TEST", pretext = Sys.info()["user"])
}
}
\seealso{
Other API: 
\code{\link{bring_api}()},
\code{\link{fb_accounts}()},
\code{\link{fb_ads}()},
\code{\link{fb_creatives}()},
\code{\link{fb_insights}()},
\code{\link{fb_process}()},
\code{\link{fb_report_check}()},
\code{\link{fb_rf}()},
\code{\link{fb_token}()},
\code{\link{gemini_ask}()},
\code{\link{gpt_ask}()},
\code{\link{queryGA}()}

Other Credentials: 
\code{\link{db_download}()},
\code{\link{db_upload}()},
\code{\link{encrypt_file}()},
\code{\link{get_credentials}()},
\code{\link{get_tweets}()},
\code{\link{mail_send}()},
\code{\link{queryDB}()},
\code{\link{queryGA}()},
\code{\link{stocks_file}()}
}
\concept{API}
\concept{Credentials}
